Package: inlabru
Type: Package
Title: Bayesian Latent Gaussian Modelling using INLA and Extensions
Version: 2.14.1
Authors@R: c(
  person("Finn", "Lindgren", email = "finn.lindgren@gmail.com",
         role = c("aut", "cre", "cph"),
         comment = c(ORCID = "0000-0002-5833-2011",
                     "Finn Lindgren continued development of the main code")),
  person("Fabian E.", "Bachl", email = "bachlfab@gmail.com",
         role = c("aut", "cph"),
         comment = "Fabian Bachl wrote the main code"),
  person("David L.", "Borchers", email = "dlb@st-andrews.ac.uk",
         role = c("ctb", "dtc", "cph"),
         comment = "David Borchers wrote code for Gorilla data import and sampling, multiplot tool"),
  person("Daniel", "Simpson", email = "dp.simpson@gmail.com",
         role = c("ctb", "cph"),
         comment = "Daniel Simpson wrote the basic LGCP sampling method"),
  person("Lindesay", "Scott-Howard", email = "lass@st-andrews.ac.uk",
         role = c("ctb", "dtc", "cph"),
         comment = "Lindesay Scott-Howard provided MRSea data import code"),
  person("Seaton", "Andy", email = "andy.e.seaton@gmail.com", role = c("ctb"),
         comment = "Andy Seaton provided testing, bugfixes, and vignettes"),
  person("Suen", "Man Ho", email = "m.suen19@imperial.ac.uk", role = c("ctb", "cph"),
         comment = c(ORCID = "0009-0003-2281-0776",
                     "Man Ho Suen contributed features for aggregated responses and vignette updates")),
  person("Roudier", "Pierre", email = "", role = c("ctb", "cph"),
         comment = "Pierre Roudier contributed general quantile summaries"),
  person("Meehan", "Tim", email = "tmeeha@gmail.com", role = c("ctb", "cph"),
         comment = "Tim Meehan contributed the SVC vignette and robins data"),
  person("Reddy Peddinenikalva", "Niharika", role = c("ctb", "cph"),
         comment = "Niharika Peddinenikalva contributed the LGCP residuals vignette"),
  person("Perepolkin", "Dmytro", email = "dperepolkin@gmail.com",
         role = c("ctb", "cph"),
         comment = "Dmytro Perepolkin contributed the ZIP/ZAP vignette")
  )
URL: http://www.inlabru.org, https://inlabru-org.github.io/inlabru/,
        https://github.com/inlabru-org/inlabru
BugReports: https://github.com/inlabru-org/inlabru/issues
Description: Facilitates spatial and general latent Gaussian modeling using
  integrated nested Laplace approximation via the INLA package (<https://www.r-inla.org>).
  Additionally, extends the GAM-like model class to more general nonlinear predictor
  expressions, and implements a log Gaussian Cox process likelihood for 
  modeling univariate and spatial point processes based on ecological survey data.
  Model components are specified with general inputs and mapping methods to the
  latent variables, and the predictors are specified via general R expressions,
  with separate expressions for each observation likelihood model in
  multi-likelihood models. A prediction method based on fast Monte Carlo sampling
  allows posterior prediction of general expressions of the latent variables.
  Ecology-focused introduction in Bachl, Lindgren, Borchers, and Illian (2019)
  <doi:10.1111/2041-210X.13168>.
License: GPL (>= 2)
Additional_repositories: https://inla.r-inla-download.org/R/testing
RoxygenNote: 7.3.3
Encoding: UTF-8
Depends: methods, R (>= 4.1.0), stats
Imports: dplyr, fmesher (>= 0.7.0), glue, lifecycle, MatrixModels,
        Matrix, plyr, rlang, sf, tibble, utils, withr, Rcpp
Suggests: covr, ggplot2, graphics, INLA (>= 23.01.31), knitr, maps,
        mgcv, patchwork, raster, RColorBrewer, rgl, rmarkdown, scales,
        scoringRules, shiny, sn, sp (>= 2.1), spatstat.geom,
        spatstat.data, sphereplot, splancs, terra (>= 1.7-66),
        tidyterra, testthat (>= 3.2.0), tidyr, DiagrammeR
Enhances: stars
Config/testthat/parallel: true
Config/testthat/edition: 3
Collate: '0_inlabru_envir.R' 'RcppExports.R' 'access_trace.R'
        'access_trace_methods.R' 'bru.gof.R' 'bru.inference.R'
        'bru_conversion.R' 'bru_index.R' 'deprecated.R' 'bru_input.R'
        'bru_is.R' 'bru_sp.R' 'data.Poisson1_1D.R' 'data.Poisson2_1D.R'
        'data.Poisson3_1D.R' 'data.gorillas.R' 'data.mexdolphin.R'
        'data.mrsea.R' 'data.robins_subset.R' 'data.shrimp.R'
        'data.toygroups.R' 'data.toypoints.R' 'deltaIC.R' 'effect.R'
        'environment.R' 'fmesher.R' 'ggplot.R' 'hexagon_tiling.R'
        'mappers.R' 'hierarchical_basis.R' 'inla.R' 'inlabru-package.R'
        'local_testthat.R' 'mapper_collect.R' 'mapper_repeat.R'
        'mapper_sum.R' 'model.R' 'nlinla.R' 'object_upgrade.R'
        'plotsample.R' 'pred_expr.R' 'rgl.R' 'sampling.R' 'spatstat.R'
        'spde.R' 'stack.R' 'track_plotting.R' 'transformation.R'
        'used.R' 'utils.R'
VignetteBuilder: knitr
BuildVignettes: true
LazyData: true
LazyDataCompression: xz
LinkingTo: Rcpp
NeedsCompilation: yes
Packaged: 2026-04-30 12:27:54 UTC; flindgre
Author: Finn Lindgren [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-5833-2011>, Finn Lindgren continued
    development of the main code),
  Fabian E. Bachl [aut, cph] (Fabian Bachl wrote the main code),
  David L. Borchers [ctb, dtc, cph] (David Borchers wrote code for
    Gorilla data import and sampling, multiplot tool),
  Daniel Simpson [ctb, cph] (Daniel Simpson wrote the basic LGCP sampling
    method),
  Lindesay Scott-Howard [ctb, dtc, cph] (Lindesay Scott-Howard provided
    MRSea data import code),
  Seaton Andy [ctb] (Andy Seaton provided testing, bugfixes, and
    vignettes),
  Suen Man Ho [ctb, cph] (ORCID: <https://orcid.org/0009-0003-2281-0776>,
    Man Ho Suen contributed features for aggregated responses and
    vignette updates),
  Roudier Pierre [ctb, cph] (Pierre Roudier contributed general quantile
    summaries),
  Meehan Tim [ctb, cph] (Tim Meehan contributed the SVC vignette and
    robins data),
  Reddy Peddinenikalva Niharika [ctb, cph] (Niharika Peddinenikalva
    contributed the LGCP residuals vignette),
  Perepolkin Dmytro [ctb, cph] (Dmytro Perepolkin contributed the ZIP/ZAP
    vignette)
Maintainer: Finn Lindgren <finn.lindgren@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-01 05:10:22 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-01 21:34:04 UTC; unix
Archs: inlabru.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",bm_collect)
S3method("[",bm_list)
S3method("[",bm_multi)
S3method("[",bm_sum)
S3method("[",bru_comp_list)
S3method("[",bru_log)
S3method("[",bru_mapper_collect)
S3method("[",bru_mapper_multi)
S3method("[",bru_mapper_sum)
S3method("[",bru_obs_list)
S3method("ibm_names<-",bm_collect)
S3method("ibm_names<-",bm_multi)
S3method("ibm_names<-",bm_sum)
S3method("ibm_names<-",bru_mapper_collect)
S3method("ibm_names<-",bru_mapper_multi)
S3method("ibm_names<-",bru_mapper_sum)
S3method(add_mappers,bru_comp)
S3method(add_mappers,bru_comp_list)
S3method(as.character,bru_log)
S3method(as_bm_list,bm_list)
S3method(as_bm_list,bru_comp_list)
S3method(as_bm_list,list)
S3method(as_bru_comp,bru_comp)
S3method(as_bru_comp_list,bru)
S3method(as_bru_comp_list,bru_comp)
S3method(as_bru_comp_list,bru_comp_list)
S3method(as_bru_comp_list,bru_info)
S3method(as_bru_comp_list,bru_model)
S3method(as_bru_comp_list,formula)
S3method(as_bru_comp_list,list)
S3method(as_bru_info,bru)
S3method(as_bru_info,bru_info)
S3method(as_bru_mapper,bru_comp)
S3method(as_bru_mapper,bru_mapper)
S3method(as_bru_mapper,bru_subcomp)
S3method(as_bru_obs,bru_obs)
S3method(as_bru_obs_list,bru)
S3method(as_bru_obs_list,bru_info)
S3method(as_bru_obs_list,bru_model)
S3method(as_bru_obs_list,bru_obs)
S3method(as_bru_obs_list,bru_obs_list)
S3method(as_bru_obs_list,list)
S3method(bru_comp,character)
S3method(bru_comp_list,bru_comp)
S3method(bru_comp_list,bru_comp_list)
S3method(bru_comp_list,formula)
S3method(bru_comp_list,list)
S3method(bru_compute_linearisation,bru_comp)
S3method(bru_compute_linearisation,bru_model)
S3method(bru_compute_linearisation,bru_obs)
S3method(bru_compute_linearisation,bru_obs_list)
S3method(bru_get_mapper,inla.cgeneric)
S3method(bru_get_mapper,inla.rgeneric)
S3method(bru_get_mapper,inla.spde)
S3method(bru_get_mapper,inla_model_reparam)
S3method(bru_index,bru)
S3method(bru_index,bru_comp)
S3method(bru_index,bru_comp_list)
S3method(bru_index,bru_info)
S3method(bru_index,bru_model)
S3method(bru_index,bru_obs)
S3method(bru_index,bru_obs_list)
S3method(bru_info,bru)
S3method(bru_info,character)
S3method(bru_inla_formula,bru)
S3method(bru_inla_formula,bru_comp_list)
S3method(bru_inla_formula,bru_info)
S3method(bru_input,bm_collect)
S3method(bru_input,bm_multi)
S3method(bru_input,bm_pipe)
S3method(bru_input,bm_repeat)
S3method(bru_input,bm_sum)
S3method(bru_input,bru_comp)
S3method(bru_input,bru_comp_list)
S3method(bru_input,bru_input)
S3method(bru_input,bru_mapper)
S3method(bru_input,bru_model)
S3method(bru_input,bru_obs)
S3method(bru_input,bru_obs_list)
S3method(bru_input_text,bm_collect)
S3method(bru_input_text,bm_multi)
S3method(bru_input_text,bm_pipe)
S3method(bru_input_text,bm_repeat)
S3method(bru_input_text,bm_sum)
S3method(bru_input_text,bru_comp)
S3method(bru_input_text,bru_comp_list)
S3method(bru_input_text,bru_input)
S3method(bru_input_text,bru_mapper)
S3method(bru_is_additive,bru_obs)
S3method(bru_is_additive,bru_obs_list)
S3method(bru_is_additive,bru_pred_expr)
S3method(bru_is_additive,character)
S3method(bru_is_additive,expression)
S3method(bru_is_additive,formula)
S3method(bru_is_linear,bru)
S3method(bru_is_linear,bru_comp)
S3method(bru_is_linear,bru_comp_list)
S3method(bru_is_linear,bru_info)
S3method(bru_is_linear,bru_mapper)
S3method(bru_is_linear,bru_model)
S3method(bru_is_linear,bru_obs)
S3method(bru_is_linear,bru_obs_list)
S3method(bru_is_linear,bru_pred_expr)
S3method(bru_is_rowwise,bru_comp)
S3method(bru_is_rowwise,bru_comp_list)
S3method(bru_is_rowwise,bru_mapper)
S3method(bru_is_rowwise,bru_obs)
S3method(bru_is_rowwise,bru_obs_list)
S3method(bru_is_rowwise,bru_pred_expr)
S3method(bru_log,bru)
S3method(bru_log,bru_log)
S3method(bru_log,character)
S3method(bru_log,iinla)
S3method(bru_make_stack,bru_obs)
S3method(bru_make_stack,bru_obs_list)
S3method(bru_mapper,fm_mesh_1d)
S3method(bru_mapper,fm_mesh_2d)
S3method(bru_names,bru)
S3method(bru_names,inla)
S3method(bru_obs_control_family,bru_obs)
S3method(bru_obs_control_family,bru_obs_list)
S3method(bru_obs_control_gcpo,bru_obs)
S3method(bru_obs_control_gcpo,bru_obs_list)
S3method(bru_obs_inla_family,bru_obs)
S3method(bru_obs_inla_family,bru_obs_list)
S3method(bru_obs_list,bru_obs)
S3method(bru_obs_list,bru_obs_list)
S3method(bru_obs_list,list)
S3method(bru_pred_expr,bru)
S3method(bru_pred_expr,bru_info)
S3method(bru_pred_expr,bru_obs)
S3method(bru_pred_expr,bru_obs_list)
S3method(bru_pred_expr,bru_pred_expr)
S3method(bru_response_size,bru)
S3method(bru_response_size,bru_info)
S3method(bru_response_size,bru_obs)
S3method(bru_response_size,bru_obs_list)
S3method(bru_response_size,default)
S3method(bru_response_size,inla.surv)
S3method(bru_response_size,list)
S3method(bru_set_missing,bru)
S3method(bru_set_missing,bru_info)
S3method(bru_set_missing,bru_obs)
S3method(bru_set_missing,bru_obs_list)
S3method(bru_set_missing,data.frame)
S3method(bru_set_missing,default)
S3method(bru_set_missing,inla.surv)
S3method(bru_timings,bru)
S3method(bru_used,"NULL")
S3method(bru_used,bru)
S3method(bru_used,bru_info)
S3method(bru_used,bru_obs)
S3method(bru_used,bru_pred_expr)
S3method(bru_used,bru_used)
S3method(bru_used,default)
S3method(bru_used,list)
S3method(bru_used_update,bru_obs)
S3method(bru_used_update,bru_obs_list)
S3method(bru_used_update,bru_pred_expr)
S3method(bru_used_update,bru_used)
S3method(bru_used_vars,character)
S3method(bru_used_vars,expression)
S3method(bru_used_vars,formula)
S3method(bru_used_vars,quosure)
S3method(c,bm_list)
S3method(c,bru_comp)
S3method(c,bru_comp_list)
S3method(c,bru_log)
S3method(c,bru_mapper)
S3method(c,bru_obs)
S3method(c,bru_obs_list)
S3method(eval_spatial,SpatRaster)
S3method(eval_spatial,SpatialGridDataFrame)
S3method(eval_spatial,SpatialPixelsDataFrame)
S3method(eval_spatial,SpatialPolygonsDataFrame)
S3method(eval_spatial,sf)
S3method(eval_spatial,stars)
S3method(evaluate_effect_single_state,bm_list)
S3method(evaluate_effect_single_state,bru_comp_list)
S3method(evaluate_effect_single_state,bru_mapper)
S3method(format,bm_collect)
S3method(format,bm_list)
S3method(format,bm_multi)
S3method(format,bm_pipe)
S3method(format,bm_reparam)
S3method(format,bm_repeat)
S3method(format,bm_sum)
S3method(format,bru_input)
S3method(format,bru_log)
S3method(format,bru_mapper)
S3method(format,bru_pred_expr)
S3method(format,bru_subcomp)
S3method(format,bru_used)
S3method(generate,bru)
S3method(gg,RasterLayer)
S3method(gg,SpatRaster)
S3method(gg,SpatialGridDataFrame)
S3method(gg,SpatialLines)
S3method(gg,SpatialPixels)
S3method(gg,SpatialPixelsDataFrame)
S3method(gg,SpatialPoints)
S3method(gg,SpatialPolygons)
S3method(gg,bru_prediction)
S3method(gg,data.frame)
S3method(gg,fm_mesh_1d)
S3method(gg,fm_mesh_2d)
S3method(gg,matrix)
S3method(gg,prediction)
S3method(gg,sf)
S3method(glplot,SpatialLines)
S3method(glplot,SpatialPoints)
S3method(glplot,fm_mesh_2d)
S3method(ibm_eval,bm_aggregate)
S3method(ibm_eval,bm_collect)
S3method(ibm_eval,bm_const)
S3method(ibm_eval,bm_logitaverage)
S3method(ibm_eval,bm_logsumexp)
S3method(ibm_eval,bm_marginal)
S3method(ibm_eval,bm_multi)
S3method(ibm_eval,bm_pipe)
S3method(ibm_eval,bm_reparam)
S3method(ibm_eval,bm_repeat)
S3method(ibm_eval,bm_scale)
S3method(ibm_eval,bm_shift)
S3method(ibm_eval,bm_sum)
S3method(ibm_eval,bm_taylor)
S3method(ibm_eval,default)
S3method(ibm_eval2,bm_pipe)
S3method(ibm_eval2,default)
S3method(ibm_inla_subset,default)
S3method(ibm_invalid_output,bm_collect)
S3method(ibm_invalid_output,bm_index)
S3method(ibm_invalid_output,bm_multi)
S3method(ibm_invalid_output,bm_repeat)
S3method(ibm_invalid_output,bm_sum)
S3method(ibm_invalid_output,default)
S3method(ibm_is_linear,bm_collect)
S3method(ibm_is_linear,bm_multi)
S3method(ibm_is_linear,bm_sum)
S3method(ibm_is_linear,default)
S3method(ibm_is_rowwise,default)
S3method(ibm_jacobian,bm_aggregate)
S3method(ibm_jacobian,bm_collect)
S3method(ibm_jacobian,bm_const)
S3method(ibm_jacobian,bm_factor)
S3method(ibm_jacobian,bm_fm_mesh_1d)
S3method(ibm_jacobian,bm_fm_mesh_2d)
S3method(ibm_jacobian,bm_fmesher)
S3method(ibm_jacobian,bm_harmonics)
S3method(ibm_jacobian,bm_index)
S3method(ibm_jacobian,bm_inla_mesh_1d)
S3method(ibm_jacobian,bm_inla_mesh_2d)
S3method(ibm_jacobian,bm_linear)
S3method(ibm_jacobian,bm_logitaverage)
S3method(ibm_jacobian,bm_logsumexp)
S3method(ibm_jacobian,bm_marginal)
S3method(ibm_jacobian,bm_matrix)
S3method(ibm_jacobian,bm_multi)
S3method(ibm_jacobian,bm_pipe)
S3method(ibm_jacobian,bm_reparam)
S3method(ibm_jacobian,bm_repeat)
S3method(ibm_jacobian,bm_scale)
S3method(ibm_jacobian,bm_shift)
S3method(ibm_jacobian,bm_sum)
S3method(ibm_jacobian,bm_taylor)
S3method(ibm_jacobian,default)
S3method(ibm_linear,bm_collect)
S3method(ibm_linear,bm_list)
S3method(ibm_linear,bm_multi)
S3method(ibm_linear,bm_repeat)
S3method(ibm_linear,bm_sum)
S3method(ibm_linear,bru_comp)
S3method(ibm_linear,bru_comp_list)
S3method(ibm_linear,bru_model)
S3method(ibm_linear,default)
S3method(ibm_n,bm_aggregate)
S3method(ibm_n,bm_collect)
S3method(ibm_n,bm_const)
S3method(ibm_n,bm_factor)
S3method(ibm_n,bm_fm_mesh_1d)
S3method(ibm_n,bm_fm_mesh_2d)
S3method(ibm_n,bm_fmesher)
S3method(ibm_n,bm_harmonics)
S3method(ibm_n,bm_inla_mesh_1d)
S3method(ibm_n,bm_inla_mesh_2d)
S3method(ibm_n,bm_linear)
S3method(ibm_n,bm_marginal)
S3method(ibm_n,bm_matrix)
S3method(ibm_n,bm_multi)
S3method(ibm_n,bm_pipe)
S3method(ibm_n,bm_reparam)
S3method(ibm_n,bm_repeat)
S3method(ibm_n,bm_scale)
S3method(ibm_n,bm_shift)
S3method(ibm_n,bm_sum)
S3method(ibm_n,bm_taylor)
S3method(ibm_n,default)
S3method(ibm_n_output,bm_aggregate)
S3method(ibm_n_output,bm_collect)
S3method(ibm_n_output,bm_marginal)
S3method(ibm_n_output,bm_multi)
S3method(ibm_n_output,bm_pipe)
S3method(ibm_n_output,bm_reparam)
S3method(ibm_n_output,bm_repeat)
S3method(ibm_n_output,bm_scale)
S3method(ibm_n_output,bm_shift)
S3method(ibm_n_output,bm_sum)
S3method(ibm_n_output,bm_taylor)
S3method(ibm_n_output,default)
S3method(ibm_names,bm_collect)
S3method(ibm_names,bm_multi)
S3method(ibm_names,bm_sum)
S3method(ibm_names,default)
S3method(ibm_simplify,bm_list)
S3method(ibm_simplify,bm_pipe)
S3method(ibm_simplify,bru_comp)
S3method(ibm_simplify,bru_comp_list)
S3method(ibm_simplify,bru_model)
S3method(ibm_simplify,default)
S3method(ibm_values,bm_aggregate)
S3method(ibm_values,bm_collect)
S3method(ibm_values,bm_const)
S3method(ibm_values,bm_factor)
S3method(ibm_values,bm_fm_mesh_1d)
S3method(ibm_values,bm_fm_mesh_2d)
S3method(ibm_values,bm_fmesher)
S3method(ibm_values,bm_inla_mesh_1d)
S3method(ibm_values,bm_inla_mesh_2d)
S3method(ibm_values,bm_linear)
S3method(ibm_values,bm_marginal)
S3method(ibm_values,bm_matrix)
S3method(ibm_values,bm_multi)
S3method(ibm_values,bm_pipe)
S3method(ibm_values,bm_reparam)
S3method(ibm_values,bm_repeat)
S3method(ibm_values,bm_scale)
S3method(ibm_values,bm_shift)
S3method(ibm_values,bm_sum)
S3method(ibm_values,bm_taylor)
S3method(ibm_values,default)
S3method(inla_subset_eval,bru_comp_list)
S3method(length,bru_log)
S3method(new_bru_used,"NULL")
S3method(new_bru_used,character)
S3method(new_bru_used,expression)
S3method(new_bru_used,formula)
S3method(new_bru_used,quosure)
S3method(plot,bru)
S3method(plot,bru_prediction)
S3method(plot,prediction)
S3method(predict,bru)
S3method(print,bm_list)
S3method(print,bru)
S3method(print,bru_comp)
S3method(print,bru_comp_list)
S3method(print,bru_info)
S3method(print,bru_input)
S3method(print,bru_log)
S3method(print,bru_mapper)
S3method(print,bru_model)
S3method(print,bru_obs)
S3method(print,bru_obs_list)
S3method(print,bru_pred_expr)
S3method(print,bru_subcomp)
S3method(print,bru_used)
S3method(print,summary_bru)
S3method(print,summary_bru_comp_list)
S3method(print,summary_bru_info)
S3method(print,summary_bru_input)
S3method(print,summary_bru_mapper)
S3method(print,summary_bru_model)
S3method(print,summary_bru_obs)
S3method(print,summary_bru_obs_list)
S3method(print,summary_bru_options)
S3method(print,summary_bru_subcomp)
S3method(print,summary_component)
S3method(summary,bru)
S3method(summary,bru_comp)
S3method(summary,bru_comp_list)
S3method(summary,bru_info)
S3method(summary,bru_input)
S3method(summary,bru_mapper)
S3method(summary,bru_model)
S3method(summary,bru_obs)
S3method(summary,bru_obs_list)
S3method(summary,bru_options)
S3method(summary,bru_subcomp)
S3method(summary,bru_used)
export("bru_comp_env<-")
export("bru_comp_env_extra<-")
export("bru_set_missing<-")
export("ibm_names<-")
export(add_mappers)
export(as.bru_options)
export(as_bm_list)
export(as_bru_comp)
export(as_bru_comp_list)
export(as_bru_info)
export(as_bru_mapper)
export(as_bru_obs)
export(as_bru_obs_list)
export(bincount)
export(bm_aggregate)
export(bm_autodetect)
export(bm_collect)
export(bm_const)
export(bm_factor)
export(bm_fmesher)
export(bm_harmonics)
export(bm_index)
export(bm_linear)
export(bm_logitaverage)
export(bm_logsumexp)
export(bm_marginal)
export(bm_matrix)
export(bm_mesh_B)
export(bm_multi)
export(bm_pipe)
export(bm_reparam)
export(bm_repeat)
export(bm_repeat_indexing)
export(bm_repeat_indexing_matrix)
export(bm_scale)
export(bm_shift)
export(bm_sum)
export(bm_taylor)
export(bru)
export(bru_call_options)
export(bru_comp)
export(bru_comp_env)
export(bru_comp_env_extra)
export(bru_comp_list)
export(bru_compute_linearisation)
export(bru_convergence_plot)
export(bru_eval_in_data_context)
export(bru_fill_missing)
export(bru_forward_transformation)
export(bru_get_mapper)
export(bru_get_mapper_safely)
export(bru_index)
export(bru_info)
export(bru_inla.stack.mexpand)
export(bru_inla.stack.mjoin)
export(bru_inla_formula)
export(bru_input)
export(bru_input_text)
export(bru_inverse_transformation)
export(bru_is_additive)
export(bru_is_linear)
export(bru_is_rowwise)
export(bru_like_list)
export(bru_log)
export(bru_log_abort)
export(bru_log_bookmark)
export(bru_log_bookmarks)
export(bru_log_index)
export(bru_log_message)
export(bru_log_new)
export(bru_log_offset)
export(bru_log_reset)
export(bru_log_warn)
export(bru_make_stack)
export(bru_mapper)
export(bru_mapper_aggregate)
export(bru_mapper_collect)
export(bru_mapper_const)
export(bru_mapper_define)
export(bru_mapper_factor)
export(bru_mapper_fmesher)
export(bru_mapper_harmonics)
export(bru_mapper_index)
export(bru_mapper_linear)
export(bru_mapper_logsumexp)
export(bru_mapper_marginal)
export(bru_mapper_matrix)
export(bru_mapper_mesh_B)
export(bru_mapper_multi)
export(bru_mapper_pipe)
export(bru_mapper_repeat)
export(bru_mapper_scale)
export(bru_mapper_shift)
export(bru_mapper_sum)
export(bru_mapper_taylor)
export(bru_model)
export(bru_names)
export(bru_obs)
export(bru_obs_control_family)
export(bru_obs_control_gcpo)
export(bru_obs_inla_family)
export(bru_obs_list)
export(bru_options)
export(bru_options_check)
export(bru_options_default)
export(bru_options_get)
export(bru_options_reset)
export(bru_options_set)
export(bru_options_set_local)
export(bru_pred_expr)
export(bru_rerun)
export(bru_response_size)
export(bru_safe_inla)
export(bru_safe_sp)
export(bru_safe_terra)
export(bru_set_missing)
export(bru_standardise_names)
export(bru_summarise)
export(bru_timings)
export(bru_timings_plot)
export(bru_used)
export(bru_used_update)
export(bru_used_vars)
export(cv_hex)
export(cv_partition)
export(deltaIC)
export(delta_IC)
export(devel.cvmeasure)
export(eval_spatial)
export(evaluate_effect_single_state)
export(evaluate_model)
export(evaluate_state)
export(fm_int)
export(fm_pixels)
export(generate)
export(gg)
export(globe)
export(glplot)
export(gmap)
export(gorillas_sf_gcov)
export(gorillas_sp)
export(ibm_eval)
export(ibm_eval2)
export(ibm_inla_subset)
export(ibm_input_available)
export(ibm_input_get)
export(ibm_input_new)
export(ibm_input_set)
export(ibm_invalid_output)
export(ibm_is_linear)
export(ibm_is_rowwise)
export(ibm_jacobian)
export(ibm_linear)
export(ibm_n)
export(ibm_n_output)
export(ibm_names)
export(ibm_simplify)
export(ibm_values)
export(iinla)
export(index_eval)
export(inla.spde2.pcmatern_B)
export(inla_subset_eval)
export(input_eval)
export(lgcp)
export(lgcp_IC)
export(like)
export(like_list)
export(local_basic_fixed_effect_testdata)
export(local_basic_intercept_testdata)
export(local_bru_options_set)
export(local_bru_safe_inla)
export(local_bru_testthat_assign)
export(local_bru_testthat_setup)
export(local_bru_testthat_tolerances)
export(make_hierarchical_mesh_basis)
export(materncov.bands)
export(mexdolphin_sp)
export(multiplot)
export(new_bru_input)
export(new_bru_pred_expr)
export(new_bru_used)
export(plotmarginal.inla)
export(plotsample)
export(point2count)
export(row_kron)
export(sample.lgcp)
export(sline)
export(spatial.to.ppp)
export(spde.posterior)
export(spoly)
import(fmesher)
import(methods)
import(stats)
importFrom(Matrix,diag)
importFrom(Rcpp,sourceCpp)
importFrom(fmesher,fm_int)
importFrom(fmesher,fm_pixels)
importFrom(glue,glue)
importFrom(glue,glue_collapse)
importFrom(glue,glue_data)
importFrom(graphics,hist)
importFrom(lifecycle,deprecated)
importFrom(rlang,.data)
importFrom(rlang,.env)
importFrom(utils,modifyList)
useDynLib(inlabru, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# inlabru

<!-- badges: start -->

[![CRAN
Status](http://www.r-pkg.org/badges/version-last-release/inlabru)](https://cran.r-project.org/package=inlabru)
[![inlabru status
badge](https://inlabru-org.r-universe.dev/badges/inlabru)](https://inlabru-org.r-universe.dev)
[![R build
status](https://github.com/inlabru-org/inlabru/workflows/R-CMD-check/badge.svg)](https://github.com/inlabru-org/inlabru/actions)
[![R code coverage
status](https://github.com/inlabru-org/inlabru/workflows/test-coverage/badge.svg)](https://github.com/inlabru-org/inlabru/actions)
[![lintr
status](https://github.com/inlabru-org/inlabru/workflows/lint/badge.svg)](https://github.com/inlabru-org/inlabru/actions)
[![Codecov test
coverage](https://codecov.io/gh/inlabru-org/inlabru/graph/badge.svg)](https://app.codecov.io/gh/inlabru-org/inlabru)
<!-- badges: end -->

The goal of [inlabru](http://inlabru.org) is to facilitate spatial
modeling using integrated nested Laplace approximation via the [R-INLA
package](https://www.r-inla.org). Additionally, extends the GAM-like
model class to more general nonlinear predictor expressions, and
implements a log Gaussian Cox process likelihood for modeling univariate
and spatial point processes based on ecological survey data. Model
components are specified with general inputs and mapping methods to the
latent variables, and the predictors are specified via general R
expressions, with separate expressions for each observation likelihood
model in multi-likelihood models. A prediction method based on fast
Monte Carlo sampling allows posterior prediction of general expressions
of the latent variables. See Fabian E. Bachl, Finn Lindgren, David L.
Borchers, and Janine B. Illian (2019), inlabru: an R package for
Bayesian spatial modelling from ecological survey data, Methods in
Ecology and Evolution, British Ecological Society, 10, 760–766,
[doi:10.1111/2041-210X.13168](https://doi.org/10.1111/2041-210X.13168),
and `citation("inlabru")`.

The [inlabru.org](http://inlabru.org) website has links to old tutorials
with code examples for versions up to 2.1.13. For later versions,
updated versions of these tutorials, as well as new examples, can be
found at <https://inlabru-org.github.io/inlabru/articles/>

## Installation

You can install the current [CRAN
version](https://cran.r-project.org/package=inlabru) version of inlabru:

``` r
options(repos = c(
  INLA = "https://inla.r-inla-download.org/R/testing",
  getOption("repos")
))
install.packages("inlabru")
```

### Installation using [pak](https://pak.r-lib.org/)

You can install the latest bugfix release of inlabru from
[GitHub](https://github.com/inlabru-org/inlabru) with:

``` r
# install.packages("pak")
pak::repo_add(INLA = "https://inla.r-inla-download.org/R/testing")
pak::pkg_install("inlabru-org/inlabru@stable")
```

You can install the development version of inlabru from
[GitHub](https://github.com/inlabru-org/inlabru) with

``` r
pak::pkg_install("inlabru-org/inlabru")
```

or track the development version builds via
[inlabru-org.r-universe.dev](https://inlabru-org.r-universe.dev/builds):

``` r
# Enable universe(s) by inlabru-org
pak::repo_add(inlabruorg = "https://inlabru-org.r-universe.dev")
pak::pkg_install("inlabru")
```

This will pick the r-universe version if it is more recent than the CRAN
version.

### Installation using `remotes`

You can install the latest bugfix release of inlabru from
[GitHub](https://github.com/inlabru-org/inlabru) with:

``` r
# install.packages("remotes")
remotes::install_github("inlabru-org/inlabru", ref = "stable")
```

You can install the development version of inlabru from
[GitHub](https://github.com/inlabru-org/inlabru) with

``` r
remotes::install_github("inlabru-org/inlabru", ref = "devel")
```

or track the development version builds via
[inlabru-org.r-universe.dev](https://inlabru-org.r-universe.dev/builds):

``` r
# Enable universe(s) by inlabru-org
options(repos = c(
  inlabruorg = "https://inlabru-org.r-universe.dev",
  INLA = "https://inla.r-inla-download.org/R/testing",
  CRAN = "https://cloud.r-project.org"
))

# Install some packages
install.packages("inlabru")
```

## Example

This is a basic example which shows how fit a simple spatial Log
Gaussian Cox Process (LGCP) and predicts its intensity:

``` r
# Load libraries
library(INLA)
#> Loading required package: Matrix
#> 
library(inlabru)
#> Loading required package: fmesher
library(fmesher)
library(ggplot2)

# Construct latent model components
matern <- inla.spde2.pcmatern(
  gorillas_sf$mesh,
  prior.sigma = c(0.1, 0.01),
  prior.range = c(0.01, 0.01)
)
cmp <- ~ mySmooth(geometry, model = matern) + Intercept(1)
# Fit LGCP model
# This particular bru/bru_obs combination has a shortcut function lgcp() as well
fit <- bru(
  cmp,
  bru_obs(
    formula = geometry ~ .,
    family = "cp",
    data = gorillas_sf$nests,
    samplers = gorillas_sf$boundary,
    domain = list(geometry = gorillas_sf$mesh)
  ),
  options = list(control.inla = list(int.strategy = "eb"))
)

# Predict Gorilla nest intensity
lambda <- predict(
  fit,
  fm_pixels(gorillas_sf$mesh, mask = gorillas_sf$boundary),
  ~ exp(mySmooth + Intercept)
)
```

``` r
# Plot the result
ggplot() +
  geom_fm(data = gorillas_sf$mesh) +
  gg(lambda, geom = "tile") +
  gg(gorillas_sf$nests, color = "red", size = 0.5, alpha = 0.5) +
  ggtitle("Nest intensity per km squared") +
  xlab("") +
  ylab("")
```

<div class="figure">

<img src="man/figures/README-plot-1.png" alt="Nest intensity per km squared" width="100%" />
<p class="caption">

Nest intensity per km squared
</p>

</div>

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
suppressPackageStartupMessages(library(inlabru))
suppressPackageStartupMessages(library(fmesher))
suppressPackageStartupMessages(library(tibble))

convert_fun_link <- function(x, webref) {
  pkg_name <- sub(
    pattern = "^(([^:]*)::)?([^(]+)(\\(.*\\))(<-)?$",
    replacement = "\\2",
    x = x
  )
  if (identical(pkg_name, "")) {
    pkg_name <- "inlabru"
  }
  nm <- sub(
    pattern = "^(([^:]*)::)?([^(]+)(\\(.*\\))(<-)?$",
    replacement = "\\3",
    x = x
  )
  args <- sub(
    pattern = "^(([^:]*)::)?([^(]+)(\\(.*\\))(<-)?$",
    replacement = "\\4\\5",
    x = x
  )
  help_name <- tryCatch(
    {
      basename(utils::help((nm), package = (pkg_name)))
    },
    error = function(e) {
      nm
    }
  )
  paste0(
    '<a href="', webref[[pkg_name]], help_name, '.html">Y</a>',
  )
}

convert_fun_links <- function(df) {
  webref <- list(
    inlabru = "https://inlabru-org.github.io/inlabru/reference/",
    fmesher = "https://inlabru-org.github.io/fmesher/reference/"
  )
  df$Fun <- vapply(
    df[["Fun"]],
    function(x) convert_fun_link(x, webref),
    ""
  )
  df
}

convert_NAs <- function(df) {
  for (k in colnames(df)) {
    df[[k]] <- vapply(
      df[[k]],
      function(x) if (is.na(x)) "" else paste0("`", x, "`"),
      ""
    )
  }
  df
}

## ----echo=FALSE---------------------------------------------------------------
special_classes <- c(
  "bru",
  "bru_comp",
  "bru_comp_list",
  "bru_obs",
  "bru_obs_list",
  "bru_predictor",
  "bru_model",
  "bru_info",
  "bru_mapper",
  "bm_list"
)

bm_classes <- names(rlang::pkg_env("inlabru"))
bm_classes <- bm_classes[grepl("^bm_", bm_classes)]
all_classes <- unique(c(special_classes, bm_classes))
# TODO: further filtering

names(all_classes) <- all_classes
methods_list <- c(
  #  "predict",
  #  "generate",
  "gg"
)
names(methods_list) <- methods_list

## ----echo=FALSE---------------------------------------------------------------
class_methods <- lapply(all_classes, function(x) {
  y <- unclass(methods(class = x))
  if (NROW(y) == 0) {
    NULL
  } else {
    y <- cbind(class = x, attr(y, "info"))
    rownames(y) <- NULL
    y
  }
})
method_classes <- lapply(methods_list, function(x) {
  y <- unclass(methods(generic.function = x))
  if (NROW(y) == 0) {
    NULL
  } else {
    y <- cbind(class = gsub(paste0("^", x, "\\."), "", y), attr(y, "info"))
    rownames(y) <- NULL
    y
  }
})

## ----echo=FALSE---------------------------------------------------------------
info <- dplyr::bind_rows(
  dplyr::bind_rows(class_methods),
  dplyr::bind_rows(method_classes)
) |> # dplyr::filter(from == "inlabru") |>
  dplyr::group_by(generic, class) |>
  dplyr::summarise(generic = generic[1], class = class[1], .groups = "drop") |>
  dplyr::select(Generic = generic, Class = class) |>
  dplyr::mutate(Support = "Y") |>
  dplyr::arrange(Class, Generic)

info_sep <- list(
  gg = info |>
    dplyr::filter(Generic == "gg"),
  bm = info |>
    dplyr::filter(Class %in% bm_classes),
  special = info |>
    dplyr::filter(Class %in% special_classes)
)

info_sep_wide <- lapply(
  info_sep,
  function(x) {
    x |>
      tidyr::pivot_wider(names_from = Class, values_from = Support) |>
      convert_NAs() |>
      dplyr::arrange(Generic)
  }
)

## ----echo = FALSE,results="asis"----------------------------------------------
x <- knitr::kable(
  info_sep_wide$special,
  caption = "General classes and methods in `inlabru`",
  #  col.names = c("Method", "Class"),
  row.names = FALSE,
  format = "html"
)

gsub('<th style="text-align:(left|center|right);"> ([^(>| )]+) </th>',
  '<th class="rotate"><div><span>\\2</span></div></th>',
  x = x
) |>
  cat(sep = "\n")

## ----echo = FALSE,results="asis"----------------------------------------------
x <- knitr::kable(
  info_sep_wide$bm,
  caption = "`bru_mapper` classes and methods in `inlabru`",
  #  col.names = c("Method", "Class"),
  row.names = FALSE,
  format = "html"
)

gsub('<th style="text-align:(left|center|right);"> ([^(>| )]+) </th>',
  '<th class="rotate"><div><span>\\2</span></div></th>',
  x = x
) |>
  cat(sep = "\n")

\donttest{
if (bru_safe_inla() &&
  require("ggplot2", quietly = TRUE) &&
  require("patchwork", quietly = TRUE) &&
  require("sn", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE) &&
  require("RColorBrewer", quietly = TRUE) &&
  require("dplyr", quietly = TRUE)) {
  # Load Gorilla data

  gorillas <- gorillas_sf
  gorillas$gcov <- gorillas_sf_gcov()

  # Use RColorBrewer

  # Fit a model with two components:
  # 1) A spatial smooth SPDE
  # 2) A spatial covariate effect (vegetation)

  pcmatern <- INLA::inla.spde2.pcmatern(
    gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(0.01, 0.01)
  )

  cmp <- geometry ~ 0 +
    vegetation(gorillas$gcov$vegetation, model = "factor_contrast") +
    spde(geometry, model = pcmatern) +
    Intercept(1)

  fit <- lgcp(
    cmp,
    gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(control.inla = list(int.strategy = "eb"))
  )

  # Predict SPDE and vegetation at a grid covering the domain of interest
  pred_loc <- fm_pixels(
    gorillas$mesh,
    mask = gorillas$boundary,
    dims = c(200, 200),
    format = "sf"
  )
  pred <- predict(
    fit,
    pred_loc,
    ~ list(
      joint = spde + vegetation,
      field = spde,
      veg = vegetation
    )
  )

  pred_collect <-
    rbind(
      pred$joint |> dplyr::mutate(component = "joint"),
      pred$field |> dplyr::mutate(component = "field"),
      pred$veg |> dplyr::mutate(component = "veg")
    ) |>
    dplyr::mutate(var = sd^2)

  # Plot component mean

  ggplot(pred_collect) +
    geom_tile(aes(geometry = geometry, fill = mean), stat = "sf_coordinates") +
    coord_equal() +
    facet_wrap(~component, nrow = 1) +
    theme(legend.position = "bottom")

  # Plot component variance

  ggplot(pred_collect) +
    geom_tile(aes(geometry = geometry, fill = var), stat = "sf_coordinates") +
    coord_equal() +
    facet_wrap(~component, nrow = 1) +
    theme(legend.position = "bottom")

  # Calculate variance and correlation measure

  vm <- devel.cvmeasure(pred$joint, pred$field, pred$veg)
  # Compute nominal relative variance contributions; note that these can be
  # greater than 100%!
  vm <- dplyr::mutate(
    vm,
    var1_rel = if_else(var1 <= 0, NA, var1 / var.joint),
    var2_rel = if_else(var2 <= 0, NA, var2 / var.joint)
  )
  lprange <- range(vm$var.joint, vm$var1, vm$var2)
  vm <- tidyr::pivot_longer(vm,
    cols = c(var.joint, var1, var2, cov, cor, var1_rel, var2_rel),
    names_to = "component",
    values_to = "value"
  )

  # Variance contribution of the components

  csc <- scale_fill_gradientn(
    colours = brewer.pal(9, "YlOrRd"),
    limits = lprange
  )

  vm_ <- dplyr::filter(vm, component %in% c("var.joint", "var1", "var2"))
  ggplot(vm_) +
    geom_tile(aes(geometry = geometry, fill = value),
      stat = "sf_coordinates"
    ) +
    csc +
    coord_equal() +
    facet_wrap(~component, nrow = 1) +
    theme(legend.position = "bottom")

  # Relative variance contribution of the components
  # When bo

  vm_ <- dplyr::filter(vm, component %in% c("var1_rel", "var2_rel"))
  ggplot(vm_) +
    geom_tile(aes(geometry = geometry, fill = value),
      stat = "sf_coordinates"
    ) +
    scale_fill_gradientn(
      colours = brewer.pal(9, "YlOrRd"),
      trans = "log"
    ) +
    coord_equal() +
    facet_wrap(~component, nrow = 1)

  # Where both relative contributions are larger than 1, the posterior
  # correlations are strongly negative.

  # Covariance and correlation of field and vegetation

  vm_cov <- dplyr::filter(vm, component %in% "cov")
  vm_cor <- dplyr::filter(vm, component %in% "cor")
  (ggplot(vm_cov) +
    geom_tile(aes(geometry = geometry, fill = value),
      stat = "sf_coordinates"
    ) +
    scale_fill_gradientn(
      colours = rev(brewer.pal(9, "RdBu")),
      limits = c(-1, 1) * max(abs(vm_cov$value), na.rm = TRUE)
    ) +
    coord_equal() +
    theme(legend.position = "bottom") +
    ggtitle("Covariances") |
    ggplot(vm_cor) +
      geom_tile(aes(geometry = geometry, fill = value),
        stat = "sf_coordinates"
      ) +
      scale_fill_gradientn(
        colours = rev(brewer.pal(9, "RdBu")),
        limits = c(-1, 1) * max(abs(vm_cor$value), na.rm = TRUE)
      ) +
      coord_equal() +
      theme(legend.position = "bottom") +
      ggtitle("Correlations")
  )

  # Variance and correlation integrated over space

  vrt <- fm_vertices(gorillas$mesh, format = "sf")
  pred_vrt <- predict(
    fit,
    vrt,
    ~ list(
      joint = spde + vegetation,
      field = spde,
      veg = vegetation
    )
  )

  vm.int <- devel.cvmeasure(
    pred_vrt$joint,
    pred_vrt$field,
    pred_vrt$veg,
    samplers = fm_int(gorillas$mesh, gorillas$boundary),
    mesh = gorillas$mesh
  )
  vm.int
}
}

\donttest{
if (bru_safe_inla() &&
    require("sn", quietly = TRUE) &&
    require("ggplot2", quietly = TRUE) &&
    require("patchwork", quietly = TRUE)) {
  # Generate some data

  input.df <- data.frame(x = cos(1:10))
  input.df <- within(
    input.df,
    y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
  )

  # Fit a model with fixed effect 'x' and intercept 'Intercept'

  fit <- bru(y ~ x, family = "gaussian", data = input.df)

  # Predict posterior statistics of 'x'

  xpost <- predict(fit, NULL, formula = ~x_latent)

  # The statistics include mean, standard deviation, the 2.5% quantile, the
  # median, the 97.5% quantile, minimum and maximum sample drawn from the
  # posterior as well as the coefficient of variation and the variance.

  xpost

  # For a single variable like 'x' the default plotting method invoked by gg()
  # will show these statistics in a fashion similar to a box plot:
  ggplot() +
    gg(xpost)


  # The predict function can also be used to simultaneously estimate posteriors
  # of multiple variables:

  xipost <- predict(fit,
    newdata = NULL,
    formula = ~ c(
      Intercept = Intercept_latent,
      x = x_latent
    )
  )
  xipost

  # If we still want a plot in the previous style we have to set the bar
  # parameter to TRUE

  p1 <- ggplot() +
    gg(xipost, bar = TRUE)
  p1

  # Note that gg also understands the posterior estimates generated while
  # running INLA

  p2 <- ggplot() +
    gg(fit$summary.fixed, bar = TRUE)
  (p1 / p2)

  # By default, if the prediction has more than one row, gg will plot the column
  # 'mean' against the row index. This is for instance useful for predicting and
  # plotting function but not very meaningful given the above example:

  ggplot() +
    gg(xipost)

  # For ease of use we can also type

  plot(xipost)

  # This type of plot will show a ribbon around the mean, which visualizes the
  # upper and lower quantiles mentioned above (2.5% and 97.5%). Plotting the
  # ribbon can be turned of using the \code{ribbon} parameter

  ggplot() +
    gg(xipost, ribbon = FALSE)

  # Much like the other geomes produced by gg we can adjust the plot using
  # ggplot2 style commands, for instance

  ggplot() +
    gg(xipost) +
    gg(xipost, mapping = aes(y = median), ribbon = FALSE, color = "red")
}
}

\donttest{
if (bru_safe_inla()) {
  # Simulate some covariates x and observations y
  input.df <- data.frame(x = cos(1:10))
  input.df <- within(input.df, {
    y <- 5 + 2 * x + rnorm(10, mean = 0, sd = 0.1)
  })

  # Fit a Gaussian likelihood model
  fit <- bru(y ~ x + Intercept(1), family = "gaussian", data = input.df)

  # Obtain summary
  fit$summary.fixed
}


if (bru_safe_inla()) {
  # Alternatively, we can use the bru_obs() function to construct the
  # likelihood:

  lik <- bru_obs(family = "gaussian",
              formula = y ~ x + Intercept,
              data = input.df)
  fit <- bru(~ x + Intercept(1), lik)
  fit$summary.fixed
}

# An important addition to the INLA methodology is bru's ability to use
# non-linear predictors. Such a predictor can be formulated via bru_obs()'s
# \code{formula} parameter. The z(1) notation is needed to ensure that
# the z component should be interpreted as single latent variable and not
# a covariate:

if (bru_safe_inla()) {
  z <- 2
  input.df <- within(input.df, {
    y <- 5 + exp(z) * x + rnorm(10, mean = 0, sd = 0.1)
  })
  lik <- bru_obs(
    family = "gaussian", data = input.df,
    formula = y ~ exp(z) * x + Intercept
  )
  fit <- bru(~ z(1) + Intercept(1), lik)

  # Check the result (z posterior should be around 2)
  fit$summary.fixed
}
}

\donttest{
if (bru_safe_inla() &&
    require(ggplot2, quietly = TRUE)) {

  # Generate some data
  input.df <- data.frame(x = cos(1:10))
  input.df <- within(
    input.df,
    y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
  )

  # Fit a model with fixed effect 'x' and intercept 'Intercept'

  fit <- bru(y ~ x, family = "gaussian", data = input.df)

  # Predict posterior statistics of 'x'

  xpost <- predict(fit, formula = ~x)

  # The result is a data.frame inheriting from class 'prediction'

  class(xpost)

  # The statistics include mean, standard deviation, the 2.5% quantile, the
  # median, the 97.5% quantile, minimum and maximum sample drawn from the
  # posterior as well as the coefficient of variation and the variance.

  xpost

  # For a single variable like 'x' the default plotting method invoked by gg()
  # will show these statistics in a fashion similar to a box plot:

  ggplot() +
    gg(xpost)


  # The predict function can also be used to simultaneously estimate posteriors
  # of multiple variables:

  xipost <- predict(fit, formula = ~ data.frame(post = c(x, Intercept)))
  xipost

  # If we still want a plot in the previous style we have to set the 'bar'
  # parameter to TRUE:

  rownames(xipost) <- c("x", "Intercept")
  ggplot() +
    gg(xipost, bar = TRUE)
}
}

## ----include = FALSE------------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(inlabru)
inla_available <- inlabru::bru_safe_inla(
  multicore = interactive(),
  quietly = TRUE
)

## ----setup, include=TRUE, eval=inla_available-----------------------------------------------------
library(dplyr)
library(scoringRules)

## ----eval=inla_available--------------------------------------------------------------------------
df_pois <- tibble::tibble(
  x = rnorm(50),
  y = rpois(length(x), exp(2 + 1 * x))
)
fit_pois <- bru(
  y ~ Intercept(1) + x,
  family = "poisson",
  data = df_pois
)
newdata_pois <- tibble::tibble(
  x = rnorm(100),
  y = rpois(100, exp(2 + 1 * x))
)

## ----eval=inla_available--------------------------------------------------------------------------
pred_pois <- predict(
  fit_pois,
  newdata_pois,
  formula = ~ exp(Intercept + x),
  n.samples = 2000
)
post_E <- pred_pois$mean
post_Var <- pred_pois$mean + pred_pois$sd^2
scores <- data.frame(
  SE = (newdata_pois$y - post_E)^2,
  DS = (newdata_pois$y - post_E)^2 / post_Var + log(post_Var)
)
summary(scores)

## ----eval=inla_available--------------------------------------------------------------------------
pred_pois <- predict(fit_pois,
  newdata_pois,
  formula = ~ dpois(y, lambda = exp(Intercept + x)),
  n.samples = 2000
)
log_score <- -log(pred_pois$mean)
summary(log_score)

## ----eval=inla_available--------------------------------------------------------------------------
head(
  data.frame(
    log_S = log_score,
    lower = -log(pred_pois$mean + 2 * pred_pois$mean.mc_std_err),
    upper = -log(pred_pois$mean - 2 * pred_pois$mean.mc_std_err)
  )
)

## ----eval=inla_available--------------------------------------------------------------------------
# some large value, so that 1-F(K) is small
max_K <- ceiling(max(newdata_pois$y) + 4 * sqrt(max(newdata_pois$y)))
k <- seq(0, max_K)
kk <- rep(k, times = length(newdata_pois$y))
i <- seq_along(newdata_pois$y)
pred_pois <- generate(fit_pois, newdata_pois,
  formula = ~ {
    lambda <- exp(Intercept + x)
    ppois(kk, lambda = rep(lambda, each = length(k)))
  },
  n.samples = 2000
)
results <- data.frame(
  i = rep(i, each = length(k)),
  k = kk,
  Fpred = rowMeans(pred_pois),
  residuals =
    rowMeans(pred_pois) - (rep(newdata_pois$y, each = length(k)) <= kk)
)
# Check that the cutoff point K has nearly probability mass 1 below it,
# for all i:
min(results |> dplyr::filter(k == max_K) |> pull(Fpred))

crps_scores <-
  (results |>
    group_by(i) |>
    summarise(crps = sum(residuals^2), .groups = "drop") |>
    pull(crps))
summary(crps_scores)

## ----eval=inla_available--------------------------------------------------------------------------
true_sigma_y <- 0.1
df_normal <- tibble::tibble(
  x = rnorm(50),
  y = rnorm(length(x), 2 + 1 * x, sd = true_sigma_y)
)
fit_normal <- bru(
  y ~ Intercept(1) + x,
  family = "normal",
  data = df_normal
)
newdata_normal <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100, 2 + 1 * x, sd = true_sigma_y)
)

## ----eval=inla_available--------------------------------------------------------------------------
pred_normal <- predict(fit_normal, newdata_normal,
  formula = ~ {
    E <- Intercept + x
    V <- 1 / Precision_for_the_Gaussian_observations
    list(
      eta = E,
      sigma_y_2 = V,
      dens = dnorm(y, mean = E, sd = sqrt(V))
    )
  },
  n.samples = 2000
)
post_E <- pred_normal$eta$mean
post_Var <- pred_normal$sigma_y_2$mean + pred_normal$eta$sd^2
scores_normal <- data.frame(
  SE = (newdata_normal$y - post_E)^2,
  DS = (newdata_normal$y - post_E)^2 / post_Var + log(post_Var),
  LS = -log(pred_normal$dens$mean)
)
summary(scores_normal)

## ----eval=inla_available--------------------------------------------------------------------------
head(
  data.frame(
    log_S = scores_normal$LS,
    lower = -log(pred_normal$dens$mean + 2 * pred_normal$dens$mean.mc_std_err),
    upper = -log(pred_normal$dens$mean - 2 * pred_normal$dens$mean.mc_std_err)
  )
)

## ----eval=inla_available--------------------------------------------------------------------------
pred_pois <- predict(fit_pois, newdata_pois, formula = ~ exp(Intercept + x))
scores <- (newdata_pois$y - pred_pois$mean)^2

## ----eval=inla_available--------------------------------------------------------------------------
pred_pois <- predict(fit_pois, newdata_pois,
  formula = ~ {
    lambda <- exp(Intercept + x)
    list(
      cond_scores = (y - lambda)^2,
      lambda = lambda
    )
  },
  n.samples = 2000
)
scores <- pred_pois$cond_scores$mean - pred_pois$lambda$sd^2
summary(scores)

## ----eval=inla_available--------------------------------------------------------------------------
max_K <- 100 # some large value, so that 1-F(K) is small
pred_pois <- predict(fit_pois, newdata_pois,
  formula = ~ {
    lambda <- exp(Intercept + x)
    list(
      crps = scoringRules::crps_pois(y, lambda),
      F = do.call(
        c,
        lapply(
          seq_along(y),
          function(i) {
            ppois(seq(0, max_K), lambda = lambda[i])
          }
        )
      )
    )
  },
  n.samples = 2000
)
crps_score <-
  pred_pois$crps$mean -
  (pred_pois$F |>
    mutate(i = rep(seq_along(newdata_pois$y), each = max_K + 1)) |>
    group_by(i) |>
    summarise(F_var = sum(sd^2), .groups = "drop") |>
    pull(F_var))
summary(crps_score)

## ----include = FALSE------------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "png",
  dev.args = list(type = "cairo-png"),
  fig.width = 7,
  fig.height = 5
)

## ----setup, include = FALSE-----------------------------------------------------------------------
library(inlabru)
library(ggplot2)

## -------------------------------------------------------------------------------------------------
lambda <- function(u, gamma) {
  -pnorm(u, lower.tail = FALSE, log.p = TRUE) / gamma
}
lambda_inv <- function(lambda, gamma) {
  qnorm(-lambda * gamma, lower.tail = FALSE, log.p = TRUE)
}
D1lambda <- function(u, gamma) {
  exp(dnorm(u, log = TRUE) - pnorm(u, lower.tail = FALSE, log.p = TRUE)) / gamma
}
D2lambda <- function(u, gamma) {
  D1L <- D1lambda(u, gamma)
  -D1L * (u - gamma * D1L)
}
D1log_lambda <- function(u, gamma) {
  D1lambda(u, gamma) / lambda(u, gamma)
}
D2log_lambda <- function(u, gamma) {
  D1logL <- D1log_lambda(u, gamma)
  -D1logL * (u - gamma * D1lambda(u, gamma = gamma) + D1logL)
}

## -------------------------------------------------------------------------------------------------
log_p <- function(u, y, gamma) {
  L <- lambda(u, gamma)
  n <- length(y)
  dnorm(u, log = TRUE) - n * L + n * mean(y) * log(L) - sum(lgamma(y + 1))
}
D1log_p <- function(u, y, gamma) {
  n <- length(y)
  -u + n * D1log_lambda(u, gamma) * (mean(y) - lambda(u, gamma))
}
D2log_p <- function(u, y, gamma) {
  n <- length(y)
  -1 +
    n * D2log_lambda(u, gamma) * (mean(y) - lambda(u, gamma)) -
    n * D1log_lambda(u, gamma) * D1lambda(u, gamma)
}

## -------------------------------------------------------------------------------------------------
g <- 1
y <- c(0, 1, 2)
y <- c(0, 0, 0, 0, 0)
y <- rpois(5, 5)
mu_quad <- uniroot(
  D1log_p,
  lambda_inv((1 + sum(y)) / (g + length(y)) * c(1 / 10, 10), gamma = g),
  y = y, gamma = g
)$root
sd_quad <- (-D2log_p(mu_quad, y = y, gamma = g))^-0.5
sd_lin <- (1 + length(y) * D1log_lambda(mu_quad, gamma = g)^2 *
  lambda(mu_quad, gamma = g))^-0.5
lambda0 <- lambda(mu_quad, gamma = g)

## -------------------------------------------------------------------------------------------------
ggplot() +
  xlim(
    lambda(mu_quad - 4 * sd_quad, gamma = g),
    lambda(mu_quad + 4 * sd_quad, gamma = g)
  ) +
  labs(x = "lambda", y = "Density") +
  geom_function(
    fun = function(x) {
      exp(log_p(lambda_inv(x, gamma = g), y = y, gamma = g)) /
        D1lambda(lambda_inv(x, gamma = g), gamma = g) / (
          exp(log_p(lambda_inv(lambda0, gamma = g), y = y, gamma = g)) /
            D1lambda(lambda_inv(lambda0, gamma = g), gamma = g)
        ) *
        dgamma(lambda0, shape = 1 + sum(y), rate = g + length(y))
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = dgamma,
    args = list(shape = 1 + sum(y), rate = g + length(y)),
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      dnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_quad) /
        D1lambda(lambda_inv(x, gamma = g), gamma = g)
    },
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      dnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_lin) /
        D1lambda(lambda_inv(x, gamma = g), gamma = g)
    },
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(mapping = aes(
    xintercept = (1 + sum(y)) / (g + length(y)),
    lty = "Bayes mean"
  )) +
  geom_vline(mapping = aes(xintercept = lambda0, lty = "Bayes mode")) +
  geom_vline(mapping = aes(xintercept = mean(y), lty = "Plain mean"))

## -------------------------------------------------------------------------------------------------
ggplot() +
  xlim(
    lambda_inv(lambda0, gamma = g) - 4 * sd_quad,
    lambda_inv(lambda0, gamma = g) + 4 * sd_quad
  ) +
  xlab("u") +
  ylab("Density") +
  geom_function(
    fun = function(x) {
      exp(log_p(x, y = y, gamma = g) -
        log_p(lambda_inv(lambda0, gamma = g), y = y, gamma = g)) *
        (dgamma(lambda0, shape = 1 + sum(y), rate = g + length(y)) *
          D1lambda(lambda_inv(lambda0, gamma = g), gamma = g))
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      dgamma(lambda(x, gamma = g), shape = 1 + sum(y), rate = g + length(y)) *
        D1lambda(x, gamma = g)
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = dnorm,
    args = list(mean = mu_quad, sd = sd_quad),
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = dnorm,
    args = list(mean = mu_quad, sd = sd_lin),
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(
    mapping = aes(
      xintercept = lambda_inv((1 + sum(y)) / (g + length(y)),
        gamma = g
      ),
      lty = "Bayes mean"
    )
  ) +
  geom_vline(
    mapping = aes(
      xintercept = lambda_inv(lambda0, gamma = g),
      lty = "Bayes mode"
    )
  ) +
  geom_vline(
    mapping = aes(
      xintercept = lambda_inv(mean(y), gamma = g),
      lty = "Plain mean"
    )
  )

## -------------------------------------------------------------------------------------------------
ggplot() +
  xlim(
    lambda(mu_quad - 4 * sd_quad, gamma = g),
    lambda(mu_quad + 4 * sd_quad, gamma = g)
  ) +
  xlab("lambda") +
  ylab("CDF") +
  geom_function(
    fun = pgamma,
    args = list(shape = 1 + sum(y), rate = g + length(y)),
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      pnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_quad)
    },
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      pnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_lin)
    },
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(mapping = aes(
    xintercept = (1 + sum(y)) / (g + length(y)),
    lty = "Bayes mean"
  )) +
  geom_vline(mapping = aes(xintercept = lambda0, lty = "Bayes mode")) +
  geom_vline(mapping = aes(xintercept = mean(y), lty = "Plain mean"))

## -------------------------------------------------------------------------------------------------
ggplot() +
  xlim(
    lambda_inv(lambda0, gamma = g) - 4 * sd_quad,
    lambda_inv(lambda0, gamma = g) + 4 * sd_quad
  ) +
  xlab("u") +
  ylab("CDF") +
  geom_function(
    fun = function(x) {
      pgamma(lambda(x, gamma = g), shape = 1 + sum(y), rate = g + length(y))
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = pnorm,
    args = list(mean = mu_quad, sd = sd_quad),
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = pnorm,
    args = list(mean = mu_quad, sd = sd_lin),
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(mapping = aes(
    xintercept = lambda_inv((1 + sum(y)) / (g + length(y)),
      gamma = g
    ),
    lty = "Bayes mean"
  )) +
  geom_vline(mapping = aes(
    xintercept = lambda_inv(lambda0, gamma = g),
    lty = "Bayes mode"
  )) +
  geom_vline(mapping = aes(
    xintercept = lambda_inv(mean(y), gamma = g),
    lty = "Plain mean"
  ))

\donttest{
  if (require("ggplot2", quietly = TRUE) &&
      bru_safe_terra(quietly = TRUE) &&
      bru_safe_sp() &&
      require("sp")) {
    # Load Gorilla data

    gorillas <- inlabru::gorillas_sf

    gcov <- gorillas_sf_gcov()
    elev <- terra::as.data.frame(gcov$elevation, xy = TRUE)
    elev <- sf::as_Spatial(sf::st_as_sf(elev, coords = c("x", "y")))

    # Turn elevation covariate into SpatialGridDataFrame
    elev <- sp::SpatialPixelsDataFrame(elev, data = as.data.frame(elev))

    # Plot Gorilla elevation covariate provided as SpatialPixelsDataFrame.
    # The same syntax applies to SpatialGridDataFrame objects.

    ggplot() +
      gg(elev)

    # Add Gorilla survey boundary and nest sightings

    ggplot() +
      gg(elev) +
      gg(gorillas$boundary, alpha = 0.0, col = "red") +
      gg(gorillas$nests)

    # Load pantropical dolphin data

    mexdolphin <- inlabru::mexdolphin_sp()

    # Plot the pantropical survey boundary, ship transects and dolphin sightings

    ggplot() +
      gg(mexdolphin$ppoly) + # survey boundary as SpatialPolygon
      gg(mexdolphin$samplers) + # ship transects as SpatialLines
      gg(mexdolphin$points) # dolphin sightings as SpatialPoints

    # Change color

    ggplot() +
      gg(mexdolphin$ppoly, color = "green") + # survey boundary; SpatialPolygon
      gg(mexdolphin$samplers, color = "red") + # ship transects; SpatialLines
      gg(mexdolphin$points, color = "blue") # dolphin sightings; SpatialPoints


    # Visualize data annotations: line width by segment number

    names(mexdolphin$samplers) # 'seg' holds the segment number
    ggplot() +
      gg(mexdolphin$samplers, aes(color = seg))

    # Visualize data annotations: point size by dolphin group size

    names(mexdolphin$points) # 'size' holds the group size
    ggplot() +
      gg(mexdolphin$points, aes(size = size))
  }
}

# Niharika Reddy Peddinenikalva
# Vacation Scholarship project
# Updated for sf by Finn Lindgren 2024

suppressPackageStartupMessages(library("RColorBrewer"))
suppressPackageStartupMessages(library("ggplot2"))
suppressPackageStartupMessages(library("dplyr"))
suppressPackageStartupMessages(library("lwgeom"))
suppressPackageStartupMessages(library("patchwork"))
suppressPackageStartupMessages(library("terra"))
suppressPackageStartupMessages(library("sf"))
suppressPackageStartupMessages(library("INLA"))
suppressPackageStartupMessages(library("inlabru"))
suppressPackageStartupMessages(library("fmesher"))
theme_set(theme_bw())



#' ----------------------------------
#' prepare_residual_calculations
#' ----------------------------------
#'
#' Computes the A_sum, A_integrate matrices and the data frame used for
#' calculating the residuals for the given set of polygons B

#' Input:
#' @param samplers `sf` polygons containing partitions for which
#' residuals are to be calculated
#' @param domain A mesh object
#' @param observations `sf` points containing observed data
#'
#' Output:
#' @return A_sum - matrix used to compute the summation term of the residuals
#' @return A_integrate - matrix used to compute the integral term
#' @return df - `sf` containing all the locations 'u' for
#' calculating residuals
#'
prepare_residual_calculations <- function(samplers, domain, observations) {
  # Calculate the integration weights for A_integrate
  ips <- fm_int(domain = domain, samplers = samplers)

  # Set-up the A_integrate matrix
  # A_integrate has as many rows as polygons in the samplers,
  # as many columns as mesh points
  A_integrate <- inla.spde.make.A(
    mesh = domain, ips, weights = ips$weight,
    block = ips$.block, block.rescale = "none"
  )


  # Set-up the A_sum matrix
  # A_sum has as many rows as polygons in the samplers,
  # as many columns as observed points
  # each row has 1s for the points in the corresponding polygon
  idx <- sf::st_within(
    observations,
    samplers,
    sparse = TRUE
  )
  A_sum <- sparseMatrix(
    i = unlist(idx),
    j = rep(
      seq_len(nrow(observations)),
      lengths(idx)
    ),
    x = rep(1, length(unlist(idx))),
    dims = c(nrow(samplers), nrow(observations))
  )


  # Setting up the data frame for calculating residuals
  observations$obs <- TRUE
  df <- st_as_sf(
    data.frame(
      obs = rep(FALSE, domain$n),
      x = domain$loc[, 1],
      y = domain$loc[, 2]
    ),
    coords = c("x", "y"),
    crs = fm_crs(domain)
  )
  df <- dplyr::bind_rows(df, observations)

  # Return A-sum, A_integrate and the data frame for predicting the residuals
  list(A_sum = A_sum, A_integrate = A_integrate, df = df)
}



#' ------------
#' residual_df
#' ------------
#'
#' Computes the three types if residuals and returns a data frame containing
#' information about all 3 residuals for each partition of 'B'
#'
#' Inputs:
#' @param model fitted model for which residuals need to be calculated
#' @param df sf object containing all the locations 'u'
#' for calculating residuals
#' @param expr an expression object containing the formula of the model
#' @param A_sum matrix used to compute the summation term of the residuals
#' @param A_integrate matrix that computes the integral term of the residuals
#'
#' Outputs:
#' @return Data frame containing residual information for each of the
#' partitions of the subset 'B'
#'

residual_df <- function(model, df, expr, A_sum, A_integrate) {
  # Compute residuals
  res <- predict(
    object = model,
    newdata = df,
    ~ {
      lambda <- eval(expr)
      h1 <- lambda * 0 + 1
      h2 <- 1 / lambda
      h3 <- 1 / sqrt(lambda)
      data.frame(
        Scaling_Residuals =
          as.vector(A_sum %*% h1[obs]) -
            as.vector(A_integrate %*% (h1 * lambda)[!obs]),
        Inverse_Residuals =
          as.vector(A_sum %*% h2[obs]) -
            as.vector(A_integrate %*% (h2 * lambda)[!obs]),
        Pearson_Residuals =
          as.vector(A_sum %*% h3[obs]) -
            as.vector(A_integrate %*% (h3 * lambda)[!obs])
      )
    },
    used = bru_used(expr)
  )

  # Label the three types of residuals
  res$Scaling_Residuals$Type <- "Scaling Residuals"
  res$Inverse_Residuals$Type <- "Inverse Residuals"
  res$Pearson_Residuals$Type <- "Pearson Residuals"
  do.call(rbind, res)
}





#' --------------
#' set_csc
#' --------------
#'
#' Sets the colour scale for the three types of residuals
#'
#' Inputs:
#' @param residuals frame containing residual information for each of the
#' partitions of the subset 'B'
#' @param col_theme vector of themes for each type of residual
#'
#' Outputs:
#' @return a list of 3 colour scales for each type of residual
#'

set_csc <- function(residuals, col_theme) {
  # Store data for the colour scale of the plots for each type of residual
  cscrange <- data.frame(
    residuals |>
      group_by(Type) |>
      summarise(maxabs = max(abs(mean)))
  )

  # Set the colour scale for all three types of residuals
  scaling_csc <-
    scale_fill_gradientn(
      colours = brewer.pal(9, col_theme[1]),
      name = "Scaling Residual",
      limits =
        cscrange[cscrange$Type == "Scaling Residuals", 2] *
          c(-1, 1)
    )

  inverse_csc <-
    scale_fill_gradientn(
      colours = brewer.pal(9, col_theme[2]),
      name = "Inverse Residual",
      limits =
        cscrange[cscrange$Type == "Inverse Residuals", 2] *
          c(-1, 1)
    )

  pearson_csc <-
    scale_fill_gradientn(
      colours = brewer.pal(9, col_theme[3]),
      name = "Pearson Residual",
      limits =
        cscrange[cscrange$Type == "Pearson Residuals", 2] *
          c(-1, 1)
    )

  list("Scaling" = scaling_csc,
       "Inverse" = inverse_csc,
       "Pearson" = pearson_csc)
}



#' ---------------
#' residual_plot
#' ---------------
#'
#' plots the three types of residuals for each polygon
#'
#' Input:
#' @param samplers A sf containing partitions for which
#' residuals are to be calculated
#' @param residuals frame containing residual information for each of the
#' partitions of the subset 'B'
#' @param csc list of three colour scales for the three types of residuals
#' @param model_name string containing the name of the model being assessed
#'
#' Output:
#' @return a list of three subplots scaling, inverse and Pearson residuals
#' for the different partitions of samplers


residual_plot <- function(samplers, residuals, csc, model_name) {
  # Initialise the scaling residuals plot
  samplers$Residual <- residuals |>
    filter(Type == "Scaling Residuals") |>
    pull(mean)
  scaling <- ggplot() +
    gg(samplers, aes(fill = Residual), alpha = 1, colour = NA) +
    csc["Scaling"] +
    theme(legend.position = "bottom") +
    labs(subtitle = paste(model_name, "Scaling"))

  # Initialise the inverse residuals plot
  samplers$Residual <- residuals |>
    filter(Type == "Inverse Residuals") |>
    pull(mean)
  inverse <- ggplot() +
    gg(samplers, aes(fill = Residual), alpha = 1, colour = NA) +
    csc["Inverse"] +
    theme(legend.position = "bottom") +
    labs(subtitle = paste(model_name, "Inverse"))

  # Initialise the Pearson residuals plot
  samplers$Residual <- residuals |>
    filter(Type == "Pearson Residuals") |>
    pull(mean)
  pearson <- ggplot() +
    gg(samplers, aes(fill = Residual), alpha = 1, colour = NA) +
    csc["Pearson"] +
    theme(legend.position = "bottom") +
    labs(subtitle = paste(model_name, "Pearson"))

  # Return the three plots in a list
  list(
    Scaling = scaling, Inverse = inverse,
    Pearson = pearson
  )
}




#' ------------------
#' partition
#' ------------------
#'
#' Partitions the region based on the given criteria for calculating residuals
#' in each partition. Parts of this function are taken from concepts in
#' https://rpubs.com/huanfaChen/grid_from_polygon
#'
#' Input:
#' @param samplers A sf polygon containing region for which
#' partitions need to be created
#' @param resolution resolution of the grids that are required
#' @param nrows number of rows of grids that are required
#' @param ncols number of columns of grids that are required
#'
#' Output:
#' @return a partitioned sf with polygons as required
#'
#'
partition <- function(samplers, resolution = NULL, nrows = NULL, ncols = NULL) {
  # Create a grid for the given boundary
  if (is.null(resolution)) {
    grid <- terra::rast(
      terra::ext(samplers),
      crs = fm_proj4string(samplers),
      nrows = nrows, ncols = ncols
    )
  }

  if (is.null(c(nrows, ncols))) {
    grid <- terra::rast(
      terra::ext(samplers),
      crs = fm_proj4string(samplers),
      resolution = resolution
    )
  }

  gridPolygon <- terra::as.polygons(grid)

  # Extract the boundary with subpolygons only
  sf::st_as_sf(
    terra::intersect(
      gridPolygon,
      terra::vect(samplers)
    )
  )
}


#' ------------------
#' edit_df
#' ------------------
#'
#' Edits the residual data frames to remove columns that need not be displayed
#'
#' Input:
#' @param df the data frame which needs to be edited
#' @param columns a vector of columns that need to be deleted from df
#'
#' Output:
#' @return the edited data frame with only the desired columns
#'
#'
edit_df <- function(df, columns) {
  # Remove the columns that are not required
  df[, !(colnames(df) %in% columns), drop = FALSE]
}

\donttest{
if (require(fmesher, quietly = TRUE) &&
    require(ggplot2, quietly = TRUE)) {

  # Load Gorilla data
  gorillas <- inlabru::gorillas_sf

  # Plot mesh using default edge colors

  ggplot() +
    gg(gorillas$mesh)

  # Don't show interior and exterior boundaries

  ggplot() +
    gg(gorillas$mesh, interior = FALSE, exterior = FALSE)

  # Change the edge colors

  ggplot() +
    gg(gorillas$mesh,
      edge.color = "green",
      int.color = "black",
      ext.color = "blue"
    )

  # Use the x-coordinate of the vertices to colorize the triangles and
  # mask the plotted area by the survey boundary, i.e. only plot the inside

  xcoord <- gorillas$mesh$loc[, 1]
  ggplot() +
    gg(gorillas$mesh, color = (xcoord - 580), mask = gorillas$boundary) +
    gg(gorillas$boundary, alpha = 0)
}
}

\donttest{
if (interactive() &&
    require("rgl", quietly = TRUE) &&
    require("sphereplot", quietly = TRUE) &&
    bru_safe_sp() &&
    require("sp")) {
  # Show the globe
  globe()

  # Load pantropoical dolphin data
  mexdolphin <- inlabru::mexdolphin_sp()

  # Add mesh, ship transects and dolphin sightings stored
  # as fm_mesh_2d, SpatialLines and SpatialPoints objects, respectively

  glplot(mexdolphin$mesh, alpha = 0.2)
  glplot(mexdolphin$samplers, lwd = 5)
  glplot(mexdolphin$points, size = 10)
}
}

\donttest{
if (bru_safe_inla() && require(ggplot2, quietly = TRUE)) {

  # Load 1D Poisson process data

  data(Poisson2_1D, package = "inlabru")


  # Take a look at the point (and frequency) data

  ggplot(pts2) +
    geom_histogram(
      aes(x = x),
      binwidth = 55 / 20,
      boundary = 0,
      fill = NA,
      color = "black"
    ) +
    geom_point(aes(x), y = 0, pch = "|", cex = 4) +
    coord_fixed(ratio = 1)

  # Fit an LGCP model with  and SPDE component

  x <- seq(0, 55, length.out = 20)
  mesh1D <- fm_mesh_1d(x, boundary = "free")
  constrained_matern <- INLA::inla.spde2.matern(mesh1D, constr = TRUE)
  mdl <- x ~ spde1D(x, model = constrained_matern) + Intercept(1)
  fit <- lgcp(mdl, data = pts2, domain = list(x = mesh1D))

  # Calculate and plot the posterior range

  range <- spde.posterior(fit, "spde1D", "range")
  plot(range)

  # Calculate and plot the posterior log range

  lrange <- spde.posterior(fit, "spde1D", "log.range")
  plot(lrange)

  # Calculate and plot the posterior variance

  variance <- spde.posterior(fit, "spde1D", "variance")
  plot(variance)

  # Calculate and plot the posterior log variance

  lvariance <- spde.posterior(fit, "spde1D", "log.variance")
  plot(lvariance)

  # Calculate and plot the posterior Matern correlation

  matcor <- spde.posterior(fit, "spde1D", "matern.correlation")
  plot(matcor)

  # Calculate and plot the posterior Matern covariance

  matcov <- spde.posterior(fit, "spde1D", "matern.covariance")
  plot(matcov)
}
}

\donttest{
  if (require("ggplot2", quietly = TRUE) &&
      bru_safe_terra(quietly = TRUE) &&
      require("tidyterra", quietly = TRUE)) {
    # Load Gorilla data

    gorillas <- inlabru::gorillas_sf
    gorillas$gcov <- gorillas_sf_gcov()

    # Plot Gorilla elevation covariate provided as terra::rast.

    ggplot() +
      gg(gorillas$gcov$elevation)

    # Add Gorilla survey boundary and nest sightings

    ggplot() +
      gg(gorillas$gcov$elevation) +
      gg(gorillas$boundary, alpha = 0) +
      gg(gorillas$nests)

    # Load pantropical dolphin data

    mexdolphin <- inlabru::mexdolphin_sf

    # Plot the pantropical survey boundary, ship transects and dolphin sightings

    ggplot() +
      gg(mexdolphin$ppoly, alpha = 0.5) + # survey boundary
      gg(mexdolphin$samplers) + # ship transects
      gg(mexdolphin$points) # dolphin sightings

    # Change color

    ggplot() +
      gg(mexdolphin$ppoly, color = "green", alpha = 0.5) + # survey boundary
      gg(mexdolphin$samplers, color = "red") + # ship transects
      gg(mexdolphin$points, color = "blue") # dolphin sightings


    # Visualize data annotations: line width by segment number

    names(mexdolphin$samplers) # 'seg' holds the segment number
    ggplot() +
      gg(mexdolphin$samplers, aes(color = seg))

    # Visualize data annotations: point size by dolphin group size

    names(mexdolphin$points) # 'size' holds the group size
    ggplot() +
      gg(mexdolphin$points, aes(size = size))
  }
}

\donttest{
if (require(ggplot2, quietly = TRUE)) {
  # Load the data

  data("toygroups", package = "inlabru")

  # The data set is a simulation of animal groups residing in a 1D space. Their
  # locations in x-space are sampled from a Cox process with intensity

  ggplot(toygroups$df.intensity) +
    geom_line(aes(x = x, y = g.lambda))

  # Adding the simulated group locations to this plot we obtain

  ggplot(toygroups$df.intensity) +
    geom_line(aes(x = x, y = g.lambda)) +
    geom_point(data = toygroups$groups, aes(x, y = 0), pch = "|")

  # Each group has a size mark attached to it.
  # These group sizes are sampled from an exponential distribution
  # for which the rate parameter depends on the x-coordinate

  ggplot(toygroups$groups) +
    geom_point(aes(x = x, y = size))

  ggplot(toygroups$df.rate) +
    geom_line(aes(x, rate))
}
}

\donttest{
if (bru_safe_inla() &&
    require(ggplot2, quietly = TRUE) &&
    require(patchwork, quietly = TRUE)) {

  # The 'bru_obs()' (previously 'like()') function's main purpose is to set up
  # observation models, both for single- and multi-likelihood models.
  # The following example generates some random covariates which are observed
  # through two different random effect models with different likelihoods

  # Generate the data

  set.seed(123)

  n1 <- 200
  n2 <- 10

  x1 <- runif(n1)
  x2 <- runif(n2)
  z2 <- runif(n2)

  y1 <- rnorm(n1, mean = 2 * x1 + 3)
  y2 <- rpois(n2, lambda = exp(2 * x2 + z2 + 3))

  df1 <- data.frame(y = y1, x = x1)
  df2 <- data.frame(y = y2, x = x2, z = z2)

  # Single likelihood models and inference using bru are done via

  cmp1 <- y ~ -1 + Intercept(1) + x
  fit1 <- bru(cmp1, family = "gaussian", data = df1)
  summary(fit1)

  cmp2 <- y ~ -1 + Intercept(1) + x + z
  fit2 <- bru(cmp2, family = "poisson", data = df2)
  summary(fit2)

  # A joint model has two likelihoods, which are set up using the bru_obs
  # function

  lik1 <- bru_obs(
    "gaussian",
    formula = y ~ x + Intercept,
    data = df1,
    tag = "norm"
  )
  lik2 <- bru_obs(
    "poisson",
    formula = y ~ x + z + Intercept,
    data = df2,
    tag = "pois"
  )

  # The union of effects of both models gives the components needed to run bru

  jcmp <- ~ x + z + Intercept(1)
  jfit <- bru(jcmp, lik1, lik2)

  bru_index(jfit, "norm")
  bru_index(jfit, "pois")

  # Compare the estimates

  p1 <- ggplot() +
    gg(fit1$summary.fixed, bar = TRUE) +
    ylim(0, 4) +
    ggtitle("Model 1")
  p2 <- ggplot() +
    gg(fit2$summary.fixed, bar = TRUE) +
    ylim(0, 4) +
    ggtitle("Model 2")
  pj <- ggplot() +
    gg(jfit$summary.fixed, bar = TRUE) +
    ylim(0, 4) +
    ggtitle("Joint model")

  (p1 / p2 / pj)
}
}

\donttest{
if (bru_safe_inla() &&
    require("sn", quietly = TRUE) &&
    require("ggplot2", quietly = TRUE) &&
    bru_safe_terra(quietly = TRUE) &&
    require("sf", quietly = TRUE)) {

  # Load the Gorilla data

  gorillas <- gorillas_sf

  # Plot the Gorilla nests, the mesh and the survey boundary

  ggplot() +
    gg(gorillas$mesh) +
    gg(gorillas$nests) +
    gg(gorillas$boundary, alpha = 0.1)

  # Define SPDE prior

  matern <- INLA::inla.spde2.pcmatern(
    gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(0.01, 0.01)
  )

  # Define domain of the LGCP as well as the model components (spatial SPDE
  # effect and Intercept)

  cmp <- geometry ~ field(geometry, model = matern) + Intercept(1)

  # Fit the model, with "eb" instead of full Bayes
  fit <- lgcp(
    cmp,
    data = gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(control.inla = list(int.strategy = "eb"))
  )

  # Once we obtain a fitted model the predict function can serve various
  # purposes.
  # The most basic one is to determine posterior statistics of a univariate
  # random variable in the model, e.g. the intercept

  icpt <- predict(fit, NULL, ~ c(Intercept = Intercept_latent))
  plot(icpt)

  # The formula argument can take any expression that is valid within the model,
  # for instance a non-linear transformation of a random variable

  exp.icpt <- predict(fit, NULL, ~ c(
    "Intercept" = Intercept_latent,
    "exp(Intercept)" = exp(Intercept_latent)
  ))
  plot(exp.icpt, bar = TRUE)

  # The intercept is special in the sense that it does not depend on other
  # variables or covariates. However, this is not true for the smooth spatial
  # effects 'field'.
  # In order to predict 'field' we have to define where (in space) to predict.
  # For this purpose, the second argument of the predict function can take
  # \code{data.frame} objects as well as sf (and legacy sp/Spatial) objects. For
  # instance, we might want to predict 'field' at the locations of the mesh
  # vertices. Using

  vrt <- fm_vertices(gorillas$mesh, format = "sf")

  # we obtain these vertices as an sf object with POINT geometries

  ggplot() +
    gg(gorillas$mesh) +
    gg(vrt, color = "red")

  # Predicting 'field' at these locations works as follows

  field <- predict(fit, vrt, ~field)

  # Note that just like the input also the output will be a sf object with
  # points and that the predicted statistics are simply added as columns

  class(field)
  head(vrt)
  head(field)

  # Plotting the mean, for instance, at the mesh node is straight forward

  ggplot() +
    gg(gorillas$mesh) +
    gg(field, aes(color = mean), size = 2)

  # However, we are often interested in a spatial field and thus a linear
  # interpolation, which can be achieved by using the gg mechanism for meshes

  ggplot() +
    gg(gorillas$mesh, color = field$mean)

  # Alternatively, we can predict the spatial field at a grid of locations, e.g.
  # a sf object with a grid of points covering the relevant part of mesh

  pxl <- fm_pixels(gorillas$mesh, format = "sf", mask = gorillas$boundary)
  field2 <- predict(fit, pxl, ~field)

  # This will give us a sf with the columns we are looking for

  head(field2)
  ggplot() +
    gg(gorillas$boundary) +
    gg(data = field2, geom = "tile")
}
}

\donttest{
if (bru_safe_inla() &&
    require("sn", quietly = TRUE)) {

  # Generate data for a simple linear model

  input.df <- data.frame(x = cos(1:10))
  input.df <- within(
    input.df,
    y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
  )

  # Fit the model

  fit <- bru(y ~ xeff(main = x, model = "linear"),
    family = "gaussian", data = input.df
  )
  summary(fit)

  # Generate samples for some predefined x

  df <- data.frame(x = seq(-4, 4, by = 0.1))
  smp <- generate(fit, df, ~ xeff + Intercept, n.samples = 10)

  # Plot the resulting realizations

  plot(df$x, smp[, 1], type = "l")
  for (k in 2:ncol(smp)) points(df$x, smp[, k], type = "l")

  # We can also draw samples form the joint posterior

  df <- data.frame(x = 1)
  smp <- generate(fit, df, ~ data.frame(xeff, Intercept), n.samples = 10)
  smp[[1]]

  # ... and plot them
  if (require(ggplot2, quietly = TRUE)) {
    plot(do.call(rbind, smp))
  }
}
}

# GENERICS ----

#' @title Add component input/latent mappers
#' @description Add missing mappers between input data and latent variables,
#' based on likelihood data
#' @param \dots Parameters passed on to other methods
#' @export
#' @rdname add_mappers
add_mappers <- function(...) {
  UseMethod("add_mappers")
}


# CONSTRUCTORS ----

#' @title Latent model component construction
#'
#' @description
#' Similar to `glm()`, `gam()` and `inla()`, [bru()] models can be constructed
#' via a formula-like syntax, where each latent effect is specified. However, in
#' addition to the parts of the syntax compatible with `INLA::inla`, `bru`
#' components offer additional functionality which facilitates modelling, and
#' the predictor expression can be specified separately, allowing more complex
#' and non-linear predictors to be defined. The formula syntax is just a way to
#' allow all model components to be defined in a single line of code, but the
#' definitions can optionally be split up into separate component definitions.
#' See Details for more information.
#'
#' The `bru_comp` methods all rely on the [bru_comp.character()]
#' method, that defines a model component with a given label/name. The user
#' usually doesn't need to call these methods directly, but can instead supply a
#' formula expression that can be interpreted by the
#' [bru_comp_list.formula()] method, called inside [bru()].
#'
#' @details
#' As shorthand, [bru()] will understand basic additive formulae describing
#' fixed effect models. For instance, the components specification `y ~ x` will
#' define the linear combination of an effect named `x` and an intercept to the
#' response `y` with respect to the likelihood family stated when calling
#' [bru()]. Mathematically, the linear predictor \eqn{\eta}{eta} would be
#' written as
#'
#' \deqn{\eta = \beta x + c,}{eta = beta * x + c,}
#'
#' where:
#'
#' \describe{
#' \item{\eqn{c}{c}}{is the *intercept*}
#' \item{\eqn{x}{x}}{is a *covariate*}
#' \item{\eqn{\beta}{beta}}{is a *latent variable* associated with \eqn{x}{x}
#' and} \item{\eqn{\psi = \beta x}{psi = beta * x}}{ is called the *effect* of
#' \eqn{x}{x}}
#' }
#'
#' A problem that arises when using this kind of R formula is that it does not
#' clearly reflect the mathematical
#' formula. For instance, when providing the formula to inla, the resulting
#' object will refer to the random
#' effect \eqn{\psi = \beta * x } as `x`.
#' Hence, it is not clear when `x` refers to the covariate
#' or the effect of the covariate.
#'
#' @section Naming random effects:
#'
#' In INLA, the `f()` notation is used to define more complex models, but
#' a simple linear effect model can also be expressed as
#'
#' \itemize{\item{`formula = y ~ f(x, model = "linear")`,}}
#'
#' where `f()` is the inla specific function to set up random effects of all
#' kinds. The underlying predictor would again be \eqn{\eta = \beta * x + c} but
#' the result of fitting the model would state `x` as the random effect's name.
#' bru allows rewriting this formula in order to explicitly state the name of
#' the random effect and the name of the associated covariate. This is achieved
#' by replacing `f` with an arbitrary name that we wish to assign to the effect,
#' e.g.
#'
#' \itemize{\item{`components = y ~ psi(x, model = "linear")`.}}
#'
#' Being able to discriminate between \eqn{x} and \eqn{\psi} is relevant because
#' of two functionalities bru offers. The formula parameters of both [bru()] and
#' the prediction method [predict.bru] are interpreted in the mathematical
#' sense. For instance, `predict` may be used to analyze the analytical
#' combination of the covariate \eqn{x} and the intercept using
#'
#' \itemize{\item{`predict(fit, data.frame(x=2)), ~ exp(psi + Intercept)`.}}
#'
#' which corresponds to the mathematical expression \ifelse{html}{\out{e
#' <sup>&#946; + c</sup>}}{\eqn{e^{x \beta + c}}}.
#'
#' On the other hand, predict may be used to only look at a transformation of
#' the latent variable \eqn{\beta_\psi}
#'
#' \itemize{\item{`predict(fit, NULL, ~ exp(psi_latent))`.}}
#'
#' which corresponds to the mathematical expression \ifelse{html}{\out{e
#' <sup>&#946;</sup>}}{\eqn{e^{\beta}}}.
#'
#' @param \dots Parameters passed on to other methods
#'
#' @rdname bru_comp
#' @aliases component
#' @seealso [bru_input()], [summary.bru_comp()]
#'
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com} and
#' Finn Lindgren \email{Finn.Lindgren@@gmail.com}
#' @family component constructors
#' @export
#'
#' @examples
#' # As an example, let us create a linear component. Here, the component is
#' # called "myLinearEffectOfX" while the covariate the component acts on is
#' # called "x". Note that a list of components is returned because the
#' # formula may define multiple components
#'
#' cmp <- bru_comp_list(~ myLinearEffectOfX(main = x, model = "linear"))
#' summary(cmp)
#' # Equivalent shortcuts:
#' cmp <- bru_comp_list(~ myLinearEffectOfX(x, model = "linear"))
#' cmp <- bru_comp_list(~ myLinearEffectOfX(x))
#' # Individual component
#' cmp <- bru_comp("myLinearEffectOfX", main = x, model = "linear")
#' summary(cmp)
bru_comp <- function(...) {
  UseMethod("bru_comp")
}

#' @describeIn bru_comp
#' Backwards compatibility alias for `bru_comp()`
bru_component <- function(...) {
  bru_comp(...)
}

#' @export
#' @param object A character label for the component
#' @param main
#' `main` takes an R expression that evaluates to where the latent variables
#' should be evaluated (coordinates, indices, continuous scalar (for rw2 etc)).
#' Arguments starting with weights, group, replicate behave similarly to main,
#' but for the corresponding features of `INLA::f()`. `main` supports `tidy`
#' evaluation expressions, with `.data` and `.env` pronouns.
#' @param model Either one of "const" (same as "offset"), "factor_full",
#' "factor_contrast", "linear",
#' "fixed", or a model name or
#' object accepted by INLA's `f` function. If set to NULL, then "linear" is used
#' for vector inputs, and "fixed" for matrix input (converted internally to
#' an iid model with fixed precision)
#' @param mapper
#' Information about how to do the mapping from the values evaluated in `main`,
#' and to the latent variables. Auto-detects spde model objects in model and
#' extracts the mesh object to use as the mapper, and auto-generates mappers
#' for indexed models. (Default: NULL, for auto-determination)
#' @param main_layer,main_selector
#' The `_layer` input should evaluate to a numeric index or character name or
#' vector of which
#' layer/variable to extract from a covariate data object given in `main`.
#' (Default: NULL if `_selector` is given. Otherwise the effect component name,
#'  if it exists in the covariate object, and otherwise the first column of
#'  the covariate data frame)
#'
#' The `_selector` value should be a character name of a variable
#' whose contents determines which layer to extract from a covariate for each
#' data point. (Default: NULL)
#' @param n The number of latent variables in the model. Should be auto-detected
#' for most or all models. Default: NULL, for auto-detection. Models with
#' matrix input for `Cmatrix` or `graph` will use the matrix size.
#' An error is given if it realises it can't figure it out by itself.
#' @param values Specifies for what covariate/index values INLA should build
#' the latent model. Normally generated internally based on the mapping details.
#' (Default: NULL, for auto-determination)
#' @param season.length Passed on to `INLA::f()` for model `"seasonal"`
#' (TODO: check if this parameter is still fully handled)
#' @param nrow,ncol Number of rows and columns for `model` types
#'  "rw2d", "rw2diid", and "matern2d". Default is `NULL`.
# Copy feature
#' @param copy character; label of other component that this component should
#' be a copy of. If the `fixed = FALSE`, a scaling constant is estimated, via a
#' hyperparameter. If `fixed = TRUE`, the component scaling is fixed, by
#' default to 1; for fixed scaling, it's more efficient to express the scaling
#' in the predictor expression instead of making a copy component.
# Weights
#' @param weights,weights_layer,weights_selector
#' Optional specification of effect scaling weights.
#' Same syntax as for `main`.
# Group model parameters
#' @param group,group_mapper,group_layer,group_selector,ngroup
#' Optional specification of kronecker/group model indexing.
#' @param control.group `list` of kronecker/group model parameters, currently
#' passed directly on to `INLA::f`
# Replicate model parameters
#' @param replicate,replicate_mapper,replicate_layer,replicate_selector,nrep
#' Optional specification of indices for an independent
#' replication model. Same syntax as for `main`
#' @param marginal May specify a `bm_marginal()` mapper,
#' that is applied before scaling by `weights`.
#' @param A.msk `r lifecycle::badge("deprecated")` and has no effect.
#' @param .envir Evaluation environment. Can later be accessed via
#'   [bru_comp_env()]. Default: `parent.frame()`
#' @param envir_extra Environment for storing `INLA::f()` argument values.
#'   If NULL, an new environment is created. Can be accessed via
#'  [bru_comp_env_extra()]
#'
#' @details The `bru_comp.character` method is inlabru's equivalent to
#'   `INLA`'s `f()` function but adds functionality that is unique to inlabru.
#'
#'   The `main`, `weights`, `group`, `replicate`, and `*_layer` arguments
#'   support `tidy` evaluation expressions, with `.data` and `.env` pronouns.
#'
#' @rdname bru_comp
#' @aliases bru_comp
#' @aliases bru_component
#'
#' @examples
#' \donttest{
#' if (bru_safe_inla()) {
#'   # As an example, let us create a linear component. Here, the component is
#'   # called "myEffectOfX" while the covariate the component acts on is called
#'   # "x":
#'
#'   cmp <- bru_comp("myEffectOfX", main = x, model = "linear")
#'   summary(cmp)
#'
#'   # A more complicated component:
#'   cmp <- bru_comp("myEffectOfX",
#'     main = x,
#'     model = INLA::inla.spde2.matern(fmesher::fm_mesh_1d(1:10))
#'   )
#'
#'   # Compound fixed effect component, where x and z are in the input data.
#'   # The formula will be passed on to MatrixModels::model.Matrix:
#'   cmp <- bru_comp("eff", ~ -1 + x:z, model = "fixed")
#'   summary(cmp)
#' }
#' }
#'
bru_comp.character <- function(object,
                               # Main model parameters
                               main = NULL, # This must be kept as 1st arg.
                               weights = NULL, # This must be 2nd arg.
                               ..., # Prevent partial matching
                               model = NULL,
                               mapper = NULL,
                               main_layer = NULL,
                               main_selector = NULL,
                               n = NULL,
                               values = NULL,
                               season.length = NULL,
                               nrow = NULL,
                               ncol = NULL,
                               # Copy feature
                               copy = NULL,
                               # Weights
                               weights_layer = NULL,
                               weights_selector = NULL,
                               # Group model parameters
                               group = 1L,
                               group_mapper = NULL,
                               group_layer = NULL,
                               group_selector = NULL,
                               ngroup = NULL,
                               control.group = NULL,
                               # Replicate model parameters
                               replicate = 1L,
                               replicate_mapper = NULL,
                               replicate_layer = NULL,
                               replicate_selector = NULL,
                               nrep = NULL,
                               # Marginal transformation
                               marginal = NULL,
                               A.msk = deprecated(),
                               .envir = parent.frame(),
                               envir_extra = NULL) {
  # INLA models:
  # itypes = c(linear, iid, mec, meb, rgeneric, rw1, rw2, crw2, seasonal, besag,
  # besag2, bym, bym2, besagproper, besagproper2, fgn, fgn2, ar1, ar1c, ar, ou,
  # generic, generic0, generic1, generic2, generic3, spde, spde2, spde3, iid1d,
  # iid2d, iid3d, iid4d, iid5d, 2diid, z, rw2d, rw2diid, slm, matern2d, copy,
  # clinear, sigm, revsigm, log1exp, logdist)
  #
  # Supported:
  # btypes = c("const", "offset", "factor_full", "factor_contrast", "linear",
  # "clinear", "iid", "seasonal", "rw1", "rw2", "ar", "ar1", "ou", "spde")

  # The label
  label <- object

  if (is.null(model) && is.null(copy)) {
    # TODO: may need a special marker to autodetect factor and matrix variables,
    #       which can only be autodetected in a data-aware pass.
    model <- "linear"
  }

  # Force evaluation of explicit inputs
  force(values)

  include_weights <- !missing(weights) &&
    (!is.null(substitute(weights)))
  include_group <- !missing(group) ||
    !missing(ngroup) || !missing(group_mapper)
  include_repl <- !missing(replicate) ||
    !missing(nrep) || !missing(replicate_mapper)
  if (include_group) {
    if (!is.null(substitute(group)) &&
      !identical(deparse1(substitute(group)), "1L")) {
      if (is.null(control.group)) {
        control.group <- INLA::inla.set.control.group.default()
      }
      group_model <- control.group$model
    } else {
      group_model <- "exchangeable"
    }
  }

  if (is.null(envir_extra)) {
    envir_extra <- new.env(parent = .envir)
  }

  if (is.null(n)) {
    arg_names <- names(list(...))
    if ("Z" %in% arg_names) {
      Zmat <- list(...)[["Z"]]
      if (is.matrix(Zmat) || inherits(Zmat, "Matrix")) {
        n <- nrow(Zmat) + ncol(Zmat)
      }
    } else if ("Cmatrix" %in% arg_names) {
      Cmat <- list(...)[["Cmatrix"]]
      if (is.matrix(Cmat) || inherits(Cmat, "Matrix")) {
        n <- nrow(Cmat)
      }
    } else if ("graph" %in% arg_names) {
      n <- INLA::inla.read.graph(list(...)[["graph"]], size.only = TRUE)
    }
  }

  # Convert weights, ngroup, and nrep to bru_mapper info
  if (include_group) {
    if (!is.null(ngroup)) {
      if (!is.null(group_mapper)) {
        stop("At most one of 'ngroup' and 'group_mapper' should be supplied.")
      }
      group_mapper <- bm_index(ngroup)
      ngroup <- NULL
    }
  }
  if (include_repl) {
    if (!is.null(nrep)) {
      if (!is.null(replicate_mapper)) {
        stop("At most one of 'nrep' and 'replicate_mapper' should be supplied.")
      }
      replicate_mapper <- bm_index(nrep)
      nrep <- NULL
    }
  }

  inp <- new_bru_input(
    {{ main }},
    label = label,
    layer = {{ main_layer }},
    selector = main_selector
  )
  subcomp <- list(
    main = bru_subcomp(
      input = inp,
      mapper = mapper,
      n = n,
      model = model,
      values = values,
      season.length = season.length,
      nrow = nrow,
      ncol = ncol
    )
  )
  if (include_group) {
    subcomp$group <-
      bru_subcomp(
        input = new_bru_input(
          {{ group }},
          label = glue("{label}.group"),
          layer = {{ group_layer }},
          selector = group_selector
        ),
        mapper = group_mapper,
        n = NULL,
        model = group_model
      )
  }
  if (include_repl) {
    subcomp$replicate <-
      bru_subcomp(
        input = new_bru_input(
          {{ replicate }},
          label = glue("{label}.repl"),
          layer = {{ replicate_layer }},
          selector = replicate_selector
        ),
        mapper = replicate_mapper,
        n = NULL,
        model = "iid"
      )
  }
  if (include_weights) {
    weights_input <- new_bru_input(
      {{ weights }},
      label = glue("{label}.weights"),
      layer = {{ weights_layer }},
      selector = weights_selector
    )
    subcomp$weights <- ibm_input_set(bm_scale(), weights_input)
  }

  # Default component (to be filled)
  component <- c(
    list(
      label = label,
      inla.formula = NULL
    ),
    subcomp,
    list(
      copy = copy,
      marginal =
        if (is.null(marginal)) {
          NULL
        } else if (ibm_input_available(marginal)) {
          marginal
        } else {
          ibm_input_new(
            marginal,
            NULL,
            label = glue("{label}.marginal"),
            layer = NULL,
            selector = NULL
          )
        },
      env = .envir,
      env_extra = envir_extra
    )
  )

  # Main bit
  # Construct a call to the f function from the parameters provided
  # Ultimately, this call will be converted to the model formula presented to
  # INLA
  fcall <- sys.call()
  fcall[[1]] <- "f"
  fcall[[2]] <- as.symbol(label)
  names(fcall)[2] <- ""
  # 'main' and 'weights' are the only regular parameter allowed to be nameless,
  # and only if they are the first parameters (position 3 and 4 in fcall)
  if ((length(fcall) >= 3L) &&
    (is.null(names(fcall)) || (names(fcall)[3] %in% c("", NA_character_)))) {
    names(fcall)[3] <- "main"
  }
  if ((length(fcall) >= 4L) &&
    (is.null(names(fcall)) || (names(fcall)[4] %in% c("", NA_character_)))) {
    names(fcall)[4] <- "weights"
  }
  unnamed_arguments <- which(names(fcall[-c(1, 2)]) %in% c("", NA_character_))
  if (length(unnamed_arguments) > 0) {
    # Without this check, R gives the error
    #   'In str2lang(s) : parsing result not of length one, but 0'
    # in the INLA call instead, which isn't very informative.
    stop(glue(
      "Unnamed arguments detected in component '{label}'.\n",
      "  Only 'main' and 'weights' parameters may be unnamed.\n",
      "  Unnamed arguments at position(s) ",
      glue_collapse(unnamed_arguments, sep = ", ")
    ))
  }

  # Special and general cases:
  if (component$main$type %in% c("offset", "const")) {
    # The offset is included either automatically for ~ . linear models,
    # or explicitly by name in the predictor expression, so no INLA formula
    # component is needed.
    component$inla.formula <- as.formula("~ .", env = .envir)
    component$main$mapper <-
      ibm_input_set(bm_const(), component$main$mapper)
    component$group <- NULL
    component$replicate <- NULL
    # Add scalable multi-mapper
    component[["mapper"]] <-
      bm_pipe(
        c(
          list(
            core = bm_multi(list(
              main = component[["main"]][["mapper"]]
            ))
          ),
          if (include_weights) {
            list(scale = component[["weights"]])
          } else {
            NULL
          }
        )
      )
  } else {
    if (!is.null(copy)) {
      # Store copy-model name or object in the environment
      #    model_name <- paste0("BRU_", label, "_copy_model")
      #    fcall[["copy"]] <- component$copy # as.symbol(model_name)
      #    bru_comp_env_extra(component, model_name) <- component$copy
    } else {
      # Store model name or object in the environment
      model_name <- glue("BRU_{label}_main_model")
      fcall[["model"]] <- as.symbol(model_name)
      bru_comp_env_extra(component, model_name) <- component$main$model
    }

    # Remove parameters inlabru supports but INLA doesn't,
    # and substitute parameters that inlabru will transform
    fcall <- fcall[!(names(fcall) %in% c(
      "main",
      "weights",
      "mapper",
      "group_mapper",
      "replicate_mapper",
      "A.msk",
      "map",
      "mesh",
      "main_layer",
      "group_layer",
      "replicate_layer",
      "weights_layer",
      "main_selector",
      "group_selector",
      "replicate_selector",
      "weights_selector",
      "marginal"
    ))]

    # Replace arguments that will be evaluated by a mapper
    # TODO: make a more general system
    suffixes <- list(
      "group" = "group",
      "replicate" = "repl"
    )
    for (arg in names(suffixes)) {
      if (arg %in% names(fcall)) {
        fcall[[arg]] <- as.symbol(glue("{label}.{suffixes[[arg]]}"))
      }
    }

    if (is.null(copy)) {
      # These values were previously initialised by an INLA::f call, but ::f
      # should only be called by INLA, and never by inlabru, since it sets up
      # temporary files that will not be removed, and also requires data not
      # available at this point!
      if (!is.null(component$main[["n"]])) {
        fcall[["n"]] <- component$main$n
      }
      if (!is.null(season.length)) {
        fcall[["season.length"]] <- season.length
      }
      if (!is.null(component$main[["nrow"]])) {
        fcall[["nrow"]] <- component$main$nrow
      }
      if (!is.null(component$main[["ncol"]])) {
        fcall[["ncol"]] <- component$main$ncol
      }

      # Make sure 'values' is setup properly.
      if (is.null(component$main$values)) {
        fcall <- fcall[!("values" %in% names(fcall))]
      } else {
        values_name <- glue("BRU_{label}_values")
        fcall[["values"]] <- as.symbol(values_name)
        bru_comp_env_extra(component, values_name) <- component$main$values
      }

      # Setup factor precision parameter
      if (component$main$type %in% c("factor", "fixed")) {
        if (is.null(fcall[["hyper"]])) {
          # TODO: allow configuration of the precision via prec.linear
          fixed_hyper_name <- glue("BRU_{label}_main_fixed_hyper")
          fcall[["hyper"]] <- as.symbol(fixed_hyper_name)
          bru_comp_env_extra(component, fixed_hyper_name) <-
            list(prec = list(
              initial = log(INLA::inla.set.control.fixed.default()$prec),
              fixed = TRUE
            ))
        }
      }
    }

    component$inla.formula <-
      as.formula(
        paste0(
          "~ . + ",
          as.character(parse(text = deparse(fcall)))
        ),
        env = .envir
      )

    component$fcall <- fcall
  }

  class(component) <- "bru_comp"

  component <- bru_comp_update_mapper(component)

  component <- bru_compat_pre_2_14_bru_comp(component)

  component
}

bru_compat_pre_2_14_bru_comp <- function(comp) {
  if (!isTRUE(bru_options_get("bru_compat_pre_2_14_enable"))) {
    return(comp)
  }
  if (!inherits(comp, "bru_comp")) {
    class(comp) <- c("bru_comp", class(comp))
  }
  comp$main$input <- ibm_input_get(comp$main$mapper)
  if (!is.null(comp$group)) {
    comp$group$input <- ibm_input_get(comp$group$mapper)
  }
  if (!is.null(comp$replicate)) {
    comp$replicate$input <- ibm_input_get(comp$replicate$mapper)
  }
  comp
}


#' Methods for inlabru component lists
#'
#' Constructor methods for inlabru component lists. Syntax details are given in
#' [bru_comp()].
#'
#' @param \dots Parameters passed on to other methods. Also see Details.
#' @family component constructors
#' @param object The object to operate on
#' @param lhoods A [bru_obs_list] object
#' @param .envir An evaluation environment for non-formula input
#' @export
#' @rdname bru_comp_list
#' @aliases bru_component_list
#' @aliases component_list
bru_comp_list <- function(object,
                          ...,
                          .envir = parent.frame()) {
  UseMethod("bru_comp_list")
}

#' @describeIn bru_comp_list Convert a component formula
#' into a `bru_comp_list` object
#'
#' @export
#' @family component constructors
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com} and
#' Finn Lindgren \email{finn.lindgren@@gmail.com}
#'
#' @examples
#' # As an example, let us create a linear component. Here, the component is
#' # called "myLinearEffectOfX" while the covariate the component acts on is
#' # called "x". Note that a list of components is returned because the
#' # formula may define multiple components
#'
#' eff <- bru_comp_list(~ myLinearEffectOfX(main = x, model = "linear"))
#' summary(eff[[1]])
#' # Equivalent shortcuts:
#' eff <- bru_comp_list(~ myLinearEffectOfX(x, model = "linear"))
#' eff <- bru_comp_list(~ myLinearEffectOfX(x))
#' # Individual component
#' eff <- bru_comp("myLinearEffectOfX", main = x, model = "linear")
bru_comp_list.formula <- function(object,
                                  ...,
                                  lhoods = NULL,
                                  .envir = parent.frame()) {
  if (!is.null(environment(object))) {
    .envir <- environment(object)
  }
  # Automatically add Intercept and remove -1 unless -1
  # or -Intercept is in components formula
  object <- auto_intercept(object)

  code <- bru_formula_to_bru_obs_code(object)
  parsed <- lapply(code, function(x) rlang::parse_expr(x))
  components <- lapply(
    parsed,
    function(component.expression) {
      rlang::eval_bare(component.expression,
        env = .envir
      )
    }
  )
  environment(object) <- .envir
  bru_comp_list(components, lhoods = lhoods, .envir = .envir)
}


#' @describeIn bru_comp_list Combine a list of components, component lists,
#'   and/or component formulas into a single `bru_comp_list` object
#' @param inputs A tree-like list of component input evaluations,
#' from [bru_input.bru_obs_list()].
#' @export
bru_comp_list.list <- function(object,
                               ...,
                               lhoods = NULL,
                               .envir = parent.frame(),
                               inputs = NULL) {
  # Maybe the list has been given an environment?
  if (!is.null(environment(object))) {
    .envir <- environment(object)
  } else if (is.null(.envir)) {
    # Later code needs an actual environment
    .envir <- new.env()
  }
  is_comp <- vapply(object, function(x) inherits(x, "bru_comp"), TRUE)
  if (!all(is_comp)) {
    is_comp_list <- vapply(object, function(x) {
      inherits(x, "bru_comp_list")
    }, TRUE)
    if (!all(is_comp_list)) {
      object <- lapply(seq_along(object), function(k) {
        bru_comp_list(object[[k]], .envir = .envir)
      })
    }
    object <- unlist(object, recursive = FALSE)
  }
  object <- structure(
    object,
    class = c("bru_comp_list", "list")
  )
  environment(object) <- .envir

  object <- set_list_names(object, tag = "label", priority = "name")
  if (anyDuplicated(names(object))) {
    stop(paste0(
      "Duplicated component labels detected: ",
      glue_collapse(
        glue(
          "'{sort(unique(names(object)[duplicated(names(object))]))}'"
        ),
        sep = ", "
      )
    ))
  }

  if (!is.null(lhoods)) {
    lhoods <- bru_used_update(lhoods, names(object))
    object <- add_mappers(object, lhoods = lhoods, inputs = inputs)
  }
  object
}

#' @export
#' @describeIn bru_comp_list Place a single `bru_comp` object into a
#'   `bru_comp_list` object.
bru_comp_list.bru_comp <- function(object,
                                   ...,
                                   .envir = parent.frame()) {
  bru_comp_list(list(object), .envir = .envir, ...)
}

#' @export
#' @describeIn bru_comp_list Make sure a `bru_comp_list` object is fully
#'   configured.
bru_comp_list.bru_comp_list <- function(object,
                                        ...,
                                        .envir = parent.frame()) {
  bru_comp_list(list(object), .envir = .envir, ...)
}


#' @export
#' @describeIn bru_comp_list The `...` arguments should be `bru_comp_list`
#' objects. The environment from the first argument will be applied to the
#' resulting `bru_comp_list`.
`c.bru_comp_list` <- function(...) {
  bru_comp_list(list(...), .envir = environment(list(...)[[1]]))
}

#' @export
#' @describeIn bru_comp_list The `...` arguments should be `component`
#'   objects from [bru_comp()]. The environment from the first argument
#'   will be applied to the resulting `bru_comp_list`.
`c.bru_comp` <- function(...) {
  bru_comp_list(list(...), .envir = environment(list(...)[[1]]))
}

#' @export
#' @param x `bru_comp_list` object from which to extract a sub-list
#' @param i indices specifying elements to extract
#' @rdname bru_comp_list
`[.bru_comp_list` <- function(x, i) {
  env <- environment(x)
  object <- NextMethod()
  class(object) <- c("bru_comp_list", "list")
  environment(object) <- env
  object
}


#' Get/set component environment data
#'
#' @description Get or set data in the component's `env_extra` and `env`
#'   environments. If `name` is `NULL`, the entire environment is returned.
#' @param x A `bru_comp` object
#' @param name The name of the variable to get or set. Names
#'   prefixed by `"BRU_"` are reserved for internal use and should not be set
#'   by users or external package authors.
#' @param value The value to set for `name` in `env_extra` (for the setter
#'   function).
#' @return For the getter, either the entire environment or the
#'   value of `name` in the environment. For the setters, the modified
#'   `bru_comp` object.
#' @export
#' @rdname bru_comp_env_extra
#' @seealso bru_comp
#' @examples
#' if (bru_safe_inla()) {
#'   cmp <- bru_comp("x", x)
#'   as.list(bru_comp_env_extra(cmp))
#'
#'   cmp <- bru_comp("x", x, model = "fixed")
#'   as.list(bru_comp_env_extra(cmp))
#'
#'   bru_comp_env(cmp)
#' }
#'
bru_comp_env_extra <- function(x, name = NULL) {
  if (is.null(name)) {
    return(x[["env_extra"]])
  }
  get(name, envir = x[["env_extra"]], inherits = FALSE)
}

#' @describeIn bru_comp_env_extra Set data in the component's `env_extra`
#' environment
#' @export
`bru_comp_env_extra<-` <- function(x, name, value) {
  assign(name, value, envir = x[["env_extra"]])
  x
}

#' @describeIn bru_comp_env_extra Get the component's `env` environment, or an
#'   element from that environment.
#' Note that in most cases this environment is the global R environment.
#' @export
bru_comp_env <- function(x, name = NULL) {
  if (is.null(name)) {
    return(x[["env"]])
  }
  get(name, envir = x[["env"]], inherits = FALSE)
}

#' @describeIn bru_comp_env_extra Set data in the component's `env` environment.
#' Note that in most cases this environment is the global R environment, so
#' modifying it is normally not recommended.
#' @export
`bru_comp_env<-` <- function(x, name, value) {
  assign(name, value, envir = x[["env"]])
  x
}


bru_comp_update_mapper <- function(component) {
  # Core component multi-mapper
  core_mapper <- list(
    core = bm_multi(
      lapply(
        component[intersect(
          c("main", "group", "replicate"),
          names(component)
        )],
        function(subcomp) {
          subcomp[["mapper"]]
        }
      )
    )
  )
  # Add marginal transformation mapper
  if (!is.null(component[["marginal"]])) {
    marginal_mapper <- list(marginal = component[["marginal"]])
  } else {
    marginal_mapper <- NULL
  }
  # Add weights-mapper
  if (!is.null(component[["weights"]])) {
    weights_mapper <- list(scale = component[["weights"]])
  } else {
    weights_mapper <- NULL
  }
  # Combine the mappers
  component[["mapper"]] <-
    bm_pipe(
      c(
        core_mapper,
        marginal_mapper,
        weights_mapper
      )
    )
  component
}


#' @title Equip components with mappers
#' @description Equip component(s) with mappers for subcomponents that do not
#'   have predefined mappers. When needed, the data in `lhoods` is used to
#'   determine the appropriate mapper(s).
#' @param component A [component] object
#' @param lhoods A [bru_obs_list] object
#' @param inputs A precomputed list of inputs, as returned by
#'   [bru_input.bru_comp_list()]
#' @param lh_data A list of data object, one for each likelihood in `lhoods`
#'   that is used to determine the mapper(s). If `NULL`, the data is
#'   or inputs are used instead.
#' @return A `component` object with completed mapper information
#' @examples
#' \dontrun{
#' if (interactive() && bru_safe_inla()) {}
#' }
#' @rdname add_mappers
#' @keywords internal
#' @export

add_mappers.bru_comp <- function(component,
                                 lhoods,
                                 inputs = NULL,
                                 lh_data = NULL,
                                 ...) {
  if (is.null(lh_data)) {
    # Filter out lhoods that don't use/support the component
    keep_lh <-
      vapply(lhoods,
        function(lh, label) {
          label %in% bru_used(lh)[["effect"]]
        },
        TRUE,
        label = component$label
      )
    lh <- lhoods[keep_lh]
    inputs <- inputs[keep_lh]

    lh_data <- lapply(lh, function(x) x[["data"]])
  }

  component$main <- add_mapper(
    component$main,
    label = paste0(component$label, "(main)"),
    data = lh_data,
    inputs = lapply(
      inputs,
      function(x) x[[component$label]][["core"]][["main"]]
    ),
    env = component$env,
    require_indexed = FALSE
  )
  if (!is.null(component[["group"]])) {
    component$group <- add_mapper(
      component$group,
      label = paste0(component$label, "(group)"),
      data = lh_data,
      inputs = lapply(
        inputs,
        function(x) x[[component$label]][["core"]][["group"]]
      ),
      env = component$env,
      require_indexed = TRUE
    )
  }
  if (!is.null(component[["replicate"]])) {
    component$replicate <- add_mapper(
      component$replicate,
      label = paste0(component$label, "(replicate)"),
      data = lh_data,
      inputs = lapply(
        inputs,
        function(x) x[[component$label]][["core"]][["replicate"]]
      ),
      env = component$env,
      require_indexed = TRUE
    )
  }
  component <- bru_comp_update_mapper(component)

  fcall <- component$fcall

  # Set ngroup and nrep defaults
  if (!is.null(component[["group"]])) {
    if (is.null(component$group[["n"]])) {
      fcall[["ngroup"]] <- 1L
    } else {
      fcall[["ngroup"]] <- component$group$n
    }
  }
  if (!is.null(component[["replicate"]])) {
    if (is.null(component$replicate[["n"]])) {
      fcall[["nrep"]] <- 1L
    } else {
      fcall[["nrep"]] <- component$replicate$n
    }
  }

  if (is.null(component$main$values)) {
    fcall <- fcall[!("values" %in% names(fcall))]
  } else {
    values_name <- glue("BRU_{component$label}_values")
    fcall[["values"]] <- as.symbol(values_name)
    bru_comp_env_extra(component, values_name) <- component$main$values
  }

  if (!(component[["main"]][["type"]] %in% c("offset", "const"))) {
    # Update the formula that will be presented to INLA
    component$inla.formula <-
      as.formula(
        glue(
          "~ . + ",
          glue_collapse(deparse(fcall), sep = "\n")
        ),
        env = component$env
      )
  }

  component$fcall <- fcall

  component
}

#' @param components A `bru_comp_list` object
#' @export
#' @rdname add_mappers
add_mappers.bru_comp_list <- function(components,
                                      lhoods,
                                      inputs = NULL,
                                      lh_data = NULL,
                                      ...) {
  if (is.null(inputs)) {
    inputs <- bru_input(lhoods, components = components)
  }
  is_copy <- vapply(components, function(x) !is.null(x[["copy"]]), TRUE)
  for (k in which(!is_copy)) {
    components[[k]] <- add_mappers(
      components[[k]], lhoods,
      inputs = inputs, lh_data = lh_data
    )
  }
  for (k in which(is_copy)) {
    if (is.null(components[[k]][["copy"]])) {
      stop(paste0(
        "Internal error: copy model detected, ",
        "but no copy information available."
      ))
    }
    if (is.null(components[[components[[k]][["copy"]]]])) {
      stop(glue(
        "Could not find component '{components[[k]][['copy']]}'",
        " to use as copy for component '{components[[k]][['label']]}'."
      ))
    }
    components[[k]]$mapper <-
      components[[components[[k]][["copy"]]]][["mapper"]]
  }
  components
}

bru_subcomp <- function(input = NULL,
                        mapper = NULL,
                        model = NULL,
                        n = NULL,
                        values = NULL,
                        season.length = NULL,
                        nrow = NULL,
                        ncol = NULL) {
  type <- model
  factor_mapping <- NULL
  if (inherits(model, "inla.spde")) {
    type <- "spde"
  } else if (inherits(model, "inla.rgeneric")) {
    type <- "rgeneric"
  } else if (inherits(model, "inla.cgeneric")) {
    type <- "cgeneric"
  } else if (is.character(model)) {
    if (identical(model, "factor")) {
      model <- "factor_contrast"
      bru_log_warn(
        paste0(
          "Deprecated model 'factor'. Please use 'factor_full' or ",
          "'factor_contrast' instead.\n",
          "Defaulting to 'factor_contrast' that matches the old 'factor' model."
        )
      )
    }
    if (model %in%
      c(
        "clinear", "sigm", "revsigm",
        "log1exp", "logdist"
      )
    ) {
      type <- "specialnonlinear"
    } else if (model %in%
      c(
        "rw2d", "rw2diid", "matern2d"
      )
    ) {
      type <- "inlalattice"
      if (is.null(n)) {
        n <- nrow * ncol
      } else {
        if (n != nrow * ncol) {
          stop(
            "n = ",
            n,
            " must be equal to nrow * ncol = ",
            nrow,
            " * ",
            ncol,
            " = ",
            nrow * ncol,
            " for inla lattice model '",
            model,
            "'."
          )
        }
      }
    } else if (identical(model, "factor_full")) {
      model <- "iid"
      type <- "factor"
      factor_mapping <- "full"
    } else if (identical(model, "factor_contrast")) {
      model <- "iid"
      type <- "factor"
      factor_mapping <- "contrast"
    } else if (identical(model, "fixed")) {
      model <- "iid"
      type <- "fixed"
    } else if (model %in% c("offset", "const")) {
      model <- "const"
      type <- "const"
    } else {
      type <- model
    }
  } else {
    type <- "unknown"
  }
  if (is.null(mapper)) {
    mapper <- bm_autodetect()
  }
  if (!inherits(mapper, "bru_mapper")) {
    stop(
      "Unknown mapper class '",
      glue_collapse(class(mapper), sep = ", "),
      "'"
    )
  }
  mapper <- ibm_input_set(mapper, input)
  subcomponent <-
    structure(
      list(
        mapper = mapper,
        model = model,
        type = type,
        n = n,
        nrow = nrow,
        ncol = ncol,
        values = values,
        season.length = season.length,
        factor_mapping = factor_mapping
      ),
      class = "bru_subcomp"
    )

  subcomponent
}


make_unique_inputs <- function(inp, allow_list = FALSE, label = "") {
  local_stop <- function(msg) {
    stop(paste0(
      "Error in making unique inputs for component '", label, "':\n  ",
      msg
    ))
  }
  is_spatial <- vapply(inp, function(x) inherits(x, "Spatial"), TRUE)
  is_sfc <- vapply(inp, function(x) inherits(x, "sfc"), TRUE)
  is_matrix <- vapply(inp, function(x) is.matrix(x), TRUE)
  is_Matrix <- vapply(inp, function(x) inherits(x, "Matrix"), TRUE)
  is_factor <- vapply(inp, function(x) is.factor(x), TRUE)
  is_data_frame <- vapply(inp, function(x) is.data.frame(x), TRUE)
  is_list <- vapply(inp, function(x) is.list(x), TRUE) &
    !is_data_frame
  if (any(is_spatial)) {
    bru_safe_sp(force = TRUE)
    if (!all(is_spatial)) {
      local_stop(paste0(
        "Inconsistent spatial/non-spatial input. ",
        "Unable to infer mapper information."
      ))
    }
    inconsistent_crs <- FALSE
    inp_crs <- lapply(inp, fm_CRS)
    crs_info <- lapply(inp_crs, fm_wkt)
    null_crs <- vapply(crs_info, is.null, logical(1))
    inconsistent_crs <-
      (length(unique(unlist(crs_info))) > 1) ||
        (any(null_crs) && !all(null_crs))
    if (inconsistent_crs) {
      local_stop(paste0(
        "Inconsistent spatial CRS information. ",
        "Unable to infer mapper information."
      ))
    }
    inp_values <- unique(do.call(
      rbind,
      lapply(
        inp,
        function(x) sp::coordinates(x)
      )
    ))
    n_values <- nrow(inp_values)
  } else if (any(is_sfc)) {
    if (!all(is_sfc)) {
      local_stop(paste0(
        "Inconsistent spatial/non-spatial input. ",
        "Unable to infer mapper information."
      ))
    }
    inconsistent_crs <- FALSE
    inp_crs <- lapply(inp, fm_crs)
    null_crs <- vapply(inp_crs, is.na, logical(1))
    inconsistent_crs <-
      (length(unique(inp_crs)) > 1) ||
        (any(null_crs) && !all(null_crs))
    if (inconsistent_crs) {
      local_stop(paste0(
        "Inconsistent spatial crs information. ",
        "Unable to infer mapper information."
      ))
    }
    inp_values <- sf::st_sfc(unique(do.call(c, inp)),
      crs = inp_crs[[1]]
    )
    n_values <- NROW(inp_values)
  } else if (any(is_matrix | is_Matrix)) {
    if (!all(is_matrix | is_Matrix)) {
      local_stop("Inconsistent input types; matrix and non-matrix")
    }
    # Add extra column to work around bug in unique.matrix for single-column
    # Matrix and ModelMatrix matrices:
    inp_values <-
      unique.matrix(cbind(1, do.call(rbind, inp)))[, -1, drop = FALSE]
    n_values <- nrow(inp_values)
  } else if (any(is_data_frame)) {
    if (!all(is_data_frame)) {
      local_stop("Inconsistent input types; data.frame and non-data.frame")
    }
    inp_values <- unique(do.call(rbind, inp))
    n_values <- nrow(inp_values)
  } else if (any(is_factor)) {
    if (!all(is_factor)) {
      local_stop("Inconsistent input types; factor and non-factor")
    }
    inp_values <- unique(lapply(inp, levels))
    if (length(inp_values) > 1) {
      local_stop(
        paste0(
          "Inconsistent factor levels. Unable to infer mapper information.\n",
          "  Level combinations:\n    ",
          paste0(
            seq_along(inp_values),
            ": ",
            vapply(
              inp_values,
              function(x) paste0("'", x, "'", collapse = ", "),
              ""
            ),
            collapse = "\n    "
          )
        )
      )
    }
    inp_values <- inp_values[[1]]
    n_values <- length(inp_values)
  } else if (any(is_list)) {
    if (!all(is_list)) {
      local_stop("Inconsistent input types; list and non-list")
    }
    if (!allow_list) {
      local_stop(paste0(
        "Unable to create automatic mapper. List data at this level requires ",
        "an explicit mapper definition."
      ))
    }
    inp_values <- list()
    n_values <- list()
    for (i in seq_along(inp[[1]])) {
      inp_ <- list()
      for (j in seq_along(inp)) {
        inp_[[j]] <- inp[[j]][[i]]
      }
      result <- make_unique_inputs(inp_, allow_list = FALSE)
      inp_values[[i]] <- result$inp_values
      n_values[[i]] <- result$n_values
    }
  } else {
    inp_values <- sort(unique(unlist(inp)), na.last = NA)
    n_values <- length(inp_values)
  }

  list(
    inp_values = inp_values,
    n_values = n_values,
    is_list
  )
}


# @param inputs list with precomputed inputs, for each lhoods element;
# @param data list of data objects, one for each lhoods element
# lapply(full_inputs,
#        function(x) x[[component$label]][["core"]][[subcomponent_label]])
add_mapper <- function(subcomp, label, data = NULL, env = NULL,
                       inputs = NULL,
                       require_indexed = FALSE) {
  if (inherits(subcomp[["mapper"]], "bm_autodetect") &&
    ibm_input_available(subcomp[["mapper"]])) {
    mapper_input <- ibm_input_get(subcomp[["mapper"]])
  } else {
    mapper_input <- NULL
  }
  if (is.null(subcomp[["mapper"]]) ||
    inherits(subcomp[["mapper"]], "bm_autodetect")) {
    if (!inherits(subcomp[["model"]], "character")) {
      subcomp[["mapper"]] <- bru_get_mapper_safely(subcomp[["model"]])
      if (inherits(subcomp[["mapper"]], "bru_mapper") &&
        ibm_input_available(subcomp[["mapper"]])) {
        mapper_input <- ibm_input_get(subcomp[["mapper"]])
      }
    }
  }
  if (!is.null(subcomp[["mapper"]]) &&
    !inherits(subcomp[["mapper"]], "bm_autodetect")) {
    if (!inherits(subcomp[["mapper"]], "bru_mapper")) {
      stop(glue(
        "Unknown mapper of type '",
        glue_collapse(class(subcomp[["mapper"]]), sep = ", "),
        "' for {label}"
      ))
    }
  } else if (is.null(mapper_input)) {
    stop(glue(
      "This code should be unused for {label}: ",
      "add_mapper() with isTRUE(is.null(mapper_input))"
    ))
  } else {
    if (!is.null(data) || !is.null(inputs)) {
      if (is.null(inputs) || (length(inputs) == 0)) {
        if (!is.null(data) && (length(data) > 0)) {
          inputs <- lapply(
            data,
            function(dat) {
              bru_input(
                mapper_input,
                data = dat,
                .envir = env,
                label = mapper_input$label,
                null.on.fail = TRUE
              )
            }
          )
        } else {
          # Component not directly used in any likelihood.
          # Attempt to evaluate with no data;
          # useful for intercept-like components only used via the
          # *_latent technique.
          inputs <- list(
            bru_input(
              mapper_input,
              data = NULL,
              .envir = env,
              label = mapper_input$label,
              null.on.fail = TRUE
            )
          )
        }
      }
      # Check for
      # 1) All NULL; Deprecated unless input is NULL. Since version 2.1.14,
      #              intercepts should be notated explicitly with label(1)
      # 2) Some NULL; exclude NULL results
      # TODO: Check for vector/matrix/coordinate inconsistency
      null.results <- vapply(inputs, function(x) is.null(x), TRUE)
      if (all(null.results)) {
        msg <- glue(
          "All covariate evaluations for '{label}'",
          " are NULL; an intercept component was likely intended.\n",
          "  Implicit latent intercept component specification is ",
          "deprecated since version 2.1.14.\n",
          "  Use explicit notation '+ {label}(1)' instead",
          if (identical(label, "Intercept")) {
            " (or '+1' for '+ Intercept(1)')"
          },
          "."
        )
        bru_log_message(msg)
        unique_inputs <- list(
          inp_values = 1,
          n_values = 1
        )
      } else {
        if (any(null.results)) {
          inp_ <- inputs[!null.results]
        } else {
          inp_ <- inputs
        }

        unique_inputs <- make_unique_inputs(
          inp_,
          allow_list = TRUE,
          label = label
        )
      }
      if (sum(unlist(unique_inputs$n_values)) < 1) {
        subcomp$n <- 1
        subcomp$values <- NULL
        inp_values <- NULL
      }
      subcomp[["mapper"]] <- make_mapper(
        subcomp,
        label,
        input_values = unique_inputs$inp_values,
        strict = TRUE,
        require_indexed = require_indexed
      )
    }
  }
  if (!is.null(subcomp[["mapper"]]) &&
    !inherits(subcomp[["mapper"]], "bm_autodetect")) {
    if (!is.null(mapper_input) && !ibm_input_available(subcomp[["mapper"]])) {
      subcomp[["mapper"]] <- ibm_input_set(subcomp[["mapper"]], mapper_input)
    }
    # Check internal consistency of user specified n and the mapper:
    mapper_n <- ibm_n(subcomp[["mapper"]], inla_f = TRUE)
    if (!is.null(subcomp[["n"]]) &&
      subcomp[["n"]] != mapper_n) {
      stop(glue(
        "Size mismatch, n={subcomp[['n']]}  != ",
        "ibm_n(inla_f = TRUE)={mapper_n}, ",
        " in mapper for label '{label}'."
      ))
    }
    subcomp[["n"]] <- mapper_n
    subcomp[["values"]] <- ibm_values(subcomp[["mapper"]], inla_f = TRUE)
  }
  subcomp
}


make_values <- function(subcomp_n, subcomp_values, input_values, label) {
  if (!is.null(subcomp_n)) {
    if (!is.null(subcomp_values)) {
      warning(
        "Both 'n' and 'values' provided. Ignoring 'values'",
        immediate. = TRUE
      )
    }
    return(seq_len(subcomp_n))
  }
  if (!is.null(subcomp_values)) {
    return(subcomp_values)
  }
  if (!is.null(input_values)) {
    return(input_values)
  }

  stop(glue("No mapper, no n, and no values given for '{label}'"))
}

make_submapper <- function(subcomp_n,
                           subcomp_nrow_ncol,
                           subcomp_values,
                           input_values,
                           label,
                           subcomp_type,
                           subcomp_factor_mapping,
                           require_indexed,
                           allow_interpolation = TRUE) {
  if (identical(subcomp_type, "inlalattice")) {
    if (is.null(subcomp_nrow_ncol)) {
      stop(
        glue("nrow and ncol must be provided for inla lattice model '{label}'.")
      )
    }
    n <- prod(subcomp_nrow_ncol)
    if (!is.null(subcomp_n)) {
      stopifnot(n == subcomp_n)
    }
    subcomp_n <- n
    return(
      bm_index(n = subcomp_n)
    )
  }

  values <- make_values(subcomp_n, subcomp_values, input_values, label)

  if (is.factor(values) ||
    is.character(values) ||
    (!is.null(subcomp_type) && (subcomp_type %in% "factor"))) {
    return(
      bm_factor(
        values,
        factor_mapping = subcomp_factor_mapping,
        indexed = require_indexed
      )
    )
  }

  values <- sort(unique(values), na.last = NA)
  if (length(values) > 1) {
    if (allow_interpolation) {
      return(
        bru_mapper(
          fm_mesh_1d(values),
          indexed = require_indexed
        )
      )
    }
    mapper <- bm_index(n = ceiling(max(values)))
    return(mapper)
  }
  if (all(values == 1)) {
    if (require_indexed) {
      return(bm_index(n = 1L))
    }
    return(bm_linear())
  }
  if (require_indexed) {
    return(
      bm_factor(values,
        factor_mapping = "full",
        indexed = TRUE
      )
    )
  }

  bm_linear()
}


#' @title
#' Extract mapper information from INLA model component objects
#'
#' @description
#' The component definitions will automatically attempt to extract mapper
#' information from any model object by calling the generic `bru_get_mapper`.
#' Any class method implementation should return a [bru_mapper] object suitable
#' for the given latent model.
#'
#' @details
#' Before implementing your own `bru_get_mapper` method, check if there
#' is already a general method available that handles your model class, such as
#' [bru_get_mapper.inla.spde()], [bru_get_mapper.inla.rgeneric()], and
#' [bru_get_mapper.inla.cgeneric()].
#'
#' @param model A model component object
#' @param \dots Arguments passed on to other methods
#' @return A [bru_mapper] object defined by the model component
#' @seealso [bru_mapper] for mapper constructor methods, and
#' the individual mappers for specific implementation details.
#' @family mappers
#' @export
#' @examples
#' if (bru_safe_inla()) {
#'   library(INLA)
#'   mesh <- fmesher::fm_rcdt_2d_inla(globe = 2)
#'   spde <- inla.spde2.pcmatern(mesh,
#'     prior.range = c(1, 0.5),
#'     prior.sigma = c(1, 0.5)
#'   )
#'   mapper <- bru_get_mapper(spde)
#'   ibm_n(mapper)
#' }
bru_get_mapper <- function(model, ...) {
  UseMethod("bru_get_mapper", model)
}


#' @describeIn bru_get_mapper Extract an indexed mapper for
#' the `model$mesh` object contained in the model object,
#' which is assumed to be of a class supporting relevant `fmesher` methods.
#' @export
bru_get_mapper.inla.spde <- function(model, ...) {
  bm_fmesher(model[["mesh"]])
}

#' @describeIn bru_get_mapper Returns the mapper given by a pre-computed mapper,
#'   or an index mapper mapping the size of the model graph.
#'
#'   The easiest method to define a mapper for an `inla.rgeneric` model is to
#'   store the mapper in the object.
#'   Alternatively, define your model using a subclass and define a
#'   `bru_get_mapper.<subclass>` method that should return the
#'   corresponding `bru_mapper` object.
#'
#' The order of precedence for the mapper construction when calling
#' `bru_get_mapper(model)` has the following precedence:
#'
#' 1. `bru_get_mapper.<subclass>`, if `model` has a subclass, otherwise
#' 2. `model[["mapper"]]` if that is `NULL`, and otherwise
#' 3. [bm_index()] using the size of the graph returned by `model[["f"]][["n"]]`
#' @export
bru_get_mapper.inla.rgeneric <- function(model, ...) {
  if (!is.null(model[["mapper"]])) {
    return(model[["mapper"]])
  }
  if (!is.null(model[["f"]][["n"]])) {
    return(bm_index(n = model[["f"]][["n"]]))
  }
  NULL
}

#' @describeIn bru_get_mapper Works the same as the method of `inla.rgeneric`
#' @export
bru_get_mapper.inla.cgeneric <- function(model, ...) {
  if (!is.null(model[["mapper"]])) {
    return(model[["mapper"]])
  }
  if (!is.null(model[["f"]][["n"]])) {
    return(bm_index(n = model[["f"]][["n"]]))
  }
  NULL
}

#' @describeIn bru_get_mapper Tries to call the `bru_get_mapper`,
#' and returns `NULL` if it fails (e.g. due to no available class method).
#' If the call succeeds and returns non-`NULL`, it checks that the object
#' inherits from the `bru_mapper` class, and gives an error if it does not.
#' @export
bru_get_mapper_safely <- function(model, ...) {
  m <- tryCatch(
    bru_get_mapper(model, ...),
    error = function(e) {
      NULL
    }
  )
  if (!is.null(m) && !inherits(m, "bru_mapper")) {
    stop(paste0(
      "The bru_get_mapper method for model class '",
      glue_collapse(class(model), sep = ", "),
      "' did not return a bru_mapper object."
    ))
  }
  m
}


# Defines a default mapper given the type of model and parameters provided
# Checks subcomp$mapper, subcomp$model (for "bym2" and other multicomponent
# models), subcomp$model$mesh (for spde models), subcomp$n,
# subcomp$values, or input_values, in that order.
make_mapper <- function(subcomp,
                        label,
                        input_values = NULL,
                        strict = TRUE,
                        require_indexed = FALSE) {
  mapper <- subcomp[["mapper"]]
  if ((is.null(mapper) ||
    inherits(mapper, "bm_autodetect")) &&
    !inherits(subcomp[["model"]], "character")) {
    mapper <- bru_get_mapper_safely(subcomp[["model"]])
  }
  if (!is.null(mapper) && !inherits(mapper, "bm_autodetect")) {
    if (inherits(mapper, "bru_mapper")) {
      return(mapper)
    }
    stop(paste0(
      "Unknown mapper of type '",
      glue_collapse(class(mapper), sep = ", "),
      "' for ", label
    ))
  }
  if (subcomp[["type"]] %in% c("linear", "specialnonlinear")) {
    return(bm_linear())
  }
  if (subcomp[["type"]] %in% c("offset", "const")) {
    return(bm_const())
  }
  if (subcomp[["type"]] %in% c("fixed")) {
    if (!is.null(subcomp[["values"]])) {
      labels <- subcomp[["values"]]
    } else if (!is.null(input_values)) {
      tmp <- as(input_values, "Matrix")
      labels <- colnames(tmp)
      if (is.null(labels)) {
        labels <- as.character(seq_len(ncol(tmp)))
      }
    } else if (!is.null(subcomp[["n"]])) {
      labels <- as.character(seq_len(subcomp[["n"]]))
    } else {
      stop(glue(
        "Need to specify at least one of values (labels), input, or n for ",
        "'{subcomp[['label']]}'",
        " of component '{label}'."
      ))
    }
    return(bm_matrix(labels))
  }
  allow_interpolation <- TRUE
  stopifnot(bru_safe_inla(multicore = TRUE))
  inla_model <- INLA::inla.models()[["latent"]][[subcomp[["model"]]]]
  if (is.null(inla_model) ||
    !is.integer(inla_model[["aug.factor"]]) ||
    (inla_model[["aug.factor"]] == 1L)) {
    the_mapper <- subcomp[["mapper"]]
    the_input <- ibm_input_get(the_mapper)
    mapper_names <- the_input[["label"]]
    labels <- the_input[["label"]]
  } else {
    if (subcomp[["model"]] %in% c("bym", "bym2")) {
      mapper_names <- c("u", "v")
      allow_interpolation <- FALSE
    } else {
      mapper_names <- glue("u{seq_len(inla_model[['aug.factor']])}")
    }
    the_mapper <- subcomp[["mapper"]]
    the_input <- ibm_input_get(the_mapper)
    labels <- glue("{the_input[['label']]}_{mapper_names}")
  }

  mappers <- lapply(
    labels,
    function(lab) {
      make_submapper(
        subcomp_n = subcomp[["n"]],
        subcomp_nrow_ncol = c(subcomp[["nrow"]], subcomp[["ncol"]]),
        subcomp_values = subcomp[["values"]],
        input_values = input_values,
        label = lab,
        subcomp_type = subcomp[["type"]],
        subcomp_factor_mapping = subcomp[["factor_mapping"]],
        require_indexed = require_indexed,
        allow_interpolation = allow_interpolation
      )
    }
  )

  if (length(mapper_names) > 1L) {
    names(mappers) <- mapper_names
    return(bm_collect(mappers, hidden = TRUE))
  }
  mappers[[1]]
}

#' Convert components to R code
#'
#' @aliases bru_formula_to_bru_obs_code
#' @keywords internal
#' @param components A [formula] describing latent model components.
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#'

bru_formula_to_bru_obs_code <- function(components, add = "") {
  fname <- "inlabru:::bru_comp.character"
  # If rhs is "~1", make sure there's at least one component to parse
  # and that offsets show up in "factors"
  tms <- terms(update.formula(components, . ~ . + BRU_DUMMY_COMPONENT + 1))
  codes <- attr(tms, "term.labels")

  # Check for offset()
  offset_idx <- attr(tms, "offset")
  if (length(offset_idx) > 0) {
    isoff <- as.vector(unlist(lapply(
      rownames(attr(tms, "factors")),
      function(s) startsWith(s, "offset")
    )))
    if (!any(isoff)) {
      stop("Internal error: multiple offsets indicated but none extracted")
    }
    codes <- c(codes, rownames(attr(tms, "factors"))[isoff])
  }
  codes <- codes[codes != "BRU_DUMMY_COMPONENT"]

  for (k in seq_along(codes)) {
    code <- codes[[k]]

    # Function syntax or linear component?
    ix <- regexpr("(", text = code, fixed = TRUE)
    is.offset <- FALSE
    if (ix > 0) {
      label <- substr(code, 1, ix - 1)
      is.fixed <- FALSE
      if (label %in% c("offset", "const")) {
        is.offset <- TRUE
      }
    } else {
      label <- code
      is.fixed <- TRUE
    }

    # Make code
    if (is.fixed) {
      codes[[k]] <- glue(
        '{fname}("{label}", main = {label}, model = "linear"{add})'
      )
    } else {
      # Add extra code before final bracket
      ix <- max(gregexpr(")", text = code, fixed = TRUE)[[1]])
      code <-
        paste0(
          substr(code, 1, ix - 1),
          add,
          substr(code, ix, nchar(code))
        )

      if (is.offset) {
        codes[[k]] <-
          sub(
            glue("{label}("),
            glue('{fname}("{label}", model = "const", main = '),
            code,
            fixed = TRUE
          )
      } else {
        codes[[k]] <- sub(
          glue("{label}("),
          glue('{fname}("{label}", '),
          code,
          fixed = TRUE
        )
      }
    }
  }

  codes
}


# OPERATORS ----


#' Summarise components
#'
#' @keywords internal
#' @param object Object to be summarised.
#' @param \dots Passed on to other summary methods.
#' @param depth The depth of which to expand the component mapper.
#' Default `Inf`, to traverse the entire mapper tree.
#' @param verbose logical; If `TRUE`, includes more details of the
#' component definitions. When `FALSE`, only show basic component
#' definition information.  Default `TRUE`.
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @seealso [bru_comp()], [summary.bru_input()]
#' @export
#' @method summary bru_comp

summary.bru_comp <- function(object, ..., depth = Inf, verbose = TRUE) {
  result <- list(
    "Label" = object[["label"]],
    "Copy_of" = object[["copy"]],
    "Type" = paste0(
      unlist(lapply(
        c("main", "group", "replicate"),
        function(x) {
          if (is.null(object[[x]][["input"]][["input"]])) {
            NULL
          } else {
            glue("{x} = {object[[x]][['type']]}")
          }
        }
      )),
      collapse = ", "
    ),
    "Mapper" =
      if (is.null(object[["mapper"]])) {
        "Not yet initialised"
      } else {
        strwrap(
          summary(object[["mapper"]],
            prefix = "    ",
            initial = "",
            depth = depth
          ),
          prefix = "    ",
          initial = ""
        )
      },
    "INLA formula" =
      paste0(
        "\n",
        strwrap(
          paste(as.character(object$inla.formula), collapse = " "),
          width = 0.9 * getOption("width"),
          indent = 4,
          exdent = 6
        )
      )
  )
  if (!verbose) {
    Summary <- paste0(
      object[["label"]],
      if (is.null(object[["copy"]])) {
        NULL
      } else {
        glue("(={object[['copy']]})")
      },
      ": ",
      paste0(
        unlist(lapply(
          c("main", "group", "replicate", "marginal", "weights"),
          function(x) {
            if (!is.null(object[[x]])) {
              format(object[[x]],
                verbose = verbose,
                ...,
                label.override = x
              )
            }
          }
        )),
        collapse = ", "
      )
    )
    result$Summary <- Summary
  }
  class(result) <- "summary_component"
  result
}

#' @export
#' @method summary bru_comp_list
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#' @rdname summary.bru_comp

summary.bru_comp_list <- function(object, verbose = TRUE, ...) {
  result <- lapply(
    object,
    function(x) {
      summary(x, verbose = verbose, ...)
    }
  )
  for (cp in which(vapply(result, function(x) !is.null(x$Copy_of), FALSE))) {
    result[[cp]]$Type <- result[[result[[cp]][["Copy_of"]]]][["Type"]]
  }
  class(result) <- c("summary_bru_comp_list", "list")
  result
}

#' @export
#' @method print bru_comp
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp
print.bru_comp <- function(x, ...) {
  print(summary(x, ...))
  invisible(x)
}

#' @export
#' @method print bru_comp_list
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp
print.bru_comp_list <- function(x, ...) {
  print(summary(x, ...))
  invisible(x)
}


#' @export
#' @method format bru_subcomp
#' @param label.override character; If not `NULL`, use this label instead of
#' the object's label.
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp
format.bru_subcomp <- function(x,
                               verbose = TRUE,
                               ...,
                               label.override = NULL) {
  text <- format(
    ibm_input_get(x[["mapper"]]),
    verbose = verbose,
    ...,
    label.override = label.override,
    type = if (verbose) NULL else x[["type"]]
  )
  text
}

#' @export
#' @method summary bru_subcomp
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp
summary.bru_subcomp <- function(object,
                                verbose = TRUE,
                                ...,
                                label.override = NULL) {
  text <- format(
    object,
    verbose = verbose,
    ...,
    label.override = label.override
  )
  res <- structure(
    list(
      text = text
    ),
    class = "summary_bru_subcomp"
  )
  res
}

#' @export
#' @method print bru_subcomp
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp
print.bru_subcomp <- function(x,
                              verbose = TRUE,
                              ...,
                              label.override = NULL) {
  text <- format(x, verbose = verbose, ..., label.override = label.override)
  cat(text, "\n", sep = "")
  invisible(x)
}


#' @export
#' @param x Object to be printed.
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp

print.summary_component <- function(x, ...) {
  if (is.null(x[["Summary"]])) {
    for (name in names(x)) {
      if (!is.null(x[[name]])) {
        if (name %in% "Label") {
          cat(glue("Label:\t{x[[name]]}"), "\n")
        } else if (name %in% "Mapper") {
          cat(glue("  Map:\t{glue::glue_collapse(x[[name]])}"), "\n")
        } else {
          cat(glue("  {name}:\t{glue::glue_collapse(x[[name]])}"), "\n")
        }
      }
    }
  } else {
    cat(x[["Summary"]], "\n", sep = "")
  }
  invisible(x)
}

#' @export
#' @param x A summary object to be printed.
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_comp

print.summary_bru_comp_list <- function(x, ...) {
  lapply(x, print)
  invisible(x)
}

#' @export
#' @rdname summary.bru_comp

print.summary_bru_subcomp <- function(x, ...) {
  if (!is.null(x[["text"]])) {
    cat(x[["text"]], "\n", sep = "")
  }
  invisible(x)
}

test_that("Multiple likelihoods: basic model", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123L)

  lik1 <- bru_obs("gaussian",
    formula = y ~ .,
    data = data.frame(
      x = rep(c(1, 1.5, 2, 3, 4), 2),
      y = rep(c(11, 12, 13, 14, 12), 2) + rnorm(10, sd = 0.05)
    ),
    used = bru_used(effect = c("int1", "effect")),
    # Checks that control.family is handled
    control.family = list(hyper = list(prec = list(fixed = TRUE)))
  )
  lik2 <- bru_obs("poisson",
    formula = y ~ .,
    data = data.frame(
      x = c(2, 2.5, 3, 4, 5),
      y = ceiling(exp(c(13, 13.5, 14, 12, 12) - 10))
    ),
    used = bru_used(effect = c("int2", "effect")),
    tag = "B"
  )

  cmp1 <- bru_comp_list(
    ~ effect(x, model = "rw2", scale.model = TRUE) + int1(1) + int2(1) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = c(lik1, lik2))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(union(lik1$data$x, lik2$data$x))
  )

  cmp <- bru_comp_list(~ -1 +
    effect(x,
      model = "rw2",
      values = seq(1, 5, by = 0.25),
      scale.model = TRUE
    ) +
    int1(1) + int2(1))

  fit <- bru(cmp, lik1, lik2)

  expect_equal(nrow(fit$summary.hyperpar), 1)
  expect_equal(fit$summary.hyperpar["Precision for effect", "mean"],
    2.0459,
    tolerance = midtol
  )
})


test_that("Predictor indexing", {
  skip_on_cran()
  local_bru_safe_inla()

  fit <- bru(
    ~ 0 + x,
    bru_obs(
      y ~ .,
      data = data.frame(x = 1:3, y = 1:3 + rnorm(3)),
      tag = "A"
    ),
    bru_obs(
      y ~ .,
      data = data.frame(x = 1:4, y = c(NA, NA, 3:4) + rnorm(4)),
      tag = "B"
    )
  )

  expect_equal(
    bru_index(fit),
    1L:7L
  )
  expect_equal(
    bru_index(fit, "A"),
    1L:3L
  )
  expect_equal(
    bru_index(fit, "B"),
    4L:7L
  )
  expect_equal(
    bru_index(fit, c("B", "A")),
    c(4L:7L, 1L:3L)
  )
  expect_equal(
    bru_index(fit, what = "observed"),
    c(1L:3L, 6L:7L)
  )
  expect_equal(
    bru_index(fit, what = "missing"),
    4L:5L
  )
})

test_that("Georeferenced data with sf", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123)
  mydata <- expand.grid(
    Easting = seq(5, 45, by = 20),
    Northing = seq(10, 30, by = 10),
    KEEP.OUT.ATTRS = FALSE
  )
  mydata[["obs"]] <- (mydata$Easting - 20) / 10 + rnorm(NROW(mydata))
  mydata <- sf::st_as_sf(mydata, coords = c("Easting", "Northing"))

  mesh <- fm_mesh_2d_inla(
    loc = mydata,
    offset = 5,
    max.edge = 4,
    n = 16
  )

  matern <- INLA::inla.spde2.pcmatern(
    mesh,
    prior.sigma = c(10, 0.01),
    prior.range = c(4, 0.01)
  )

  # Check that mistaken empty or unnamed arguments are detected
  cmp <- obs ~ Intercept(1) + field(geometry, model = matern, )
  expect_error(
    bru_comp_list(cmp),
    "Unnamed arguments detected in component .* position\\(s\\) 3"
  )

  cmp <- obs ~ Intercept(1) + field(geometry, model = matern)

  fit <- bru(
    cmp,
    data = mydata,
    options = list(
      control.inla = list(
        int.strategy = "eb"
      )
    )
  )

  # Check Intercept
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.5398535,
    tolerance = midtol
  )

  # Check SPDE
  expect_equal(
    fit$summary.random$field$mean[mesh$idx$loc[1:3]],
    c(-2.6003077, -0.2699909, 3.5188725),
    tolerance = midtol
  )


  # Check that explicit access to the data object works
  cmp <- obs ~ Intercept(1) + field(sf::st_coordinates(.data.), model = matern)

  fit <- bru(
    cmp,
    data = mydata,
    options = list(
      control.inla = list(
        int.strategy = "eb"
      )
    )
  )

  # Check Intercept
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.5398535,
    tolerance = midtol
  )

  # Check SPDE
  expect_equal(
    fit$summary.random$field$mean[mesh$idx$loc[1:3]],
    c(-2.6003077, -0.2699909, 3.5188725),
    tolerance = midtol
  )

  pred_df <- fm_pixels(mesh, dims = c(8, 8), format = "sf")

  skip_if_not_installed("sn")
  pred <- predict(fit, pred_df, ~ exp(Intercept + field), n.samples = 5)
  expect_s3_class(pred, "sf")
})


test_that("Georeferenced data with sf, with groups", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123)
  mydata <- expand.grid(
    Easting = seq(5, 45, by = 20),
    Northing = seq(10, 30, by = 10),
    KEEP.OUT.ATTRS = FALSE
  )
  mydata <- rbind(
    cbind(mydata, season = 1L),
    cbind(mydata, season = 2L)
  )
  mydata[["obs"]] <- (mydata$Easting - 20) / 10 + rnorm(NROW(mydata))
  mydata <- sf::st_as_sf(mydata, coords = c("Easting", "Northing"))

  mesh <- fm_mesh_2d_inla(
    loc = mydata,
    offset = 5,
    max.edge = 4,
    n = 16
  )

  matern <- INLA::inla.spde2.pcmatern(
    mesh,
    prior.sigma = c(10, 0.01),
    prior.range = c(4, 0.01)
  )

  cmp <- obs ~ Intercept(1) + field(geometry,
    model = matern,
    group = season,
    ngroup = 2
  )

  fit <- bru(
    cmp,
    data = mydata,
    options = list(
      control.inla = list(
        int.strategy = "eb"
      )
    )
  )

  # Check Intercept
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.6343082,
    tolerance = midtol
  )

  # Check SPDE
  expect_equal(
    fit$summary.random$field$mean[mesh$idx$loc[1:3]],
    c(-2.6947679, -0.3644125, 3.4244099),
    tolerance = midtol
  )

  pred_df <- fmesher::fm_pixels(mesh, dims = c(8, 8), format = "sf")
  expect_s3_class(pred_df, "sf")
  pred_df <- fmesher::fm_cprod(pred_df, data.frame(season = seq_len(2)))

  skip_if_not_installed("sn")
  pred <- predict(fit, pred_df, ~ exp(Intercept + field), n.samples = 5)
  expect_s3_class(pred, "sf")
})

#' @title Unit test helpers
#' @name local_testthat
#' @description Local helper functions for package unit tests
#' @param envir environment for exit handlers
#' @rdname local_testthat
#' @keywords internal
NULL

#' @param x character; Name of variable to assign to
#' @param values the object to assign to `x`
#' @export
#' @describeIn local_testthat Assign local variable. Useful for easy cleanup
#' of global workspace with `withr::deferred_run()` when running tests
#' interactively.
local_bru_testthat_assign <- function(x, values, envir = parent.frame()) {
  exist <- exists(x, envir = envir)
  if (exist) {
    old_value <- envir[[x]]
    withr::defer(assign(x, old_value, envir = envir), envir = envir)
  } else {
    withr::defer(rm(list = x, envir = envir), envir = envir)
  }
  assign(x, values, envir = envir)
}

#' @param tolerances numeric vector of length 3; `[lowtol, midtol, hitol]`
#' @export
#' @describeIn local_testthat Assign test tolerances
#' Assign local tolerance variables. Useful for easy cleanup
#' of global workspace with `withr::deferred_run()` when running tests
#' interactively.
local_bru_testthat_tolerances <- function(tolerances = c(1e-4, 1e-2, 1e-1),
                                          envir = parent.frame()) {
  local_bru_testthat_assign("lowtol", tolerances[1], envir = envir)
  local_bru_testthat_assign("midtol", tolerances[2], envir = envir)
  local_bru_testthat_assign("hitol", tolerances[3], envir = envir)
}


#' @describeIn local_testthat Wrapper for [bru_options_set_local()],
#' to locally override the global package options.
#' @return `local_bru_options_set()` returns a copy of the global override
#' options (not including the defaults), invisibly.
#' @seealso [bru_options_set_local()], [bru_options_default()],
#'   [bru_options_get()]
#' @param .reset For `local_bru_options_set`, logical indicating if the global
#' override options list should be emptied before setting the new option(s).
#'
#' @export

local_bru_options_set <- function(...,
                                  .reset = FALSE,
                                  envir = parent.frame()) {
  bru_options_set_local(..., .reset = .reset, .envir = envir)
}


#' @export
#' @rdname local_testthat
local_basic_intercept_testdata <- function() {
  withr::local_seed(123)
  data.frame(
    Intercept = 1,
    y = rnorm(100)
  )
}

#' @export
#' @rdname local_testthat
local_basic_fixed_effect_testdata <- function() {
  withr::local_seed(123)
  cbind(
    local_basic_intercept_testdata(),
    data.frame(x1 = rnorm(100))
  )
}


# @returns logical; If the option setting was successful, `TRUE` is returned,
# otherwise `FALSE`.
local_inla_options_set <- function(...,
                                   .envir = parent.frame(),
                                   .save_only = FALSE) {
  # Set INLA options
  inla_options <- list(...)
  old_inla_options <- list()
  if (length(inla_options) > 0) {
    for (name in names(inla_options)) {
      old_inla_options[[name]] <- tryCatch(
        INLA::inla.getOption(name),
        error = function(e) {
          e
        }
      )

      if (inherits(old_inla_options[[name]], "error")) {
        return(FALSE)
      }

      if (!.save_only) {
        e <- tryCatch(
          INLA::inla.setOption(name, inla_options[[name]]),
          error = function(e) {
            e
          }
        )
        if (inherits(e, "error")) {
          return(FALSE)
        }
      }

      withr::defer(
        INLA::inla.setOption(name, old_inla_options[[name]]),
        .envir
      )
    }
  }
  TRUE
}


#' @describeIn local_testthat Tests should set num.threads = "1:1:1" to ensure
#'   within-system repeatability by calling `local_bru_safe_inla()`; see also
#'   [bru_safe_inla()]
#' @param multicore logical; if `TRUE`, multiple cores are allowed, and the INLA
#'   `num.threads` option is not checked or altered. Default: `FALSE`, multicore
#'   not allowed (used for examples and unit tests).
#' @param quietly logical; if `TRUE`, prints diagnostic messages. A message is
#'   always printed if the INLA `num.threads` option is altered, regardless of
#'   the `quietly` argument. Default: TRUE.
#' @export
local_bru_safe_inla <- function(multicore = FALSE,
                                quietly = TRUE,
                                envir = parent.frame()) {
  if (requireNamespace("INLA", quietly = TRUE)) {
    inla.call <- tryCatch(
      INLA::inla.getOption("inla.call"),
      error = function(e) {
        e
      }
    )
    if (inherits(inla.call, "error")) {
      return(testthat::skip(
        "inla.getOption('inla.call') failed, skipping INLA tests."
      ))
    }

    if (all(grepl("\\.run$", inla.call))) {
      inla.binary <- gsub("\\.run$", "", inla.call)
      if (!file.exists(inla.binary)) {
        return(testthat::skip(
          paste0(
            "INLA binary '", inla.binary, "' not found. ",
            "INLA not installed correctly, or with platform mismatch.\n",
            "Skipping INLA tests."
          )
        ))
      }
    }


    # Save the num.threads option so it can be restored
    local_inla_options_set(
      num.threads = NULL,
      .envir = envir,
      .save_only = TRUE
    )

    local_inla_options_set(
      inla.timeout = 60,
      fmesher.evolution = 2L,
      fmesher.evolution.warn = TRUE,
      fmesher.evolution.verbosity = "stop",
      .envir = envir,
      .save_only = FALSE
    )

    # withr::local_options(lifecycle_verbosity = "quiet", .local_envir = envir)
  }

  if (!multicore) {
    local_bru_options_set(num.threads = "1:1:1", envir = envir)
  }
  testthat::skip_if_not(bru_safe_inla(multicore = multicore, quietly = quietly))
  # Ignore spurious warnings for INLA 23.12.17. Is fixed in later INLA versions:
  assign(
    "processed.status.for.model.scopy.in.section.latent",
    TRUE,
    INLA::inla.get.inlaEnv()
  )
}


#' @describeIn local_testthat Initialise environment for tests.
#' Assigns tolerance variables.
#' To be called either at the top of a testfile, or inside tests.
#' Does *not* call [local_bru_safe_inla()], since that may invoke a skip and
#' should be called inside each test that relies on INLA.
#' @export
local_bru_testthat_setup <- function(envir = parent.frame()) {
  local_bru_testthat_tolerances(envir = envir)
  local_bru_options_set(
    # Need to specify specific smtp to ensure consistent tests.
    # To specifically test pardiso, need to override locally
    control.compute = list(smtp = "taucs"),
    inla.mode = "compact",
    bru_compat_pre_2_14_enable = FALSE,
    envir = envir
  )
  sp_version <- tryCatch(
    getNamespaceVersion("sp"),
    error = function(e) {
      NULL
    }
  )
  if (!is.null(sp_version) &&
    (utils::compareVersion(sp_version, "1.6-0") >= 0)) {
    old_sp_evolution_status <- tryCatch(
      sp::get_evolution_status(),
      error = function(e) {
        2L
      },
      warning = function(e) {
        2L
      }
    )
    withr::defer(
      tryCatch(sp::set_evolution_status(old_sp_evolution_status),
        warning = function(e) invisible(NULL),
        error = function(e) invisible(NULL)
      ),
      envir = envir
    )
    bru_safe_sp(quietly = TRUE, force = TRUE)
  }

  invisible()
}

test_that("2D LGCP fitting and prediction: Plot sampling", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not(bru_safe_terra())
  skip_if_not_installed("sf")

  options <- list(
    control.inla = list(
      int.strategy = "eb"
    )
  )

  gorillas <- gorillas_sf

  matern <- INLA::inla.spde2.pcmatern(gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(5, 0.01)
  )

  #  if (FALSE) {
  #    bm <- bench::mark(
  #      ips_old = ipoints(
  #        samplers = gorillas$plotsample$plots,
  #        domain = gorillas$mesh,
  #        int.args = list(use_new = FALSE)
  #      ),
  #      ips_new = ipoints(
  #        samplers = gorillas$plotsample$plots,
  #        domain = gorillas$mesh,
  #        int.args = list(use_new = TRUE)
  #      ),
  #      check = FALSE)
  #    ips_old <- ipoints(
  #      samplers = gorillas$plotsample$plots,
  #      domain = gorillas$mesh,
  #      int.args = list(use_new = FALSE)
  #    )
  #    pl <- gorillas$plotsample$plots
  #    pl$ID <- seq_len(NROW(pl))
  #    ips_new <- ipoints(
  #      samplers = pl,
  #      domain = gorillas$mesh,
  #      int.args = list(use_new = TRUE),
  #      group = "ID"
  #    )
  #    ggplot() +
  #      gg(gorillas$mesh) +
  #      gg(ips_old, aes(size=weight, col="old")) +
  #      gg(ips_new, aes(size=weight, col = "new"))
  #    identical(ips_old, ips_new)
  #    ggplot() +
  #      gg(gorillas$mesh) +
  #      gg(ips_new, aes(size=weight, col = factor(ID)))
  #  }

  cmp <- geometry ~ my.spde(main = geometry, model = matern)
  fit <- lgcp(cmp,
    data = gorillas$plotsample$nests,
    samplers = gorillas$plotsample$plots,
    domain = list(geometry = gorillas$mesh),
    options = options
  )

  expect_equal(
    sum(as_bru_obs_list(fit)[[1]]$response_data$BRU_E),
    7.092,
    tolerance = lowtol
  )
  # Test points selected with sd+sd less than 1.2, for a more stable check.
  mean_est <- fit$summary.fixed["Intercept", "mean"] +
    fit$summary.random$my.spde$mean[c(19, 100, 212)]
  expect_snapshot_value(
    mean_est,
    tolerance = midtol,
    style = "serialize"
  )
  sd_est <- fit$summary.fixed["Intercept", "sd"] +
    fit$summary.random$my.spde$sd[c(19, 100, 212)]
  expect_snapshot_value(
    sd_est,
    tolerance = hitol,
    style = "serialize"
  )
})

test_that("sp data detected", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not_installed("sp")
  skip_if_not_installed("terra")

  local_bru_options_set(bru_run = FALSE)
  withr::local_options(lifecycle_verbosity = "error")
  expect_error(
    {
      fit_sp <- bru(
        as.integer(date) ~ field(sp::coordinates, model = "linear"),
        data = gorillas_sp()$nests
      )
    },
    paste0(
      "`data` argument has deprecated support for `Spatial` input; ",
      "was deprecated in inlabru 2.12.0.9023."
    )
  )
})

test_that("basic intercept model", {
  skip_on_cran()
  local_bru_safe_inla()
  options <- list(
    control.inla = list(h = 0.005)
  )
  mycomp <- y ~ 1
  mydata <- local_basic_intercept_testdata()
  mydata$y <- mydata$y + 1
  fit <- bru(mycomp,
    family = "normal",
    data = mydata,
    options = options
  )

  expect_equal(
    fit$summary.fixed["Intercept", ]$mean,
    1.09140515,
    tolerance = midtol
  )
})

test_that("basic intercept model, spatial data", {
  skip_on_cran()
  local_bru_safe_inla()

  options <- list(
    control.inla = list(h = 0.005)
  )
  mycomp <- y ~ 1
  mydata <- local_basic_intercept_testdata()
  mydata$coord1 <- 11
  mydata$coord2 <- 12
  mydata <- sf::st_as_sf(mydata, coords = c("coord1", "coord2"))

  fit <- bru(mycomp,
    family = "normal",
    data = mydata,
    options = options
  )

  expect_equal(
    fit$summary.fixed["Intercept", ]$mean + 1,
    0.09140515 + 1,
    tolerance = midtol
  )
})

test_that("basic fixed effect model", {
  skip_on_cran()
  local_bru_safe_inla()
  options <- list(
    control.inla = list(h = 0.005)
  )
  mycomp <- y ~ 1 + x1
  mydata <- local_basic_fixed_effect_testdata()
  fit <- bru(mycomp,
    family = "normal",
    data = mydata,
    options = options
  )

  expect_equal(
    fit$summary.fixed["Intercept", ]$mean + 1,
    0.08637662 + 1,
    tolerance = midtol
  )
})

test_that("basic fixed effect model, order relevance", {
  skip_on_cran()
  local_bru_safe_inla()
  options <- list(
    control.inla = list(h = 0.005)
  )
  mydata <- local_basic_fixed_effect_testdata()
  mycomp1 <- y ~ Intercept(1) + x1
  fit1 <- bru(mycomp1,
    family = "normal",
    data = mydata,
    options = options
  )
  mycomp2 <- y ~ x1 + Intercept(1)
  fit2 <- bru(mycomp2,
    family = "normal",
    data = mydata,
    options = options
  )

  expect_equal(
    fit2$summary.fixed["Intercept", ],
    fit1$summary.fixed["Intercept", ],
    tolerance = lowtol
  )
})

test_that("interaction fixed effect model", {
  skip_on_cran()
  local_bru_safe_inla()
  options <- list()
  mydata <- local_basic_fixed_effect_testdata()
  withr::local_seed(123L)
  mydata <- cbind(mydata, x2 = sample(
    x = factor(c("A", "B")),
    size = nrow(mydata),
    replace = TRUE
  ))
  mycomp1 <- y ~ -1 + mix(~ -1 + x1:x2, model = "fixed")
  fit0 <- INLA::inla(y ~ -1 + x1:x2, data = mydata, family = "normal")
  fit1 <- bru(mycomp1,
    family = "normal",
    data = mydata,
    options = options
  )
  expect_equal(
    fit1$summary.random$mix$ID,
    rownames(fit0$summary.fixed)
  )
  expect_equal(
    fit1$summary.random$mix$ID,
    c("x1:x2A", "x1:x2B")
  )
  expect_equal(
    fit1$summary.random$mix$mean,
    fit0$summary.fixed$mean,
    tolerance = lowtol
  )
  expect_equal(
    fit1$summary.random$mix$sd,
    fit0$summary.fixed$sd,
    tolerance = midtol
  )
})

test_that("bru: clinear component", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123)

  input.df <- data.frame(x = cos(1:100))
  input.df <- within(input.df, {
    y <- 5 + 2 * x + rnorm(100, mean = 0, sd = 0.1)
  })
  fit <- bru(
    y ~ myLin(main = x, model = "clinear", range = c(0, Inf)) +
      Intercept(1),
    family = "gaussian",
    data = input.df,
    options = list(bru_initial = list(myLin = 2))
  )

  expect_equal(
    fit$summary.random[["myLin"]][1, "mean"],
    2.002517,
    tolerance = hitol
  )
  expect_equal(
    fit$summary.random[["myLin"]][1, "sd"],
    0.013,
    tolerance = hitol
  )

  skip_if_not_installed("sn")
  pr <- predict(
    fit,
    data.frame(x = c(1, 2)),
    ~ myLin + 2,
    n.samples = 5,
    seed = 1L
  )

  expect_equal(pr[, "mean"], c(4.0, 6.0), tolerance = midtol)
})

latent_ar1_testdata <- function() {
  data1 <- data.frame(
    time = c(1, 2, 3, 5, 4),
    obs = c(1, 2, 1, 4, 2)
  )
  data1 <- data.frame(
    time = c(1, 2, 3.5, 5, 4),
    obs = c(1, 2, 1, 4, 2)
  )
  data2 <- data1[5:1, , drop = FALSE]
  data3 <- data1[c(2, 5, 4, 3, 1), , drop = FALSE]

  cmp <- obs ~ time(time, model = "ar1") - Intercept
  fit1 <- bru(cmp,
    data = data1, family = "gaussian"
  )
  fit2 <- bru(cmp,
    data = data2, family = "gaussian"
  )
  fit3 <- bru(cmp,
    data = data3, family = "gaussian"
  )

  list(
    data = list(data1 = data1, data2 = data2, data3 = data3),
    cmp = cmp,
    fit = list(fit1 = fit1, fit2 = fit2, fit3 = fit3)
  )
}

old <- function() {
  fit <- latent_ar1_testdata()

  cbind(
    fit$fit[[1]]$summary.random$time$mean,
    fit$fit[[2]]$summary.random$time$mean,
    fit$fit[[3]]$summary.random$time$mean
  )
  cbind(
    fit$fit[[1]]$summary.random$time$sd,
    fit$fit[[2]]$summary.random$time$sd,
    fit$fit[[3]]$summary.random$time$sd
  )
  cbind(
    fit$fit[[1]]$summary.hyperpar[, "mean"],
    fit$fit[[2]]$summary.hyperpar[, "mean"],
    fit$fit[[3]]$summary.hyperpar[, "mean"]
  )

  fit$fit_inla <- list(
    INLA::inla(obs ~ f(time, model = "ar1") - 1,
      data = fit$data[[1]], family = "gaussian"
    ),
    INLA::inla(obs ~ f(time, model = "ar1") - 1,
      data = fit$data[[2]], family = "gaussian"
    ),
    INLA::inla(obs ~ f(time, model = "ar1") - 1,
      data = fit$data[[3]], family = "gaussian"
    )
  )

  cbind(
    fit$fit_inla[[1]]$summary.hyperpar[, "mean"],
    fit$fit_inla[[2]]$summary.hyperpar[, "mean"],
    fit$fit_inla[[3]]$summary.hyperpar[, "mean"]
  )
  cbind(
    fit$fit_inla[[1]]$summary.random$time$mean,
    fit$fit_inla[[2]]$summary.random$time$mean,
    fit$fit_inla[[3]]$summary.random$time$mean
  )
}

test_that("Latent models: AR1 bru ordering", {
  skip_on_cran()
  local_bru_safe_inla()
  data <- latent_ar1_testdata()

  # Check AR1
  expect_equal(
    data$fit[[1]]$summary.random$time$mean,
    c(1, 2, 1, 2, 4),
    tolerance = midtol
  )
  expect_equal(
    data$fit[[2]]$summary.random$time$mean,
    c(1, 2, 1, 2, 4),
    tolerance = midtol
  )
  expect_equal(
    data$fit[[3]]$summary.random$time$mean,
    c(1, 2, 1, 2, 4),
    tolerance = midtol
  )
})

# Note:
# On non-CRAN systems, INLA should be installed when running the tests
# but all inla-requiring tests should be protected specifically for missing
# INLA.  The check below should NOT be done, as it will fail when testing
# without Suggests packages, which includes INLA.
# The test is now disabled since it's empty
#
# Note 2: This test was likely dubious, as testthat might run it in the same
#         thread as some previous tests.

# test_that("INLA loading", {
#   skip_on_cran()
#   # expect_error(local_bru_safe_inla(), NA, class = "skip")
# })

test_that("2D modelling on the globe", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123L)

  options <- list(
    control.inla = list(
      int.strategy = "eb"
    )
  )

  mesh <- fm_rcdt_2d_inla(globe = 2)

  data <- data.frame(
    Long = rep(seq(-179, 179, length.out = 10), times = 10),
    Lat = rep(seq(-89, 89, length.out = 10), each = 10)
  )
  data <- within(
    data,
    expr = {
      y <- rpois(nrow(data), exp(1 + sin(Lat / 180 * pi) * 2))
    }
  )
  data <- sf::st_as_sf(data,
    coords = c("Long", "Lat"),
    crs = fm_crs("longlat_globe")
  )
  data <- fm_transform(data, crs = fm_crs("sphere"))

  matern <- INLA::inla.spde2.pcmatern(
    mesh,
    prior.range = c(0.1, 0.01),
    prior.sigma = c(0.1, 0.01)
  )
  cmp <- y ~ mySmooth(main = geometry, model = matern) - 1

  fit <- bru(
    cmp,
    data = data,
    family = "poisson",
    options = options
  )

  expect_s3_class(fit, "bru")

  expect_equal(
    fit$summary.hyperpar["Range for mySmooth", "mean"],
    2.7076690,
    tolerance = midtol
  )

  expect_equal(
    fit$summary.hyperpar["Stdev for mySmooth", "mean"],
    0.5524327,
    tolerance = midtol
  )
})


test_that("2D LGCP modelling on the globe", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123L)

  options <- list(
    control.inla = list(
      int.strategy = "eb"
    )
  )

  mesh <- fm_rcdt_2d_inla(globe = 2, crs = fm_crs("sphere"))

  data <- data.frame(
    Long = rep(seq(0, 360 * 9 / 10, length.out = 10), times = 10),
    Lat = 180 / pi * asin(rep(
      seq(1 / 90, 89 / 90, length.out = 10)^0.5,
      each = 10
    ))
  )
  data <- sf::st_as_sf(data,
    coords = c("Long", "Lat"),
    crs = fm_crs("longlat_globe")
  )
  data <- fm_transform(data, crs = fm_crs("sphere"))

  cmp <- geometry ~
    field(main = sf::st_coordinates(geometry), model = "fixed") +
    Intercept(1)

  expect_equal(
    sum(fm_int(mesh)$weight),
    4 * pi
  )

  fit <- bru(
    cmp,
    data = data,
    family = "cp",
    domain = list(geometry = mesh),
    options = options
  )

  expect_s3_class(fit, "bru")

  expect_equal(
    fit$summary.random$field[, "mean"],
    c(0, 0, 2.840911),
    tolerance = midtol
  )
  expect_equal(
    fit$summary.random$field[, "sd"],
    c(0.2089972, 0.2089972, 0.2968185),
    tolerance = midtol
  )
})

latent_spde1D_testdata <- function() {
  local_bru_safe_inla()
  data(Poisson2_1D, package = "inlabru", envir = environment())
  x <- seq(0, 55, length.out = 50)
  mesh1D <- fm_mesh_1d(x, boundary = "free")

  matern <- INLA::inla.spde2.pcmatern(mesh1D,
    prior.range = c(1, 0.01),
    prior.sigma = c(1, 0.01)
  )

  cmp <- count ~ field(main = x, model = matern) + Intercept(1)
  # This model is sensitive to the integration strategy; "eb" is too smooth.
  fit <- bru(cmp,
    data = countdata2, family = "poisson",
    options = list(
      E = countdata2$exposure,
      control.inla = list(h = 0.005)
    )
  )

  list(
    data = countdata2,
    matern = matern,
    cmp = cmp,
    fit = fit
  )
}

test_that("Latent models: SPDE 1D", {
  skip_on_cran()
  local_bru_safe_inla()
  data <- latent_spde1D_testdata()

  # Check Intercept + SPDE (highly negatively correlated)
  expect_snapshot_value(
    data$fit$summary.fixed["Intercept", "mean"] +
      data$fit$summary.random$field$mean[c(1, 25, 50)],
    tolerance = hitol,
    style = "serialize"
  )
})

test_that("Linear predictor offset", {
  skip_on_cran()
  local_bru_safe_inla()

  dat <- data.frame(
    id = c(1:8),
    deaths = c(5, 30, 2, 4, 5, 4, 7, 10),
    pop = c(1000, 2300, 300, 400, 500, 700, 1000, 700)
  )

  # Used to fail
  m1 <- bru(
    deaths ~ 1 + myoffset(log(pop), model = "offset"),
    data = dat,
    family = "poisson"
  )
  m2 <- bru(
    ~ 1 + myoffset(log(pop), model = "offset"),
    formula = deaths ~ Intercept + myoffset,
    data = dat,
    family = "poisson"
  )

  expect_equal(
    m1$summary.fixed$mean,
    m2$summary.fixed$mean,
    ignore_attr = TRUE,
    tolerance = lowtol
  )
  expect_equal(
    m1$summary.fixed$sd,
    m2$summary.fixed$sd,
    ignore_attr = TRUE,
    tolerance = midtol
  )

  # Give an error if an 'offset' option is specified.
  # Note the since 2.11.1.9013 there is no need to make the have the full
  # predictor length for a non-linear model.
  expect_error(
    {
      m3 <- bru(
        ~ Intercept(1),
        formula = deaths ~ Intercept,
        data = dat,
        family = "poisson",
        options = list(offset = log(dat$pop))
      )
    },
    "An offset option was specified which may interfere"
  )
})

test_that("bru: factor component", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123L)

  # Factor models
  input.df <- data.frame(
    x1 = factor(rep(c("Alpha", "Beta", "Mu"), c(70, 60, 70))),
    x2 = factor(rep(
      c("Delta", "Gamma", "H1", "H2", "H3", "H4", "H5"),
      c(75, 75, 10, 10, 10, 10, 10)
    ))
  )
  input.df$y <-
    c(1, -2, 0)[as.numeric(input.df$x1)] +
    c(1, -2, 1, -2, 1, -2, 1)[as.numeric(input.df$x2)] +
    rnorm(200, mean = 0, sd = 0.1)

  # Fit the model
  fit <- bru(
    components = ~ Intercept(main = 1) +
      fac1(main = x1, model = "factor_contrast") +
      fac2(main = x2, model = "iid"),
    formula = y ~ Intercept + fac1 + fac2,
    family = "gaussian",
    data = input.df,
    options = list(
      verbose = FALSE,
      control.inla = list(
        int.strategy = "eb"
      )
    )
  )

  #  fit0$summary.hyperpar
  #  fit$summary.hyperpar
  #  rbind(
  #    cbind(ID = rownames(fit0$summary.fixed), fit0$summary.fixed),
  #    fit0$summary.random$x2
  #  )
  #  rbind(
  #    cbind(ID = rownames(fit$summary.fixed), fit$summary.fixed),
  #    fit$summary.random$fac1,
  #    fit$summary.random$fac2
  #  )

  # Check factor effect results
  expect_equal(
    fit$summary.random$fac1$mean,
    c(-3.072345, -1.092111),
    tolerance = midtol
  )
  expect_equal(
    fit$summary.random$fac1$sd,
    c(0.04367061, 0.05013299),
    tolerance = midtol
  )
  expect_equal(
    fit$summary.random$fac2$mean,
    c(
      1.213695, -1.723471,  1.310547, -1.671748,
      1.302147, -1.700772,  1.271110
    ),
    tolerance = midtol
  )
  expect_equal(
    fit$summary.random$fac2$sd,
    c(
      0.5225545, 0.5212273, 0.5216542, 0.5216542,
      0.5216542, 0.5216542, 0.5216542
    ),
    tolerance = hitol
  )

  # Check if prediction works
  skip_if_not_installed("sn")
  pr <- predict(
    fit,
    data.frame(
      x1 = factor(c("Alpha", "Beta", "Mu")),
      x2 = factor(c("Delta", "Delta", "Gamma"))
    ),
    ~ fac1 + fac2,
    n.samples = 500,
    seed = 1,
    num.threads = "1:1:1"
  )
  expect_equal(
    pr[, "mean"],
    c(1.223399, -1.849207, -2.807144),
    tolerance = midtol
  )
  expect_equal(
    pr[, "sd"],
    c(0.5064775, 0.5113702, 0.5070317),
    tolerance = hitol
  )
})


test_that("bru: indexed factor component", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123)

  # Factor models
  input.df <- data.frame(
    x3 = rep(c(1, 2, 3, 4), times = 2)
  )
  input.df$y <-
    c(11, 12, 13, 14)[input.df$x3] +
    rnorm(8, mean = 0, sd = 0.1)

  # Fit the model
  fit <- bru(
    components = ~ Intercept(main = 1) +
      fac3(main = x3, model = "factor_contrast"),
    formula = y ~ Intercept + fac3,
    family = "gaussian",
    data = input.df,
    options = list(
      verbose = FALSE,
      control.inla = list(
        int.strategy = "eb",
        h = 0.005
      )
    )
  )


  # Check factor effect results
  expect_equal(
    fit$summary.random$fac3$mean,
    c(1.095837, 2.122574, 2.961865),
    tolerance = midtol
  )
  expect_equal(
    fit$summary.random$fac3$sd,
    c(0.07778122, 0.07778122, 0.07778122),
    tolerance = midtol
  )
})

test_that("sf gorillas lgcp vignette", {
  ##  skip("Feature not yet implemented")

  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not_installed("sf")

  # Code adapted from the lgcp vignette with an additional mesh construction
  # step instead of using the mesh provided in the gorillas data.

  # All sp objects in the example are replaced with sf equivalents.

  data(gorillas_sf, package = "inlabru", envir = environment())

  # ## Not needed anymore for this test, but kept for future reference
  # # Bug in st_as_sf making check_ring_dir=TRUE have no effect, as
  # # st_as_sfc.SpatialPolygons ignores it. To get around it, need to convert to
  # # sfc_POLYGON first, and then do a separate st_sfc call, which does use
  # # check_ring_dir.
  # # No effect:
  # b1 <- sf::st_as_sf(gorillas$boundary, check_ring_dir = TRUE)
  # # st_sfc gives the proper effect:
  # b2 <- b1
  # b2$geometry <- sf::st_sfc(sf::st_geometry(b2$geometry),
  #   check_ring_dir = TRUE)
  #
  # gorillas_sf$boundary <- b2

  ## Build boundary information:
  boundary <- list(
    gorillas_sf$boundary,
    NULL
  )

  ## Build the mesh:
  mesh_sf <- fm_mesh_2d_inla(
    boundary = boundary,
    max.edge = c(0.54, 0.97),
    min.angle = c(30, 21),
    ## Safeguard against large meshes.
    max.n = c(48000, 16000),
    ## Don't build a huge mesh!
    max.n.strict = c(128000, 128000),
    ## Filter away adjacent points.
    cutoff = 0.01,
    ## Offset for extra boundaries, if needed.
    offset = c(0.73, 1.55),
    ## Build mesh in this crs:
    crs = fm_crs(gorillas_sf$nests)
  )

  # library(ggplot2)
  # ggplot() +
  #   gg(mesh_sf) +
  #   geom_sf(data = gorillas_sf$boundary, alpha = 0.2, fill = "blue")

  matern <- INLA::inla.spde2.pcmatern(mesh_sf,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(0.1, 0.01)
  )

  cmp <- geometry ~ mySmooth(geometry, model = matern) +
    Intercept(1)

  expect_no_error({
    fit <- lgcp(
      cmp,
      data = gorillas_sf$nests,
      samplers = gorillas_sf$boundary,
      domain = list(geometry = mesh_sf)
    )
  })
})

test_that("Latent models: RW2 mapping", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123L)

  data1 <- data.frame(
    time = rep(c(1, 2, 4, 8, 16), times = 4),
    obs = rep(c(1, 2, 1, 4, 2), times = 4) + rnorm(20, sd = 0.5)
  )

  cmp1 <- obs ~ time(time,
    model = "rw2", values = 2^(0:4),
    constr = FALSE, scale.model = TRUE
  ) - Intercept
  fit1 <- bru(cmp1, data = data1, family = "gaussian")

  expect_equal(
    fit1$summary.random$time$mean,
    c(1.631781, 1.895681, 1.025671, 3.959789, 1.841140),
    tolerance = midtol
  )
  expect_equal(
    fit1$summary.random$time$sd,
    c(0.2350089, 0.2438991, 0.2749531, 0.2542052, 0.2473781),
    tolerance = hitol
  )
})


test_that("Latent models: RW2 mapping, data is list with different I/O sizes", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(123L)

  data1 <- list(
    time = c(1, 2, 4, 8, 16),
    obs = rep(c(1, 2, 1, 4, 2), times = 4) + rnorm(20, sd = 0.5)
  )

  cmp1 <- obs ~ time(time,
    model = "rw2", values = 2^(0:4),
    constr = FALSE, scale.model = TRUE
  ) - Intercept
  formula <- obs ~ rep(time, times = 4)
  fit1 <- bru(cmp1,
    formula = formula, data = data1, family = "gaussian",
    is_rowwise = FALSE
  )

  expect_equal(
    fit1$summary.random$time$mean,
    c(1.631781, 1.895681, 1.025671, 3.959789, 1.841140),
    tolerance = midtol
  )
  expect_equal(
    fit1$summary.random$time$sd,
    c(0.2350011, 0.2438999, 0.2749631, 0.2542013, 0.2473662),
    tolerance = hitol
  )
})

#' @include deprecated.R
#' @include mappers.R

## _collect ####

#' @title Mapper for concatenated variables
#' @param mappers A list of `bru_mapper` objects
#' @param hidden `logical`, set to `TRUE` to flag that the mapper is to be used
#' as a first level input mapper for `INLA::f()` in a model that requires making
#' only the first mapper visible to `INLA::f()` and `INLA::inla.stack()`, such
#' as for "bym2" models, as activated by the `inla_f` argument to `ibm_n`,
#' `ibm_values`, and `ibm_jacobian`. Set to `FALSE` to always access the full
#' mapper, e.g. for `rgeneric` models
#' @description
#' Constructs a concatenated collection mapping
#' @export
#' @inheritParams bru_mapper_generics
#' @inheritParams bm_scale
#' @inheritParams bm_multi
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' (m <- bm_collect(list(
#'   a = bm_index(2),
#'   b = bm_index(3)
#' ), hidden = FALSE))
#' ibm_eval2(m, list(a = c(1, 2), b = c(1, 3, 2)), 1:5)
#'
#' @rdname bm_collect
#'
bm_collect <- function(mappers, hidden = FALSE) {
  mapper <- list(
    mappers = as_bm_list(mappers),
    n_multi = lapply(mappers, ibm_n),
    values_multi = lapply(mappers, ibm_values),
    hidden = hidden,
    is_linear_multi = lapply(mappers, ibm_is_linear),
    is_rowwise_multi = lapply(mappers, ibm_is_rowwise)
  )
  mapper[["n"]] <- sum(unlist(mapper[["n_multi"]]))
  mapper[["values"]] <- seq_len(mapper[["n"]])
  mapper[["is_linear"]] <- all(unlist(mapper[["is_linear_multi"]]))
  mapper[["is_rowwise"]] <- hidden && all(mapper[["is_rowwise_multi"]][[1]])
  bru_mapper_define(mapper, new_class = "bm_collect")
}

#' @export
#' @rdname bm_collect
bru_mapper_collect <- function(...) {
  bm_collect(...)
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_collect <- function(mapper,
                             inla_f = FALSE,
                             multi = FALSE,
                             ...) {
  if (multi) {
    mapper[["n_multi"]]
  } else if (mapper[["hidden"]] && inla_f) {
    mapper[["n_multi"]][[1]]
  } else {
    mapper[["n"]]
  }
}


# Only for inla_f = FALSE or not hidden
bm_collect_indexing <- function(mapper, input) {
  if (is.matrix(input)) {
    nms <- colnames(input)
    len <- ncol(input)
  } else {
    stopifnot(is.list(input))
    nms <- names(input)
    len <- length(input)
  }

  nms_mapper <- names(mapper[["mappers"]])
  if (is.null(nms)) {
    indexing <- seq_len(min(length(nms_mapper), length(input)))
  } else {
    indexing <- intersect(nms_mapper, nms)
  }
  names(indexing) <- nms_mapper[indexing]
  indexing
}


#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_collect <- function(mapper, input,
                                    state = NULL,
                                    inla_f = FALSE,
                                    multi = FALSE, ...) {
  if (mapper[["hidden"]] && inla_f) {
    return(ibm_n_output(mapper[["mappers"]][[1]], input = input))
  }

  indexing <- bm_collect_indexing(mapper, input)
  if (is.matrix(input)) {
    n <- vapply(
      indexing,
      function(x) {
        as.integer(
          ibm_n_output(mapper[["mapper"]][[x]], input[, x], ...)
        )
      },
      0L
    )
  } else {
    n <- vapply(
      indexing,
      function(x) {
        as.integer(
          ibm_n_output(mapper[["mapper"]][[x]], input[[x]], ...)
        )
      },
      0L
    )
  }

  if (!multi) {
    n <- sum(n)
  }

  n
}


#' @export
#' @rdname ibm_values
#'
ibm_values.bm_collect <- function(mapper,
                                  inla_f = FALSE,
                                  multi = FALSE,
                                  ...) {
  if (multi) {
    mapper[["values_multi"]]
  } else if (mapper[["hidden"]] && inla_f) {
    mapper[["values_multi"]][[1]]
  } else {
    mapper[["values"]]
  }
}

#' @export
#' @rdname ibm_is_linear
#'
ibm_is_linear.bm_collect <- function(mapper,
                                     inla_f = FALSE,
                                     multi = FALSE,
                                     ...) {
  if (mapper[["hidden"]] && inla_f && !multi) {
    ibm_is_linear(mapper[["mappers"]][[1]])
  } else if (multi) {
    mapper[["is_linear_multi"]]
  } else {
    mapper[["is_linear"]]
  }
}


bm_collect_sub_lin <- function(mapper, input, state,
                               inla_f = FALSE,
                               ...) {
  if (mapper[["hidden"]] && inla_f) {
    input <- list(input)
  }
  indexing <- bm_collect_indexing(mapper, input)
  if (is.matrix(input)) {
    nms <- colnames(input)
    input <- as.data.frame(input)
  } else {
    nms <- names(input)
  }
  if (is.null(nms)) {
    names(input) <- names(indexing)
  }

  # We need all the sub_lin objects even if some input is NULL
  indexing <- names(mapper[["mappers"]])
  n_multi <- unlist(mapper[["n_multi"]])
  n_offset <- c(0, cumsum(n_multi))
  names(n_offset) <- c(names(n_multi), "THEEND")
  sub_lin <-
    lapply(
      indexing,
      function(x) {
        state_subset <- state[n_offset[x] + seq_len(n_multi[x])]
        ibm_linear(
          mapper[["mappers"]][[x]],
          input =
            if (is.numeric(x) && (x > length(input))) {
              NULL
            } else {
              input[[x]]
            },
          state = state_subset
        )
      }
    )
  sub_lin
}


#' @describeIn ibm_jacobian
#' Accepts a list with
#' named entries, or a list with unnamed but ordered elements.
#' The names must match the sub-mappers, see [ibm_names.bm_collect()].
#' Each list element should take a format accepted by the corresponding
#' sub-mapper. In case each element is a vector, the input can be given as a
#' data.frame with named columns, a matrix with named columns, or a matrix
#' with unnamed but ordered columns. When `inla_f=TRUE` and `hidden=TRUE` in
#' the mapper definition, the input format should instead match that of
#' the first, non-hidden, sub-mapper.
#' @export
#'
ibm_jacobian.bm_collect <- function(mapper, input, state = NULL,
                                    inla_f = FALSE, multi = FALSE,
                                    ...,
                                    sub_lin = NULL) {
  if (is.null(sub_lin)) {
    sub_lin <- bm_collect_sub_lin(mapper, input, state, inla_f = inla_f)
  }
  A <- lapply(sub_lin, function(x) x[["jacobian"]])

  if (multi) {
    return(A)
  }

  # Combine the matrices (A1, A2, A3, ...) -> bdiag(A1, A2, A3, ...)
  A <- Matrix::.bdiag(A)
  A
}


#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_collect <- function(mapper, input, state,
                                inla_f = FALSE, multi = FALSE,
                                ...,
                                sub_lin = NULL) {
  if (is.null(sub_lin)) {
    sub_lin <- bm_collect_sub_lin(mapper, input, state, inla_f = inla_f)
  }
  val <- lapply(sub_lin, function(x) x[["offset"]])
  if (multi) {
    return(val)
  }

  # Combine the vectors (b1, b2, b3, ...) -> c(b1, b2, b3, ...)
  val <- do.call(c, val)
  val
}


#' @export
#' @rdname ibm_linear
#'
ibm_linear.bm_collect <- function(mapper, input, state,
                                  inla_f = FALSE,
                                  ...) {
  if (mapper[["hidden"]] && inla_f) {
    input <- list(input)
  }
  sub_lin <-
    bm_collect_sub_lin(mapper, input, state,
      inla_f = FALSE,
      ...
    )
  eval2 <- ibm_eval2(
    mapper,
    input = input,
    state = state,
    inla_f = FALSE,
    multi = FALSE,
    ...,
    sub_lin = sub_lin
  )
  bm_taylor(
    offset = eval2$offset,
    jacobian = eval2$jacobian,
    state0 = state,
    values_mapper = mapper
  )
}


#' @describeIn ibm_invalid_output
#' Accepts a list with
#' named entries, or a list with unnamed but ordered elements.
#' The names must match the sub-mappers, see [ibm_names.bm_collect()].
#' Each list element should take a format accepted by the corresponding
#' sub-mapper. In case each element is a vector, the input can be given as a
#' data.frame with named columns, a matrix with named columns, or a matrix
#' with unnamed but ordered columns.
#' @export
#'
ibm_invalid_output.bm_collect <- function(mapper, input, state,
                                          inla_f = FALSE,
                                          multi = FALSE, ...) {
  if (mapper[["hidden"]] && inla_f) {
    return(
      ibm_invalid_output(
        mapper[["mappers"]][[1]],
        input = input,
        multi = FALSE
      )
    )
  }

  indexing <- bm_collect_indexing(mapper, input)
  if (is.matrix(input)) {
    invalid <-
      lapply(
        indexing,
        function(x) {
          ibm_invalid_output(
            mapper[["mappers"]][[x]],
            input = input[, x],
            multi = FALSE
          )
        }
      )
  } else if (is.list(input)) {
    invalid <-
      lapply(
        indexing,
        function(x) {
          ibm_invalid_output(
            mapper[["mappers"]][[x]],
            input = input[[x]],
            multi = FALSE
          )
        }
      )
  } else {
    # TODO: check and fix this!
    invalid <- as.list(rep(TRUE, length(mapper[["mappers"]])))
  }

  if (multi) {
    return(invalid)
  }

  # Combine the vectors (v1, v2, v3) -> c(v1, v2, v3)
  invalid_ <- do.call(c, invalid)
  invalid_
}

#' @return
#' * `[`-indexing a `bm_collect` extracts a subset
#'   `bm_collect` object (for drop `FALSE`) or an individual sub-mapper
#'   (for drop `TRUE`, and `i` identifies a single element)
#' @export
#' @param x object from which to extract element(s)
#' @param i indices specifying element(s) to extract
#' @param drop logical;
#' For `[.bm_collect`, whether to extract an individual mapper when
#' `i` identifies a single element. If `FALSE`, a list of sub-mappers is
#' returned (suitable e.g. for creating a new `bm_collect` object).
#' Default: `TRUE`
#' @rdname bm_collect
#'
`[.bm_collect` <- function(x, i, drop = TRUE) {
  if (is.logical(i)) {
    i <- which(i)
  }
  mapper <- x[["mappers"]][i]
  if (drop) {
    if (length(mapper) == 1) {
      mapper <- mapper[[1]]
    } else if (length(mapper) == 0) {
      mapper <- NULL
    }
  }
  mapper
}

#' @export
#' @rdname bm_collect
`[.bru_mapper_collect` <- function(x, i, drop = TRUE) {
  if (is.logical(i)) {
    i <- which(i)
  }
  mapper <- x[["mappers"]][i]
  if (drop) {
    if (length(mapper) == 1) {
      mapper <- mapper[[1]]
    } else if (length(mapper) == 0) {
      mapper <- NULL
    }
  }
  mapper
}

#' @describeIn ibm_names Returns the names from the
#' sub-mappers list
#' @export
#'
`ibm_names.bm_collect` <- function(mapper) {
  names(mapper[["mappers"]])
}

#' @export
#' @rdname ibm_names
#'
`ibm_names<-.bm_collect` <- function(mapper, value) {
  names(mapper[["mappers"]]) <- value
  names(mapper[["n_multi"]]) <- value
  names(mapper[["values_multi"]]) <- value
  mapper
}

#' @export
#' @rdname ibm_names
#'
`ibm_names<-.bru_mapper_collect` <- function(mapper, value) {
  names(mapper[["mappers"]]) <- value
  names(mapper[["n_multi"]]) <- value
  names(mapper[["values_multi"]]) <- value
  mapper
}

#' Update used_component information objects
#'
#' Merge available component labels information with used components
#' information.
#'
#' @param x Object to be updated
#' @param labels character vector of component labels
#' @param \dots Unused
#' @returns An updated version of `x`
#' @keywords internal
#' @export
#' @family bru_used
bru_used_update <- function(x, labels, ...) {
  UseMethod("bru_used_update")
}

#' @rdname bru_used_update
#' @export
bru_used_update.bru_obs_list <- function(x, labels, ...) {
  for (k in seq_along(x)) {
    x[[k]] <- bru_used_update(x[[k]], labels = labels, ...)
  }
  x
}

#' @rdname bru_used_update
#' @export
bru_used_update.bru_obs <- function(x, labels, ...) {
  x[["pred_expr"]] <- bru_used_update(
    x[["pred_expr"]],
    labels = labels,
    ...
  )
  x <- bru_compat_pre_2_14_bru_obs(x)
  x
}

#' @rdname bru_used_update
#' @export
bru_used_update.bru_pred_expr <- function(x, labels, ...) {
  pre_used <- bru_used(x)
  used <- bru_used_update(pre_used, labels = labels, ...)
  if (isTRUE(x[["is_additive"]])) {
    if ((length(used$latent) > 0) ||
      (length(setdiff(pre_used$effect, used$effect)) > 0)) {
      x[["is_additive"]] <- FALSE
      x[["is_linear"]] <- FALSE

      if (is.null(x[["pred_expr"]])) {
        pred_text <- paste0(pre_used$effect, collapse = " + ")
        if (length(pre_used$latent) > 0) {
          x[["is_rowwise"]] <- FALSE
          pred_text <- paste(
            pred_text,
            paste0(pre_used$latent, "_latent", collapse = " + "),
            sep = " + "
          )
        }
        x[["pred_text"]] <- pred_text
        x[["pred_expr"]] <- rlang::parse_expr(x[["pred_text"]])
      }
    } else {
      if (is.null(x[["pred_expr"]])) {
        x[["pred_text"]] <- paste0(used$effect, collapse = " + ")
        x[["pred_expr"]] <- rlang::parse_expr(x[["pred_text"]])
      }
    }
  }
  x[["used"]] <- used
  x
}

#' @rdname bru_used_update
#' @export
bru_used_update.bru_used <- function(x, labels, ...) {
  used <- x
  used$effect <-
    parse_inclusion(
      labels,
      include = used[["effect"]],
      exclude = used[["effect_exclude"]]
    )
  used[["effect_exclude"]] <- NULL
  if (is.null(used[["latent"]])) {
    used$latent <- character(0)
  }
  used$latent <-
    parse_inclusion(
      labels,
      include = used[["latent"]],
      exclude = NULL
    )
  used
}


#' List components used in a model
#'
#' Create or extract information about which components are used by a model, or
#' its individual observation models. If a non-NULL `labels` argument is
#' supplied, also calls [bru_used_update()] on the `bru_used` objects.
#'
#' @param x An object that contains information about used components
#' @param labels character; component labels passed on to
#' [bru_used_update()]

#' @param join Whether to join list output into a single object; Default
#' may depend on the input object class
#' @param \dots Parameters passed on to the other methods
#' @returns A `bru_used` object (a list with elements `effect`
#' and `latent`), or a list of such objects
#' (for methods with `join = FALSE`)
#'
#' @examples
#' (used <- new_bru_used(~.))
#' bru_used(used, labels = c("a", "c"))
#' (used <- new_bru_used(~ a + b + c_latent + d_latent))
#' bru_used(used, labels = c("a", "c"))
#' (used <- new_bru_used(expression(a + b + c_latent + d_latent)))
#' bru_used(used, labels = c("a", "c"))
#'
#' @export
#' @keywords internal
#' @family bru_used
bru_used <- function(x = NULL, ...) {
  # Need to specify the dispatch object explicitly to handle the NULL case:
  UseMethod("bru_used", x)
}

#' @title Store information about used components
#'
#' @description Create a `bru_used` object from effect name character
#'   vectors.
#' Create information about which components are used by a model, or
#' its individual observation models. If a non-NULL `labels` argument is
#' supplied, also calls [bru_used_update()] on the `bru_used` object.
#'
#' @param x `NULL`, or an object representing an expression
#' @param effect character; components used as effects. When `NULL`, auto-detect
#' components to include all components in a predictor expression.
#' @param effect_exclude character; components to specifically exclude from
#' effect evaluation. When `NULL`, do not specifically exclude any components.
#' @param latent character; components used as `_latent` or `_eval()`. When
#' `NULL`, auto-detect components.
#' @param labels character; component labels passed on to
#' [bru_used_update()]

#' @param \dots Parameters passed on to the other methods
#' @returns A `bru_used` object (a list with elements `effect`
#' and `latent`)
#'
#' @details
#' The arguments `effect`, `effect_exclude`, and `latent` control what
#' components and effects are available for use in predictor expressions.
#' \describe{
#'   \item{`effect`}{
#'   Character vector of component labels that are used as effects
#'   by the predictor expression; If `NULL` (default), the names
#'   are extracted from the formula.
#'   }
#'   \item{`exclude`}{
#'   Character vector of component labels to be excluded from the effect list
#'   even if they have been auto-detected as being necessary.
#'   Default is `NULL`; do not remove any components from the inclusion list.
#'   }
#'   \item{`include_latent`}{Character vector.
#'   Specifies which latent state variables need to be directly available to the
#'   predictor expression, with a `_latent` suffix. This also makes evaluator
#'   functions with suffix `_eval` available, taking parameters `main`, `group`,
#'   and `replicate`, taking values for where to evaluate the component effect
#'   that are different than those defined in the component definition itself
#'   (see [bru_comp_eval()]). If `NULL`, the use of `_latent` and `_eval`
#'   in the predictor expression is detected automatically.
#'   }
#' }
#'
#' @examples
#' (used <- new_bru_used(~.))
#' bru_used(used, labels = c("a", "c"))
#' (used <- new_bru_used(~ a + b + c_latent + d_latent))
#' bru_used(used, labels = c("a", "c"))
#' (used <- new_bru_used(expression(a + b + c_latent + d_latent)))
#' bru_used(used, labels = c("a", "c"))
#'
#' @export
#' @keywords internal
#' @family bru_used
#' @export
new_bru_used <- function(x = NULL,
                         ...,
                         effect = NULL,
                         effect_exclude = NULL,
                         latent = NULL,
                         labels = NULL) {
  # Need to specify the dispatch object explicitly to handle the NULL case:
  UseMethod("new_bru_used", x)
}

#' @describeIn new_bru_used Create a `bru_used` object from effect name
#'   character vectors.
#' @export
new_bru_used.NULL <- function(x = NULL,
                              ...,
                              effect = NULL,
                              effect_exclude = NULL,
                              latent = NULL,
                              labels = NULL) {
  used <- structure(
    list(
      effect = effect,
      latent = latent
    ),
    class = "bru_used"
  )
  used[["effect_exclude"]] <- effect_exclude

  if (!is.null(labels)) {
    used <- bru_used_update(used, labels = labels)
  }

  used
}

#' @describeIn bru_used Create a `bru_used` object by calling [new_bru_used()].
#' @export
bru_used.NULL <- function(x = NULL, ...) {
  new_bru_used(x = x, ...)
}


# Function from
# https://stackoverflow.com/questions/63580260/
#   is-there-a-way-to-stop-all-vars-returning-names-from-the-right-hand-side-of
# corrected to handle multiple $ correctly
replace_dollar <- function(expr) {
  if (!is.language(expr) || length(expr) == 1L) {
    return(expr)
  }
  if (expr[[1]] == quote(`$`)) {
    expr[[1]] <- quote(`[[`)
    expr[[3]] <- as.character(expr[[3]])
    expr[[2]] <- replace_dollar(expr[[2]])
    expr[[3]] <- replace_dollar(expr[[3]])
  } else {
    for (i in seq_along(expr)[-1]) {
      if (!is.null(expr[[i]])) {
        expr[[i]] <- replace_dollar(expr[[i]])
      }
    }
  }
  expr
}

#' @title Extract basic variable names from expression
#'
#' @description
#' Extracts the variable names from an R expression by pre- and post-processing
#' around [all.vars()].
#' First replaces `$` with `[[` indexing, so that internal column/variable names
#' are ignored, then calls `all.vars()`.
#'
#' @param x A `formula`, `expression`, or `character`
#' @param functions logical; if TRUE, include function names
#'
#' @returns If successful, a character vector, otherwise `NULL`
#'
#' @examples
#' bru_used_vars(~.)
#' bru_used_vars(~ a + b + c_latent + d_eval())
#' bru_used_vars(expression(a + b + c_latent + d_eval()))
#'
#' bru_used_vars(~., functions = TRUE)
#' bru_used_vars(~ a + b + c_latent + d_eval(), functions = TRUE)
#' bru_used_vars(expression(a + b + c_latent + d_eval()), functions = TRUE)
#'
#' bru_used_vars(a ~ b)
#' bru_used_vars(expression(a ~ b))
#'
#' @keywords internal
#' @export
#' @family bru_used
bru_used_vars <- function(x, functions = FALSE) {
  UseMethod("bru_used_vars")
}

#' @rdname bru_used_vars
#' @export
bru_used_vars.character <- function(x, functions = FALSE) {
  ex <- paste0(x, collapse = "\n")
  ex <- str2lang(ex)
  ex <- replace_dollar(ex)
  vars <- all.vars(ex, functions = functions, unique = TRUE)
  # Map '.' to NULL, to defer decision until later
  # Do not map character(0) to NULL, as it is a valid result
  if (identical(vars, ".")) {
    vars <- NULL
  }
  vars
}

#' @rdname bru_used_vars
#' @export
bru_used_vars.expression <- function(x, functions = FALSE) {
  attributes(x) <- NULL
  ex <- deparse1(x, collapse = "\n")
  bru_used_vars(ex, functions = functions)
}

#' @rdname bru_used_vars
#' @export
bru_used_vars.quosure <- function(x, functions = FALSE) {
  ex <- deparse1(rlang::quo_get_expr(x), collapse = "\n")
  bru_used_vars(ex, functions = functions)
}

#' @describeIn bru_used_vars Only the right-hand side is used.
#' @export
bru_used_vars.formula <- function(x, functions = FALSE) {
  form <- x[[length(x)]]
  ex <- deparse1(form, collapse = "\n")
  bru_used_vars(ex, functions = functions)
}


#' @describeIn bru_used Create a `bru_used` object by calling [new_bru_used()]
#' @export
bru_used.default <- function(x, ...) {
  new_bru_used(x = x, ...)
}

#' @describeIn new_bru_used Create a `bru_used` object from a `character`
#' representation of an expression.
#' @export
new_bru_used.character <- function(x, ...,
                                   effect = NULL,
                                   effect_exclude = NULL,
                                   latent = NULL,
                                   labels = NULL) {
  form <- x
  if (is.null(effect)) {
    effect <- bru_used_vars(form, functions = FALSE)
    effect <- effect[
      !grepl("^.*_latent$", effect) &
        !grepl("^.*_eval$", effect)
    ]
  }
  if (is.null(latent)) {
    latent <- bru_used_vars(form, functions = TRUE)

    include_latent <- latent[grepl("^.*_latent$", latent)]
    include_latent <- gsub("_latent$", "", include_latent)
    include_eval <- latent[grepl("^.*_eval$", latent)]
    include_eval <- gsub("_eval$", "", include_eval)
    latent <- union(include_latent, include_eval)
    if (length(latent) == 0) {
      include_latent <- character(0)
    }
  }

  new_bru_used(
    x = NULL,
    ...,
    effect = effect,
    effect_exclude = effect_exclude,
    latent = latent,
    labels = labels
  )
}


#' @describeIn new_bru_used Create a `bru_used` object from an expression
#'   object.
#' @export
new_bru_used.expression <- function(x, ...,
                                    effect = NULL,
                                    effect_exclude = NULL,
                                    latent = NULL,
                                    labels = NULL) {
  attributes(x) <- NULL
  y <- deparse1(x, collapse = "\n")
  new_bru_used(
    x = y,
    ...,
    effect = effect,
    effect_exclude = effect_exclude,
    latent = latent,
    labels = labels
  )
}

#' @describeIn new_bru_used Create a `bru_used` object from an `rlang` `expr`
#'   object.
#' @export
new_bru_used.quosure <- function(x, ...,
                                 effect = NULL,
                                 effect_exclude = NULL,
                                 latent = NULL,
                                 labels = NULL) {
  y <- deparse1(rlang::quo_get_expr(x), collapse = "\n")
  attributes(y) <- NULL
  new_bru_used(
    x = y,
    ...,
    effect = effect,
    effect_exclude = effect_exclude,
    latent = latent,
    labels = labels
  )
}


#' @describeIn new_bru_used Create a `bru_used` object from a formula (only the
#' right-hand side is used).
#' @export
new_bru_used.formula <- function(x, ...,
                                 effect = NULL,
                                 effect_exclude = NULL,
                                 latent = NULL,
                                 labels = NULL) {
  form <- x[[length(x)]]
  ex <- deparse1(form, collapse = "\n")
  new_bru_used(
    x = ex,
    ...,
    effect = effect,
    effect_exclude = effect_exclude,
    latent = latent,
    labels = labels
  )
}

#' @describeIn bru_used Extract the `bru_used` information for the collection
#' of observation models used in a `bru` object.
#' @export
bru_used.bru <- function(x, ..., join = TRUE) {
  bru_used(x[["bru_info"]], ..., join = join)
}

#' @describeIn bru_used Extract the `bru_used` information for the collection
#' of observation models used in a `bru_info` object.
#' @export
bru_used.bru_info <- function(x, ..., join = TRUE) {
  bru_used(x[["lhoods"]], ..., join = join)
}

#' @describeIn bru_used Extract the `bru_used` information for each element
#'   of a list, and optionally join into a single `bru_used` object.
#' @export
bru_used.list <- function(x, ..., join = TRUE) {
  used <- lapply(x, function(y) bru_used(y, ...))
  if (join) {
    effect_exclude <- unique(unlist(lapply(used, function(y) {
      y[["effect_exclude"]]
    })))
    if (length(effect_exclude) > 0) {
      stop(paste0(
        "Cannot join 'bru_used' objects with non-null ",
        "'effect_exclude' information."
      ))
    }
    used <-
      bru_used(
        effect = unique(unlist(lapply(used, function(y) y[["effect"]]))),
        latent = unique(unlist(lapply(used, function(y) y[["latent"]])))
      )
  }
  used
}

#' @describeIn bru_used Extract the `bru_used` information for the observation
#'   model predictor used in a `bru` observation model `bru_obs` object.
#' @export
bru_used.bru_obs <- function(x, ...) {
  bru_used(x[["pred_expr"]], ...)
}

#' @describeIn bru_used Extract the `bru_used` information for an observation
#'   model predictor `bru_pred_expr` object.
#' @export
bru_used.bru_pred_expr <- function(x, ...) {
  bru_used(x[["used"]], ...)
}

#' @describeIn bru_used Convenience method that takes
#' an existing `bru_used` object and calls [bru_used_update()]
#' if `labels` is non-NULL.
#' @export
bru_used.bru_used <- function(x, labels = NULL, ...) {
  if (!is.null(labels)) {
    x <- bru_used_update(x, labels = labels)
  }
  x
}


#' @describeIn bru_used Text formatting method for `bru_used` objects.
#' @export
format.bru_used <- function(x, ...) {
  if (is.null(x[["effect"]])) {
    s <- paste0("effect[<not yet initialised>]")
  } else {
    s <- paste0("effect[", paste0(x$effect, collapse = ", "), "]")
  }
  if (is.null(x[["latent"]])) {
    s <- paste0(s, ", latent[<not yet initialised>]")
  } else {
    s <- paste0(s, ", latent[", paste0(x$latent, collapse = ", "), "]")
  }
  if (!is.null(x[["effect_exclude"]])) {
    s <- paste0(s, ", exclude[", paste0(x$effect_exclude, collapse = ", "), "]")
  }
  s
}

#' @describeIn bru_used Summary method for `bru_used` objects.
#' @export
#' @method summary bru_used
summary.bru_used <- function(object, ...) {
  object
}

#' @describeIn bru_used Print method for `bru_used` objects.
#' @export
print.bru_used <- function(x, ...) {
  cat("Used ", format(x), "\n", sep = "")
  invisible(x)
}

#' @name shrimp
#' @title Blue and red shrimp in the Western Mediterranean Sea
#' @docType data
#' @description Blue and red shrimp in the Western Mediterranean Sea.
#'
#' @usage data(shrimp)
#'
#' @format A list of objects:
#'  \describe{
#'    \item{`hauls`:}{ An `sf` object containing haul locations}
#'    \item{`mesh`:}{ An `fm_mesh_2d` object containing a Delaunay triangulation
#'      mesh (a type of discretization of continuous space) covering the haul
#'      locations.}
#'    \describe{
#'       \item{`catch`}{Catch in Kg.}
#'       \item{`landing`}{Landing in Kg.}
#'       \item{`depth`}{Mean depth (in metres) of the fishery haul.}
#'     }
#'  }
#' @source
#' Pennino, Maria Grazia. Personal communication.
#'
#' @references
#' Pennino, M. G., Paradinas, I., Munoz, F., Illian, J.,Quilez-Lopez, A.,
#' Bellido, J.M., Conesa, D. Accounting for preferential sampling in species
#' distribution models. Ecology and Evolution, 9(1), p653-663, 2019
#' \doi{10.1002/ece3.4789}
#'
#' @examples
#' \donttest{
#' if (require(ggplot2, quietly = TRUE)) {
#'   data(shrimp, package = "inlabru", envir = environment())
#'   ggplot() +
#'     fmesher::geom_fm(data = shrimp$mesh) +
#'     gg(shrimp$hauls, aes(col = catch)) +
#'     coord_sf(datum = fmesher::fm_crs(shrimp$hauls))
#' }
#' }
"shrimp"

# Re-export fmesher functions for temporary backwards compatibility.
# These functions are not intended for public use and will be removed in a
# future release.
# For 2.14.1, fm_int and fm_pixels are kept, as they are used
# by intSDM without explicitly importing from fmesher.

# @export
# fmesher::fm_cprod
# @export
# fmesher::fm_crs
#' @export
fmesher::fm_int
#' @export
fmesher::fm_pixels

#' @name robins_subset
#' @title robins_subset
#' @docType data
#' @description This is the `robins_subset` dataset, which is a subset of the
#'   full robins data set used to demonstrate a spatially varying trend
#'   coefficient model in Meehan et al. 2019. The dataset includes American
#'   Robin counts, along with time, location, and effort information, from
#'   Audubon Christimas Bird Counts (CBC) conducted in six US states between
#'   1987 and 2016.
#'
#' @format The data are a data.frame with variables
#'  \describe{
#'    \item{`circle`:}{
#'    Four-letter code of the CBC circle.
#'    }
#'    \item{`bcr`:}{
#'    Numeric code for the bird conservation region encompassing the
#'    count circle.
#'    }
#'    \item{`state`:}{
#'    US state encompassing the count circle.
#'    }
#'    \item{`year`:}{
#'    calendar year the count was conducted.
#'    }
#'    \item{`std_yr`:}{
#'    transformed year, with 2016 = 0.
#'    }
#'    \item{`count`:}{
#'    number of robins recorded.
#'    }
#'    \item{`log_hrs`:}{
#'    the natural log of party hours.
#'    }
#'    \item{`lon`:}{
#'    longitude of the count circle centroid.
#'    }
#'    \item{`lat`:}{
#'    latitude of the count circle centroid.
#'    }
#'    \item{`obs`:}{
#'    unique record identifier.
#'    }
#'  }
#' @source
#' https://github.com/tmeeha/inlaSVCBC
#'
#'
#' @references
#' Meehan, T.D., Michel, N.L., and Rue, H. 2019. Spatial modeling of Audubon
#' Christmas Bird Counts reveals fine-scale patterns and drivers of relative
#' abundance trends. Ecosphere, 10(4), p.e02707.
#'
#' @examples
#' if (require(ggplot2, quietly = TRUE)) {
#'   data(robins_subset, package = "inlabru") # get the data
#'
#'   # plot the counts for one year of data
#'   ggplot(robins_subset[robins_subset$std_yr == 0, ]) +
#'     geom_point(aes(lon, lat, colour = count + 1)) +
#'     scale_colour_gradient(low = "blue", high = "red", trans = "log")
#' }
"robins_subset"

# Linearisation ----

#' Compute inlabru model linearisation information
#'
#' @param \dots Parameters passed on to other methods
#' @export
#' @rdname bru_compute_linearisation
#' @keywords internal
bru_compute_linearisation <- function(...) {
  UseMethod("bru_compute_linearisation")
}

#' @param cmp A [bru_comp] object
#' @param model A [bru_model] object
#' @param lhood_expr A predictor expression
#' @param data Input data
#' @param data_extra Additional data for the predictor
#' @param input Precomputed component inputs from `bru_input()`
#' @param state The state information, as a list of named vectors
#' @param comp_simple Component evaluation information
#' * For `bru_comp`: A [bm_taylor] object
#' * For `bru_obs`: A [bm_list] object
#'   for the components in the likelihood
#' * For `bru_obs_list`: A list of [bm_list] objects
#' @param effects
#' * For `bru_comp`:
#' Precomputed effect list for all components involved in the likelihood
#' expression
#' @param pred0 Precomputed predictor for the given state
#' @param used A [bru_used()] object for the predictor expression
#' @param is_rowwise logical; If `FALSE`, the predictor expression may
#' involve several rows of the input data to influence the same row.
#' @param eps The finite difference step size
#' @param n_pred The length of the predictor expression. If not `NULL`, scalar
#' predictor evaluations are expanded to vectors of length `n_pred`.
#'
#' @export
#' @rdname bru_compute_linearisation
bru_compute_linearisation.bru_comp <- function(cmp,
                                               model,
                                               lhood_expr,
                                               data,
                                               data_extra,
                                               input,
                                               state,
                                               comp_simple,
                                               effects,
                                               pred0,
                                               used,
                                               is_rowwise,
                                               eps,
                                               n_pred = NULL,
                                               ...) {
  label <- cmp[["label"]]
  bru_log_message(
    paste0("Linearise with respect to component '", label, "'"),
    verbosity = 5
  )

  if (cmp[["main"]][["type"]] %in% c("offset", "const")) {
    # Zero-column matrix, since a const/offset has no latent state variables.
    return(Matrix::sparseMatrix(
      i = c(),
      j = c(),
      x = c(1),
      dims = c(NROW(pred0), 0)
    ))
  }

  allow_latent <- label %in% used[["latent"]]

  if (is.null(comp_simple)) {
    A <- NULL
    assume_rowwise <- FALSE
  } else {
    A <- ibm_jacobian(
      comp_simple,
      input = input[[label]],
      state = state[[label]]
    )

    assume_rowwise <- !allow_latent && is_rowwise && is.data.frame(data)

    if (assume_rowwise) {
      if (!is.null(n_pred) && (NROW(pred0) != n_pred)) {
        stop(
          "Number of rows (",
          NROW(pred0),
          ") in the predictor for component '",
          label,
          "' does not match the length implied by the response data (",
          n_pred,
          ")."
        )
      }
      if (NROW(A) == 1L) {
        A <- Matrix::kronecker(rep(1, NROW(pred0)), A)
      }
    }
  }

  triplets <- list(
    i = integer(0),
    j = integer(0),
    x = numeric(0)
  )

  if (!all(is.finite(pred0))) {
    warning(
      "Non-finite (-Inf/Inf/NaN) entries detected in predictor.\n",
      immediate. = TRUE
    )
  }

  symmetric_diffs <- FALSE
  for (k in seq_len(NROW(state[[label]]))) {
    if (is.null(A)) {
      row_subset <- seq_len(NROW(pred0))
    } else {
      Ak <- A[, k, drop = TRUE]
      row_subset <- which(Ak != 0.0)
    }
    if (length(row_subset) > 0) {
      if (symmetric_diffs) {
        state_eps <- list(state, state)
        state_eps[[1]][[label]][k] <- state[[label]][k] - eps
        state_eps[[2]][[label]][k] <- state[[label]][k] + eps
      } else {
        state_eps <- state
        state_eps[[label]][k] <- state[[label]][k] + eps
      }
      # TODO:
      # Option: filter out the data and effect rows for which
      # the rows of A have some non-zeros, or all if !is_rowwise
      # Option: compute predictor for multiple different states. This requires
      # constructing multiple states and corresponding effects before calling
      # evaluate_predictor

      if (symmetric_diffs) {
        effects_eps <- list(effects, effects)
      } else {
        effects_eps <- effects
      }
      if (!is.null(A)) {
        if (assume_rowwise) {
          if (symmetric_diffs) {
            for (label_loop in names(effects)) {
              if (NROW(effects[[label_loop]]) == 1) {
                effects_eps[[1]][[label_loop]] <-
                  rep(effects[[label_loop]], length(row_subset))
                effects_eps[[2]][[label_loop]] <-
                  rep(effects[[label_loop]], length(row_subset))
              } else {
                effects_eps[[1]][[label_loop]] <-
                  effects[[label_loop]][row_subset]
                effects_eps[[2]][[label_loop]] <-
                  effects[[label_loop]][row_subset]
              }
            }
            effects_eps[[1]][[label]] <-
              effects_eps[[1]][[label]] - Ak[row_subset] * eps
            effects_eps[[2]][[label]] <-
              effects_eps[[2]][[label]] + Ak[row_subset] * eps
          } else {
            for (label_loop in names(effects)) {
              if (NROW(effects[[label_loop]]) == 1) {
                effects_eps[[label_loop]] <-
                  rep(effects[[label_loop]], length(row_subset))
              } else {
                effects_eps[[label_loop]] <- effects[[label_loop]][row_subset]
              }
            }
            effects_eps[[label]] <- effects_eps[[label]] + Ak[row_subset] * eps
          }
        } else {
          if (symmetric_diffs) {
            effects_eps <- list(effects, effects)
            effects_eps[[1]][[label]] <- effects_eps[[1]][[label]] - Ak * eps
            effects_eps[[2]][[label]] <- effects_eps[[2]][[label]] + Ak * eps
          } else {
            effects_eps <- effects
            effects_eps[[label]] <- effects_eps[[label]] + Ak * eps
          }
        }
      }
      pred_eps <- evaluate_predictor(
        model,
        state = if (symmetric_diffs) {
          state_eps
        } else {
          list(state_eps)
        },
        data =
          if (assume_rowwise) {
            data[row_subset, , drop = FALSE]
          } else {
            data
          },
        data_extra = data_extra,
        effects =
          if (symmetric_diffs) {
            effects_eps
          } else {
            list(effects_eps)
          },
        predictor = lhood_expr,
        used = used,
        format = "matrix",
        n_pred =
          if (assume_rowwise) {
            length(row_subset)
          } else {
            n_pred
          }
      )
      # Store sparse triplet information
      if (symmetric_diffs) {
        if (assume_rowwise) {
          values <- (pred_eps[, 2] - pred_eps[, 1]) / 2
        } else {
          values <- (pred_eps[, 2] - pred_eps[, 1]) / 2
        }
      } else {
        if (!all(is.finite(pred_eps))) {
          warning(
            "Non-finite (-Inf/Inf/NaN) entries detected in predictor '",
            label,
            "' plus eps.\n",
            immediate. = TRUE
          )
        }
        if (assume_rowwise) {
          values <- (pred_eps - pred0[row_subset])
        } else {
          values <- (pred_eps - pred0)
        }
      }
      nonzero <- is.finite(values)
      if (!all(nonzero)) {
        warning(
          "Non-finite (-Inf/Inf/NaN) entries detected in predictor ",
          "derivatives for '",
          label,
          "'; treated as 0.0.\n",
          immediate. = TRUE
        )
      }
      nonzero[nonzero] <- (values[nonzero] != 0.0) # Detect exact (non)zeros
      if (assume_rowwise) {
        triplets$i <- c(triplets$i, row_subset[nonzero])
      } else {
        triplets$i <- c(triplets$i, which(nonzero))
      }
      triplets$j <- c(triplets$j, rep(k, sum(nonzero)))
      triplets$x <- c(triplets$x, values[nonzero] / eps)
    }
  }
  B <- Matrix::sparseMatrix(
    i = triplets$i,
    j = triplets$j,
    x = triplets$x,
    dims = c(NROW(pred0), NROW(state[[label]]))
  )
  if (NROW(B) != NROW(pred0)) {
    stop(
      "Jacobian matrix for component '",
      label,
      "' has ",
      NROW(B),
      " rows, but expected ",
      NROW(pred0),
      " rows based on the predictor length."
    )
  }
  B
}

#' @param lhood A `bru_obs` object
#' @param model A `bru_model` object
#' @export
#' @rdname bru_compute_linearisation
bru_compute_linearisation.bru_obs <- function(lhood,
                                              model,
                                              data,
                                              input,
                                              state,
                                              comp_simple,
                                              eps,
                                              ...) {
  used <- bru_used(lhood)
  pred_expr <- bru_pred_expr(lhood)
  is_rowwise <- bru_is_rowwise(pred_expr)
  effects <- evaluate_effect_single_state(
    comp_simple[used[["effect"]]],
    input = input[used[["effect"]]],
    state = state[used[["effect"]]]
  )

  lhood_expr <- bru_pred_expr(pred_expr, format = "expr")
  n_pred <- bru_response_size(lhood)

  pred0 <- evaluate_predictor(
    model,
    state = list(state),
    data = data,
    data_extra = lhood[["data_extra"]],
    effects = list(effects),
    predictor = lhood_expr,
    used = used,
    format = "matrix",
    n_pred = n_pred
  )

  # Compute derivatives for each non-const/offset component
  B <- list()
  offset <- pred0
  # Either this loop or the internal bru_comp specific loop
  # can in principle be parallelised.
  comp_lst <- as_bru_comp_list(model)
  additive_and_rowwise <- bru_is_additive(lhood) && is_rowwise
  for (label in union(used[["effect"]], used[["latent"]])) {
    if (ibm_n(comp_lst[[label]][["mapper"]]) > 0) {
      if (additive_and_rowwise) {
        # If additive and no combinations allowed, just need to copy the
        # non-offset A matrix, and possibly expand to full size
        A <- ibm_jacobian(
          comp_simple[[label]],
          input[[label]],
          state[[label]]
        )
        if (NROW(A) == 1) {
          if (NROW(offset) > 1) {
            B[[label]] <- Matrix::kronecker(rep(1, NROW(offset)), A)
          } else {
            B[[label]] <- A
          }
        } else {
          B[[label]] <- A
        }
      } else {
        B[[label]] <-
          bru_compute_linearisation(
            comp_lst[[label]],
            model = model,
            lhood_expr = lhood_expr,
            data = data,
            data_extra = lhood[["data_extra"]],
            input = input,
            state = state,
            comp_simple = comp_simple[[label]],
            effects = effects,
            pred0 = pred0,
            used = used,
            is_rowwise = is_rowwise,
            eps = eps,
            n_pred = n_pred,
            ...
          )
      }
      if ((NROW(offset) == 1L) && (NROW(B[[label]]) > 1L)) {
        offset <- matrix(offset, NROW(B[[label]]), 1)
      }
      offset <- offset - B[[label]] %*% state[[label]]
    }
  }

  bm_taylor(offset = offset, jacobian = B, state0 = NULL)
}

#' @param lhoods A `bru_obs_list` object
#' @export
#' @rdname bru_compute_linearisation
bru_compute_linearisation.bru_obs_list <- function(lhoods,
                                                   model,
                                                   input,
                                                   state,
                                                   comp_simple,
                                                   eps = 1e-5,
                                                   ...) {
  # TODO: set the eps default more intelligently
  lapply(seq_along(lhoods), function(idx) {
    x <- lhoods[[idx]]
    bru_compute_linearisation(
      x,
      model = model,
      data = x[["data"]],
      input = input[[idx]],
      state = state,
      comp_simple = comp_simple[[idx]],
      eps = eps,
      ...
    )
  })
}

#' @export
#' @rdname bru_compute_linearisation
bru_compute_linearisation.bru_model <- function(model, lhoods,
                                                input, state,
                                                comp_simple, ...) {
  bru_compute_linearisation(lhoods,
    model = model,
    input = input, state = state,
    comp_simple = comp_simple, ...
  )
}

#' inlabru
#'
#' Convenient model fitting using (iterated) INLA.
#'
#' @details
#'
#' `inlabru` facilitates Bayesian spatial modelling using integrated nested
#' Laplace approximations. It is heavily based on R-inla
#' (<https://www.r-inla.org>) but adds additional modelling abilities and
#' simplified syntax for (in particular) spatial models.
#' Tutorials and more information can be found at
#' <https://inlabru-org.github.io/inlabru/> and <http://www.inlabru.org/>.
#' The iterative method used for non-linear predictors is documented in the
#' `method` vignette.
#'
#' The main function for inference using inlabru is [bru()].
#' The general model specification details is documented in [bru_comp()]
#' and [bru_obs()].
#' Posterior quantities beyond the basic summaries can be calculated with
#' a `predict()` method, documented in [predict.bru()].
#' For point process inference [lgcp()] can be used as a shortcut to
#' `bru(..., bru_obs(model="cp", ...))`.
#'
#' The package comes with multiple real world data sets, namely [gorillas],
#' [gorillas_sf], [mexdolphin_sf]. Plotting these data
#' sets is straight forward using inlabru's extensions
#' to `ggplot2`, e.g. the [gg()] function. For educational purposes some
#' simulated data sets are available as well, e.g. [Poisson1_1D],
#' [Poisson2_1D], [Poisson2_1D] and [toygroups].
#'
#' @aliases inlabru
#' @import stats
#' @import methods
#' @importFrom Matrix diag
#' @import fmesher
#' @importFrom glue glue
#' @importFrom glue glue_data
#' @importFrom glue glue_collapse
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#'   and Finn Lindgren \email{finn.lindgren@@gmail.com}
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
#' @importFrom Rcpp sourceCpp
#' @useDynLib inlabru, .registration = TRUE
## usethis namespace: end
NULL

#' Sample from an inhomogeneous Poisson process
#'
#' This function provides point samples from one- and two-dimensional
#' inhomogeneous Poisson processes. The log intensity has to be provided via its
#' values at the nodes of an `fm_mesh_1d` or `fm_mesh_2d` object. In between
#' mesh nodes the log intensity is assumed to be linear.
#'
#' For 2D processes on a sphere the `R` parameter can be used to adjust to
#' sphere's radius implied by the mesh. If the intensity is very high the
#' standard `strategy` "spherical" can cause memory issues. Using the
#' "sliced-spherical" strategy can help in this case.
#'
#' @export
#'
#' @param mesh An [fmesher::fm_mesh_1d] or [fmesher::fm_mesh_2d] object
#' @param loglambda vector or matrix; A vector of log intensities at the mesh
#'   vertices (for higher order basis functions, e.g. for `fm_mesh_1d` meshes,
#'   `loglambda` should be given as `mesh$m` basis function weights rather than
#'   the values at the `mesh$n` vertices) A single scalar is expanded to a
#'   vector of the appropriate length. If a matrix is supplied, one process
#'   sample for each column is produced.
#' @param strategy Only relevant for 2D meshes. One of `'triangulated'`,
#'   `'rectangle'`, `'sliced-spherical'`, `'spherical'`. The `'rectangle'`
#'   method is only valid for CRS-less flat 2D meshes.
#'   If `NULL` or `'auto'`, the the likely fastest method is chosen;
#'   `'rectangle'` for flat 2D meshes with no CRS,
#'   `'sliced-spherical'` for CRS `'longlat'` meshes, and
#'   `'triangulated'` for all other meshes.
#' @param R Numerical value only applicable to spherical and geographical
#'   meshes. It is interpreted as `R` is the equivalent Earth radius, in km,
#'   used to scale the lambda intensity. For CRS enabled meshes, the default is
#'   6371. For CRS-less spherical meshes, the default is 1.
#' @param samplers A `SpatialPolygonsDataFrame` or `fm_mesh_2d` object.
#'   Simulated points that fall outside these polygons are discarded.
#' @param ignore.CRS logical; if `TRUE`, ignore any CRS information in the mesh.
#'   Default `FALSE`. This affects `R` and the permitted values for `strategy`.
#'
#' @return A `data.frame` (1D case),
#'   or `sf` (2D flat and 3D spherical surface cases,
#'   or 2D/2.5D surface cases with multiple samples).
#'   For multiple samples, the `data.frame` output has a
#'   column `'sample'` giving the index for each sample.
#' object of point locations.
#'
#' For the old (pre version `2.13.0.9034`) `sp` output format, use `as(ret,
#' "Spatial")` or `sf::as_Spatial(ret)`.
#'
#' @details
#' * For crs-less meshes on R2: Lambda is interpreted in the raw coordinate
#'   system. Output has an NA CRS.
#' * For crs-less meshes on S2: Lambda with raw units, after scaling the mesh
#'   to radius `R`, if specified.
#'   Output is given on the same domain as the mesh, with an NA CRS.
#' * For crs meshes on R2: Lambda is interpreted as per km^2, after scaling the
#'   globe to the Earth radius 6371 km, or `R`, if specified. Output given in
#'   the same CRS as the mesh.
#' * For crs meshes on S2: Lambda is interpreted as per km^2, after scaling the
#'   globe to the Earth radius 6371 km, or `R`, if specified. Output given in
#'   the same CRS as the mesh.
#'
#' @author Daniel Simpson \email{dp.simpson@@gmail.com} (base rectangle and
#'   spherical algorithms), Fabian E. Bachl \email{bachlfab@@gmail.com}
#'   (inclusion in inlabru, sliced spherical sampling), Finn Lindgren
#'   \email{finn.lindgren@@gmail.com} (extended CRS support, triangulated
#'   sampling)
#'
#' @examples
#' \donttest{
#' # The INLA package is required
#' if (bru_safe_inla()) {
#'   vertices <- seq(0, 3, by = 0.1)
#'   mesh <- fmesher::fm_mesh_1d(vertices)
#'   loglambda <- 5 - 0.5 * vertices
#'   pts <- sample.lgcp(mesh, loglambda)
#'   pts$y <- 0
#'   plot(vertices, exp(loglambda), type = "l", ylim = c(0, 150))
#'   points(pts, pch = "|")
#' }
#' }
#'
#' \donttest{
#' # The INLA package is required
#' if (bru_safe_inla() &&
#'   require(ggplot2, quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("sf", quietly = TRUE)) {
#'   gorillas <- gorillas_sf
#'   pts <- sample.lgcp(gorillas$mesh,
#'     loglambda = 1.5,
#'     samplers = gorillas$boundary
#'   )
#'   ggplot() +
#'     gg(gorillas$mesh) +
#'     gg(pts)
#'
#'   pts <- sample.lgcp(gorillas$mesh,
#'     loglambda = base::scale(gorillas$mesh$loc) * 2,
#'     samplers = gorillas$boundary,
#'     ignore.CRS = TRUE
#'   )
#'   ggplot() +
#'     gg(gorillas$mesh) +
#'     gg(pts, aes(color = as.factor(sample))) +
#'     labs(color = "Sample")
#' }
#' }
#'
sample.lgcp <- function(mesh,
                        loglambda,
                        strategy = NULL,
                        R = NULL,
                        samplers = NULL,
                        ignore.CRS = FALSE) {
  mesh <- fm_as_fm(mesh)
  if (inherits(mesh, "fm_mesh_1d")) {
    xmin <- mesh$interval[1]
    xmax <- mesh$interval[2]
    area <- xmax - xmin

    multi.samples <- is.matrix(loglambda)
    if (multi.samples) {
      result <- list()
      n.samples <- ncol(loglambda)
      loglambda.matrix <- loglambda
    } else {
      n.samples <- 1
    }

    multi.sample <- list()
    for (sample in seq_len(n.samples)) {
      if (multi.samples) {
        loglambda <- as.vector(loglambda.matrix[, sample, drop = TRUE])
      }

      # The B-spline basis expansion has a convex hull property which implies
      # that the maximum weight is not greater or equal to the maximum function
      # value.
      wmax <- max(loglambda)

      Npoints <- rpois(1, lambda = area * exp(wmax))
      if (Npoints > 0) {
        points <- runif(n = Npoints, min = xmin, max = xmax)
        proj <- fm_basis(mesh, points, full = TRUE)
        if (length(loglambda) == 1) {
          lambda_ratio <- exp(as.vector(
            Matrix::rowSums(proj$A) * loglambda
          ) - wmax)
        } else {
          lambda_ratio <- exp(as.vector(proj$A %*% loglambda) - wmax)
        }
        keep <- (runif(Npoints) <= lambda_ratio)
        waste_ratio <- sum(keep) / length(keep)
        if (multi.samples) {
          multi.sample[[sample]] <- data.frame(
            x = points[keep],
            sample = sample
          )
        } else {
          multi.sample[[sample]] <- data.frame(x = points[keep])
        }
      } else {
        waste_ratio <- 0
        if (multi.samples) {
          multi.sample[[sample]] <- data.frame(
            x = numeric(0),
            sample = integer(0)
          )
        } else {
          multi.sample[[sample]] <- data.frame(x = numeric(0))
        }
      }
    }
    ret <- do.call(rbind, multi.sample)
  } else if (inherits(mesh, "fm_mesh_2d")) {
    multi.samples <- is.matrix(loglambda)
    if (multi.samples) {
      result <- list()
      n.samples <- ncol(loglambda)
      if (nrow(loglambda) == 1) {
        loglambda.matrix <- matrix(loglambda, mesh$n, n.samples, byrow = TRUE)
      } else {
        loglambda.matrix <- loglambda
      }
    } else {
      n.samples <- 1
      loglambda <- as.vector(loglambda)
      if (length(loglambda) == 1) {
        loglambda <- rep(loglambda, mesh$n)
      }
    }

    if (ignore.CRS) {
      mesh$crs <- NULL
    }
    input.crs <- fm_crs(mesh)
    use.crs <- !is.na(input.crs) && !ignore.CRS
    is.geocent <- fmesher::fm_crs_is_geocent(mesh)

    if (is.geocent || use.crs) {
      strategies <- c("triangulated", "sliced-spherical", "spherical")
    } else {
      strategies <- c("triangulated", "rectangle")
    }
    if (!is.null(strategy) && (strategy == "auto")) {
      strategy <- NULL
    }
    if (is.null(strategy)) {
      if (is.geocent) {
        strategy <- "triangulated"
      } else if (!use.crs) {
        strategy <- "rectangle"
      } else if (identical(sf::st_crs(input.crs)$proj, "longlat")) {
        strategy <- "sliced-spherical"
      }
    }
    strategy <- match.arg(strategy, strategies)

    if (is.geocent) {
      space.R <- mean(rowSums(mesh$loc^2)^0.5)
      internal.crs <- fm_crs("sphere")
      mesh$loc <- mesh$loc / space.R
      mesh$crs <- internal.crs
    } else {
      if (use.crs) {
        internal.crs <- fm_crs("sphere")
      } else {
        internal.crs <- fm_crs(NA_character_)
        mesh$crs <- NULL
      }
    }

    area.R <- 1
    if (!missing(R) && !is.null(R)) {
      area.R <- R
    } else {
      if (use.crs) {
        #        if (is.na(input.crs)) {
        #          space.units <- fm_length_unit(input.crs)
        #        }
        area.R <- 6371
        if (is.geocent) {
          if (abs(1 - space.R / area.R) > 1e-2) {
            warning(
              "The mesh has radius '", space.R,
              "', but crs information is available. ",
              "Using radius 6371 for area calculations."
            )
          }
        }
      } else if (is.geocent) {
        area.R <- space.R
      }
    }

    for (sample in seq_len(n.samples)) {
      if (multi.samples) {
        loglambda <- as.vector(loglambda.matrix[, sample, drop = TRUE])
      }

      if (strategy == "triangulated") {
        if (is.geocent) {
          area.mesh <- mesh
        } else if (use.crs) {
          area.mesh <- fm_transform(mesh, crs = internal.crs)
        } else {
          area.mesh <- mesh
          area.R <- 1
        }
        areas <- fm_fem(area.mesh, order = 0)$ta * area.R^2

        loglambda_tri <- matrix(
          loglambda[mesh$graph$tv],
          nrow(mesh$graph$tv),
          ncol(mesh$graph$tv)
        )
        loglambda_max <- apply(loglambda_tri, 1, max)

        Npoints <- rpois(length(areas), lambda = exp(loglambda_max) * areas)

        if (sum(Npoints) > 0) {
          triangle <- rep(seq_along(areas), Npoints)

          # Sample uniformly on base triangle (0,0),(1,0),(0,1)
          points <- matrix(runif(2 * sum(Npoints)), sum(Npoints), 2)
          flip <- rowSums(points) > 1.0
          points[flip, ] <- 1 - points[flip, 2:1, drop = FALSE]
          # Convert to in-triangle points
          points <-
            mesh$loc[mesh$graph$tv[triangle, 1], , drop = FALSE] +
            (mesh$loc[mesh$graph$tv[triangle, 2], , drop = FALSE] -
              mesh$loc[mesh$graph$tv[triangle, 1], , drop = FALSE]) *
              points[, 1] +
            (mesh$loc[mesh$graph$tv[triangle, 3], , drop = FALSE] -
              mesh$loc[mesh$graph$tv[triangle, 1], , drop = FALSE]) *
              points[, 2]
          if (is.geocent) {
            points <- points / rowSums(points^2)^0.5
          }
        }

        # Construct SpatialPoints
        if (is.geocent) {
          target.crs <- internal.crs
        } else if (use.crs) {
          target.crs <- fm_crs(input.crs)
        } else {
          target.crs <- fm_crs(NA_character_)
        }
        if (sum(Npoints) > 0) {
          points <- sf::st_as_sf(
            as.data.frame(points),
            coords = seq_len(ncol(points)),
            crs = target.crs
          )

          A <- fm_basis(mesh, points)
          lambda_ratio <- exp(as.vector(A %*% loglambda) -
            loglambda_max[triangle])
          keep <- (runif(sum(Npoints)) <= lambda_ratio)
          ret <- points[keep, , drop = FALSE]
          waste_ratio <- sum(keep) / length(keep)
        } else {
          points <- sf::st_as_sf(
            as.data.frame(matrix(0, 1, 3)),
            coords = seq_len(3),
            crs = target.crs
          )[-1, , drop = FALSE]
          ret <- points
          waste_ratio <- 0
        }
      } else if (strategy == "rectangle") {
        # Construct bounding rectangle
        loc <- mesh$loc
        xmin <- min(loc[, 1])
        xmax <- max(loc[, 1])
        ymin <- min(loc[, 2])
        ymax <- max(loc[, 2])
        area <- (xmax - xmin) * (ymax - ymin)

        # Simulate number of points
        lambda_max <- max(loglambda)
        Npoints <- rpois(1, lambda = area * exp(lambda_max))

        if (Npoints == 0) {
          ret <- sf::st_as_sf(
            as.data.frame(matrix(0, 1, 2)),
            coords = seq_len(2),
            crs = internal.crs
          )[-1, , drop = FALSE]
          waste_ratio <- 0
        } else {
          # Simulate uniform points on the bounding rectangle
          points <- sf::st_as_sf(
            data.frame(
              x = runif(n = Npoints, min = xmin, max = xmax),
              y = runif(n = Npoints, min = ymin, max = ymax)
            ),
            coords = seq_len(2),
            crs = internal.crs
          )

          # Do some thinning
          proj <- fm_basis(mesh, points, full = TRUE)
          lambda_ratio <- exp(as.vector(proj$A %*% loglambda) - lambda_max)
          keep <- proj$ok & (runif(Npoints) <= lambda_ratio)
          ret <- points[keep, , drop = FALSE]
          waste_ratio <- sum(keep) / length(keep)
        }
      } else if (strategy == "spherical") {
        # Simulate number of points
        area <- 4 * pi * area.R^2
        lambda_max <- max(loglambda)
        Npoints <- rpois(n = 1, lambda = area * exp(lambda_max))
        if (Npoints > 5000000) {
          stop(paste0("Too many points!: ", Npoints))
        }

        if (Npoints == 0) {
          ret <- sf::st_as_sf(
            as.data.frame(matrix(0, 1, 3)),
            coords = seq_len(3),
            crs = internal.crs
          )[-1, , drop = FALSE]
          waste_ratio <- 0
        } else {
          # Choose z uniformly distributed in [-1,1].
          z <- runif(n = Npoints, min = -1, max = 1)
          # Choose t uniformly distributed on [0, 2*pi).
          t <- runif(n = Npoints, min = 0, max = 2 * pi)
          r <- sqrt(1 - z^2)
          x <- r * cos(t)
          y <- r * sin(t)

          points <- sf::st_as_sf(
            as.data.frame(x, y, z),
            coords = seq_len(3),
            crs = internal.crs
          )

          proj <- fm_basis(mesh, points, full = TRUE)
          lambda_ratio <- exp(as.vector(proj$A %*% loglambda) - lambda_max)
          keep <- proj$ok & (runif(Npoints) <= lambda_ratio)
          ret <- points[keep, , drop = FALSE]
          waste_ratio <- sum(keep) / length(keep)
        }
      } else if (strategy == "sliced-spherical") {
        if (identical(input.crs$proj, "longlat")) {
          # Simulate number of points
          lon.range <- range(mesh$loc[, 1])
          lat.range <- range(mesh$loc[, 2])
          lon.rel <- diff(lon.range) / 360
          lat.rel <- diff(lat.range) / 180
        } else {
          lon.range <- c(-180, 180)
          lat.range <- c(-90, 90)
          lon.rel <- 1
          lat.rel <- 1
        }
        slat.range <- pmax(-1, pmin(1, sin(lat.range * pi / 180)))
        radlon.range <- lon.range * pi / 180
        area <- (4 * pi * area.R^2) * lon.rel * lat.rel

        lambda_max <- max(loglambda)
        Npoints <- rpois(n = 1, lambda = area * exp(lambda_max))

        sampled.points <- list()
        nblocks <- max(1, Npoints / 1000000)
        sp <- ceiling(seq(1, Npoints + 1, length.out = max(2, nblocks)))
        n.keep <- 0
        n.sampled <- 0

        for (k in seq_len(length(sp) - 1)) {
          n.points <- sp[k + 1] - sp[k]
          if (n.points == 0) {
            sampled.points[[k]] <-
              sf::st_as_sf(
                as.data.frame(matrix(0, 1, 3)),
                coords = seq_len(3),
                crs = internal.crs
              )[-1, , drop = FALSE]
            break
          }

          # Choose z uniformly distributed in [-1,1].
          # Transforms into latitude
          z <- runif(n = n.points, min = slat.range[1], max = slat.range[2])
          # Choose t uniformly distributed on [0, 2*pi).
          angle <- runif(
            n = n.points,
            min = radlon.range[1],
            max = radlon.range[2]
          )
          r <- sqrt(1 - z^2)
          x <- r * cos(angle)
          y <- r * sin(angle)

          points <- sf::st_as_sf(
            as.data.frame(x, y, z),
            coords = seq_len(3),
            crs = internal.crs
          )

          proj <- fm_basis(mesh, points, full = TRUE)
          lambda_ratio <- exp(as.vector(proj$A %*% loglambda) - lambda_max)
          keep <- proj$ok & (runif(Npoints) <= lambda_ratio)
          sampled.points[[k]] <- points[keep, , drop = FALSE]

          n.keep <- n.keep + sum(keep)
          n.sampled <- n.sampled + length(keep)
        }

        # What to return
        if (sum(lengths(sampled.points)) > 0) {
          ret <- do.call(rbind, sampled.points)
        } else {
          ret <- sampled.points[[1]]
        }
        waste_ratio <- n.keep / n.sampled
      }

      if (multi.samples) {
        result[[sample]] <- cbind(ret, sample = rep(sample, NROW(ret)))
      }
    }
    if (multi.samples) {
      ret <- do.call(rbind, result)
    }

    if (is.geocent) {
      if (NROW(ret) > 0) {
        ret <- fm_transform(ret, crs = fm_crs("globe"))
      } else if (multi.samples) {
        if (use.crs) {
          ret_crs <- fm_crs(input.crs)
        } else {
          ret_crs <- fm_crs(NA_character_)
        }
        ret <- sf::st_as_sf(
          cbind(as.data.frame(matrix(0, 1, 3)), sample = 1),
          coords = seq_len(3),
          crs = ret_crs
        )[-1, , drop = FALSE]
      } else {
        ret <- sf::st_as_sf(
          as.data.frame(matrix(0, 1, 3)),
          coords = seq_len(3),
          crs = ret_crs
        )[-1, , drop = FALSE]
      }
    } else {
      if (use.crs) {
        if (NROW(ret) > 0) {
          ret <- fm_transform(ret, input.crs)
        } else if (multi.samples) {
          ret <- sf::st_as_sf(
            cbind(as.data.frame(matrix(0, 1, 2)), sample = 1),
            coords = seq_len(2),
            crs = fm_crs(input.crs)
          )[-1, , drop = FALSE]
        } else {
          ret <- sf::st_as_sf(
            as.data.frame(matrix(0, 1, 2)),
            coords = seq_len(2),
            crs = fm_crs(input.crs)
          )[-1, , drop = FALSE]
        }
      } else {
        sf::st_crs(ret) <- fm_crs(NA_character_)
      }
    }

    # Only retain points within the samplers
    if (!is.null(samplers) && (NROW(ret) > 0)) {
      if (inherits(samplers, "fm_mesh_2d")) {
        proj <- fm_basis(samplers, ret, full = TRUE)
        ret <- ret[proj$ok, , drop = FALSE]
      } else if (inherits(samplers, "Spatial")) {
        ret_ <- sf::as_Spatial(ret)
        if (use.crs) {
          ret_ <- fm_transform(ret_, fm_crs(samplers))
        } else {
          fm_crs(ret_) <- fm_crs(NA_character_)
          fm_crs(samplers) <- fm_crs(NA_character_)
        }
        ok <- !is.na(sp::over(ret_, samplers))
        ret <- ret[ok, , drop = FALSE]
      } else {
        ret_ <- sf::st_as_sf(ret)
        if (use.crs) {
          ret_ <- fm_transform(ret_, fm_crs(samplers))
        } else {
          fm_crs(ret_) <- fm_crs(NA_character_)
          fm_crs(samplers) <- fm_crs(NA_character_)
        }
        idx <- sf::st_within(ret_, samplers)
        ok <- vapply(idx, function(x) length(x) > 0, TRUE)
        ret <- ret[ok, , drop = FALSE]
      }
    }
  } else {
    stop(paste0(
      "The `mesh` must be convertible by `fm_as_fm()` to `fm_mesh_1d` ",
      "or `fm_mesh_2d`.\n",
      "  class = c('",
      paste0(class(mesh), collapse = "', '"),
      "')"
    ))
  }

  ret
}

#' Standardise inla hyperparameter names
#'
#' The inla hyperparameter output uses parameter names that can include
#' whitespace and special characters. This function replaces those characters
#' with underscores.
#'
#' @param x character vector; names to be standardised
#' @returns A character vector with standardised names
#' @examples
#' bru_standardise_names("Precision for the Gaussian observations")
#' @seealso [bru_names()]
#' @export
#' @keywords internal
bru_standardise_names <- function(x) {
  new_names <- vapply(
    x,
    function(x) {
      gsub("\\[|\\]|[-() ]", "_", x = x, fixed = FALSE)
    },
    "name"
  )
  not_ok <- grepl("__", x = new_names)
  while (any(not_ok)) {
    new_names[not_ok] <- vapply(
      new_names[not_ok],
      function(x) {
        gsub("__", "_", x = x, fixed = FALSE)
      },
      "name"
    )
    not_ok <- grepl("__", x = new_names)
  }
  new_names
}


#' @title Extract standardised names from a bru or inla result object
#' @description
#' Extracts the names of fixed effects, random effects, and hyperparameters,
#' converted with [bru_standardise_names()]
#' @param x an `inla` or [bru()] result object
#' @returns A character vector with standardised names
#' @export
#' @examples
#' if (bru_safe_inla()) {
#'   fit <- bru(y ~ 1 + x + z(z, model = "iid"),
#'     data = data.frame(
#'       y = rnorm(10),
#'       x = rnorm(10),
#'       z = rep(seq_len(2), 5)
#'     )
#'   )
#'   bru_names(fit)
#' }
bru_names <- function(x) {
  UseMethod("bru_names")
}
#' @export
#' @rdname bru_names
bru_names.inla <- function(x) {
  bru_standardise_names(c(
    rownames(x[["summary.fixed"]]),
    names(x[["summary.random"]]),
    rownames(x[["summary.hyperpar"]])
  ))
}
#' @export
#' @rdname bru_names
bru_names.bru <- function(x) {
  bru_standardise_names(c(
    rownames(x[["summary.fixed"]]),
    names(x[["summary.random"]]),
    rownames(x[["summary.hyperpar"]])
  ))
}


inla_result_latent_idx <- function(result) {
  do.call(
    c,
    lapply(
      result$misc$configs$contents$tag,
      function(x) {
        idx <-
          list(
            result$misc$configs$contents$start[
              result$misc$configs$contents$tag == x
            ] - 1 + seq_len(
              result$misc$configs$contents$length[
                result$misc$configs$contents$tag == x
              ]
            )
          )
        names(idx) <- x
        idx
      }
    )
  )
}

#' Extract a summary property from all results of an inla result
#'
#' @param result an `inla` result object with a `bru_info` element
#' @param property character; "mean", "sd", "mode", or some other column
#' identifier for inla result `$summary.fixed`, `$summary.random$label`, and
#' `$summary.hyperpar`, or "joint_mode". For "joint_mode", the joint latent mode
#' is extracted, and the joint hyperparameter mode, in the internal scale.
#' For "predictor_sd" the posterior standard deviations of the linear predictor
#' are returned.
#' @param internal_hyperpar logical; if `TRUE`, use internal scale for
#' hyperparamter properties. Default is `FALSE`, except when `property` is
#' "joint_mode" which forces `internal_hyperpar=TRUE`.
#' @return named list for each estimated fixed effect coefficient,
#' random effect vector, and hyperparameter. The hyperparameter names are
#' standardised with [bru_standardise_names()]
#' @keywords internal
extract_property <- function(result, property,
                             internal_hyperpar = FALSE) {
  stopifnot(inherits(result, "inla"))
  ret <- list()

  if (property == "joint_mode") {
    mode_idx <- inla_result_latent_idx(result)
    for (label in c(
      rownames(result$summary.fixed),
      names(result$summary.random)
    )) {
      ret[[label]] <- result$mode$x[mode_idx[[label]]]
    }
    theta_names <- bru_standardise_names(result$mode$theta.tags)
    for (idx in seq_along(theta_names)) {
      ret[[theta_names[idx]]] <- result$mode$theta[idx]
    }
    return(ret)
  } else if (property == "predictor_sd") {
    idx <- grepl(
      pattern = "^APredictor",
      x = rownames(result$summary.linear.predictor)
    )
    ret <- result$summary.linear.predictor$sd[idx]
    return(ret)
  }

  for (label in rownames(result$summary.fixed)) {
    ret[[label]] <- result$summary.fixed[label, property]
  }

  for (label in names(result$summary.random)) {
    ret[[label]] <- result$summary.random[[label]][, property]
  }

  if (internal_hyperpar) {
    for (label in rownames(result$internal.summary.hyperpar)) {
      new.label <- bru_standardise_names(label)
      ret[[new.label]] <- result$internal.summary.hyperpar[label, property]
    }
  } else {
    for (label in rownames(result$summary.hyperpar)) {
      new.label <- bru_standardise_names(label)
      ret[[new.label]] <- result$summary.hyperpar[label, property]
    }
  }

  comp_lst <- as_bru_comp_list(result$bru_info)
  fac.names <- names(comp_lst)[
    vapply(
      comp_lst,
      function(e) {
        identical(e$type, "factor")
      },
      TRUE
    )
  ]

  # TODO: Consider whether to extract/convert effect$type == "factor" models
  # from random effects into fixed effects
  #
  # Old code:
  # For factors we add a data.frame with column names equivalent to the
  # factor levels
  #    for (name in fac.names) {
  #      tmp <- unlist(ret[startsWith(names(ret), name)])
  #      names(tmp) <- lapply(names(tmp), function(nm) {
  #        substring(nm, nchar(name) + 1)
  #      })
  #      ret[[name]] <- tmp
  #    }

  ret
}


##
## A wrapper for inla.posterior.sample()
##
## Converts each sample into a list of sub-samples representing the
## latent variables
##
## Example: samples[[1]]$somefield is the value of the field "somefield"
##          in the first sample
##

post.sample.structured <- function(result, n, seed = NULL,
                                   num.threads = NULL, ...) {
  if (!is.null(seed) && (seed != 0L)) {
    num.threads <- "1:1:1"
  }
  # Workaround for older versions of INLA
  if ("hyper.user.scale" %in% formalArgs(INLA::inla.posterior.sample)) {
    samples <- INLA::inla.posterior.sample(
      n = n,
      result = result,
      seed = seed,
      num.threads = num.threads
    )
  } else if ("parallel.configs" %in% formalArgs(INLA::inla.posterior.sample)) {
    samples <- INLA::inla.posterior.sample(
      n = n,
      result = result,
      seed = seed,
      intern = FALSE,
      num.threads = num.threads,
      parallel.configs = FALSE,
      add.names = FALSE
    )
  } else {
    samples <- INLA::inla.posterior.sample(
      n = n,
      result = result,
      seed = seed,
      intern = FALSE,
      num.threads = num.threads,
      add.names = FALSE
    )
  }

  ssmpl <- list()
  .contents <- attr(samples, ".contents")
  for (i in seq_along(samples)) {
    smpl.latent <- samples[[i]]$latent
    smpl.hyperpar <- samples[[i]]$hyperpar
    vals <- list()

    # Extract simulated predictor and fixed effects
    for (name in unique(c("Predictor", result$names.fixed))) {
      vals[[name]] <- extract_entries(name, smpl.latent, .contents = .contents)
    }

    # Extract simulated latent variables.
    # If the model is "clinear", however, we might extract the realisations
    # from the hyperpar field. TODO: check if all the special models now have
    # their results available as latent random effects, and avoid special code,
    # since the hyperpar name definition has changed
    if (length(result$summary.random) > 0) {
      for (k in seq_along(result$summary.random)) {
        name <- names(result$summary.random)[k]
        #        model <- result$model.random[k]
        #        if (!(model == "Constrained linear")) {
        vals[[name]] <- extract_entries(name,
          smpl.latent,
          .contents = .contents
        )
        #        }
        #        else {
        #         vals[[name]] <- smpl.hyperpar[glue("Beta for {name}")]
        #        }
      }
    }

    # For effects that were modeled via factors we attach an extra vector
    # holding the samples
    comp_lst <- as_bru_comp_list(result)
    fac.names <- names(comp_lst)[
      vapply(
        comp_lst,
        function(e) {
          identical(e$main$type, "factor")
        },
        TRUE
      )
    ]
    for (name in fac.names) {
      # TODO: figure out how to interact this with group and replicate info
      names(vals[[name]]) <- comp_lst[[name]]$main$values
    }

    if (length(smpl.hyperpar) > 0) {
      ## Sanitize the variable names; replace problems with "_".
      ## Needs to handle whatever INLA uses to describe the hyperparameters.
      ## Known to include " " and "-" and potentially "(" and ")".
      names(smpl.hyperpar) <- bru_standardise_names(names(smpl.hyperpar))
    }
    ssmpl[[i]] <- c(vals, smpl.hyperpar)
  }

  #
  # Return
  #
  ssmpl
}

extract_entries <- function(name, smpl, .contents = NULL) {
  if (is.null(.contents)) {
    ename <- gsub("\\.", "\\\\.", name)
    ename <- gsub("\\(", "\\\\(", ename)
    ename <- gsub("\\)", "\\\\)", ename)
    ptn <- paste("^", ename, "[\\:]*[\\.]*[0-9]*[\\.]*[0-9]*$", sep = "")
    vals <- smpl[grep(ptn, rownames(smpl))]
    return(vals)
  }
  idx <- match(name, .contents[["tag"]])
  if (is.na(idx)) {
    warning(glue("Element '{name}' not found in posterior sample."))
    return(numeric(0L))
  }
  vals <- smpl[.contents[["start"]][idx] +
    seq_len(.contents[["length"]][idx]) - 1L]
  vals
}

#' Backwards compatibility to handle mexpand for INLA <= 24.06.02
#'
#' Expand observation vectors/matrices in stacks into to a multicolumn matrix
#' for multiple likelihoods
#'
#' @export
#' @param \dots List of stacks that contain vector observations (existing
#'   multilikelihood observation matrices are also permitted)
#' @param old.names A vector of strings with the names of the observation
#'   vector/matrix for each stack. If a single string, this is assumed for all
#'   the stacks. (default "BRU.response")
#' @param new.name The name to be used for the expanded observation matrix,
#'        possibly the same as an old name. (default "BRU.response")
#' @return a list of modified stacks with multicolumn observations
#' @author Fabian E. Bachl \email{f.e.bachl@@bath.ac.uk} and Finn Lindgren
#'   \email{finn.lindgren@@gmail.com}
#'
#' @keywords internal
#' @rdname bru_inla.stack.mexpand

bru_inla.stack.mexpand <- function(...,
                                   old.names = "BRU.response",
                                   new.name = "BRU.response") {
  stacks <- list(...)
  if (length(old.names) == 1) {
    old.names <- rep(old.names, length(stacks))
  }
  y.cols <- unlist(lapply(
    seq_along(stacks),
    function(x, stacks, old.names) {
      LHS <- INLA::inla.stack.LHS(stacks[[x]])[[old.names[x]]]
      ifelse(is.vector(LHS), 1, NCOL(LHS))
    },
    stacks = stacks, old.names = old.names
  ))
  y.offset <- c(0, cumsum(y.cols))
  y.cols.total <- sum(y.cols)
  for (j in seq_along(stacks)) {
    LHS <- INLA::inla.stack.LHS(stacks[[j]])
    RHS <- INLA::inla.stack.RHS(stacks[[j]])
    A <- INLA::inla.stack.A(stacks[[j]])
    responses <- stacks[[j]][["responses"]]
    # Access the raw tag indexing information
    tags <- list(
      data = stacks[[j]]$data$index,
      effects = stacks[[j]]$effects$index
    )

    # Expand the observation vector/matrix into a multilikelihood observation
    # matrix:
    y.rows <- NROW(A)
    if (!is.null(LHS[[old.names[j]]])) {
      LHS[[new.name]] <-
        cbind(
          matrix(NA, nrow = y.rows, ncol = y.offset[j]),
          LHS[[old.names[j]]],
          matrix(NA, nrow = y.rows, ncol = y.cols.total - y.offset[j + 1])
        )
    }

    # Create the modified stack, with model compression disabled to prevent
    # modifications:
    if (utils::packageVersion("INLA") <= "24.06.02") {
      stacks[[j]] <-
        INLA::inla.stack.sum(
          data = LHS,
          A = A,
          effects = RHS,
          compress = FALSE,
          remove.unused = FALSE
        )
    } else {
      stacks[[j]] <-
        INLA::inla.stack.sum(
          data = LHS,
          A = A,
          effects = RHS,
          compress = FALSE,
          remove.unused = FALSE,
          responses = responses
        )
    }

    # Since the row indexing is unchanged, copy the tag index information:
    stacks[[j]]$data$index <- tags$data
    stacks[[j]]$effects$index <- tags$effects
  }
  stacks
}


#' @title Join stacks intended to be run with different likelihoods
#'
#' @description
#' Helper functions for multi-likelihood models
#'
#' @param \dots List of stacks that contain vector observations (existing
#'   multi-likelihood observation matrices are also permitted)
#' @param compress If `TRUE`, compress the model by removing duplicated rows of
#' effects, replacing the corresponding A-matrix columns with a single column
#' containing the sum.
#' @param remove.unused	If `TRUE`, compress the model by removing rows of
#'   effects corresponding to all-zero columns in the A matrix (and removing
#'   those columns).
#' @param old.names A vector of strings with the names of the observation
#'   vector/matrix for each stack. If a single string, this is assumed for all
#'   the stacks. (default "BRU.response")
#' @param new.name The name to be used for the expanded observation matrix,
#'        possibly the same as an old name. (default "BRU.response")
#' @export
#' @keywords internal
#' @rdname bru_inla.stack.mjoin
#'

bru_inla.stack.mjoin <- function(...,
                                 compress = TRUE,
                                 remove.unused = TRUE,
                                 old.names = "BRU.response",
                                 new.name = "BRU.response") {
  if (utils::packageVersion("INLA") <= "24.06.02") {
    stacks <- bru_inla.stack.mexpand(...,
      old.names = old.names,
      new.name = new.name
    )
    do.call(INLA::inla.stack.join, c(
      stacks,
      list(
        compress = compress,
        remove.unused = remove.unused
      )
    ))
  } else {
    stacks <- bru_inla.stack.mexpand(...,
      old.names = old.names,
      new.name = new.name
    )
    do.call(INLA::inla.stack.join, c(
      stacks,
      list(
        compress = compress,
        remove.unused = remove.unused,
        multi.family = TRUE
      )
    ))
  }
}


#' @rdname plot.bru
#' @param result an `inla` or `bru` result object
#' @param varname character; name of the variable to plot
#' @param index integer; index of the random effect to plot
#' @param link function; link function to apply to the variable
#' @param add logical; if `TRUE`, add to an existing plot
#' @param ggp logical; unused
#' @param lwd numeric; line width
#' @export
plotmarginal.inla <- function(result,
                              varname = NULL,
                              index = NULL,
                              link = function(x) {
                                x
                              },
                              add = FALSE,
                              ggp = TRUE,
                              lwd = 3,
                              ...) {
  requireNamespace("ggplot2")
  vars <- variables.inla(result)
  ovarname <- varname

  if (varname %in% c(result$names.fixed, rownames(result$summary.hyperpar)) ||
    (!is.null(index) && (varname %in% names(result$summary.random)))) {
    if (varname %in% rownames(vars) &&
      vars[varname, "type"] == "fixed") {
      marg <- INLA::inla.tmarginal(link, result$marginals.fixed[[varname]])
    } else if (varname %in% names(result$summary.random)) {
      marg <- INLA::inla.tmarginal(
        link,
        result$marginals.random[[varname]][[index]]
      )
      ovarname <- glue("{ovarname} {index}")
      if (ovarname %in% rownames(vars)) {
        if (!identical(vars[ovarname, "ID"], as.character(index - 1))) {
          # Use factor level name
          ovarname <- vars[ovarname, "ID"]
        }
      }
    } else if (varname %in% rownames(vars) &&
      vars[varname, "type"] == "hyperpar") {
      marg <- INLA::inla.tmarginal(link, result$marginals.hyperpar[[varname]])
    }
    uq <- INLA::inla.qmarginal(0.975, marg)
    uqy <- INLA::inla.dmarginal(uq, marg)
    lq <- INLA::inla.qmarginal(0.025, marg)
    lqy <- INLA::inla.dmarginal(lq, marg)
    inner.x <- seq(lq, uq, length.out = 100)
    inner.marg <- data.frame(
      x = inner.x,
      y = INLA::inla.dmarginal(inner.x, marg)
    )

    df <- data.frame(marg)
    ggplot2::ggplot(
      data = df,
      ggplot2::aes(x = .data[["x"]], y = .data[["y"]])
    ) +
      ggplot2::geom_path() +
      ggplot2::geom_ribbon(
        ymin = 0,
        ggplot2::aes(ymax = .data[["y"]]),
        alpha = 0.1
      ) +
      ggplot2::geom_segment(x = lq, y = 0, xend = lq, yend = lqy) +
      ggplot2::geom_segment(x = uq, y = 0, xend = uq, yend = uqy) +
      ggplot2::geom_ribbon(
        data = inner.marg, ymin = 0,
        ggplot2::aes(ymax = .data[["y"]]), alpha = 0.1
      ) +
      ggplot2::xlab(ovarname) +
      ggplot2::ylab("pdf")
  } else {
    df <- result$summary.random[[varname]]
    colnames(df) <- c(
      "ID", "mean", "sd", "lower", "mid", "upper", "mode", "kld"
    )
    df$mean <- link(df$mean)
    h <- 1e-6
    df$sd <- link(df$sd) * abs((link(df$mean + h) - link(df$mean - h)) /
      (2 * h))
    df$lower <- link(df$lower)
    df$mid <- link(df$mid)
    df$upper <- link(df$upper)
    df$mode <- link(df$mode)
    p <- ggplot2::ggplot(df, ggplot2::aes(.data[["ID"]], .data[["mode"]]))
    p +
      ggplot2::geom_ribbon(ggplot2::aes(
        ymin = .data[["lower"]],
        ymax = .data[["upper"]]
      ), alpha = 0.1) +
      ggplot2::geom_line() +
      ggplot2::geom_point() +
      ggplot2::geom_line(ggplot2::aes(y = .data[["mean"]]), col = 2) +
      ggplot2::geom_line(ggplot2::aes(y = .data[["mid"]]), col = 2, lty = 2) +
      ggplot2::ylab("mode and quantiles") +
      ggplot2::xlab(glue("{varname} ID"))
  }
}

variables.inla <- function(result, include.random = TRUE) {
  handle.missing <- function(col.names) {
    cbind(data.frame(
      type = character(0),
      model = character(0),
      as.data.frame(
        matrix(NA, 0, length(col.names),
          dimnames = list(c(), col.names)
        )
      )
    ))
  }

  handle.missing.columns <- function(data, col.names) {
    missing.names <- setdiff(col.names, colnames(data))
    if (length(missing.names) > 0) {
      df <- as.data.frame(matrix(NA, nrow(data), length(missing.names),
        dimnames = list(NULL, missing.names)
      ))
      data <- dplyr::bind_cols(data, df)
    }
    data
  }

  handle.data.frame <- function(data, type, model, col.names) {
    ##    rownames(data) ???
    cbind(
      data.frame(
        type = rep(type, nrow(data)),
        model = rep(model, nrow(data))
      ),
      handle.missing.columns(data, col.names)
    )
  }

  ## Get column names, handling possibly empty output:
  fixed.missing <- (is.null(result$summary.fixed) ||
    (ncol(result$summary.fixed) == 0))
  hyperpar.missing <- (is.null(result$summary.hyperpar) ||
    (ncol(result$summary.hyperpar) == 0))
  random.missing <- (is.null(result$summary.random) ||
    (length(result$summary.random) == 0) ||
    (ncol(result$summary.random[[1]]) == 0))
  col.names <- c()
  if (!fixed.missing) {
    col.names <- union(col.names, colnames(result$summary.fixed))
  }
  if (!hyperpar.missing) {
    col.names <- union(col.names, colnames(result$summary.hyperpar))
  }
  if (!random.missing) {
    col.names <- union(col.names, colnames(result$summary.random[[1]]))
  }
  if (fixed.missing) {
    fixed <- handle.missing(col.names)
  } else {
    fixed <- handle.data.frame(
      result$summary.fixed,
      "fixed", "fixed", col.names
    )
  }
  if (hyperpar.missing) {
    hyperpar <- handle.missing(col.names)
  } else {
    hyperpar <- handle.data.frame(
      result$summary.hyperpar,
      "hyperpar", NA, col.names
    )
  }
  if (random.missing) {
    random <- list(handle.missing(col.names))
  } else {
    if (!include.random) {
      random <-
        lapply(
          seq_along(result$summary.random),
          function(x) {
            dat <- data.frame(
              mean = mean(result$summary.random[[x]]$mean),
              sd = mean(result$summary.random[[x]]$sd)
            )
            output <- handle.data.frame(
              dat,
              "random",
              result$model.random[x],
              col.names
            )
            rownames(output) <-
              names(result$summary.random)[x]
            output
          }
        )
    } else {
      random <-
        lapply(
          seq_along(result$summary.random),
          function(x) {
            output <- handle.data.frame(
              result$summary.random[[x]],
              "random",
              result$model.random[x],
              col.names
            )
            rownames(output) <-
              paste(
                names(result$summary.random)[x],
                seq_len(nrow(result$summary.random[[x]]))
              )
            output
          }
        )
    }
  }

  variables <- do.call(rbind, c(list(fixed, hyperpar), random))
  variables
}

#' @title Predictor expression handler object
#'
#' @description
#' Create and access predictor expression information, keeping track of what
#' latent variables are used, whether the predictor is additive/linear/rowwise.
#'
#' @param x For creation, a formula or character string giving the predictor
#' expression. If the character string is `"."`, it is taken to mean an
#' additive expression including all latent variables. For access, an object
#' containing a `bru_pred_expr` object.
#' @param \dots Passed on to submethods
#'
#' @keywords internal
#' @export
#' @rdname bru_pred_expr
#' @name bru_pred_expr
NULL

#' @describeIn bru_pred_expr Create a `bru_pred_expr` object from a formula or
#' character string.
#' @param used An optional `bru_used` object, overriding the automated
#'  detection. Default: `NULL`
#' @param is_rowwise logical; whether the predictor can be assumed to use the
#' input data rowwise, so that blockwise evaluation is possible. If `NULL`, it
#' needs to be set before use. Default: `NULL`
#' @param .envir The environment in which to evaluate the expression.
#' @export
#' @examples
#' (new_bru_pred_expr(~ x + z + Intercept))
#'
new_bru_pred_expr <- function(x, ..., used = NULL, is_rowwise = NULL,
                              .envir = parent.frame()) {
  pred_text <- NULL
  resp_text <- NULL
  if (is.null(x)) {
    is_additive <- NA
    is_additive_dot <- NA
  } else {
    if (inherits(x, "formula")) {
      form.envir <- environment(x)
      if (!is.null(form.envir)) {
        .envir <- form.envir
      }
      formula_text <- deparse1(x, collapse = "\n")
      x <- as.character(x)
      pred_text <- x[length(x)]
      if (length(x) > 2) {
        resp_text <- x[2]
      }
    } else {
      pred_text <- x
      formula_text <- glue::glue("~ {pred_text}")
    }
    is_additive <- bru_is_additive(pred_text)
    is_additive_dot <- identical(pred_text, ".")
    if (is_additive_dot) {
      pred_text <- NULL
    }
  }
  if (is.null(pred_text)) {
    pred_expr <- NULL
  } else {
    pred_expr <- rlang::parse_expr(pred_text)
  }
  if (is.null(resp_text)) {
    resp_expr <- NULL
  } else {
    resp_expr <- rlang::parse_expr(resp_text)
  }
  structure(
    list(
      # TODO: Encode type of expression construction; full expression (expr),
      # additive (additive), list of effects and/or latent state vectors (list),
      # the latter also needing "hyper" information, so that generate.bru
      # can rely on this class for all use cases.
      type = "expr", # One of "expr", "additive", "list"
      pred_text = pred_text,
      pred_expr = pred_expr,
      is_additive = is_additive,
      # Also depends on component defs, so may be changed later:
      is_linear = is_additive,
      is_rowwise = is_rowwise,
      used = if (is.null(used)) bru_used(pred_text) else used,
      .envir = .envir,
      resp_text = resp_text,
      resp_expr = resp_expr
    ),
    class = "bru_pred_expr"
  )
}

bru_compat_pre_2_14_bru_obs <- function(lh) {
  if (!isTRUE(bru_options_get("bru_compat_pre_2_14_enable"))) {
    return(lh)
  }
  lh$is_additive <- lh$pred_expr$is_additive
  # Also depends on component defs, so may be changed later:
  lh$linear <- lh$pred_expr$is_linear
  lh$expr_text <- lh$pred_expr$pred_text
  lh$expr <- lh$pred_expr$pred_expr
  lh$used <- lh$pred_expr$used
  lh$allow_combine <- !isTRUE(lh$pred_expr$is_rowwise)
  lh
}

#' @describeIn bru_pred_expr Accessor generic for `bru_pred_expr` objects,
#'   including ones stored inside other objects.
#'
#' @keywords internal
#' @export
bru_pred_expr <- function(x, ...) {
  UseMethod("bru_pred_expr")
}


#' @describeIn bru_pred_expr Accessor for the `bru_pred_expr` object stored
#' inside a `bru_obs` object.
#' @export
bru_pred_expr.bru_obs <- function(x, ...) {
  bru_pred_expr(x[["pred_expr"]], ...)
}

#' @describeIn bru_pred_expr Accessor for the `bru_pred_expr` objects stored
#' inside a `bru_obs_list` object.
#' @export
bru_pred_expr.bru_obs_list <- function(x, ...) {
  lapply(x, function(y) bru_pred_expr(y[["pred_expr"]], ...))
}

#' @describeIn bru_pred_expr Accessor for the `bru_pred_expr` object stored
#' inside a `bru_info` object.
#' @export
bru_pred_expr.bru_info <- function(x, ...) {
  bru_pred_expr(as_bru_obs_list(x), ...)
}

#' @describeIn bru_pred_expr Accessor for the `bru_pred_expr` object stored
#' inside a `bru` object.
#' @export
bru_pred_expr.bru <- function(x, ...) {
  bru_pred_expr(as_bru_obs_list(x), ...)
}

#' @describeIn bru_pred_expr Access a `bru_pred_expr` object or convert it to
#' text or expression object.
#' @param format Character; one of
#' * `"object"` (the default),
#' * `"text"` (plain text),
#' * `"quo"` (an `rlang` quosure),
#' * `"expr"` (an `rlang` expression),
#' * `"formula"` (the original formula input if available, otherwise
#'   constructed from the expression),
#' * `"formula_text"` (a text version of the "formula"),
#' * `"text_raw"` (plain text, without substituting `BRU_EXPRESSION`), or
#' * `"resp_text"` (plain text of the response side of the formula).
#' @export
bru_pred_expr.bru_pred_expr <- function(x, ..., format = "object") {
  format <- match.arg(
    format,
    c(
      "object",
      "text",
      "quo",
      "expr",
      "formula",
      "formula_text",
      "text_raw",
      "resp_text"
    )
  )
  if (identical(format, "object")) {
    return(x)
  }

  pred_text <- x[["pred_text"]]
  if (is.null(pred_text)) {
    pred_text <- "BRU_EXPRESSION"
  }
  if (!identical(format, "text_raw") &&
    grepl(
      pattern = "BRU_EXPRESSION",
      x = pred_text,
      fixed = TRUE
    )) {
    included <- bru_used(x)[["effect"]]
    pred_text <-
      gsub(
        pattern = "BRU_EXPRESSION",
        replacement = paste0(included, collapse = " + "),
        x = pred_text,
        fixed = TRUE
      )
  }

  switch(format,
    text = pred_text,
    quo = rlang::parse_quo(pred_text, env = x[[".envir"]]),
    expr = rlang::parse_expr(pred_text),
    formula = {
      formula_text <- bru_pred_expr(x, format = "formula_text")
      as.formula(formula_text, env = x[[".envir"]])
    },
    formula_text = if (is.null(x[["formula_text"]])) {
      if (is.null(pred_text) || identical(pred_text, "")) {
        p_txt <- "."
      } else {
        p_txt <- pred_text
      }
      if (!is.null(x[["resp_text"]])) {
        glue::glue("{x$resp_text} ~ {p_txt}")
      } else {
        glue::glue("~ {p_txt}")
      }
    } else {
      x[["formula_text"]]
    },
    text_raw = pred_text,
    resp_text = x[["resp_text"]],
    stop(glue::glue("Unknown format '{format}'"))
  )
}


#' @rdname bru_pred_expr
#' @export
#' @importFrom glue glue
format.bru_pred_expr <- function(x, ...) {
  glue::glue(
    "    Predictor: {predictor}\n",
    "    Additive/Linear/Rowwise: {is_additive}/{is_linear}/{is_rowwise}\n",
    "    Used components: {format(used)}",
    predictor = bru_pred_expr(x, format = "formula_text"),
    is_additive = bru_is_additive(x),
    is_linear = bru_is_linear(x),
    is_rowwise = bru_is_rowwise(x),
    used = bru_used(x)
  )
}

#' @rdname bru_pred_expr
#' @export
#' @importFrom glue glue
print.bru_pred_expr <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

#' @include deprecated.R

#' @title Load INLA safely for examples and tests
#'
#' @description
#' Loads the INLA package with `requireNamespace("INLA", quietly = TRUE)`, and
#' optionally checks and sets the multicore `num.threads` INLA option.
#'
#' @param multicore logical; if `TRUE`, multiple cores are allowed, and the
#' INLA `num.threads` option is not checked or altered.
#' If `FALSE`, forces `num.threads="1:1:1"`. Default: NULL, checks
#' if running in testthat or non-interactively, in which case sets
#' `multicore=FALSE`, otherwise `TRUE`.
#' @param quietly logical; if `FALSE` and `multicore` is `FALSE`,
#' prints a message if the `num.threads` option
#' isn't already "1.1:1" to alert the user to the change.
#' Default: FALSE.
#' @param minimum_version character; the minimum required INLA version.
#' Default 23.1.31 (should always match the requirement in the package
#' DESCRIPTION)
#' @export
#' @return logical; `TRUE` if INLA was loaded safely, otherwise FALSE
#'
#' @examples
#' \dontrun{
#' if (bru_safe_inla()) {
#'   # Run inla dependent calculations
#' }
#' }
#'
#' @keywords internal
bru_safe_inla <- function(multicore = NULL,
                          quietly = FALSE,
                          minimum_version = "23.1.31") {
  inla_version <-
    check_package_version_and_load(
      pkg = "INLA",
      minimum_version = minimum_version,
      quietly = quietly
    )
  if (is.na(inla_version)) {
    return(FALSE)
  }

  inla.call <- tryCatch(
    INLA::inla.getOption("inla.call"),
    error = function(e) {
      e
    }
  )
  if (inherits(inla.call, "error")) {
    if (!quietly) {
      message(
        "inla.getOption('inla.call') failed. INLA not installed correctly."
      )
    }
    return(FALSE)
  }
  if (all(grepl("\\.run$", inla.call))) {
    inla.binary <- gsub("\\.run$", "", inla.call)
    if (!file.exists(inla.binary)) {
      if (!quietly) {
        message(
          paste0(
            "INLA binary '", inla.binary, "' not found. ",
            "INLA not installed correctly, or with platform mismatch."
          )
        )
      }
      return(FALSE)
    }
  }

  if (is.null(multicore)) {
    multicore <-
      interactive() && !identical(Sys.getenv("TESTTHAT"), "true")
  }
  if (!multicore) {
    n.t <- tryCatch(
      INLA::inla.getOption("num.threads"),
      error = function(e) {
        e
      }
    )
    if (inherits(n.t, "error")) {
      if (!quietly) {
        message("inla.getOption() failed. INLA not installed correctly.")
      }
      return(FALSE)
    }
    if (!identical(n.t, "1:1:1")) {
      if (!quietly) {
        message(paste0(
          "Changing INLA option num.threads from '", n.t, "' to '1:1:1'."
        ))
      }
      INLA::inla.setOption(num.threads = "1:1:1")
    }
  }
  TRUE
}





check_package_version_and_load <-
  function(pkg, minimum_version, quietly = FALSE) {
    version <- tryCatch(utils::packageVersion(pkg),
      error = function(e) NA_character_
    )
    if (is.na(version)) {
      if (!quietly) {
        message(paste0("Package '", pkg, "' is not installed."))
      }
      return(NA_character_)
    }
    if (version < minimum_version) {
      if (!quietly) {
        message(paste0(
          "Installed '", pkg, "' version is ", version, " but ",
          "version >= ", minimum_version, " is required."
        ))
      }
      return(NA_character_)
    }
    if (!requireNamespace(pkg, quietly = TRUE)) {
      if (!quietly) {
        message("Package '", pkg, "' not loaded safely.")
      }
      return(NA_character_)
    }
    version
  }


#' Check for potential `sp` version compatibility issues
#'
#' Loads the sp package with `requireNamespace("sp", quietly = TRUE)`, and
#' checks and optionally sets the `sp` evolution status flag if `rgdal` is
#' unavailable.
#'
#' @param quietly logical; if `TRUE`, prints diagnostic messages. Default
#'   `FALSE`
#' @param force logical; If `rgdal` is unavailable
#' and evolution status is less that `2L`, return `FALSE` if `force` is `FALSE`.
#' If `force` is `TRUE`, return `TRUE` if the package configuration is safe,
#' potentially after forcing the evolution status to `2L`.
#' Default `FALSE`
#' @param minimum_version character; the minimum required version.
#' Default 2.1 (should always match the requirement in the package
#' DESCRIPTION)
#' @return Returns (invisibly) `FALSE` if a potential issue is detected, and
#'   give a message if `quietly` is `FALSE`. Otherwise returns `TRUE`
#' @export
#' @examples
#' \dontrun{
#' if (bru_safe_sp()) {
#'   # Run sp dependent calculations
#' }
#' }
#'
#' @keywords internal
bru_safe_sp <- function(quietly = FALSE,
                        force = FALSE,
                        minimum_version = "2.1") {
  sp_version <-
    check_package_version_and_load(
      pkg = "sp",
      minimum_version = minimum_version,
      quietly = quietly
    )
  if (is.na(sp_version)) {
    return(invisible(FALSE))
  }

  if (sp_version >= "2.1.4") {
    return(invisible(TRUE))
  }

  if (sp_version >= "1.6-0") {
    # Default to 2L to allow future sp to stop supporting
    # get_evolution_status; assume everything is fine if it fails.
    evolution_status <- tryCatch(
      sp::get_evolution_status(),
      error = function(e) 2L,
      warning = function(e) 2L
    )
    rgdal_version <- tryCatch(utils::packageVersion("rgdal"),
      error = function(e) NA_character_
    )
    if ((evolution_status < 2L) && is.na(rgdal_version)) {
      if (!quietly) {
        message(
          "'sp' version >= 1.6-0 detected, rgdal isn't installed, and ",
          "evolution status is < 2L."
        )
      }
      if (!force) {
        if (!quietly) {
          message(
            "This may cause issues with some CRS handling code. ",
            "To avoid this, use 'sp::set_evolution_status(2L)'"
          )
        }
        return(invisible(FALSE))
      }

      sp::set_evolution_status(2L)
      if (!quietly) {
        message(
          "Ran 'sp::set_evolution_status(2L)' to avoid ",
          "issues with some CRS handling code."
        )
      }
    }
  }
  invisible(TRUE)
}


#' @title Check for potential `terra` installation issues
#'
#' @description
#' Loads the terra package with `requireNamespace("terra", quietly = TRUE)`,
#' and checks whether `eval_spatial()` successfully extracts non-NA values.
#' NA values may appear when there is a GDAL/PROJ installation issue that causes
#' it to not find the `proj.db` database file.
#'
#' For use in package tests and examples that depend on `terra`.
#'
#' @param quietly logical; if `TRUE`, prints diagnostic messages. Default
#'   `FALSE`
#' @param minimum_version character; the minimum required version.
#' Default 1.7-66 (should always match the requirement in the package
#' DESCRIPTION)
#' @return Returns (invisibly) `FALSE` if a potential issue is detected, and
#'   give a message if `quietly` is `FALSE`. Otherwise returns `TRUE`
#' @export
#' @examples
#' \dontrun{
#' if (bru_safe_terra()) {
#'   # Run terra dependent calculations
#' }
#' }
#'
#' @keywords internal
bru_safe_terra <- function(quietly = FALSE,
                           minimum_version = "1.7-66") {
  terra_version <-
    check_package_version_and_load(
      pkg = "terra",
      minimum_version = minimum_version,
      quietly = quietly
    )
  if (is.na(terra_version)) {
    return(invisible(FALSE))
  }

  gcov <- gorillas_sf_gcov()
  where <- inlabru::gorillas_sf$nests[seq_len(5), , drop = FALSE]
  values <- eval_spatial(gcov, where = where)
  if (anyNA(values)) {
    if (!quietly) {
      message(
        paste0(
          "Potential issue detected with 'terra' package spatial evaluation,\n",
          "likely due to GDAL/PROJ installation issues."
        )
      )
    }
    return(invisible(FALSE))
  }

  invisible(TRUE)
}


#' Expand labels
#'
#' @param labels character vector; original labels
#' @param expand character vector; subset of labels to expand
#' @param suffix character; the suffix to add to the labels selected by `expand`
#' @return a vector of labels with suffix appended to the selected labels
#' @keywords internal
expand_labels <- function(labels, expand, suffix) {
  labels[labels %in% expand] <- paste0(labels[labels %in% expand], suffix)
  labels
}

extract_matrixlist_column <- function(thelist, col) {
  vapply(
    names(thelist),
    function(x) {
      list(as.vector(thelist[[x]][, col]))
    },
    list(1)
  )
}

extract_vectorlist_column <- function(thelist) {
  vapply(
    names(thelist),
    function(x) {
      list(as.vector(thelist[[x]]))
    },
    list(1)
  )
}


check_layer <- function(data, where, layer) {
  # This works for both SpatialGrid/PixelDataFrame and SpatRaster
  names_data <- names(data)
  if (is.character(layer)) {
    unique_layer <- unique(layer)
    if (!all((unique_layer %in% names_data))) {
      stop(
        paste0(
          "Input layer name(s) '",
          paste0(unique_layer, collapse = "', '"),
          "' mismatch with available variable names.\n",
          "Available names are '",
          paste0(names_data, collapse = "', '"),
          "'.\n",
          "Use *_layer for the input component to specify a valid name."
        )
      )
    }
  } else if (is.numeric(layer)) {
    ok_layer <- (layer >= 1) & (layer <= length(names_data))
    if (!all(ok_layer)) {
      stop(
        paste0(
          "Input layer(s) nr ",
          paste0(unique(layer[!ok_layer]), collapse = ", "),
          " not in the valid range, [",
          1, ", ", length(names_data), "]"
        )
      )
    }
  } else {
    stop(
      paste0(
        "Unable to identify the spatial data frame layer to evaluate.\n",
        "Available names are '",
        paste0(names(data), collapse = "', '"),
        "'.\n",
        "Use *_layer for the input component to specify a valid name."
      )
    )
  }
  TRUE
}


extract_selector <- function(where, selector) {
  if (is.null(selector)) {
    return(NULL)
  }
  if (inherits(where, "SpatVector")) {
    requireNamespace("terra")
    layer <- terra::values(where)[[selector]]
  } else {
    layer <- where[[selector]]
  }
  if (is.null(layer)) {
    stop(
      "'selector' is non-null, but no such label found in the 'where' object"
    )
  }
  layer
}

extract_layer <- function(where, layer, selector) {
  if (!is.null(layer)) {
    if (!is.null(selector)) {
      warning("Both layer and selector specified. Ignoring selector",
        immediate. = TRUE
      )
    }
    selector <- NULL
  } else if (!is.null(selector)) {
    layer <- extract_selector(where, selector)
  } else if (is.null(layer)) {
    layer <- 1
  }
  if (length(layer) == 1) {
    layer <- rep(layer, NROW(where))
  }
  layer
}


#' Evaluate spatial covariates
#'
#' @param data Spatial data
#' @param where Where to evaluate the data
#' @param layer Which `data` layer to extract (as integer or character).
#' May be a vector, specifying a separate layer for each `where` item.
#' @param selector The name of a variable in `where` specifying the `layer`
#' information.
#'
#' @export
eval_spatial <- function(data, where, layer = NULL, selector = NULL) {
  UseMethod("eval_spatial")
}


#' @export
#' @describeIn eval_spatial Compatibility wrapper for `eval_spatial.sf`
eval_spatial.SpatialPolygonsDataFrame <- function(data,
                                                  where,
                                                  layer = NULL,
                                                  selector = NULL) {
  eval_spatial(
    sf::st_as_sf(data),
    where = where,
    layer = layer,
    selector = selector
  )
}

#' @export
#' @rdname eval_spatial
eval_spatial.SpatialPixelsDataFrame <- function(data,
                                                where,
                                                layer = NULL,
                                                selector = NULL) {
  eval_spatial_Spatial(
    data = data,
    where = where,
    layer = layer,
    selector = selector
  )
}

#' @export
#' @rdname eval_spatial
eval_spatial.SpatialGridDataFrame <- function(data,
                                              where,
                                              layer = NULL,
                                              selector = NULL) {
  eval_spatial_Spatial(
    data = data,
    where = where,
    layer = layer,
    selector = selector
  )
}

eval_spatial_Spatial <- function(data, where, layer = NULL, selector = NULL) {
  stopifnot(inherits(
    data,
    c(
      "SpatialPixelsDataFrame",
      "SpatialGridDataFrame"
    )
  ))
  if (inherits(where, "sf")) {
    where <- sf::as_Spatial(where)
  }

  if (inherits(where, "SpatialPoints")) {
    where <- fm_transform(where, crs = fm_CRS(data), passthrough = TRUE)
    if (ncol(sp::coordinates(where)) >= 3) {
      where <- sp::SpatialPoints(
        coords = sp::coordinates(where)[, 1:2, drop = FALSE],
        proj4string = fm_CRS(where)
      )
    }
  }
  layer <- extract_layer(where, layer, selector)
  check_layer(data, where, layer)
  unique_layer <- unique(layer)
  if (length(unique_layer) == 1) {
    val <- sp::over(
      where,
      data
    )[, unique_layer, drop = TRUE]
  } else {
    val <- numeric(NROW(where))
    for (l in unique(layer)) {
      val[layer == l] <- sp::over(
        where[layer == l, , drop = FALSE],
        data
      )[, l, drop = TRUE]
    }
  }
  val
}


#' @export
#' @describeIn eval_spatial Supports point-in-polygon information lookup.
#' Other combinations are untested.
eval_spatial.sf <- function(data, where, layer = NULL, selector = NULL) {
  if (!inherits(where, c("sf", "sfc", "sfg"))) {
    where <- sf::st_as_sf(where)
  }
  where <- fm_transform(where, crs = sf::st_crs(data), passthrough = TRUE)

  layer <- extract_layer(where, layer, selector)
  check_layer(data, where, layer)
  unique_layer <- unique(layer)
  data_example <- data[1, unique_layer[1], drop = TRUE]
  data_NA <- data_example
  is.na(data_NA) <- TRUE
  if (length(unique_layer) == 1) {
    idx <- sf::st_intersects(where, data, sparse = TRUE)
    val <- vapply(
      idx,
      function(i) {
        if (is.null(i) || (length(i) == 0)) {
          data_NA
        } else {
          data[min(i), unique_layer, drop = TRUE]
        }
      },
      data_example
    )
  } else {
    val <- numeric(NROW(where))
    idx <- sf::st_intersects(where, data, sparse = TRUE)
    for (l in unique(layer)) {
      val[layer == l] <- vapply(
        idx[layer == l],
        function(i) {
          if (is.null(i) || (length(i) == 0)) {
            data_NA
          } else {
            data[min(i), unique_layer, drop = TRUE]
          }
        },
        data_example
      )
    }
  }
  val
}



terra_factor_levels <- function(data) {
  layers <- names(data)
  is_factor <- terra::is.factor(data)
  if (!any(is_factor)) {
    return(NULL)
  }
  if (!all(is_factor)) {
    stop(
      glue::glue(
        "Some layers are factors ",
        "({paste(sort(layers[is_factor]), collapse = ', ')})",
        " and some are not ",
        "({paste(sort(layers[!is_factor]), collapse = ', ')})."
      )
    )
  }
  factor_levels <- lapply(
    terra::levels(data),
    function(l) {
      if (is.data.frame(l)) {
        l[, 2]
      } else {
        NULL
      }
    }
  )
  unique_levels_idx <- which(!duplicated(factor_levels))
  factor_levels <- factor_levels[unique_levels_idx]
  layers <- layers[unique_levels_idx]
  if (length(unique_levels_idx) > 1) {
    stop(
      glue::glue(
        "Factor levels differ between layers. ",
        "Found factor levels:\n",
        glue::glue_collapse(
          glue::glue("{layers}: {factor_levels}", sep = ", "),
          sep = "\n"
        )
      )
    )
  }

  factor_levels[[1]]
}


#' @export
#' @rdname eval_spatial
eval_spatial.SpatRaster <- function(data,
                                    where,
                                    layer = NULL,
                                    selector = NULL) {
  requireNamespace("terra")
  layer <- extract_layer(where, layer, selector)
  check_layer(data, where, layer)
  if (!inherits(where, "SpatVector")) {
    handle_crs <- (!fm_crs_is_null(fm_crs(where)) &&
      !fm_crs_is_null(fm_crs(data)))
    if (handle_crs) {
      where <- fmesher::fm_transform(where, fm_crs(data))
    }
    where <- terra::vect(where)
    if (handle_crs) {
      # Terra's data crs might not be detected as the same as fm_crs(data)
      where <- terra::project(where, data)
    }
  } else {
    if (!fm_crs_is_null(fm_crs(where)) &&
      !fm_crs_is_null(fm_crs(data))) {
      where <- terra::project(where, data)
    }
  }
  val <- terra::extract(
    data,
    where,
    ID = FALSE,
    layer = layer
  )
  val <- val[["value"]]

  layer_set <- sort(unique(layer))
  factor_levels <- terra_factor_levels(data[[layer_set]])
  if (!is.null(factor_levels)) {
    if (is.factor(val)) {
      val <- as.character(val)
      levs <- factor_levels
    } else {
      levs <- seq_along(factor_levels)
    }
    val <- factor(val, levels = levs, labels = factor_levels)
  }
  val
}


#' @export
#' @rdname eval_spatial
eval_spatial.stars <- function(data, where, layer = NULL, selector = NULL) {
  layer <- extract_layer(where, layer, selector)
  check_layer(data, where, layer)
  if (!inherits(where, c("sf", "sfc"))) {
    where <- sf::st_as_sf(where)
  }
  val <- stars::st_extract(
    data,
    sf::st_geometry(where)
  )
  val <- sf::st_as_sf(val)

  unique_layer <- unique(layer)
  if (length(unique_layer) == 1) {
    val <- val[, unique_layer, drop = TRUE]
  } else {
    val_ <- numeric(NROW(where))
    for (l in unique(layer)) {
      val_[layer == l] <- val[layer == l, l, drop = TRUE]
    }
    val <- val_
  }

  val
}


#' Fill in missing values in Spatial grids
#'
#' Computes nearest-available-value imputation for missing values in space
#'
#' @param data A SpatialPointsDataFrame, SpatialPixelsDataFrame,
#' SpatialGridDataFrame, SpatRaster, Raster, or sf object
#' containing data to use for filling
#' @param where A, matrix, data.frame, or SpatialPoints or
#'   SpatialPointsDataFrame, or sf object, containing the locations of the
#'   evaluated values
#' @param values A vector of values to be filled in where `is.na(values)` is
#' `TRUE`
#' @param layer,selector Specifies what data column or columns from which to
#' extract data, see [bru_comp()] for details.
#' @param batch_size `r lifecycle::badge("deprecated")` due to improved
#'   algorithm.
#' Size of nearest-neighbour calculation blocks, to limit the
#' memory and computational complexity.
#' @return An infilled vector of values
#' @export
#' @examples
#' \dontrun{
#' if (require("sf", quietly = TRUE)) {
#'   points <-
#'     sf::st_as_sf(
#'       data.frame(
#'         x = 1:3,
#'         y = 4:6,
#'         val = c(NA, NA, NA)
#'       ),
#'       coords = c("x", "y")
#'     )
#'   input_coord <- expand.grid(x = 0:7, y = 0:7)
#'   input <-
#'     sf::st_as_sf(
#'       cbind(input_coord, val = as.vector(input_coord$y)),
#'       coords = c("x", "y")
#'     )
#'   points$val <- bru_fill_missing(input, points, points$val)
#'   print(points)
#'
#'   # To fill in missing values in a grid:
#'   print(input$val[c(3, 30)])
#'   input$val[c(3, 30)] <- NA # Introduce missing values
#'   input$val <- bru_fill_missing(input, input, input$val)
#'   print(input$val[c(3, 30)])
#' }
#' }
bru_fill_missing <- function(data, where, values,
                             layer = NULL, selector = NULL,
                             batch_size = deprecated()) {
  stopifnot(inherits(
    data,
    c(
      "SpatialPointsDataFrame",
      "SpatialPixelsDataFrame",
      "SpatialGridDataFrame",
      "Raster",
      "SpatRaster",
      "sf"
    )
  ))
  # Convert to sf and terra
  if (inherits(data, c("SpatialGrid", "SpatialPixelsDataFrame", "Raster"))) {
    requireNamespace("terra")
    data <- terra::rast(data)
  } else if (inherits(data, "SpatialPointsDataFrame")) {
    data <- sf::st_as_sf(data)
  }
  # Convert to sf
  if (inherits(where, c("Spatial", "SpatVector"))) {
    where <- sf::st_as_sf(where)
  } else if (!inherits(where, "sf")) {
    where <- sf::st_as_sf(where, coords = seq_len(min(3, NCOL(where))))
  }

  layer <- extract_layer(where, layer, selector)
  check_layer(data, where, layer)

  if (anyNA(layer)) {
    stop("NAs detected in the 'layer' information.")
  }

  layers <- unique(layer)
  if (length(layers) > 1) {
    for (l in layers) {
      values[layer == l] <-
        bru_fill_missing(
          data = data,
          where = where[layer == l, , drop = FALSE],
          values = values[layer == l],
          layer = l,
          batch_size = batch_size
        )
    }
    return(values)
  }

  # Only one layer from here on.
  layer <- layers

  data_crs <- fm_crs(data)
  if (inherits(data, "SpatRaster")) {
    requireNamespace("terra")
    data_values <- terra::values(
      data[[layer]],
      dataframe = TRUE,
      na.rm = TRUE
    )[[layer]]
    data_coord <- as.data.frame(terra::crds(data[[layer]], na.rm = TRUE))
    data_coord <- sf::st_as_sf(data_coord,
      coords = seq_len(ncol(data_coord)),
      crs = data_crs
    )
  } else {
    data_values <- data[[layer]]
    data_coord <- data
  }
  data_notok <- is.na(data_values)
  if (all(data_notok)) {
    stop("All data values are NA; unable to fill missing values.")
  }
  data_values <- data_values[!data_notok]
  if (inherits(data_coord, "sf")) {
    data_coord <- data_coord[!data_notok, , drop = FALSE]
  } else {
    data_coord <- data_coord[!data_notok, , drop = FALSE]
  }

  where <- fm_transform(where, crs = data_crs, passthrough = TRUE)

  values_notok <- is.na(values)

  nn <- sf::st_nearest_feature(
    where[values_notok, , drop = FALSE],
    data_coord
  )

  values[values_notok] <- data_values[nn]

  values
}


# Resave data
resave_package_data <- function() {
  name_list <- c(
    "gorillas_sf", "mexdolphin_sf",
    "mrsea",
    "Poisson1_1D", "Poisson2_1D", "Poisson3_1D",
    "shrimp", "toygroups",
    "robins_subset",
    "toypoints"
  )
  compress_values <- c("gzip", "bzip2", "xz")

  do_compress <- function(env, compress, the_path) {
    if (length(names(env)) == 1) {
      thetext <- paste0(
        "usethis::use_data(",
        paste0(names(env), collapse = ", "),
        ", compress = '", compress, "', overwrite = TRUE)"
      )
    } else {
      thetext <- paste0(
        "save(",
        paste0(names(env), collapse = ", "),
        ", file = '",
        the_path,
        "', compress = '", compress, "')"
      )
    }
    eval(
      parse(text = thetext),
      envir = env
    )
  }

  new_info <- NULL
  old_info <- NULL
  for (name in name_list) {
    message(paste0("Data: ", name))
    env <- new.env()
    utils::data(list = name, package = "inlabru", envir = env)

    # Find paths
    the_path <- file.path("data", paste0(name, ".rda"))

    old_info <- rbind(old_info, file.info(the_path))

    smallest_size <- Inf
    smallest_compress <- NULL
    for (compress in compress_values) {
      do_compress(env, compress, the_path)
      new_info_ <- file.info(the_path)

      if (new_info_$size < smallest_size) {
        smallest_size <- new_info_$size
        smallest_compress <- compress
      }
    }

    do_compress(env, smallest_compress, the_path)

    new_info <- rbind(
      new_info,
      cbind(
        file.info(the_path),
        method = smallest_compress
      )
    )
  }
  df <- data.frame(
    path = rownames(old_info),
    size_old = old_info$size,
    size_new = new_info$size,
    method = new_info$method
  )
  df$"new/old" <- round(df$size_new / df$size_old * 1000) / 10
  df
}


#' @describeIn inlabru-deprecated Row-wise Kronecker products
#' `r lifecycle::badge("deprecated")` in favour of [fmesher::fm_row_kron()].
#'
#' Takes two Matrices and computes the row-wise Kronecker product.  Optionally
#' applies row-wise weights and/or applies an additional 0/1 row-wise Kronecker
#' matrix product.
#'
#' Returns a `Matrix::sparseMatrix` object.
#'
#' @param M1 A matrix that can be transformed into a sparse Matrix.
#' @param M2 A matrix that can be transformed into a sparse Matrix.
#' @param repl An optional index vector.  For each entry, specifies which
#' replicate the row belongs to, in the sense used in
#' `INLA::inla.spde.make.A`
#' @param n.repl The maximum replicate index, in the sense used in
#' `INLA::inla.spde.make.A()`.
#' @param weights Optional scaling weights to be applied row-wise to the
#' resulting matrix.
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @export row_kron
#' @keywords internal
row_kron <- function(M1, M2, repl = NULL, n.repl = NULL, weights = NULL) {
  lifecycle::deprecate_warn(
    "2.10.0",
    "row_kron()",
    "fmesher::fm_row_kron()"
  )
  fm_row_kron(M1 = M1, M2 = M2, repl = repl, n.repl = n.repl, weights = weights)
}

test_that("Missing data infilling", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not(bru_safe_sp())
  skip_if_not(bru_safe_terra())

  points <-
    sp::SpatialPointsDataFrame(
      matrix(1:6, 3, 2),
      data = data.frame(val = c(NA, NA, NA))
    )
  input_coord <- expand.grid(x = 0:6, y = 0:7)
  input_data <- data.frame(val = as.vector(input_coord$y))
  input <-
    sp::SpatialPixelsDataFrame(
      input_coord,
      data = input_data
    )

  val0 <- bru_fill_missing(input, points, points$val)

  input <-
    sp::SpatialPointsDataFrame(
      input_coord,
      data = input_data
    )
  val2 <- bru_fill_missing(input, points, points$val)

  expect_equal(val2, val0)

  input <- sp::SpatialGrid(sp::GridTopology(c(0, 0), c(1, 1), c(7, 8)))
  input_data <- data.frame(val = as.vector(sp::coordinates(input)[, 2]))
  input <-
    sp::SpatialGridDataFrame(
      input,
      data = input_data
    )
  val2 <- bru_fill_missing(input, points, points$val)

  expect_equal(val2, val0)
})


test_that("Laplace distribution", {
  q <- -5:5
  rate <- 2
  p <- plaplace(q, rate = rate)

  expect_equal(plaplace(q, rate = rate, lower.tail = TRUE), p)
  expect_equal(plaplace(-q, rate = rate, lower.tail = FALSE), p)
  expect_equal(
    plaplace(q, rate = rate, lower.tail = TRUE, log.p = TRUE),
    log(p)
  )
  expect_equal(
    plaplace(-q, rate = rate, lower.tail = FALSE, log.p = TRUE),
    log(p)
  )
  expect_equal(qlaplace(p, rate = rate, lower.tail = TRUE), q)
  expect_equal(qlaplace(p, rate = rate, lower.tail = FALSE), -q)
  expect_equal(
    qlaplace(log(p), rate = rate, lower.tail = TRUE, log.p = TRUE),
    q
  )
  expect_equal(
    qlaplace(log(p), rate = rate, lower.tail = FALSE, log.p = TRUE),
    -q
  )
})

test_that("2D LGCP fitting (sf)", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not(bru_safe_terra())
  skip_if_not_installed("sf")

  withr::local_seed(123L)

  options <- list(
    control.inla = list(
      int.strategy = "eb",
      h = 0.005
    )
  )

  gorillas <- gorillas_sf

  matern <- INLA::inla.spde2.pcmatern(gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(5, 0.01)
  )
  cmp <- geometry ~ mySmooth(main = geometry, model = matern) +
    Intercept(1)

  fit <- lgcp(
    cmp,
    data = gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = options
  )

  expect_s3_class(fit, "bru")

  # test_that("2D LGCP fitting: INLA intercept", {
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.6882955,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "sd"],
    0.5167925,
    tolerance = midtol
  )

  index <- c(1, 456, 789, 1058, 1479)
  # test_that("2D LGCP fitting: INLA random field", {
  expect_snapshot_value(
    fit$summary.random$mySmooth$mean[index],
    tolerance = midtol,
    style = "serialize"
  )
  expect_snapshot_value(
    fit$summary.random$mySmooth$sd[index],
    tolerance = midtol,
    style = "serialize"
  )
  expect_equal(
    fit$summary.hyperpar["Range for mySmooth", "mean"],
    2.137037,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.hyperpar["Stdev for mySmooth", "mean"],
    1.133754,
    tolerance = midtol
  )

  # test_that("2D LGCP fitting: predicted random field", {
  loc <- sf::st_sf(geometry = fm_as_sfc(gorillas$mesh, format = "loc"))
  withr::local_seed(123L)
  skip_if_not_installed("sn")
  pr <- predict(fit, loc, ~mySmooth,
    n.samples = 5, seed = 5657L,
    parallel.configs = FALSE
  )
  # Prediction variability includes reordering differences, so need large
  # tolerances unless n.samples is large
  expect_snapshot_value(
    pr$mean[c(1, 255, 778, 1000)],
    tolerance = 1,
    style = "serialize"
  )
  expect_snapshot_value(
    pr$sd[c(2, 215, 656, 1010)],
    tolerance = 1,
    style = "serialize"
  )

  # test_that("2D LGCP fitting: predicted intensity integral", {
  ips <- fm_int(gorillas$mesh, gorillas$boundary)
  withr::local_seed(123L)
  Lambda <- predict(fit, ips, ~ sum(weight * exp(mySmooth + Intercept)),
    n.samples = 10, seed = 5657L
  )

  expect_equal(
    Lambda$mean,
    670,
    tolerance = max(Lambda$mean.mc_std_err) * 6 / 670
  )
  expect_equal(
    Lambda$sd,
    27,
    tolerance = max(Lambda$sd.mc_std_err) * 6 / 27
  )

  # test_that("Supplying integration points instead of samplers&domains", {
  fit_ips <- lgcp(
    cmp,
    gorillas$nests,
    ips = ips,
    options = options
  )

  expect_equal(
    fit_ips$summary.fixed["Intercept", "mean"],
    fit$summary.fixed["Intercept", "mean"],
    tolerance = midtol
  )
  expect_equal(
    fit_ips$summary.fixed["Intercept", "sd"],
    fit$summary.fixed["Intercept", "sd"],
    tolerance = midtol
  )
  expect_equal(
    fit_ips$summary.random$mySmooth$mean,
    fit$summary.random$mySmooth$mean,
    tolerance = midtol
  )
  expect_equal(
    fit_ips$summary.random$mySmooth$sd,
    fit$summary.random$mySmooth$sd,
    tolerance = midtol
  )
})

test_that("2D LGCP fitting (sp)", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not(bru_safe_sp())
  skip_if_not(bru_safe_terra())
  skip_if_not_installed("sf")
  withr::local_options(lifecycle_verbosity = "quiet")

  withr::local_seed(123L)

  options <- list(
    control.inla = list(
      int.strategy = "eb",
      h = 0.005
    )
  )

  gorillas <- gorillas_sp()

  matern <- INLA::inla.spde2.pcmatern(gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(5, 0.01)
  )
  cmp <- coordinates ~ mySmooth(main = sp::coordinates, model = matern) +
    Intercept(1)

  ips <- fm_int(gorillas$mesh, gorillas$boundary)
  fit <- lgcp(
    cmp,
    data = gorillas$nests,
    ips = ips,
    options = options
  )

  expect_s3_class(fit, "bru")

  # test_that("2D LGCP fitting: INLA intercept", {
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.6882955,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "sd"],
    0.516925,
    tolerance = midtol
  )

  index <- c(1, 456, 789, 1058, 1479)
  # test_that("2D LGCP fitting: INLA random field", {
  expect_snapshot_value(
    fit$summary.random$mySmooth$mean[index],
    tolerance = midtol,
    style = "serialize"
  )
  expect_snapshot_value(
    fit$summary.random$mySmooth$sd[index],
    tolerance = midtol,
    style = "serialize"
  )
  expect_equal(
    fit$summary.hyperpar["Range for mySmooth", "mean"],
    2.137037,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.hyperpar["Stdev for mySmooth", "mean"],
    1.133754,
    tolerance = midtol
  )

  # test_that("2D LGCP fitting: predicted random field", {
  loc <- sp::SpatialPoints(
    gorillas$mesh$loc[, c(1, 2)],
    proj4string = fm_CRS(gorillas$nests)
  )
  withr::local_seed(123L)
  skip_if_not_installed("sn")
  pr <- predict(fit, loc, ~mySmooth,
    n.samples = 5, seed = 5657L,
    parallel.configs = FALSE
  )
  # Prediction variability includes reordering differences, so need large
  # tolerances unless n.samples is large
  expect_snapshot_value(
    pr$mean[c(1, 255, 778, 1000)],
    tolerance = 1,
    style = "serialize"
  )
  expect_snapshot_value(
    pr$sd[c(2, 215, 656, 1010)],
    tolerance = 1,
    style = "serialize"
  )

  # test_that("2D LGCP fitting: predicted intensity integral", {
  withr::local_seed(123L)
  Lambda <- predict(fit, ips, ~ sum(weight * exp(mySmooth + Intercept)),
    n.samples = 10, seed = 5657L
  )

  expect_equal(
    Lambda$mean,
    670,
    tolerance = max(Lambda$mean.mc_std_err) * 6 / 670
  )
  expect_equal(
    Lambda$sd,
    27,
    tolerance = max(Lambda$sd.mc_std_err) * 6 / 27
  )
})

test_that("2D LGCP fitting", {
  skip_on_cran()
  local_bru_safe_inla()

  # test_that("2D LGCP fitting: Factor covariate (as SpatialPixelsDataFrame)", {
  skip_if_not(bru_safe_terra())
  skip_if_not_installed("sf")
  gorillas <- gorillas_sf
  gorillas$gcov <- gorillas_sf_gcov()

  # Uses the component label to pick the covariate layer to extract,
  # so doesn't need an explicit main_layer="vegetation".

  mdl <- geometry ~ vegetation(
    main = gorillas$gcov$vegetation,
    model = "iid"
  ) - Intercept

  fit <- lgcp(
    mdl, gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(
      control.inla = list(
        int.strategy = "eb",
        h = 0.005
      )
    )
  )

  expect_snapshot_value(
    fit$summary.random$veg$mean,
    tolerance = midtol,
    style = "serialize"
  )
  expect_snapshot_value(
    fit$summary.random$veg$sd,
    tolerance = midtol,
    style = "serialize"
  )

  # test_that("2D LGCP fitting: Continuous covariate (as function)", {
  elev <- gorillas$gcov$elevation
  elev <- elev - mean(terra::values(elev), na.rm = TRUE)

  mdl2 <- geometry ~ beta.elev(
    main = elev,
    main_layer = "elevation",
    model = "linear"
  ) + Intercept(1)
  fit2 <- lgcp(mdl2, gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(
      control.inla = list(
        int.strategy = "eb",
        h = 0.005
      )
    )
  )

  expect_equal(fit2$summary.fixed["beta.elev", "mean"], 0.004180126,
    tolerance = midtol
  )
  expect_equal(fit2$summary.fixed["beta.elev", "sd"], 0.0002472825,
    tolerance = midtol
  )
  expect_equal(fit2$summary.fixed["Intercept", "mean"], 2.967071,
    tolerance = midtol
  )
  expect_equal(fit2$summary.fixed["Intercept", "sd"], 0.06031667,
    tolerance = midtol
  )

  f.elev <- function(xy) {
    eval_spatial(elev, xy)
  }

  mdl3 <- geometry ~ beta.elev(
    main = f.elev(.data.$geometry),
    model = "linear"
  ) +
    Intercept(1)
  fit3 <- lgcp(mdl3, gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(
      control.inla = list(
        int.strategy = "eb",
        h = 0.005
      )
    )
  )

  expect_equal(fit2$summary.fixed["beta.elev", "mean"], 0.004180126,
    tolerance = midtol
  )
  expect_equal(fit2$summary.fixed["beta.elev", "sd"], 0.0002472825,
    tolerance = midtol
  )
  expect_equal(fit2$summary.fixed["Intercept", "mean"], 2.967071,
    tolerance = midtol
  )
  expect_equal(fit2$summary.fixed["Intercept", "sd"], 0.06031667,
    tolerance = midtol
  )
})

# > bru_timings(fit)
# Task Iteration       Time     System    Elapsed
# user.self Preprocess         0 0.195 secs 0.002 secs 0.222 secs
# 1         Preprocess         1 0.454 secs 0.002 secs 0.479 secs
# 2         Run inla()         1 0.866 secs 0.016 secs 0.707 secs
# > bru_timings(fit2)
# Task Iteration       Time     System    Elapsed
# user.self Preprocess         0 0.420 secs 0.006 secs 0.430 secs
# 1         Preprocess         1 0.267 secs 0.008 secs 0.272 secs
# 2         Run inla()         1 0.695 secs 0.018 secs 0.439 secs
# > bru_timings(fit3)
# Task Iteration       Time     System    Elapsed
# user.self Preprocess         0 0.067 secs 0.002 secs 0.068 secs
# 1         Preprocess         1 0.327 secs 0.002 secs 0.328 secs
# 2         Run inla()         1 0.619 secs 0.016 secs 0.407 secs

# > bru_timings(fit)
# Task Iteration       Time     System    Elapsed
# 1 Preprocess         0 0.449 secs 0.022 secs 0.514 secs
# 2 Preprocess         1 0.199 secs 0.062 secs 0.231 secs
# 3 Run inla()         1 0.962 secs 0.044 secs 0.753 secs
# > bru_timings(fit2)
# Task Iteration       Time     System    Elapsed
# 1 Preprocess         0 0.507 secs 0.028 secs 0.525 secs
# 2 Preprocess         1 0.220 secs 0.000 secs 0.219 secs
# 3 Run inla()         1 0.648 secs 0.012 secs 0.405 secs
# > bru_timings(fit3)
# Task Iteration       Time     System    Elapsed
# 1 Preprocess         0 0.137 secs 0.021 secs 0.148 secs
# 2 Preprocess         1 0.060 secs 0.000 secs 0.059 secs
# 3 Run inla()         1 0.626 secs 0.020 secs 0.455 secs

test_that("eval_spatial.SpatRaster", {
  skip_if_not(bru_safe_terra())

  # Load the Gorilla data
  data(gorillas_sf, package = "inlabru", envir = environment())
  gcov <- gorillas_sf_gcov()

  covs <- c(
    gcov$elevation,
    gcov$elevation
  )
  names(covs) <- c("A", "B")
  covs$B <- covs$B + 1

  # Multirow
  where <- gorillas_sf$nests[seq_len(5), , drop = FALSE]
  A_val <- eval_spatial(covs,
    where = where,
    layer = rep("A", nrow(where))
  )
  B_val <- eval_spatial(covs,
    where = where,
    layer = rep("B", nrow(where))
  )
  AB_val <- eval_spatial(covs,
    where = where,
    layer = rep(
      c("A", "B"),
      c(2, nrow(where) - 2)
    )
  )

  expect_equal(A_val + 1, B_val)
  expect_equal(c(A_val[1:2], B_val[3:5]), AB_val)

  # Multirow
  where <- gorillas_sf$nests[seq_len(5), , drop = FALSE]
  A1_val <- eval_spatial(covs,
    where = where,
    layer = "A"
  )
  B1_val <- eval_spatial(covs,
    where = where,
    layer = "B"
  )

  expect_equal(A1_val, A_val)
  expect_equal(B1_val, B_val)

  # Single row
  where <- gorillas_sf$nests[5, , drop = FALSE]
  A2_val <- eval_spatial(covs,
    where = where,
    layer = "A"
  )
  B2_val <- eval_spatial(covs,
    where = where,
    layer = "B"
  )

  expect_equal(A2_val, A_val[5])
  expect_equal(B2_val, B_val[5])
})


test_that("eval_spatial.Spatial*", {
  skip_if_not(bru_safe_sp())

  # Load the Gorilla data
  skip_if_not(bru_safe_terra())
  skip_if_not_installed("sf")
  gorillas <- gorillas_sp()

  covs <- gorillas$gcov$elevation
  names(covs) <- "A"
  covs$B <- covs$A + 1

  # Multirow
  where <- gorillas$nests[seq_len(5), , drop = FALSE]
  A_val <- eval_spatial(covs,
    where = where,
    layer = rep("A", nrow(where))
  )
  B_val <- eval_spatial(covs,
    where = where,
    layer = rep("B", nrow(where))
  )
  AB_val <- eval_spatial(covs,
    where = where,
    layer = rep(
      c("A", "B"),
      c(2, nrow(where) - 2)
    )
  )

  expect_equal(A_val + 1, B_val)
  expect_equal(c(A_val[1:2], B_val[3:5]), AB_val)

  # Multirow
  where <- gorillas$nests[seq_len(5), , drop = FALSE]
  A1_val <- eval_spatial(covs,
    where = where,
    layer = "A"
  )
  B1_val <- eval_spatial(covs,
    where = where,
    layer = "B"
  )

  expect_equal(A1_val, A_val)
  expect_equal(B1_val, B_val)

  # Single row
  where <- gorillas$nests[5, , drop = FALSE]
  A2_val <- eval_spatial(covs,
    where = where,
    layer = "A"
  )
  B2_val <- eval_spatial(covs,
    where = where,
    layer = "B"
  )

  expect_equal(A2_val, A_val[5])
  expect_equal(B2_val, B_val[5])
})


test_that("eval_spatial.sf", {
  # Load the Gorilla data
  data(gorillas_sf, package = "inlabru", envir = environment())

  nests <- gorillas_sf$nests
  plots <- gorillas_sf$plotsample$plots
  plots$something_num <- seq_len(nrow(plots))
  nests$something_num <- eval_spatial(plots, nests, layer = "something_num")
  something_num <- eval_spatial(plots, plots, layer = "something_num")

  expect_equal(
    sort(unique(nests$something_num), na.last = TRUE),
    c(3, 6, 8, 9, 16, 17, 19, 20, 21, 22, 25, NA)
  )

  plots$something_char <- as.character(seq_len(nrow(plots)))
  nests$something_char <- eval_spatial(plots, nests, layer = "something_char")

  expect_equal(nests$something_char, as.character(nests$something_num))
})


test_that("eval_spatial.stars", {
  skip_if_not(bru_safe_terra())
  skip_if_not_installed("stars")

  # Load the Gorilla data
  data(gorillas_sf, package = "inlabru", envir = environment())
  gcov <- gorillas_sf_gcov()

  covs <- c(
    gcov$elevation,
    gcov$elevation
  )
  names(covs) <- c("A", "B")
  covs$B <- covs$B + 10000L

  # Multirow
  where <- gorillas_sf$nests[seq_len(5), , drop = FALSE]
  A_val <- eval_spatial(covs,
    where = where,
    layer = rep("A", nrow(where))
  )
  B_val <- eval_spatial(covs,
    where = where,
    layer = rep("B", nrow(where))
  )
  AB_val <- eval_spatial(covs,
    where = where,
    layer = rep(
      c("A", "B"),
      c(2, nrow(where) - 2)
    )
  )

  expect_equal(A_val + 10000L, B_val)
  expect_equal(c(A_val[1:2], B_val[3:5]), AB_val)

  # Multirow
  where <- gorillas_sf$nests[seq_len(5), , drop = FALSE]
  A1_val <- eval_spatial(covs,
    where = where,
    layer = "A"
  )
  B1_val <- eval_spatial(covs,
    where = where,
    layer = "B"
  )

  expect_equal(A1_val, A_val)
  expect_equal(B1_val, B_val)

  # Single row
  where <- gorillas_sf$nests[5, , drop = FALSE]
  A2_val <- eval_spatial(covs,
    where = where,
    layer = "A"
  )
  B2_val <- eval_spatial(covs,
    where = where,
    layer = "B"
  )

  expect_equal(A2_val, A_val[5], ignore_attr = "class")
  expect_equal(B2_val, B_val[5], ignore_attr = "class")
})

access_trace <- function(name, package = NULL) {
  method <- fmesher::fm_caller_name(-1)
  access_method_ <- list(
    "^\\$<-" = "$<-",
    "^\\$" = "$",
    "^\\[\\[<-" = "[[<-",
    "^\\[\\[" = "[[",
    "^\\[<-" = "[<-",
    "^\\[" = "["
  )
  access_method <- NA_character_
  for (pattern in names(access_method_)) {
    if (grepl(pattern, method)) {
      access_method <- access_method_[[pattern]]
      break
    }
  }
  method_text_ <-
    list(
      "$<-" = "\\$<-",
      "$" = "\\$",
      "[[<-" = "\\[\\[<-",
      "[[" = "\\[\\[",
      "[<-" = "\\[<-",
      "[" = "\\["
    )
  method_text <- if (access_method %in% names(method_text_)) {
    method_text_[[access_method]]
  } else {
    ""
  }
  open_text_ <- list(
    "$<-" = "$",
    "$" = "$",
    "[[<-" = "[['",
    "[[" = "[['",
    "[<-" = "['",
    "[" = "['"
  )
  open_text <- if (access_method %in% names(open_text_)) {
    open_text_[[access_method]]
  } else {
    ""
  }
  close_text_ <- list(
    "$<-" = " <-",
    "$" = "",
    "[[<-" = "']] <-",
    "[[" = "']]",
    "[<-" = "'] <-",
    "[" = "']"
  )
  close_text <- if (access_method %in% names(close_text_)) {
    close_text_[[access_method]]
  } else {
    ""
  }

  fun_package <- function(fun_name) {
    fun <- tryCatch(
      get(fun_name, mode = "function", inherits = TRUE),
      error = function(e) NULL
    )
    if (is.null(fun)) {
      return("")
    }
    env <- environment(fun)
    pkg <- ""
    while (!is.null(env)) {
      if (identical(env, .GlobalEnv)) {
        pkg <- "global"
        break
      }
      env_name <- environmentName(env)
      if (!is.null(env_name) && !identical(env_name, "")) {
        pkg <- env_name
        break
      }
      env <- parent.env(env)
    }
    pkg
  }

  class_name <- sub(paste0("^", method_text, "\\."), "", method)
  idx <- -3
  caller <- fmesher::fm_caller_name(idx)
  if (caller %in% c("", "FUN")) {
    idx <- idx - 1L
    cal <- fmesher::fm_caller_name(idx)
    caller <- c(caller, cal)
  }
  for (i in seq_len(1)) {
    idx <- idx - 1L
    cal <- fmesher::fm_caller_name(idx)
    caller <- c(caller, cal)
  }
  caller_packages <- vapply(caller, fun_package, character(1))
  if (any(caller_packages %in% package)) {
    return(invisible())
  }
  caller <- paste0(caller, collapse = ":")

  if (!is.character(name)) {
    if (is.integer(name)) {
      name <- "<integer>"
    } else if (is.numeric(name)) {
      name <- "<numeric>"
    } else if (is.logical(name)) {
      name <- "<logical>"
    } else {
      name <- "<unknown>"
    }
  }

  print(glue::glue("{caller}: {class_name}{open_text}{name}{close_text}"))
}


access_trace_reset <- function(file = "R/access_trace_methods.R") {
  write(
    c(
      "# This file is auto-generated by access_trace_reset/add() calls.",
      "# Do not edit!"
    ),
    file = file,
    append = FALSE,
    sep = ""
  )
  invisible()
}


# @param class_name
# @param method_names
# @param file
# @param package character vector; If non-NULL, only trace access outside of
# this package or packages
access_trace_add <- function(
  class_name,
  method_names = c("$", "[[", "[", "$<-", "[[<-", "[<-"),
  file = "R/access_trace_methods.R",
  package = NULL
) {
  package <- union(package, c("base", "utils"))

  write(
    c("", glue::glue("# Class {class_name} ####"), ""),
    file = file,
    append = TRUE,
    sep = ""
  )
  arguments_ <-
    list(
      "$" = "name",
      "[[" = "i",
      "[" = "i",
      "$<-" = "name, value",
      "[[<-" = "i, value",
      "[<-" = "i, value"
    )
  trace_name_ <-
    list(
      "$" = "name",
      "[[" = "i",
      "[" = "i",
      "$<-" = "name",
      "[[<-" = "i",
      "[<-" = "i"
    )
  package_text <-
    if (is.null(package)) {
      "NULL"
    } else {
      paste0("c(", paste0('\"', package, '\"', collapse = ", "), ")")
    }
  for (method_name in method_names) {
    full_name <- paste0(method_name, ".", class_name)
    if (exists(full_name, mode = "function")) {
      warning("Method ", full_name, " already exists; skipping generation.")
      next
    }
    arguments <- arguments_[[method_name]]
    trace_name <- trace_name_[[method_name]]
    code <- c(glue::glue(
      "#' @export
       #'
       `{full_name}` <- function(x, {arguments}) {{
         access_trace({trace_name}, package = {package_text})
         NextMethod()
       }}"
    ), "")
    write(
      code,
      file = file,
      append = TRUE,
      sep = ""
    )
  }
  invisible()
}

# General conversion to `bru_obs`, `bru_obs_list` objects ####

#' @title Conversion methods for `bru_obs` and `bru_obs_list` objects
#' @description Methods for converting to `bru_obs` and `bru_obs_list`
#'   objects.
#' @param x An object to convert to [bru_obs] or [bru_obs_list]
#' @param \dots Additional arguments passed to sub-methods.
#' @returns An object of class [bru_obs] or [bru_obs_list].
#' @export
#' @seealso [as_bru_comp_list()]
as_bru_obs <- function(x, ...) {
  if (is.null(x)) {
    return(NULL)
  }
  UseMethod("as_bru_obs")
}

#' @rdname as_bru_obs
#' @export
as_bru_obs_list <- function(x, .tag = NULL) {
  if (is.null(x)) {
    return(NULL)
  }
  UseMethod("as_bru_obs_list")
}

#' @rdname as_bru_obs
#' @export
as_bru_obs.bru_obs <- function(x, ...) {
  x
}

#' @rdname as_bru_obs
#' @param .tag character; optional name for the single observation model in the
#'   returned [bru_obs_list] object. Default is `NULL`, which results in
#'   automatic naming based on the `tag` attribute of `x`, if present.
#' @export
as_bru_obs_list.bru_obs <- function(x, .tag = NULL) {
  bru_obs_list(x, .tag = .tag)
}

#' @rdname as_bru_obs
#' @export
as_bru_obs_list.list <- function(x, .tag = NULL) {
  bru_obs_list(x)
}

#' @rdname as_bru_obs
#' @export
as_bru_obs_list.bru_obs_list <- function(x, .tag = NULL) {
  bru_obs_list(x)
}

#' @rdname as_bru_obs
#' @export
as_bru_obs_list.bru <- function(x, .tag = NULL) {
  as_bru_obs_list(x[["bru_info"]])
}

#' @rdname as_bru_obs
#' @export
as_bru_obs_list.bru_info <- function(x, .tag = NULL) {
  # When/if lhoods is moved to the bru_model class, this code can be simplified.
  if (is.null(x[["lhoods"]])) {
    as_bru_obs_list(x[["model"]])
  } else {
    bru_obs_list(x[["lhoods"]])
  }
}

#' @rdname as_bru_obs
#' @export
as_bru_obs_list.bru_model <- function(x, .tag = NULL) {
  # When/if lhoods is moved to the bru_model class, this code can be simplified.
  if (is.null(x[["lhoods"]])) {
    bru_obs_list(list())
  } else {
    bru_obs_list(x[["lhoods"]])
  }
}

#' @title Conversion methods for `bru_comp` and `bru_comp_list` objects
#' @description Methods for converting to `bru_comp` and `bru_comp_list`
#'   objects.
#' @param x An object to convert to [bru_comp] or [bru_comp_list]
#' @param \dots Additional arguments passed on to [bru_comp_list()].
#' @returns An object of class [bru_comp_list].
#' @export
#' @rdname as_bru_comp
#' @name as_bru_comp
#' @seealso [as_bru_obs()], [as_bru_obs_list()]
as_bru_comp <- function(x, ...) {
  if (is.null(x)) {
    return(NULL)
  }
  UseMethod("as_bru_comp")
}

#' @rdname as_bru_comp
#' @export
as_bru_comp_list <- function(x, ...) {
  if (is.null(x)) {
    return(NULL)
  }
  UseMethod("as_bru_comp_list")
}

#' @rdname as_bru_comp
#' @export
as_bru_comp.bru_comp <- function(x, ...) {
  x
}

#' @rdname as_bru_comp
#' @export
as_bru_comp_list.bru_comp_list <- function(x, ...) {
  bru_comp_list(x, ...)
}

#' @rdname as_bru_comp
#' @export
as_bru_comp_list.bru_comp <- function(x, ...) {
  bru_comp_list(list(x), ...)
}

#' @describeIn as_bru_comp Extract the component list from a [bru()] object.
#' @export
as_bru_comp_list.bru <- function(x, ...) {
  as_bru_comp_list(x[["bru_info"]], ...)
}

#' @describeIn as_bru_comp Extract the component list from a [bru_info()]
#'   object.
#' @export
as_bru_comp_list.bru_info <- function(x, ...) {
  as_bru_comp_list(x[["model"]], ...)
}

#' @describeIn as_bru_comp Extract the component list from a [bru_model()]
#'   object.
#' @export
as_bru_comp_list.bru_model <- function(x, ...) {
  as_bru_comp_list(x[["effects"]], ...)
}

#' @rdname as_bru_comp
#' @export
as_bru_comp_list.list <- function(x, ...) {
  bru_comp_list(x, ...)
}

#' @rdname as_bru_comp
#' @export
as_bru_comp_list.formula <- function(x, ...) {
  bru_comp_list(x, ...)
}

#' @include deprecated.R

#' @title Store and evaluate component inputs
#' @description Store and evaluate component inputs, including support for
#'   non-standard-evaluation with `rlang`, automatic transfer to
#'   function calls, [eval_spatial()], and [MatrixModels::model.Matrix()].
#'
#'   These functions are normally only called internally, but users may find
#'   it useful to call them directly to experiment, and to make advanced
#'   model definitions using [ibm_input_set]/[ibm_input_new()].
#' @name bru_input
#' @rdname bru_input
NULL

#' @describeIn bru_input Create a `bru_input` object.
#' @param input An expression to be evaluated.
#' @param label character; optional label used to identify the object in
#'   informational messages.
#' @param layer Optional expression that evaluates layer names or indices for
#'   use with [eval_spatial()]
#' @param selector character or integer; optional selector for use with
#'   use with [eval_spatial()]
#' @param \dots Passed on to sub-methods.
#' @seealso [bru_input_text()]
#' @export
#' @examples
#' (inp <- new_bru_input(x, "LABEL"))
#' bru_input(inp, data.frame(x = 1:3))
#'
new_bru_input <- function(input,
                          label = NULL,
                          layer = NULL,
                          selector = NULL,
                          ...) {
  inp <- structure(
    list(
      input = rlang::enquo(input),
      label = label,
      layer = rlang::enquo(layer),
      selector = selector
    ),
    class = "bru_input"
  )
  inp
}

#' @describeIn bru_input Evaluate inputs
#'
#' @export
#' @seealso [summary.bru_input()], [bru_comp()]
bru_input <- function(...) {
  UseMethod("bru_input")
}

#' @title Extract input definitions as text
#'
#' @description
#' Extract input definitions from the [bru_input] information stored in bru
#' model components.  Primarily intended for diagnostic use, as the output does
#' not currently contain all information, such as the layer and selector
#' information for spatial covariates.
#'
#' @inheritParams bru_input
#'
#' @seealso [bru_input()]
#' @keywords internal
#' @rdname bru_input_text
#' @export
#' @examples
#' (inp <- new_bru_input(x, "LABEL"))
#' bru_input_text(inp)
#' if (bru_safe_inla()) {
#'   bru_input_text(bru_comp("x", cos(y)))
#' }
#'
bru_input_text <- function(...) {
  UseMethod("bru_input_text")
}

#' @describeIn bru_input Attempts to evaluate a component input (e.g. `main`,
#' `group`, `replicate`, or `weight`), and process the results:
#' 1. If `rlang::eval_tidy()` failed, return NULL or map everything to 1
#'    (see the `null.on.fail` argument). This should normally not
#'    happen, unless the component use logic is incorrect,
#'    (via `used`)
#'    leading to missing columns for a certain likelihood in a
#'    multi-`bru_obs()` model.
#' 2. If we obtain a function, apply the function to the data object
#' 3. If we obtain an object supported by [eval_spatial()], extract the values
#'    of that data frame at the point locations
#' 4. If we obtain a formula, call `ModelMatrix::model.Matrix()`
#' 5. Else we obtain a vector and return as-is. This happens when input
#'    references a column of the data points, or some other complete expression
#'
#' @param x A `bru_input` object, or other object for recursive evaluation.
#' @param data A `data.frame`, `tibble`, `sf`, `list`, or `Spatial*` object of
#' covariates and/or point locations.
#' @param mask A `bru_data_mask` object, constructed internally.
#' @param null.on.fail logical; if `TRUE`, return `NULL` if the input evaluation
#'   fails. If `FALSE` (default), return a vector of 1s and issue a warning
#'   (only for deprecated use of `coordinates` without having loaded the `sp`
#'   package), or stop with an error.
#' @param .envir environment in which to evaluate the input expression. Default
#'   is `parent.frame()`
#' @export
bru_input.bru_input <- function(x,
                                data = NULL,
                                mask = NULL,
                                null.on.fail = FALSE,
                                .envir = parent.frame(),
                                ...) {
  bru_log_message(
    glue("bru_input(bru_input) for ({x$label})"),
    verbosity = 5
  )
  # Evaluate the map with the data in an environment
  if (is.null(mask)) {
    mask <- bru_data_mask(list(data = data))
  }
  orig_data <- data
  if (!is.null(data) && !is.list(data)) {
    data <- tibble::as_tibble(data)
  }

  inp <- rlang::as_quosure(x$input, env = .envir)
  e_input <- tryCatch(
    rlang::eval_tidy(inp, data = mask, env = .envir),
    error = function(e) {
      e
    }
  )

  # ## Need to handle varying input lengths.
  # ## auto-expansion of scalars needs to happen elsewhere, where it's needed,
  # ## e.g. when using component A-matrices to construct a default linear model
  # ## A matrix.
  #  n <- nrow(as.data.frame(data))

  handle_problems <- function(val) {
    if (is.null(val)) {
      return(NULL)
    }
    if (inherits(e_input, "error")) {
      if (null.on.fail) {
        return(NULL)
      }

      val <- 1
      input_string <- paste0(rlang::as_label(x$input), collapse = "\n")
      if (identical(input_string, "coordinates")) {
        warning(
          glue(
            "The input evaluation '{input_string}' for '{x$label}'",
            " failed. Perhaps you need to load the 'sp' package ",
            "with 'library(sp)'?",
            " Attempting 'sp::coordinates'."
          ),
          immediate. = TRUE
        )
        x$input <- rlang::as_quosure(sp::coordinates, env = .envir)
        return(bru_input(
          x,
          data = orig_data,
          mask = mask,
          .envir = .envir,
          null.on.fail = null.on.fail,
          ...
        ))
      } else {
        stop(
          glue(
            "The input evaluation '{input_string}' for '{x$label}' failed.",
            "\n  Perhaps the data object doesn't contain the needed variables?"
          )
        )
      }
    }
    val
  }

  e_input <- handle_problems(e_input)
  if (is.null(e_input)) {
    return(NULL)
  }

  if (is.function(e_input)) {
    # Allow but detect failures:
    val <- tryCatch(
      e_input(orig_data),
      error = function(e) {
        e
      }
    )

    val <- handle_problems(val)
    if (is.null(val)) {
      return(NULL)
    }

    if (identical(as.character(x$input), "coordinates")) {
      tryCatch(
        expr = {
          # Return SpatialPoints instead of a matrix
          val <- as.data.frame(val)
          sp::coordinates(val) <- seq_len(ncol(val))
          # Allow proj4string failures:
          data_crs <- tryCatch(fm_CRS(orig_data),
            error = function(e) {}
          )
          if (!fm_crs_is_null(data_crs)) {
            sp::proj4string(val) <- data_crs
          }
          val
        },
        error = function(e) {
          NULL
        }
      )
    }
  } else if (inherits(e_input, "formula")) {
    # Allow but detect failures:
    val <- tryCatch(
      MatrixModels::model.Matrix(e_input, data = data, sparse = TRUE),
      error = function(e) {
        e
      }
    )
    val <- handle_problems(val)
    if (is.null(val)) {
      return(NULL)
    }
    val <- as(val, "Matrix")
  } else if (inherits(
    e_input,
    gsub(
      pattern = "^eval_spatial\\.([^*]*)\\*?",
      replacement = "\\1",
      x = format(utils::.S3methods("eval_spatial"))
    )
  )) {
    input_layer <-
      bru_input_layer(
        layer = x[["layer"]],
        selector = x[["selector"]],
        mask = mask,
        .envir = .envir,
        label = x[["label"]],
        e_input = e_input
      )
    layer <- extract_layer(
      data,
      input_layer,
      x[["selector"]]
    )
    check_layer(e_input, data, layer)
    val <- eval_spatial(
      e_input,
      orig_data,
      layer = layer,
      selector = NULL
    )
    #  } else if ((x$label %in% c("offset", "const")) &&
    #    is.numeric(e_input) &&
    #    (length(e_input) == 1)) {
    #    val <- rep(e_input, n)
  } else {
    val <- e_input
  }

  # ## Need to allow different sizes; K %*% effect needs to have same length as
  # response, but the input and effect effect itself doesn't.
  #  # Always return as many rows as data has
  #  if (is.vector(val) && (length(val) == 1) && (n > 1)) {
  #    val <- rep(val, n)
  #  }

  # Check if any of the locations are NA. If we are dealing with
  # SpatialGridDataFrame try to fix that by filling in nearest neighbour values.
  # # TODO: Check how to deal with this fully in the case of multilikelihood
  # models
  # Answer: should respect the lhood "include/exclude" info for the component
  # list
  if ((inherits(e_input, c(
    "SpatialGridDataFrame",
    "SpatialPixelsDataFrame",
    "SpatRaster"
  ))) &&
    anyNA(as.data.frame(val))) {
    msg <- glue(
      "Model input '",
      glue_collapse(rlang::as_label(x$input), sep = "\n"),
      "' for '{x$label}' returned some NA values.\n",
      "Attempting to fill in spatially by nearest available value.\n",
      "To avoid this basic covariate imputation, supply complete data."
    )
    bru_log_warn(msg)

    val <- bru_fill_missing(
      data = e_input, where = orig_data, values = val,
      layer = layer, selector = NULL
    )
  }

  # Check for NA values.
  # TODO: update 2022-11-04:
  # iid models now by default are given _index mappers, which treat NAs
  # as zero.  An option to turn on a warning could be useful for checking
  # models that aren't supposed to have NA inputs.
  #
  #  if (anyNA(unlist(as.vector(val)))) {
  #    # TODO: remove this check and make sure NAs are handled properly
  #    # elsewhere,
  #    # if possible. Problem: treating NA as "no effect" can have negative side
  #    # effects. For spatial covariates, can be handled by infill, but a
  #    # general
  #    # solution doesn't exist, so not sure how to deal with this.
  #    stop(sprintf(
  #      "Input '%s' of component '%s' has returned NA values. Please design
  #       your argument as to return non-NA for all points in your model
  #       domain/mesh.",
  #      as.character(x$input)[[1]], label
  #    ))
  #  }

  val
}

#' @describeIn bru_input_text Extract the input definitions from a `bru_input`
#'   object, as a named character vector
#' @export
bru_input_text.bru_input <- function(x,
                                     .envir = parent.frame(),
                                     ...) {
  bru_log_message(
    glue("bru_input_text(bru_input) for ({x$label})"),
    verbosity = 5
  )

  inp <- paste0(rlang::as_label(x$input), collapse = "\n")

  inp
}

#' @section Simple covariates and the input parameters:
#'
#' It is not unusual for a random effect act on a transformation of a covariate.
#' In other frameworks this would mean that the transformed covariate would have
#' to be calculated in advance and added to the data frame that is usually
#' provided via the `data` parameter. inlabru provides the option to do this
#' transformation automatically. For instance, one might be interested in the
#' effect of a covariate \eqn{x^2}. In inla and other frameworks this would
#' require to add a column `xsquared` to the input data frame and use the
#' formula
#'
#' \itemize{\item{`formula = y ~ f(xsquared, model = "linear")`,}}
#'
#' In inlabru this can be achieved in several ways of using the `main` parameter
#' (`map` in version 2.1.13 and earlier), which does not need to be named.
#'
#' \itemize{
#' \item{`components = y ~ psi(x^2, model = "linear")`}
#' \item{`components = y ~ psi(main = x^2, model = "linear")`}
#' \item{`components = y ~ psi(mySquareFun(x), model = "linear")`,}
#' \item{`components = y ~ psi(myOtherSquareFun, model = "linear")`,}
#' }
#'
#' In the first example inlabru will interpret the map parameter as an
#' expression to be evaluated within the data provided. Since \eqn{x} is a known
#' covariate it will know how to calculate it. The second example is an
#' expression as well but it uses a function called `mySquareFun`. This function
#' is defined by user but has to be accessible within the work space when
#' setting up the components. The third example provides the function
#' `myOtherSquareFun`. In this case, inlabru will call the function as
#' `myOtherSquareFun(.data.)`, where `.data.` is the data provided via the
#' [bru_obs()] `data` parameter. The function needs to know what parts of the
#' data to use to construct the needed output. For example,
#' ```
#' myOtherSquareFun <- function(data) {
#'   data[ ,"x"]^2
#' }
#' ```
#'
#' Interactions can be handled by a formula input and `model = "fixed"`:
#' `components = y ~ 0 + name(~ 1 + x:z, model = "fixed")`
#'
#' @section Spatial Covariates:
#'
#' When fitting spatial models it is common to work with covariates that depend
#' on space, e.g. sea surface temperature or elevation. Although it is
#' straightforward to add this data to the input data frame or write a covariate
#' function like in the previous section there is an even more convenient way in
#' inlabru. Spatial covariates are often stored as `SpatialPixelsDataFrame`,
#' `SpatialPixelsDataFrame` or `RasterLayer` objects. These can be provided
#' directly via the input expressions if they are supported by [eval_spatial()],
#' and the [bru_obs()] data is an `sf` or `SpatialPointsDataFrame` object.
#' `inlabru` will then automatically evaluate and/or interpolate the covariate
#' at your data locations when using code like
#' ```
#' components = y ~ psi(mySpatialPixels, model = "linear")
#' ```
#' For more precise control, use the the `layer` and `selector` arguments (see
#' [bru_comp()]), or call [eval_spatial()] directly, e.g.:
#' ```
#' components = y ~ psi(eval_spatial(mySpatialPixels, where = .data.),
#'                      model = "linear")
#' ```
#'
#' @section Coordinates:
#'
#' A common spatial modelling component when using inla are SPDE models. An
#' important feature of inlabru is that it will automatically calculate the so
#' called A-matrix (a component model matrix) which maps SPDE values at the mesh
#' vertices to values at the data locations. For this purpose, the input can be
#' set to `coordinates`, which is the `sp` package function that extracts point
#' coordinates from the `SpatialPointsDataFrame` that was provided as input to
#' [bru_obs()]. The code for this would look as follows:
#' ```
#' components = y ~ field(coordinates, model = inla.spde2.matern(...))
#' ```
#' Since `coordinates` is a function from the `sp` package, this results in
#' evaluation of `sp::coordinates(.data.)`, which loses any CRS information
#' from the data object.
#'
#' For `sf` data with a geometry column (by default named `geometry`), use
#' ```
#' components = y ~ field(geometry, model = inla.spde2.matern(...))
#' ```
#' Since the CRS information is part of the geometry column of the `sf` object,
#' this retains CRS information, so this is more robust, and allows the model
#' to be built on a different CRS than the observation data.
#'
#'
#' @export
#' @return * `bru_input(bru_comp)`: A list of mapper input values, formatted
#'   for the full component mapper (of type [bm_pipe])
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}, Finn Lindgren
#'   \email{finn.lindgren@@gmail.com}
#' @rdname bru_input
bru_input.bru_comp <- function(x, ..., label = x$label) {
  bru_log_message(
    glue("bru_input.bru_comp({label})"),
    verbosity = 4
  )

  if (is.null(x[["mapper"]])) {
    stop(glue("Component mapper for '{label}' is NULL"))
  }

  bru_input(x[["mapper"]], ..., label = label)
}

#' @describeIn bru_input_text Extract the input definition as a named character
#' vector.
#' @return * `bru_input(bru_comp)`: A character vector of mapper input values.
#' @export
bru_input_text.bru_comp <- function(x, ..., label = x$label) {
  bru_log_message(
    glue("bru_input.bru_comp({label})"),
    verbosity = 4
  )

  if (is.null(x[["mapper"]])) {
    stop(glue("Component mapper for '{label}' is NULL"))
  }

  bru_input_text(x[["mapper"]], ..., label = label)
}

#' @return * `bru_input(bru_comp_list)`: A list of mapper input values,
#'   with one entry for each component.
#' @export
#' @rdname bru_input
bru_input.bru_comp_list <- function(x, ...) {
  bru_log_message("bru_input(bru_comp_list)", verbosity = 4)
  lapply(x, function(xx) bru_input(xx, ...))
}

#' @return * `bru_input_text(bru_comp_list)`: A list of mapper input definition
#'   text strings, with one entry for each component.
#' @export
#' @rdname bru_input_text
bru_input_text.bru_comp_list <- function(x, ...) {
  bru_log_message("bru_input_text(bru_comp_list)", verbosity = 4)
  lapply(x, function(xx) bru_input_text(xx, ...))
}

#' @describeIn bru_input Computes the component inputs for included components
#' for each observation model.
#'
#' @param lhoods A `bru_obs_list` object containing all observations models.
#' @return * `bru_input(bru_model)`: A list of mapper input values,
#'   with one entry for each observation model, each containing a list
#'   of inputs for the components used by the corresponding observation model.
#' @export
bru_input.bru_model <- function(x, lhoods, ...) {
  bru_input(lhoods, components = as_bru_comp_list(x), ...)
}

#' @rdname bru_input
#' @param components A `bru_comp_list` object containing all components
#'   defined in the model.
#' @return * `bru_input(bru_obs_list)`: A list of mapper input values,
#'   with one entry for each observation model, each containing a list
#'   of inputs for the components used by the corresponding observation model.
#' @export
bru_input.bru_obs <- function(x, components, ...) {
  included <- bru_used(x)[["effect"]]

  bru_input(
    components[included],
    data = x[["data"]],
    mask = bru_data_mask(
      list(data = x[["data"]], data_extra = x[["data_extra"]])
    ),
    ...
  )
}

#' @rdname bru_input
#' @return * `bru_input(bru_obs_list)`: A list of mapper input values,
#'   with one entry for each observation model, each containing a list
#'   of inputs for the components used by the corresponding observation model.
#' @export
bru_input.bru_obs_list <- function(x, components, ...) {
  bru_log_message(
    "Evaluate component inputs for each observation model",
    verbosity = 3L
  )
  lapply(x, function(xx) bru_input(xx, components = components, ...))
}


bru_input_layer <- function(layer,
                            selector = NULL,
                            mask,
                            .envir = parent.frame(),
                            label,
                            e_input) {
  input_layer <- tryCatch(
    rlang::eval_tidy(layer, data = mask, env = .envir),
    error = function(e) {
      e
    }
  )
  if (inherits(input_layer, "error")) {
    stop(paste0(
      "Failed to evaluate 'layer' input '",
      glue_collapse(rlang::as_label(layer), sep = "\n"),
      glue("' for '{label}:layer'.")
    ))
  }
  if (is.null(input_layer) && is.null(selector)) {
    if (label %in% names(e_input)) {
      input_layer <- label
    }
  }
  input_layer
}


#' @export
#' @describeIn inlabru-deprecated `r lifecycle::badge("deprecated")` from
#'   `2.12.0.9023`.
#' Use `bru_input()` instead.
input_eval <- function(...) {
  lifecycle::deprecate_warn(
    "2.12.0.9023",
    "input_eval()",
    "bru_input()"
  )
  bru_input(...)
}

#' @describeIn inlabru-deprecated `r lifecycle::badge("deprecated")` since
#'   version `2.12.0.9023`. Use [bru_input()] instead. Computes the component
#'   inputs for included components for each observation model.
#' @keywords internal
evaluate_inputs <- function(...) {
  lifecycle::deprecate_warn(
    when = "2.12.9023",
    "evaluate_inputs()",
    "bru_input()"
  )
  bru_input(...)
}


#' Summarise component inputs
#'
#' @param object Object to be summarised.
#' @param \dots Passed on to other summary methods.
#' @param verbose logical; If `TRUE`, includes more details of the
#' component definitions. When `FALSE`, only show basic component
#' definition information.  Default `TRUE`.
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @seealso [bru_input()], [bru_comp()]
#' @param label.override character; If not `NULL`, use this label instead of
#' the object's label.
#' @param type character; if non-NULL, added to the output'; `label =
#'   type(input)`.
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_input
#' @export
#' @method format bru_input
format.bru_input <- function(x, verbose = TRUE, ..., label.override = NULL,
                             type = NULL) {
  inp <- x[["input"]]
  if (!is.null(label.override)) {
    lab <- label.override
  } else {
    lab <- x[["label"]]
  }
  if (is.null(inp)) {
    text <- ""
  } else {
    the_input <- glue_collapse(rlang::as_label(inp), sep = "\n")
    if (is.null(lab) || identical(lab, "")) {
      the_lab <- ""
    } else {
      the_lab <- glue("{lab} = ")
    }
    if (is.null(type)) {
      text <- glue(the_lab, the_input)
    } else {
      text <- glue(the_lab, "{type}(", the_input, ")")
    }
  }
  text
}

#' @export
#' @method summary bru_input
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_input
summary.bru_input <- function(object,
                              verbose = TRUE,
                              ...,
                              label.override = NULL) {
  res <- structure(
    list(
      text = format(object,
        verbose = verbose,
        ...,
        label.override = label.override
      )
    ),
    class = "summary_bru_input"
  )
  res
}

#' @export
#' @param x Object to be printed
#' @method print bru_input
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname summary.bru_input
print.bru_input <- function(x, verbose = TRUE, ..., label.override = NULL) {
  cat(
    format(x,
      verbose = verbose,
      ...,
      label.override = label.override
    ),
    "\n",
    sep = ""
  )
  invisible(x)
}

#' @export
#' @rdname summary.bru_comp

print.summary_bru_input <- function(x, ...) {
  if (!is.null(x[["text"]])) {
    cat(x[["text"]], "\n", sep = "")
  }
  invisible(x)
}


#' @title Interface between `bru_input` and `bru_mapper`
#' @description Associate [bru_input] objects with [bru_mapper] objects.
#' @name ibm_input
#' @seealso [bru_input()]
NULL

#' @describeIn ibm_input Add an existing `bru_input` to a `bru_mapper`.
#' @param mapper A `bru_mapper` object.
#' @param input A `bru_input` object, or `NULL` to remove any existing input.
#'   Alternatively, an existing `bru_mapper` object with or without associated
#'   input can be provided, in which case the input from that mapper
#'   is copied.
#' @export
#' @examples
#' (m <- bm_autodetect())
#' ibm_input_available(m)
#' (m <- ibm_input_new(m, cos(x)))
#' ibm_input_available(m)
#' ibm_input_set(bm_linear(), m)
#'
ibm_input_set <- function(mapper, input) {
  stopifnot(inherits(mapper, "bru_mapper"))
  inp <- input
  if (inherits(input, "bru_mapper")) {
    if (ibm_input_available(input)) {
      inp <- ibm_input_get(input)
    } else {
      inp <- NULL
    }
  }
  stopifnot(is.null(inp) || inherits(inp, "bru_input"))
  # Use a special name that is unlikely to conflict with user-defined names.
  mapper[[".input"]] <- inp
  mapper
}
#' @describeIn ibm_input Create and add a `bru_input` to a `bru_mapper`.
#' @param \dots Passed on to [new_bru_input()].
#' @export
ibm_input_new <- function(mapper, ...) {
  stopifnot(inherits(mapper, "bru_mapper"))
  ibm_input_set(mapper, new_bru_input(...))
}
#' @describeIn ibm_input Check if a `bru_input` is associated with a
#' `bru_mapper`.
#' @export
ibm_input_available <- function(mapper) {
  !is.null(mapper[[".input"]])
}
#' @describeIn ibm_input Get the `bru_input` associated with a `bru_mapper`.
#' @export
ibm_input_get <- function(mapper) {
  stopifnot(ibm_input_available(mapper))
  mapper[[".input"]]
}
#' @describeIn bru_input Evaluate the input associated with a `bru_mapper`.
#' @seealso [ibm_input]
#' @export
bru_input.bru_mapper <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input.bru_mapper({label})"),
    verbosity = 5
  )
  if (!ibm_input_available(x)) {
    stop(glue(
      "No input defined for mapper '{label}'.",
      "\n Use `ibm_input_new()` or `ibm_input_set()` to assign one,",
      "\n like `bm_<type>(...) |> ibm_input_new(<expr>, ...)`."
    ))
  }
  inp <- ibm_input_get(x)
  bru_input(inp, ..., label = label)
}
#' @describeIn bru_input Evaluate the inputs for each sub-mapper in a `bm_pipe`
#' object.
#' @export
bru_input.bm_pipe <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input.bm_pipe({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}
#' @describeIn bru_input Evaluate the inputs for each sub-mapper in a `bm_multi`
#' object.
#' @export
bru_input.bm_multi <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input.bm_multi({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}
#' @describeIn bru_input Evaluate the inputs for each sub-mapper in a
#'   `bm_collect` object.
#' @export
bru_input.bm_collect <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input.bm_collect({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}
#' @describeIn bru_input Evaluate the inputs for the sub-mapper in a `bm_repeat`
#' object.
#' @export
bru_input.bm_repeat <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input.bm_repeat({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input(inp, ..., label = label))
  }
  bru_input(x[["mapper"]], ..., label = label)
}
#' @describeIn bru_input Evaluate the inputs for each sub-mapper in a `bm_sum`
#' object.
#' @export
bru_input.bm_sum <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input.bm_sum({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}


#' @describeIn bru_input_text Extract the input definitions associated with a
#'   `bru_mapper`.
#' @seealso [ibm_input]
#' @export
bru_input_text.bru_mapper <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input_text.bru_mapper({label})"),
    verbosity = 5
  )
  if (!ibm_input_available(x)) {
    stop(glue(
      "No input defined for mapper '{label}'.",
      "\n Use `ibm_input_new()` or `ibm_input_set()` to assign one,",
      "\n like `bm_<type>(...) |> ibm_input_new(<expr>, ...)`."
    ))
  }
  inp <- ibm_input_get(x)
  bru_input_text(inp, ..., label = label)
}
#' @rdname bru_input_text
#' @export
bru_input_text.bm_pipe <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input_text.bm_pipe({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input_text(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input_text(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}
#' @rdname bru_input_text
#' @export
bru_input_text.bm_multi <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input_text.bm_multi({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input_text(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input_text(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}
#' @rdname bru_input_text
#' @export
bru_input_text.bm_collect <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input_text.bm_collect({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input_text(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input_text(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}
#' @rdname bru_input_text
#' @export
bru_input_text.bm_repeat <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input_text.bm_repeat({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input_text(inp, ..., label = label))
  }
  bru_input_text(x[["mapper"]], ..., label = label)
}
#' @rdname bru_input_text
#' @export
bru_input_text.bm_sum <- function(x, ..., label = "<unknown>") {
  bru_log_message(
    glue("bru_input_text.bm_sum({label})"),
    verbosity = 5
  )
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    return(bru_input_text(inp, ..., label = label))
  }
  indexing <- names(x$mappers)
  if (is.null(indexing)) {
    indexing <- seq_along(x$mappers)
  } else {
    names(indexing) <- indexing
  }
  lapply(indexing, function(idx) {
    bru_input_text(x$mappers[[idx]], ..., label = glue("{label}:{idx}"))
  })
}


#' @describeIn ibm_input Create a `bru_mapper` placeholder object of class
#'   `bm_autodetect`. The main purpose of this class is to attach [bru_input]
#'   information to it, which is later used to determine a suitable 'real'
#'   mapper type.
#' @export
bm_autodetect <- function() {
  bru_mapper_define(
    list(n = 1L),
    new_class = "bm_autodetect"
  )
}

#' Create an inlabru model object from model components
#'
#' The [inlabru] syntax for model formulae is different from what
#' `INLA::inla` considers a valid.
#' In inla most of the effects are defined by adding an `f(...)` expression to
#' the formula.
#' In [inlabru] the `f` is replaced by an arbitrary (exceptions: `const` and
#' `offset`) string that will determine the label of the effect. See Details for
#' further information.
#'
#' @details
#' For instance
#'
#' `y ~ f(myspde, ...)`
#'
#' in INLA is equivalent to
#'
#' `y ~ myspde(...)`
#'
#' in inlabru.
#'
#' A disadvantage of the inla way is that there is no clear separation between
#' the name of the covariate and the label of the effect. Furthermore, for some
#' models like SPDE it is much more natural to use spatial coordinates as
#' covariates rather than an index into the SPDE vertices. For this purpose
#' [inlabru] provides the `main` argument. For convenience, the `main` argument
#' can be used like the first argument of the f function, e.g., and is the first
#' argument of the component definition.
#' The `INLA` model formula
#'
#' `y ~ f(temperature, model = 'linear')`
#'
#' is equivalent to the `inlabru` component and formula definition
#'
#' `y ~ temperature(temperature, model = 'linear')`
#' and
#' `y ~ temperature(main = temperature, model = 'linear')`
#' as well as
#' `y ~ temperature(model = 'linear')`
#' which sets `main = temperature`.
#'
#' On the other hand, `main` can also be a function mapping, e.g the
#' `sp::coordinates()` function:
#'
#' `y ~ mySPDE(coordinates, ...)`
#'
#' This extracts the coordinates from the data object, and maps it to the latent
#' field via the information given in the `mapper`, which by default is
#' extracted from the `model` object, in the case of `spde` model objects.
#'
#' Morevover, `main` can be any expression that evaluates within your data as an
#' environment.
#' For instance, if your data has columns 'a' and 'b', you can create a fixed
#' effect of 'sin(a+b)' by setting `map` in the following way:
#'
#' `y ~ myEffect(sin(a+b))`
#'
#'
#' @export
#' @param components A [bru_comp_list] object
#' @param lhoods Either a [bru_obs()] object, a [bru_obs_list()] object, or
#'   a list of one or more [bru_obs()] or [bru_obs_list()] objects,
#'   or a list of arguments for a call to [bru_obs()].
#' @param options A [bru_options] options object or a list of options passed
#' on to [bru_options()]
#' @param .envir The environment in which the components are evaluated.
#'
#' @return A [bru_model] object
#' @keywords internal

bru_model <- function(components,
                      lhoods,
                      options = list(),
                      .envir = parent.frame()) {
  options <- bru_call_options(options)

  .response <- extract_response(components)

  # Turn input into a list of components (from existing list, or a special
  # formula)
  components <- bru_comp_list(components, .envir = .envir)

  lhoods <- bru_obs_list_construct(
    lhoods,
    options = options,
    .envir = .envir,
    .response = .response,
    .components = components
  )

  if (length(lhoods) == 0) {
    stop("No observation models provided.")
  }

  inputs <- bru_input(lhoods, components = components, null.on.fail = FALSE)

  # Back up environment
  env <- environment(components)

  # Complete the used component definitions based on data
  included <- bru_used(lhoods)
  included <- union(included[["effect"]], included[["latent"]])
  components <- bru_comp_list(
    components[included],
    lhoods = lhoods,
    inputs = inputs
  )

  # Create joint formula that will be used by inla
  formula <- bru_inla_formula(components)
  formula <- update.formula(formula, BRU_response ~ .)

  # Restore environment
  environment(components) <- env
  environment(formula) <- env

  # Check linearity
  for (lh in seq_along(lhoods)) {
    if (bru_is_linear(lhoods[[lh]])) {
      used_lh <- bru_used(lhoods[[lh]])
      lhoods[[lh]][["pred_expr"]][["is_linear"]] <-
        (length(used_lh$latent) == 0) &&
          all(vapply(components[used_lh$effect], function(cmp) {
            ibm_is_linear(cmp$mapper)
          }, TRUE))
    }
    lhoods[[lh]] <- bru_compat_pre_2_14_bru_obs(lhoods[[lh]])
  }

  # Make model
  mdl <- structure(
    list(
      effects = components,
      lhoods = lhoods,
      inputs = inputs,
      formula = formula
    ),
    class = "bru_model"
  )

  mdl
}


#' @export
#' @method summary bru_model
#' @param object Object to operate on
#' @param \dots Arguments passed on to other methods
#' @rdname bru_model
summary.bru_model <- function(object, ...) {
  result <- structure(
    list(
      components =
        summary(as_bru_comp_list(object), ...),
      lhoods = {
        # Once lhoods is moved into bru_model, the NULL part will no longer be
        # needed, and as_bru_obs_list() can be used instead.
        bru_obs_lst <- object[["lhoods"]]
        if (is.null(bru_obs_lst)) {
          NULL
        } else {
          summary(bru_obs_lst, ...)
        }
      }
    ),
    class = "summary_bru_model"
  )
  result
}

#' @export
#' @param x An object to be printed
#' @rdname bru_model
print.summary_bru_model <- function(x, ...) {
  cat("Latent components:\n")
  print(x[["components"]])
  if (!is.null(x[["lhoods"]])) {
    cat("Observation models:\n")
    print(x[["lhoods"]])
  }
  invisible(x)
}

#' @export
#' @rdname bru_model
print.bru_model <- function(x, ...) {
  print(summary(x))
  invisible(x)
}


#' Evaluate or sample from a posterior result given a model and locations
#'
#' @export
#' @param model A [bru] model
#' @param state list of state lists, as generated by [evaluate_state()]
#' @param data A `list`, `data.frame`, or `Spatial*DataFrame`, with coordinates
#' and covariates needed to evaluate the predictor.
#' @param data_extra Additional data for the predictor evaluation
#' @param input Precomputed inputs list for the components
#' @param comp_simple Precomputed [bm_list] of simplified mappers for the
#' components
#' @param predictor A formula or a [bru_pred_expr] expression to be evaluated
#'   given the posterior or for each sample thereof. The default (`NULL`)
#'   returns a `data.frame` containing the sampled effects. In case of a formula
#'   the right hand side is used for evaluation.
#' @param format character; determines the storage format of predictor output.
#' Available options:
#' * `"auto"` If the first evaluated result is a vector or single-column matrix,
#'   the "matrix" format is used, otherwise "list".
#' * `"matrix"` A matrix where each column contains the evaluated predictor
#' expression for a state.
#' * `"list"` A list where each element contains the evaluated predictor
#' expression for a state.
#' @param n Number of samples to draw.
#' @param seed If seed != 0L, the random seed
#' @param num.threads Specification of desired number of threads for parallel
#' computations. Default NULL, leaves it up to INLA.
#' When seed != 0, overridden to "1:1:1"
#' @param used A [bru_used()] object, or NULL (default)
#' @param n_pred integer. If provided, scalar predictor results are expanded to
#' vectors of length `n_pred`.
#' @param \dots Additional arguments passed on to `inla.posterior.sample`
#' @details * `evaluate_model` is a wrapper to evaluate model state, A-matrices,
#' effects, and predictor, all in one call.
#'
#' @keywords internal
#' @rdname evaluate_model
evaluate_model <- function(model,
                           state,
                           data = NULL,
                           data_extra = NULL,
                           input = NULL,
                           comp_simple = NULL,
                           predictor = NULL,
                           format = NULL,
                           used = NULL,
                           n_pred = NULL,
                           ...) {
  comp_lst <- as_bru_comp_list(model)
  if (inherits(predictor, "bru_pred_expr")) {
    if (!is.null(used)) {
      warning(
        paste0(
          "Overriding `used` argument to `evaluate_model()` ",
          "in favour of `bru_used(predictor)`."
        )
      )
    }
    used <- bru_used(predictor, labels = names(comp_lst))
    predictor <- bru_pred_expr(predictor, format = "formula")
  } else {
    used <- bru_used(used, labels = names(comp_lst))
  }

  if (is.null(state)) {
    stop("Not enough information to evaluate model states.")
  }
  if (is.null(input)) {
    input <- bru_input(
      comp_lst[used$effect],
      data = data
    )
  }
  if (is.null(comp_simple) && !is.null(input)) {
    comp_simple <- ibm_simplify(
      comp_lst[used$effect],
      input = input,
      inla_f = TRUE
    )
  }
  if (is.null(comp_simple)) {
    effects <- NULL
  } else {
    effects <-
      lapply(
        state,
        function(x) {
          evaluate_effect_single_state(
            comp_simple,
            state = x,
            input = input
          )
        }
      )
  }

  if (is.null(predictor)) {
    return(effects)
  }

  values <- evaluate_predictor(
    model,
    state = state,
    data = data,
    data_extra = data_extra,
    effects = effects,
    predictor = predictor,
    format = format,
    used = used,
    n_pred = n_pred
  )

  values
}


#' @details * `evaluate_state` evaluates model state properties or samples
#' @param result A `bru` object from [bru()] or [lgcp()]
#' @param property Property of the model components to obtain value from.
#' Default: "mode". Other options are "mean", "0.025quant", "0.975quant",
#' "sd" and "sample". In case of "sample" you will obtain samples from the
#' posterior (see `n` parameter). If `result` is `NULL`, all-zero vectors are
#' returned for each component.
#' @param internal_hyperpar logical; If `TRUE`, return hyperparameter properties
#' on the internal scale. Currently ignored when `property="sample"`.
#' Default is `FALSE`.
#' @export
#' @rdname evaluate_model
#' @keywords internal
evaluate_state <- function(model,
                           result,
                           property = "mode",
                           n = 1,
                           seed = 0L,
                           num.threads = NULL,
                           internal_hyperpar = FALSE,
                           ...) {
  # Evaluate random states, or a single property
  if (property == "sample") {
    state <- post.sample.structured(result,
      n = n, seed = seed,
      num.threads = num.threads,
      ...
    )
  } else if (is.null(result)) {
    state <- list(lapply(
      as_bru_comp_list(model),
      function(x) {
        rep(0.0, ibm_n(x[["mapper"]]))
      }
    ))
  } else {
    state <- list(extract_property(
      result = result,
      property = property,
      internal_hyperpar = internal_hyperpar
    ))
  }

  state
}


#' @export
#' @rdname evaluate_effect
evaluate_effect_single_state <- function(...) {
  UseMethod("evaluate_effect_single_state")
}

#' Evaluate a component effect
#'
#' Calculate latent component effects given some data and the state of the
#' component's internal random variables.
#'
#' @export
#' @keywords internal
#' @param component A [bru_mapper], [bru_comp], or
#'   [bm_list].
#' @param input Pre-evaluated component input
#' @param state Specification of one latent variable state:
#' * `evaluate_effect_single_state.bru_mapper`:
#'   A vector of the latent component state.
#' * `evaluate_effect_single_state.*_list`: list of named state vectors.
#' @param \dots Optional additional parameters, e.g. `inla_f`. Normally unused.
#' @param label Option label used for any warning messages, specifying the
#' affected component.
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com} and
#' Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname evaluate_effect
#' @keywords internal

evaluate_effect_single_state.bru_mapper <- function(component, input, state,
                                                    ...,
                                                    label = NULL) {
  values <- ibm_eval(component, input = input, state = state, ...)

  not_ok <- ibm_invalid_output(
    component,
    input = input,
    state = state
  )
  if (any(not_ok)) {
    if (is.null(label)) {
      warning("Inputs for a mapper give some invalid outputs.",
        immediate. = TRUE
      )
    } else {
      warning("Inputs for '", label, "' give some invalid outputs.",
        immediate. = TRUE
      )
    }
  }

  as.vector(as.matrix(values))
}

#' @return * `evaluate_effect_single_state.bm_list`: A list of
#'   evaluated component effect values
#' @export
#' @rdname evaluate_effect
#' @keywords internal
evaluate_effect_single_state.bm_list <- function(components,
                                                 input,
                                                 state,
                                                 ...) {
  result <- list()
  for (label in names(components)) {
    result[[label]] <- evaluate_effect_single_state(
      components[[label]],
      input = input[[label]],
      state = state[[label]],
      ...,
      label = label
    )
  }
  result
}

#' @export
#' @rdname evaluate_effect
#' @keywords internal
evaluate_effect_single_state.bru_comp_list <- function(components,
                                                       input,
                                                       state,
                                                       ...) {
  comp_simple <- ibm_simplify(components, input = input, state = state, ...)
  evaluate_effect_single_state(comp_simple, input = input, state = state, ...)
}


#' Evaluate component effects or expressions
#'
#' Evaluate component effects or expressions, based on a bru model and one or
#' several states of the latent variables and hyperparameters.
#'
#' @param data A `list`, `data.frame`, or `Spatial*DataFrame`, with coordinates
#' and covariates needed to evaluate the model.
#' @param data_extra Additional data for the predictor evaluation. Variables
#' with the same name as in `data` will be ignored, unless accessed via
#' `.data_extra.[["name"]]` or `.data_extra.$name`, or via pronouns;
#'  see Details.
#' @param state A list where each element is a list of named latent state
#' information, as produced by [evaluate_state()]
#' @param effects A list where each element is list of named evaluated effects,
#' each computed by [evaluate_effect_single_state.bru_comp_list()]
#' @param predictor Either a formula or [bru_pred_expr] expression
#' @param used A [bru_used()] object, or NULL (default)
#' @param format character; determines the storage format of the output.
#' Available options:
#' * `"auto"` If the first evaluated result is a vector or single-column matrix,
#'   the "matrix" format is used, otherwise "list".
#' * `"matrix"` A matrix where each column contains the evaluated predictor
#' expression for a state.
#' * `"list"` A list where each column contains the evaluated predictor
#' expression for a state.
#'
#' Default: "auto"
#' @param n_pred integer. If provided, scalar predictor results are expanded to
#' vectors of length `n_pred`.
#' @details For each component, e.g. "name", the latent state values are
#'   available as `name_latent`, and arbitrary evaluation can be done with
#'   `name_eval(...)`, see [bru_comp_eval()].
#'
#'   The evaluation supports several [rlang::as_data_pronoun()] data masking
#'   pronouns, to access variables from different data sources, and some of
#'   these also have corresponding full objects, with an appended `.` in the
#'   name. The full objects can be passed as arguments to functions.
#'   \describe{
#'   \item{.effect/.effect.}{refers to the `effects` vectors}
#'   \item{.latent/.latent.}{refers to the latent state vectors}
#'   \item{.data/.data.}{refers to the main `data` argument}
#'   \item{.data_extra/.data_extra.}{refers to the `data_extra` argument}
#'   \item{.env}{refers to the evaluation environment of the predictor}
#'   }
#' @return A list or matrix is returned, as specified by `format`
#' @keywords internal
#' @rdname evaluate_predictor
evaluate_predictor <- function(model,
                               state,
                               data,
                               data_extra,
                               effects,
                               predictor,
                               used = NULL,
                               format = "auto",
                               n_pred = NULL) {
  stopifnot(inherits(model, "bru_model"))
  format <- match.arg(format, c("auto", "matrix", "list"))
  comp_lst <- as_bru_comp_list(model)
  if (inherits(predictor, "bru_pred_expr")) {
    if (!is.null(used)) {
      warning(
        paste0(
          "Overriding `used` argument to `evaluate_predictor()` ",
          "in favour of `bru_used(predictor)`."
        )
      )
    }
    used <- bru_used(predictor, labels = names(comp_lst))
    predictor <- bru_pred_expr(predictor, format = "formula")
  }
  pred.envir <- environment(predictor)
  if (inherits(predictor, "formula")) {
    pred_text <- as.character(predictor)
    pred_text <- pred_text[length(pred_text)]
    predictor <- rlang::parse_expr(pred_text)
  }
  formula.envir <- environment(model$formula)
  enclos <-
    if (!is.null(pred.envir)) {
      pred.envir
    } else if (!is.null(formula.envir)) {
      formula.envir
    } else {
      parent.frame()
    }

  used <- bru_used(used, labels = names(comp_lst))

  # General evaluation environment
  envir <- new.env(parent = enclos)
  # Find .data. first,
  # then data variables,
  # then pred.envir variables (via enclos),
  # then formula.envir (via enclos if pred.envir is NULL):
  #  for (nm in names(pred.envir)) {
  #    assign(nm, pred.envir[[nm]], envir = envir)
  #  }
  #
  # Note: Since 2.7.0.9019, no longer converts Spatial*DataFrame to data frame
  # here; coordinates must be accessed via sp::coordinates() if needed.

  # Rename component states from label to label_latent
  state_names <- as.list(expand_labels(
    names(state[[1]]),
    names(comp_lst),
    suffix = "_latent"
  ))
  names(state_names) <- names(state[[1]])

  # Construct _eval function names
  eval_names <- as.list(expand_labels(
    intersect(names(state[[1]]), names(comp_lst)),
    intersect(names(state[[1]]), names(comp_lst)),
    suffix = "_eval"
  ))
  names(eval_names) <- intersect(names(state[[1]]), names(comp_lst))

  eval_fun_factory <-
    function(.comp, .envir, .enclos) {
      .is_offset <- .comp$main$type %in% c("offset", "const")
      .is_iid <- .comp$main$type %in% c("iid")
      .mapper <- .comp$mapper
      .label <- paste0(.comp$label, "_latent")
      .iid_precision <- paste0("Precision_for_", .comp$label)
      .iid_cache <- list()
      .iid_cache_index <- NULL
      eval_fun <- function(main, group = NULL,
                           replicate = NULL,
                           weights = NULL,
                           .state = NULL) {
        n_input <- ibm_n_output(
          .mapper[["mappers"]][["core"]][["mappers"]][["main"]],
          input = main
        )
        if (is.null(group)) {
          group <- rep(1, n_input)
        }
        if (is.null(replicate)) {
          replicate <- rep(1, n_input)
        }
        if (!.is_offset && is.null(.state)) {
          .state <- rlang::eval_tidy(
            rlang::parse_expr(.label),
            data = data_mask,
            env = .envir
          )
        }
        .input <- list(
          core = list(
            main = main,
            group = group,
            replicate = replicate
          ),
          scale = weights
        )

        .values <- ibm_eval(
          .mapper,
          input = .input,
          state = .state
        )
        if (!.is_iid) {
          not_ok <- ibm_invalid_output(
            .mapper[["mappers"]][[1]],
            input = .input[[1]],
            state = .state
          )
          if (any(not_ok)) {
            warning(
              "Inputs for `ibm_eval()` for '",
              .comp[["label"]],
              '" give some invalid outputs.'
            )
          }
        } else { # .is_iid, invalid indices give new samples
          # Check for known invalid output elements, based on the
          # initial mapper (subsequent mappers in the component pipe
          # are assumed to keep the same length and validity)
          not_ok <- ibm_invalid_output(
            .mapper[["mappers"]][[1]],
            input = .input[[1]],
            state = .state
          )
          if (any(not_ok)) {
            .cache_state_index <- rlang::eval_tidy(
              rlang::parse_expr(".cache_state_index"),
              data = data_mask,
              env = .envir
            )
            if (!identical(.cache_state_index, .iid_cache_index)) {
              .iid_cache_index <<- .cache_state_index
              .iid_cache <<- list()
            }
            key <- as.character(main[not_ok])
            not_cached <- !(key %in% names(.iid_cache))
            if (any(not_cached)) {
              .prec <- rlang::eval_tidy(
                rlang::parse_expr(.iid_precision),
                data = data_mask,
                env = .envir
              )
              for (k in unique(key[not_cached])) {
                .iid_cache[k] <<- rnorm(1, mean = 0, sd = .prec^-0.5)
              }
            }
            .values[not_ok] <- vapply(
              key,
              function(k) .iid_cache[[k]],
              0.0
            )
          }
        }

        as.matrix(.values)
      }
      eval_fun
    }
  eval_list <- list()
  for (nm in names(eval_names)) {
    eval_list[[eval_names[[nm]]]] <-
      eval_fun_factory(
        comp_lst[[nm]],
        .envir = envir,
        .enclos = enclos
      )
  }

  # Remove problematic objects:
  problems <- c(".Random.seed")
  remove(list = intersect(names(envir), problems), envir = envir)

  n <- length(state)
  for (k in seq_len(n)) {
    state_df <- state[[k]]
    names(state_df) <- state_names[names(state_df)]
    data_mask <- bru_data_mask(
      list(
        effect = effects[[k]],
        state_df,
        latent = state[[k]],
        data = data,
        data_extra = data_extra,
        eval_list,
        # Keep track of the iteration index so the iid cache can be
        # invalidated
        list(.cache_state_index = k)
      )
    )

    result_ <- rlang::eval_tidy(predictor,
      data = data_mask,
      env = envir
    )
    if (!is.null(n_pred) && is.numeric(result_) && length(result_) == 1) {
      result_ <- rep(result_, n_pred)
    }
    if (k == 1) {
      if (identical(format, "auto")) {
        if ((is.vector(result_) && !is.list(result_)) ||
          (is.matrix(result_) && (NCOL(result_) == 1))) {
          format <- "matrix"
        } else {
          format <- "list"
        }
      }
      if (identical(format, "matrix")) {
        result <- matrix(0.0, NROW(result_), n)
        rownames(result) <- row.names(as.matrix(result_))
      } else if (identical(format, "list")) {
        result <- vector("list", n)
      }
    }
    if (identical(format, "list")) {
      result[[k]] <- result_
    } else {
      result[, k] <- result_
    }
  }

  result
}


#' Evaluate component values in predictor expressions
#'
#' In predictor expressions, `name_eval(...)` can be used to evaluate
#' the effect of a component called "name".
#'
#' @aliases bru_component_eval
#' @param main,group,replicate,weights Specification of where to evaluate a
#'   component. The four inputs are passed on to the joint `bru_mapper` for the
#'   component, as
#'  ```
#'  list(mapper = list(
#'         main = main,
#'         group = group,
#'         replicate = replicate),
#'       scale = weights)
#'  ```
#'  NOTE: If you have model component with the same name as a data variable you
#'  want to supply as input to `name_eval()`, you need to use
#'  `.data.[["myvar"]]` to access it. Otherwise, it will try to use the other
#'  component effect as input, which is ill-defined.
#' @param .state The internal component state. Normally supplied automatically
#' by the internal methods for evaluating inlabru predictor expressions.
#' @return A vector of values for a component
#' @examples
#' if (bru_safe_inla() &&
#'   require("sf", quietly = TRUE) &&
#'   requireNamespace("sn", quietly = TRUE)) {
#'   mesh <- fmesher::fm_mesh_2d_inla(
#'     cbind(0, 0),
#'     offset = 2,
#'     max.edge = 2.5
#'   )
#'   spde <- INLA::inla.spde2.pcmatern(
#'     mesh,
#'     prior.range = c(1, NA),
#'     prior.sigma = c(0.2, NA)
#'   )
#'   set.seed(12345L)
#'   data <- sf::st_as_sf(
#'     data.frame(
#'       x = runif(50),
#'       y = runif(50),
#'       z = rnorm(50)
#'     ),
#'     coords = c("x", "y")
#'   )
#'   fit <- bru(
#'     z ~ -1 + field(geometry, model = spde),
#'     family = "gaussian", data = data,
#'     options = list(control.inla = list(int.strategy = "eb"))
#'   )
#'   pred <- generate(
#'     fit,
#'     newdata = data.frame(A = 0.5, B = 0.5),
#'     formula = ~ field_eval(cbind(A, B)),
#'     n.samples = 1L
#'   )
#' }
bru_comp_eval <- function(main,
                          group = NULL,
                          replicate = NULL,
                          weights = NULL,
                          .state = NULL) {
  stop(paste0(
    "In your predictor expression, use 'mylabel_eval(...)' instead of\n",
    "'bru_comp_eval(...)'.  See ?bru_comp_eval for more information."
  ))
}


#' @include mappers.R

#' @title Mapper methods for model objects
#' @description
#' Methods for the `ibm_linear()` and `ibm_simplify()` methods for
#' [bru] model objects and related classes.
#'
#' @inheritParams bru_mapper_generics
#' @inheritParams ibm_linear
#' @inheritParams ibm_simplify
#'
#' @name bru_model_mapper_methods
#' @rdname bru_model_mapper_methods
NULL

#' @describeIn bru_model_mapper_methods Returns a list (one element per
#'   observation model) of [bm_list] objects, each with one [bm_taylor]
#'   entry for each included component.
#' @export
ibm_linear.bru_model <- function(mapper, input, state = NULL, ...) {
  model <- mapper
  stopifnot(inherits(model, "bru_model"))
  bru_log_message(
    paste0("Linearise components for each observation model"),
    verbosity = 3
  )
  mappers <-
    lapply(
      input,
      function(inp) {
        ibm_linear(
          as_bru_comp_list(model),
          input = inp,
          state = state,
          ...
        )
      }
    )

  mappers
}

#' @rdname bru_model_mapper_methods
#' @export
ibm_linear.bru_comp_list <- function(mapper, input, state = NULL, ...) {
  comp <- mapper
  included <- parse_inclusion(
    names(comp),
    names(input),
    NULL
  )

  mappers <- ibm_linear(
    as_bm_list(comp[included]),
    input = input[included],
    state = state[included],
    ...
  )

  mappers
}

#' @rdname ibm_linear
#' @export
#'
ibm_linear.bru_comp <- function(mapper,
                                input,
                                state = NULL,
                                ...) {
  bru_log_message(
    paste0("Linearise component '", mapper[["label"]], "'"),
    verbosity = 5
  )
  if (is.null(state)) {
    state <- rep(0, ibm_n(mapper[["mapper"]]))
  }
  ibm_linear(mapper[["mapper"]], input = input, state = state, ...)
}

#' @describeIn bru_model_mapper_methods Returns a list (one element per
#'   observation model) of [bm_list] objects, each with one [bru_mapper]
#'   entry for each included component.
#' @export
ibm_simplify.bru_model <- function(mapper, input = NULL, state = NULL, ...) {
  model <- mapper
  bru_log_message(
    paste0("Simplify component mappers for each observation model"),
    verbosity = 3
  )
  mappers <-
    lapply(
      input,
      function(inp) {
        ibm_simplify(
          as_bru_comp_list(model),
          input = inp,
          state = state,
          ...
        )
      }
    )

  mappers
}

#' @rdname bru_model_mapper_methods
#' @export
ibm_simplify.bru_comp <- function(mapper,
                                  input = NULL,
                                  state = NULL,
                                  ...) {
  bru_log_message(
    paste0("Linearise component '", mapper[["label"]], "'"),
    verbosity = 5
  )
  if (is.null(state)) {
    state <- rep(0, ibm_n(mapper[["mapper"]]))
  }
  ibm_simplify(mapper[["mapper"]], input = input, state = state, ...)
}


#' @rdname bru_model_mapper_methods
#' @export
#'
ibm_simplify.bru_comp_list <- function(mapper,
                                       input = NULL,
                                       state = NULL,
                                       ...) {
  comp <- mapper
  included <- parse_inclusion(names(comp), names(input), NULL)

  mappers <- ibm_simplify(
    as_bm_list(comp[included]),
    input = input[included],
    state = state[included],
    ...
  )

  mappers
}


# @title Mapper methods for model objects
# @description
# Methods for the `ibm_linear()` and `ibm_simplify()` methods for
# [bru] model objects and related classes.
#
#' @export
#' @rdname bru_model_mapper_methods
#' @export
#'
ibm_linear.bm_list <- function(mapper, input, state = NULL, ...) {
  label <- names(mapper)
  if (is.null(label)) {
    label <- as.character(seq_along(mapper))
  }
  mappers <- lapply(
    seq_along(mapper),
    function(k) {
      bru_log_message(
        paste0("Linearise component '", label[k], "'"),
        verbosity = 4
      )
      ibm_linear(
        mapper[[k]],
        input[[k]],
        state = NULL,
        ...
      )
    }
  )
  names(mappers) <- names(mapper)
  as_bm_list(mappers)
}

#' @rdname bru_model_mapper_methods
#' @export
ibm_simplify.bm_list <- function(mapper, input = NULL, state = NULL, ...) {
  label <- names(mapper)
  if (is.null(label)) {
    label <- as.character(seq_along(mapper))
  }
  mappers <- lapply(
    seq_along(mapper),
    function(k) {
      bru_log_message(
        paste0("Simplify component '", label[k], "'"),
        verbosity = 4
      )
      ibm_simplify(
        mapper[[k]],
        input[[k]],
        state = NULL,
        ...
      )
    }
  )
  names(mappers) <- names(mapper)
  as_bm_list(mappers)
}

#' @include deprecated.R

# Generics ----

#' @title Constructors for `bru_mapper` objects
#' @param \dots Arguments passed on to sub-methods, or used for special
#'   purposes, see details for each function below.
#' @export
#' @seealso [bru_mapper_generics] for generic methods,
#' the individual mapper pages for special method implementations, and
#' [bru_get_mapper] for hooks to extract mappers from latent model object
#' class objects.
#' @describeIn bru_mapper
#' Generic mapper S3 constructor, used for constructing mappers for special
#' objects. See below for details of the default constructor
#' [bru_mapper_define()] that can be used to define new mapper clesses in user
#' code. To extracting mappers for latent component models, see
#' [bru_get_mapper()].
#'
#' @returns
#' A `bru_mapper` object.
#' @family mappers
#' @examples
#' mapper <- bm_index(5)
#' ibm_jacobian(mapper, input = c(1, 3, 4, 5, 2))
bru_mapper <- function(...) {
  UseMethod("bru_mapper")
}

#' Generic methods for bru_mapper objects
#'
#' @description
#' A `bru_mapper` sub-class implementation must provide an
#' [ibm_jacobian()] method. If the model size 'n' and definition
#' values 'values' are stored in the object itself, default methods [ibm_n()]
#' and [ibm_values()] are
#' available. Otherwise the
#' [ibm_n()] and [ibm_values()] methods also need to be provided.
#'
#' @seealso [bru_mapper] for constructor methods, and
#' [bru_get_mapper] for hooks to extract mappers from latent model object
#' class objects.
#'
#' @name bru_mapper_generics
#' @family mapper methods
NULL


#' @param mapper A mapper S3 object, inheriting from `bru_mapper`.
#' @param inla_f logical; when `TRUE` for `ibm_n()` and `ibm_values()`, the
#' result must be compatible with the `INLA::f(...)` and corresponding
#' `INLA::inla.stack(...)` constructions.  For `ibm_{eval,jacobian,linear}`,
#' the `input` interpretation may be different.
#' Implementations do not normally need to do anything different, except
#' for mappers of the type needed for hidden multicomponent models such
#' as "bym2", which can be handled by `bm_collect`.
#' @param \dots Arguments passed on to other methods
#' @title Size of the latent vector of a mapping
#' @description
#' Implementations must return the size of the latent vector
#' being mapped to.
#' @export
#' @inheritParams ibm_n_output
#' @family mapper methods
ibm_n <- function(mapper, inla_f = FALSE, ...) {
  UseMethod("ibm_n")
}

#' @title Output size of a mapping
#' @description
#' Implementations must return an integer denoting the
#' mapper output length.
#' The default implementation returns `NROW(input)`.
#' Mappers such as `bm_multi` and `bm_collect`,
#' that can accept `list()` inputs require their own method implementations.
#' @param state A vector of latent state values for the mapping,
#' of length `ibm_n(mapper, inla_f = FALSE)`
#' @param input Data input for the mapper.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_n_output <- function(mapper, input, state = NULL, inla_f = FALSE, ...) {
  UseMethod("ibm_n_output")
}


#' @title Value vector for a mapping
#' @description
#' When `inla_f=TRUE`, implementations must return a vector that
#' would be interpretable by an `INLA::f(..., values = ...)` specification.
#' The exception is the method for `bm_multi`, that returns a
#' multi-column data frame if `multi=TRUE`.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
ibm_values <- function(mapper, inla_f = FALSE, ...) {
  UseMethod("ibm_values")
}


#' @title Check if a mapper is linear/affine
#' @description
#' Implementations must return `TRUE` or `FALSE`.
#' If `TRUE` (returned by the default method unless the mapper
#' contains an `is_linear` variable), users of the mapper
#' may assume the mapper is linear/affine.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
ibm_is_linear <- function(mapper, ...) {
  UseMethod("ibm_is_linear")
}
#' @title Check if a mapper is rowwise
#' @description
#' Implementations must return `TRUE` or `FALSE`.
#' If `TRUE` (returned by the default method unless the mapper
#' contains an `is_rowwise` variable), users of the mapper
#' may assume the mapper uses its inputs in "rowwise" manner, so that
#' blockwise evaluation is always possible.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
ibm_is_rowwise <- function(mapper, ...) {
  UseMethod("ibm_is_rowwise")
}

#' @title Jacobian of a mapper
#' @description
#' Implementations must return a (sparse) matrix of size
#' `ibm_n_output(mapper, input, inla_f)`
#' by `ibm_n(mapper, inla_f = FALSE)`. The `inla_f=TRUE` argument should
#' only affect the allowed type of input format.
#' @export
#' @family mapper methods
#' @param sub_lin Internal, optional pre-computed sub-mapper information
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_jacobian <- function(mapper, input, state = NULL, inla_f = FALSE, ...) {
  UseMethod("ibm_jacobian")
}


#' @title Compute a mapper linearisation
#' @description
#' Implementations must return a [bm_taylor] object
#' The linearisation information includes `offset`, `jacobian`, and `state0`.
#' The state information indicates for which state the `offset` was evaluated,
#' with `NULL` meaning all-zero.
#' The linearised mapper output is defined as
#' ```
#' effect(input, state) =
#'   offset(input, state0) + jacobian(input, state0) %*% (state - state0)
#' ```
#' The default method calls [ibm_eval()] and [ibm_jacobian()] to generate
#' the needed information.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_linear <- function(mapper, input, state = NULL, ...) {
  UseMethod("ibm_linear")
}

#' @title Simplify a mapper
#' @description
#' Implementations must return a [bru_mapper] object.
#' The default method returns the result of [ibm_linear()] for linear/affine
#' mappers, and the original `mapper` for non-linear mappers.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_simplify <- function(mapper, input = NULL, state = NULL, ...) {
  UseMethod("ibm_simplify")
}

#' @title Evaluate a mapping
#' @description
#' Implementations must return a vector of length [ibm_n_output()].
#' The `input` contents must
#' be in a format accepted by [ibm_jacobian()]
#' for the mapper.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_eval <- function(mapper, input, state = NULL, ...) {
  UseMethod("ibm_eval")
}

#' @title Evaluate a mapper and its Jacobian
#' @description
#' Implementations must return a list with elements `offset` and `jacobian`.
#' The `input` contents must
#' be in a format accepted by [ibm_jacobian()]
#' for the mapper.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_eval2 <- function(mapper, input, state = NULL, ...) {
  UseMethod("ibm_eval2")
}

#' @title Names of submapper
#' @description
#' Implementations must return a character vector of sub-mapper names, or
#' `NULL`. Intended for providing information about multi-mappers and mapper
#' collections.
#' @export
#' @examples
#' # ibm_names
#' mapper <- bm_multi(list(A = bm_index(2), B = bm_index(2)))
#' ibm_names(mapper)
#' ibm_names(mapper) <- c("new", "names")
#' ibm_names(mapper)
#' @family mapper methods
#' @inheritParams ibm_n
ibm_names <- function(mapper) {
  UseMethod("ibm_names")
}

#' @param value a character vector of the same length as the number
#' of sub-mappers in the mapper
#' @describeIn ibm_names Set mapper names.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
`ibm_names<-` <- function(mapper, value) {
  UseMethod("ibm_names<-")
}


#' @title Find index subset of INLA visible states
#' @description
#' Implementations must return a logical vector of `TRUE/FALSE` for
#' the subset such that, given the full A matrix and values output,
#' `A[, subset, drop = FALSE]` and `values[subset]`
#' (or `values[subset, , drop = FALSE]` for data.frame values) are equal
#' to the `inla_f = TRUE` version of A and values. The default method uses
#' the `ibm_values` output to construct the subset indexing.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
ibm_inla_subset <- function(mapper, ...) {
  UseMethod("ibm_inla_subset")
}

#' @title Detect invalid input to a mapper
#' @description
#' Implementations should return a logical vector of length
#' `ibm_n_output(mapper, input, state, ...)` indicating which, if any,
#' output elements of `ibm_eval(mapper, input, state, ...)` are known to be
#' invalid.
#' For for multi/collect mappers, a list, when given a `multi=TRUE` argument.
#' @export
#' @family mapper methods
#' @inheritParams ibm_n
#' @inheritParams ibm_n_output
ibm_invalid_output <- function(mapper, input, state, ...) {
  UseMethod("ibm_invalid_output")
}


#' Methods for mapper lists
#'
#' `bru_mapper` lists can be combined into `bm_list` lists.
#' @name bm_list
#' @param \dots Objects to be combined.
#' @param x Object to convert
#' @export
as_bm_list <- function(x) {
  UseMethod("as_bm_list")
}

#' @rdname bm_list
#' @export
as_bm_list.list <- function(x) {
  stopifnot(all(vapply(
    x,
    function(x) inherits(x, "bru_mapper"),
    TRUE
  )))
  class(x) <- c("bm_list", "list")
  x
}

#' @rdname bm_list
#' @export
as_bm_list.bm_list <- function(x) {
  x
}

#' @rdname bm_list
#' @export
as_bm_list.bru_comp_list <- function(x) {
  if (is.null(names(x))) {
    names(x) <- vapply(x, function(y) {
      y[["label"]]
    }, "")
  }
  mappers <- as_bm_list(lapply(x, function(y) {
    y[["mapper"]]
  }))
  mappers
}

#' Methods for mapper extraction
#'
#' Extract a mapper from another object
#' @name as_bru_mapper
#' @param x Object to convert/extract
#' @returns A `bru_mapper` object
#' @export
as_bru_mapper <- function(x) {
  UseMethod("as_bru_mapper")
}

#' @rdname as_bru_mapper
#' @export
as_bru_mapper.bru_mapper <- function(x) {
  x
}

#' @rdname as_bru_mapper
#' @export
as_bru_mapper.bru_comp <- function(x) {
  x[["mapper"]]
}

#' @rdname as_bru_mapper
#' @export
#' @examples
#' # Extract a mapper from a `bru_subcomp` object
#' as_bru_mapper(bru_comp("x", x, mapper = bm_index(4))$main)
#'
as_bru_mapper.bru_subcomp <- function(x) {
  x[["mapper"]]
}

#' @export
#' @describeIn bm_list The `...` arguments should be `bru_mapper`
#' objects.
#' @returns A `bm_list` object
#' @examples
#' m <- c(A = bm_const(), B = bm_scale())
#' str(m)
#' str(m[2])
`c.bru_mapper` <- function(...) {
  as_bm_list(list(...))
}

#' @export
#' @describeIn bm_list The `...` arguments should be `bm_list`
#' objects.
`c.bm_list` <- function(...) {
  stopifnot(all(vapply(
    list(...),
    function(x) inherits(x, "bm_list"),
    TRUE
  )))
  object <- NextMethod()
  class(object) <- c("bm_list", "list")
  object
}

#' @export
#' @param x `bm_list` object from which to extract element(s)
#' @param i indices specifying elements to extract
#' @describeIn bm_list Extract sub-list
`[.bm_list` <- function(x, i) {
  object <- NextMethod()
  class(object) <- c("bm_list", "list")
  object
}


# Summaries ----

ibm_shortname <- function(mapper, ...) {
  cls1 <- class(mapper)[1]
  if (any(grepl("^bm_", cls1))) {
    shortname <- sub("^bm_", "", cls1)
  } else if (any(grepl("^bru_mapper_", cls1))) {
    shortname <- sub("^bru_mapper_", "", cls1)
  } else {
    shortname <- character(0)
  }
  if (nchar(shortname) == 0) {
    shortname <- "default"
  }
  shortname
}


#' @title mapper object summaries
#'
#' @param x Object to format/print
#' @param object Object to summarise
#' @param \dots Unused arguments
#' @param prefix character prefix for each line. Default `""`.
#' @param initial character prefix for the first line. Default `initial=prefix`.
#' @param depth The recursion depth for multi/collection/pipe mappers. Default
#'   1, to only show the collection, and not the contents of the sub-mappers.
#' @export
#' @method format bru_mapper
#' @rdname bm_summary
format.bru_mapper <- function(x, ...,
                              prefix = "",
                              initial = prefix,
                              depth = 1) {
  mapper_new <- make_bm_class_from_old(x)
  if (!is.null(mapper_new)) {
    return(format(
      mapper_new,
      ...,
      prefix = prefix,
      initial = initial,
      depth = depth
    ))
  }
  text <- paste0(initial, ibm_shortname(x))
  if (ibm_input_available(x)) {
    inp <- ibm_input_get(x)
    text <- glue("{text}({format(inp, label.override = '')})")
  }
  as.character(text)
}

#' @export
#' @method format bm_list
#' @param collapse character or NULL, as in [base::paste()].
#' @param labels logical; if TRUE, include mapper names or numerical indices.
#'   Default `TRUE`
#'
#' @rdname bm_summary
format.bm_list <- function(x, ...,
                           prefix = "",
                           initial = prefix,
                           depth = 1,
                           collapse = ", ",
                           labels = TRUE) {
  if (labels) {
    paste0(
      vapply(
        seq_along(x),
        function(k) {
          nm <- names(x)[k]
          if (is.null(nm) || (identical(nm, ""))) {
            nm <- as.character(k)
          }
          paste0(
            nm,
            " = ",
            format(
              x[[k]],
              prefix = prefix,
              initial = initial,
              depth = depth
            )
          )
        },
        ""
      ),
      collapse = collapse
    )
  } else {
    paste0(
      vapply(
        seq_along(x),
        function(k) {
          paste0(
            format(
              x[[k]],
              prefix = prefix,
              initial = initial,
              depth = depth
            )
          )
        },
        ""
      ),
      collapse = collapse
    )
  }
}

#' @rdname bm_summary
#' @export
#' @method summary bru_mapper
summary.bru_mapper <- function(object, ...,
                               prefix = "",
                               initial = prefix,
                               depth = 1) {
  mapper_new <- make_bm_class_from_old(object)
  if (!is.null(mapper_new)) {
    return(summary(
      mapper_new,
      ...,
      prefix = prefix,
      initial = initial,
      depth = depth
    ))
  }
  summary_object <- format(
    object,
    prefix = prefix,
    initial = initial,
    depth = depth
  )
  class(summary_object) <- c(
    "summary_bru_mapper",
    class(summary_object)
  )
  summary_object
}

#' @export
#' @method format bm_multi
#' @rdname bm_summary
#' @examples
#' mapper <-
#'   bm_pipe(
#'     list(
#'       bm_multi(list(
#'         A = bm_index(2),
#'         B = bm_index(3)
#'       )),
#'       bm_index(2)
#'     )
#'   )
#' summary(mapper, depth = 2)
format.bm_multi <- function(x, ...,
                            prefix = "",
                            initial = prefix,
                            depth = 1) {
  txt <- NextMethod()
  if (depth <= 0) {
    return(txt)
  }
  sub_prefix <- paste0(prefix, "      ")
  txt <-
    paste0(
      txt,
      "(",
      format(x[["mappers"]],
        prefix = sub_prefix,
        initial = "",
        depth = depth - 1,
        collapse = ", "
      ),
      ")"
    )
  txt
}


#' @export
#' @method format bm_pipe
#' @rdname bm_summary
format.bm_pipe <- function(x, ...,
                           prefix = "",
                           initial = prefix,
                           depth = 1) {
  txt <- NextMethod()
  if (depth <= 0) {
    return(txt)
  }
  sub_prefix <- paste0(prefix, "      ")
  txt <-
    paste0(
      txt,
      " = ",
      format(x[["mappers"]],
        prefix = sub_prefix,
        initial = "",
        depth = depth - 1,
        collapse = " -> ",
        labels = FALSE
      )
    )
  txt
}

#' @export
#' @method format bm_collect
#' @rdname bm_summary
format.bm_collect <- function(x, ...,
                              prefix = "",
                              initial = prefix,
                              depth = 1) {
  txt <- NextMethod()
  if (depth <= 0) {
    return(txt)
  }
  sub_prefix <- paste0(prefix, "      ")
  txt <-
    paste0(
      txt,
      "(",
      paste0(
        vapply(
          seq_along(x[["mappers"]]),
          function(k) {
            nm <- names(x[["mappers"]])[k]
            if (is.null(nm)) {
              nm <- as.character(k)
            }
            paste0(
              nm,
              " = ",
              summary(
                x[["mappers"]][[k]],
                prefix = sub_prefix,
                initial = "",
                depth = depth - 1
              ),
              if (x[["hidden"]] && (k > 1)) {
                "(hidden)"
              } else {
                NULL
              }
            )
          },
          ""
        ),
        collapse = ", "
      ),
      ")"
    )
  txt
}

#' @export
#' @method format bm_sum
#' @rdname bm_summary
format.bm_sum <- function(x, ...,
                          prefix = "",
                          initial = prefix,
                          depth = 1) {
  txt <- NextMethod()
  if (depth <= 0) {
    return(txt)
  }
  sub_prefix <- paste0(prefix, "      ")
  txt <-
    paste0(
      txt,
      "(",
      format(
        x[["mappers"]],
        prefix = sub_prefix,
        initial = "",
        depth = depth - 1,
        collapse = ", "
      ),
      ")"
    )
  txt
}

#' @export
#' @method format bm_repeat
#' @rdname bm_summary
#' @examples
#' mapper <-
#'   bm_repeat(
#'     bm_multi(
#'       list(
#'         A = bm_index(2),
#'         B = bm_index(3)
#'       )
#'     ),
#'     3
#'   )
#' summary(mapper)
#' summary(mapper, depth = 0)
format.bm_repeat <- function(x, ...,
                             prefix = "",
                             initial = prefix,
                             depth = 1) {
  txt <- NextMethod()
  sub_prefix <- paste0(prefix, "      ")
  txt <-
    paste0(
      txt,
      "(",
      x[["n_rep"]],
      " x ",
      paste0(
        format(
          x[["mapper"]],
          prefix = sub_prefix,
          initial = "",
          depth = depth
        )
      ),
      if (isTRUE(x[["interleaved"]])) {
        ", interleaved"
      } else {
        NULL
      },
      ")"
    )
  txt
}

#' @export
#' @method format bm_reparam
#' @rdname bm_summary
#' @examples
#' mapper <-
#'   bm_reparam(
#'     bm_multi(
#'       list(
#'         A = bm_index(2),
#'         B = bm_index(3)
#'       )
#'     ),
#'     matrix(1:36, nrow = 6)
#'   )
#' summary(mapper)
#' summary(mapper, depth = 0)
format.bm_reparam <- function(x, ...,
                              prefix = "",
                              initial = prefix,
                              depth = 1) {
  txt <- NextMethod()
  sub_prefix <- paste0(prefix, "      ")
  txt <-
    paste0(
      txt,
      "(",
      paste0(
        format(
          x[["mapper"]],
          prefix = sub_prefix,
          initial = "",
          depth = depth
        )
      ),
      ")"
    )
  txt
}


#' @param sep character; separator for printing the summary.
#' @export
#' @method print summary_bru_mapper
#' @rdname bm_summary
print.summary_bru_mapper <- function(x, ..., sep = "\n") {
  cat(x, sep = sep)
  invisible(x)
}

#' @export
#' @method print bru_mapper
#' @rdname bm_summary
print.bru_mapper <- function(x, ...,
                             sep = "\n",
                             prefix = "",
                             initial = prefix,
                             depth = 1) {
  cat(
    format(x,
      prefix = "",
      initial = prefix,
      depth = depth
    ),
    sep = sep
  )
  invisible(x)
}

#' @export
#' @method print bm_list
#' @rdname bm_summary
print.bm_list <- function(x, ...,
                          sep = "\n",
                          prefix = "",
                          initial = prefix,
                          depth = 1,
                          labels = TRUE,
                          collapse = ", ") {
  cat(
    format(x,
      prefix = "",
      initial = prefix,
      depth = depth,
      collapse = collapse
    ),
    sep = sep
  )
  invisible(x)
}


# MAPPERS ----
## Constructor ----

#' @param mapper For `bru_mapper_define`, a prototype mapper object, see
#'   Details.
#' @param new_class If non-`NULL`, this is added at the front of the class
#'   definition
#' @param remove_class If non-`NULL`, this class or classes is removed from the
#'   class definition before adding the `new_class` names. Default is `"list"`.
#'
#' @describeIn bru_mapper Adds the `new_class` and `"bru_mapper"` class names to
#'   the inheritance list for the input `mapper` object, unless the object
#'   already inherits from these.
#'
#' To register mapper classes and methods in scripts, use `.S3method()`
#' to register the methods, e.g.
#' `.S3method("ibm_jacobian", "my_mapper_class", ibm_jacobian.my_mapper_class)`.
#'
#' In packages with `Suggests: inlabru`, add method information for delayed
#' registration, e.g.:
#' ```
#' #' @rawNamespace S3method(inlabru::bru_get_mapper, inla_rspde)
#' #' @rawNamespace S3method(inlabru::ibm_n, bru_mapper_inla_rspde)
#' #' @rawNamespace S3method(inlabru::ibm_values, bru_mapper_inla_rspde)
#' #' @rawNamespace S3method(inlabru::ibm_jacobian, bru_mapper_inla_rspde)
#' ```
#' or before each method, use `@exportS3Method`:
#' ```
#' #' @exportS3Method inlabru::bru_get_mapper
#' ```
#' etc., which semi-automates it.
#'
#' @export
bru_mapper_define <- function(mapper,
                              new_class = NULL,
                              remove_class = "list",
                              ...) {
  old_cls <- class(mapper)

  if ((length(remove_class) > 0) && (length(old_cls) > 0)) {
    old_cls <- old_cls[!(old_cls %in% remove_class)]
  }

  new_class <- c(new_class, "bru_mapper")
  for (cls in rev(new_class)) {
    if (!(cls %in% old_cls)) {
      old_cls <- c(cls, old_cls)
    }
  }
  class(mapper) <- old_cls

  mapper
}

## Default methods ----

# Conversion method from bru_mapper_<type> to bm_<type> class names.
# Only converts known inlabru mapper classes.
# Returns NULL if no conversion was needed.
make_bm_class_from_old <- function(mapper) {
  cls <- class(mapper)
  idx <- grepl(paste0("^bru_mapper_",
    c(
      "aggregate",
      "collect",
      "const",
      "factor",
      "fmesher",
      "harmonics",
      "index",
      "linear",
      "logsumexp",
      "marginal",
      "matrix",
      "mesh_B",
      "multi",
      "pipe",
      "repeat",
      "scale",
      "shift",
      "sum",
      "taylor",
      "fm_mesh_1d",
      "fm_mesh_2d",
      "inla_mesh_1d",
      "inla_mesh_2d"
    ),
    "$",
    collapse = "|"
  ), cls)
  if (any(idx)) {
    idx <- min(which(idx))
    # Remove the bru_mapper_ prefix, and add bm_ prefix
    cls[idx] <- sub("^bru_mapper_", "bm_", cls[idx])
    class(mapper) <- cls

    if (inherits(
      mapper,
      c("bm_aggregate", "bm_multi", "bm_collect", "bm_sum", "bm_pipe")
    )) {
      mapper[["is_rowwise"]] <- FALSE
    }

    return(mapper)
  }
  NULL
}

#' @describeIn ibm_n
#' Returns a non-null element 'n' from the
#' mapper object, and gives an error if it doesn't exist. If `inla_f=TRUE`,
#' first checks for a 'n_inla' element.
#' @export
ibm_n.default <- function(mapper, inla_f = FALSE, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_n(mapper_new, inla_f = inla_f, ...))
  }
  if (inla_f && !is.null(mapper[["n_inla"]])) {
    mapper[["n_inla"]]
  } else if (!is.null(mapper[["n"]])) {
    mapper[["n"]]
  } else {
    stop(paste0(
      "Default 'ibm_n()' method called but mapper doesn't have an 'n' element;",
      "\nClass: ",
      paste0(class(mapper), collapse = ", ")
    ))
  }
}

#' @export
#' @describeIn ibm_n_output Returns `NROW(input)`
ibm_n_output.default <- function(mapper,
                                 input,
                                 state = NULL,
                                 inla_f = FALSE,
                                 ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_n_output(
      mapper_new,
      input = input,
      state = state,
      inla_f = inla_f,
      ...
    ))
  }
  NROW(input)
}


#' @describeIn ibm_values
#' Returns a non-null element
#' 'values' from the mapper object, and `seq_len(ibm_n(mapper))` if
#' it doesn't exist.
#' @export
ibm_values.default <- function(mapper, inla_f = FALSE, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_values(mapper_new, inla_f = inla_f, ...))
  }
  if (inla_f && !is.null(mapper[["values_inla"]])) {
    mapper[["values_inla"]]
  } else if (!inla_f && !is.null(mapper[["values"]])) {
    mapper[["values"]]
  } else {
    seq_len(ibm_n(mapper, inla_f = inla_f))
  }
}

#' @describeIn ibm_is_linear
#' Returns logical
#' `is_linear` from the mapper object if it exists, and otherwise `TRUE`.
#' @export
ibm_is_linear.default <- function(mapper, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_is_linear(mapper_new, ...))
  }
  if (!is.null(mapper[["is_linear"]])) {
    mapper[["is_linear"]]
  } else {
    TRUE
  }
}

#' @describeIn ibm_is_rowwise
#' Returns logical
#' `is_rowwise` from the mapper object if it exists, and otherwise `TRUE`.
#' @export
ibm_is_rowwise.default <- function(mapper, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_is_rowwise(mapper_new, ...))
  }
  if (!is.null(mapper[["is_rowwise"]])) {
    mapper[["is_rowwise"]]
  } else {
    TRUE
  }
}

#' @describeIn ibm_jacobian
#' Mapper classes must implement their own `ibm_jacobian` method.
#' @export
ibm_jacobian.default <- function(mapper, input, state = NULL, ...) {
  if (is.null(mapper)) {
    stop(paste0(
      "NULL mapper detected."
    ))
  }
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_jacobian(mapper_new, input = input, state = state, ...))
  }
  stop(paste0(
    "Missing ibm_jacobian() method for class '",
    paste0(class(mapper), collapse = ", "),
    "'."
  ))
}

#' @describeIn ibm_linear
#' Calls [ibm_eval2()] and returns a [bm_taylor] object.
#' @returns A [bm_taylor] object.
#' The `state0` information in the affine mapper indicates for which state
#' the `offset` was evaluated; The affine mapper output is defined as
#' ```
#' effect(input, state) =
#'   offset(input, state0) + jacobian(input, state0) %*% (state - state0)
#' ```
#' @export
ibm_linear.default <- function(mapper, input, state, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_linear(mapper_new, input = input, state = state, ...))
  }
  eval2 <- ibm_eval2(mapper, input = input, state = state, ...)
  bm_taylor(
    offset = eval2$offset,
    jacobian = eval2$jacobian,
    state0 = state,
    values_mapper = mapper
  )
}

#' @describeIn ibm_simplify
#' Calls [ibm_linear()] for linear mappers, and returns the original mapper
#' for non-linear mappers.
#' @returns The original mapper is returned for non-linear mappers, and the
#' output of [ibm_linear()] is returned for linear mappers.
#' @export
ibm_simplify.default <- function(mapper, input = NULL, state = NULL, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_simplify(mapper_new, input = input, state = state, ...))
  }
  if (ibm_is_linear(mapper)) {
    if (is.null(state)) {
      state <- rep(0, ibm_n(mapper, input = input, state = NULL, ...))
    }
    return(ibm_linear(mapper, input = input, state = state, ...))
  }
  mapper
}


#' @describeIn ibm_eval
#' Verifies that the mapper is linear
#' with [ibm_is_linear()], and then computes a linear mapping
#' as `ibm_jacobian(...) %*% state`.  When `state` is `NULL`,
#' a zero vector of length [ibm_n_output()] is returned.
#' @param jacobian
#' For [ibm_eval()] methods, an optional pre-computed Jacobian, typically
#' supplied by internal methods that already have the Jacobian.
#' @export
ibm_eval.default <- function(mapper,
                             input,
                             state = NULL,
                             ...,
                             jacobian = NULL) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_eval(
      mapper_new,
      input = input,
      state = state,
      ...,
      jacobian = jacobian
    ))
  }
  if (!ibm_is_linear(mapper)) {
    stop("Non-linear mappers must implement their own ibm_eval() method.")
  }

  val <- numeric(ibm_n_output(mapper, input, state = state, ...))

  if ((ibm_n(mapper) > 0) && !is.null(state)) {
    if (is.null(jacobian)) {
      jacobian <- ibm_jacobian(mapper, input = input, state = state, ...)
    }
    val <- val + as.vector(jacobian %*% state)
  }

  val
}


#' @describeIn ibm_eval2
#' Calls `jacobian <- ibm_jacobian(...)` and
#' `offset <- ibm_eval(..., jacobian = jacobian)`
#' and returns a list with elements `offset` and `jacobian`, as needed by
#' [ibm_linear.default()] and similar methods. Mapper classes can implement
#' their own `ibm_eval2` method if joint construction of evaluation and Jacobian
#' is more efficient than separate or sequential construction.
#' @export
ibm_eval2.default <- function(mapper, input, state, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_eval2(mapper_new, input = input, state = state, ...))
  }
  jacobian <- ibm_jacobian(mapper, input, state, ...)
  offset <- ibm_eval(
    mapper,
    input,
    state,
    ...,
    jacobian = jacobian
  )
  list(offset = offset, jacobian = jacobian)
}


#' @export
#' @describeIn ibm_names Returns `NULL`
ibm_names.default <- function(mapper, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_names(mapper_new, ...))
  }
  NULL
}


#' @describeIn ibm_inla_subset
#' Uses the [ibm_values`()] output to construct the inla subset indexing as the
#' difference between `inla_f=FALSE` and `inla_f=TRUE`. Extra arguments such as
#' `multi` are passed on to [ibm_values()]. This means it supports both regular
#' vector values and `multi=1` data.frame values.
#' @export
ibm_inla_subset.default <- function(mapper, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_inla_subset(mapper_new, ...))
  }
  values_full <- ibm_values(mapper, inla_f = FALSE, ...)
  values_inla <- ibm_values(mapper, inla_f = TRUE, ...)
  if (is.data.frame(values_full)) {
    subset <- logical(NROW(values_full))
    if (length(subset) > 0) {
      subset[
        plyr::match_df(
          cbind(
            .inla_subset = seq_len(NROW(values_full)),
            values_full
          ),
          values_inla,
          on = names(values_full)
        )[[".inla_subset"]]
      ] <- TRUE
    }
  } else {
    subset <- base::match(values_full, values_inla, nomatch = 0) > 0
  }
  subset
}


#' @describeIn ibm_invalid_output
#' Returns an all-`FALSE` logical vector.
#' @export
ibm_invalid_output.default <- function(mapper, input, state, ...) {
  mapper_new <- make_bm_class_from_old(mapper)
  if (!is.null(mapper_new)) {
    return(ibm_invalid_output(mapper_new, input = input, state = state, ...))
  }
  rep(FALSE, ibm_n_output(mapper, input = input, state = state, ...))
}


## _fmesher ####

#' @title Mapper for general `fmesher` function space objects
#' @param mesh An `fmesher` object to map, supported by
#' [fmesher::fm_basis]`(mesh, input)` and
#' [fmesher::fm_dof]`(mesh)`.
#' @export
#' @description Creates a mapper for general `fmesher` function space objects.
#' @details Handles indexed mapping for all `fmesher` classes that support
#'   `fm_dof()` and `fm_basis()` methods. For non-indexed mapping of
#'   `fm_mesh_1d` objects, use `bru_mapper(mesh, indexed = FALSE)` which
#'   invokes the [bru_mapper.fm_mesh_1d()] method.
#' @returns A `bm_fmesher` object.
#' @rdname bm_fmesher
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#'
#' @examples
#' m <- bm_fmesher(fmesher::fmexample$mesh)
#' ibm_n(m)
#' ibm_eval(m, as.matrix(expand.grid(-2:2, -2:2)), seq_len(ibm_n(m)))
#'
bm_fmesher <- function(mesh) {
  if (is.null(mesh)) {
    return(NULL)
  }
  mapper <- list(mesh = mesh)
  bru_mapper_define(mapper, new_class = "bm_fmesher")
}

#' @export
#' @rdname bm_fmesher
#' @param \dots Arguments passed on to [bm_fmesher()]
bru_mapper_fmesher <- function(...) {
  bm_fmesher(...)
}

#' @export
#' @describeIn ibm_n Returns the [fmesher::fm_dof()] value of the mesh being
#' mapped.
#'
ibm_n.bm_fmesher <- function(mapper, ...) {
  fmesher::fm_dof(mapper[["mesh"]])
}
#' @export
#' @describeIn ibm_values Returns an index vector for the mesh basis functions.
#'
ibm_values.bm_fmesher <- function(mapper, ...) {
  seq_len(fmesher::fm_dof(mapper[["mesh"]]))
}
#' @export
#'
#' @describeIn ibm_jacobian Returns the [fmesher::fm_basis()] matrix of the
#' mesh being mapped.
ibm_jacobian.bm_fmesher <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  fmesher::fm_basis(mapper[["mesh"]], input)
}


## fm_mesh_2d ####

#' @describeIn bm_fmesher
#' Equivalent to calling [bm_fmesher()].
#' Note: Prior to version `2.12.0.9021`,
#'   this returned a `bru_mapper_fm_mesh_2d` object. Also see the note for
#'   [bru_mapper.fm_mesh_1d()].
#' @export
bru_mapper.fm_mesh_2d <- function(mesh, ...) {
  bm_fmesher(mesh, ...)
}

## The following methods are only used for old stored mapper objects

#' @title Deprecated methods
#' @description `r lifecycle::badge("deprecated")`
#'   Deprecated methods only used for old stored
#'   `bru_mapper_fm_mesh_1d` and `bru_mapper_fm_mesh_2d` objects.
#' @keywords internal
#' @name bm_fm_mesh_old
#' @rdname bm_fm_mesh_old
NULL

#' @export
#' @rdname bm_fm_mesh_old
ibm_n.bm_fm_mesh_2d <- function(mapper, ...) {
  fmesher::fm_dof(mapper[["mesh"]])
}
#' @export
#' @rdname bm_fm_mesh_old
ibm_values.bm_fm_mesh_2d <- function(mapper, ...) {
  seq_len(fmesher::fm_dof(mapper[["mesh"]]))
}
#' @export
#' @rdname bm_fm_mesh_old
ibm_jacobian.bm_fm_mesh_2d <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  fm_basis(mapper[["mesh"]], input)
}

## The following methods are only used for old stored mapper objects

#' @export
#' @rdname bm_fm_mesh_old
ibm_n.bm_inla_mesh_2d <- function(mapper, ...) {
  mapper[["mesh"]]$n
}
#' @export
#' @rdname bm_fm_mesh_old
ibm_values.bm_inla_mesh_2d <- function(mapper, ...) {
  seq_len(mapper[["mesh"]]$n)
}
#' @export
#' @rdname bm_fm_mesh_old
ibm_jacobian.bm_inla_mesh_2d <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  fm_basis(mapper[["mesh"]], input)
}


## fm_mesh_1d ####

#' @title Mapper for `fm_mesh_1d`
#' @param indexed logical; If `TRUE` (default), the `ibm_values()` output will
#'   be the integer indexing sequence for the latent variables (needed for
#'   `spde` models). If `FALSE`, points representative of the basis centres
#'   are returned (useful for an interpolator for `rw2` models and similar,
#'   for `fmesher` versions `>= 0.3.0.9002`).
#' @returns A `bm_fm_mesh_1d` or `bm_fmesher` object. The the
#'   general [bm_fmesher()] mapper handles all indexed `fmesher`
#'   objects.
#' @export
#' @description Create mapper for an `fm_mesh_1d` object
#' @param mesh An `fm_mesh_1d` object to use as a mapper
#' @rdname bm_fm_mesh_1d
#' @name bm_fm_mesh_1d
#' @inheritParams bru_mapper_generics
#' @inheritParams bm_fmesher
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bru_mapper(fmesher::fm_mesh_1d(c(1:3, 5, 7)))
#' ibm_values(m)
#' ibm_eval(m, 1:7, 1:5)
#'
#' m <- bru_mapper(fmesher::fm_mesh_1d(c(1:3, 5, 7)), indexed = FALSE)
#' ibm_values(m)
#' ibm_eval(m, 1:7, 1:5)
#'
#' m <- bru_mapper(
#'   fmesher::fm_mesh_1d(c(1:3, 5, 7), degree = 2, boundary = "free"),
#'   indexed = FALSE
#' )
#' ibm_values(m)
#' ibm_eval(m, 1:7, 1:6)
#'
bru_mapper.fm_mesh_1d <- function(mesh, indexed = TRUE, ...) {
  if (indexed) {
    mapper <- bm_fmesher(mesh)
    return(mapper)
  }
  mapper <- list(
    mesh = mesh,
    indexed = indexed
  )
  bru_mapper_define(mapper, new_class = "bm_fm_mesh_1d")
}

#' @export
#' @describeIn ibm_n Returns the [fmesher::fm_dof()] value of the mesh being
#' mapped.
#'
ibm_n.bm_fm_mesh_1d <- function(mapper, ...) {
  fmesher::fm_dof(mapper[["mesh"]])
}
#' @export
#' @describeIn ibm_values Returns an index vector into the basis functions for
#'   an indexed mapper. Otherwise, the `mid` values if present in the mesh being
#'   mapped, and otherwise returns the `loc` values of the mesh.
#'
ibm_values.bm_fm_mesh_1d <- function(mapper, ...) {
  if (mapper[["indexed"]]) {
    seq_len(fmesher::fm_dof(mapper[["mesh"]]))
  } else if (!is.null(mapper[["mesh"]][["mid"]])) {
    mapper[["mesh"]]$mid
  } else {
    mapper[["mesh"]]$loc
  }
}
#' @export
#' @describeIn ibm_jacobian Returns the [fmesher::fm_basis()] matrix of the
#' mesh being mapped.
#'
ibm_jacobian.bm_fm_mesh_1d <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  # Note: Handles NA input from fmesher 0.2.0.9002, with fixes in 0.4.0.9003
  fm_basis(mapper[["mesh"]], input)
}

## The following methods are only used for old stored mapper objects

#' @export
#' @rdname bm_fm_mesh_old
ibm_n.bm_inla_mesh_1d <- function(mapper, ...) {
  mapper[["mesh"]]$m
}
#' @export
#' @rdname bm_fm_mesh_old
ibm_values.bm_inla_mesh_1d <- function(mapper, ...) {
  if (mapper[["indexed"]]) {
    seq_len(mapper[["mesh"]]$m)
  } else {
    mapper[["mesh"]]$loc
  }
}
#' @export
#' @rdname bm_fm_mesh_old
ibm_jacobian.bm_inla_mesh_1d <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  fm_basis(mapper[["mesh"]], input)
}

## _index ####

#' @title Mapper for indexed variables
#' @param n Size of a model for `bm_index`
#' @export
#' @description Create a an indexing mapper
#' @rdname bm_index
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#'
#' @examples
#' m <- bm_index(4)
#' ibm_eval(m, -2:6, 1:4)
bm_index <- function(n = 1L) {
  bru_mapper_define(list(n = n), new_class = "bm_index")
}

#' @export
#' @rdname bm_index
#' @param \dots Arguments passed on to [bm_index()]
bru_mapper_index <- function(...) {
  bm_index(...)
}

#' @export
#' @describeIn ibm_invalid_output Returns `TRUE` out-of-range input values.
#'   Returns `FALSE` for in-range and `NA` input values (to support NA giving
#'   zero effect).
#'
ibm_invalid_output.bm_index <- function(mapper, input, state, ...) {
  # Allow NA to give effect zero
  nok <- !is.na(input)
  nok[nok] <-
    !((input[nok] >= 1) &
      (input[nok] <= ibm_n(mapper)))
  nok
}


#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_index <- function(mapper, input, state, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  nok <- ibm_invalid_output(mapper, input = input, state = state, ...)
  ok <- which(!is.na(input) & !nok)
  Matrix::sparseMatrix(
    i = ok,
    j = input[ok],
    x = rep(1, length(ok)),
    dims = c(NROW(input), ibm_n(mapper))
  )
}

## _taylor ####

#' @param offset For `bm_taylor`, an offset vector giving the value of
#' the linearisation at `state0`.
#' May be `NULL`, interpreted as an all-zero vector of length determined by
#' a non-null Jacobian.
#' @param jacobian For `bm_taylor()`, the Jacobian matrix,
#' evaluated at `state0`, or, a named list of such matrices.
#' May be `NULL` or an empty list, for a constant mapping.
#' @param state0 For `bm_taylor`, the reference `state` for the
#' linearisation, or a list of such states matching the `jacobian` list.
#' `NULL` is interpreted as 0.
#' @param values_mapper mapper object to be used for `ibm_n` and
#' `ibm_values` for `inla_f=TRUE` (experimental, currently unused)
#' @title Mapper for linear Taylor approximations
#' @description
#' Provides a pre-computed affine mapping,
#' internally used to represent and evaluate linearisation information.
#' The `state0` information indicates for which state the `offset` was
#' evaluated;
#' The affine mapper output is defined as
#' `effect(state) = offset + jacobian %*% (state - state0)`
#' @export
#' @rdname bm_taylor
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#'
#' @examples
#' m <- bm_taylor(
#'   offset = rep(2, 3),
#'   jacobian = matrix(1:6, 3, 2),
#'   state0 = c(1, 2)
#' )
#' ibm_eval2(m, state = 2:3)
bm_taylor <- function(offset = NULL, jacobian = NULL, state0 = NULL,
                      values_mapper = NULL) {
  stopifnot(!is.null(offset) || !is.null(jacobian))
  if (is.null(state0)) {
    n_state <- 0
  } else if (is.list(state0)) {
    if (is.null(names(state0))) {
      stop("`bm_taylor(..., state0)` lists must have named elements.")
    }
    n_state <- vapply(
      state0,
      function(x) {
        ifelse(is.null(x),
          0L,
          length(x)
        )
      },
      0L
    )
  } else {
    n_state <- length(state0)
  }

  if (is.null(jacobian)) {
    n_jacobian <- 0
  } else if (is.list(jacobian)) {
    if (is.null(names(jacobian))) {
      stop("`bm_taylor(..., jacobian)` lists must have named elements.")
    }
    n_jacobian <- vapply(
      jacobian,
      function(x) {
        ifelse(is.null(x),
          0L,
          ncol(x)
        )
      },
      0L
    )
  } else {
    n_jacobian <- ncol(jacobian)
  }

  if (!is.null(state0)) {
    if ((sum(n_jacobian) == 0) && (sum(n_state) > 0)) {
      stop(paste0(
        "Jacobian information indicates an empty state, but the state ",
        "information has length > 0"
      ))
    }
    stopifnot(sum(n_jacobian) == sum(n_state))
  }

  n_multi <- n_jacobian
  if (is.null(jacobian)) {
    state0 <- NULL
    n_state <- 0
  } else if (!is.list(jacobian) && is.list(state0)) {
    state0 <- unlist(state0)
    n_state <- length(state0)
  } else if (is.list(jacobian) && !is.null(state0) && !is.list(state0)) {
    # Split vector
    state0 <- lapply(
      seq_along(n_multi),
      function(k) {
        idx <- sum(n_multi[seq_len(k - 1)])
        idx <- idx + seq_len(n_multi[k]) - 1
        state0[idx]
      }
    )
    names(state0) <- names(jacobian)
    n_state <- n_multi
  }
  if (!is.null(state0)) {
    stopifnot(all(n_jacobian == n_state))
  }

  if (is.null(offset)) {
    if (is.list(jacobian)) {
      offset <- numeric(nrow(jacobian[[1]]))
    } else {
      offset <- numeric(nrow(jacobian))
    }
  }
  bru_mapper_define(
    list(
      offset = as.vector(offset),
      jacobian = jacobian,
      state0 = state0,
      n_multi = n_multi,
      n = sum(n_multi),
      n_output = length(offset),
      values_mapper = NULL
    ),
    # TODO: maybe allow values_mapper
    new_class = "bm_taylor"
  )
}

#' @export
#' @rdname bm_taylor
#' @param \dots Arguments passed on to [bm_taylor()]
bru_mapper_taylor <- function(...) {
  bm_taylor(...)
}

#' @export
#' @rdname ibm_n
#' @inheritParams bm_multi
#'
ibm_n.bm_taylor <- function(mapper,
                            inla_f = FALSE,
                            multi = FALSE,
                            ...) {
  if (inla_f && !is.null(mapper[["values_mapper"]])) {
    stop("ibm_n.bm_taylor should not be used with inla_f = TRUE")
  } else if (multi) {
    mapper[["n_multi"]]
  } else {
    mapper[["n"]]
  }
}

#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_taylor <- function(mapper, input, ...) {
  mapper[["n_output"]]
}


#' @export
#' @rdname ibm_values
#'
ibm_values.bm_taylor <- function(mapper,
                                 inla_f = FALSE,
                                 multi = FALSE,
                                 ...) {
  if (inla_f && !is.null(mapper[["values_mapper"]])) {
    stop("ibm_values.bm_taylor should not be used with inla_f = TRUE")
  } else if (multi) {
    lapply(mapper[["n_multi"]], function(k) seq_len(k))
  } else {
    seq_len(mapper[["n"]])
  }
}

#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_taylor <- function(mapper, ..., multi = FALSE) {
  if (is.null(mapper[["jacobian"]])) {
    return(Matrix::Matrix(0, mapper[["n_output"]], 0))
  }
  if (multi) {
    mapper[["jacobian"]]
  } else if (is.list(mapper[["jacobian"]])) {
    do.call(cbind, mapper[["jacobian"]])
  } else {
    mapper[["jacobian"]]
  }
}


#' @describeIn ibm_eval
#' Evaluates linearised
#' mapper information at the given `state`. The `input` argument is ignored,
#' so that the usual argument order
#' `ibm_eval(mapper, input, state)` syntax can be used, but also
#' `ibm_eval(mapper, state = state)`.  For a mapper with a named jacobian list,
#' the `state` argument must also be a named list.  If `state` is `NULL`,
#' all-zero is assumed.
#' @export
#'
ibm_eval.bm_taylor <- function(mapper,
                               input = NULL,
                               state = NULL,
                               ...) {
  if (is.null(mapper[["jacobian"]]) ||
    (mapper[["n"]] == 0) ||
    (is.null(state) && is.null(mapper[["state0"]]))) {
    val <- mapper[["offset"]]
  } else if (is.list(mapper[["jacobian"]])) {
    stopifnot(is.null(state) || is.list(state))
    val <- mapper[["offset"]]
    if (is.null(mapper[["state0"]])) {
      missing_names <- setdiff(names(mapper[["jacobian"]]), names(state))
      if (length(missing_names) > 0) {
        stop(
          "Names in jacobian list missing from state list: ",
          paste0(missing_names, collapse = ", ")
        )
      }
      for (nm in names(mapper[["jacobian"]])) {
        val <- val + mapper[["jacobian"]][[nm]] %*% state[[nm]]
      }
    } else if (is.null(state)) {
      for (nm in names(mapper[["jacobian"]])) {
        val <- val - mapper[["jacobian"]][[nm]] %*% mapper[["state0"]][[nm]]
      }
    } else {
      for (nm in names(mapper[["jacobian"]])) {
        val <- val + mapper[["jacobian"]][[nm]] %*%
          (state[[nm]] - mapper[["state0"]][[nm]])
      }
    }
  } else {
    stopifnot(is.null(state) || !is.list(state))
    if (is.null(mapper[["state0"]])) {
      val <- mapper[["offset"]] + mapper[["jacobian"]] %*% state
    } else if (is.null(state)) {
      val <- mapper[["offset"]] - mapper[["jacobian"]] %*% mapper[["state0"]]
    } else {
      val <- mapper[["offset"]] +
        mapper[["jacobian"]] %*% (state - mapper[["state0"]])
    }
  }
  as.vector(val)
}


## _linear ####

#' @title Mapper for a linear effect
#' @export
#' @description Create a mapper for linear effects
#' @rdname bm_linear
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_linear()
#' ibm_eval(m, input = 1:4, state = 2)
#'
bm_linear <- function() {
  bru_mapper_define(list(), new_class = "bm_linear")
}

#' @export
#' @rdname bm_linear
bru_mapper_linear <- function() {
  bm_linear()
}

#' @export
#' @describeIn ibm_n Returns `1L`
#'
ibm_n.bm_linear <- function(mapper, ...) {
  1L
}
#' @export
#' @describeIn ibm_values Returns `1.0`
#'
ibm_values.bm_linear <- function(mapper, ...) {
  1.0
}

#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_linear <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  if (!is.null(input) &&
    !(is.numeric(input) ||
      is.logical(input) ||
      is(input, "Matrix"))) {
    stop(paste0(
      "The input to a bm_linear evaluation must be numeric or logical."
    ))
  }
  ok <- !is.na(input)
  A <- Matrix::sparseMatrix(
    i = which(ok),
    j = rep(1, sum(ok)),
    x = input[ok],
    dims = c(NROW(input), ibm_n(mapper))
  )
  A
}

## _matrix ####

#' @title Mapper for matrix multiplication
#' @param labels Column labels for matrix mappings; Can be factor, character, or
#'   a single integer specifying the number of columns for integer column
#'   indexing.
#' @export
#' @description Create a matrix mapper, for a given number of columns
#' @rdname bm_matrix
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#'
#' @examples
#' m <- bm_matrix(labels = c("a", "b"))
#' ibm_values(m)
#' ibm_eval2(m, input = matrix(1:6, 3, 2), state = 2:3)
#'
#' m <- bm_matrix(labels = 2L)
#' ibm_values(m)
#' ibm_eval2(m, input = matrix(1:6, 3, 2), state = 2:3)
#'
bm_matrix <- function(labels) {
  if (is.factor(labels)) {
    mapper <- list(
      labels = levels(labels)
    )
  } else if (is.integer(labels) && (length(labels) == 1L)) {
    mapper <- list(
      labels = seq_len(labels)
    )
  } else {
    mapper <- list(
      labels = as.character(labels)
    )
  }
  bru_mapper_define(mapper, new_class = "bm_matrix")
}

#' @export
#' @rdname bm_matrix
#' @param \dots Arguments passed on to [bm_matrix()]
bru_mapper_matrix <- function(...) {
  bm_matrix(...)
}

#' @export
#' @describeIn ibm_n Returns the number of columns in the matrix mapper.
#'
ibm_n.bm_matrix <- function(mapper, ...) {
  length(mapper$labels)
}
#' @export
#' @describeIn ibm_values For integer labels, the vector of labels.
#' For character labels, the labels as a factor variable.
#'
ibm_values.bm_matrix <- function(mapper, ...) {
  if (is.integer(mapper$labels)) {
    mapper$labels
  } else {
    factor(x = mapper$labels, levels = mapper$labels)
  }
}

#' @export
#' @describeIn ibm_jacobian Accepts `input` as a matrix, `Matrix`, `Spatial`,
#' or `sfc_POINT` object.
#'
ibm_jacobian.bm_matrix <- function(mapper, input, state = NULL,
                                   inla_f = FALSE, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  } else if (is.matrix(input)) {
    A <- as(input, "Matrix")
  } else if (inherits(input, "Matrix")) {
    A <- input
  } else if (inherits(input, "Spatial")) {
    A <- sp::coordinates(input)
    A <- as(A, "Matrix")
  } else if (inherits(input, "sfc_POINT")) {
    A <- sf::st_coordinates(input)
    nm <- intersect(colnames(A), c("X", "Y", "Z", "M"))
    A <- as(A[, nm, drop = FALSE], "Matrix")
  } else {
    A <- as(input, "Matrix")
  }
  if (ncol(A) != ibm_n(mapper)) {
    stop(paste0(
      "Input to matrix mapper has ", ncol(A),
      " columns but should have ", ibm_n(mapper),
      " columns."
    ))
  }
  colnames(A) <- mapper$labels
  A
}


## _factor ####

#' @title Mapper for factor variables
#' @param values Input values calculated by [bru_input.bru_input()]
#' @param factor_mapping character; selects the type of factor mapping.
#' * `'contrast'` for leaving out the first factor level.
#' * `'full'` for keeping all levels.
#' @param indexed logical; if `TRUE`, the `ibm_values()` method
#' will return an integer vector instead of the factor levels.
#' This is needed e.g. for `group` and `replicate` mappers, since
#' `INLA::f()` doesn't accept factor values. Default: `FALSE`, which
#' works for the main input mappers. The default mapper constructions
#' will set it the required setting.
#' @export
#' @description Create a factor mapper
#' @rdname bm_factor
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_factor(factor(c("a", "b")), "full")
#' ibm_eval2(m, input = c("b", "a", "a", "b"), state = c(1, 3))
#'
#' m <- bm_factor(factor(c("a", "b")), "contrast")
#' ibm_eval2(m, input = factor(c("b", "a", "a", "b")), state = 2)
#'
bm_factor <- function(values, factor_mapping, indexed = FALSE) {
  factor_mapping <- match.arg(factor_mapping, c("full", "contrast"))
  if (is.factor(values)) {
    mapper <- list(
      levels = levels(values),
      factor_mapping = factor_mapping,
      indexed = indexed
    )
  } else if (is.character(values)) {
    mapper <- list(
      levels = unique(values),
      factor_mapping = factor_mapping,
      indexed = indexed
    )
  } else {
    mapper <- list(
      levels = as.character(sort(unique(values))),
      factor_mapping = factor_mapping,
      indexed = indexed
    )
  }
  mapper$n <- length(mapper$levels) - identical(factor_mapping, "contrast")
  if (indexed) {
    bru_mapper_define(mapper,
      new_class = c(
        "bm_factor_index",
        "bm_factor"
      )
    )
  } else {
    bru_mapper_define(mapper, new_class = "bm_factor")
  }
}

#' @export
#' @rdname bm_factor
#' @param \dots Arguments passed on to [bm_factor()]
bru_mapper_factor <- function(...) {
  bm_factor(...)
}

#' @export
#' @describeIn ibm_n Returns the number of levels when `factor_mapping` is
#'   "full", and the number of levels minus one if `factor_mapping` is
#'   "contrast".
#'
ibm_n.bm_factor <- function(mapper, ...) {
  length(mapper[["levels"]]) - identical(mapper[["factor_mapping"]], "contrast")
}
#' @export
#' @describeIn ibm_values Returns the factor levels (minus the first level for
#' `factor_mapping` "contrast"), or an integer vector (if `indexed = TRUE` in
#' [bm_factor()]).
#'
ibm_values.bm_factor <- function(mapper, ...) {
  if (is.null(mapper[["indexed"]]) || !mapper[["indexed"]]) {
    if (identical(mapper[["factor_mapping"]], "contrast")) {
      mapper$levels[-1L]
    } else {
      mapper$levels
    }
  } else {
    seq_len(ibm_n(mapper))
  }
}

#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_factor <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  if (is.factor(input)) {
    if (!identical(levels(input), mapper$levels)) {
      input <- factor(as.character(input), levels = mapper$levels)
    }
  } else {
    input <- factor(input, levels = mapper$levels)
  }
  if (mapper$factor_mapping == "full") {
    input <- as.numeric(input)
  } else if (mapper$factor_mapping == "contrast") {
    input <- as.numeric(input) - 1L
  } else {
    stop("Unknown factor mapping '", mapper$factor_mapping, "'.")
  }
  ok <- !is.na(input)
  ok[which(ok)] <- (input[ok] > 0L)
  A <- Matrix::sparseMatrix(
    i = which(ok),
    j = input[ok],
    x = rep(1.0, sum(ok)),
    dims = c(NROW(input), ibm_n(mapper))
  )
  A
}


## _const ####

#' @title Constant mapper
#' @export
#' @description Create a constant mapper
#' @rdname bm_const
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_const()
#' ibm_eval2(m, input = 1:4)
#'
bm_const <- function() {
  bru_mapper_define(list(), new_class = "bm_const")
}

#' @export
#' @rdname bm_const
bru_mapper_const <- function() {
  bm_const()
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_const <- function(mapper, ...) {
  0L
}
#' @export
#' @rdname ibm_values
#'
ibm_values.bm_const <- function(mapper, ...) {
  NULL
}

#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_const <- function(mapper, input, ...) {
  Matrix::Matrix(0, ibm_n_output(mapper, input, ...), 0L)
}

#' @export
#' @describeIn ibm_eval Returns the input values, with NA replaced by 0.
#'
ibm_eval.bm_const <- function(mapper, input, state = NULL, ...) {
  if (is.null(input)) {
    return(numeric(0))
  }
  input <- as.vector(input)
  ok <- !is.na(input)
  input[!ok] <- 0
  input
}


## _shift ####

#' @title Mapper for element-wise shifting
#' @export
#' @description Create a standalone
#' shift mapper that can be used as part of a `bm_pipe`.
#' If `mapper` is non-null, the `bm_shift()` constructor
#' returns
#' `bm_pipe(list(mapper = mapper, shift = bm_shift()))`
#' @param mapper If non-NULL, a `bru_mapper` to be shifted.
#' @rdname bm_shift
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_shift()
#' ibm_eval2(m, c(1, 2, 1, 2), 1:4)
#'
bm_shift <- function(mapper = NULL) {
  m <- bru_mapper_define(
    list(),
    new_class = "bm_shift"
  )
  if (!is.null(mapper)) {
    m <- bm_pipe(list(mapper = mapper, shift = m))
  }
  m
}

#' @export
#' @rdname bm_shift
#' @param \dots Arguments passed on to [bm_shift()]
bru_mapper_shift <- function(...) {
  bm_shift(...)
}

#' @param n_state integer giving the length of the state vector for mappers
#' that have state dependent output size.
#' @export
#' @rdname ibm_n
#'
ibm_n.bm_shift <- function(mapper, ..., state = NULL, n_state = NULL) {
  # Output size depends on the state size
  if (!is.null(state)) {
    length(state)
  } else if (is.null(n_state)) {
    NA_integer_
  } else {
    n_state
  }
}
#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_shift <- function(mapper, input, state = NULL, ...,
                                  n_state = NULL) {
  if (!is.null(state)) {
    length(state)
  } else if (is.null(n_state)) {
    # To allow scalar input shifts, do not assume NROW(input) size
    NA_integer_
  } else {
    n_state
  }
}
#' @export
#' @rdname ibm_values
#'
ibm_values.bm_shift <- function(mapper, ...,
                                state = NULL, n_state = NULL) {
  n_state <- ibm_n(mapper, state = state, n_state = n_state)
  if (is.na(n_state)) {
    # Don't know how big the mapper will be
    NULL
  } else {
    seq_len(n_state)
  }
}

#' @export
#' @describeIn ibm_jacobian `input` NULL values are interpreted as no shift.
#'
ibm_jacobian.bm_shift <- function(mapper, input, state = NULL, ...) {
  stopifnot(!is.null(state))
  Matrix::Diagonal(n = length(state), 1.0)
}


#' @export
#' @rdname ibm_eval
#' @inheritParams ibm_jacobian
#'
ibm_eval.bm_shift <- function(mapper, input, state = NULL, ...) {
  stopifnot(!is.null(state))
  if (is.null(input)) {
    return(state)
  }
  shift <- as.vector(input)
  ok <- !is.na(shift)
  shift[!ok] <- 0
  state + shift
}


## _scale ####

#' @title Mapper for element-wise scaling
#' @export
#' @description Create a standalone
#' scaling mapper that can be used as part of a `bm_pipe`.
#' If `mapper` is non-null, the `bm_scale()` constructor
#' returns
#' `bm_pipe(list(mapper = mapper, scale = bm_scale()))`
#' @rdname bm_scale
#' @param mapper An optional `bru_mapper` to be scaled.
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_scale()
#' ibm_eval2(m, c(1, 2, 1, 2), 1:4)
#'
bm_scale <- function(mapper = NULL) {
  m <- bru_mapper_define(
    list(),
    new_class = "bm_scale"
  )
  if (!is.null(mapper)) {
    m <- bm_pipe(list(mapper = mapper, scale = m))
  }
  m
}

#' @export
#' @rdname bm_scale
#' @param \dots Arguments passed on to [bm_scale()]
bru_mapper_scale <- function(...) {
  bm_scale(...)
}

#' @param n_state integer giving the length of the state vector for mappers
#' that have state dependent output size.
#' @export
#' @rdname ibm_n
#'
ibm_n.bm_scale <- function(mapper, ..., state = NULL, n_state = NULL) {
  # Output size depends on the state size
  if (!is.null(state)) {
    length(state)
  } else if (is.null(n_state)) {
    NA_integer_
  } else {
    n_state
  }
}
#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_scale <- function(mapper, input, state = NULL, ...,
                                  n_state = NULL) {
  if (!is.null(state)) {
    length(state)
  } else if (is.null(n_state)) {
    # To allow scalar input weights, do not assume NROW(input) size
    NA_integer_
  } else {
    n_state
  }
}
#' @export
#' @rdname ibm_values
#'
ibm_values.bm_scale <- function(mapper, ...,
                                state = NULL, n_state = NULL) {
  n_state <- ibm_n(mapper, state = state, n_state = n_state)
  if (is.na(n_state)) {
    # Don't know how big the mapper will be
    NULL
  } else {
    seq_len(n_state)
  }
}

#' @export
#' @describeIn ibm_jacobian `input` NULL values
#' are interpreted as no scaling.
#'
ibm_jacobian.bm_scale <- function(mapper, input, state = NULL, ...) {
  stopifnot(!is.null(state))
  if (is.null(input)) {
    # No scaling
    return(Matrix::Diagonal(n = length(state), 1.0))
  }
  if (!(is.numeric(input) ||
    is.logical(input) ||
    is(input, "Matrix"))) {
    stop(
      "The input to a bm_scale evaluation must be numeric or logical."
    )
  }
  scale <- as.vector(input)
  ok <- !is.na(scale)
  scale[!ok] <- 0
  Matrix::Diagonal(n = length(state), scale)
}


#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_scale <- function(mapper, input, state = NULL, ...) {
  stopifnot(!is.null(state))
  if (is.null(input)) {
    return(state)
  }
  if (!is.null(input) &&
    !(is.numeric(input) ||
      is.logical(input) ||
      is(input, "Matrix"))) {
    stop(
      "The input to a bm_scale evaluation must be numeric or logical."
    )
  }
  scale <- as.vector(input)
  ok <- !is.na(scale)
  scale[!ok] <- 0
  scale * state
}


## _aggregate ####

#' @title Mapper for aggregation
#' @export
#' @param rescale
#' logical; For `bm_aggregate` and `bm_logsumexp`,
#' specifies if the blockwise sums should be normalised by the blockwise weight
#' sums or not:
#' * `FALSE`: (default) Straight weighted sum, no rescaling.
#' * `TRUE`: Divide by the sum of the weight values within each block.
#'   This is useful for integration averages, when the given weights are plain
#'   integration weights. If the weights are `NULL` or all ones, this is
#'   the same as dividing by the number of entries in each block.
#' @param n_block Predetermined number of output blocks. If `NULL`, overrides
#' the maximum block index in the inputs. The priority order is `input$n_block`,
#' the mapper definition `n_block`, then `max(input$block)`.
#' @param type character; if non-NULL, overrides the `rescale` argument, and
#' constructs an aggregation mapper of the given type instead. Supported
#' values are "sum", "average" (for regular [bm_aggregate()]),
#' "logsumexp", "logaverageexp" (for [bm_logsumexp()]), and
#' "logitaverage" (for [bm_logitaverage()]).
#' @description
#' Constructs a mapper
#' that aggregates elements of the input state, so it can be used e.g.
#' for weighted summation or integration over blocks of values.
#' @rdname bm_aggregate
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_aggregate()
#' ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4), 11:14)
#' ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4, n_block = 3), 11:14)
#'
bm_aggregate <- function(rescale = FALSE,
                         n_block = NULL,
                         type = NULL) {
  if (!is.null(type)) {
    type <- match.arg(
      type,
      c("sum", "average", "logsumexp", "logaverageexp", "logitaverage")
    )
    if (type == "sum") {
      rescale <- FALSE
    } else if (type == "average") {
      rescale <- TRUE
    } else if (type == "logsumexp") {
      return(bm_logsumexp(
        rescale = FALSE,
        n_block = n_block
      ))
    } else if (type == "logaverageexp") {
      return(bm_logsumexp(
        rescale = TRUE,
        n_block = n_block
      ))
    } else if (type == "logitaverage") {
      return(bm_logitaverage(
        n_block = n_block
      ))
    }
  }
  bru_mapper_define(
    list(
      rescale = rescale,
      n_block = n_block,
      is_rowwise = FALSE
    ),
    new_class = "bm_aggregate"
  )
}

#' @export
#' @param \dots Arguments passed on to [bm_aggregate()]
#' @rdname bm_aggregate
bru_mapper_aggregate <- function(...) {
  bm_aggregate(...)
}

#' @param n_state integer giving the length of the state vector for mappers
#' that have state dependent output size.
#' @export
#' @rdname ibm_n
#'
ibm_n.bm_aggregate <- function(mapper, ...,
                               input = NULL,
                               state = NULL,
                               n_state = NULL) {
  # Output size depends on the state size
  if (!is.null(state)) {
    length(state)
  } else if (!is.null(n_state)) {
    n_state
  } else if (!is.null(input) &&
    (!is.null(input[["block"]]) ||
      !is.null(input[["weights"]]) ||
      !is.null(input[["log_weights"]]))) {
    max(
      length(input[["block"]]),
      length(input[["weights"]]),
      length(input[["log_weights"]])
    )
    length(input[["block"]])
  } else {
    NA_integer_
  }
}

bm_aggregate_n_block <- function(mapper, input = NULL) {
  if (is.null(input[["n_block"]])) {
    # Note: Allowed to be NULL
    n_block <- mapper[["n_block"]]
  } else {
    n_block <- input[["n_block"]]
  }
  if (is.null(n_block) && !is.null(input[["block"]])) {
    if (is.character(input[["block"]])) {
      lifecycle::deprecate_stop(
        "2.12.0.9001",
        "bm_aggregate(input = 'can not be a `character`')",
        I("`integer` values for the `input` argument.")
      )
    } else {
      n_block <- max(input[["block"]])
    }
  }
  n_block
}

#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_aggregate <- function(mapper,
                                      input = NULL, ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  if (is.null(n_block)) {
    return(NA_integer_)
  }
  n_block
}
#' @export
#' @rdname ibm_values
#'
ibm_values.bm_aggregate <- function(mapper, ...,
                                    state = NULL,
                                    n_state = NULL) {
  n_state <- ibm_n(mapper, state = state, n_state = n_state)
  if (is.na(n_state)) {
    # Don't know how big the mapper will be
    NULL
  } else {
    seq_len(n_state)
  }
}


#' @export
#' @describeIn ibm_jacobian
#' `input` should be a list with elements `block`
#' and `weights`. `block`
#' should be a vector of the same length as the `state`, or `NULL`, with `NULL`
#' equivalent to all-1.
#' If `weights` is `NULL`, it's interpreted as all-1.
#'
ibm_jacobian.bm_aggregate <- function(mapper,
                                      input,
                                      state = NULL,
                                      ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  fm_block(
    block = input[["block"]],
    weights = input[["weights"]],
    log_weights = input[["log_weights"]],
    n_block = n_block,
    rescale = mapper[["rescale"]]
  )
}


#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_aggregate <- function(mapper, input, state = NULL, ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  val <-
    fm_block_eval(
      block = input[["block"]],
      log_weights = input[["log_weights"]],
      weights = input[["weights"]],
      rescale = mapper[["rescale"]],
      n_block = n_block,
      values = state
    )
  val
}


## _logsumexp ####

#' @title Mapper for log-sum-exp aggregation
#' @export
#' @description
#' Constructs a mapper
#' that aggregates elements of `exp(state)`, with optional non-negative
#' weighting, and then takes the `log()`, so it can be used e.g.
#' for  \eqn{v_k=\log[\sum_{i\in I_k} w_i \exp(u_i)]
#' }{log(blocksum(w*exp(state)))}
#' and \eqn{v_k=\log[\sum_{i\in I_k} w_i \exp(u_i) / \sum_{i\in I_k} w_i]
#' }{log(blocksum(w*exp(state)/blocksum(w)))}
#' calculations.  Relies on the input handling methods for
#' `bm_aggregate`, but also allows the weights to be supplied on a
#' logarithmic scale as `log_weights`. To avoid numerical overflow, it uses the
#' common method of internally shifting the state blockwise;
#' \eqn{v_k=s_k+\log[\sum_{i\in I_k} \exp(u_i + \log(w_i)- s_k)]
#' }{log(blocksum(w*exp(state)))},
#' where \eqn{s_k=\max_{i\in I_k} u_i + \log(w_i)}{s=blockmax(u+log(w))} is the
#' shift for block \eqn{k}{k}.
#' @rdname bm_logsumexp
#' @inheritParams bm_aggregate
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_logsumexp()
#' ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4), 11:14)
#' ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4, n_block = 3), 11:14)
#'
bm_logsumexp <- function(rescale = FALSE,
                         n_block = NULL) {
  # Arguments documented for bm_aggregate
  # Inherit class bm_aggregate to reuse common methods
  bru_mapper_define(
    list(
      rescale = rescale,
      n_block = n_block,
      is_rowwise = FALSE,
      is_linear = FALSE
    ),
    new_class = c("bm_logsumexp", "bm_aggregate")
  )
}

#' @export
#' @rdname bm_logsumexp
#' @param \dots Arguments passed on to [bm_logsumexp()]
bru_mapper_logsumexp <- function(...) {
  bm_logsumexp(...)
}


#' @export
#' @describeIn ibm_jacobian `input` should be a list with elements
#'   `block` and `weights`. `block` should be a vector of the same length as the
#'   `state`, or `NULL`, with `NULL` equivalent to all-1.
#' If `weights` is `NULL`, it's interpreted as all-1.
#'
ibm_jacobian.bm_logsumexp <- function(mapper,
                                      input,
                                      state = NULL,
                                      ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  input <- fm_block_prep(
    block = input[["block"]],
    log_weights = input[["log_weights"]],
    weights = input[["weights"]],
    force_log = TRUE,
    values = state,
    n_block = n_block
  )

  n_state <- length(input$block)
  n_out <- input$n_block

  log_weights <- fm_block_log_weights(
    block = input[["block"]],
    log_weights = input[["log_weights"]],
    weights = input[["weights"]],
    n_block = n_out,
    rescale = mapper[["rescale"]]
  )

  # Compute shift for stable log-sum-exp
  w_state <- state + log_weights
  shift <- fm_block_log_shift(
    log_weights = w_state,
    block = input[["block"]],
    n_block = n_out
  )

  sum_values <-
    as.vector(
      Matrix::sparseMatrix(
        i = input[["block"]],
        j = rep(1L, n_state),
        x = exp(w_state - shift[input[["block"]]]),
        dims = c(n_out, 1)
      )
    )
  scale <- sum_values[input[["block"]]]

  Matrix::sparseMatrix(
    i = input[["block"]],
    j = seq_len(n_state),
    x = exp(w_state - shift[input[["block"]]]) / scale,
    dims = c(n_out, n_state)
  )
}

#' @export
#' @param log logical; control `log` output. Default `TRUE`, see the
#'   `ibm_eval()` details for `logsumexp` mappers.
#' @describeIn ibm_eval When `log` is `TRUE` (default), `ibm_eval()`
#'   for `logsumexp` returns the log-sum-weight-exp value. If `FALSE`, the
#'   `sum-weight-exp` value is returned.
#'
ibm_eval.bm_logsumexp <- function(mapper, input, state = NULL,
                                  log = TRUE, ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  val <-
    fm_block_logsumexp_eval(
      block = input[["block"]],
      log_weights = input[["log_weights"]],
      weights = input[["weights"]],
      rescale = mapper[["rescale"]],
      n_block = n_block,
      values = state,
      log = log
    )
  val
}


## _logitaverage ####

#' @title Mapper for logit-sum-inverse-logit aggregation
#' @export
#' @description `r lifecycle::badge("experimental")`
#' Constructs a mapper that averages elements of `plogis(state)`, with optional
#' non-negative weighting, and then takes the `qlogis()`.  Relies on the input
#' handling methods for `bm_aggregate`. To avoid numerical issues, it uses
#' `plogis(x, log.p = TRUE)`, `plogis(-x, log.p = TRUE)`, and the equivalent of
#' two applications of [bm_logsumexp()] to evaluate
#' \eqn{\log(p_k)-\log(1-p_k)}{log(p[k])-log(1-p[k])}, where
#'
#' \eqn{p_k=\sum_{i\in I_k} w_i / (1+e^{-\eta_i}) / \sum_{i\in I_k} w_i
#' }{p[k]=sum(w[i] * plogis(eta[i])) / sum(w[i])}
#' @rdname bm_logitaverage
#' @inheritParams bm_aggregate
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_logitaverage()
#' ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4), 11:14)
#' ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4, n_block = 3), 11:14)
#'
bm_logitaverage <- function(n_block = NULL) {
  # Arguments documented for bm_aggregate
  # Inherit class bm_aggregate to reuse common methods
  bru_mapper_define(
    list(
      rescale = TRUE,
      n_block = n_block,
      is_rowwise = FALSE,
      is_linear = FALSE
    ),
    new_class = c("bm_logitaverage", "bm_aggregate")
  )
}

#' @export
#' @describeIn ibm_jacobian `input` should be a list with elements
#'   `block` and `weights`. `block` should be a vector of the same length as the
#'   `state`, or `NULL`, with `NULL` equivalent to all-1.
#' If `weights` is `NULL`, it's interpreted as all-1.
#'
ibm_jacobian.bm_logitaverage <- function(mapper,
                                         input,
                                         state = NULL,
                                         ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  input <- fm_block_prep(
    block = input[["block"]],
    log_weights = input[["log_weights"]],
    weights = input[["weights"]],
    force_log = TRUE,
    values = state,
    n_block = n_block
  )

  n_state <- length(input$block)
  n_out <- input$n_block

  log_weights <- fm_block_log_weights(
    block = input[["block"]],
    log_weights = input[["log_weights"]],
    weights = input[["weights"]],
    n_block = n_out,
    rescale = TRUE
  )

  state1 <- plogis(state, log.p = TRUE)
  state2 <- plogis(-state, log.p = TRUE)

  # Compute shift for stable log-sum-exp
  w_state1 <- state1 + log_weights
  w_state2 <- state2 + log_weights
  shift1 <- fm_block_log_shift(
    log_weights = w_state1,
    block = input[["block"]],
    n_block = n_out
  )
  shift2 <- fm_block_log_shift(
    log_weights = w_state2,
    block = input[["block"]],
    n_block = n_out
  )

  sum_values1 <-
    as.vector(
      Matrix::sparseMatrix(
        i = input[["block"]],
        j = rep(1L, n_state),
        x = exp(w_state1 - shift1[input[["block"]]]),
        dims = c(n_out, 1)
      )
    )
  scale1 <- exp(state2[input[["block"]]]) / sum_values1[input[["block"]]]
  sum_values2 <-
    as.vector(
      Matrix::sparseMatrix(
        i = input[["block"]],
        j = rep(1L, n_state),
        x = exp(w_state2 - shift2[input[["block"]]]),
        dims = c(n_out, 1)
      )
    )
  scale2 <- -exp(state1[input[["block"]]]) / sum_values2[input[["block"]]]

  A <- Matrix::sparseMatrix(
    i = input[["block"]],
    j = seq_len(n_state),
    x = exp(w_state1 - shift1[input[["block"]]]) * scale1 -
      exp(w_state2 - shift2[input[["block"]]]) * scale2,
    dims = c(n_out, n_state)
  )
  A
}

#' @export
#' @param logit logical; control `logit` output. Default `TRUE`, see the
#'   `ibm_eval()` details for `logitaverage` mappers.
#' @describeIn ibm_eval When `logit` is `TRUE` (default), `ibm_eval()`
#'   for `logitaverage` returns the logit-sum-weight-inverse-logit value.
#'   If `FALSE`, the `sum-weights=invere-logit` value is returned.
#'
ibm_eval.bm_logitaverage <- function(mapper, input, state = NULL,
                                     logit = TRUE, ...) {
  n_block <- bm_aggregate_n_block(mapper = mapper, input = input)
  state1 <- plogis(state, log.p = TRUE)
  state2 <- plogis(-state, log.p = TRUE)
  val_1 <-
    fm_block_logsumexp_eval(
      block = input[["block"]],
      log_weights = input[["log_weights"]],
      weights = input[["weights"]],
      rescale = TRUE,
      n_block = n_block,
      values = state1,
      log = TRUE
    )
  val_2 <-
    fm_block_logsumexp_eval(
      block = input[["block"]],
      log_weights = input[["log_weights"]],
      weights = input[["weights"]],
      rescale = TRUE,
      n_block = n_block,
      values = state2,
      log = TRUE
    )
  val <- val_1 - val_2
  if (!logit) {
    val <- plogis(val)
  }
  val
}


## _marginal ####

require_args <- function(fun, req) {
  label <- deparse1(substitute(fun))
  if (is.null(fun)) {
    return(invisible())
  }
  fun_args <- formals(args(fun))
  subset <- req %in% names(fun_args)
  if (!all(subset)) {
    subset <- req[!subset]
    stop(paste0(
      "The '", label, "' function is missing the argument",
      if (length(subset) > 1) "s" else "",
      " ",
      paste0("'", subset, "'", collapse = ", "),
      "."
    ))
  }
  invisible()
}

#' @title Mapper for marginal distribution transformation
#' @export
#' @description
#' Constructs a mapper that transforms the marginal distribution `state` from
#' \eqn{\textrm{N}(0,1)}{N(0, 1)} to the distribution of a given (continuous)
#' quantile function. The `...` arguments are used as parameter arguments to
#' `qfun`, `pfun`, `dfun`, and `dqfun`.
#' @param qfun A quantile function, supporting arguments `p`, `lower.tail`, and
#'   `log.p`, like [stats::qnorm()].
#' @param pfun A CDF, supporting arguments `q`, `lower.tail`, and `log.p`,
#' like [stats::pnorm()].  Only needed and used when
#' `xor(mapper[["inverse"]], reverse)` is `TRUE` in a method call.
#' Default `NULL`
#' @param dfun A pdf, supporting arguments `x` and `log`,
#' like [stats::dnorm()]. If `NULL` (default), uses finite
#' differences on `qfun` or `pfun` instead.
#' @param dqfun A function evaluating the reciprocal of the derivative of
#'   `qfun`, i.e. the density at a given CDF value. If `NULL` (default), uses
#'   `dfun(qfun(...),...)` or finite differences on `qfun` or `pfun` instead.
#'   Must support the same arguments as `qfun`.
#' @param inverse logical; If `FALSE` (default), [bm_marginal()]
#' defines a mapping from standard Normal to a specified distribution.
#' If `TRUE`, it defines a mapping from the specified distribution to a standard
#' Normal.
#' @examples
#' m <- bm_marginal(qexp, pexp, rate = 1 / 8)
#' (val <- ibm_eval(m, state = -5:5))
#' ibm_eval(m, state = val, reverse = TRUE)
#' @rdname bm_marginal
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_marginal(qexp, pexp, dexp, rate = 1 / 8)
#' ibm_eval2(m, state = -3:3)
#'
bm_marginal <- function(qfun,
                        pfun = NULL,
                        dfun = NULL,
                        dqfun = NULL,
                        ...,
                        inverse = FALSE) {
  arg_names <- ...names()
  require_args(qfun, c("p", "lower.tail", "log.p", arg_names))
  require_args(pfun, c("q", "lower.tail", "log.p", arg_names))
  require_args(dfun, c("x", "log", arg_names))
  require_args(dqfun, c("p", "lower.tail", "log.p", "log", arg_names))
  bru_mapper_define(
    list(
      qfun = qfun,
      pfun = pfun,
      dfun = dfun,
      dqfun = dqfun,
      inverse = inverse,
      param = list(...),
      is_linear = FALSE
    ),
    new_class = "bm_marginal"
  )
}

#' @export
#' @rdname bm_marginal
#' @param \dots Arguments passed on to [bm_marginal()]
bru_mapper_marginal <- function(...) {
  bm_marginal(...)
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_marginal <- function(mapper,
                              ...,
                              state = NULL,
                              n_state = NULL) {
  # Output size depends on the state size
  if (!is.null(state)) {
    length(state)
  } else if (is.null(n_state)) {
    NA_integer_
  } else {
    n_state
  }
}
#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_marginal <- function(mapper,
                                     input,
                                     state = NULL,
                                     ...,
                                     n_state = NULL) {
  if (!is.null(state)) {
    length(state)
  } else if (is.null(n_state)) {
    # To allow scalar input weights, do not assume NROW(input) size
    NA_integer_
  } else {
    n_state
  }
}
#' @export
#' @rdname ibm_values
#'
ibm_values.bm_marginal <- function(mapper, ...,
                                   state = NULL, n_state = NULL) {
  n_state <- ibm_n(mapper, state = state, n_state = n_state)
  if (is.na(n_state)) {
    # Don't know how big the mapper will be
    NULL
  } else {
    seq_len(n_state)
  }
}

#' @export
#' @param reverse logical; control `bm_marginal` evaluation. Default
#'   `FALSE`. When `TRUE`, reverses the direction of the mapping, see details
#'   for `marginal` mappers.
#' @describeIn ibm_jacobian Non-NULL `input` values are interpreted
#' as a parameter list for `qfun`, overriding that of the mapper itself.
#'
ibm_jacobian.bm_marginal <- function(mapper, input, state = NULL,
                                     ...,
                                     reverse = FALSE) {
  stopifnot(!is.null(state))
  if (!missing(input) && !is.null(input)) {
    mapper$param <- input
  }
  if (!xor(isTRUE(mapper[["inverse"]]), reverse)) {
    if (!is.null(mapper[["dqfun"]])) {
      val <- pnorm(state)
      dq <- do.call(mapper[["dqfun"]], c(list(val), mapper$param))
      der <- dnorm(state) / dq
    } else if (!is.null(mapper[["dfun"]])) {
      mapper[["inverse"]] <- FALSE
      val <- ibm_eval(
        mapper,
        input = NULL,
        state = state,
        reverse = FALSE
      )
      log_dens <-
        do.call(mapper[["dfun"]], c(list(val, log = TRUE), mapper$param))
      der <- exp(dnorm(state, log = TRUE) - log_dens)
    } else {
      der <- NULL
    }
  } else if (!is.null(mapper[["dqfun"]]) ||
    !is.null(mapper[["dfun"]])) {
    mapper[["inverse"]] <- FALSE
    val <- ibm_eval(
      mapper,
      input = NULL,
      state = state,
      reverse = TRUE
    )
    if (!is.null(mapper[["dqfun"]])) {
      dq <- do.call(mapper[["dqfun"]], c(list(val), mapper$param))
      der <- dq / dnorm(val)
    } else { # if (!is.null(mapper[["dfun"]])) {
      log_dens <-
        do.call(mapper[["dfun"]], c(list(state, log = TRUE), mapper$param))
      der <- exp(log_dens - dnorm(val, log = TRUE))
    }
  } else {
    der <- NULL
  }
  if (is.null(der)) {
    eps <- 1e-4 * (1 + abs(state))
    der <-
      (
        ibm_eval(
          mapper,
          input = NULL,
          state = state + eps,
          reverse = reverse
        ) -
          ibm_eval(
            mapper,
            input = NULL,
            state = state - eps,
            reverse = reverse
          )
      ) /
        (2 * eps)
  }
  Matrix::Diagonal(n = length(state), der)
}


#' @export
#' @describeIn ibm_eval When `xor(mapper[["inverse"]], reverse)` is
#' `FALSE`, `ibm_eval()`
#' for `marginal` returns `qfun(pnorm(x), param)`, evaluated in a numerically
#' stable way. Otherwise, evaluates the inverse `qnorm(pfun(x, param))` instead.
#'
ibm_eval.bm_marginal <- function(mapper, input, state = NULL,
                                 ...,
                                 reverse = FALSE) {
  stopifnot(!is.null(state))
  if (!missing(input) && !is.null(input)) {
    mapper$param <- input
  }
  if (!xor(isTRUE(mapper[["inverse"]]), reverse)) {
    val <- do.call(
      bru_forward_transformation,
      c(
        list(
          qfun = mapper[["qfun"]],
          x = state
        ),
        mapper[["param"]]
      )
    )
  } else {
    val <- do.call(
      bru_inverse_transformation,
      c(
        list(
          pfun = mapper[["pfun"]],
          x = state
        ),
        mapper[["param"]]
      )
    )
  }
  as.vector(val)
}


## _pipe ####

#' @title Mapper for linking several mappers in sequence
#' @export
#' @description
#' Create a pipe mapper, where `mappers` is a list of mappers,
#' and the evaluated output of each mapper is handed as the state to the next
#' mapper.
#' The `input` format for the `ibm_eval` and `ibm_jacobian` methods is
#' a list of inputs, one for each mapper.
#' @rdname bm_pipe
#' @inheritParams bm_multi
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_pipe(list(
#'   scale = bm_scale(),
#'   shift = bm_shift()
#' ))
#' ibm_eval2(m, input = list(scale = 2, shift = 1:4), state = 1:4)
#'
bm_pipe <- function(mappers) {
  mappers <- as_bm_list(mappers)
  is_linear_multi <- vapply(mappers, function(x) ibm_is_linear(x), TRUE)
  is_rowwise_multi <- vapply(mappers, function(x) ibm_is_rowwise(x), TRUE)
  n_multi <- vapply(mappers, function(x) as.integer(ibm_n(x)), 0L)
  n <- ibm_n(mappers[[1]])
  is_linear <- all(is_linear_multi)
  is_rowwise <- all(is_rowwise_multi)
  the_names <- names(mappers)
  if (is.null(the_names)) {
    the_names <- as.character(seq_along(mappers))
    names(mappers) <- the_names
  } else if ("" %in% names(mappers)) {
    stop("Either all or none of the pipe sub-mappers should be named.")
  }
  bru_mapper_define(
    list(
      mappers = mappers,
      is_linear_multi = is_linear_multi,
      is_linear = is_linear,
      is_rowwise = is_rowwise,
      n_multi = n_multi,
      n = n,
      names = the_names
    ),
    new_class = "bm_pipe"
  )
}

#' @export
#' @rdname bm_pipe
#' @param \dots Arguments passed on to [bm_pipe()]
bru_mapper_pipe <- function(...) {
  bm_pipe(...)
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_pipe <- function(mapper, ..., input = NULL, state = NULL) {
  if (is.null(mapper[["mappers"]][[1]]) && is.null(state)) {
    return(NA_integer_)
  }
  if (is.null(mapper[["names"]])) {
    mapper[["names"]] <- names(mapper[["mappers"]])
  }
  if (!is.null(input) && is.null(names(input))) {
    names(input) <- mapper[["names"]][seq_along(input)]
  }
  ibm_n(mapper[["mappers"]][[1]], ...,
    input = input[[names(mapper[["mappers"]])[1]]],
    state = state
  )
}

#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_pipe <- function(mapper,
                                 input,
                                 state = NULL,
                                 ...,
                                 n_state = NULL) {
  if (is.null(mapper[["names"]])) {
    mapper[["names"]] <- names(mapper[["mappers"]])
  }
  if (!is.null(input) && is.null(names(input))) {
    names(input) <- mapper[["names"]][seq_along(input)]
  }
  n <- ibm_n(
    mapper[["mappers"]][[1]],
    input = input[[names(mapper[["mappers"]])[1]]],
    state = state,
    ...,
    n_state = n_state
  )
  for (k in names(mapper[["mappers"]])) {
    n <- ibm_n_output(
      mapper[["mappers"]][[k]],
      input = input[[k]],
      ..., n_state = n
    )
  }
  n
}

#' @export
#' @rdname ibm_values
#'
ibm_values.bm_pipe <- function(mapper, ...) {
  ibm_values(mapper[["mappers"]][[1]], ...)
}

#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_pipe <- function(mapper, input, state = NULL, ...) {
  ibm_eval2(mapper, input = input, state = state, ...)$jacobian
}


#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_pipe <- function(mapper, input, state = NULL, ...) {
  if (is.null(mapper[["names"]])) {
    mapper[["names"]] <- names(mapper[["mappers"]])
  }
  if (!is.null(input) && is.null(names(input))) {
    names(input) <- mapper[["names"]][seq_along(input)]
  }
  state_k <- state
  for (k in names(mapper[["mappers"]])) {
    state_k <- ibm_eval(
      mapper[["mappers"]][[k]],
      input = input[[k]],
      state = state_k,
      ...
    )
  }
  state_k
}


#' @export
#' @rdname ibm_eval2
#'
ibm_eval2.bm_pipe <- function(mapper, input, state = NULL, ...) {
  if (is.null(mapper[["names"]])) {
    mapper[["names"]] <- names(mapper[["mappers"]])
  }
  if (!is.null(input) && is.null(names(input))) {
    names(input) <- mapper[["names"]][seq_along(input)]
  }
  state_k <- state
  first <- names(mapper[["mappers"]])[1]
  for (k in names(mapper[["mappers"]])) {
    eval2_k <- ibm_eval2(
      mapper[["mappers"]][[k]],
      input = input[[k]],
      state = state_k,
      ...
    )
    state_k <- eval2_k$offset
    if (k == first) {
      A <- eval2_k$jacobian
    } else {
      A <- eval2_k$jacobian %*% A
    }
  }
  list(offset = state_k, jacobian = A)
}


#' @describeIn ibm_simplify
#' Constructs a simplified `pipe` mapper. For fully linear pipes, calls
#' [ibm_linear()].
#' For partially non-linear pipes, replaces each sequence of linear mappers with
#' a single [bm_taylor()] mapper, while keeping the full list of
#' original mapper names, allowing the original `input` structure to be used
#' also with the simplified mappers, since the `taylor` mappers are not
#' dependent on inputs.
#' @export
#'
ibm_simplify.bm_pipe <- function(mapper,
                                 input = NULL,
                                 state = NULL,
                                 inla_f = FALSE,
                                 ...,
                                 n_state = NULL) {
  if (is.null(mapper[["names"]])) {
    mapper[["names"]] <- names(mapper[["mappers"]])
  }
  if (!is.null(input) && is.null(names(input))) {
    names(input) <- mapper[["names"]][seq_along(input)]
  }
  if (ibm_is_linear(mapper)) {
    if (is.null(state)) {
      n <- ibm_n(mapper[["mappers"]][[1]],
        ...,
        inla_f = FALSE,
        input = input[[names(mapper[["mappers"]])[1]]],
        n_state = n_state
      )
      state <- rep(0, n)
    }
    return(ibm_linear(mapper,
      input = input, state = state, ...,
      inla_f = inla_f, n_state = n_state
    ))
  }

  # Basic version, that just replaces each individual mapper with a
  # simplification.
  n <- ibm_n(
    mapper[["mappers"]][[1]],
    input = input[[names(mapper[["mappers"]])[1]]],
    state = state,
    ...,
    inla_f = FALSE,
    n_state = n_state
  )
  for (k in names(mapper[["mappers"]])) {
    if (ibm_is_linear(mapper[["mappers"]][[k]])) {
      mapper[["mappers"]][[k]] <- ibm_simplify(
        mapper[["mappers"]][[k]],
        input = input[[k]],
        ...,
        n_state = n
      )
    }
    n <- ibm_n_output(
      mapper[["mappers"]][[k]],
      input = input[[k]],
      ...,
      inla_f = FALSE,
      n_state = n
    )
  }

  mapper
}


## _multi ####

#' @title Mapper for tensor product domains
#' @param mappers A list of `bru_mapper` objects
#' @param simplify logical; If `TRUE`, removes trivial submappers.
#' Currently only sub-mappers of class [bm_index()] with `ibm_n() == 1L` are
#' removed, and only if the mappers are named (to avoid ordering mismatches).
#' Default: `FALSE`
#' @description Constructs a row-wise Kronecker product mapping of linear/affine
#'   mappers. Any offset in sub-mappers is added into a combined offset. Only
#'   linear/affine sub-mappers are allowed.
#' @rdname bm_multi
#' @export
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' (m <- bm_multi(list(a = bm_index(2), b = bm_index(3))))
#' ibm_eval2(m, list(a = c(1, 2, 1), b = c(1, 3, 2)), 1:6)
#'
bm_multi <- function(mappers, simplify = FALSE) {
  if (!is.list(mappers)) {
    stop("bm_multi requires a list of sub-mappers.")
  }
  if (is.null(names(mappers))) {
    names(mappers) <- as.character(seq_along(mappers))
  } else if (any(names(mappers) == "")) {
    stop("Either all or none of the multi sub-mappers should be named.")
  }
  mappers <- as_bm_list(mappers)

  if (simplify) {
    trivial <- vapply(mappers, function(x) {
      inherits(x, "bm_index") && (ibm_n(x) == 1L)
    }, logical(1))
    if (any(trivial)) {
      if (is.null(names(mappers))) {
        warning("Not removing trivial sub-mappers in unnamed bm_multi.")
      } else {
        mappers <- mappers[!trivial]
      }
    }
  }

  mapper <- list(
    mappers = mappers,
    n_multi = lapply(mappers, ibm_n),
    n_inla_multi = lapply(mappers, function(x) {
      ibm_n(x, inla_f = TRUE)
    }),
    values_multi = lapply(mappers, ibm_values),
    values_inla_multi = lapply(mappers, function(x) {
      ibm_values(x, inla_f = TRUE)
    }),
    is_linear_multi = lapply(mappers, ibm_is_linear),
    is_rowwise_multi = lapply(mappers, ibm_is_rowwise)
  )
  mapper[["n"]] <- prod(unlist(mapper[["n_multi"]]))
  mapper[["n_inla"]] <- prod(unlist(mapper[["n_inla_multi"]]))
  mapper[["is_linear"]] <- all(unlist(mapper[["is_linear_multi"]]))
  mapper[["is_rowwise"]] <- all(unlist(mapper[["is_rowwise_multi"]]))

  if (!mapper[["is_linear"]]) {
    stop("bm_multi sub-mappers must be linear mappers")
  }
  bru_mapper_define(mapper, new_class = "bm_multi")
}

#' @export
#' @rdname bm_multi
#' @param \dots Arguments passed on to [bm_multi()]
bru_mapper_multi <- function(...) {
  bm_multi(...)
}

#' @param multi logical;
#' If `TRUE` (or positive), recurse one level into sub-mappers
#' @export
#' @rdname ibm_n
#'
ibm_n.bm_multi <- function(mapper, inla_f = FALSE, multi = FALSE, ...) {
  if (multi) {
    if (inla_f) {
      mapper[["n_inla_multi"]]
    } else {
      mapper[["n_multi"]]
    }
  } else if (inla_f) {
    mapper[["n_inla"]]
  } else {
    mapper[["n"]]
  }
}

#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_multi <- function(mapper, input, ...) {
  input <- bm_multi_prepare_input(mapper, input)
  # Assume that the first mapper fully handles the output size
  nm <- names(mapper[["mappers"]])[1]
  if (is.matrix(input)) {
    ibm_n_output(mapper[["mappers"]][[nm]], input[, nm, drop = TRUE], ...)
  } else {
    stopifnot(is.list(input)) # data.frame or list
    ibm_n_output(mapper[["mappers"]][[nm]], input[[nm]], ...)
  }
}


#' @export
#' @rdname ibm_values
#'
ibm_values.bm_multi <- function(mapper,
                                inla_f = FALSE,
                                multi = FALSE,
                                ...) {
  if (multi) {
    vals <- if (inla_f) {
      mapper[["values_inla_multi"]]
    } else {
      mapper[["values_multi"]]
    }
    if (any(vapply(vals, is.null, logical(1)))) {
      return(NULL)
    }
    # Expand indices/values. First sub-mapper varies fastest
    as.data.frame(
      do.call(
        expand.grid,
        c(
          vals,
          list(
            KEEP.OUT.ATTRS = FALSE,
            stringsAsFactors = FALSE
          )
        )
      )
    )
  } else if (inla_f) {
    seq_len(mapper[["n_inla"]])
  } else {
    seq_len(mapper[["n"]])
  }
}

#' @export
#' @rdname ibm_is_linear
#'
ibm_is_linear.bm_multi <- function(mapper, multi = FALSE, ...) {
  if (multi) {
    mapper[["is_linear_multi"]]
  } else {
    mapper[["is_linear"]]
  }
}

bm_multi_prepare_input <- function(mapper, input) {
  if (is.matrix(input)) {
    input_names <- colnames(input)
    input <- as.data.frame(input)
  } else {
    input_names <- names(input)
  }
  if (is.null(input_names)) {
    if (is.data.frame(input)) {
      names(input) <- names(mapper[["mappers"]])[seq_len(ncol(input))]
    } else {
      if (is.null(input)) {
        input <- rep(list(NULL), length(mapper[["mappers"]]))
      }
      # input should now be a list
      names(input) <- names(mapper[["mappers"]])[seq_along(input)]
    }
  }
  input
}

#' @describeIn ibm_jacobian
#' Accepts a list with
#' named entries, or a list with unnamed but ordered elements.
#' The names must match the sub-mappers, see [ibm_names.bm_multi()].
#' Each list element should take a format accepted by the corresponding
#' sub-mapper. In case each element is a vector, the input can be given as a
#' data.frame with named columns, a matrix with named columns, or a matrix
#' with unnamed but ordered columns.
#' @param sub_A Internal; precomputed Jacobian matrices.
#' @export
#'
ibm_jacobian.bm_multi <- function(mapper,
                                  input,
                                  state = NULL,
                                  inla_f = FALSE,
                                  multi = FALSE,
                                  ...,
                                  sub_A = NULL) {
  input <- bm_multi_prepare_input(mapper, input)

  # Note: the multi-mapper only handles linear/affine sub-mappers.
  if (is.null(sub_A)) {
    sub_A <-
      lapply(
        names(mapper[["mappers"]]),
        function(x) {
          ibm_jacobian(mapper[["mappers"]][[x]],
            input = input[[x]],
            state = NULL,
            inla_f = inla_f,
            multi = FALSE
          )
        }
      )
  }
  if (multi) {
    return(sub_A)
  }
  # Combine the matrices
  # (A1, A2, A3) -> rowkron(A3, rowkron(A2, A1))
  A_ <- sub_A[[1]]
  for (k in seq_len(length(mapper[["mappers"]]) - 1)) {
    A_ <- fm_row_kron(sub_A[[k + 1]], A_)
  }
  A_
}


#' @export
#' @rdname ibm_linear
#'
ibm_linear.bm_multi <- function(mapper, input, state,
                                inla_f = FALSE,
                                ...) {
  eval2 <- ibm_eval2(
    mapper, input,
    state = state,
    inla_f = inla_f,
    multi = FALSE,
    ...
  )
  bru_mapper_taylor(
    offset = eval2$offset,
    jacobian = eval2$jacobian,
    state0 = state,
    values_mapper = mapper
  )
}


#' @param jacobian
#' For `ibm_eval()` methods, an optional pre-computed Jacobian, typically
#' supplied by internal methods that already have the Jacobian.
#' @param pre_A `r lifecycle::badge("deprecated")` in favour of `jacobian`.
#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_multi <- function(mapper, input, state = NULL,
                              inla_f = FALSE, ...,
                              jacobian = NULL,
                              pre_A = deprecated()) {
  input <- bm_multi_prepare_input(mapper, input)

  if (ibm_n(mapper) == 0L) {
    # TODO: Ensure that this code is only reached when the first sub-mapper
    # is constant (mapper_const or a pipe that is state-invariant)
    val <- 0
    m <- names(mapper[["mappers"]])[1]
    val_add <- ibm_eval(mapper[["mappers"]][[m]],
      input = input[[m]],
      state = NULL,
      inla_f = inla_f,
      multi = FALSE
    )
    val <- val + val_add
    return(val)
  }

  val <- rep(0, ibm_n_output(mapper, input, state = state, inla_f = inla_f))
  for (m in names(mapper[["mappers"]])) {
    val_add <- ibm_eval(mapper[["mappers"]][[m]],
      input = input[[m]],
      state = NULL,
      inla_f = inla_f,
      multi = FALSE
    )
    val <- val + val_add
  }
  if ((ibm_n(mapper) > 0) && !is.null(state)) {
    if (is.null(jacobian)) {
      jacobian <- ibm_jacobian(mapper,
        input = input, state = state,
        inla_f = inla_f, multi = FALSE, ...
      )
    }
    val <- val + jacobian %*% state
  }
  as.vector(val)
}


bm_multi_indexing <- function(mapper, input) {
  if (is.matrix(input)) {
    nms <- colnames(input)
    len <- ncol(input)
  } else {
    stopifnot(is.list(input))
    nms <- names(input)
    len <- length(input)
  }

  nms_mapper <- names(mapper[["mappers"]])
  if (is.null(nms)) {
    indexing <- seq_len(min(length(nms_mapper), length(input)))
  } else {
    indexing <- intersect(nms_mapper, nms)
  }
  names(indexing) <- indexing
  indexing
}


#' @describeIn ibm_invalid_output
#' Accepts a list with
#' named entries, or a list with unnamed but ordered elements.
#' The names must match the sub-mappers, see [ibm_names.bm_multi()].
#' Each list element should take a format accepted by the corresponding
#' sub-mapper. In case each element is a vector, the input can be given as a
#' data.frame with named columns, a matrix with named columns, or a matrix
#' with unnamed but ordered columns.
#' @export
#'
ibm_invalid_output.bm_multi <- function(mapper,
                                        input,
                                        state,
                                        inla_f = FALSE,
                                        multi = FALSE,
                                        ...) {
  indexing <- bm_multi_indexing(mapper, input)
  if (is.matrix(input)) {
    invalid <-
      lapply(
        indexing,
        function(x) {
          ibm_invalid_output(
            mapper[["mappers"]][[x]],
            input = input[, x],
            inla_f = inla_f,
            multi = FALSE
          )
        }
      )
  } else if (is.list(input)) {
    invalid <-
      lapply(
        indexing,
        function(x) {
          ibm_invalid_output(
            mapper[["mappers"]][[x]],
            input = input[[x]],
            inla_f = inla_f,
            multi = FALSE
          )
        }
      )
  } else {
    invalid <- as.list(rep(TRUE, length(mapper[["mappers"]])))
    names(invalid) <- names(mapper[["mappers"]])
  }
  if (!multi) {
    # Combine the vectors (v1, v2, v3) -> v1 | v2 | v3
    invalid_ <- invalid[[1]]
    for (k in seq_len(length(mapper[["mappers"]]) - 1)) {
      invalid_ <- invalid_ | invalid[[k + 1]]
    }
    return(invalid_)
  }
  invalid
}

#' @return
#' * `[`-indexing a `bm_multi` extracts a subset
#'   `bm_multi` object (for drop `FALSE`) or an individual sub-mapper
#'   (for drop `TRUE`, and `i` identifies a single element)
#' @export
#' @param x object from which to extract element(s)
#' @param i indices specifying element(s) to extract
#' @param drop logical;
#' For `[.bm_multi`, whether to extract an individual mapper when
#' `i` identifies a single element. If `FALSE`, a list of sub-mappers is
#' returned (suitable e.g. for creating a new `bm_multi` object).
#' Default: `TRUE`
#' @rdname bm_multi
`[.bm_multi` <- function(x, i, drop = TRUE) {
  if (is.logical(i)) {
    i <- which(i)
  }
  mapper <- x[["mappers"]][i]
  if (drop) {
    if (length(mapper) == 1) {
      mapper <- mapper[[1]]
    } else if (length(mapper) == 0) {
      mapper <- NULL
    }
  }
  mapper
}

#' @rdname bm_multi
#' @export
`[.bru_mapper_multi` <- function(x, i, drop = TRUE) {
  if (is.logical(i)) {
    i <- which(i)
  }
  mapper <- x[["mappers"]][i]
  if (drop) {
    if (length(mapper) == 1) {
      mapper <- mapper[[1]]
    } else if (length(mapper) == 0) {
      mapper <- NULL
    }
  }
  mapper
}

#' @describeIn ibm_names
#' Returns the names from the sub-mappers list
#' @export
#'
`ibm_names.bm_multi` <- function(mapper) {
  names(mapper[["mappers"]])
}

#' @param value a character vector of up to the same length as the number
#' of mappers in the multi-mapper x
#' @export
#' @rdname ibm_names
#'
`ibm_names<-.bm_multi` <- function(mapper, value) {
  names(mapper[["mappers"]]) <- value
  names(mapper[["n_multi"]]) <- value
  names(mapper[["n_inla_multi"]]) <- value
  names(mapper[["values_multi"]]) <- value
  names(mapper[["values_inla_multi"]]) <- value
  mapper
}

#' @export
#' @rdname ibm_names
`ibm_names<-.bru_mapper_multi` <- function(mapper, value) {
  names(mapper[["mappers"]]) <- value
  names(mapper[["n_multi"]]) <- value
  names(mapper[["n_inla_multi"]]) <- value
  names(mapper[["values_multi"]]) <- value
  names(mapper[["values_inla_multi"]]) <- value
  mapper
}

## _harmonics ####

#' @title Mapper for cos/sin functions
#' @param order For `bm_harmonics`, specifies the maximum `cos`/`sin`
#' order. (Default 1)
#' @param scaling For `bm_harmonics`, specifies an optional vector of
#'  scaling factors of length `intercept + order`, or a common single scalar.
#' @param intercept logical; For `bm_harmonics`, if `TRUE`, the first
#'   basis function is a constant. (Default `TRUE`)
#' @param interval numeric length-2 vector specifying a domain interval.
#'   Default `c(0, 1)`.
#' @description Constructs a mapper for `cos`/`sin` functions of orders 1 (if
#'   `intercept` is `TRUE`, otherwise 0) through `order`. The total number of
#'   basis functions is `intercept + 2 * order`.
#'
#'   Optionally, each order can be given a non-unit scaling, via the `scaling`
#'   vector, of length `intercept + order`. This can be used to
#'   give an effective spectral prior. For example, let
#'   ```
#'   scaling = 1 / (1 + (0:4)^2)
#'   x <- seq(0, 1, length.out = 11)
#'   bmh1 = bm_harmonics(order = 4, interval = c(0, 1))
#'   u1 <- ibm_eval(
#'     bmh1,
#'     input = x,
#'     state = rnorm(9, sd = rep(scaling, c(1, 2, 2, 2, 2)))
#'   )
#'   ```
#'   Then, with
#'   ```
#'   bmh2 = bm_harmonics(order = 4, scaling = scaling)
#'   u2 = ibm_eval(bmh2, input = x, state = rnorm(9))
#'   ```
#'   the stochastic properties of `u1` and `u2` will be the same, with
#'   `scaling^2` determining the variance for each frequency contribution.
#'
#'   The period for the first order harmonics is shifted and scaled to match
#'   `interval`.
#' @export
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' m <- bm_harmonics(2)
#' ibm_eval2(m, input = c(0, pi / 4, pi / 2, 3 * pi / 4), 1:5)
#'
#' @rdname bm_harmonics
#'
bm_harmonics <- function(order = 1,
                         scaling = 1,
                         intercept = TRUE,
                         interval = c(0, 1)) {
  if (length(scaling) == 1) {
    scaling <- rep(scaling, intercept + order)
  } else {
    stopifnot(length(scaling) == intercept + order)
  }
  mapper <- list(
    order = order,
    scaling = scaling,
    intercept = intercept,
    interval = interval
  )
  bru_mapper_define(mapper, new_class = "bm_harmonics")
}

#' @export
#' @rdname bm_harmonics
#' @param \dots Arguments passed on to [bm_harmonics()]
bru_mapper_harmonics <- function(...) {
  bm_harmonics(...)
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_harmonics <- function(mapper, inla_f = FALSE, ...) {
  mapper[["order"]] * 2 + mapper[["intercept"]]
}

#' @export
#' @rdname ibm_jacobian
#'
ibm_jacobian.bm_harmonics <- function(mapper,
                                      input,
                                      state = NULL,
                                      inla_f = FALSE,
                                      ...) {
  # Indexing into sparseMatrix is slower than into a dense Matrix,
  # and indexing into a dense Matrix is slower than for a plain matrix(),
  # including conversion to dense Matrix afterwards.
  A <- matrix(0.0, NROW(input), ibm_n(mapper))
  off <- 0
  if (mapper[["intercept"]]) {
    A[, 1] <- 1.0 * mapper[["scaling"]][1]
    off <- off + 1
  }
  if (mapper[["order"]] > 0) {
    input <- (input - mapper[["interval"]][1]) / diff(mapper[["interval"]])
    for (ord in seq_len(mapper[["order"]])) {
      scale <- mapper[["scaling"]][mapper[["intercept"]] + ord]
      angle <- (2 * pi * ord) * input
      A[, off + 1] <- cos(angle) * scale
      A[, off + 2] <- sin(angle) * scale
      off <- off + 2
    }
  }
  as(A, "Matrix")
}


## _reparam ####

#' @title Mapper for reparameterising mapper states
#' @param mapper A `bru_mapper` object
#' @param B a square or rectangular basis conversion matrix
# @param \dots Arguments passed on to submethods
#' @export
#' @description Creates a mapper for handling basis conversions. Functionally
#'   equivalent to `bm_pipe(list(bm_matrix(ncol(B)), mapper))`, but with an
#'   internally stored matrix input `B` for efficiency, and allowing the mapper
#'   `input` format to be identical to that of the original `mapper`.
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @rdname bm_reparam
#' @examples
#' # 2->2 reparameterisation; (u1,u2) -> (u1, u1 + u2)
#' (m <- bm_reparam(bm_index(2), B = matrix(c(1, 1, 0, 1), 2, 2)))
#'
#' # 2->3 reparameterisation; (u1,u2) -> (u1, u2, u1+u2)
#' # This is an example of a low-rank representation of a higher-dimensional
#' # state vector.
#' (m <- bm_reparam(bm_index(3), B = cbind(c(1, 0, 1), c(0, 1, 1))))
#'
bm_reparam <- function(mapper, B) {
  map_n <- ibm_n(mapper)
  stopifnot(is.na(map_n) || (map_n == nrow(B)))
  map <- list(
    mapper = mapper,
    B = B,
    n = ncol(B),
    is_linear = ibm_is_linear(mapper),
    is_rowwise = ibm_is_rowwise(mapper)
  )
  bru_mapper_define(map, new_class = "bm_reparam")
}


#' @title Deprecated basis conversion mapper
#' @description `r lifecycle::badge("deprecated")`
#'  Old deprecated name for [bm_reparam()]
#' @param mesh Object supported by a [bru_mapper()] method.
#' @param B a square or tall basis conversion matrix
#' @export
#' @keywords internal
bm_mesh_B <- function(mesh, B) {
  bm_reparam(bru_mapper(mesh), B)
}

#' @export
#' @describeIn bm_mesh_B `r lifecycle::badge("deprecated")`
#'   Deprecated name for `bm_mesh_B`
bru_mapper_mesh_B <- function(...) {
  bm_mesh_B(...)
}

#' @export
#' @rdname ibm_n
#' @inheritParams ibm_n
ibm_n.bm_reparam <- function(mapper, ...) {
  mapper[["n"]]
}
#' @export
#' @rdname ibm_values
ibm_values.bm_reparam <- function(mapper, ...) {
  seq_len(mapper[["n"]])
}
#' @export
#' @rdname ibm_n_output
ibm_n_output.bm_reparam <- function(mapper, ...) {
  ibm_n_output(mapper[["mapper"]], ...)
}
#' @export
#' @rdname ibm_jacobian
ibm_jacobian.bm_reparam <- function(mapper, input, state = NULL, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  if (!is.null(state)) {
    state <- mapper[["B"]] %*% state
  }
  A <- ibm_jacobian(mapper[["mapper"]], input = input, state = state, ...)
  A %*% mapper[["B"]]
}
#' @export
#' @rdname ibm_eval
ibm_eval.bm_reparam <- function(mapper, input, state = NULL, ...,
                                jacobian = NULL) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  if (!is.null(state)) {
    state <- mapper[["B"]] %*% state
  }
  # Note: Have to prevent the jacobian to be fed through, as it's the jacobian
  # for the remapped state, not the original state.
  val <- ibm_eval(mapper[["mapper"]], input = input, state = state, ...)
  val
}

test_that("Mexdolphin: Hazard rate detection function", {
  skip_on_cran()
  local_bru_safe_inla()

  mexdolphin <- inlabru::mexdolphin_sf

  sig <- function(x) bru_forward_transformation(qexp, x, rate = 1 / 8)
  hr <- function(distance, sigma) {
    1 - exp(-(distance / sigma)^-1)
  }
  log_hr <- function(distance, sigma) {
    log1p(-exp(-(distance / sigma)^-1))
  }
  cmp <- ~ sig_theta(1, prec.linear = 1) + Intercept(1)
  form <- distance ~ log_hr(distance, sig(sig_theta)) + Intercept
  form_list <- list(distance = distance) ~
    log(hr(distance, sig(sig_theta))) + Intercept

  pts <-
    mexdolphin$points

  fit <- bru(
    components = cmp,
    bru_obs(
      formula = form,
      family = "cp",
      data = pts,
      domain = list(distance = fm_mesh_1d(seq(0, 8, by = 0.1)))
    ),
    options = list(
      bru_verbose = 0,
      bru_compress_cp = TRUE,
      bru_max_iter = 10,
      verbose = FALSE,
      bru_initial = list(Intercept = 0, sig_theta = -1),
      inla.mode = "compact"
    )
  )

  expect_equal(
    fit$summary.fixed["sig_theta", "mean"],
    -0.4653159,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["sig_theta", "sd"],
    0.3583088,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    2.2593613,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "sd"],
    0.2720904,
    tolerance = midtol
  )

  fit_list <- bru(
    components = cmp,
    bru_obs(
      formula = form_list,
      family = "cp",
      data = pts,
      domain = list(distance = fm_mesh_1d(seq(0, 8, by = 0.1)))
    ),
    options = list(
      bru_verbose = 0,
      bru_compress_cp = TRUE,
      bru_max_iter = 10,
      verbose = FALSE,
      bru_initial = list(Intercept = 0, sig_theta = -1),
      inla.mode = "compact"
    )
  )

  expect_equal(
    fit_list$summary.fixed["sig_theta", "mean"],
    -0.4653159,
    tolerance = midtol
  )
  expect_equal(
    fit_list$summary.fixed["sig_theta", "sd"],
    0.3583088,
    tolerance = midtol
  )
  expect_equal(
    fit_list$summary.fixed["Intercept", "mean"],
    2.2593613,
    tolerance = midtol
  )
  expect_equal(
    fit_list$summary.fixed["Intercept", "sd"],
    0.2720904,
    tolerance = midtol
  )

  ##########
  # pred <- predict(
  #   fit,
  #   newdata = data.frame(distance = seq(-8, 8, by = 0.01)),
  #   formula = ~ exp(Intercept + log_hr(abs(distance), sig(sig_theta)))
  # )
  #
  # library(ggplot2)
  # library(tidyverse)
  # ggplot(data.frame(distance = c(
  #   pts$distance,
  #   -pts$distance
  # ))) +
  #   geom_density(aes(distance, after_stat(count))) +
  #   geom_line(aes(distance, est, col = "Plugin"),
  #             data = data.frame(distance = seq(-8, 8, by = 0.001)) |>
  #               mutate(est = hr(
  #                 abs(distance),
  #                 sig(fit$summary.fixed["sig_theta", "mean"])
  #               ) *
  #                 exp(fit$summary.fixed["Intercept", "mean"]))
  #   ) +
  #   geom_line(aes(distance, mean, col = "Pred"),
  #             data = pred
  #   ) +
  #   geom_line(aes(distance, est, col = "1"),
  #             data = data.frame(distance = seq(-8, 8, by = 0.001)) |>
  #               mutate(est = hr(
  #                 abs(distance),
  #                 1
  #               ) *
  #                 exp(2.3))
  #   )
  #
  # pred <- predict(
  #   fit,
  #   data = ips,
  #   formula = ~ exp(Intercept + log_hr(abs(distance), sig(sig_theta)))
  # )
  # ggplot(data.frame(pts)) +
  #   stat_ecdf(aes(distance)) +
  #   geom_line(
  #     aes(distance,
  #         cumsum(hr(distance, sig(fit$summary.fixed["sig_theta", "mean"]))) /
  #           sum(hr(distance, sig(fit$summary.fixed["sig_theta", "mean"]))),
  #         col = "Plugin"
  #     ),
  #     data = data.frame(ips) |> arrange(distance)
  #   ) +
  #   geom_line(
  #     aes(distance,
  #         cumsum(mean) / sum(mean),
  #         col = "Pred"
  #     ),
  #     data = pred |> arrange(distance)
  #   ) +
  #   geom_line(
  #     aes(distance,
  #         cumsum(hr(distance, 1)) /
  #           sum(hr(distance, 1)),
  #         col = "1"
  #     ),
  #     data = data.frame(ips) |> arrange(distance)
  #   )
  ##########
})


test_that("Marginal parameter transformation", {
  skip_on_cran()
  local_bru_safe_inla()
  data(mexdolphin_sf, package = "inlabru", envir = environment())

  hr <- function(distance, sigma) {
    1 - exp(-(distance / sigma)^-1)
  }
  log_hr <- function(distance, sigma) {
    log1p(-exp(-(distance / sigma)^-1))
  }
  cmp <- bru_comp_list(~
    sigma(
      1,
      prec.linear = 1,
      marginal = bm_marginal(
        qfun = qexp,
        pfun = pexp,
        dfun = dexp,
        rate = 1 / 8
      )
    ) + Intercept(1))
  form <- distance ~ log_hr(distance, sigma = sigma) + Intercept

  pts <- mexdolphin_sf$points

  fit <- bru(
    components = cmp,
    bru_obs(
      formula = form,
      family = "cp",
      data = pts,
      domain = list(distance = fm_mesh_1d(seq(0, 8, by = 0.1)))
    ),
    options = list(
      bru_verbose = 0,
      bru_compress_cp = TRUE,
      bru_max_iter = 10,
      verbose = FALSE,
      bru_initial = list(
        Intercept = 0,
        sigma = ibm_eval(cmp$sigma$marginal, state = 1, inverse = TRUE)
      ),
      inla.mode = "compact"
    )
  )

  expect_equal(
    fit$summary.fixed["sigma", "mean"],
    -0.4653159,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["sigma", "sd"],
    0.3583088,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    2.2593613,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "sd"],
    0.2720904,
    tolerance = midtol
  )
})

test_that("Aggregated Gaussian observations, using fm_block_eval", {
  skip_on_cran()
  local_bru_safe_inla()

  obs <- data.frame(
    x = c(10, 20, 30),
    y = c(10, 20, 30),
    z = c(10, 20, 30)
  )
  pred <- data.frame(
    x = c(1, 2, 3, 4, 5, 6),
    y = c(1, 20, 3, 40, 5, 60),
    weights = c(1, 1, 1, 1, 1, 1),
    grp = c(1, 1, 2, 2, 2, 3)
  )

  comp <- ~ Intercept(1) + x

  fit <- bru(
    comp,
    bru_obs(
      z ~ fm_block_eval(
        block = grp,
        n_block = nrow(obs),
        weights = weights,
        rescale = TRUE,
        values = Intercept + x
      ),
      family = "normal",
      response_data = obs,
      data = pred,
      control.family = list(
        hyper = list(
          prec = list(
            initial = 6,
            fixed = TRUE
          )
        )
      ),
      is_rowwise = FALSE
    )
  )

  expect_equal(
    fit$summary.fixed$mean,
    c(3.033, 4.426),
    tolerance = midtol
  )

  # With basic sf storage:

  obs_sf <- sf::st_as_sf(obs, coords = c("x", "y"))
  pred_sf <- sf::st_as_sf(pred, coords = c("x", "y"))

  comp_sf <- ~ Intercept(1) + x(sf::st_coordinates(pred_sf)[, "X"])

  fit_sf <- bru(
    comp_sf,
    bru_obs(
      z ~ fm_block_eval(
        block = grp,
        weights = weights,
        rescale = TRUE,
        n_block = nrow(obs),
        values = Intercept + x
      ),
      family = "normal",
      response_data = obs_sf,
      data = pred_sf,
      control.family = list(
        hyper = list(prec = list(initial = 6, fixed = TRUE))
      ),
      is_rowwise = FALSE
    )
  )

  expect_equal(
    fit_sf$summary.fixed$mean,
    c(3.033, 4.426),
    tolerance = midtol
  )
})


test_that("Aggregated Gaussian observations, using aggregate feature", {
  skip_on_cran()
  local_bru_safe_inla()

  obs <- data.frame(
    x = c(10, 20, 30),
    y = c(1000, 2000, 3000),
    z = c(10, 20, 30),
    the_group_char = c("a", "b", "c")
  )
  pred <- data.frame(
    x = c(1, 2, 3, 4, 5, 6),
    y = c(1, 20, 3, 40, 5, 60),
    weights = c(1, 1, 1, 1, 1, 1),
    grp = c(1, 1, 2, 2, 2, 3),
    grp_char = c("a", "a", "b", "b", "b", "c")
  )
  domain <- list()

  comp <- ~ Intercept(1) + x + y

  fit <- bru(
    comp,
    bru_obs(
      z ~ Intercept + x + y,
      family = "normal",
      response_data = obs,
      data = pred,
      aggregate = "average",
      aggregate_input = list(
        weights = weights,
        block = grp,
        n_block = bru_response_size(.response_data.)
      ),
      control.family = list(
        hyper = list(
          prec = list(
            initial = 6,
            fixed = TRUE
          )
        )
      )
    )
  )

  expect_equal(
    fit$summary.fixed$mean,
    c(3.636, 3.889, 0.0505),
    tolerance = midtol
  )

  expect_no_error({
    bru_obs(
      z ~ Intercept + x + y,
      family = "normal",
      response_data = obs,
      data = pred,
      aggregate = "average",
      aggregate_input = list(
        weights = weights,
        block = grp
      ),
      control.family = list(
        hyper = list(
          prec = list(
            initial = 6,
            fixed = TRUE
          )
        )
      )
    )
  })

  # Missing block information
  expect_error(
    {
      bru_obs(
        z ~ Intercept + x + y,
        family = "normal",
        response_data = obs,
        data = c(as.list(pred), list(.block = NULL)),
        aggregate = "average",
        aggregate_input = list(
          weights = weights,
          n_block = bru_response_size(.response_data.)
        ),
        control.family = list(
          hyper = list(
            prec = list(
              initial = 6,
              fixed = TRUE
            )
          )
        )
      )
    },
    paste0(
      "Aggregation requested, but `aggregate_input[['block']]` ",
      "evaluates to NULL."
    ),
    fixed = TRUE
  )

  # Character block information without matching response block information
  expect_error(
    {
      bru_obs(
        z ~ Intercept + x + y,
        family = "normal",
        response_data = obs,
        data = pred,
        aggregate = "average",
        aggregate_input = list(
          weights = weights,
          block = grp_char,
          n_block = bru_response_size(.response_data.)
        ),
        control.family = list(
          hyper = list(
            prec = list(
              initial = 6,
              fixed = TRUE
            )
          )
        )
      )
    },
    paste0(
      "'character' aggregation block information detected."
    ),
    fixed = TRUE
  )

  # Character block information without match response block information
  expect_no_error({
    bru_obs(
      z ~ Intercept + x + y,
      family = "normal",
      response_data = obs,
      data = pred,
      aggregate = "average",
      aggregate_input = list(
        weights = weights,
        block = grp_char,
        n_block = bru_response_size(.response_data.),
        block_response = "the_group_char"
      ),
      control.family = list(
        hyper = list(
          prec = list(
            initial = 6,
            fixed = TRUE
          )
        )
      )
    )
  })
})


test_that("Aggregated Gaussian observations, using domain/samplers feature", {
  skip_on_cran()
  local_bru_safe_inla()

  obs <- data.frame(
    x = c(10, 20, 30),
    z = c(10, 20, 30)
  )
  pred <- data.frame(
    y = c((1 + 20) / 2, (3 + 40 + 5) / 3, 60)
  )
  domain <- list(x = 1:6)
  samplers <- list(x = list(1:2, 3:5, 6))

  comp <- ~ Intercept(1) + x + y

  fit <- bru(
    comp,
    bru_obs(
      z ~ Intercept + x + y,
      family = "normal",
      response_data = obs,
      data = pred,
      aggregate = "average",
      domain = domain,
      samplers = samplers,
      control.family = list(
        hyper = list(
          prec = list(
            initial = 6,
            fixed = TRUE
          )
        )
      )
    ),
    options = list(control.inla = list(
      int.strategy = "eb"
    ))
  )

  expect_equal(
    fit$summary.fixed$mean,
    c(3.636, 3.889, 0.0505),
    tolerance = midtol
  )

  expect_error(
    {
      bru_model(
        comp,
        bru_obs(
          z ~ Intercept + x + y,
          family = "normal",
          response_data = obs,
          data = NULL,
          aggregate = "average",
          domain = domain,
          samplers = samplers,
          control.family = list(
            hyper = list(
              prec = list(
                initial = 6,
                fixed = TRUE
              )
            )
          )
        )
      )
    },
    paste0(
      "The input evaluation 'y' for 'y' failed.\n",
      "Perhaps the data object doesn't contain the needed variables?"
    ),
    fixed = TRUE
  )
})


test_that("Aggregated Poisson observations, using mapper", {
  skip_on_cran()
  local_bru_safe_inla()

  obs <- data.frame(y = c(10, 20, 30))
  pred <- data.frame(
    x = c(1, 2, 3, 4, 5, 6),
    weights = c(1, 1, 1, 1, 1, 1),
    grp = c(1, 1, 2, 2, 2, 3)
  )

  # Aggregation by summation on the intensity/expectation scale
  # (log-sum-exp since the predictor is log-intensity)
  agg <- bm_logsumexp(rescale = FALSE, n_block = nrow(obs))

  comp <- ~ Intercept(1) + x

  fit <- bru(
    comp,
    bru_obs(
      y ~ ibm_eval(
        agg,
        input = list(weights = weights, block = grp),
        state = Intercept + x
      ),
      family = "poisson",
      response_data = obs,
      data = pred,
      is_rowwise = FALSE
    )
  )

  expect_equal(
    fit$summary.fixed$mean,
    c(0.337, 0.470),
    tolerance = midtol
  )

  # With E specification:

  obs <- data.frame(
    y = c(10, 20, 30),
    E = c(1, 2, 3)
  )

  fit <- bru(
    comp,
    bru_obs(
      y ~ ibm_eval(
        agg,
        input = list(weights = weights, block = grp),
        state = Intercept + x
      ),
      family = "poisson",
      E = E,
      response_data = obs,
      data = pred,
      is_rowwise = FALSE
    )
  )

  expect_equal(
    fit$summary.fixed$mean,
    c(0.639, 0.237),
    tolerance = midtol
  )
})

test_that("Aggregated Gaussian observations, using mapper", {
  skip_on_cran()
  local_bru_safe_inla()

  obs <- data.frame(
    x = c(10, 20, 30),
    y = c(10, 20, 30),
    z = c(10, 20, 30)
  )
  pred <- data.frame(
    x = c(1, 2, 3, 4, 5, 6),
    y = c(1, 20, 3, 40, 5, 60),
    weights = c(1, 1, 1, 1, 1, 1),
    grp = c(1, 1, 2, 2, 2, 3)
  )

  # Aggregation by average:
  agg <- bm_aggregate(rescale = TRUE, n_block = nrow(obs))

  comp <- ~ Intercept(1) + x

  fit <- bru(
    comp,
    bru_obs(
      z ~ ibm_eval(
        agg,
        input = list(weights = weights, block = grp),
        state = Intercept + x
      ),
      family = "normal",
      response_data = obs,
      data = pred,
      control.family = list(
        hyper = list(
          prec = list(
            initial = 6,
            fixed = TRUE
          )
        )
      ),
      is_rowwise = FALSE
    )
  )

  expect_equal(
    fit$summary.fixed$mean,
    c(3.033, 4.426),
    tolerance = midtol
  )

  # With basic sf storage:

  obs_sf <- sf::st_as_sf(obs, coords = c("x", "y"))
  pred_sf <- sf::st_as_sf(pred, coords = c("x", "y"))

  comp_sf <- ~ Intercept(1) + x(sf::st_coordinates(pred_sf)[, "X"])

  fit_sf <- bru(
    comp_sf,
    bru_obs(
      z ~ ibm_eval(
        agg,
        input = list(weights = weights, block = grp),
        state = Intercept + x
      ),
      family = "normal",
      response_data = obs_sf,
      data = pred_sf,
      control.family = list(
        hyper = list(prec = list(initial = 6, fixed = TRUE))
      ),
      is_rowwise = FALSE
    )
  )

  expect_equal(
    fit_sf$summary.fixed$mean,
    c(3.033, 4.426),
    tolerance = midtol
  )
})

test_that("Component construction: linear model", {
  skip_on_cran()
  local_bru_safe_inla()

  df <- data.frame(x = 1:10, W = rep(1, 10), response = 1:10)

  llik <- bru_obs_list(list(bru_obs(formula = response ~ ., data = df)))

  # Using label as input:
  cmp0 <- bru_comp_list(
    list(bru_comp("x", x)),
    lhoods = llik
  )[["x"]]

  expect_equal(cmp0$label, "x")
  expect_equal(cmp0$main$model, "linear")
  expect_equal(rlang::as_label(ibm_input_get(cmp0$main$mapper)$input), "x")

  # Using label as input:
  cmp0 <- bru_comp_list(
    ~x,
    lhoods = llik
  )[["x"]]

  expect_equal(cmp0$label, "x")
  expect_equal(cmp0$main$model, "linear")
  expect_equal(rlang::as_label(ibm_input_get(cmp0$main$mapper)$input), "x")

  cmp <- bru_comp_list(
    ~ beta(main = x, weights = W, model = "linear", values = 1),
    lhoods = llik
  )[["beta"]]

  expect_equal(cmp$label, "beta")
  expect_equal(cmp$main$model, "linear")
  expect_equal(rlang::as_label(ibm_input_get(cmp$main$mapper)$input), "x")

  # Covariate mapping
  df <- data.frame(x = 1:10, W = 1)
  inp <- bru_input(cmp, data = df)
  expect_equal(
    inp,
    list(
      core = list(
        main = 1:10 # ,
        #        group = 1,
        #        replicate = 1
      ),
      scale = rep_len(1, 10)
    )
  )

  idx <- bru_index(cmp, inla_f = FALSE)
  expect_type(idx, "list")
  expect_equal(names(idx)[1], "beta")
  #  expect_equal(names(idx)[2], "beta.group")
  #  expect_equal(names(idx)[3], "beta.repl")
  expect_equal(idx$beta, 1)

  # A-matrix
  comp_lin <- ibm_linear(cmp, input = inp, inla_f = FALSE)
  A <- ibm_jacobian(comp_lin)
  expect_s4_class(A, "dgCMatrix")
  expect_equal(nrow(A), 10)
  expect_equal(ncol(A), 1)
  expect_equal(as.vector(A), 1:10)

  # Value
  v <- evaluate_effect_single_state(comp_lin, state = 2)
  expect_equal(v, 2 * df$x, ignore_attr = TRUE)

  v <- evaluate_effect_single_state(comp_lin, state = 2, A = A)
  expect_equal(v, 2 * df$x, ignore_attr = TRUE)

  cmps <- bru_comp_list(list(cmp))
  inps <- bru_input(cmps, data = df)
  v <- evaluate_effect_single_state(
    cmps,
    input = inps,
    state = list(beta = 2)
  )
  expect_equal(v[["beta"]], 2 * df$x, ignore_attr = TRUE)
})


test_that("Component construction: duplicate detection", {
  skip_on_cran()
  local_bru_safe_inla()
  expect_error(
    bru_comp_list(
      ~ -1 +
        beta(main = x, model = "linear", values = 1) +
        beta(main = x, model = "linear", values = 2)
    ),
    regexp = "Duplicated component labels detected: 'beta'"
  )
})


test_that("Component construction: offset", {
  skip_on_cran()
  local_bru_safe_inla()
  cmp <- bru_comp_list(~ -1 + something(a, model = "offset"))
  inp <- bru_input(cmp, data = data.frame(a = 11:15))
  val <- evaluate_effect_single_state(cmp,
    input = inp,
    state = NULL
  )
  expect_equal(
    val$something,
    11:15,
    ignore_attr = TRUE
  )
})


test_that("Component construction: terra", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not_installed("sf")
  skip_if_not(bru_safe_terra())

  f <- system.file("ex", "elev.tif", package = "terra")
  r <- terra::rast(f)

  data <- sf::st_sf(
    data.frame(
      geometry = sf::st_sfc(
        sf::st_point(cbind(6, 50)),
        crs = "epsg:4326"
      ),
      response = 1
    )
  )

  expect_equal(eval_spatial(r, data), 406)

  llik <- bru_obs_list(list(bru_obs(formula = response ~ ., data = data)))

  skip_on_cran()
  local_bru_safe_inla()
  cmp <- bru_comp_list(
    ~ -1 + something(eval_spatial(r, geometry), model = "linear"),
    lhoods = llik
  )
  inp <- bru_input(cmp, data = data)
  comp_lin <- ibm_linear(cmp,
    input = inp,
    state = list(something = 2),
    inla_f = FALSE
  )
  val <- evaluate_effect_single_state(
    comp_lin,
    state = list(something = 2)
  )
  expect_equal(
    val$something,
    2 * 406
  )

  cmp <- bru_comp_list(~ -1 + something(r, model = "linear"),
    lhoods = llik
  )
  inp <- bru_input(cmp, data = data)
  comp_lin <- ibm_linear(cmp,
    input = inp,
    state = list(something = 2),
    inla_f = FALSE
  )
  val <- evaluate_effect_single_state(
    comp_lin,
    state = list(something = 2)
  )
  expect_equal(
    val$something,
    2 * 406
  )

  cmp <- bru_comp_list(
    ~ -1 + something(r, model = "linear", main_layer = 1),
    lhoods = llik
  )
  inp <- bru_input(cmp, data = data)
  comp_lin <- ibm_linear(cmp,
    input = inp,
    state = list(something = 2),
    inla_f = FALSE
  )
  val <- evaluate_effect_single_state(
    comp_lin,
    state = list(something = 2)
  )
  expect_equal(
    val$something,
    2 * 406
  )

  cmp <- bru_comp_list(
    ~ -1 + something(r, model = "linear", main_layer = "elevation"),
    lhoods = llik
  )
  inp <- bru_input(cmp, data = data)
  comp_lin <- ibm_linear(cmp,
    input = inp,
    state = list(something = 2),
    inla_f = FALSE
  )
  val <- evaluate_effect_single_state(
    comp_lin,
    state = list(something = 2)
  )
  expect_equal(
    val$something,
    2 * 406
  )

  expect_error(
    bru_comp_list(~ something(r, model = "linear", main_layer = 2),
      lhoods = llik
    ),
    NULL
  )

  expect_error(
    bru_comp_list(~ something(r, model = "linear", main_layer = "elev"),
      lhoods = llik
    ),
    NULL
  )
})


test_that("Component construction: default index/mesh/mapping construction", {
  skip_on_cran()
  local_bru_safe_inla()

  lik <- bru_obs(
    "gaussian",
    formula = y ~ .,
    data = data.frame(x = c(1, 1.5, 2, NA, 4), y = 11:15),
    used = bru_used(effect = "effect")
  )

  cmp1 <- bru_comp_list(~ effect(c(1, 1.5, 2, NA, 4), model = "iid") - 1)
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(unique(lik$data$x), na.last = NA)
  )
  expect_equal(
    ibm_eval(cmp2$effect$mapper,
      input = list(core = list(main = c(1, NA, 4))),
      state = c(11, 12, 13, 14)
    ),
    c(11, 0, 14)
  )

  cmp1 <- bru_comp_list(~ effect(x, model = "rw2") - 1)
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(unique(lik$data$x), na.last = NA)
  )

  mesh1 <- fm_mesh_1d(
    sort(unique(lik$data$x), na.last = NA)
  )
  expect_error(
    bru_comp_list(
      ~ effect(x, model = "rw2", mapper = mesh1) - 1
    ),
    regexp = "Unknown mapper"
  )

  cmp1 <- bru_comp_list(
    ~ effect(x,
      model = "rw2",
      mapper = bru_mapper(mesh1, indexed = FALSE)
    ) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(unique(lik$data$x), na.last = NA)
  )

  cmp1 <- bru_comp_list(
    ~ effect(x,
      model = "rw2",
      mapper = bru_mapper(mesh1, indexed = TRUE)
    ) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    seq_along(sort(unique(lik$data$x), na.last = NA))
  )
})


test_that("Component construction: main iid factor construction", {
  skip_on_cran()
  local_bru_safe_inla()

  lik <- bru_obs("gaussian",
    formula = y ~ .,
    data = data.frame(x = as.factor(c(1, 1.5, 2, 3, 4)), y = 11:15),
    used = bru_used(effect = "effect")
  )

  cmp1 <- bru_comp_list(~ effect(as.factor(c(1, 1.5, 2, 3, 4)),
    model = "iid"
  ) - 1)
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    as.character(lik$data$x)
  )

  expect_equal(
    ibm_eval(cmp2$effect$mapper,
      input = list(core = list(main = as.factor(c(1, NA, 4)))),
      state = c(11, 12, 13, 14, 15)
    ),
    c(11, 0, 15)
  )
})

test_that("Component construction: group iid factor construction", {
  skip_on_cran()
  local_bru_safe_inla()

  lik <- bru_obs("gaussian",
    formula = y ~ .,
    data = data.frame(x = as.factor(c(1, 1.5, 2, 3, 4)), y = 11:15),
    used = bru_used(effect = "effect")
  )

  cmp1 <- bru_comp_list(
    ~ -1 +
      effect(rep(1, 5),
        group = x,
        model = "iid",
        control.group = list(model = "iid")
      )
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$group,
    as.numeric(lik$data$x)
  )

  expect_equal(
    ibm_eval(cmp2$effect$mapper,
      input = list(core = list(
        main = rep(1, 3),
        group = as.factor(c(1, NA, 4))
      )),
      state = c(11, 12, 13, 14, 15)
    ),
    c(11, 0, 15)
  )

  skip_on_cran()
  local_bru_safe_inla()
  expect_no_error(bru(cmp2, lik))
})

test_that("Component construction: replicate iid factor construction", {
  skip_on_cran()
  local_bru_safe_inla()

  lik <- bru_obs("gaussian",
    formula = y ~ .,
    data = data.frame(x = as.factor(c(1, 1.5, 2, 3, 4)), y = 11:15),
    used = bru_used(effect = "effect")
  )

  cmp1 <- bru_comp_list(
    ~ -1 +
      effect(rep(1, 5),
        replicate = x,
        model = "iid"
      )
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$replicate,
    as.numeric(lik$data$x)
  )

  expect_equal(
    ibm_eval(cmp2$effect$mapper,
      input = list(core = list(
        main = rep(1, 3),
        replicate = as.factor(c(1, NA, 4))
      )),
      state = c(11, 12, 13, 14, 15)
    ),
    c(11, 0, 15)
  )

  skip_on_cran()
  local_bru_safe_inla()
  expect_no_error(bru(cmp2, lik))
})


test_that("Component construction: unsafe intercepts", {
  skip_on_cran()
  local_bru_safe_inla()

  cmp <- bru_comp_list(~ something_unknown - 1)
  lik <- bru_obs(formula = response ~ ., data = data.frame(response = 1:5))
  lik <- bru_used_update(lik, labels = names(cmp))
  expect_error(
    object = {
      model <- bru_model(cmp, bru_obs_list(list(lik)))
    },
    paste0(
      "The input evaluation 'something_unknown' for 'something_unknown' ",
      "failed.\nPerhaps the data object doesn't contain the needed variables?"
    )
  )
})

test_that("Component inputs: non-numeric input detection", {
  skip_on_cran()
  local_bru_safe_inla()

  bnd <- spoly(data.frame(x = c(-15, 15, 15, -15), y = c(25, 25, 50, 50)),
    format = "sf"
  )
  mesh <- fm_mesh_2d_inla(boundary = bnd, max.edge = 100)
  matern <-
    INLA::inla.spde2.pcmatern(mesh,
      prior.sigma = c(0.1, 0.01),
      prior.range = c(1, 0.01)
    )

  df <- sf::st_as_sf(
    data.frame(x = runif(10, -10, 10), y = runif(10, 30, 45), obs = rnorm(10)),
    coords = c("x", "y")
  )

  # No model specified (argument unnamed), geometry invalid for
  # bm_linear:
  cmp <- bru_comp_list(~ Intercept(1) + field(geometry, matern))
  lk <- bru_obs(formula = obs ~ ., data = df, family = "gaussian")
  lk <- bru_used_update(lk, labels = names(cmp))
  model <- bru_model(cmp, c(lk))

  expect_error(
    ibm_eval(bm_linear(), input = df$geometry, state = 1),
    "The input to a bm_linear evaluation must be numeric or logical."
  )

  cmp <- bru_comp_list(
    ~ Intercept(1) + field(geometry, matern, model = matern)
  )
  lk <- bru_obs(formula = obs ~ ., data = df, family = "gaussian")
  lk <- bru_used_update(lk, labels = names(cmp))
  model <- bru_model(cmp, c(lk))
  input <- bru_input(as_bru_comp_list(model)$field, data = df)

  expect_error(
    ibm_eval(
      as_bru_comp_list(model)$field$mapper,
      input = input,
      state = rep(1, fm_dof(mesh))
    ),
    "The input to a bm_scale evaluation must be numeric or logical."
  )
})

test_that("Linear mapper", {
  mapper <- bm_linear()

  input <- seq_len(4)
  state <- 10
  input_length <- length(input)
  expect_equal(ibm_n(mapper), 1)
  expect_equal(ibm_n_output(mapper, input = input), input_length)
  expect_equal(ibm_values(mapper), 1)
  expect_equal(ibm_values(mapper, inla_f = TRUE), 1)
  expect_equal(
    ibm_eval(mapper, input = input, state = state),
    input * state
  )
  expect_equal(
    ibm_eval(
      ibm_linear(mapper, input = input, state = state),
      state = state
    ),
    input * state
  )
})


test_that("Index mapper", {
  values <- seq_len(4)
  state <- c(10, 20, 30, 40)
  mapper <- bm_index(max(values))
  input <- c(2, 2, 1, 3)
  val <- state[input]

  values_length <- length(values)
  input_length <- length(input)
  expect_equal(ibm_n(mapper), values_length)
  expect_equal(ibm_n_output(mapper, input = input), input_length)
  expect_equal(ibm_values(mapper), values)
  expect_equal(ibm_values(mapper, inla_f = TRUE), values)
  expect_equal(
    ibm_eval(mapper, input = input, state = state),
    val
  )
  expect_equal(
    ibm_eval(
      ibm_linear(mapper, input = input, state = state),
      state = state
    ),
    val
  )
})


test_that("Factor mapper", {
  all_values <-
    list(
      full = c("a", "b", "c"),
      contrast = c("b", "c")
    )
  all_state <-
    list(
      full = c(a = 10, b = 30, c = 60),
      contrast = c(b = 20, c = 50)
    )
  for (factor_mapping in rev(c("full", "contrast"))) {
    values <- all_values[[factor_mapping]]
    mapper <- bm_factor(all_values[["full"]],
      factor_mapping = factor_mapping
    )

    input <- c("b", "b", "a", "c")
    state <- all_state[[factor_mapping]]
    val <- {
      val <- rep(0, length(input))
      ok <- input %in% names(state)
      val[ok] <- state[input[ok]]
      val
    }

    values_length <- length(values)
    input_length <- length(input)
    expect_equal(ibm_n(mapper), values_length)
    expect_equal(ibm_n_output(mapper, input = input), input_length)
    expect_equal(ibm_values(mapper), values)
    expect_equal(ibm_values(mapper, inla_f = TRUE), values)
    expect_equal(
      ibm_eval(mapper, input = input, state = state),
      val
    )
    expect_equal(
      ibm_eval(
        ibm_linear(mapper, input = input, state = state),
        state = state
      ),
      val
    )
  }
})


test_that("Automated factor releveling", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not_installed("sf")

  pts <- sf::st_as_sf(
    data.frame(
      x = runif(20),
      y = runif(20),
      category = factor(
        sample(c("Alpha", "Beta", "Gamma"), 20, TRUE),
        levels = c("Alpha", "Beta", "Gamma")
      )
    ),
    coords = c("x", "y"),
    crs = 5070
  )

  boundary <- sf::st_sf(
    geometry =
      sf::st_sfc(
        sf::st_polygon(list(rbind(
          c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0)
        ))),
        crs = 5070
      )
  )

  mesh <- fmesher::fm_mesh_2d(
    boundary = boundary,
    max.edge = 0.5,
    cutoff = 0.1,
    crs = sf::st_crs(boundary)
  )

  ips <- fmesher::fm_int(list(geometry = mesh), boundary)

  ips$category <- factor(
    sample(c("Alpha", "Beta", "Gamma"), nrow(ips), TRUE),
    levels = c("Alpha", "Beta", "Gamma")
  )

  cmp <- geometry ~ Intercept(1) +
    category(category, model = "factor_contrast")

  # Relevel pts only, which is ok since dplyr::bind_rows() relevels the ips
  # to the levels of pts.
  pts$category <- relevel(pts$category, ref = "Beta")

  fit <- bru(
    cmp,
    bru_obs(
      geometry ~ .,
      data = pts,
      family = "cp",
      ips = ips
    ),
    options = list(bru_run = FALSE)
  )

  lev <- as_bru_comp_list(fit)$category$mapper$mappers$core$mappers$main$levels
  expect_equal(lev, c("Beta", "Alpha", "Gamma"))

  # Also relevel ips:
  ips$category <- relevel(ips$category, ref = "Beta")

  fit <- bru(
    cmp,
    bru_obs(
      geometry ~ .,
      data = pts,
      family = "cp",
      ips = ips
    ),
    options = list(bru_run = FALSE)
  )

  lev <- as_bru_comp_list(fit)$category$mapper$mappers$core$mappers$main$levels
  expect_equal(lev, c("Beta", "Alpha", "Gamma"))

  # Relevel again, to check inconsistent levels are noticed:
  ips$category <- relevel(ips$category, ref = "Gamma")

  pts$z <- rnorm(nrow(pts))
  ips$z <- rnorm(nrow(ips))

  expect_error(
    bru(
      cmp,
      bru_obs(
        z ~ .,
        data = pts
      ),
      bru_obs(
        z ~ .,
        data = ips
      ),
      options = list(bru_run = FALSE)
    ),
    "Inconsistent factor levels. Unable to infer mapper information."
  )
})


test_that("Pipe mapper", {
  mapper <-
    bm_pipe(
      list(
        bm_multi(list(
          space = bm_index(4),
          time = bm_index(3)
        )),
        bm_scale()
      )
    )

  expect_equal(ibm_n(mapper), 12)
  expect_equal(
    ibm_n(mapper, multi = TRUE),
    list(space = 4, time = 3)
  )
  expect_equal(ibm_values(mapper), seq_len(12))
  expect_equal(
    as.data.frame(ibm_values(mapper, multi = TRUE)),
    expand.grid(space = seq_len(4), time = seq_len(3)),
    ignore_attr = TRUE
  )

  olist_data <-
    list(
      list(space = 2:4, time = 1:3),
      2
    )
  A <- Matrix::sparseMatrix(
    i = 1:3,
    j = c(2, 7, 12),
    x = 1 * 2,
    dims = c(3, 12)
  )
  expect_equal(ibm_jacobian(mapper, olist_data), A)
})


test_that("Aggregate mapper", {
  mapper <- bm_aggregate(rescale = FALSE)

  expect_equal(ibm_n(mapper), NA_integer_)
  expect_equal(
    ibm_n(mapper, state = c(1, 1)),
    2
  )
  expect_equal(
    ibm_n_output(mapper, input = list(block = c(1, 2, 2))),
    2
  )
  expect_equal(ibm_values(mapper, state = c(1, 1)), seq_len(2))

  input <-
    list(block = c(1, 2, 2, 1, 3), weights = c(1, 1, 1, 2, 3))
  state <- c(1, 2, 3, 4, 5)

  val <- c(9, 5, 15)
  A <- Matrix::sparseMatrix(
    i = c(1, 1, 2, 2, 3),
    j = c(1, 4, 2, 3, 5),
    x = c(1, 2, 1, 1, 3),
    dims = c(3, 5)
  )
  expect_equal(ibm_eval(mapper, input = input, state = state), val)
  expect_equal(ibm_jacobian(mapper, input = input, state = state), A)


  mapper <- bm_aggregate(rescale = TRUE)

  expect_equal(ibm_n(mapper), NA_integer_)
  expect_equal(
    ibm_n(mapper, state = c(1, 1)),
    2
  )
  expect_equal(
    ibm_n_output(mapper, input = list(block = c(1, 2, 2))),
    2
  )
  expect_equal(ibm_values(mapper, state = c(1, 1)), seq_len(2))

  val <- c(3, 2.5, 5)
  A <- Matrix::sparseMatrix(
    i = c(1, 1, 2, 2, 3),
    j = c(1, 4, 2, 3, 5),
    x = c(1 / 3, 2 / 3, 1 / 2, 1 / 2, 3 / 3),
    dims = c(3, 5)
  )
  expect_equal(ibm_eval(mapper, input = input, state = state), val)
  expect_equal(ibm_jacobian(mapper, input = input, state = state), A)

  withr::local_seed(123L)
  delta <- (runif(length(state)) * 2 - 1) * 1e-6
  num_deriv <-
    (ibm_eval(mapper, input = input, state = state + delta) -
      ibm_eval(mapper, input = input, state = state - delta)) /
      (2 * sum(delta^2)^0.5)
  jac_deriv <-
    as.vector(ibm_jacobian(mapper, input = input, state = state) %*%
      (delta / sum(delta^2)^0.5))
  expect_equal(
    num_deriv,
    jac_deriv,
    tolerance = lowtol
  )
})


test_that("logsumexp mapper", {
  mapper <- bm_logsumexp(rescale = FALSE)

  expect_equal(ibm_n(mapper), NA_integer_)
  expect_equal(
    ibm_n(mapper, state = c(1, 1)),
    2
  )
  expect_equal(
    ibm_n_output(mapper, input = list(block = c(1, 2, 2))),
    2
  )
  expect_equal(ibm_values(mapper, state = c(1, 1)), seq_len(2))

  state <- c(1, 2, 3, 4, 5)

  # Test default input handling
  # Should treat NULL group and weights as all-1, but input needs to be supplied
  expect_equal(
    ibm_eval(mapper, input = NULL, state = state),
    5.451914,
    tolerance = lowtol
  )
  expect_equal(
    ibm_eval(mapper, input = NULL, state = state),
    5.451914,
    tolerance = lowtol
  )


  input <-
    list(block = c(1, 2, 2, 1, 3), weights = c(1, 1, 1, 2, 3))

  val <- c(4.717736, 3.313262, 6.098612)
  A <- Matrix::sparseMatrix(
    i = c(1, 1, 2, 2, 3),
    j = c(1, 4, 2, 3, 5),
    x = c(0.0242889, 0.9757111, 0.2689414, 0.7310586, 1),
    dims = c(3, 5)
  )
  expect_equal(
    ibm_eval(mapper, input = input, state = state),
    val,
    tolerance = lowtol
  )
  expect_equal(
    ibm_eval(mapper, input = input, state = state, log = FALSE),
    exp(val),
    tolerance = lowtol
  )
  expect_equal(
    ibm_jacobian(mapper, input = input, state = state),
    A,
    tolerance = midtol
  )


  mapper <- bm_logsumexp(rescale = TRUE)

  expect_equal(ibm_n(mapper), NA_integer_)
  expect_equal(
    ibm_n(mapper, state = c(1, 1)),
    2
  )
  expect_equal(
    ibm_n_output(mapper, input = list(block = c(1, 2, 2))),
    2
  )
  expect_equal(ibm_values(mapper, state = c(1, 1)), seq_len(2))

  val <- c(3.619124, 2.620115, 5)
  A <- Matrix::sparseMatrix(
    i = c(1, 1, 2, 2, 3),
    j = c(1, 4, 2, 3, 5),
    x = c(0.0242889, 0.9757111, 0.2689414, 0.7310586, 1),
    dims = c(3, 5)
  )
  expect_equal(
    ibm_eval(mapper, input = input, state = state),
    val,
    tolerance = lowtol
  )
  expect_equal(
    ibm_eval(mapper, input = input, state = state, log = FALSE),
    exp(val),
    tolerance = lowtol
  )
  expect_equal(
    ibm_jacobian(mapper, input = input, state = state),
    A,
    tolerance = midtol
  )

  withr::local_seed(123L)
  delta <- (runif(length(state)) * 2 - 1) * 1e-6
  num_deriv <-
    (ibm_eval(mapper, input = input, state = state + delta) -
      ibm_eval(mapper, input = input, state = state - delta)) /
      (2 * sum(delta^2)^0.5)
  jac_deriv <-
    as.vector(ibm_jacobian(mapper, input = input, state = state) %*%
      (delta / sum(delta^2)^0.5))
  expect_equal(
    num_deriv,
    jac_deriv,
    tolerance = lowtol
  )
})


test_that("Multi-mapper bru input", {
  mapper <- bm_multi(list(
    space = bm_index(4),
    time = bm_index(3)
  ))
  expect_equal(ibm_n(mapper), 12)
  expect_equal(ibm_n(mapper, multi = 1), list(space = 4, time = 3))
  expect_equal(ibm_values(mapper), seq_len(12))
  expect_equal(
    as.data.frame(ibm_values(mapper, multi = 1)),
    expand.grid(space = seq_len(4), time = seq_len(3)),
    ignore_attr = TRUE
  )

  list_data <- list(time = 1:3, space = 2:4)
  olist_data <- list(space = 2:4, time = 1:3)
  df_data <- as.data.frame(list_data)
  matrix_data <- cbind(time = 1:3, space = 2:4)
  omatrix_data <- cbind(2:4, 1:3)
  A <- Matrix::sparseMatrix(
    i = 1:3,
    j = c(2, 7, 12),
    x = 1,
    dims = c(3, 12)
  )
  withr::local_seed(123L)
  state <- rnorm(12)
  val <- as.vector(A %*% state)
  expect_equal(ibm_eval(mapper, list_data, state = state), val)
  expect_equal(ibm_eval(mapper, olist_data, state = state), val)
  expect_equal(ibm_eval(mapper, df_data, state = state), val)
  expect_equal(ibm_eval(mapper, matrix_data, state = state), val)
  expect_equal(ibm_eval(mapper, omatrix_data, state = state), val)

  expect_equal(ibm_jacobian(mapper, list_data), A)
  expect_equal(ibm_jacobian(mapper, olist_data), A)
  expect_equal(ibm_jacobian(mapper, df_data), A)
  expect_equal(ibm_jacobian(mapper, matrix_data), A)
  expect_equal(ibm_jacobian(mapper, omatrix_data), A)
})

test_that("Multi-mapper bru input with offset", {
  mapper <- bm_multi(
    list(
      space = bm_pipe(
        list(mapper = bm_index(4), offset = bm_shift())
      ),
      time = bm_index(3)
    )
  )
  expect_equal(ibm_n(mapper), 12)
  expect_equal(ibm_n(mapper, multi = 1), list(space = 4, time = 3))
  expect_equal(ibm_values(mapper), seq_len(12))
  expect_equal(
    as.data.frame(ibm_values(mapper, multi = 1)),
    expand.grid(space = seq_len(4), time = seq_len(3)),
    ignore_attr = TRUE
  )

  list_data <- list(time = 1:3, space = list(mapper = 2:4, offset = 3:5))
  olist_data <- list(space = list(mapper = 2:4, offset = 3:5), time = 1:3)
  A <- Matrix::sparseMatrix(
    i = 1:3,
    j = c(2, 7, 12),
    x = 1,
    dims = c(3, 12)
  )
  withr::local_seed(123L)
  state <- rnorm(12)
  val <- list_data$space$offset + as.vector(A %*% state)
  expect_equal(ibm_eval(mapper, list_data, state = state), val)
  expect_equal(ibm_eval(mapper, olist_data, state = state), val)
  expect_equal(
    ibm_eval(mapper, list_data, state = NULL),
    list_data$space$offset
  )
  expect_equal(
    ibm_eval(mapper, olist_data, state = NULL),
    list_data$space$offset
  )

  expect_equal(ibm_jacobian(mapper, list_data), A)
  expect_equal(ibm_jacobian(mapper, olist_data), A)
  expect_equal(ibm_jacobian(mapper, list_data, state = NULL), A)
  expect_equal(ibm_jacobian(mapper, olist_data, state = NULL), A)
})


test_that("User defined mappers", {
  # .S3method was unavailable in R 3.6!
  skip_if_not(utils::compareVersion("4", R.Version()$major) <= 0)

  # User defined mapper objects

  ibm_jacobian.bm_test <- function(mapper, input, ...) {
    message("---- IBM_JACOBIAN from inner environment ----")
    Matrix::sparseMatrix(
      i = seq_along(input),
      j = input,
      x = rep(2, length(input)),
      dims = c(length(input), mapper$n)
    )
  }
  .S3method("ibm_jacobian", "bm_test", ibm_jacobian.bm_test)

  bm_test <- function(n, ...) {
    bru_mapper_define(
      list(n = n),
      new_class = "bm_test"
    )
  }

  m <- bm_test(n = 20)
  cmp <- y ~ -1 + indep(x, model = "iid", mapper = m)
  mydata <- data.frame(y = rnorm(15) + 2 * (1:15), x = 1:15)
  suppressMessages({
    expect_message(
      object = {
        ibm_jacobian(m, mydata$x)
      },
      "---- IBM_JACOBIAN from inner environment ----",
      label = "ibm_jacobian generic call"
    )
  })

  skip_on_cran()
  local_bru_safe_inla()

  suppressMessages({
    expect_message(
      object = {
        fit <- bru(cmp, data = mydata, family = "gaussian")
      },
      "---- IBM_JACOBIAN from inner environment ----",
      label = "Non-interactive bru() call"
    )
  })
})


test_that("Collect mapper, direct construction", {
  skip_on_cran()
  withr::local_seed(1234L)

  mapper <- bm_collect(
    list(
      u = bm_index(4),
      v = bm_index(4)
    ),
    hidden = TRUE
  )
  expect_equal(ibm_n(mapper), 8)
  expect_equal(ibm_n(mapper, inla_f = TRUE), 4)
  expect_equal(ibm_n(mapper, multi = 1), list(u = 4, v = 4))
  expect_equal(ibm_values(mapper), seq_len(8))
  expect_equal(ibm_values(mapper, inla_f = TRUE), seq_len(4))
  expect_equal(
    as.data.frame(ibm_values(mapper, multi = 1)),
    list(u = seq_len(4), v = seq_len(4)),
    ignore_attr = TRUE
  )

  list_data <- list(u = 1:3, v = 2:4)
  A <- Matrix::bdiag(
    Matrix::sparseMatrix(
      i = 1:3,
      j = 1:3,
      x = 1,
      dims = c(3, 4)
    ),
    Matrix::sparseMatrix(
      i = 1:3,
      j = 2:4,
      x = 1,
      dims = c(3, 4)
    )
  )
  A <- as(as(as(A, "dMatrix"), "generalMatrix"), "TsparseMatrix")
  expect_equal(ibm_jacobian(mapper, list_data), A)
  expect_equal(
    as(
      ibm_jacobian(mapper, list_data[["u"]], inla_f = TRUE)[
        , ibm_inla_subset(mapper),
        drop = FALSE
      ],
      "dgTMatrix"
    ),
    as(A[
      seq_along(list_data[["u"]]),
      ibm_inla_subset(mapper),
      drop = FALSE
    ], "dgTMatrix")
  )
})


test_that("Collect mapper, automatic construction", {
  skip_on_cran()
  local_bru_safe_inla()

  data <- data.frame(val = 1:3, y = 1:3)

  mapper <- bm_collect(
    list(
      u = bm_index(4),
      v = bm_index(4)
    ),
    hidden = TRUE
  )

  cmp1 <- y ~
    -1 +
    indep(val,
      model = "bym2",
      mapper = mapper,
      graph = Matrix::Diagonal(4) + 1
    )

  cmp2 <- y ~
    -1 +
    indep(val,
      model = "bym2",
      n = 4,
      graph = Matrix::Diagonal(4) + 1
    )

  cmp3 <- y ~
    -1 +
    indep(val,
      model = "bym2",
      graph = Matrix::Diagonal(4) + 1
    )

  lik <- bru_obs(formula = y ~ ., data = data)

  cmp1 <- bru_comp_list(cmp1, lhoods = bru_obs_list(list(lik)))
  cmp2 <- bru_comp_list(cmp2, lhoods = bru_obs_list(list(lik)))
  cmp3 <- bru_comp_list(cmp3, lhoods = bru_obs_list(list(lik)))

  for (inla_f in c(FALSE, TRUE)) {
    n <- ibm_n(cmp1$indep$mapper, inla_f = inla_f)
    expect_identical(
      n,
      if (inla_f) 4 else 8
    )
    expect_identical(
      ibm_n(cmp2$indep$mapper, inla_f = inla_f),
      n
    )
    expect_identical(
      ibm_n(cmp3$indep$mapper, inla_f = inla_f),
      n
    )

    values <- ibm_values(cmp1$indep$mapper, inla_f = inla_f)
    expect_identical(
      ibm_values(cmp2$indep$mapper, inla_f = inla_f),
      values
    )
    expect_identical(
      ibm_values(cmp3$indep$mapper, inla_f = inla_f),
      values
    )

    if (inla_f) {
      input <- list(core = list(
        main = data$val,
        group = rep(1, 3),
        replicate = rep(1, 3)
      ))
    } else {
      input <- list(core = list(
        main = list(u = data$val),
        group = rep(1, 3),
        replicate = rep(1, 3)
      ))
    }

    J <- as.matrix(ibm_jacobian(
      cmp1$indep$mapper,
      input = input,
      inla_f = inla_f
    ))

    expect_identical(
      as.matrix(ibm_jacobian(cmp2$indep$mapper,
        input = input,
        inla_f = inla_f
      )),
      J
    )
    expect_identical(
      as.matrix(ibm_jacobian(cmp3$indep$mapper,
        input = input,
        inla_f = inla_f
      )),
      J
    )
  }
})


test_that("Collect mapper works", {
  skip_on_cran()
  local_bru_safe_inla()

  graph <- Matrix::sparseMatrix(
    i = c(1, 2, 3, 2, 3, 4),
    j = c(2, 3, 4, 1, 2, 3),
    x = rep(1, 6),
    dims = c(4, 4)
  )

  withr::local_seed(12345L)
  data <- data.frame(
    y = 4 + rnorm(6),
    x = c(1, 2, 3, 2, 3, 4)
  )

  fit_inla <- INLA::inla(
    y ~ 1 + f(x, model = "bym", graph = graph),
    data = data
  )

  expect_no_error({
    fit_bru <-
      bru(~ Intercept(1) + field(x, model = "bym", graph = graph),
        formula = y ~ Intercept + field,
        data = data,
        options = list(bru_initial = list(field = rep(10, 8)))
      )
  })
})


test_that("Marginal mapper", {
  m1 <- bm_marginal(qexp, pexp, rate = 1 / 8)
  state0 <- -5:5
  val1 <- ibm_eval(m1, state = state0)

  state1 <- ibm_eval(m1, state = val1, reverse = TRUE)
  expect_equal(state1, state0)

  m2 <- bm_marginal(qexp, pexp, rate = 1 / 8, inverse = TRUE)
  state2 <- ibm_eval(m2, state = val1)
  expect_equal(state2, state0)

  val2 <- ibm_eval(m2, state = state2, reverse = TRUE)
  expect_equal(val2, val1)

  dqexp <- function(p,
                    rate = 1,
                    lower.tail = TRUE,
                    log.p = FALSE,
                    log = FALSE) {
    if (log.p) {
      if (lower.tail) {
        val <- log1p(-exp(p)) + log(rate)
      } else {
        val <- p + log(rate)
      }
    } else {
      if (lower.tail) {
        val <- log1p(-p) + log(rate)
      } else {
        val <- log(p) + log(rate)
      }
    }
    if (log) {
      return(val)
    }
    exp(val)
  }
  m1_d <- bm_marginal(qexp, pexp, dexp, rate = 1 / 8)
  m1_dq <- bm_marginal(qexp, pexp, NULL, dqexp, rate = 1 / 8)
  jac1 <- ibm_jacobian(m1, state = state0)
  jac1_d <- ibm_jacobian(m1_d, state = state0)
  jac1_dq <- ibm_jacobian(m1_dq, state = state0)
  # rbind(diag(jac1), diag(jac1_d), diag(jac1_dq))
  expect_equal(sum(abs(diag(jac1) - diag(jac1_d))), 0, tolerance = lowtol)
  expect_equal(sum(abs(diag(jac1_d) - diag(jac1_dq))), 0, tolerance = lowtol)
})


test_that("Mapper lists", {
  m1 <- bm_index(4L)
  m2 <- bm_index(3L)

  expect_s3_class(c(m1), "bm_list")
  expect_s3_class(c(m1, m2), "bm_list")
  expect_s3_class(c(c(m1, m2), c(m2, m1)), "bm_list")
})

test_that("Mesh 1d mapper", {
  m <- bru_mapper(fmesher::fm_mesh_1d(1:5, boundary = "f"), indexed = TRUE)

  loc <- seq(-2, 7, by = 0.5)
  val <- ibm_eval(m, input = loc, state = seq_len(ibm_n(m)))
  expect_length(val, length(loc))
  expect_equal(val, loc)

  m <- bru_mapper(
    fmesher::fm_mesh_1d(c(1, 2, 4, 6, 9), boundary = "f"),
    indexed = FALSE
  )
  expect_equal(ibm_values(m), c(1, 2, 4, 6, 9))
})

test_that("Mesh 2d mapper", {
  m <- bru_mapper(fmesher::fmexample$mesh)

  loc <- fm_pixels(fmesher::fmexample$mesh, dims = c(5, 5), mask = FALSE)
  val <- ibm_eval(m, input = loc, state = seq_len(ibm_n(m)))
  expect_length(val, 25)

  loc <- fm_pixels(fmesher::fmexample$mesh, dims = c(5, 5), mask = TRUE)
  val <- ibm_eval(m, input = loc, state = seq_len(ibm_n(m)))
  expect_length(val, 9)
})


test_that("Repeat mapper, direct construction", {
  withr::local_seed(1234L)

  mapper <- bm_repeat(
    bm_index(4),
    n_rep = 3
  )
  expect_equal(ibm_n(mapper), 12)
  expect_equal(ibm_values(mapper), seq_len(12))

  data <- 3:1
  A <- Matrix::sparseMatrix(
    i = 1:3,
    j = data,
    x = 1,
    dims = c(3, 4)
  )
  A <- cbind(A, A, A)
  A <- as(as(as(A, "dMatrix"), "generalMatrix"), "CsparseMatrix")
  expect_equal(ibm_jacobian(mapper, data), A)
})


test_that("Repeat mapper works", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(12345L)
  u <- rnorm(8)
  data <- data.frame(
    x = c(1, 2, 3, 2, 3, 4)
  )
  data$y <- 4 + u[data$x] + u[data$x + 4L]

  expect_no_error({
    fit_bru <-
      bru(
        y ~ Intercept(1) + field(
          x,
          model = "iid",
          mapper = bm_repeat(bm_index(4), n_rep = 2)
        ),
        data = data,
        control.family = list(hyper = list(prec = list(
          initial = log(1e8), fixed = TRUE
        ))),
        options = list(bru_initial = list(field = rep(10, 8)))
      )
  })
})

test_that("Reparam mapper works", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(12345L)
  u <- rnorm(8)
  data <- data.frame(
    x = rep_len(c(1, 2, 3, 2, 3, 4, 5, 6, 7, 8), 10 * 2)
  )
  data$y <- c(
    u[1L:4L] + cumsum(u[5L:8L]),
    cumsum(u[5L:8L])
  )[data$x]

  B <- rbind(
    c(1, 0, 0, 0, 1, 0, 0, 0),
    c(0, 1, 0, 0, 1, 1, 0, 0),
    c(0, 0, 1, 0, 1, 1, 1, 0),
    c(0, 0, 0, 1, 1, 1, 1, 1),
    c(0, 0, 0, 0, 1, 0, 0, 0),
    c(0, 0, 0, 0, 1, 1, 0, 0),
    c(0, 0, 0, 0, 1, 1, 1, 0),
    c(0, 0, 0, 0, 1, 1, 1, 1)
  )

  expect_no_error({
    fit_bru_pipe <-
      bru(
        ~ 0 + field(
          list(AA = B, BB = x),
          model = "iid",
          mapper = bm_pipe(list(AA = bm_matrix(ncol(B)), BB = bm_index(8))),
          constr = FALSE
        ),
        bru_obs(
          formula = y ~ .,
          data = data,
          control.family = list(hyper = list(prec = list(
            initial = log(1e8), fixed = TRUE
          )))
        ),
        options = list(bru_initial = list(field = rep(10, 8)))
      )
  })
  expect_no_error({
    fit_bru_reparam <-
      bru(
        ~ 0 + field(
          x,
          model = "iid",
          mapper = bm_reparam(bm_index(8), B = B)
        ),
        bru_obs(
          formula = y ~ .,
          data = data,
          control.family = list(hyper = list(prec = list(
            initial = log(1e8), fixed = TRUE
          )))
        ),
        options = list(bru_initial = list(field = rep(10, 8)))
      )
  })
  expect_equal(
    fit_bru_pipe$summary.fixed,
    fit_bru_reparam$summary.fixed,
    tolerance = midtol
  )

  expect_equal(
    fit_bru_pipe$summary.random$field,
    fit_bru_reparam$summary.random$field,
    tolerance = midtol
  )

  expect_equal(
    fit_bru_pipe$summary.hyperpar,
    fit_bru_reparam$summary.hyperpar,
    tolerance = midtol
  )
})

test_that("Linearisation", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(12345L)
  data <- data.frame(x = seq_len(10) / 1)
  data <- within(data, {
    y <- exp(x / 5) - 2 + rnorm(length(x), sd = 0.1)
    z <- rpois(length(x), (exp(x / 5) + exp(2)))
  })

  cmp <- ~ -1 + x + Int_y(1) + Int_z(1)
  lhoods <-
    c(
      bru_obs(
        formula = y ~ exp(x) + Int_y_latent,
        data = data
      ),
      bru_obs(
        formula = z ~ log(exp(x) + exp(Int_z_latent)),
        data = data,
        family = "poisson"
      )
    )
  lhoods <- bru_used_update(lhoods, names(bru_comp_list(cmp)))

  used <- bru_used(lhoods[[1]])
  expect_equal(used[["effect"]], "x")
  expect_equal(used[["latent"]], "Int_y")

  used <- bru_used(lhoods[[2]])
  expect_equal(used[["effect"]], "x")
  expect_equal(used[["latent"]], "Int_z")

  used <- bru_used(lhoods)
  expect_equal(used[["effect"]], "x")
  expect_equal(used[["latent"]], c("Int_y", "Int_z"))

  model <- bru_model(bru_comp_list(cmp), lhoods)
  lhoods <- model$lhoods

  idx <- bru_index(model, used = bru_used(lhoods))
  inp <- bru_input(model, lhoods)
  comp_lin <- ibm_linear(model, input = inp, state = NULL)
  lin0 <- bru_compute_linearisation(
    model,
    lhoods = lhoods,
    input = inp,
    state = list(Int_y = 0, Int_z = 0, x = 0),
    comp_simple = comp_lin
  )
  lin <- bru_compute_linearisation(
    model,
    lhoods = lhoods,
    input = inp,
    state = list(x = 1 / 5, Int_y = -4, Int_z = log(4)),
    comp_simple = comp_lin
  )

  if (utils::packageVersion("INLA") > "24.06.02") {
    stks0 <-
      lapply(
        seq_along(lhoods),
        function(lh_idx) {
          lh <- lhoods[[lh_idx]]
          lin_off <- ibm_eval(lin0[[lh_idx]], multi = TRUE, inla_f = TRUE)
          lin_A <- ibm_jacobian(lin0[[lh_idx]], multi = TRUE, inla_f = TRUE)
          nms <- names(lin_A)
          INLA::inla.stack(
            list(
              BRU.E = lh[["response_data"]][["BRU_E"]],
              BRU.Ntrials = lh[["response_data"]][["BRU_Ntrials"]],
              BRU.weights = lh[["response_data"]][["BRU_weights"]],
              BRU.scale = lh[["response_data"]][["BRU_scale"]],
              BRU.offset = as.vector(lin_off)
            ),
            A = lapply(nms, function(nm) {
              lin_A[[nm]][, idx[["inla_subset"]][[nm]], drop = FALSE]
            }),
            effects = idx[["idx_inla"]][nms],
            responses = list(lh$response_data[[lh$response]])
          )
        }
      )
  } else {
    stks0 <-
      lapply(
        seq_along(lhoods),
        function(lh_idx) {
          lh <- lhoods[[lh_idx]]
          lin_off <- ibm_eval(lin0[[lh_idx]], multi = TRUE, inla_f = TRUE)
          lin_A <- ibm_jacobian(lin0[[lh_idx]], multi = TRUE, inla_f = TRUE)
          nms <- names(lin_A)
          INLA::inla.stack(
            list(
              BRU.response = lh$response_data[[lh$response]],
              BRU.E = lh[["response_data"]][["BRU_E"]],
              BRU.Ntrials = lh[["response_data"]][["BRU_Ntrials"]],
              BRU.weights = lh[["response_data"]][["BRU_weights"]],
              BRU.scale = lh[["response_data"]][["BRU_scale"]],
              BRU.offset = as.vector(lin_off)
            ),
            A = lapply(nms, function(nm) {
              lin_A[[nm]][, idx[["inla_subset"]][[nm]], drop = FALSE]
            }),
            effects = idx[["idx_inla"]][nms]
          )
        }
      )
  }

  stk0 <-
    do.call(
      bru_inla.stack.mjoin,
      c(stks0, list(compress = TRUE, remove.unused = FALSE))
    )

  stk0_ <- bru_make_stack(lhoods, lin0, idx)

  expect_s3_class(stk0, "inla.data.stack")

  expect_no_error(
    object = {
      fit <- bru(
        components = cmp,
        lhoods,
        options = list(
          #          bru_initial = list(
          #            x = 1 / 5, Int_y = -4, Int_z = log(4)
          #          ),
          control.inla = list(int.strategy = "eb"),
          bru_verbose = FALSE,
          bru_method = list(
            taylor = "pandemic",
            search = "all"
          )
        )
      )
    }
  )
})

test_that("Response and predictor mismatch handling", {
  skip_on_cran()
  local_bru_safe_inla()
  n <- 10
  df <- data.frame(Y = rpois(n, lambda = 3))

  cmpA <- ~ -1 + beta(1, model = "linear")
  cmpB <- ~ -1 + beta(rep(1, NROW(.data.)), model = "linear")

  lik1 <- bru_obs("poisson",
    formula = Y ~ .,
    data = df
  )

  expect_no_error({
    fit1A <- bru(components = cmpA, lik1)
  })

  expect_no_error({
    fit1B <- bru(components = cmpB, lik1)
  })

  lik2 <- bru_obs("poisson",
    formula = Y ~ beta,
    data = df
  )

  expect_no_error({
    fit2A <- bru(components = cmpA, lik2)
  })
  expect_no_error({
    fit2B <- bru(components = cmpB, lik2)
  })

  lik3 <- bru_obs("poisson",
    formula = Y ~ c(beta, beta),
    data = df
  )

  expect_error(
    {
      fit3A <- bru(components = cmpA, lik3)
    },
    paste0(
      "Number of rows \\(2\\) in the predictor for component 'beta' ",
      "does not match the length implied by the response data \\(10\\)"
    )
  )
  expect_error(
    {
      fit3B <- bru(components = cmpB, lik3)
    },
    paste0(
      "Number of rows \\(20\\) in the predictor for component 'beta' ",
      "does not match the length implied by the response data \\(10\\)"
    )
  )


  # INLA prior to mdata/surv stack support will give a different error message
  # than the one below, so we skip the test for old INLA versions.
  skip_if(utils::packageVersion("INLA") <= "24.06.26")

  expect_error(
    {
      fit4 <- bru(y ~ 0 + comp(1:3),
        data = data.frame(y = rnorm(2)),
        family = "gaussian"
      )
    },
    paste0(
      "The total number of response values \\(N=2\\) and predictor ",
      "values \\(N=3\\) do not match.\n",
      "  This is likely due to a mistake in the component or predictor ",
      "constructions."
    )
  )
})


test_that("Complex list data handling", {
  skip_on_cran()
  local_bru_safe_inla()
  withr::local_seed(12345L)
  n <- 6
  m <- 3
  data <- list(
    x1 = rnorm(m * n),
    weight = runif(m * n),
    .block = rep(1:n, m),
    x2 = rnorm(n)
  )
  agg <- bm_logsumexp()
  resp_data <- with(data, data.frame(Y = rpois(n, lambda = exp(
    1 +
      ibm_eval(
        agg,
        input = list(weight = weight, block = .block),
        state = x1
      ) + x2
  ))))
  cmpA <- ~ 0 + Intercept(1) + x1 + x2

  lik1 <- bru_obs(
    "poisson",
    formula = Y ~ Intercept + ibm_eval(
      agg,
      input = list(weight = weight, block = .block),
      state = x1
    ) + x2,
    data = data,
    is_rowwise = FALSE,
    response_data = resp_data
  )

  expect_no_error({
    fit <- bru(components = cmpA, lik1)
  })
})

test_that("bru: factor component", {
  skip_on_cran()
  local_bru_safe_inla()

  # Required for reproducible predict() and generate() output.
  withr::local_seed(1234L)

  input.df <- data.frame(x = cos(1:10), zz = rep(c(1, 10), each = 5))
  input.df <- within(input.df, {
    y <- 5 + 2 * cos(1:10) +
      rnorm(10, mean = 0, sd = 1)[zz] +
      rnorm(10, mean = 0, sd = 0.1)
  })

  # Fit a model with fixed effect 'x' and intercept 'Intercept'

  fit <- bru(y ~ x + z(zz, model = "iid", mapper = bm_index(10)),
    family = "gaussian", data = input.df
  )

  # Predict posterior statistics of 'x'

  skip_if_not_installed("sn")
  xpost <- predict(
    fit,
    newdata = NULL,
    formula = ~ x_latent + fit$summary.random$z$mean[1],
    n.samples = 5,
    seed = 12345L
  )

  xpost2 <- predict(
    fit,
    newdata = NULL,
    formula = ~ {
      tmp <- c(
        a = x_latent,
        b = exp(x_latent)
      )
      c(tmp, a_b = sum(tmp), c = as.vector(diff(tmp)))
    },
    n.samples = 5,
    seed = 12345L
  )

  xpost3 <- predict(
    fit,
    newdata = NULL,
    formula = ~ {
      tmp <- c(
        a = x_latent,
        b = exp(x_latent)
      )
      data.frame(A = 2, B = tmp, C = c(sum(tmp), as.vector(diff(tmp))))
    },
    n.samples = 5,
    seed = 12345L
  )

  # The default statistics include mean, standard deviation,
  # the 2.5% quantile, the median, the 97.5% quantile
  expect_true(is.data.frame(xpost))
  expect_equal(nrow(xpost), 1)

  expect_true(is.data.frame(xpost2))
  expect_equal(nrow(xpost2), 4)
  expect_equal(rownames(xpost2), c("a", "b", "a_b", "c"))

  xipost <- generate(fit,
    newdata = NULL,
    formula = ~ c(
      Intercept = Intercept_latent,
      x = x_latent
    ),
    n.samples = 5,
    seed = 12345L
  )

  expect_true(is.matrix(xipost))
  expect_equal(rownames(xipost), c("Intercept", "x"))


  # Evaluate effect with _eval feature

  xpost4 <- generate(
    fit,
    newdata = NULL,
    formula = ~ c(
      z_eval(c(1, 2, 11, 12, 12)),
      z_eval(c(1, 2, 11, 12, 12))
    ),
    n.samples = 5,
    seed = 12345L
  )

  # The first four rows should equal the last 5 rows.
  expect_equal(
    xpost4[1:5, , drop = FALSE],
    xpost4[6:10, , drop = FALSE]
  )
  # The index 12 values should be equal.
  expect_equal(
    xpost4[4, , drop = FALSE],
    xpost4[5, , drop = FALSE]
  )
  # The columns should all different values
  expect_equal(
    sum(xpost4[, 1] == xpost4[, 2]),
    0
  )


  xpost5 <- predict(
    fit,
    newdata = NULL,
    formula = ~ z_eval(c(1, 11)) * Precision_for_z^0.5,
    n.samples = 500,
    seed = 12345L
  )

  # sd for z(11) should be close to 1
  expect_equal(mean(xpost5[2, "sd"]),
    1,
    tolerance = hitol
  )
})


test_that("bru: predict with _eval", {
  skip_on_cran()
  local_bru_safe_inla()

  # Required for reproducible predict() and generate() output.
  withr::local_seed(1234L)

  data <- data.frame(
    z = rnorm(5),
    u = 1:5
  )
  u_mapper <- bru_mapper(
    fm_mesh_1d(
      seq(1, 5, length.out = 51)
    ),
    indexed = FALSE
  )
  fit <- bru(
    z ~ -1 + fun(u, model = "rw2", mapper = u_mapper),
    family = "gaussian",
    data = data
  )

  skip_if_not_installed("sn")
  pred <- predict(
    fit,
    newdata = data.frame(u = seq(-1, 6, by = 0.1)),
    formula = ~ data.frame(
      A = fun,
      B = fun_eval(u)
    ),
    n.samples = 10L
  )

  expect_equal(pred$B, pred$A)
})

#' Deprecated functions in inlabru
#'
#' Some of these functions may still attempt to do their job, but will be
#' removed in a future version.
#'
#' @param \dots Usually passed on to other methods
#'
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#'
#' @name inlabru-deprecated
NULL

#' @include deprecated.R
#' @include effect.R
#' @include mappers.R

# pcmatern_B ####

#' @title Make hierarchical mesh basis functions
#' @description `r lifecycle::badge("experimental")` Construct hierarchical
#'   basis functions. This is highly experimental and may change or be removed
#'   in future versions.
#' @param mesh A `fmesher` object.
#' @param forward logical; if `TRUE` (default), the basis functions are
#'   constructed in forward order; otherwise in reverse order.
#' @param method Character; one of `"laplace"` (default), `"graphdistance"` or
#'   `"graphlaplace"`. The method used to construct the basis functions. See
#'   Details.
#' @param alpha Numeric; only used if `method = "laplace"`. The approximate
#'   Laplacian power used to construct the basis functions. Default is
#'   `(1 + fm_manifold_dim(mesh)) / 2`, which gives an approximately linear
#'   function along the shortest path. Must be in the range `[1, 2]`. See
#'   Details.
#' @details The hierarchical basis functions are constructed by iteratively
#'   adding basis functions centered at the point furthest away from the
#'   previously added basis functions. The radius of each basis function is
#'   equal to the distance to the previously added basis functions at the time
#'   of addition. Three methods are available to construct the basis functions:
#'   \describe{
#'     \item{`laplace`}{The basis functions are constructed using a combination
#'       of
#'  two operators, the Laplacian and the squared Laplacian (the biharmonic
#'  operator). This approximates the Laplacian raised to the power `alpha`. The
#'  default value for `alpha` gives an approximately linear function along the
#'  shortest path.one based on the standard FEM stiffness matrix, and one based
#'  on the graph Laplacian. The `alpha` parameter controls the weight between
#'  the two operators, with `alpha = 1` using only the Laplacian and `alpha = 2`
#'  using only the squared Laplacian.}
#'    \item{`graphdistance`}{The basis functions are constructed using a simple
#'  neighbour distance-based approach, where the basis function value
#'  decreases by equal amounts for each traversed edge.}
#'    \item{`graphlaplace`}{The basis functions are constructed using the
#'  graph Laplacian only.}
#'  }
#' @returns A sparse matrix of class `dgCMatrix`, where each column
#'   corresponds to a basis function.
#' @export
#' @keywords internal
#' @rdname pcmatern_B
#' @examples
#' m <- fmesher::fmexample$mesh
#' B <- make_hierarchical_mesh_basis(m, method = "laplace2", alpha = 1.5)
#'
#' \donttest{
#' m <- fmesher::fm_subdivide(fmesher::fmexample$mesh, 1)
#' B <- list()
#' for (methodB in c(
#'   "laplace", "laplace2",
#'   "graphdistance",
#'   "graphlaplace", "graphlaplace2"
#' )) {
#'   B[[methodB]] <- make_hierarchical_mesh_basis(
#'     m,
#'     method = methodB, alpha = 1.5
#'   )
#' }
#' if (require("ggplot2") && require("patchwork")) {
#'   print(
#'     ggplot(
#'       data =
#'         data.frame(
#'           idx = seq_len(ncol(B$laplace)),
#'           nnz = as.vector(Matrix::colSums(0 != B$laplace)),
#'           laplace = as.vector(Matrix::colSums(B$laplace)),
#'           laplace2 = as.vector(Matrix::colSums(B$laplace2)),
#'           distance = as.vector(Matrix::colSums(B$graphdistance)),
#'           graph = as.vector(Matrix::colSums(B$graphlaplace)),
#'           graph2 = as.vector(Matrix::colSums(B$graphlaplace2))
#'         )
#'     ) +
#'       geom_point(aes(x = idx, y = nnz, color = "nnz")) +
#'       geom_point(aes(x = idx, y = laplace, color = "laplace")) +
#'       geom_point(aes(x = idx, y = laplace2, color = "laplace2")) +
#'       geom_point(aes(x = idx, y = distance, color = "graphdistance")) +
#'       geom_point(aes(x = idx, y = graph, color = "graphlaplace")) +
#'       scale_y_log10() +
#'       scale_x_log10()
#'   )
#'
#'   idx <- seq_len(fmesher::fm_dof(m))
#'   idx <- seq_len(10)
#'   method0 <- "laplace"
#'   theta <- qr.solve(B[[method0]][, idx, drop = FALSE], m$loc[, 1])
#'   print(
#'     ggplot() +
#'       gg(m,
#'         col = B[[method0]][, idx, drop = FALSE] %*% theta - m$loc[, 1],
#'         nx = 60, ny = 60
#'       ) +
#'       fmesher::geom_fm(
#'         data = m,
#'         color = ggplot2::alpha("black", 0.1),
#'         alpha = 0
#'       ) +
#'       scale_fill_distiller(palette = "RdBu")
#'   )
#'
#'   ev <- fmesher::fm_evaluator(m, dims = c(60, 60))
#'   df <- NULL
#'   for (method in c(
#'     "laplace", "laplace2",
#'     "graphdistance",
#'     "graphlaplace", "graphlaplace2"
#'   )) {
#'     fun_cumulative <- 1e-16
#'     for (k in 1:28) {
#'       fun_orig <- as.vector(B[[method]][, k, drop = FALSE])
#'       fun_cumulative <- fun_cumulative + fun_orig
#'       fun_norm <- fun_orig / fun_cumulative
#'       df <- rbind(
#'         df,
#'         data.frame(
#'           x = ev$lattice$loc[, 1],
#'           y = ev$lattice$loc[, 2],
#'           orig = as.vector(fmesher::fm_evaluate(ev, fun_orig)),
#'           norm = as.vector(fmesher::fm_evaluate(ev, fun_norm)),
#'           index = k,
#'           fun = paste0(
#'             "fun",
#'             formatC(k, width = 2, format = "d", flag = "0")
#'           ),
#'           method = method
#'         )
#'       )
#'     }
#'   }
#'   df$norm[df$fun == "fun01"] <- NA
#'   df$orig[df$orig == 0] <- NA
#'   df$norm[df$norm == 0] <- NA
#'   print(
#'     ggplot(data = df[df$index <= 4, ]) +
#'       geom_tile(aes(x, y, fill = orig),
#'         na.rm = TRUE
#'       ) +
#'       fmesher::geom_fm(
#'         data = m,
#'         color = ggplot2::alpha("black", 0.1),
#'         alpha = 0
#'       ) +
#'       geom_contour(aes(x, y, z = orig),
#'         breaks = seq_len(29 - 15) / (30 - 15), #* 2-0.5,
#'         na.rm = TRUE, col = "black", alpha = 0.5
#'       ) +
#'       geom_contour(aes(x, y, z = norm),
#'         breaks = seq_len(29 - 15) / (30 - 15), #* 2-0.5,
#'         na.rm = TRUE, col = "red", alpha = 0.5
#'       ) +
#'       scale_fill_distiller(palette = "RdBu", limits = c(0, 1)) + #* 2-0.5) +
#'       facet_wrap(vars(fun, method))
#'   )
#' }
#' }
#'
make_hierarchical_mesh_basis <- function(mesh, forward = TRUE, method = NULL,
                                         alpha = 1) {
  if (is.null(alpha)) {
    alpha <- (1 + fmesher::fm_manifold_dim(mesh)) / 2
  }
  stopifnot((alpha >= 1) && (alpha <= 2))
  # Construct neighbour matrix in a way that doesn't involve the mesh specifics;
  # only the computational neighbourhood structure:
  fem <- fm_fem(mesh, order = 2)
  G1 <- fem$g1
  G2 <- fem$g2
  G <- G1
  # Adjacency matrix:
  Gadj <- (G1 != 0) * 1.0
  Gadj <- Gadj - Matrix::Diagonal(nrow(G), diag(Gadj))

  # First point for each disconnected mesh component
  # Calculate graph distances
  ii <- list()
  jj <- list()
  xx <- list()
  D <- rep(Inf, nrow(G))
  while (!all(is.finite(D))) {
    set <- rep(FALSE, nrow(G))
    front <- rep(FALSE, nrow(G))
    start <- min(which(!is.finite(D)))
    front[start] <- TRUE
    max_dist <- -1
    while (any(front)) {
      max_dist <- max_dist + 1
      D[front] <- max_dist
      set <- set | front
      front <- (as.vector(Gadj %*% front) > 0.5) & !set
    }
    set[set] <- (D[set] < max_dist)
    ijx <- make_basis_fcn(
      start = start,
      inner = set,
      D = D,
      max_dist = max_dist,
      G = G,
      Gadj = Gadj,
      method = method,
      G2 = G2,
      alpha = alpha
    )
    ii[[length(ii) + 1]] <- ijx[["ii"]]
    jj[[length(jj) + 1]] <- length(jj) + ijx[["jj"]]
    xx[[length(xx) + 1]] <- ijx[["xx"]]
  }

  # Iteratively add basis functions for the point furthest away from the the
  # previous core points, i.e. where D is maximal.  The radius of each is equal
  # to the initial D-value for the new point.
  while (any(D > 0)) {
    D_local <- rep(Inf, nrow(G))
    set <- rep(FALSE, nrow(G))
    front <- rep(FALSE, nrow(G))
    start <- which.max(D) # The first maximal distance point
    front[start] <- TRUE
    max_dist <- D[start]
    for (the_dist in c(0, seq_len(max_dist))) {
      D[front] <- pmin(D[front], the_dist)
      D_local[front] <- the_dist
      set <- set | front
      front <- (as.vector(Gadj %*% front) > 0.5) & !set
    }
    set[set] <- (D_local[set] < max_dist)
    ijx <- make_basis_fcn(
      start = start,
      inner = set,
      D = D_local,
      max_dist = max_dist,
      G = G,
      Gadj = Gadj,
      method = method,
      G2 = G2,
      alpha = alpha
    )
    ii[[length(ii) + 1]] <- ijx[["ii"]]
    jj[[length(jj) + 1]] <- length(jj) + ijx[["jj"]]
    xx[[length(xx) + 1]] <- ijx[["xx"]]
  }

  if (forward) {
    B <- Matrix::sparseMatrix(
      i = unlist(ii),
      j = unlist(jj),
      x = unlist(xx),
      dims = c(nrow(G), length(ii))
    )
  } else {
    B <- Matrix::sparseMatrix(
      i = unlist(ii),
      j = length(ii) + 1 - unlist(jj),
      x = unlist(xx),
      dims = c(nrow(G), length(ii))
    )
  }
  B
}

make_basis_fcn <- function(start, inner, D, max_dist, G, Gadj, method = NULL,
                           G2 = NULL, alpha = 1) {
  method <- match.arg(method, c(
    "graphdistance",
    "laplace", "laplace2",
    "graphlaplace", "graphlaplace2"
  ))
  front <- D == max_dist
  if (is.logical(inner)) {
    inner <- which(inner)
  }
  # Make sure the start and end are not part of inner
  inner <- setdiff(inner, start)
  if (length(inner) == 0L) {
    return(list(ii = start, jj = 1L, xx = 1.0))
  }
  ii <- c(start, inner)
  jj <- rep(1, length(ii))
  xx <- switch(method,
    "graphdistance" = {
      c(1, (max_dist - D[inner]) / max_dist)
    },
    "laplace" = {
      val <- 0
      if (alpha < 2) {
        b1 <- G[inner, start, drop = FALSE]
        L1 <- G[inner, inner, drop = FALSE]
        val <- val + (2 - alpha) * c(1, -as.vector(Matrix::solve(L1, b1)))
      }
      if (alpha > 1) {
        b2 <- G2[inner, start, drop = FALSE]
        L2 <- G2[inner, inner, drop = FALSE]
        val <- val + (alpha - 1) * c(1, -as.vector(Matrix::solve(L2, b2)))
      }
      val
    },
    "laplace2" = {
      M <- G * (2 - alpha)
      M <- M + G2 * (alpha - 1)
      b <- M[inner, start, drop = FALSE]
      L <- M[inner, inner, drop = FALSE]
      val <- c(1, -as.vector(Matrix::solve(L, b)))
      val
    },
    "graphlaplace" = {
      G1 <- Matrix::Diagonal(nrow(G), Matrix::rowSums(Gadj)) - Gadj
      G2 <- G1 %*% G1
      val <- 0
      if (alpha < 2) {
        b1 <- G1[inner, start, drop = FALSE]
        L1 <- G1[inner, inner, drop = FALSE]
        val <- val + (2 - alpha) * c(1, -as.vector(Matrix::solve(L1, b1)))
      }
      if (alpha > 1) {
        b2 <- G2[inner, start, drop = FALSE]
        L2 <- G2[inner, inner, drop = FALSE]
        val <- val + (alpha - 1) * c(1, -as.vector(Matrix::solve(L2, b2)))
      }
      val
    },
    "graphlaplace2" = {
      G1 <- Matrix::Diagonal(nrow(G), Matrix::rowSums(Gadj)) - Gadj
      G2 <- G1 %*% G1
      M <- (2 - alpha) * G1 + (alpha - 1) * G2
      b <- M[inner, start, drop = FALSE]
      L <- M[inner, inner, drop = FALSE]
      val <- c(1, -as.vector(Matrix::solve(L, b)))
      val
    }
  )
  list(ii = ii, jj = jj, xx = xx)
}


#' @export
#' @keywords internal
#' @describeIn pcmatern_B Construct a pcmatern model with basis change
#' @seealso [bru_get_mapper.inla_model_reparam()]
#' `r lifecycle::badge("experimental")`
inla.spde2.pcmatern_B <- function(mesh, ..., B) {
  model <- INLA::inla.spde2.pcmatern(mesh, ...)
  model$n.spde <- ncol(B)
  model$f$n <- ncol(B)
  if (nrow(B) != ncol(B)) {
    stop("Rectangular B not supported")
  }
  # TODO: check that it's a stationary model, since non-stationary would need a
  # different precision structure (should use rgeneric or cgeneric) and
  # different B0, B1, B2 matrices
  model$param.inla$M0 <- Matrix::t(B) %*% model$param.inla$M0 %*% B
  model$param.inla$M1 <- Matrix::t(B) %*% model$param.inla$M1 %*% B
  model$param.inla$M2 <- Matrix::t(B) %*% model$param.inla$M2 %*% B
  model$B <- B
  class(model) <- c("inla_model_reparam", class(model))
  model
}

#' @describeIn bru_get_mapper Reparameterised inla model mapper, see
#'  [inla.spde2.pcmatern_B()]
#' `r lifecycle::badge("experimental")`
#' @export
bru_get_mapper.inla_model_reparam <- function(model, ...) {
  map <- NextMethod()
  bm_reparam(map, model[["B"]])
}

#' @name Poisson3_1D
#' @title 1-Dimensional NonHomogeneous Poisson example.
#' @docType data
#' @description Point data and count data, together with intensity function and
#'   expected counts for a multimodal nonhomogeneous 1-dimensional Poisson
#'   process example. Counts are given for two different gridded data interval
#'   widths.
#'
#' @aliases lambda3_1D E_nc3a E_nc3b pts3 countdata3a countdata3b
#'
#' @usage data(Poisson3_1D)
#'
#' @format The data contain the following `R` objects:
#'  \describe{
#'    \item{`lambda3_1D`}{A function defining the intensity function of a
#'    nonhomogeneous Poisson process. Note that this function is only defined on
#'    the interval (0,55).}
#'    \item{`E_nc3a`}{The expected counts of gridded data for the wider bins
#'      (10 bins).}
#'    \item{`E_nc3b`}{The expected counts of gridded data for the wider bins
#'      (20 bins).}
#'    \item{`pts3`}{The locations of the observed points (a data frame with one
#'      column, named `x`).}
#'    \item{`countdata3a`}{A data frame with three columns, containing the
#'      count data for the 10-interval case:}
#'    \item{`countdata3b`}{A data frame with three columns, containing the
#'      count data for the 20-interval case:}
#'    \describe{
#'      \item{`x`}{The grid cell midpoint.}
#'      \item{`count`}{ The number of detections in the cell.}
#'      \item{`exposure`}{ The width of the cell.}
#'    }
#'  }
#'
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE) &&
#'   require("patchwork", quietly = TRUE)) {
#'   data(Poisson3_1D)
#'   # first the plots for the 10-bin case:
#'   p1a <- ggplot(countdata3a) +
#'     geom_point(data = countdata3a, aes(x = x, y = count), col = "blue") +
#'     ylim(0, max(countdata3a$count, E_nc3a)) +
#'     geom_point(
#'       data = countdata3a, aes(x = x), y = 0, shape = "+",
#'       col = "blue", cex = 4
#'     ) +
#'     geom_point(
#'       data = data.frame(x = countdata3a$x, y = E_nc3a),
#'       aes(x = x), y = E_nc3a, shape = "_", cex = 5
#'     ) +
#'     xlab(expression(bold(s))) +
#'     ylab("count")
#'   ss <- seq(0, 55, length.out = 200)
#'   lambda <- lambda3_1D(ss)
#'   p2a <- ggplot() +
#'     geom_line(
#'       data = data.frame(x = ss, y = lambda), aes(x = x, y = y),
#'       col = "blue"
#'     ) +
#'     ylim(0, max(lambda)) +
#'     geom_point(data = pts3, aes(x = x), y = 0.2, shape = "|", cex = 4) +
#'     xlab(expression(bold(s))) +
#'     ylab(expression(lambda(bold(s))))
#'   (p1a / p2a)
#'
#'   # Then the plots for the 20-bin case:
#'   p1a <- ggplot(countdata3b) +
#'     geom_point(data = countdata3b, aes(x = x, y = count), col = "blue") +
#'     ylim(0, max(countdata3b$count, E_nc3b)) +
#'     geom_point(
#'       data = countdata3b, aes(x = x), y = 0, shape = "+",
#'       col = "blue", cex = 4
#'     ) +
#'     geom_point(
#'       data = data.frame(x = countdata3b$x, y = E_nc3b),
#'       aes(x = x), y = E_nc3b, shape = "_", cex = 5
#'     ) +
#'     xlab(expression(bold(s))) +
#'     ylab("count")
#'   ss <- seq(0, 55, length.out = 200)
#'   lambda <- lambda3_1D(ss)
#'   p2a <- ggplot() +
#'     geom_line(
#'       data = data.frame(x = ss, y = lambda), aes(x = x, y = y),
#'       col = "blue"
#'     ) +
#'     ylim(0, max(lambda)) +
#'     geom_point(data = pts3, aes(x = x), y = 0.2, shape = "|", cex = 4) +
#'     xlab(expression(bold(s))) +
#'     ylab(expression(lambda(bold(s))))
#'   (p1a / p2a)
#' }
#' }
NULL

#' @title Summarise DIC from `lgcp` estimates.
#'
#' @description
#' Calculates DIC differences and produces an ordered summary.
#'
#' @param \dots Comma-separated objects inheriting from class `inla` and
#'   obtained from a run of `INLA::inla()`, [bru()] or [lgcp()]
#' @param criterion character vector.
#' If it includes 'DIC', computes DIC differences;
#' 'WAIC' is also allowed, but note that plain WAIC values from inla are not
#' well-defined for point process models. Default 'WAIC'
#'
#' @return A data frame with each row containing the Model name,
#' DIC and Delta.DIC.
#'
#' @export
#'
#' @examples
#' \donttest{
#' if (bru_safe_inla()) {
#'   # Generate some data
#'   input.df <- data.frame(idx = 1:10) |>
#'     dplyr::mutate(
#'       x = cos(idx),
#'       y = rpois(length(x), 5 + 2 * x + rnorm(length(x), mean = 0, sd = 0.1))
#'     )
#'
#'   # Fit two models
#'   fit1 <- bru(
#'     y ~ x,
#'     family = "poisson",
#'     data = input.df,
#'     options = list(control.compute = list(dic = TRUE))
#'   )
#'   fit2 <- bru(
#'     y ~ x + rand(idx, model = "iid"),
#'     family = "poisson",
#'     data = input.df,
#'     options = list(control.compute = list(dic = TRUE))
#'   )
#'
#'   # Compare DIC
#'
#'   deltaIC(fit1, fit2)
#' }
#' }
deltaIC <- function(..., criterion = "DIC") {
  criterion <- match.arg(criterion, c("DIC", "WAIC"), several.ok = TRUE)

  if ("WAIC" %in% criterion) {
    warning(
      paste0(
        "WAIC values from INLA are not well-defined for point process models.",
        " Use with caution, or better, not at all."
      ),
      .immediately = TRUE
    )
  }

  names <- as.character(substitute(list(...)))[-1L]
  model <- eval(list(...))
  nmod <- length(model)
  dic <- waic <- rep(NA, nmod)
  for (i in 1:nmod) {
    mod <- model[[i]]
    if ("DIC" %in% criterion) {
      if (is.null(mod$dic$dic)) {
        stop("Object ", i, " does not have DIC information.")
      }
      dic[i] <- mod$dic$dic
    }
    if ("WAIC" %in% criterion) {
      if (is.null(mod$waic$waic)) {
        stop("Object ", i, " does not have WAIC information.")
      }
      waic[i] <- mod$waic$waic
    }
  }
  if ("DIC" %in% criterion) {
    ord <- order(dic)
  } else if ("WAIC" %in% criterion) {
    ord <- order(waic)
  }
  dic <- dic[ord]
  waic <- waic[ord]
  names <- names[ord]
  ddic <- dic - min(dic)
  dwaic <- waic - min(waic)
  result <- data.frame(Model = names)
  if ("DIC" %in% criterion) {
    result <- cbind(result, data.frame(DIC = dic, Delta.DIC = ddic))
  }
  if ("WAIC" %in% criterion) {
    result <- cbind(result, data.frame(WAIC = waic, Delta.WAIC = dwaic))
  }
  result
}


#' @title Information criteria for point processes
#' @description `r lifecycle::badge("experimental")`
#'   Calculate information criteria for point process models, using a fitted
#'   model and optionally new data. The interface is still under development
#'   and may change without warning in future versions.
#' @param fit A fitted `bru` model object, from [bru()] or [lgcp()].
#' @param data A data frame containing the observed points, with columns for
#'  coordinates and any covariates used in the model. The data frame may also
#'  include a column named `.block` that indicates the block membership of each
#'  point, if blockwise criteria are to be calculated.
#' @param predictor An expression that evaluates to the linear predictor of the
#'  model, given the data. This should be an expression that can be evaluated
#'  in the context of the data frame and model components, just as in the
#'  `formula` argument of [bru_obs()] and [lgcp()].
#' @param domain,samplers Arguments defining the integration domain(s) and
#'   sampling regions, as for [bru_obs()] and [lgcp()].
#' @param n.samples Integer. The number of posterior samples to draw for
#'   estimating the criteria. Default is 2000.
#' @returns A `tibble` with columns for the `Criterion` name, `Method`,
#' log pointwise predictive density (`lppd`), effective number of parameters
#' (`p_eff`), and the information criterion value (`IC`)
#' @export
#' @keywords internal
# lgcp_IC(fit_veg,nests,vegetation,list(geometry=mesh_int),boundary)
# lgcp_IC(fit_veg,nests,vegetation,list(geometry=mesh_int),cvpart)
lgcp_IC <- function(fit, data, predictor, domain, samplers, n.samples = 2000L) {
  int_points <- fm_int(domain = domain, samplers = samplers)
  agg <- bm_aggregate(type = "sum", n_block = 1L)
  agg_block <- bm_aggregate(type = "sum", n_block = max(int_points$.block))
  if (is.null(data[[".block"]]) || (max(int_points$.block) == 1L)) {
    # TODO: implement automated linking of points to integration blocks.
    data$.block <- 1L
  }
  pred_quo <- rlang::enquo(predictor)
  pred_text <- as.character(as.formula(deparse1(pred_quo)))
  pred_text <- pred_text[length(pred_text)]
  form <- as.formula(
    glue::glue(
      '~ {{
      eta <- {{
               {pred_text}
             }}
      log_lambda_y <- eta[part == "Y"]
      lambda_y <- exp(log_lambda_y)
      # Full domain
      lambda_int <- sum(weight[part == "int"] * exp(eta[part == "int"]))
      log_lambda_int <- log(lambda_int)
      area <- sum(weight[part == "int"])
      log_like <- sum(log_lambda_y) + area - lambda_int
      # Blockwise
      lambda_block_int <-
        ibm_eval(agg_block,
                 input = list(
                   block = .block[part == "int"],
                   weights = weight[part == "int"]
                 ),
                 state = exp(eta[part == "int"])
                )
      block_area <-
        ibm_eval(agg_block,
                 input = list(
                   block = .block[part == "int"],
                   weights = weight[part == "int"]
                 ),
                 state = rep(1, sum(part == "int"))
                )
      block_log_lambda_y <-
        ibm_eval(agg_block,
                 input = list(block = .block[part == "Y"]),
                 state = log(lambda_block_int / block_area)[.block[part == "Y"]]
                )
      block_n <-
        ibm_eval(agg_block,
                 input = list(block = .block[part == "Y"]),
                 state = rep(1, sum(part == "Y"))
                )

      block_log_like <- block_area - lambda_block_int + block_log_lambda_y
      block_like <-
        exp(block_log_like - block_area +
              block_n * log(block_area) - lgamma(block_n + 1L))
      list(
        lambda_y = lambda_y,
        log_lambda_y = log_lambda_y,
        lambda_int = lambda_int,
        log_lambda_int = log_lambda_int,
        area = area,
        log_like = log_like,
        block_log_like = block_log_like,
        block_like = block_like
      )
    }}'
    )
  )
  newdata <- cbind(
    dplyr::bind_rows(data, int_points),
    part = c(
      rep("Y", nrow(data)),
      rep("int", nrow(int_points))
    )
  )
  pred <- predict(fit, newdata = newdata, formula = form, n.samples = n.samples)
  # Limit:
  lppd_limit <- sum(log(pred$lambda_y$mean)) +
    pred$area$mean -
    pred$lambda_int$mean
  p_WAIC_limit <- sum(pred$log_lambda_y$sd^2)
  p_DIC <- 2 * pred$log_like$sd^2
  WAIC_limit <- -2 * (lppd_limit - p_WAIC_limit)
  DIC <- -2 * (lppd_limit - p_DIC)

  lppd_limit.se <-
    sqrt(
      sum(pred$lambda_y$mean.mc_std_err^2 / pred$lambda_y$mean^2) +
        pred$lambda_int$mean.mc_std_err^2
    )
  p_DIC.se <- sqrt(4 * (2 * pred$log_like$sd)^2 * pred$log_like$sd.mc_std_err^2)
  p_WAIC_limit.se <- sqrt(sum(
    (2 * pred$log_lambda_y$sd)^2 *
      pred$log_lambda_y$sd.mc_std_err^2
  ))

  # True Blockwise counts:
  block_area <- ibm_eval(
    agg_block,
    input = list(
      block = newdata$.block[newdata$part == "int"],
      weights = newdata$weight[newdata$part == "int"]
    ),
    state = rep(1, sum(newdata$part == "int"))
  )
  block_n <-
    ibm_eval(
      agg_block,
      input = list(block = newdata$.block[newdata$part == "Y"]),
      state = rep(1, sum(newdata$part == "Y"))
    )
  adj <- newdata$.block[newdata$part == "Y"]

  ok <- block_area > 0
  lppd_blockwise <- sum((log(pred$block_like$mean) +
    block_area -
    block_n * log(block_area) +
    lgamma(block_n + 1L))[ok])
  p_WAIC_blockwise <- sum(pred$block_log_like$sd[ok]^2)
  WAIC_blockwise <- -2 * (lppd_blockwise - p_WAIC_blockwise)

  lppd_blockwise.se <-
    sqrt(
      sum(pred$block_like$mean.mc_std_err[ok]^2 / pred$block_like$mean[ok]^2)
    )
  p_WAIC_blockwise.se <- sqrt(sum(
    (2 * pred$block_log_like$sd[ok])^2 *
      pred$block_log_like$sd.mc_std_err[ok]^2
  ))

  # INLA ICs:
  lppd_INLA_DIC <- fit$dic$dic / (-2) + fit$dic$p.eff
  lppd_INLA_WAIC <- fit$waic$waic / (-2) + fit$waic$p.eff
  lppd_INLA_DIC_adjusted <- lppd_INLA_DIC + sum(int_points$weight)
  lppd_INLA_WAIC_adjusted <- lppd_INLA_WAIC + sum(int_points$weight)

  # fmt: skip
  df <- tibble::tribble(
    ~Criterion, ~Method, ~lppd, ~p_eff, ~IC,
    ~lppd.std_err, ~p_eff.std_err, ~IC.std_err,
    "DIC", "Limit", lppd_limit, p_DIC, DIC, lppd_limit.se, p_DIC.se, NA,
    "DIC", "INLA", lppd_INLA_DIC, fit$dic$p.eff, fit$dic$dic, NA, NA, NA,
    "DIC_adjusted", "INLA", lppd_INLA_DIC_adjusted, fit$dic$p.eff,
    -2 * (lppd_INLA_DIC_adjusted - fit$dic$p.eff), NA, NA, NA,
    "WAIC", "Limit", lppd_limit, p_WAIC_limit, WAIC_limit,
    lppd_limit.se, p_WAIC_limit.se, NA,
    "WAIC", "Blockwise", lppd_blockwise, p_WAIC_blockwise, WAIC_blockwise,
    lppd_blockwise.se, p_WAIC_blockwise.se, NA,
    "WAIC", "INLA", lppd_INLA_WAIC, fit$waic$p.eff, fit$waic$waic, NA, NA, NA,
    "WAIC_adjusted", "INLA", lppd_INLA_WAIC_adjusted, fit$waic$p.eff,
    -2 * (lppd_INLA_WAIC_adjusted - fit$waic$p.eff), NA, NA, NA
  )
  df$IC.std_err <- 2 * (df$lppd.std_err + df$p_eff.std_err)
  df
}

#' @export
#' @describeIn lgcp_IC Calculate IC differences based on joined output from
#' several `lgcp_IC` calls;
#' ```
#' delta_IC(rbind(cbind(lgcp_IC(...), Model = "A"),
#'                cbind(lgcp_IC(...), Model = "B")))
#' ```
delta_IC <- function(data) {
  data <- data |>
    dplyr::group_by(.data$Criterion, .data$Method) |>
    dplyr::mutate(Delta_IC = .data$IC - min(.data$IC)) |>
    dplyr::ungroup() |>
    dplyr::arrange(.data$Criterion, .data$Method, .data$IC)
  data
}

#' @name mrsea
#' @title Marine renewables strategic environmental assessment
#' @docType data
#' @description Data imported from package MRSea, see
#'   <https://www.creem.st-andrews.ac.uk/software/>
#'
#' @format A list of objects:
#'  \describe{
#'    \item{`points`}{ A `sf` object containing the locations of
#'    XXXXX.}
#'    \item{`samplers`}{ A `sf` object containing the transect lines
#'    that were surveyed.}
#'    \item{`mesh`}{ An `fm_mesh_2d` object containing a Delaunay triangulation
#'    mesh (a type of discretization of continuous space) covering the survey
#'      region.}
#'    \item{`boundary`}{ An `sf` object defining the boundary polygon of the
#'    survey region.}
#'    \item{`covar`}{ An `sf` containing sea depth estimates.}
#'  }
#' @source
#' Library `MRSea`.
#'
#' @references
#'
#' NONE YET
#'
#' @examples
#' if (require(ggplot2, quietly = TRUE)) {
#'   ggplot() +
#'     fmesher::geom_fm(data = mrsea$mesh) +
#'     gg(mrsea$samplers) +
#'     gg(mrsea$points) +
#'     gg(mrsea$boundary)
#' }
"mrsea"

#' @title Matern correlation or covariance function approximate credible bands.
#'
#' @description Evaluate the covariance function for an inla.spde
#'     object. Calculates the posterior distribution of the range,
#'     log(range), variance, or log(variance) parameter of a model's
#'     SPDE component. Can also calculate Matern correlation or covariance
#'     function.
#'
#' @keywords internal
#'
#' @param manifold Either "R1", "S1", "R2", or "S2", from
#'     `fm_manifold(mesh)`, or an object supported by
#'     [fmesher::fm_manifold()].
#' @param dist A vector of distances at which to calculate the
#'     covariances/correlations
#' @param log.range A scalar or a list (mean, sd), such as produced by
#'     `inla.spde.result(...)$summary.log.range.nominal[[1]][c("mean","sd")]`
#' @param log.variance Either `NULL`, a scalar, or vector of the
#'     same type as for log.range. When `NULL`, the correlations
#'     are calculated instead of the covariances.
#' @param alpha The SPDE operator order. Default 2.
#' @param quantile The target credible probability. Default 0.95.
#' @param n The number of parameter combinations to use for the
#'     approximation. Default 64.
#' @param S1.L For `manifold` `"S1"`, give the length of the
#'     cyclic interval
#' @return A list with estimated covariance or correlation (when `log.variance`
#'   is `NULL`) functions:
#' \item{lower}{An approximate lower bound for the `quantile` credible region}
#' \item{median}{The function for for the approximate median parameters
#' quantile}
#' \item{upper}{An approximate upper bound for the `quantile` credible region}
#'
#' @details
#' Uses a Gaussian assumption for the internal model parameters, and finds a
#' region in parameter space with approximately `quantile` probability.
#'
#' @author Finn Lindgren \email{Finn.Lindgren@@ed.ac.uk}
#' @export
materncov.bands <- function(manifold, dist, log.range,
                            log.variance = NULL, alpha = 2,
                            quantile = 0.95, n = 64, S1.L = NULL) {
  stopifnot(bru_safe_inla(multicore = TRUE))
  calc.cov.R <- function(dist, kappa, var) {
    INLA::inla.matern.cov(nu = nu, kappa, x = dist, d = d, corr = TRUE) * var
  }
  ## Do the right thing for _nominal_ variance
  calc.cov.S1 <- function(dist, kappa, var) {
    dist <- dist %% S1.L
    out <- calc.cov.R(abs(dist), kappa, var) + calc.cov.R(
      abs(dist - S1.L),
      kappa,
      var
    )
    # Measure the relative contribution of each term:
    cov0 <- calc.cov.R(0, kappa, 1)
    loop <- 0
    max.loop <- 10
    while ((cov0 > 1e-3) && (loop < max.loop)) {
      loop <- loop + 1
      out <- out + calc.cov.R(abs(dist + S1.L * loop), kappa, var) +
        calc.cov.R(abs(dist - S1.L - S1.L * loop), kappa, var)
      cov0 <- calc.cov.R(S1.L * loop, kappa, 1)
    }
    out
  }
  ## Do the right thing for _nominal_ variance
  calc.cov.S2 <- function(dist, kappa, var) {
    INLA::inla.matern.cov.s2(nu = nu, kappa, x = dist, norm.corr = FALSE) /
      INLA::inla.matern.cov(nu = nu, kappa, x = 0, d = 2, corr = FALSE) * var
  }
  calc.corr.R <- function(dist, kappa) {
    INLA::inla.matern.cov(nu = nu, kappa, x = dist, d = d, corr = TRUE)
  }
  calc.corr.S1 <- function(dist, kappa) {
    stop("Ooops, somehow inla.matern.cov.s1() is not available.")
    # INLA::inla.matern.cov.s1(nu=nu, kappa, x=dist, norm.corr=FALSE) /
    #   INLA::inla.matern.cov.s1(nu=nu, kappa, x=0, norm.corr=FALSE)
  }
  calc.corr.S2 <- function(dist, kappa) {
    INLA::inla.matern.cov.s2(nu = nu, kappa, x = dist, norm.corr = FALSE) /
      INLA::inla.matern.cov.s2(nu = nu, kappa, x = 0, norm.corr = FALSE)
  }
  if (!is.character(manifold)) {
    if (fm_manifold(manifold, "S1") && is.null(S1.L)) {
      S1.L <- diff(manifold$interval)
    }
    manifold <- fm_manifold(manifold)
  }
  if (fm_manifold(manifold, "R")) {
    d <- fm_manifold_dim(manifold)
    calc.corr <- calc.corr.R
    calc.cov <- calc.cov.R
  } else if (manifold == "S1") {
    if (is.null(S1.L)) {
      stop(paste0("Manifold type 'S1' requires an interval length, 'S1.L'"))
    }
    d <- 1
    calc.corr <- calc.corr.S1
    calc.cov <- calc.cov.S1
  } else if (manifold == "S2") {
    d <- 2
    calc.corr <- calc.corr.S2
    calc.cov <- calc.cov.S2
  } else {
    stop(paste0(
      "materncov.bands does not support mesh manifold type ",
      manifold
    ))
  }
  nu <- alpha - d / 2
  if (!is.list(log.range)) {
    log.range <- list(mean = log.range, sd = 0)
  }
  probs <- c((1 - quantile) / 2, 0.5, (1 + quantile) / 2)
  if (is.null(log.variance)) {
    qq <- qnorm(probs)
    kappas <- sqrt(8 * nu) / exp(log.range$mean + log.range$sd * rev(qq))
    out <- data.frame(
      lower = calc.corr(dist, kappas[1]),
      median = calc.corr(dist, kappas[2]),
      upper = calc.corr(dist, kappas[3])
    )
  } else { ## !is.null(log.variance)
    if (!is.list(log.variance)) {
      log.variance <- list(mean = log.variance, sd = 0)
    }
    kappa <- sqrt(8 * nu) / exp(log.range$mean)
    out <- data.frame(
      lower = Inf,
      median = calc.cov(dist, kappa, exp(log.variance$mean)),
      upper = -Inf
    )
    if (log.variance$sd == 0) {
      out$lower <- out$median
      out$upper <- out$median
    } else {
      qq <- qchisq(quantile, 2)^0.5
      log.kappas <- log(sqrt(8 * nu)) - (log.range$mean +
        log.range$sd * c(qq, -qq))
      log.variances <- log.variance$mean + log.variance$sd * c(-qq, qq)
      for (angle in 2 * pi * (1:n) / n) {
        ca <- cos(angle)
        sa <- sin(angle)
        the.corr <- calc.cov(
          dist,
          exp(log.kappas[1] * (1 - ca) / 2 +
            log.kappas[2] * (ca + 1) / 2),
          exp(log.variances[1] * (1 - sa) / 2 +
            log.variances[2] * (sa + 1) / 2)
        )
        out$lower <- pmin(out$lower, the.corr)
        out$upper <- pmax(out$upper, the.corr)
      }
    }
  }
  out
}


#' @title Posteriors of SPDE hyper parameters and Matern correlation or
#'   covariance function.
#'
#' @description
#' Calculate posterior distribution of the range, log(range), variance, or
#' log(variance) parameter of a model's SPDE component. Can also plot Matern
#' correlation or covariance function. `inla.spde.result`.
#'
#' @param result An object inheriting from `inla`.
#' @param name Character stating the name of the SPDE effect, see
#'   `names(result$summary.random)`.
#' @param what One of "range", "log.range", "variance", "log.variance",
#'   "matern.correlation" or "matern.covariance".
#' @inheritParams materncov.bands
#' @return A `prediction` object.
#'
#' @export
#' @author Finn Lindgren \email{Finn.Lindgren@@ed.ac.uk}
#' @examples
#' \donttest{
#' if (bru_safe_inla() && require(ggplot2, quietly = TRUE)) {
#'   # Load 1D Poisson process data
#'
#'   data(Poisson2_1D, package = "inlabru")
#'
#'
#'   # Take a look at the point (and frequency) data
#'
#'   ggplot(pts2) +
#'     geom_histogram(
#'       aes(x = x),
#'       binwidth = 55 / 20,
#'       boundary = 0,
#'       fill = NA,
#'       color = "black"
#'     ) +
#'     geom_point(aes(x), y = 0, pch = "|", cex = 4) +
#'     coord_fixed(ratio = 1)
#'
#'   # Fit an LGCP model with  and SPDE component
#'
#'   x <- seq(0, 55, length.out = 20)
#'   mesh1D <- fmesher::fm_mesh_1d(x, boundary = "free")
#'   constrained_matern <- INLA::inla.spde2.matern(mesh1D, constr = TRUE)
#'   mdl <- x ~ spde1D(x, model = constrained_matern) + Intercept(1)
#'   fit <- lgcp(mdl, data = pts2, domain = list(x = mesh1D))
#'
#'   # Calculate and plot the posterior range
#'
#'   range <- spde.posterior(fit, "spde1D", "range")
#'   plot(range)
#'
#'   # Calculate and plot the posterior log range
#'
#'   lrange <- spde.posterior(fit, "spde1D", "log.range")
#'   plot(lrange)
#'
#'   # Calculate and plot the posterior variance
#'
#'   variance <- spde.posterior(fit, "spde1D", "variance")
#'   plot(variance)
#'
#'   # Calculate and plot the posterior log variance
#'
#'   lvariance <- spde.posterior(fit, "spde1D", "log.variance")
#'   plot(lvariance)
#'
#'   # Calculate and plot the posterior Matern correlation
#'
#'   matcor <- spde.posterior(fit, "spde1D", "matern.correlation")
#'   plot(matcor)
#'
#'   # Calculate and plot the posterior Matern covariance
#'
#'   matcov <- spde.posterior(fit, "spde1D", "matern.covariance")
#'   plot(matcov)
#' }
#' }
#'
spde.posterior <- function(result, name, what = "range",
                           quantile = 0.95) {
  stopifnot(bru_safe_inla(multicore = TRUE))

  spdespec <- as_bru_comp_list(result)[[name]]$main$model
  spderesult <- INLA::inla.spde.result(result, name, spdespec)

  if (what == "matern.correlation" || what == "matern.covariance") {
    xmax <- exp(spderesult$summary.log.range.nominal[["0.975quant"]]) * 1.2
    x <- seq(0, xmax, length.out = 200)
    log.range <- list(
      mean = spderesult$summary.log.range.nominal[["mean"]],
      sd = spderesult$summary.log.range.nominal[["sd"]]
    )
    log.variance <- list(
      mean = spderesult$summary.log.variance.nominal[["mean"]],
      sd = spderesult$summary.log.variance.nominal[["sd"]]
    )
    if (what == "matern.correlation") {
      corr <- TRUE
      ylab <- "Matern Correlation"
      out <- materncov.bands(
        as_bru_comp_list(result)[[name]]$main$mapper$mesh,
        dist = x,
        log.range = log.range,
        log.variance = NULL,
        alpha = 2, quantile = quantile
      )
    } else {
      corr <- FALSE
      ylab <- "Matern Covariance"
      out <- materncov.bands(
        as_bru_comp_list(result)[[name]]$main$mapper$mesh,
        dist = x,
        log.range = log.range,
        log.variance = log.variance,
        alpha = 2, quantile = quantile
      )
    }


    df <- data.frame(
      x = x, q0.5 = out$median, lower = out$lower, upper = out$upper,
      median = out$median
    )
    colnames(df)[3] <- paste0("q", paste0(round((1 - quantile) / 2, 5), ""))
    colnames(df)[4] <-
      paste0("q", paste0(round(1 - (1 - quantile) / 2, 5), ""))
    attr(df, "type") <- "1d"
    attr(df, "misc") <- list(dims = "x", predictor = c("distance", ylab))
    class(df) <- list("prediction", "data.frame")
    df
  } else {
    marg <- switch(what,
      range = spderesult$marginals.range.nominal[[1]],
      log.range = spderesult$marginals.log.range.nominal[[1]],
      variance = spderesult$marginals.variance.nominal[[1]],
      log.variance = spderesult$marginals.log.variance.nominal[[1]]
    )
    if (is.null(marg)) {
      stop("Invalid varname: ", what, ". must be one of 'range',
                           'log.range',  'variance',  'log.variance',
                           'matern.correlation', matern.covariance")
    }

    med <- INLA::inla.qmarginal(0.5, marg)
    uq <- INLA::inla.qmarginal(1 - (1 - quantile) / 2, marg)
    lq <- INLA::inla.qmarginal((1 - quantile) / 2, marg)
    inner.x <- seq(lq, uq, length.out = 100)
    inner.marg <- data.frame(
      x = inner.x,
      y = INLA::inla.dmarginal(inner.x, marg)
    )
    colnames(inner.marg) <- c(what, "pdf")
    df <- data.frame(marg)
    colnames(df) <- c(what, "pdf")
    attr(df, "type") <- "0d"
    attr(df, "summary") <- list(
      uq = uq,
      lq = lq,
      median = med,
      inner.marg = inner.marg
    )
    class(df) <- list("prediction", "data.frame")
    df
  }
}

#' @name mexdolphin_sf
#' @title Pan-tropical spotted dolphins in the Gulf of Mexico
#' @docType data
#' @description This a version of the `mexdolphins` dataset from the package
#'   `dsm`, reformatted as point process data for use with `inlabru`, with the
#'   parts stored in `sf` format. The data are from a combination of several
#'   NOAA shipboard surveys conducted on pan-tropical spotted dolphins in the
#'   Gulf of Mexico. 47 observations of groups of dolphins were detected. The
#'   group size was recorded, as well as the Beaufort sea state at the time of
#'   the observation. Transect width is 16 km, i.e. maximal detection distance 8
#'   km (transect half-width 8 km).
#'
#' @format A list of objects:
#'  \describe{
#'    \item{`points`:}{ An `sf` object containing the locations of
#'    detected dolphin groups, with their size as an attribute.}
#'    \item{`samplers`:}{ An `sf` object containing the transect lines
#'    that were surveyed.}
#'    \item{`mesh`:}{ An `fm_mesh_2d` object containing a Delaunay triangulation
#'    mesh (a type of discretization of continuous space) covering the survey
#'    region.}
#'    \item{`ppoly`:}{ An `sf` object defining the boundary of the
#'    survey region.}
#'    \item{`simulated`:}{ A `sf` object containing the locations of a
#'    *simulated* population of dolphin groups. The population was simulated
#'    from a `inlabru`
#'    model fitted to the actual survey data. Note that the simulated data
#'    do not have any associated
#'    size information.}
#'  }
#' @source
#' Library `dsm`.
#'
#' @references
#' Halpin, P.N., A.J. Read, E. Fujioka, B.D. Best, B. Donnelly, L.J. Hazen, C.
#' Kot, K. Urian, E. LaBrecque, A. Dimatteo, J. Cleary, C. Good, L.B. Crowder,
#' and K.D. Hyrenbach. 2009. OBIS-SEAMAP: The world data center for marine
#' mammal, sea bird, and sea turtle distributions. Oceanography 22(2):104-115
#'
#' NOAA Southeast Fisheries Science Center. 1996. Report of a Cetacean Survey of
#' Oceanic and Selected Continental Shelf Waters of the Northern Gulf of Mexico
#' aboard NOAA Ship Oregon II (Cruise 220)
#'
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE)) {
#'   data(mexdolphin_sf, package = "inlabru", envir = environment())
#'   ggplot() +
#'     gg(mexdolphin_sf$mesh) +
#'     gg(mexdolphin_sf$ppoly, color = "blue", alpha = 0, linewidth = 1) +
#'     gg(mexdolphin_sf$samplers) +
#'     gg(mexdolphin_sf$points, aes(size = size), color = "red") +
#'     scale_size_area()
#'
#'   ggplot() +
#'     gg(mexdolphin_sf$mesh,
#'       color = mexdolphin_sf$lambda,
#'       mask = mexdolphin_sf$ppoly
#'     )
#' }
#' }
"mexdolphin_sf"

#' @describeIn mexdolphin_sf Convert `mexdolphin_sf` to `sp` format. Replaces
#'   the old `mexdolphin` dataset.
#' @export
#'
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE) &&
#'   require("sp", quietly = TRUE)) {
#'   mexdolphin <- mexdolphin_sp()
#'   ggplot() +
#'     gg(mexdolphin$mesh) +
#'     gg(mexdolphin$ppoly, color = "blue") +
#'     gg(mexdolphin$samplers) +
#'     gg(mexdolphin$points, aes(size = size), color = "red") +
#'     scale_size_area() +
#'     coord_equal()
#'
#'   ggplot() +
#'     gg(mexdolphin$mesh,
#'       col = mexdolphin$lambda,
#'       mask = mexdolphin$ppoly
#'     ) +
#'     coord_equal()
#' }
#' }
mexdolphin_sp <- function() {
  dat <- inlabru::mexdolphin_sf
  out <- list(
    points = sf::as_Spatial(dat$points),
    samplers = sf::as_Spatial(dat$samplers),
    mesh = dat$mesh,
    ppoly = sf::as_Spatial(dat$ppoly),
    lambda = dat$lambda,
    simulated = sf::as_Spatial(dat$simulated),
    depth = sf::as_Spatial(dat$depth)
  )
  sp::coordnames(out$points) <- c("x", "y")
  sp::coordnames(out$samplers) <- c("x", "y")
  sp::coordnames(out$ppoly) <- c("x", "y")
  sp::coordnames(out$simulated) <- c("x", "y")
  sp::coordnames(out$depth) <- c("x", "y")
  out
}

#' @title Deprecated alias for sp version of the mexdolphin dataset
#' @name mexdolphin
#' @rdname mexdolphin
#' @description
#' Deprecated dataset name for the `sp` version
#' of [mexdolphin_sf]. Use [mexdolphin_sp()] instead.
#' @seealso [mexdolphin_sf]
#' @keywords internal
NULL

#' @title Gorilla nesting sites in sf format
#' @docType data
#' @description This is the `gorillas` dataset from the package `spatstat.data`,
#'   reformatted as point process data for use with `inlabru`.
#'
#' @usage
#' gorillas_sf
#' data(gorillas_sf, package = "inlabru")
#'
#' @format The data are a list that contains these elements:
#'  \describe{
#'    \item{`nests`:}{ An `sf` object containing the locations of
#'    the gorilla nests.}
#'    \item{`boundary`:}{ An `sf` object defining the boundary
#'    of the region that was searched for the nests.}
#'    \item{`mesh`:}{ An `fm_mesh_2d` object containing a mesh that can be used
#'    with function `lgcp` to fit a LGCP to the nest data.}
#'    \item{`gcov_file`:}{ The in-package filename of a `terra::SpatRaster`
#'    object, with one layer for each of these spatial covariates:
#'     \describe{
#'       \item{`aspect`}{ Compass direction of the terrain slope. Categorical,
#'       with levels
#'       N, NE, E, SE, S, SW, W and NW, which are coded as integers 1 to 8.}
#'       \item{`elevation`}{ Digital elevation of terrain, in metres.}
#'       \item{`heat`}{ Heat Load Index at each point on the surface
#'       (Beer's aspect),
#'       discretised. Categorical with values Warmest (Beer's aspect between
#'       0 and 0.999),
#'       Moderate (Beer's aspect between 1 and 1.999), Coolest (Beer's aspect
#'       equals 2). These are
#'       coded as integers 1, 2 and 3, in that order.}
#'       \item{`slopangle`}{ Terrain slope, in degrees.}
#'       \item{`slopetype`}{ Type of slope. Categorical, with values Valley,
#'       Toe (toe slope),
#'       Flat, Midslope, Upper and Ridge. These are coded as integers 1 to 6.}
#'       \item{`vegetation`}{ Vegetation type: a categorical variable with 6
#'       levels coded as
#'       integers 1 to 6 (in order of increasing expected habitat suitability)}
#'       \item{`waterdist`}{ Euclidean distance from nearest water body, in
#'       metres.}
#'     }
#'     Loading of the covariates can be done with `gorillas_sf_gcov()` or
#'
#'     gorillas_sf$gcov <- terra::rast(
#'       system.file(gorillas_sf$gcov_file, package = "inlabru")
#'     )
#'
#'    }
#'    \item{`plotsample`}{Plot sample of gorilla nests, sampling 9x9 over the
#'    region, with 60\% coverage. Components:
#'    \describe{
#'      \item{`counts`}{An `sf` object with elements
#'      `count`, `exposure`, and `geometry`, holding the point geometry for the
#'      centre of each plot, the count in each
#'      plot and the area of each plot.}
#'      \item{`plots`}{An `sf` object with `MULTIPOLYGON` objects defining the
#'      individual plot boundaries and an all-ones `weight` column.}
#'      \item{`nests`}{An `sf` giving the locations of
#'      each detected nests, `group` ("minor" or "major"),
#'      `season` ("dry" or "rainy"), and `date` (in `Date` format).}
#'    }
#'    }
#'  }
#' @source
#' Library `spatstat.data`.
#'
#'
#' @references
#' Funwi-Gabga, N. (2008) A pastoralist survey and fire impact assessment in the
#' Kagwene Gorilla Sanctuary, Cameroon. M.Sc. thesis, Geology and Environmental
#' Science, University of Buea, Cameroon.
#'
#' Funwi-Gabga, N. and Mateu, J. (2012) Understanding the nesting spatial
#' behaviour of gorillas in the Kagwene Sanctuary, Cameroon. Stochastic
#' Environmental Research and Risk Assessment 26 (6), 793-811.
#'
#' @examples
#' if (interactive() &&
#'   require(ggplot2, quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   requireNamespace("tidyterra", quietly = TRUE)) {
#'   # plot all the nests, mesh and boundary
#'   ggplot() +
#'     gg(gorillas_sf$mesh) +
#'     geom_sf(
#'       data = gorillas_sf$boundary,
#'       alpha = 0.1, fill = "blue"
#'     ) +
#'     geom_sf(data = gorillas_sf$nests)
#'
#'   # Plot the elevation covariate
#'   gorillas_sf$gcov <- gorillas_sf_gcov()
#'   ggplot() +
#'     tidyterra::geom_spatraster(data = gorillas_sf$gcov$elevation)
#'
#'   # Plot the plot sample
#'   ggplot() +
#'     geom_sf(data = gorillas_sf$plotsample$plots) +
#'     geom_sf(data = gorillas_sf$plotsample$nests)
#' }
"gorillas_sf"


#' @describeIn gorillas_sf Access the `gorillas_sf` covariates data as a
#' `terra::rast()` object.
#' @export
#' @examples
#' if (interactive() &&
#'   bru_safe_terra(quietly = TRUE)) {
#'   gorillas_sf$gcov <- gorillas_sf_gcov()
#' }
gorillas_sf_gcov <- function() {
  # Intentionally not calling bru_safe_terra() here, since this data is needed
  # by bru_safe_terra() itself.
  requireNamespace("terra")
  terra::rast(system.file(inlabru::gorillas_sf$gcov_file, package = "inlabru"))
}

#' @describeIn gorillas_sf Access the `gorillas_sf` data in `sp` format.
#' The covariate data is added as `gcov`, a list of `sp::SpatialPixelsDataFrame`
#' objects. Requires the `sp`, `sf`, and `terra` packages to be installed.
#' @export
gorillas_sp <- function() {
  requireNamespace("sf")
  requireNamespace("terra")
  stopifnot(bru_safe_sp())
  dat <- inlabru::gorillas_sf
  gcov <- gorillas_sf_gcov()
  gcov_ <- as.data.frame(gcov, xy = TRUE)
  gcov_sp <- list()
  for (nm in names(gcov)) {
    gcov_sp[[nm]] <- sp::SpatialPixelsDataFrame(
      cbind(gcov_$x, gcov_$y),
      data = gcov_[, nm, drop = FALSE],
      proj4string = fm_CRS(gcov)
    )
  }

  out <- list(
    nests = sf::as_Spatial(dat$nests),
    mesh = dat$mesh,
    boundary = sf::as_Spatial(dat$boundary),
    plotsample = list(
      counts = sf::as_Spatial(dat$plotsample$counts),
      plots = sf::as_Spatial(dat$plotsample$plots),
      nests = sf::as_Spatial(dat$plotsample$nests)
    ),
    gcov = gcov_sp
  )
  out
}

#' @title Deprecated alias for sp version of the gorillas dataset
#' @name gorillas
#' @rdname gorillas
#' @description
#' Deprecated dataset name for the `sp` version
#' of [gorillas_sf]. Use [gorillas_sp()] instead.
#' @seealso [gorillas_sf]
#' @keywords internal
NULL

#' @name toypoints
#' @title Simulated 2D point process data
#' @docType data
#' @description
#'
#' This data set serves as an example for basic inlabru.
#'
#' @usage data(toypoints)
#'
#' @format The data are a list that contains these elements:
#'  \describe{
#'    \item{`points`}{An `sf` object of point locations and and `z`
#'      measurements}
#'    \item{`mesh`}{An `fm_mesh_2d` object}
#'    \item{`boundary`}{An `sf` polygon denting the region of interest}
#'    \item{`pred_locs`}{A `sf` object with prediction point locations}
#'  }
#'
#'
#' @examples
#' if (require("ggplot2")) {
#'   ggplot() +
#'     fmesher::geom_fm(data = toypoints$mesh, alpha = 0) +
#'     geom_sf(data = toypoints$boundary, fill = "blue", alpha = 0.1) +
#'     geom_sf(data = toypoints$points, aes(color = z)) +
#'     scale_color_viridis_c()
#' }
"toypoints"

#' @title Check for predictor linearity
#' @description Checks if a predictor expression is linear (or affine)
#' @param x A object containing a predictor definition
#' @param \dots Arguments passed on recursively.
#' @return `TRUE` if the expression is detected to be linear, `FALSE`
#'   otherwise.
#' @export
#' @examples
#' bru_is_linear(new_bru_pred_expr(~ x + y))
#' bru_is_linear(new_bru_pred_expr(~ x * y))
#' bru_is_linear(bm_scale())
#' bru_is_linear(bm_logsumexp())
#'
bru_is_linear <- function(x, ...) {
  UseMethod("bru_is_linear")
}

#' @rdname bru_is_linear
#' @export
bru_is_linear.bru <- function(x, ...) {
  all(bru_is_linear(x[["bru_info"]], ...))
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_info <- function(x, ...) {
  all(bru_is_linear(x[["lhoods"]], ...)) &&
    all(bru_is_linear(x[["model"]], ...))
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_model <- function(x, ...) {
  bru_is_linear(x[["effects"]], ...)
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_obs <- function(x, ...) {
  bru_is_linear(x[["pred_expr"]])
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_obs_list <- function(x, ...) {
  vapply(x, function(lh) bru_is_linear(lh, ...), logical(1))
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_pred_expr <- function(x, ...) {
  isTRUE(x[["is_linear"]])
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_comp_list <- function(x, ...) {
  vapply(x, bru_is_linear, logical(1))
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_comp <- function(x, ...) {
  bru_is_linear(x[["mapper"]])
}
#' @rdname bru_is_linear
#' @export
bru_is_linear.bru_mapper <- function(x, ...) {
  ibm_is_linear(x)
}


#' @title Check for predictor expression additivity
#' @description Checks if a predictor expression is additive or not
#' @param x A predictor `expression`, `formula`, or parse information
#'   `data.frame`.
#' @param \dots Arguments passed on recursively.
#' @param verbose logical; if `TRUE`, print diagnostic parsing information.
#' @return `TRUE` if the expression is detected to be additive, `FALSE`
#'   otherwise.
#' @export
#' @examples
#' bru_is_additive(~ x + y)
#' bru_is_additive(~ x * y)
#'
bru_is_additive <- function(x, ...) {
  UseMethod("bru_is_additive")
}

bru_is_additive_data_frame <- function(x, root_id = 0, ..., verbose = FALSE) {
  if (root_id == 0) {
    root_id <- x$id[x$parent == 0]
    if (length(root_id) != 1L) {
      stop("Cannot determine parser root id")
    }
  }

  x_root <- x[x$id == root_id, , drop = FALSE]
  if (x_root$token == "SYMBOL") {
    if (verbose) {
      message("SYMBOL found")
    }
    return(TRUE)
  }
  if (x_root$token == "expr") {
    if (verbose) {
      message("expr found, may be additive")
    }
    x_terms <- x[x$parent == root_id, , drop = FALSE]
    if (nrow(x_terms) < 1L) {
      if (verbose) {
        message("No terms found, assuming non-additive")
      }
      return(FALSE)
    }
    if (nrow(x_terms) == 1L) {
      if (verbose) {
        message("Only one term, checking for symbol")
      }
      add <- x_terms$token[1] == "SYMBOL"
      if (add && verbose) {
        message("SYMBOL found")
      }
      return(add)
    }
    if (nrow(x_terms) == 2L) {
      if (x_terms$token[1] != "'+'") {
        if (verbose) {
          message("Only two terms and no '+' found, assuming non-additive")
        }
        return(FALSE)
      }
      if (verbose) {
        message("'+' found, may be additive")
      }
      add <- bru_is_additive_data_frame(
        x = x,
        root_id = x_terms$id[2],
        ...,
        verbose = verbose
      )
      return(all(add))
    }
    if (nrow(x_terms) >= 4L) {
      if (verbose) {
        message("More than 3 terms, assuming non-additive")
      }
      return(FALSE)
    }
    if (x_terms$token[2] == "'+'") {
      if (verbose) {
        message("'+' found, may be additive")
      }
      add <- c(
        bru_is_additive_data_frame(
          x = x,
          root_id = x_terms$id[1],
          verbose = verbose
        ),
        bru_is_additive_data_frame(
          x = x,
          root_id = x_terms$id[3],
          verbose = verbose
        )
      )
      return(all(add))
    }
    if ((x_terms$token[1] == "'('") && (x_terms$token[3] == "')'")) {
      if (x_terms$token[2] == "expr") {
        if (verbose) {
          message("(expr) found, may be additive")
        }
        add <- bru_is_additive_data_frame(
          x = x,
          root_id = x_terms$id[2],
          verbose = verbose
        )
        return(add)
      }
    }
  }

  if (verbose) {
    message("No known additive structure found, assuming non-additive")
  }

  FALSE
}

#' @rdname bru_is_additive
#' @export
bru_is_additive.character <- function(x, ..., verbose = FALSE) {
  bru_is_additive_data_frame(
    bru_get_parse_data(x),
    root_id = 0,
    ...
  )
}

#' @rdname bru_is_additive
#' @export
bru_is_additive.expression <- function(x, ..., verbose = FALSE) {
  bru_is_additive_data_frame(
    bru_get_parse_data(as.character(x)),
    root_id = 0,
    ...
  )
}

#' @rdname bru_is_additive
#' @export
bru_is_additive.formula <- function(x, ..., verbose = FALSE) {
  bru_is_additive_data_frame(
    bru_get_parse_data(as.character(x)[length(x)]),
    root_id = 0,
    ...
  )
}


#' @rdname bru_is_additive
#' @export
bru_is_additive.bru_pred_expr <- function(x, ...) {
  isTRUE(x[["is_additive"]])
}
#' @rdname bru_is_additive
#' @export
bru_is_additive.bru_obs <- function(x, ...) {
  bru_is_additive(x[["pred_expr"]], ...)
}
#' @rdname bru_is_additive
#' @export
bru_is_additive.bru_obs_list <- function(x, ...) {
  vapply(x, function(lh) bru_is_linear(lh, ...), logical(1))
}


#' @title Check for predictor rowwise evaluability
#' @description Checks if a predictor expression may be evaluated rowwise
#' @param x An object containing a predictor definition
#' @param \dots Arguments passed on to submethods.
#' @return `TRUE` if the expression is believed to be rowwise, `FALSE`
#'   otherwise.
#' @export
#' @examples
#' bru_is_rowwise(new_bru_pred_expr(~ x + y))
#'
bru_is_rowwise <- function(x, ...) {
  UseMethod("bru_is_rowwise")
}

#' @rdname bru_is_rowwise
#' @export
bru_is_rowwise.bru_pred_expr <- function(x, ...) {
  isTRUE(x[["is_rowwise"]])
}
#' @rdname bru_is_rowwise
#' @export
bru_is_rowwise.bru_obs <- function(x, ...) {
  bru_is_rowwise(x[["pred_expr"]], ...)
}
#' @rdname bru_is_rowwise
#' @export
bru_is_rowwise.bru_obs_list <- function(x, ...) {
  vapply(x, function(lh) bru_is_rowwise(lh, ...), logical(1))
}
#' @rdname bru_is_rowwise
#' @export
bru_is_rowwise.bru_comp_list <- function(x, ...) {
  vapply(x, bru_is_rowwise, logical(1))
}
#' @rdname bru_is_rowwise
#' @export
bru_is_rowwise.bru_comp <- function(x, ...) {
  bru_is_rowwise(x[["mapper"]])
}
#' @rdname bru_is_rowwise
#' @export
bru_is_rowwise.bru_mapper <- function(x, ...) {
  ibm_is_rowwise(x)
}

test_that("bru: linear component", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123)

  input.df <- data.frame(x = cos(1:100))
  input.df <- within(input.df, {
    y <- 5 + 2 * x + rnorm(100, mean = 0, sd = 0.1)
  })
  fit <- bru(
    y ~ myLin(main = x, model = "linear") + Intercept(1),
    family = "gaussian",
    data = input.df
  )

  expect_equal(fit$summary.fixed["myLin", "mean"], 2.002273, tolerance = midtol)
  expect_equal(fit$summary.fixed["myLin", "sd"], 0.01323361, tolerance = hitol)

  skip_if_not_installed("sn")
  pr <- predict(
    fit,
    data.frame(x = c(1, 2)),
    ~ myLin + 2,
    n.samples = 5,
    seed = 1L
  )

  expect_equal(pr[, "mean"], c(4.005013, 6.010026), tolerance = midtol)
})


test_that("bru: linear predictor detection", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123)

  input.df <- data.frame(x = cos(1:100), z = sin(1:100))
  input.df <- within(input.df, {
    y <- 5 + 2 * x + rnorm(100, mean = 0, sd = 0.1)
  })
  fit <- bru(
    ~ x(x) + z(z) + Intercept(1),
    formula = y ~ .,
    family = "gaussian",
    data = input.df,
    options = list(bru_run = FALSE)
  )
  expect_equal(
    bru_pred_expr(fit, format = "text")[[1]],
    "x + z + Intercept",
    info = "Should construct correct expression"
  )
  expect_true(
    bru_is_additive(as_bru_obs_list(fit)[[1]]),
    info = "Should detect additive predictor"
  )
  expect_true(
    bru_is_linear(as_bru_obs_list(fit)[[1]]),
    info = "Should detect linear predictor"
  )
  expect_equal(
    bru_used(fit),
    bru_used(effect = c("x", "z", "Intercept"), latent = character(0))
  )

  fit <- bru(
    ~ x(x) + z(z) + Intercept(1),
    formula = y ~ x + Intercept,
    family = "gaussian",
    data = input.df,
    options = list(bru_run = FALSE)
  )
  expect_equal(
    bru_pred_expr(as_bru_obs_list(fit)[[1]], format = "text"),
    "x + Intercept",
    info = "Should construct correct expression"
  )
  expect_true(
    bru_is_additive(as_bru_obs_list(fit)[[1]]),
    info = "Should detect additive predictor"
  )
  expect_true(
    bru_is_linear(as_bru_obs_list(fit)[[1]]),
    info = "Should detect linear predictor"
  )
  expect_equal(
    bru_used(fit),
    bru_used(effect = c("x", "Intercept"), latent = character(0))
  )

  # Predictor with use of _latent should be detected as non-additive
  fit <- bru(
    ~ x(x) + z(z) + Intercept(1),
    formula = y ~ x + z_latent + Intercept,
    family = "gaussian",
    data = input.df,
    options = list(bru_run = FALSE)
  )
  expect_equal(
    bru_pred_expr(as_bru_obs_list(fit)[[1]], format = "text"),
    "x + z_latent + Intercept",
    info = "Should construct correct expression"
  )
  expect_false(
    bru_is_additive(as_bru_obs_list(fit)[[1]]),
    info = "Should detect non-additive predictor"
  )
  expect_false(
    bru_is_linear(as_bru_obs_list(fit)[[1]]),
    info = "Should detect non-linear predictor, due to non-pure additivity"
  )
  expect_equal(
    bru_used(fit),
    bru_used(effect = c("x", "Intercept"), latent = "z"),
    info = "Should detect two components, 'x' and 'Intercept', and latent 'z'"
  )

  # Predictor with missing component/variable should be detected as
  # non-additive, but only give errors when attempting to evaluate the model
  fit <- bru(
    ~ x(x) + z(z) + Intercept(1),
    formula = y ~ x + Intercept + something,
    family = "gaussian",
    data = input.df,
    options = list(bru_run = FALSE)
  )
  expect_equal(
    bru_pred_expr(as_bru_obs_list(fit)[[1]], format = "text"),
    "x + Intercept + something",
    info = "Should construct correct expression"
  )
  expect_false(
    bru_is_additive(as_bru_obs_list(fit)[[1]]),
    info = "Should detect non-additive predictor"
  )
  expect_false(
    bru_is_linear(as_bru_obs_list(fit)[[1]]),
    info = "Should detect non-linear predictor, due to missing variable"
  )
  expect_equal(
    bru_used(fit),
    bru_used(effect = c("x", "Intercept"), latent = character(0)),
    info = "Should detect two components, 'x' and 'Intercept'"
  )
})

# Internal checker for spatstat packages
# Adapted from Adrian Baddeley and Ege Rubak
check_spatstat <- function(pkg = "spatstat.geom") {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    caller_name <- fm_caller_name(1)
    stop(paste0(
      "package '",
      pkg,
      "' required",
      if (identical(caller_name, "")) {
        ""
      } else {
        paste0(" by '", caller_name, "'")
      },
      "; please install it (or the full spatstat package) first"
    ))
  } else {
    spst_ver <- try(utils::packageVersion("spatstat"), silent = TRUE)
    if (!inherits(spst_ver, "try-error") && spst_ver < "2.0-0") {
      warning(paste0(
        "You have an old version of 'spatstat' installed which is ",
        "incompatible with '",
        pkg,
        "'. Please update 'spatstat' (or uninstall it)."
      ))
      return(FALSE)
    }
  }
  TRUE
}

#' Convert SpatialPoints and boundary polygon to spatstat ppp object
#'
#' Spatstat point pattern objects consist of points and an observation windows.
#' This function uses a `SpatialPoints` object and a `SpatialPolygon` object to
#' generate the points and the window. Lastly, the `ppp()` function is called to
#' create the `ppp` object.
#'
#' @export
#' @param points A `SpatialPoints[DataFrame]` object describing the point
#'   pattern.
#' @param samplers A `SpatialPolygons[DataFrame]` object describing the
#'   observation window.
#' @return A spatstat `spatstat` `ppp` object
#'
#' @examples
#' \donttest{
#' if (require("spatstat.geom") &&
#'   bru_safe_sp() &&
#'   require("sp") &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("sf", quietly = TRUE)) {
#'   # Load Gorilla data
#'
#'   gorillas <- gorillas_sp()
#'
#'   # Use nest locations and survey boundary to create a spatstat ppp object
#'
#'   gp <- spatial.to.ppp(gorillas$nests, gorillas$boundary)
#'   class(gp)
#'
#'   # Plot it
#'
#'   plot(gp)
#' }
#' }
#'
#' @keywords internal
spatial.to.ppp <- function(points, samplers) {
  check_spatstat("spatstat.geom")

  bnd <- samplers@polygons[[1]]@Polygons[[1]]@coords
  bnd <- bnd[1:(nrow(bnd) - 1), ]
  gp <- spatstat.geom::ppp(
    x = sp::coordinates(points)[, 1],
    y = sp::coordinates(points)[, 2],
    window = spatstat.geom::owin(
      poly = list(x = rev(bnd[, 1]), y = rev(bnd[, 2]))
    )
  )
}

#' @title Grid partitioning
#'
#' @description
#' `r lifecycle::badge("experimental")` Both the interface and function name
#' may change in future versions.
#'
#' Partitions the region based on the given criteria for calculating residuals
#' in each partition. Parts of this function are taken from concepts in
#' https://rpubs.com/huanfaChen/grid_from_polygon
#'
#' @param samplers A sf object containing region for which
#' partitions to be created
#' @param resolution resolution of the grids that are required
#' @param nrows number of rows of grids that are required
#' @param ncols number of columns of grids that are required
#' @param chess chessboard partitioning
#' @param \dots Unused.
#' @return a partitioned sf object as required or a list of partitioned sf
#'   objects if chess is TRUE
#'
#' @keywords internal
#' @export
#'
#' @author Man Ho Suen
cv_partition <- function(samplers,
                         resolution = NULL,
                         nrows = NULL,
                         ncols = NULL,
                         chess = TRUE,
                         ...) {
  requireNamespace("terra")
  # Create a grid for the given boundary
  if (is.null(resolution)) {
    grid <- terra::rast(
      terra::ext(sf::st_as_sf(
        fmesher::fm_nonconvex_hull(samplers, ...)
      )),
      crs = fmesher::fm_crs(samplers)$input,
      nrows = nrows,
      ncols = ncols
    )
  }

  if (is.null(c(nrows, ncols))) {
    grid <- terra::rast(
      terra::ext(sf::st_as_sf(
        fmesher::fm_nonconvex_hull(samplers, ...)
      )),
      crs = fmesher::fm_crs(samplers)$input,
      resolution = resolution
    )
  }

  gridPolygon <- terra::as.polygons(grid)
  if (isTRUE(chess)) {
    # no idea how it works
    # spatSample(x = gridPolygon, size = 0.5*nrow(gridPolygon),
    #            method="random", strata=NULL, chess="black")
    grid_chess <- terra::init(grid, "chess")
    grid_chess_sf <- sf::st_intersection(
      sf::st_cast(sf::st_as_sf(terra::as.polygons(grid_chess)), "POLYGON"),
      samplers
    )

    # I am not sure which terra version changes how they name for init()
    terra_version <- utils::packageVersion("terra")
    if (terra_version <= "1.8-42") {
      grid_chess_white_sf <- grid_chess_sf[grid_chess_sf$lyr.1 == 1, ]
      grid_chess_black_sf <- grid_chess_sf[grid_chess_sf$lyr.1 == 0, ]
    } else {
      grid_chess_white_sf <- grid_chess_sf[grid_chess_sf$chess == 1, ]
      grid_chess_black_sf <- grid_chess_sf[grid_chess_sf$chess == 0, ]
    }
    return(list(white = grid_chess_white_sf, black = grid_chess_black_sf))
    # ggplot() + gg(data = grid_chess_white_sf, aes(fill = lyr.1)) +
    #   geom_sf(data = nepal_bnd, col = "red", fill = "NA")
    # ggplot() + gg(data = grid_chess_black_sf, aes(fill = lyr.1)) +
    #   geom_sf(data = nepal_bnd, col = "red", fill = "NA")
  } else {
    # Extract the boundary with subpolygons only
    gridPolygon <- sf::st_as_sf(
      terra::intersect(gridPolygon, terra::vect(samplers))
    )
  }
  gridPolygon
}


#' @title Hexagon tiling of region
#'
#' @description
#' `r lifecycle::badge("experimental")` Both the interface and function name
#' may change in future versions.
#'
#' Partitions the region based on the hexagon tiling
#'
#' Input:
#' @param samplers A sf object containing region for which
#' partitions to be created
#' @param cellsize hexagon cellsize, see [sf::st_make_grid()] description
#' @param n_group number of cv folds.
#' @param \dots Passed on to `fm_nonconvex_hull()`, e.g. `resolution`
#' @returns a hexagonal partition covering the `samplers` object, with a `group`
#' integer variable indicating the fold assignment, 1, 2, ..., `n_group`.
#'
#' @keywords internal
#' @export
#'
#' @author Man Ho Suen
#' @examples
#' if (interactive()) {
#'   bnd <- gorillas_sf$boundary
#'   hex_cv <- cv_hex(bnd, cellsize = 0.5, n_group = 3, resolution = 100)
#'   plot(hex_cv)
#'
#'   chess <- cv_partition(bnd, resolution = 0.5, chess = TRUE)
#'   plot(chess$white)
#' }
#'
cv_hex <- function(samplers, cellsize = 0.5, n_group = 3, ...) {
  bnd <- fmesher::fm_nonconvex_hull(samplers, convex = -0.01, ...)
  hex_cell <- sf::st_make_grid(
    bnd,
    cellsize = cellsize,
    square = FALSE
  )
  hex_cell <- sf::st_sf(geometry = hex_cell)

  # get centroids for x positions
  hex_centroids <- sf::st_centroid(hex_cell)
  coords <- sf::st_coordinates(hex_centroids)[, 1]

  # get sorted unique x positions (columns)
  x_unique <- sort(unique(round(coords, 8)))

  # map each hex to a column index
  col_index <- match(round(coords, 8), x_unique)

  # assign groups by cycling 1 → 2 → 3 -> etc -> 1
  col_group <- ((col_index - 1) %% n_group) + 1

  # attach back
  hex_groups <-
    dplyr::mutate(
      hex_cell,
      group = col_group
    )

  sf::st_agr(hex_groups) <- "constant"
  hex_cell_ <- sf::st_intersection(hex_groups, sf::st_geometry(samplers))

  hex_cell_["group"]
}

#' Generate samples from fitted bru models
#'
#' @description
#'
#' Generic function for sampling for fitted models. The function invokes
#' particular methods which depend on the class of the first argument.
#'
#' @name generate
#' @export
#' @family sample generators
#' @param object a fitted model.
#' @param \dots additional arguments affecting the samples produced.
#' @return The form of the value returned by `generate()` depends on the data
#' class and prediction formula. Normally, a data.frame is returned, or a list
#' of data.frames (if the prediction formula generates a list)
#' @examples
#' \donttest{
#' if (
#'   bru_safe_inla() &&
#'     requireNamespace("sn", quietly = TRUE)
#' ) {
#'   # Generate data for a simple linear model
#'
#'   input.df <- data.frame(x = cos(1:10))
#'   input.df <- within(
#'     input.df,
#'     {
#'       y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
#'     }
#'   )
#'
#'   # Fit the model
#'
#'   fit <- bru(
#'     y ~ xeff(main = x, model = "linear"),
#'     family = "gaussian",
#'     data = input.df
#'   )
#'   summary(fit)
#'
#'   # Generate samples for some predefined x
#'
#'   df <- data.frame(x = seq(-4, 4, by = 0.1))
#'   smp <- generate(fit, df, ~ xeff + Intercept, n.samples = 10)
#'
#'   # Plot the resulting realizations
#'
#'   plot(df$x, smp[, 1], type = "l")
#'   for (k in 2:ncol(smp)) {
#'     points(df$x, smp[, k], type = "l")
#'   }
#'
#'   # We can also draw samples form the joint posterior
#'
#'   df <- data.frame(x = 1)
#'   smp <- generate(fit, df, ~ data.frame(xeff, Intercept), n.samples = 10)
#'   smp[[1]]
#'
#'   # ... and plot them
#'   if (require(ggplot2, quietly = TRUE)) {
#'     plot(do.call(rbind, smp))
#'   }
#' }
#' }
#'
generate <- function(object, ...) {
  UseMethod("generate")
}

#' Methods for bru_info objects
#'
#' The `bru_info` class is used to store metadata about `bru` models.
#'
#' @export
#' @rdname bru_info
bru_info <- function(...) {
  UseMethod("bru_info")
}

#' @export
#' @param method character; The type of estimation method used
#' @param \dots Additional arguments to be stored in the `bru_info` object. For
#' `summary` and `print` methods, arguments passed on to submethods.
#' @param inlabru_version character; inlabru package version. Default: NULL, for
#' automatically detecting the version
#' @param INLA_version character; INLA package version. Default: NULL, for
#' automatically detecting the version
#' @describeIn bru_info Create a `bru_info` object
bru_info.character <- function(method,
                               ...,
                               inlabru_version = NULL,
                               INLA_version = NULL) {
  if (is.null(inlabru_version)) {
    inlabru_version <-
      tryCatch(
        expr = {
          getNamespaceVersion("inlabru")
        },
        error = function(e) {
          "unknown"
        }
      )
  }
  if (is.null(INLA_version)) {
    INLA_version <-
      tryCatch(
        expr = {
          getNamespaceVersion("INLA")
        },
        error = function(e) {
          "unknown"
        }
      )
  }
  object <- structure(
    list(
      method = method,
      ...,
      inlabru_version = inlabru_version,
      INLA_version = INLA_version
    ),
    class = "bru_info"
  )
  object
}

#' @export
#' @describeIn bru_info Extract the `bru_info` object from an estimated [bru()]
#'   result object. The default print method shows information about model
#'   components and observation models.
bru_info.bru <- function(object, ...) {
  object <- bru_check_object_bru(object)
  object[["bru_info"]]
}

#' @export
#' @describeIn bru_info Extract the `bru_info` object from an estimated [bru()]
#'   result object. The default print method shows information about model
#'   components and observation models.
as_bru_info <- function(object, ...) {
  UseMethod("as_bru_info")
}
#' @export
#' @describeIn bru_info Extract a `bru_info` object.
as_bru_info.bru_info <- function(object, ...) {
  object <- bru_info_upgrade(object)
  object
}
#' @export
#' @describeIn bru_info Extract the `bru_info` object from an estimated [bru()]
#'   result object.
as_bru_info.bru <- function(object, ...) {
  object <- bru_check_object_bru(object)
  object[["bru_info"]]
}


#' @title Extract INLA formula
#'
#' @description
#' Extracts the INLA formula used in a `bru` model
#'
#' @param x An object containing information about an INLA formula
#' @param \dots Additional arguments passed on to submethods
#' @export
#' @keywords internal
bru_inla_formula <- function(x, ...) {
  UseMethod("bru_inla_formula")
}
#' @rdname bru_inla_formula
#' @export
bru_inla_formula.bru <- function(x, ...) {
  bru_inla_formula(as_bru_info(x), ...)
}
#' @rdname bru_inla_formula
#' @export
bru_inla_formula.bru_info <- function(x, ...) {
  formula <- x[["model"]][["formula"]]
  if (is.null(formula)) {
    components <- x[["model"]][["effects"]]
    used <- bru_used(x)
    included <- union(used$effect, used$latent)
    formula <- bru_inla_formula(components[included])
    formula <- update.formula(formula, BRU_response ~ .)
  }
  formula
}

#' @rdname bru_inla_formula
#' @export
bru_inla_formula.bru_comp_list <- function(x, ...) {
  formula <- . ~ 0
  for (cmp in seq_along(x)) {
    formula <- update.formula(formula, x[[cmp]]$inla.formula)
  }
  environment(formula) <- environment(x)
  formula
}


#' @rdname bru_info
#' @param object A `bru_info` object
#' @param verbose logical; If `TRUE`, include more details of the
#' component definitions. If `FALSE`, only show basic component
#' definition information. Default: `FALSE`
#' @export
#' @method summary bru_info
summary.bru_info <- function(object, verbose = TRUE, ...) {
  structure(
    list(
      inlabru_version = object[["inlabru_version"]],
      INLA_version = object[["INLA_version"]],
      components = summary(object[["model"]], verbose = verbose, ...),
      lhoods = summary(as_bru_obs_list(object), verbose = verbose, ...)
    ),
    class = "summary_bru_info"
  )
}

#' @export
#' @param x An  object to be printed
#' @rdname bru_info
print.summary_bru_info <- function(x, ...) {
  cat(glue("inlabru version: {x$inlabru_version}"), "\n")
  cat(glue("INLA version: {x$INLA_version}"), "\n")
  print(x$components)
  if (!is.null(x$lhoods)) {
    print(x$lhoods)
  }
  invisible(x)
}

#' @export
#' @rdname bru_info
print.bru_info <- function(x, ...) {
  print(summary(x))
  invisible(x)
}


#' @title Extract timing information from fitted [bru] object
#' @description
#' Extracts a data.frame or tibble with information about the `Time` (CPU),
#' `System`, and `Elapsed` time for each step of a `bru()` run.
#' @param object A fitted `bru` object
#' @param \dots unused
#' @export
bru_timings <- function(object, ...) {
  UseMethod("bru_timings")
}

#' @export
#' @rdname bru_timings
bru_timings.bru <- function(object, ...) {
  object <- bru_check_object_bru(object)
  object[["bru_timings"]]
}


# @param args Either a [bru_obs()] object, a [bru_obs_list()] object, or
#   a list of one or more [bru_obs()] or [bru_obs_list()] objects,
#   or a list of arguments for a call to [bru_obs()].
# Returns: A bru_obs_list object.
bru_obs_list_construct <- function(args, options, .envir = parent.frame(),
                                   .response = . ~ .,
                                   .components = NULL) {
  lhoods <- args
  if (inherits(lhoods, c("bru_obs", "bru_obs_list"))) {
    lhoods <- list(lhoods)
  }
  dot_is_lhood <- vapply(
    lhoods,
    function(lh) inherits(lh, "bru_obs"),
    TRUE
  )
  dot_is_lhood_list <- vapply(
    lhoods,
    function(lh) inherits(lh, "bru_obs_list"),
    TRUE
  )
  if (any(dot_is_lhood | dot_is_lhood_list)) {
    if (!all(dot_is_lhood | dot_is_lhood_list)) {
      stop(paste0(
        "Cannot mix `bru_obs()` parameters with `bru_obs` ",
        "and `bru_obs_list` objects.",
        "\n  Check if the argument(s) ",
        glue_collapse(
          glue("'{names(lhoods)[!(dot_is_lhood | dot_is_lhood_list)]}'"),
          sep = ", ", last = ", and "
        ),
        " were meant to be given to a call to 'bru_obs()',\n  or in the ",
        "'options' list argument instead."
      ))
    }
  } else {
    if (is.null(lhoods[["formula"]])) {
      lhoods[["formula"]] <- . ~ .
    }
    lhoods[["formula"]] <- auto_response(
      lhoods[["formula"]],
      .response
    )
    lhoods <- list(do.call(
      bru_obs,
      c(
        lhoods,
        list(
          options = options,
          .envir = .envir
        )
      )
    ))
    dot_is_lhood <- TRUE
    dot_is_lhood_list <- FALSE
  }
  lhoods <- do.call(c, lhoods[dot_is_lhood | dot_is_lhood_list])

  # Update include/exclude information to limit it to existing components
  # and check additivity
  lhoods <- bru_used_update(lhoods, labels = names(.components))

  lhoods
}


#' @title Convenient model fitting using (iterated) INLA
#'
#' @description This method is a wrapper for `INLA::inla` and provides
#'   multiple enhancements.
#'
#'   \itemize{
#'     \item
#'       Easy usage of spatial covariates and automatic construction of inla
#'       projection matrices for (spatial) SPDE models. This feature is
#'       accessible via the `components` parameter. Practical examples on how to
#'       use spatial data by means of the components parameter can also be found
#'       by looking at the [lgcp()] function's documentation.
#'     \item
#'       Constructing multiple observation models is straightforward. See
#'       [bru_obs()] for more information on how to provide additional
#'       models to `bru` using the `...` parameter list.
#'     \item
#'       Support for non-linear predictors. See example below.
#'     \item
#'       Log Gaussian Cox process (LGCP) inference is
#'       available by using the `"cp"` family or (even easier) by using the
#'       [lgcp()] function.
#'   }
#' @export
#'
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#'
#' @param components Latent component definitions, either as a [bru_comp_list()]
#'   object, or a `formula`-like specification.
#'   Also used to define a default linear additive predictor.  See
#'   [bru_comp()] for details.
#' @param \dots Observation models, each constructed by a calling [bru_obs()],
#'   or [bru_obs_list()].
#'
#'   Alternatively, for backwards compatibility, may be named parameters that
#'   can be passed to a single [bru_obs()] call. These arguments will be
#'   evaluated before calling [bru_obs()], in order to detect if they already
#'   are `bru_obs` objects. This means that special arguments that are only
#'   available in the context of `data` or `response_data` (such as `Ntrials`)
#'   will only work properly in direct calls to [bru_obs()].
#' @param .envir Environment for component evaluation (for when a non-formula
#' specification is used)
#' @param options A [bru_options] options object or a list of options passed
#' on to [bru_options()]
#'
#' @return bru returns an object of class "bru". A `bru` object inherits
#'   from `INLA::inla` (see the inla documentation for its properties) and
#'   adds additional information stored in the `bru_info` field.
#'
#' @examples
#' \donttest{
#' if (bru_safe_inla()) {
#'   # Simulate some covariates x and observations y
#'   input.df <- data.frame(x = cos(1:10))
#'   input.df <- within(input.df, {
#'     y <- 5 + 2 * x + rnorm(10, mean = 0, sd = 0.1)
#'   })
#'
#'   # Fit a Gaussian likelihood model
#'   fit <- bru(y ~ x + Intercept(1), family = "gaussian", data = input.df)
#'
#'   # Obtain summary
#'   fit$summary.fixed
#' }
#'
#'
#' if (bru_safe_inla()) {
#'   # Alternatively, we can use the bru_obs() function to construct the
#'   # likelihood:
#'
#'   lik <- bru_obs(
#'     family = "gaussian",
#'     formula = y ~ x + Intercept,
#'     data = input.df
#'   )
#'   fit <- bru(~ x + Intercept(1), lik)
#'   fit$summary.fixed
#' }
#'
#' # An important addition to the INLA methodology is bru's ability to use
#' # non-linear predictors. Such a predictor can be formulated via bru_obs()'s
#' # \code{formula} parameter. The z(1) notation is needed to ensure that
#' # the z component should be interpreted as single latent variable and not
#' # a covariate:
#'
#' if (bru_safe_inla()) {
#'   z <- 2
#'   input.df <- within(input.df, {
#'     y <- 5 + exp(z) * x + rnorm(10, mean = 0, sd = 0.1)
#'   })
#'   lik <- bru_obs(
#'     family = "gaussian", data = input.df,
#'     formula = y ~ exp(z) * x + Intercept
#'   )
#'   fit <- bru(~ z(1) + Intercept(1), lik)
#'
#'   # Check the result (z posterior should be around 2)
#'   fit$summary.fixed
#' }
#' }
#'
bru <- function(components = ~ Intercept(1),
                ...,
                options = list(),
                .envir = parent.frame()) {
  stopifnot(bru_safe_inla(multicore = TRUE))

  # Update default options
  options <- bru_call_options(options)
  bru_options_set_local(options, .reset = TRUE)

  bru_log_bookmark("bru")
  bru_log_message("bru: Preprocessing", verbosity = 1L)

  timing <- bru_timer_do(NULL, "Preprocess", 0L)

  # Turn model components and bru_obs objects into internal bru model
  bru.model <- bru_model(
    components,
    lhoods = list(...),
    options = options,
    .envir = .envir
  )

  # Set max iterations to 1 if all likelihood formulae are linear
  if (all(vapply(bru.model[["lhoods"]], bru_is_linear, TRUE))) {
    options$bru_max_iter <- 1
  }

  # Until the storage format is changed in a backwards incompatible way,
  # move lhoods and inputs out of bru.model:
  lhoods <- bru.model[["lhoods"]]
  inputs <- bru.model[["inputs"]]
  bru.model[["lhoods"]] <- NULL
  bru.model[["inputs"]] <- NULL

  info <- bru_info(
    method = "bru",
    model = bru.model,
    inputs = inputs,
    lhoods = lhoods,
    log = bru_log()["bru"],
    options = options
  )

  timing <- bru_timer_done(timing)

  # Run iterated INLA
  if (options$bru_run) {
    result <- iinla(
      model = info[["model"]],
      lhoods = as_bru_obs_list(info),
      inputs = info[["inputs"]],
      options = info[["options"]]
    )
  } else {
    result <- list()
  }

  result$bru_timings <-
    rbind(
      timing,
      result[["bru_iinla"]][["timings"]]
    )

  # Add bru information to the result
  result$bru_info <- info
  class(result) <- c("bru", class(result))
  result
}


#' @describeIn bru
#' Continue the optimisation from a previously computed estimate. The estimation
#' `options` list can be given new values to override the original settings.
#'
#' To rerun with a subset of the data (e.g. for cross validation or prior
#' sampling), use [bru_set_missing()] to set all or part of the response data
#' to `NA` before calling `bru_rerun()`.
#' @param result A previous estimation object of class `bru`
#'
#' @export
bru_rerun <- function(result, options = list()) {
  result <- bru_check_object_bru(result)

  info <- result[["bru_info"]]
  info[["options"]] <- bru_call_options(
    bru_options(
      info[["options"]],
      as.bru_options(options)
    )
  )
  bru_options_set_local(info[["options"]], .reset = TRUE)

  orig_timings <- result[["bru_timings"]]

  result <- iinla(
    model = info[["model"]],
    lhoods = as_bru_obs_list(info),
    inputs = info[["inputs"]],
    initial = result,
    options = info[["options"]]
  )

  new_timings <- result[["bru_iinla"]][["timings"]]$Iteration >
    max(orig_timings$Iteration)
  result$bru_timings <-
    rbind(
      orig_timings,
      result[["bru_iinla"]][["timings"]][new_timings, , drop = FALSE]
    )

  # Add bru information to the result
  result$bru_info <- info
  class(result) <- c("bru", class(result))
  result
}

#' @title Set missing values in observation models
#'
#' @description Set all or parts of the observation model response data
#' to `NA`, for example for use in cross validation (with [bru_rerun()])
#' or prior sampling (with [bru_rerun()] and [inlabru::generate()], but see
#' "Prior sampling caveats" below).
#'
#' @param object A `bru`, `bru_obs` or `bru_obs_list` object
#' @param keep For `bru_obs`, a single logical or an integer vector;
#'   If `TRUE`, keep all the response data, if `FALSE` (default),
#'   set all of it to `NA`. An integer vector determines which elements
#'   to keep (for positive values) or to set as missing (negative values).
#'
#'   For `bru` and `bru_obs_list`, a logical scalar or vector, or a list, see
#'   Details.
#' @param \dots Additional arguments passed on to the `bru_obs` and individual
#'   data class methods.  Currently unused.
#'
#' @details For `bru` and `bru_obs_list`,
#' \itemize{
#'   \item{`keep` must be either a single logical, which is expanded to a list,}
#'   \item{a logical vector, which is converted to a list,}
#'   \item{an unnamed list of the same length as the number of observation
#'     models, with elements compatible with the `bru_obs` method, or}
#'   \item{a named list with elements compatible with the `bru_obs` method,
#'     and only the named `bru_obs` models are acted upon, i.e. the elements
#'     not present in the list are treated as `keep = TRUE`.}
#' }
#'
#' E.g.: `keep = list(b = FALSE)` sets all observations in model `b` to missing,
#' and does not change model `a`.
#'
#' E.g.: `keep = list(a = 1:4, b = -(3:5))` keeps only observations `1:4` of
#' model `a`, marking the rest as missing, and sets observations `3:5` of model
#' `b` to missing.
#'
#' @section Prior sampling caveats:
#' Note that prior sampling requires special care for hyperparameters, as the
#' prior modes are not typically useful; in the future, we plan to have a
#' dedicated method that samples from the hyperparameters, and then uses
#' ```
#' bru_rerun(
#'   bru_set_missing(...),
#'   options = list(
#'     control.mode = list(
#'       theta = theta_sample,
#'       fixed = TRUE)
#'     )
#'   )
#' ```
#' for each sample.
#'
#' @export
#' @rdname bru_set_missing
#' @examples
#' obs <- c(
#'   A = bru_obs(y_A ~ ., data = data.frame(y_A = 1:6)),
#'   B = bru_obs(y_B ~ ., data = data.frame(y_B = 11:15))
#' )
#' bru_response_size(obs)
#' lapply(
#'   bru_set_missing(obs, keep = FALSE),
#'   function(x) {
#'     x[["response_data"]][[x[["response"]]]]
#'   }
#' )
#' lapply(
#'   bru_set_missing(obs, keep = list(B = FALSE)),
#'   function(x) {
#'     x[["response_data"]][[x[["response"]]]]
#'   }
#' )
#' lapply(
#'   bru_set_missing(obs, keep = list(1:4, -(3:5))),
#'   function(x) {
#'     x[["response_data"]][[x[["response"]]]]
#'   }
#' )
#'
bru_set_missing <- function(object, keep = FALSE, ...) {
  UseMethod("bru_set_missing")
}
#' @export
#' @param x Object on which to apply `bru_set_missing()`
#' @param value Value to be passed as `keep` to `bru_set_missing()`
#' @describeIn bru_set_missing Setter method for `bru_set_missing()`
`bru_set_missing<-` <- function(x, ..., value) {
  object <- bru_set_missing(x, keep = value, ...)
  object
}
#' @export
#' @rdname bru_set_missing
bru_set_missing.bru <- function(object, keep = FALSE, ...) {
  object <- bru_check_object_bru(object)
  bru_set_missing(object[["bru_info"]], ...) <- keep
  object
}
#' @export
#' @rdname bru_set_missing
bru_set_missing.bru_info <- function(object, keep = FALSE, ...) {
  bru_set_missing(object[["lhoods"]], ...) <- keep
  object
}
#' @export
#' @rdname bru_set_missing
bru_set_missing.bru_obs_list <- function(object, keep = FALSE, ...) {
  if (is.logical(keep)) {
    if (length(keep) == 1L) {
      keep <- rep(list(keep), length(object))
    } else {
      keep <- as.list(keep)
    }
  }
  if (is.null(names(keep))) {
    idxs <- seq_along(object)
    if (length(keep) != length(object)) {
      stop(paste0(
        "When `keep` is not named list or vector, it must have ",
        "length matching the number of observation models."
      ))
    }
  } else {
    idxs <- intersect(names(keep), names(object))
    if (length(setdiff(names(keep), names(object))) > 0) {
      stop(paste0(
        "Some observation models in `keep` were not found: ",
        glue_collapse(glue("'{setdiff(names(keep), names(object))}'"), ", ")
      ))
    }
  }
  for (idx in idxs) {
    bru_set_missing(object[[idx]], ...) <- keep[[idx]]
  }
  object
}
#' @export
#' @rdname bru_set_missing
bru_set_missing.bru_obs <- function(object, keep = FALSE, ...) {
  if (is.null(keep) || isTRUE(keep)) {
    return(object)
  }
  if (isFALSE(keep)) {
    keep <- -seq_len(bru_response_size(object))
  }
  bru_set_missing(
    object[["response_data"]][[object[["response"]]]],
    ...
  ) <- keep
  object
}

#' @export
#' @describeIn bru_set_missing From `> 2.13.0`, set missing values in any
#' object supporting [base::is.na<-()] for positive and negative indices.
bru_set_missing.default <- function(object, keep = FALSE, ...) {
  if (is.null(keep) || isTRUE(keep)) {
    return(object)
  }
  if (isFALSE(keep)) {
    keep <- -seq_len(bru_response_size(object))
  }
  is.na(object) <- -keep
  object
}

#' @export
#' @describeIn bru_set_missing From `> 2.13.0`, handles `data.frame`, tibbles,
#'   including `inla.mdata`.
#' @examplesIf bru_safe_inla()
#' (obs <- INLA::inla.mdata(y = 1:4, X = matrix(1:8, 4, 2)))
#' bru_set_missing(obs, keep = c(1, 4))
#' bru_set_missing(obs) <- -(1:2)
#' obs
#'
bru_set_missing.data.frame <- function(object, keep = FALSE, ...) {
  if (is.null(keep) || isTRUE(keep)) {
    return(object)
  }
  if (isFALSE(keep)) {
    keep <- -seq_len(bru_response_size(object))
  }
  object[-keep, ] <- NA
  object
}

#' @export
#' @describeIn bru_set_missing From `> 2.13.0`, handles `inla.surv`.
#' @examplesIf bru_safe_inla()
#' (obs <- INLA::inla.surv(time = 1:4, event = c(1, 0, 1, 0)))
#' bru_set_missing(obs, keep = c(1, 4))
#'
#' (obs <- INLA::inla.surv(
#'   time = 1:4,
#'   event = c(1, 0, 1, 0),
#'   cure = matrix(1:8, 4, 2)
#' ))
#' bru_set_missing(obs, keep = c(1, 4))
#'
#' (obs <- INLA::inla.surv(
#'   time = 1:4,
#'   event = c(1, 0, 1, 0),
#'   subject = c(1, 1, 2, 1)
#' ))
#' bru_set_missing(obs, keep = c(1, 4))
#'
bru_set_missing.inla.surv <- function(object, keep = FALSE, ...) {
  if (is.null(keep) || isTRUE(keep)) {
    return(object)
  }
  if (isFALSE(keep)) {
    keep <- -seq_len(bru_response_size(object))
  }
  cls <- class(object)
  if (!is.data.frame(object)) {
    # TODO: This block should be function for converting list-`inla.surv` to
    # `data.frame`-`inla.surv`, but inla.surv might standardise
    # to tibble or data.frame, so we should remove this when possible.
    dat <- object
    att <- attributes(dat)
    cure_null <- ("cure" %in% names(dat)) && is.null(dat[["cure"]])
    if (cure_null) {
      dat["cure"] <- NULL
    }
    dat <- tibble::as_tibble(unclass(dat))
    attr(dat, "names.ori") <- att[["names.ori"]]
    class(dat) <- c("inla.surv", class(dat))
    object <- dat
  }
  # Note: by only setting time,lower,upper to NA, printing of the object
  # still works without giving an NA indexing error.
  if ("subject" %in% names(object)) {
    object$time[-keep] <- NA
  } else {
    object$time[-keep] <- NA
    object$lower[-keep] <- NA
    object$upper[-keep] <- NA
  }
  object <- as.list(object)
  class(object) <- c("inla.surv", class(object))
  object
}

#' Parse inclusion of component labels in a predictor expression
#' @param thenames Set of labels to restrict
#' @param include Character vector of component labels that are needed by the
#'   predictor expression; Default: NULL (include all components that are not
#'   explicitly excluded)
#' @param exclude Character vector of component labels that are not used by the
#'   predictor expression. The exclusion list is applied to the list
#'   as determined by the `include` parameter; Default: NULL (do not remove
#'   any components from the inclusion list)
#' @keywords internal
parse_inclusion <- function(thenames, include = NULL, exclude = NULL) {
  if (!is.null(include)) {
    intersect(thenames, setdiff(include, exclude))
  } else {
    setdiff(thenames, exclude)
  }
}

bru_data_mask <- function(data,
                          pronouns = NULL,
                          objects = NULL) {
  if (is.null(pronouns) || is.null(objects)) {
    nms <- setdiff(names(data), "")
    if (is.null(pronouns)) {
      pronouns <- nms
    }
    if (is.null(objects)) {
      objects <- nms
    }
  }
  data_orig <- data
  data <- lapply(data, function(x) {
    if (!is.null(x) && !is.list(x)) {
      x <- tibble::as_tibble(x)
    }
    x
  })
  top_envir <- rlang::new_environment()
  if (length(objects) > 0) {
    dot2_nms <- paste0(".", objects, ".")
    names(dot2_nms) <- objects
    for (nm in objects) {
      if (is.null(data_orig[[nm]])) {
        next
      }
      assign(dot2_nms[nm], data_orig[[nm]], envir = top_envir)
    }
  }
  bottom_envir <- top_envir
  for (k in rev(seq_along(data))) {
    if (is.null(data[[k]])) {
      next
    }
    bottom_envir <- rlang::new_environment(data[[k]], parent = bottom_envir)
  }
  mask <- rlang::new_data_mask(bottom_envir, top_envir)
  if (length(pronouns) > 0) {
    dot_nms <- paste0(".", pronouns)
    names(dot_nms) <- pronouns
    for (nm in pronouns) {
      if (is.null(data[[nm]])) {
        next
      }
      mask[[dot_nms[nm]]] <- rlang::as_data_pronoun(data[[nm]])
    }
  }
  mask[[".mask"]] <- rlang::as_data_pronoun(mask)
  class(mask) <- c("bru_data_mask", class(mask))
  mask
}

#' @title Evaluate expressions in data contexts
#' @description Evaluate an expression in a series of data contexts, also making
#'   the objects directly available as names surrounded by ".", stopping when
#'   the expression evaluation completes with no error, as well as `tidy`
#'   evaluation pronouns, e.g. `.data`.
#'
#'   This is an internal inlabru method, not intended for general use.
#' @param input An expression to be evaluated
#' @param data list of data objects in priority order. Named elements will be
#'   available as whole objects `.name.` as well as pronouns `.name` in the
#'   evaluation.
#' @param default Value used if the expression is evaluated as NULL. Default
#' NULL
#' @param .envir The evaluation environment
#' @return The result of expression evaluation
#' @keywords internal
#' @examples
#' # The A values come from the 'data' element, and the B values come from
#' # the 'response_data' element, as that is listed first.
#' bru_eval_in_data_context(
#'   list(A = .data$x, B = x),
#'   list(
#'     response_data = tibble::tibble(x = 1:5),
#'     data = tibble::tibble(x = 1:10)
#'   )
#' )
#' # Both A and B come from the 'data' element, as 'x' is found there,
#' # and the data objects are listed in order of precedence.
#' bru_eval_in_data_context(
#'   list(A = .data$x, B = x),
#'   list(
#'     data = tibble::tibble(x = 1:10),
#'     response_data = tibble::tibble(x = 1:5)
#'   )
#' )
#'
#' @export
bru_eval_in_data_context <- function(input,
                                     data = NULL,
                                     default = NULL,
                                     .envir = parent.frame()) {
  input <- rlang::enquo(input)
  deparse_input <- rlang::expr_text(input)
  if (inherits(data, "bru_data_mask")) {
    mask <- data
  } else {
    mask <- bru_data_mask(data)
  }
  success <- FALSE
  result <- NULL
  result <- try(
    rlang::eval_tidy(input, data = mask, env = .envir),
    silent = TRUE
  )
  success <- !inherits(result, "try-error")
  if (!success) {
    stop(glue::glue(
      "Input '{glue::glue_collapse(deparse_input, sep = '\n')}' could ",
      "not be evaluated."
    ))
  }
  if (is.null(result)) {
    result <- default
  }

  result
}


complete_coordnames <- function(data_coordnames, ips_coordnames) {
  new_coordnames <- character(
    max(
      length(ips_coordnames),
      length(data_coordnames)
    )
  )
  from_data <- which(!(data_coordnames %in% ""))
  new_coordnames[from_data] <- data_coordnames[from_data]
  from_ips <- setdiff(
    which(!(ips_coordnames %in% "")),
    from_data
  )
  new_coordnames[from_ips] <- ips_coordnames[from_ips]

  dummies <- seq_along(new_coordnames)
  dummies <- setdiff(dummies, c(from_data, from_ips))

  new_coordnames[dummies] <-
    glue("BRU_dummy_coordinate_{seq_along(new_coordnames)}")[dummies]

  list(
    data = new_coordnames[seq_along(data_coordnames)],
    ips = new_coordnames[seq_along(ips_coordnames)]
  )
}


# Extend bind_rows to handle XY/XYZ mismatches in one or more sfc columns
# TODO: detect non-point-data and either warn/stop or handle it properly.
# E.g. POLYGON + POINT data combinations currently fail with obscure error.
extended_bind_rows <- function(...) {
  dt <- list(...)
  names_ <- lapply(dt, names)
  is_sfc_ <- lapply(
    dt,
    function(data) {
      vapply(
        data,
        function(x) {
          inherits(x, "sfc")
        },
        TRUE
      )
    }
  )
  sfc_names_ <- unique(unlist(names_)[unlist(is_sfc_)])

  for (nm in sfc_names_) {
    # Which data objects have this column?
    sf_data_idx_ <-
      which(vapply(dt, function(data) !is.null(data[[nm]]), TRUE))

    # Unify CRS
    if (!inherits(dt[[sf_data_idx_[1]]][[nm]], "sfc")) {
      stop(glue(
        "Column '{nm}' is not a simple feature column in all data objects."
      ))
    }
    the_crs <- fm_crs(dt[[sf_data_idx_[1]]][[nm]])
    for (i in sf_data_idx_) {
      if (!inherits(dt[[i]][[nm]], "sfc")) {
        stop(glue(
          "Column '{nm}' is not a simple feature column in all data objects."
        ))
      }
      dt_crs <- fm_crs(dt[[i]][[nm]])
      if (!fm_crs_is_identical(dt_crs, the_crs)) {
        dt[[i]][[nm]] <- fm_transform(dt[[i]][[nm]], crs = the_crs)
      }
    }

    # Individual elements in sf columns may have different XY/XYZ properties,
    # so need to find out if any of them have Z or M, and then extend all others
    # to have Z too. M is handled by fmesher::fm_zm, but not sf::st_zm
    dt_nm <- lapply(dt[sf_data_idx_], function(x) x[[nm]])
    dt_nm <- fmesher::fm_zm(dt_nm)
    for (i in seq_along(sf_data_idx_)) {
      dt[sf_data_idx_[i]][[nm]] <- dt_nm[[i]][[nm]]
    }
  }

  # Technically, we only need this if at least one of the objects is a tibble
  dt <- lapply(dt, function(object) {
    if (!tibble::is_tibble(object)) {
      if (inherits(object, "sf")) {
        # Convert to tibble, preserving geometry column
        geometry_name <- attr(object, "sf_column", exact = TRUE)
        object <- tibble::as_tibble(object)
        object <- sf::st_as_sf(object, sf_column_name = geometry_name)
      } else {
        object <- tibble::as_tibble(object)
      }
    }
    object
  })
  result <- do.call(dplyr::bind_rows, dt)
  if (length(sfc_names_) > 0) {
    result <- sf::st_as_sf(result)
  }

  result
}


bru_get_parse_data <- function(x) {
  withr::with_options(
    list(keep.parse.data = TRUE),
    {
      utils::getParseData(parse(text = x, keep.source = TRUE))
    }
  )
}

# Returns TRUE if any argument is a Spatial* object, and gives deprecation
# messages.
check_sp_data_deprecation <- function(...) {
  .caller <- sys.call(-1)[[1]]
  if (startsWith(deparse1(.caller)[1], "function")) {
    .caller <- NULL
  }
  obj <- list(...)
  nms <- names(obj)
  result <- vapply(
    nms,
    function(nm) {
      if (inherits(obj[[nm]], "Spatial")) {
        if (!is.null(.caller)) {
          lifecycle::deprecate_warn(
            "2.12.0.9023",
            as.character(glue::glue(
              "{deparse1(.caller)}({nm} = ",
              "'has deprecated support for `Spatial` input')"
            )),
            I("`sf` input")
          )
        } else {
          lifecycle::deprecate_warn(
            "2.12.0.9023",
            I(as.character(glue::glue(
              "`{nm}` argument has deprecated support for `Spatial` input;"
            ))),
            I("`sf` input")
          )
        }
        TRUE
      } else {
        FALSE
      }
    },
    logical(1)
  )
  invisible(any(result))
}


bru_agg_data <- function(
  data,
  ips = NULL,
  domain = NULL,
  samplers = NULL,
  options,
  .envir
) {
  if (is.null(ips)) {
    if (!is.null(domain)) {
      ips <- fm_int(
        domain = domain,
        samplers = samplers,
        int.args = options[["bru_int_args"]]
      )
    }
  } else {
    if (is.null(data)) {
      data <- ips
      ips <- NULL
    }
  }
  if (!is.null(ips)) {
    if (!is.null(data)) {
      ips$.block <- as.integer(ips$.block)
      if (".block" %in% names(data)) {
        ips <- dplyr::left_join(ips, data, by = ".block")
      } else {
        if (NROW(data) != max(ips$.block)) {
          stop(glue(
            "`data` has {NROW(data)} rows, but `ips` has ",
            "{max(ips$.block)} blocks. Cannot join."
          ))
        }
        ips <- dplyr::left_join(
          ips,
          dplyr::bind_cols(data, .block = seq_len(NROW(data))),
          by = ".block"
        )
      }
    }
    data <- ips
    ips <- NULL
  }

  data
}


bru_agg_input <- function(input, data_list, .envir) {
  aggregate_input <- bru_eval_in_data_context(
    {{ input }},
    data = data_list,
    default = NULL,
    .envir = .envir
  )
  if (is.null(aggregate_input)) {
    aggregate_input <- list()
  }
  if (is.null(aggregate_input[["block"]])) {
    aggregate_input[["block"]] <- bru_eval_in_data_context(
      .data[[".block"]],
      data = data_list,
      default = NULL,
      .envir = .envir
    )
  }
  if (is.null(aggregate_input[["weights"]])) {
    aggregate_input[["weights"]] <- bru_eval_in_data_context(
      .data[["weight"]],
      data = data_list,
      default = NULL,
      .envir = .envir
    )
  }
  if (is.null(aggregate_input[["n_block"]])) {
    agg_n_block_expr <- rlang::parse_expr(
      "bru_response_size(.response_data.)"
    )
    aggregate_input[["n_block"]] <- bru_eval_in_data_context(
      !!agg_n_block_expr,
      data = data_list,
      default = NULL,
      .envir = .envir
    )
  }
  lapply(c("block", "weights", "n_block"), function(nm) {
    if (is.null(aggregate_input[[nm]])) {
      bru_log_abort(
        glue::glue(
          "Aggregation requested, but `aggregate_input[['{nm}']]` ",
          "evaluates to NULL."
        )
      )
    }
  })

  if (!is.null(aggregate_input[["block_response"]])) {
    blk_resp <- aggregate_input[["block_response"]]
    if (is.null(data_list$response_data[[blk_resp]])) {
      msg <- glue::glue(
        '`block_response` variable "{blk_resp}" ',
        "not found in `response_data`."
      )

      bru_log_abort(msg)
    }
  } else {
    aggregate_input[["block_response"]] <- ".block"
  }
  blk_resp <- aggregate_input[["block_response"]]
  if (!is.null(data_list$response_data[[blk_resp]])) {
    agg_block_resp_expr <- rlang::parse_expr(
      glue::glue('.response_data[["{blk_resp}"]]')
    )
    block_response <- bru_eval_in_data_context(
      !!agg_block_resp_expr,
      data = data_list,
      default = NULL,
      .envir = .envir
    )

    aggregate_input[["block"]] <-
      match(aggregate_input[["block"]], block_response)
  }

  if (is.character(aggregate_input[["block"]])) {
    msg <- paste0(
      "'character' aggregation block information detected.\n",
      "Please use a numeric or integer vector for the block information ",
      "instead.\n",
      "If you want to use a character vector, supply a `block_response`\n",
      "argument with the name of a response variableto match the character ",
      "vector to."
    )

    bru_log_abort(msg)
  }

  if (anyNA(aggregate_input[["block"]])) {
    msg <- paste0(
      "'NA' aggregation block information detected.",
      "Either the `block` information was out of range, or a match was not ",
      "found when using `.response_data$block_response`."
    )

    bru_log_abort(msg)
  }

  aggregate_input
}


bru_obs_agg <- function(lh,
                        aggregate = NULL,
                        aggregate_input = NULL,
                        options = list(),
                        .envir = parent.frame()) {
  if (!is.null(aggregate) && is.character(aggregate)) {
    aggregate <- switch(aggregate,
      "none" = NULL,
      bm_aggregate(type = aggregate)
    )
  }
  if (!is.null(aggregate)) {
    lh$data <- bru_agg_data(
      data = lh$data,
      ips = lh$integration_info$ips,
      domain = lh$integration_info$domain,
      samplers = lh$integration_info$samplers,
      options = options,
      .envir = .envir
    )
    lh$integration_info$ips <- NULL

    aggregate_input <- bru_agg_input(
      {{ aggregate_input }},
      data_list = list(
        data = lh$data,
        response_data = lh$BRU_original_response_data,
        extra_data = lh$data_extra
      ),
      .envir = .envir
    )

    pred_text <- bru_pred_expr(lh, format = "text_raw")
    pred_text <- glue(
      "{{ibm_eval(BRU_aggregate_mapper, input = BRU_aggregate_input,",
      " state = {{{pred_text}}})}}"
    )
    lh$pred_expr <- new_bru_pred_expr(
      pred_text,
      used = lh$pred_expr$used,
      is_rowwise = FALSE,
      .envir = lh$pred_expr$.envir
    )
    lh$pred_expr$is_additive <- FALSE
    lh$pred_expr$is_linear <- FALSE
    lh$data_extra[["BRU_aggregate_mapper"]] <- aggregate
    lh$data_extra[["BRU_aggregate_input"]] <- aggregate_input

    lh <- bru_compat_pre_2_14_bru_obs(lh)
  }

  lh
}


bru_obs_family_cp_sp <- function(lh, options, .envir) {
  if (!is.null(lh[["aggregate"]])) {
    stop("The 'aggregate' feature cannot be used with family='cp'.")
  }

  response <- lh[["response_data"]][[lh[["response"]]]]
  orig_response_data <- lh[["BRU_original_response_data"]]
  response_data <- lh[["response_data"]]
  data <- lh[["data"]]
  formula <- lh[["formula"]]
  domain <- lh[["integration_info"]][["domain"]]
  samplers <- lh[["integration_info"]][["samplers"]]
  ips <- lh[["integration_info"]][["ips"]]

  response_expr_text <- as.character(formula)[2]

  # Catch and handle special cases:
  if (is.null(response) || !inherits(response, "list")) {
    domain_names <- trimws(strsplit(response_expr_text, split = "\\+")[[1]])
    if (!is.null(domain_names)) {
      # "a + b" conversion to list(a = a, b = b)
      domain_expr <- paste0(
        "list(",
        paste0(
          vapply(
            domain_names, function(x) {
              if (identical(x, "coordinates")) {
                glue("{x} = sp::{x}(.data.)")
              } else {
                glue("{x} = {x}")
              }
            },
            ""
          ),
          collapse = ", "
        ),
        ")"
      )
      response_expr <- rlang::parse_expr(domain_expr)
    }
    response <- tryCatch(
      expr = bru_eval_in_data_context(
        !!response_expr,
        data = list(response_data = orig_response_data, data = data),
        default = NULL,
        .envir = .envir
      ),
      error = function(e) {
        NULL
      }
    )
  }

  if (is.null(response)) {
    stop(paste0(
      "You called bru_obs() with family='cp' but the evaluated ",
      "response information is NULL"
    ))
  }

  if (is.null(ips)) {
    if (is.null(domain)) {
      stop(paste0(
        "The family='cp' model requires a 'domain' specification compatible ",
        "with 'fmesher::fm_int()'"
      ))
    }
    if (!setequal(names(response), names(domain))) {
      stop(glue("
          Mismatch between response and domain names:
            names(response) = ({glue_collapse(names(response), sep = ', ')})
            names(domain)   = ({glue_collapse(names(domain), sep = ', ')})"))
    }

    ips <- fm_int(
      domain = domain,
      samplers = samplers,
      int.args = options[["bru_int_args"]]
    )
    if ((inherits(samplers, "Spatial") ||
      inherits(data, "Spatial") ||
      inherits(response[["coordinates"]], "Spatial")) &&
      inherits(ips, "sf")) {
      ips <- sf::as_Spatial(ips)
    }
  }

  if (length(unique(response_data[["BRU_E"]])) > 1) {
    bru_log_warn(
      "Exposure/effort parameter E should be a scalar for likelihood 'cp'."
    )
  }

  ips_is_Spatial <- inherits(ips, "Spatial")
  if (ips_is_Spatial) {
    bru_safe_sp(force = TRUE)
    ips_coordnames <- sp::coordnames(ips)
    ips_crs <- fm_CRS(ips)
    # For backwards compatibility:
    data_crs <- fm_CRS(data)

    if ("coordinates" %in% names(response)) {
      data_coordnames <- colnames(response$coordinates)
      new_coordnames <- complete_coordnames(data_coordnames, ips_coordnames)
      colnames(response$coordinates) <- new_coordnames$data
      sp::coordnames(ips) <- new_coordnames$ips
    } else {
      if (inherits(response, c("sf", "sfc")) ||
        (is.list(response) &&
          any(vapply(
            response,
            function(x) inherits(x, c("sf", "sfc")), TRUE
          )))) {
        ips <- sf::st_as_sf(ips)
        ips_is_Spatial <- FALSE
      }
    }
  }
  # TODO: check that the crs info is the same

  # For non-Spatial models:
  # Use the response data list as the actual data object, since that's now the
  # canonical place where the point information is given.  This also allows
  # response_data to be used when constructing the response list. This makes
  # it a strict requirement that the predictor can be evaluated as a pure
  # function of the domain data.  When implementing sf support, might be able
  # to give explicit access to spatial coordinates, but otherwise the user can
  # extract it from the geometry with st_coordinates(geometry)[,1] or similar.
  # For Spatial models, keep the old behaviour for backwards compatibility for
  # now, but can likely realign that in the future after more testing.
  # Save general response data to add to response (precomputed covariates etc)
  if (!is.null(orig_response_data)) {
    data_ <- orig_response_data
  } else {
    data_ <- data
  }
  if (inherits(data_, "Spatial")) {
    data_ <- tibble::as_tibble(as.data.frame(data_))
  }
  if (ips_is_Spatial) {
    if ("coordinates" %in% names(response)) {
      idx <- names(response) %in% "coordinates"
      data <- as.data.frame(response$coordinates)
      if (!all(idx)) {
        data <- dplyr::bind_cols(data, as.data.frame(response[!idx]))
      }
    } else {
      data <- as.data.frame(response)
    }
  } else {
    data <- tibble::as_tibble(response)
    if (("geometry" %in% names(data)) &&
      inherits(data$geometry, "sfc")) {
      sf::st_geometry(data) <- "geometry"
    }
  }
  orig_response_data <- NULL
  n_block <- max(ips$.block)
  if (".block" %in% names(data_)) {
    n_block <- max(max(data_$.block), n_block)
    N_data <- base::tabulate(data_$.block, n_block)
  } else {
    N_data <- NROW(data)
  }

  # Add back additional data
  additional_data_names <- setdiff(names(data_), names(data))
  if ((length(additional_data_names) > 0) &&
    (NROW(data_) == sum(N_data))) {
    data <- cbind(data, tibble::as_tibble(data_)[additional_data_names])
  }
  if (is.null(data[[".block"]])) {
    data$.block <- 1L
  }

  if (ips_is_Spatial) {
    ips <- as.data.frame(ips)
  } else {
    if ("geometry" %in% names(ips)) {
      sf::st_geometry(ips) <- "geometry"
    }
  }

  # Use 'weights' for per-point weighting of eta
  if (length(response_data[["BRU_weights"]]) == 1L) {
    point_weights <- rep(response_data[["BRU_weights"]], sum(N_data))
  } else {
    stopifnot(length(response_data[["BRU_weights"]]) == sum(N_data))
    point_weights <- response_data[["BRU_weights"]]
  }
  response_data[["BRU_weights"]] <- 1L

  if (identical(options[["bru_compress_cp"]], TRUE)) {
    new_response_data <- tibble::tibble(
      BRU_E = c(
        rep(0, length(N_data[N_data > 0])),
        response_data[["BRU_E"]] * ips[["weight"]]
      ),
      BRU_response = c(
        N_data[N_data > 0],
        rep(0, NROW(ips))
      ),
      BRU_block = c(
        which(N_data > 0),
        ips[[".block"]]
      ),
      BRU_weights = response_data[["BRU_weights"]][1],
      BRU_Ntrials = response_data[["BRU_Ntrials"]][1],
      BRU_scale = response_data[["BRU_scale"]][1]
    )

    lh$data_extra[["BRU_cp_block_subset"]] <- which(N_data > 0)
    lh$data_extra[["BRU_cp_n_block"]] <- n_block
    lh$data_extra[["BRU_cp_agg"]] <-
      bm_aggregate(
        type = "average",
        n_block = n_block
      )

    pred_text <- bru_pred_expr(lh, format = "text_raw")
    pred_text <- glue("
        {{
          BRU_eta <- {{
            {pred_text}
          }}
          BRU_eta <- rep_len(BRU_eta, length(BRU_aggregate))
          c(ibm_eval(
              BRU_cp_agg,
              list(block = .block[BRU_aggregate]),
              state = BRU_eta[BRU_aggregate] *
                BRU_point_weights[BRU_aggregate])[BRU_cp_block_subset],
            BRU_eta[!BRU_aggregate])
        }}")
    old_pred_expr <- bru_pred_expr(lh)
    lh$pred_expr <- new_bru_pred_expr(
      pred_text,
      used = lh$pred_expr$used,
      is_rowwise = FALSE,
      .envir = lh$pred_expr$.envir
    )
    lh$pred_expr$is_additive <- FALSE
    lh$pred_expr$is_linear <- old_pred_expr$is_linear

    data <- extended_bind_rows(
      dplyr::bind_cols(data,
        BRU_aggregate = TRUE,
        BRU_point_weights = point_weights
      ),
      dplyr::bind_cols(ips,
        BRU_aggregate = FALSE,
        BRU_point_weights = 0.0
      )
    )

    if (!is.null(lh$control.gcpo)) {
      group_cv_block <- c(which(N_data > 0), ips$.block)
      # Would like:
      # group_cv_friends <- lapply(seq_len(max(ips$.block)), function(i) {
      #   which(group_cv_block == i)
      # })
      # Current inla.group.cv interface (2025-08-28)
      group_cv_friends <- lapply(seq_along(group_cv_block), function(i) {
        which(group_cv_block == group_cv_block[i])
      })

      lh$control.gcpo <- modifyList(
        lh$control.gcpo,
        list(friends = group_cv_friends)
      )
    }
  } else {
    new_response_data <- data.frame(
      BRU_E = c(
        rep(0, sum(N_data)),
        response_data[["BRU_E"]] * ips[["weight"]]
      ),
      BRU_response = c(
        point_weights,
        rep(0, NROW(ips))
      ),
      BRU_block = c(
        data[[".block"]],
        ips[[".block"]]
      ),
      BRU_weights = response_data[["BRU_weights"]][1],
      BRU_Ntrials = response_data[["BRU_Ntrials"]][1],
      BRU_scale = response_data[["BRU_scale"]][1]
    )
    data <- extended_bind_rows(data, ips)
  }
  if (ips_is_Spatial) {
    non_coordnames <- setdiff(names(data), data_coordnames)
    data <- sp::SpatialPointsDataFrame(
      coords = as.matrix(data[data_coordnames]),
      data = data[non_coordnames],
      proj4string = fm_CRS(data_crs),
      match.ID = FALSE
    )
  }

  lh$data <- data
  lh$response_data <- new_response_data
  lh$response <- "BRU_response"
  lh$inla.family <- "poisson"
  lh$integration_info$ips <- NULL

  lh <- bru_compat_pre_2_14_bru_obs(lh)

  lh
}

#
# > x<-sample(1:1000,size=1000000,replace=TRUE)
# > bench::mark(R = convert_group_cv_blocks_to_friends_list(x, method = "R"),
#               C = convert_group_cv_blocks_to_friends_list(x, method = "C"))
# # A tibble: 2 × 13
# expression      min   median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc
# total_time result memory time           gc
# <bch:expr> <bch:tm> <bch:tm>     <dbl> <bch:byt>    <dbl> <int> <dbl>
# <bch:tm> <list> <list> <list>         <list>
# 1 R             4.23s    4.23s     0.236        NA    27.6      1   117
# 4.23s <list> <NULL> <bench_tm [1]> <tibble>
# 2 C            97.1ms  97.97ms    10.1          NA     1.68     6     1
# 595.38ms <list> <NULL> <bench_tm [6]> <tibble>
#
convert_group_cv_blocks_to_friends_list <- function(group_cv_block,
                                                    method = "C") {
  # Would like:
  # group_cv_friends <- lapply(seq_len(max(ips$.block)), function(i) {
  #   which(group_cv_block == i)
  # })
  # Current inla.group.cv interface (2025-08-28)
  # Inefficient implementation:
  # group_cv_friends <- lapply(seq_along(group_cv_block), function(i) {
  #   which(group_cv_block == group_cv_block[i])
  # })
  # Faster version:
  if (method == "R") {
    grp_index <- seq_len(max(group_cv_block))
    group_cv_friends_groups <- lapply(grp_index, function(i) {
      which(group_cv_block == i)
    })
    group_cv_friends <- lapply(group_cv_block, function(i) {
      group_cv_friends_groups[[i]]
    })
  } else if (method == "C") {
    group_cv_friends <-
      inlabru_group_cv_block_conversion(as.integer(group_cv_block),
        as.integer(max(group_cv_block)),
        per_node = as.logical(TRUE)
      )
  }
  group_cv_friends
}


bru_obs_family_cp <- function(lh, options, .envir) {
  if (inherits(lh[["data"]], "Spatial") ||
    inherits(lh[["BRU_original_response_data"]], "Spatial") ||
    inherits(lh[["integration_info"]][["samplers"]], "Spatial") ||
    inherits(lh[["integration_info"]][["ips"]], "Spatial")) {
    return(bru_obs_family_cp_sp(lh, options, .envir))
  }

  if (!is.null(lh[["aggregate"]])) {
    stop("The 'aggregate' feature cannot be used with family='cp'.")
  }

  response <- lh[["response_data"]][[lh[["response"]]]]
  orig_response_data <- lh[["BRU_original_response_data"]]
  response_data <- lh[["response_data"]]
  data <- lh[["data"]]
  domain <- lh[["integration_info"]][["domain"]]
  samplers <- lh[["integration_info"]][["samplers"]]
  ips <- lh[["integration_info"]][["ips"]]

  response_expr_text <- bru_pred_expr(lh, format = "resp_text")

  # Catch and handle special cases:
  if (is.null(response) || !inherits(response, "list")) {
    domain_names <- trimws(strsplit(response_expr_text, split = "\\+")[[1]])
    if (!is.null(domain_names)) {
      # "a + b" conversion to list(a = a, b = b)
      domain_expr <- paste0(
        "list(",
        paste0(
          vapply(
            domain_names, function(x) {
              if (identical(x, "coordinates")) {
                glue("{x} = sp::{x}(.data.)")
              } else {
                glue("{x} = {x}")
              }
            },
            ""
          ),
          collapse = ", "
        ),
        ")"
      )
      response_expr <- rlang::parse_expr(domain_expr)
    }
    response <- tryCatch(
      expr = bru_eval_in_data_context(
        !!response_expr,
        data = list(response_data = orig_response_data, data = data),
        default = NULL,
        .envir = .envir
      ),
      error = function(e) {
        NULL
      }
    )
  }

  if (is.null(response)) {
    stop(paste0(
      "You called bru_obs() with family='cp' but the evaluated ",
      "response information is NULL"
    ))
  }

  if (is.null(ips)) {
    if (is.null(domain)) {
      stop(paste0(
        "The family='cp' model requires a 'domain' specification compatible ",
        "with 'fmesher::fm_int()'"
      ))
    }
    if (!setequal(names(response), names(domain))) {
      stop(glue("
          Mismatch between response and domain names:
            names(response) = ({glue_collapse(names(response), sep = ', ')})
            names(domain)   = ({glue_collapse(names(domain), sep = ', ')})"))
    }

    ips <- fm_int(
      domain = domain,
      samplers = samplers,
      int.args = options[["bru_int_args"]]
    )
  }

  if (length(unique(response_data[["BRU_E"]])) > 1) {
    bru_log_warn(
      "Exposure/effort parameter E should be a scalar for likelihood 'cp'."
    )
  }

  # TODO: check that the crs info is the same

  # For non-Spatial models:
  # Use the response data list as the actual data object, since that's now the
  # canonical place where the point information is given.  This also allows
  # response_data to be used when constructing the response list. This makes
  # it a strict requirement that the predictor can be evaluated as a pure
  # function of the domain data.  When implementing sf support, might be able
  # to give explicit access to spatial coordinates, but otherwise the user can
  # extract it from the geometry with st_coordinates(geometry)[,1] or similar.
  # For Spatial models, keep the old behaviour for backwards compatibility for
  # now, but can likely realign that in the future after more testing.
  # Save general response data to add to response (precomputed covariates etc)
  if (!is.null(orig_response_data)) {
    data_ <- orig_response_data
  } else {
    data_ <- data
  }

  data <- tibble::as_tibble(response)
  if (("geometry" %in% names(data)) &&
    inherits(data$geometry, "sfc")) {
    sf::st_geometry(data) <- "geometry"
  }

  orig_response_data <- NULL
  n_block <- max(ips$.block)
  if (".block" %in% names(data_)) {
    n_block <- max(max(data_$.block), n_block)
    N_data <- base::tabulate(data_$.block, n_block)
  } else {
    N_data <- NROW(data)
  }

  # Add back additional data
  additional_data_names <- setdiff(names(data_), names(data))
  if ((length(additional_data_names) > 0) &&
    (NROW(data_) == sum(N_data))) {
    data <- cbind(data, tibble::as_tibble(data_)[additional_data_names])
  }
  if (is.null(data[[".block"]])) {
    data$.block <- 1L
  }

  if ("geometry" %in% names(ips)) {
    sf::st_geometry(ips) <- "geometry"
  }

  # Use 'weights' for per-point weighting of eta
  if (length(response_data[["BRU_weights"]]) == 1L) {
    point_weights <- rep(response_data[["BRU_weights"]], sum(N_data))
  } else {
    stopifnot(length(response_data[["BRU_weights"]]) == sum(N_data))
    point_weights <- response_data[["BRU_weights"]]
  }
  response_data[["BRU_weights"]] <- 1L

  if (identical(options[["bru_compress_cp"]], TRUE)) {
    new_response_data <- tibble::tibble(
      BRU_E = c(
        rep(0, length(N_data[N_data > 0])),
        response_data[["BRU_E"]] * ips[["weight"]]
      ),
      BRU_response = c(
        N_data[N_data > 0],
        rep(0, NROW(ips))
      ),
      BRU_block = c(
        which(N_data > 0),
        ips[[".block"]]
      ),
      BRU_weights = response_data[["BRU_weights"]][1],
      BRU_Ntrials = response_data[["BRU_Ntrials"]][1],
      BRU_scale = response_data[["BRU_scale"]][1]
    )

    lh$data_extra[["BRU_cp_block_subset"]] <- which(N_data > 0)
    lh$data_extra[["BRU_cp_n_block"]] <- n_block
    lh$data_extra[["BRU_cp_agg"]] <-
      bm_aggregate(
        type = "average",
        n_block = n_block
      )

    old_pred_expr <- bru_pred_expr(lh)
    pred_text <- bru_pred_expr(old_pred_expr, format = "text_raw")
    resp_text <- bru_pred_expr(old_pred_expr, format = "resp_text")
    lh$data_extra[["BRU_cp_predictor"]] <-
      function(eta, data, data_extra) {
        eta <- rep_len(eta, length(data$BRU_aggregate))
        c(
          ibm_eval(
            data_extra$BRU_cp_agg,
            list(block = data$.block[data$BRU_aggregate]),
            state = eta[data$BRU_aggregate] *
              data$BRU_point_weights[data$BRU_aggregate]
          )[data_extra$BRU_cp_block_subset],
          eta[!data$BRU_aggregate]
        )
      }
    pred_text <-
      glue("BRU_cp_predictor({{ {pred_text} }}, .data., .data_extra.)")
    lh$pred_expr <- new_bru_pred_expr(
      pred_text,
      used = lh$pred_expr$used,
      is_rowwise = FALSE,
      .envir = lh$pred_expr$.envir
    )
    lh$pred_expr$is_additive <- FALSE
    lh$pred_expr$is_linear <- old_pred_expr$is_linear
    lh$pred_expr$resp_text <- resp_text

    data <- extended_bind_rows(
      dplyr::bind_cols(data,
        BRU_aggregate = TRUE,
        BRU_point_weights = point_weights
      ),
      dplyr::bind_cols(ips,
        BRU_aggregate = FALSE,
        BRU_point_weights = 0.0
      )
    )
  } else {
    new_response_data <- data.frame(
      BRU_E = c(
        rep(0, length(point_weights)),
        response_data[["BRU_E"]] * ips[["weight"]]
      ),
      BRU_response = c(
        point_weights,
        rep(0, NROW(ips))
      ),
      BRU_block = c(
        data[[".block"]],
        ips[[".block"]]
      ),
      BRU_weights = response_data[["BRU_weights"]][1],
      BRU_Ntrials = response_data[["BRU_Ntrials"]][1],
      BRU_scale = response_data[["BRU_scale"]][1]
    )
    data <- extended_bind_rows(data, ips)
  }
  if (!is.null(lh$control.gcpo)) {
    group_cv_block <- new_response_data$BRU_block
    group_cv_friends <- convert_group_cv_blocks_to_friends_list(group_cv_block)
    lh$control.gcpo <- modifyList(
      lh$control.gcpo,
      list(friends = group_cv_friends)
    )
  }

  if (inherits(data, "sf")) {
    data <- fmesher::fm_zm(data)
  }
  lh$data <- data
  lh$response_data <- new_response_data
  lh$response <- "BRU_response"
  lh$inla.family <- "poisson"
  lh$integration_info$ips <- NULL

  lh <- bru_compat_pre_2_14_bru_obs(lh)

  lh
}


bru_obs_check_used_deprecation <- function(
  used,
  include,
  exclude,
  include_latent,
  expr_text,
  env = rlang::caller_env(),
  user_env = rlang::caller_env(2)
) {
  if (lifecycle::is_present(include) ||
    lifecycle::is_present(exclude) ||
    lifecycle::is_present(include_latent)) {
    if (lifecycle::is_present(include)) {
      bru_log_message(
        paste0(
          "The `include` argument of `bru_obs()` is deprecated ",
          "since inlabru 2.11.0 and will be ignored.\n\t",
          "If auto-detection doesn't work, use ",
          "`used = bru_used(effect = include)` instead."
        ),
        verbosity = 1L
      )
      lifecycle::deprecate_warn(
        "2.11.0",
        "bru_obs(include)",
        "bru_obs(used)",
        c(
          "The provided `include` value will be ignored.",
          "If auto-detection doesn't work, use `bru_used(effect = include)`"
        ),
        env = env,
        user_env = user_env
      )
    }
    if (lifecycle::is_present(exclude)) {
      bru_log_message(
        paste0(
          "The `exclude` argument of `bru_obs()` is deprecated ",
          "since inlabru 2.11.0 and may be ignored.\n\t",
          "If auto-detection doesn't work, use ",
          "`used = bru_used(effect_exclude = exclude)` instead."
        ),
        verbosity = 1L
      )
      lifecycle::deprecate_warn(
        "2.11.0",
        "bru_obs(exclude)",
        "bru_obs(used)",
        c(
          "The provided `exclude` value will be ignored.",
          paste0(
            "If auto-detection doesn't work, ",
            "use `bru_used(effect_exclude = exclude)`"
          )
        ),
        env = env,
        user_env = user_env
      )
    }
    if (lifecycle::is_present(include_latent)) {
      bru_log_message(
        paste0(
          "The `include_latent` argument of `bru_obs()` is deprecated ",
          "since inlabru 2.11.0 and will be ignored.\n\t",
          "If auto-detection doesn't work, use ",
          "`used = bru_used(latent = include_latent)` instead."
        ),
        verbosity = 1L
      )
      lifecycle::deprecate_warn(
        "2.11.0",
        "bru_obs(include_latent)",
        "bru_obs(used)",
        c(
          "The provided `include_latent` value will be ignored.",
          paste0(
            "If auto-detection doesn't work, ",
            "use `bru_used(latent = include_latent)`"
          )
        ),
        env = env,
        user_env = user_env
      )
    }
    if (is.null(used)) {
      used <- bru_used(expr_text)
    }
  }
  used
}

bru_obs_handle_is_rowwise <- function(pred_expr,
                                      is_rowwise = NULL,
                                      allow_combine = NULL,
                                      data,
                                      response_data,
                                      aggregate) {
  if (is.null(is_rowwise) && is.logical(allow_combine)) {
    is_rowwise <- !allow_combine
  }
  if (!is.logical(is_rowwise)) {
    if (!is.null(aggregate)) {
      is_rowwise <- FALSE
    } else if (!is.null(response_data)) {
      bru_log_warn(
        paste0(
          "Non-null response data supplied; ",
          "guessing is_rowwise=FALSE.",
          "\n  Specify is_rowwise explicitly to avoid this warning."
        )
      )
      is_rowwise <- FALSE
    } else if (is.list(data) && !is.data.frame(data)) {
      bru_log_warn(
        paste0(
          "Non data-frame list-like data supplied; ",
          "guessing is_rowwise=FALSE.\n",
          "  Specify is_rowwise explicitly to avoid this warning."
        )
      )
      is_rowwise <- FALSE
    } else {
      is_rowwise <- TRUE
    }
  }
  if (is.null(pred_expr$is_rowwise)) {
    pred_expr$is_rowwise <- is_rowwise
  } else {
    pred_expr$is_rowwise <- pred_expr$is_rowwise && is_rowwise
  }
  pred_expr
}


#' @title Observation model construction for usage with [bru()]
#'
#' @description Observation model construction for usage with [bru()].
#'
#' Note: Prior to version `2.12.0`, this function was called `like()`, and that
#' alias will remain for a while until examples etc have been updated and users
#' made aware of the change. The name change is to avoid issues with namespace
#' clashes, e.g. with `data.table::like()`, and also to signal that the function
#' defines observation models, not just likelihood functions.
#'
#' @rdname bru_obs
#' @export
#'
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com}
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#'
#' @param formula a `formula` where the right hand side is a general R
#'   expression defines the predictor used in the model.
#' @param family A string identifying a valid `INLA::inla` likelihood family.
#' The default is
#'   `gaussian` with identity link. In addition to the likelihoods provided
#'   by inla (see `names(INLA::inla.models()$likelihood)`)
#'   inlabru supports fitting latent Gaussian Cox
#'   processes via `family = "cp"`.
#'   As an alternative to [bru()], the [lgcp()] function provides
#'   a convenient interface to fitting Cox processes.
#' @param data Predictor expression-specific data, as a `data.frame`, `tibble`,
#'  or `sf`.  Since `2.12.0.9023`, deprecated support for
#' `SpatialPoints[DataFrame]` objects.
#' @param response_data Observation/response-specific data for models that need
#'   different size/format for inputs and response variables, as a `data.frame`,
#'   `tibble`, or `sf`. Since `2.12.0.9023`, deprecated support for
#'   `SpatialPoints[DataFrame]` objects.
#' @param data_extra object convertible with `as.list()` with additional
#'   variables to be made available in predictor evaluations. Variables with
#'   the same names as the data object will be ignored, unless accessed via
#'   `.data_extra.[["name"]]` or `.data_extra.$name` in the formula.
#' @param E Exposure/effort parameter for family = 'poisson' passed on to
#'   `INLA::inla`. Special case if family is 'cp': rescale all integration
#'   weights by a scalar `E`. For sampler specific reweighting/effort, use a
#'   `weight` column in the `samplers` object instead, see [fmesher::fm_int()].
#'   Default taken from `options$E`, normally `1`.
#' @param Ntrials A vector containing the number of trials for the 'binomial'
#'   likelihood. Default taken from `options$Ntrials`, normally `1`.
#' @param weights Fixed (optional) weights parameters of the likelihood, so the
#'   log-likelihood`[i]` is changed into `weights[i] * log_likelihood[i]`.
#'   Default value is `1`. WARNING: The normalizing constant for the likelihood
#'   is NOT recomputed, so ALL marginals (and the marginal likelihood) must be
#'   interpreted with great care.
#'
#'   For `family = "cp"`, the weights are applied as `sum(weights * eta)` in
#'   the point location contribution part of the log-likelihood, where `eta` is
#'   the linear predictor, and do not affect the integration part of the
#'   likelihood. This can be used to implement approximative methods for point
#'   location uncertainty.
#' @param scale Fixed (optional) scale parameters of the precision for several
#'   models, such as Gaussian and student-t response models.
#' @param domain,samplers,ips Arguments used for `family="cp"` and `aggregate=`.
#' \describe{
#' \item{`domain`}{Named list of domain definitions, see [fmesher::fm_int()].}
#' \item{`samplers`}{Integration domain for `family="cp"` or subdomains for
#'   `aggregate=`, see [fmesher::fm_int()].}
#' \item{`ips`}{Integration points. Defaults
#'   to [fmesher::fm_int]`(domain, samplers)`. If explicitly given,
#'   overrides `domain` and `samplers`. `fmesher::new_fm_int()` (from fmesher
#'   `0.5.0.9013`) can be used for manually constructed integration schemes.}
#' }
#' @param used Either `NULL` (default) or a [bru_used()] object. When,
#'   `NULL`, the information about what effects and
#' latent vectors are made available to the predictor evaluation is defined by
#' `bru_used(formula)`, which will include all effects and latent vectors
#' used by the predictor expression.
#' @param is_rowwise,allow_combine logical; If `is_rowwise` is `FALSE`, the
#'   predictor expression may involve several rows of the input data to
#'   influence the same row. When `NULL`, it defaults to `TRUE`, unless
#'   `response_data` is non-`NULL`, or `data` is a `list`, or the likelihood
#'   construction requires it. The `allow_combine` argument is a deprecated and
#'   inverted version, such that `is_rowwise = !allow_combine`.
#' @param aggregate character ("none" or a valid name for the `type` argument of
#'   [bm_aggregate()]) or an aggregation `bru_mapper` object ([bm_aggregate()],
#'   [bm_logsumexp()], or [bm_logitaverage()]). Default `NULL`, interpreted as
#'   "none". `r lifecycle::badge("experimental")`, available from version
#'   `2.12.0.9013`.
#' @param aggregate_input `NULL` or an optional input list to the mapper
#'   defined by non-NULL `aggregate`, overriding the default,
#'   ```
#'   list(block = .data.[[".block"]],
#'        weights = .data.[["weight"]],
#'        n_block = bru_response_size(.response_data.),
#'        block_response = ".block")
#'   ```
#'   `r lifecycle::badge("experimental")`, available from version `2.12.0.9013`.
#'
#'   From `2.13.0.9016`, it will look for a `block_response` element in the
#'   list, which should be the name of a response variable in `response_data` to
#'   match the `block` information against, allowing character or factor
#'   aggregation block information to be used by replacing `block` with
#'   `match(block, block_response)`. If not supplied, the name
#'   `".block"` is tried. If that isn't available, the `block` information must
#'   be supplied directly as a numeric or integer vector, indexing into the rows
#'   of the response variable. Having no `block_response` variable is equivalent
#'   to having
#'   `response_data$.block = seq_len(bru_response_size(response_data))`.
#' @param control.family A optional `list` of `INLA::control.family` options
#' @param control.gcpo A optional `list` of `INLA::control.gcpo` options
#' @param tag character; Name that can be used to identify the relevant parts
#' of INLA predictor vector output, via [bru_index()].
#' @param options A [bru_options] options object or a list of options passed
#' on to [bru_options()]
#' @param .envir The evaluation environment to use for special arguments (`E`,
#'   `Ntrials`, `weights`, and `scale`) if not found in `response_data` or
#'   `data`. Defaults to the calling environment.
#' @param include,exclude,include_latent `r lifecycle::badge("deprecated")`, use
#'   `used` instead.
#'
#' @details
#' The `E`, `Ntrials`, `weights`, and `scale` arguments are evaluated in the
#' data context, with values from `response_data` taking precedence over `data`.
#'
#' @return A likelihood configuration which can be used to parameterise [bru()].
#' @seealso [bru_response_size()], [bru_used()], [bru_comp()],
#' [bru_comp_eval()]
#'
#' @examples
#' \donttest{
#' if (bru_safe_inla() &&
#'   require(ggplot2, quietly = TRUE) &&
#'   require(patchwork, quietly = TRUE)) {
#'   # The 'bru_obs()' (previously 'like()') function's main purpose is to set
#'   # up observation models, both for single- and multi-likelihood models.
#'   # The following example generates some random covariates which are observed
#'   # through two different random effect models with different likelihoods
#'
#'   # Generate the data
#'
#'   set.seed(123)
#'
#'   n1 <- 200
#'   n2 <- 10
#'
#'   x1 <- runif(n1)
#'   x2 <- runif(n2)
#'   z2 <- runif(n2)
#'
#'   y1 <- rnorm(n1, mean = 2 * x1 + 3)
#'   y2 <- rpois(n2, lambda = exp(2 * x2 + z2 + 3))
#'
#'   df1 <- data.frame(y = y1, x = x1)
#'   df2 <- data.frame(y = y2, x = x2, z = z2)
#'
#'   # Single likelihood models and inference using bru are done via
#'
#'   cmp1 <- y ~ -1 + Intercept(1) + x
#'   fit1 <- bru(cmp1, family = "gaussian", data = df1)
#'   summary(fit1)
#'
#'   cmp2 <- y ~ -1 + Intercept(1) + x + z
#'   fit2 <- bru(cmp2, family = "poisson", data = df2)
#'   summary(fit2)
#'
#'   # A joint model has two likelihoods, which are set up using the bru_obs
#'   # function
#'
#'   lik1 <- bru_obs(
#'     "gaussian",
#'     formula = y ~ x + Intercept,
#'     data = df1,
#'     tag = "norm"
#'   )
#'   lik2 <- bru_obs(
#'     "poisson",
#'     formula = y ~ x + z + Intercept,
#'     data = df2,
#'     tag = "pois"
#'   )
#'
#'   # The union of effects of both models gives the components needed to run
#'   # bru
#'
#'   jcmp <- ~ x + z + Intercept(1)
#'   jfit <- bru(jcmp, lik1, lik2)
#'
#'   bru_index(jfit, "norm")
#'   bru_index(jfit, "pois")
#'
#'   # Compare the estimates
#'
#'   p1 <- ggplot() +
#'     gg(fit1$summary.fixed, bar = TRUE) +
#'     ylim(0, 4) +
#'     ggtitle("Model 1")
#'   p2 <- ggplot() +
#'     gg(fit2$summary.fixed, bar = TRUE) +
#'     ylim(0, 4) +
#'     ggtitle("Model 2")
#'   pj <- ggplot() +
#'     gg(jfit$summary.fixed, bar = TRUE) +
#'     ylim(0, 4) +
#'     ggtitle("Joint model")
#'
#'   (p1 / p2 / pj)
#' }
#' }
#'
bru_obs <- function(formula = . ~ .,
                    family = "gaussian",
                    data = NULL,
                    response_data = NULL,
                    data_extra = NULL,
                    E = NULL,
                    Ntrials = NULL,
                    weights = NULL,
                    scale = NULL,
                    domain = NULL,
                    samplers = NULL,
                    ips = NULL,
                    used = NULL,
                    is_rowwise = NULL,
                    aggregate = NULL,
                    aggregate_input = NULL,
                    control.family = NULL,
                    control.gcpo = NULL,
                    tag = NULL,
                    options = list(),
                    .envir = parent.frame(),
                    include = deprecated(),
                    exclude = deprecated(),
                    include_latent = deprecated(),
                    allow_combine = deprecated()) {
  options <- bru_call_options(options)
  bru_options_set_local(options, .reset = TRUE)

  tag_text <- if (is.null(tag)) {
    "<unknown>"
  } else {
    glue("'{tag}'")
  }

  check_sp_data_deprecation(
    data = data,
    response_data = response_data,
    samplers = samplers,
    ips = ips
  )

  # Some defaults
  inla.family <- family

  # Only for deprecated include/exclude/include_latent argument handling:
  formula_char <- as.character(formula)
  used <- bru_obs_check_used_deprecation(
    used = used,
    include = rlang::maybe_missing(include),
    exclude = rlang::maybe_missing(exclude),
    include_latent = rlang::maybe_missing(include_latent),
    expr_text = formula_char[length(formula_char)]
  )

  # Build expressions from formula:
  pred_expr <- new_bru_pred_expr(formula,
    used = used,
    .envir = .envir
  )

  # Set the response name
  if (is.null(pred_expr$resp_text)) {
    formula_text <- deparse1(formula, collapse = "    \n", width.cutoff = 80L)
    stop(glue::glue(
      "Missing response variable name or expression in formula for ",
      "bru_obs(tag={tag_text}).\n",
      "  Formula: {formula_text}"
    ))
  }
  response_expr <- rlang::parse_expr(pred_expr$resp_text)
  data_list <- list(response_data = response_data, data = data)
  response <- tryCatch(
    expr = bru_eval_in_data_context(
      !!response_expr,
      data = data_list,
      default = NULL,
      .envir = .envir
    ),
    error = function(e) {
      NULL
    }
  )

  E <- bru_eval_in_data_context(
    {{ E }},
    data = data_list,
    default = options[["E"]],
    .envir = .envir
  )
  Ntrials <- bru_eval_in_data_context(
    {{ Ntrials }},
    data = data_list,
    default = options[["Ntrials"]],
    .envir = .envir
  )
  weights <- bru_eval_in_data_context(
    {{ weights }},
    data = data_list,
    default = 1,
    .envir = .envir
  )
  scale <- bru_eval_in_data_context(
    {{ scale }},
    data = data_list,
    default = 1,
    .envir = .envir
  )

  data_extra <- as.list(data_extra)

  # Decide on is_rowwise default
  pred_expr <- bru_obs_handle_is_rowwise(
    pred_expr = pred_expr,
    is_rowwise = is_rowwise,
    allow_combine = allow_combine,
    data = data,
    response_data = response_data,
    aggregate = aggregate
  )

  original_response_data <- response_data
  # Must be a list or tibble to allow inla.mdata responses
  # Until inla.surv is converted to tibble (or data.frame), must be a list
  response_data <- list(
    BRU_response = response,
    BRU_E = E,
    BRU_Ntrials = Ntrials,
    BRU_scale = scale,
    BRU_weights = weights
  )

  # Prototype object
  lh <- structure(
    list(
      family = family,
      pred_expr = pred_expr,
      response_data = response_data,
      data = data,
      data_extra = data_extra,
      integration_info = list(
        ips = ips,
        domain = domain,
        samplers = samplers
      ),
      response = "BRU_response",
      inla.family = inla.family,
      control.family = control.family,
      control.gcpo = control.gcpo,
      tag = tag,
      BRU_original_response_data = original_response_data
    ),
    class = "bru_obs"
  )

  lh$formula <- formula
  lh <- bru_compat_pre_2_14_bru_obs(lh)

  if (!is.null(aggregate)) {
    lh <- bru_obs_agg(lh,
      aggregate = aggregate,
      aggregate_input = {{ aggregate_input }},
      options = options,
      .envir = .envir
    )
  }

  # More on special bru likelihoods
  if (family == "cp") {
    lh <- bru_obs_family_cp(lh, options = options, .envir = .envir)
  }

  if (is.null(lh[["response_data"]][[lh[["response"]]]])) {
    stop(glue::glue(
      "Response variable missing or could not be evaluated in ",
      "bru_obs(tag={tag_text})"
    ))
  }

  if (is.null(lh[["used"]])) {
    lh[["used"]] <- bru_used(formula)
  }

  # Remove temporary copy of input data
  lh$BRU_original_response_data <- NULL

  # Return likelihood
  lh
}

#' @describeIn bru_obs `r lifecycle::badge("deprecated")` Legacy `like()`
#' method for `inlabru` prior to version `2.12.0`. Use [bru_obs()] instead.
#' @param \dots Arguments passed from `like()` to `bru_obs()`
#' @export
like <- function(...,
                 E = NULL,
                 Ntrials = NULL,
                 weights = NULL,
                 scale = NULL,
                 options = list(),
                 .envir = parent.frame()) {
  options <- bru_call_options(options)
  bru_options_set_local(options, .reset = TRUE)
  bru_log_message(
    paste0(
      "The `like()` function has been deprecated in favour of `bru_obs()`, ",
      "since inlabru 2.12.0."
    ),
    verbosity = 1L
  )
  lifecycle::deprecate_warn(
    "2.12.0",
    "like()",
    "bru_obs()"
  )

  bru_obs(
    ...,
    E = {{ E }},
    Ntrials = {{ Ntrials }},
    weights = {{ weights }},
    scale = {{ scale }},
    options = options,
    .envir = .envir
  )
}


#' @title Response size queries
#'
#' @description
#' Extract the number of response values from `bru` and related objects.
#' @return An `integer` vector.
#' @param object An object from which to extract response size(s).
#' @seealso [like()]
#' @export
#' @examples
#' bru_response_size(
#'   bru_obs(y ~ 1, data = data.frame(y = rnorm(10)), family = "gaussian")
#' )
bru_response_size <- function(object) {
  UseMethod("bru_response_size")
}

#' @describeIn bru_response_size Extract the number of observations from an
#'   object supporting `NROW()`.
#' @export
bru_response_size.default <- function(object) {
  NROW(object)
}

#' @describeIn bru_response_size Extract the number of observations from an
#'   object supporting `NROW(object[[1]])`.
#' @export
bru_response_size.list <- function(object) {
  NROW(object[[1]])
}

#' @describeIn bru_response_size Extract the number of observations from an
#' `inla.surv` object.
#' @export
bru_response_size.inla.surv <- function(object) {
  # This special case handling is needed because inla.surv objects are
  # lists. When the INLA package changes representation of `inla.surv`
  # to a tibble or data.frame, this special case won't be needed anymore.
  NROW(object[[1]])
}

#' @describeIn bru_response_size Extract the number of observations from a
#' `bru_obs` object.
#' @export
bru_response_size.bru_obs <- function(object) {
  bru_response_size(object[["response_data"]][[object[["response"]]]])
}

#' @describeIn bru_response_size Extract the number of observations from a
#' `bru_obs_list` object, as a vector with one value per observation model.
#' @export
bru_response_size.bru_obs_list <- function(object) {
  vapply(object, bru_response_size, 1L)
}

#' @describeIn bru_response_size Extract the number of observations from a
#' `bru_info` object, as a vector with one value per observation model.
#' @export
bru_response_size.bru_info <- function(object) {
  bru_response_size(object[["lhoods"]])
}

#' @describeIn bru_response_size Extract the number of observations from a
#' `bru` object, as a vector with one value per observation model.
#' @export
bru_response_size.bru <- function(object) {
  bru_response_size(object[["bru_info"]])
}


set_list_names <- function(x, tag, priority = "immutable") {
  priority <- match.arg(priority, c("immutable", "tag", "name"))
  list_names <- names(x)
  if (is.null(list_names)) {
    list_names <- rep(NA_character_, length(x))
  } else {
    list_names <- vapply(
      list_names,
      function(xx) {
        if (identical(xx, "")) {
          NA_character_
        } else {
          xx
        }
      },
      ""
    )
  }
  tag_names <- vapply(
    x,
    function(xx) {
      if (is.null(xx[[tag]]) ||
        identical(xx[[tag]], "")) {
        NA_character_
      } else {
        xx[[tag]]
      }
    },
    ""
  )
  new_names <- vapply(
    seq_along(x),
    function(k) {
      if (priority == "immutable") {
        if (is.na(list_names[k])) {
          return(tag_names[k])
        }
        if (is.na(tag_names[k])) {
          return(list_names[k])
        }
        if (identical(list_names[k], tag_names[k])) {
          return(list_names[k])
        }
        stop(glue("
          When `priority = 'immutable'`, the tag and name must match.
          Cannot combine objects with mismatching tags and names:
            Tag: {tag_names[k]}
            Name: {list_names[k]}
          Use `NA` for either the tag or name, or use matching tags/names."))
      } else if (priority == "tag") {
        if (!is.na(tag_names[k])) {
          return(tag_names[k])
        }
        if (!is.na(list_names[k])) {
          return(list_names[k])
        }
        NA_character_
      } else if (priority == "name") {
        if (!is.na(list_names[k])) {
          return(list_names[k])
        }
        if (!is.na(tag_names[k])) {
          return(tag_names[k])
        }
        NA_character_
      }
    },
    ""
  )
  names(x) <- new_names
  # Set tags on missing tags
  for (k in which(is.na(tag_names))) {
    x[[k]][[tag]] <- new_names[k]
  }

  x
}

#' @describeIn bru_obs
#' Combine `bru_obs` observation model object into a `bru_obs_list` object
#' @param \dots For `bru_obs_list.bru_obs`, one or more `bru_obs` objects
#' @export
bru_obs_list <- function(..., .tag = NULL) {
  UseMethod("bru_obs_list")
}

#' @describeIn bru_obs
#' Combine one or more lists of `bru_obs` observation model objects
#' into a `bru_obs_list` object
#' @param object A list of `bru_obs` and/or `bru_obs_list` objects
#' @export
bru_obs_list.list <- function(object, ..., .tag = NULL) {
  if (length(list(...)) > 0) {
    return(bru_obs_list(list(object, ...)))
  }
  if (length(object) == 0) {
    return(
      structure(
        list(),
        class = c("bru_obs_list", "list")
      )
    )
  }
  is_obs <- vapply(object, function(x) inherits(x, "bru_obs"), TRUE)
  if (!all(is_obs)) {
    is_obs_list <- vapply(object, function(x) inherits(x, "bru_obs_list"), TRUE)
    if (!all(is_obs_list)) {
      object <- lapply(seq_along(object), function(k) {
        bru_obs_list(object[[k]], .tag = names(object)[k])
      })
    }
    object <- unlist(object, recursive = FALSE)
  }
  object <- structure(
    object,
    class = c("bru_obs_list", "list")
  )

  object <- set_list_names(object, tag = "tag", priority = "immutable")

  object
}

#' @describeIn bru_obs
#' Combine one or more lists of `bru_obs` observation model objects
#' into a `bru_obs_list` object
#' @param .tag Optional name to assign to a single `bru_obs` object. Reserved
#'   for internal use.
#' @export
bru_obs_list.bru_obs <- function(..., .tag = NULL) {
  if (length(list(...)) != 1L) {
    return(bru_obs_list(list(...)))
  }

  object <- structure(
    list(...),
    class = c("bru_obs_list", "list")
  )
  if (!is.null(.tag)) {
    names(object) <- .tag
  }
  object <- set_list_names(object, tag = "tag", priority = "immutable")
  object
}

#' @describeIn bru_obs
#' Combine one or more `bru_obs_list` objects into a `bru_obs_list` object
#' @export
bru_obs_list.bru_obs_list <- function(..., .tag = NULL) {
  if (length(list(...)) != 1L) {
    return(bru_obs_list(list(...)))
  }
  object <- list(...)[[1]]
  object <- set_list_names(object, tag = "tag", priority = "immutable")
  object
}


#' @describeIn bru_obs
#' Combine several `bru_obs` objects into a `bru_obs_list` object
#' @export
c.bru_obs <- function(...) {
  bru_obs_list(list(...))
}


#' @describeIn bru_obs
#' Combine several `bru_obs_list` objects into a `bru_obs_list` object
#' @export
c.bru_obs_list <- function(...) {
  bru_obs_list(list(...))
}


#' @export
#' @param x `bru_obs_list` object from which to extract element(s)
#' @param i indices specifying elements to extract
#' @rdname bru_obs
#' @seealso [summary.bru_obs()]
`[.bru_obs_list` <- function(x, i) {
  env <- environment(x)
  object <- NextMethod()
  class(object) <- c("bru_obs_list", "list")
  environment(object) <- env
  object
}


#' Summary and print methods for observation models
#'
#' @rdname bru_obs_print
#' @seealso [bru_obs()]
#' @param object Object to operate on
#' @param verbose logical; If `TRUE`, include more details of the
#' component definitions. If `FALSE`, only show basic component
#' definition information. Default: `TRUE`
#' @param \dots Arguments passed on to other `summary` methods
#' @param x Object to be printed
#' @method summary bru_obs
#' @export
#' @examples
#' obs <- bru_obs(y ~ ., data = data.frame(y = rnorm(10)))
#' summary(obs)
#' print(obs)
#'
summary.bru_obs <- function(object, verbose = TRUE, ...) {
  structure(
    list(
      family = object[["family"]],
      data_class = class(object[["data"]]),
      response_class = class(object[["response_data"]][[object[["response"]]]]),
      predictor =
        glue::glue("{bru_pred_expr(object, format = 'formula_text')}"),
      is_additive = bru_is_additive(object),
      is_linear = bru_is_linear(object),
      is_rowwise = bru_is_rowwise(object),
      used = bru_used(object),
      tag = object[["tag"]]
    ),
    class = "summary_bru_obs"
  )
}

#' @describeIn bru_obs `r lifecycle::badge("deprecated")`
#' Backwards compatibility for versions `<= 2.12.0`. For later versions, use
#' `as_bru_obs_list()`, `bru_obs_list()`, or `c()`.
#' @export
like_list <- function(...) {
  lifecycle::deprecate_warn(
    "2.12.0",
    "like_list()",
    "bru_obs_list()",
    details = paste0(
      "Use `as_bru_obs_list()`, `bru_obs_list(...)` or ",
      "`c(...)` to construct observation model lists."
    )
  )
  bru_obs_list(list(...))
}

#' @describeIn bru_obs `r lifecycle::badge("deprecated")`
#' Backwards compatibility for versions `<= 2.12.0.9017`. For later versions,
#' use `as_bru_obs_list()`, `bru_obs_list()` or `c()`.
#' @export
bru_like_list <- function(...) {
  lifecycle::deprecate_warn(
    "2.12.0.9017",
    "bru_like_list()",
    "bru_obs_list()",
    details = paste0(
      "Use `as_bru_obs_list()`, `bru_obs_list(...)` or ",
      "`c(...)` to construct observation model lists."
    )
  )
  bru_obs_list(list(...))
}

#' @rdname bru_obs_print
#' @method summary bru_obs_list
#' @export
summary.bru_obs_list <- function(object, verbose = TRUE, ...) {
  structure(
    lapply(
      object,
      function(x) {
        summary(x, verbose = verbose, ...)
      }
    ),
    class = "summary_bru_obs_list"
  )
}


#' @rdname bru_obs_print
#' @export
#' @importFrom glue glue
print.summary_bru_obs <- function(x, ...) {
  lh <- x
  cat(
    glue_data(
      lh,
      "  Model tag: {tag}\n",
      "    Family: '{family}'\n",
      "    Data class: {data_class}\n",
      "    Response class: {response_class}\n",
      "    Predictor: {predictor}\n",
      "    Additive/Linear/Rowwise: {is_additive}/{is_linear}/{is_rowwise}\n",
      "    Used components: {format(lh$used)}",
      .trim = FALSE,
      tag = if (is.null(lh$tag) || is.na(lh$tag)) {
        "<No tag>"
      } else {
        glue::glue_collapse(glue::glue_data(lh, "'{tag}'"), sep = ", ")
      },
      data_class = glue::glue_collapse(
        glue::glue_data(lh, "'{data_class}'"),
        sep = ", "
      ),
      response_class = glue::glue_collapse(
        glue::glue_data(lh, "'{response_class}'"),
        sep = ", "
      ),
      predictor =
        if (length(lh$predictor) > 1) {
          paste0(
            "\n        ",
            paste0(
              lh$predictor,
              collapse = "\n        "
            )
          )
        } else {
          lh$predictor
        }
    ),
    "\n"
  )
  invisible(x)
}

#' @rdname bru_obs_print
#' @export
print.summary_bru_obs_list <- function(x, ...) {
  cat("Observation models:\n")
  for (lh in x) {
    print(lh)
  }
  invisible(x)
}

#' @export
#' @rdname bru_obs_print
print.bru_obs <- function(x, ...) {
  print(summary(x))
  invisible(x)
}

#' @export
#' @rdname bru_obs_print
print.bru_obs_list <- function(x, ...) {
  print(summary(x))
  invisible(x)
}


#' Utility functions for bru observation model objects
#' @param x Object of `bru_obs` or `bru_obs_list` type
#' @param \dots Further arguments passed on to the submethods
#' @export
#' @keywords internal
#' @returns * `bru_obs_inla_family()` returns a string or vector of strings
#' @rdname bru_obs_methods
#' @name bru_obs_methods
#' @seealso [summary.bru_obs()]
bru_obs_inla_family <- function(x, ...) {
  UseMethod("bru_obs_inla_family")
}
#' @export
#' @rdname bru_obs_methods
bru_obs_inla_family.bru_obs <- function(x, ...) {
  x[["inla.family"]]
}
#' @export
#' @rdname bru_obs_methods
bru_obs_inla_family.bru_obs_list <- function(x, ...) {
  vapply(x, bru_obs_inla_family, "")
}

#' @param control.family list of INLA `control.family` options to override
#' @export
#' @keywords internal
#' @returns * `bru_obs_control_family()` returns a list with
#'   `INLA::control.family` options, or a list of such lists, with one element
#'   per observation model
#' @rdname bru_obs_methods
bru_obs_control_family <- function(x, control.family = NULL, ...) {
  UseMethod("bru_obs_control_family")
}

#' @export
#' @rdname bru_obs_methods
bru_obs_control_family.bru_obs <- function(x, control.family = NULL, ...) {
  if (!is.null(control.family)) {
    control.family
  } else if (is.null(x[["control.family"]])) {
    list()
  } else {
    x[["control.family"]]
  }
}

#' @export
#' @rdname bru_obs_methods
bru_obs_control_family.bru_obs_list <- function(x,
                                                control.family = NULL,
                                                ...) {
  # Extract the control.family information for each likelihood
  if (!is.null(control.family)) {
    if (length(control.family) != length(x)) {
      stop(paste0(
        "control.family supplied as option, but format doesn't match ",
        "the number of likelihoods"
      ))
    }
    like_has_cf <- vapply(
      seq_along(x),
      function(k) !is.null(x[[k]][["control.family"]]),
      TRUE
    )
    if (any(like_has_cf)) {
      bru_log_warn(
        glue(
          "Global control.family option overrides settings in likelihood(s) ",
          "{glue_collapse(which(like_has_cf), sep = ', ', last = ' and ')}",
        )
      )
    }
  } else {
    control.family <- lapply(x, bru_obs_control_family)
    # inla() requires a unnamed list of lists
    names(control.family) <- NULL
  }
  control.family
}

#' @export
#' @keywords internal
#' @returns * `bru_obs_control_gcpo()` returns a list with
#'   `INLA::control.gcpo` options, with predictor/response variable indices
#'   unified for multi-observation models. If a `bru_obs` model has a NULL
#'   `control.gcpo` argument, an empty list is returned for that model.
#' @rdname bru_obs_methods
bru_obs_control_gcpo <- function(x, ...) {
  UseMethod("bru_obs_control_gcpo")
}

#' @param index_offset integer; offset to add to indices in `control.gcpo`
#' @param index_length integer; length of the response vector for the
#'   observation model. Defaults to `bru_response_size(x)`.
#' @param force_weights logical; if `TRUE`, ensure that `control.gcpo$weights`
#'   is populated. This is needed if any of the observation models in a
#'   `bru_obs_list` has non-null `control.gcpo$weights`.
#' @export
#' @rdname bru_obs_methods
bru_obs_control_gcpo.bru_obs <- function(x,
                                         index_offset,
                                         index_length = bru_response_size(x),
                                         force_weights,
                                         ...) {
  c.gcpo <- x[["control.gcpo"]]
  if (is.null(c.gcpo)) {
    return(list())
  }
  for (nm in intersect(
    c("groups", "selection", "group.selection", "friends"),
    names(c.gcpo)
  )) {
    if (nm == "groups") {
      c.gcpo[[nm]] <- lapply(c.gcpo[[nm]], function(v) {
        v$idx <- v$idx + index_offset
        v
      })
    } else {
      c.gcpo[[nm]] <- lapply(c.gcpo[[nm]], function(v) v + index_offset)
    }
  }
  if (!("friends" %in% names(c.gcpo))) {
    c.gcpo[["friends"]] <- as.list(index_offset + seq_len(index_length))
  }
  if (force_weights && !("weights" %in% names(c.gcpo))) {
    c.gcpo[["weights"]] <- rep(1.0, index_length)
  } else if ("weights" %in% names(c.gcpo)) {
    if (length(c.gcpo[["weights"]]) != index_length) {
      stop(glue::glue(
        "Length of control.gcpo$weights ({length(c.gcpo[['weights']])}) ",
        "does not match response length ({index_length})"
      ))
    }
  }
  c.gcpo
}

#' @param control.gcpo list of INLA `control.gcpo` default options
#' @export
#' @rdname bru_obs_methods
bru_obs_control_gcpo.bru_obs_list <- function(x,
                                              control.gcpo = NULL,
                                              ...) {
  # Update the control.gcpo information for each likelihood
  response_sizes <- bru_response_size(x)
  any_element <- vapply(
    c("groups", "selection", "group.selection", "friends", "weights"),
    function(nm) {
      any(vapply(x, function(lh) {
        !is.null(lh[["control.gcpo"]][[nm]])
      }, TRUE))
    },
    TRUE
  )
  all_element <- vapply(
    c("groups", "selection", "group.selection", "friends", "weights"),
    function(nm) {
      all(vapply(x, function(lh) {
        !is.null(lh[["control.gcpo"]][[nm]])
      }, TRUE))
    },
    TRUE
  )
  c.gcpo <- lapply(
    seq_along(x),
    function(k) {
      bru_obs_control_gcpo(x[[k]],
        index_offset = sum(response_sizes[seq_len(k - 1)]),
        index_length = response_sizes[k],
        force_weights = any_element["weights"]
      )
    }
  )

  # If given in one model, must be given in all models
  c.gcpo.combined <- list()
  for (nm in c("groups", "selection", "group.selection")) {
    if (any_element[nm]) {
      if (!all_element[nm]) {
        stop(glue::glue(
          "control.gcpo${nm} given in some, but not all, observation models"
        ))
      }
      c.gcpo.combined[[nm]] <-
        do.call("c", lapply(c.gcpo, function(lh) lh[[nm]]))
    }
  }
  # Combine list/vector
  c.gcpo.combined[["friends"]] <-
    do.call("c", lapply(c.gcpo, function(lh) lh[["friends"]]))
  if (any_element[["weights"]]) {
    c.gcpo.combined[["weights"]] <-
      unlist(lapply(c.gcpo, function(lh) lh[["weights"]]))
  }

  merge_list_check <- function(a, b, exclude = character(0)) {
    for (nm in setdiff(names(b), exclude)) {
      if (nm %in% names(a)) {
        if (identical(a[[nm]], b[[nm]])) {
          next
        }
        stop(glue::glue(
          "Cannot merge control.gcpo lists: ",
          "both contain element '{nm}' with conflicting content."
        ))
      }
      a[[nm]] <- b[[nm]]
    }
    a
  }

  exc <- names(c.gcpo.combined)
  if (!is.null(control.gcpo)) {
    c.gcpo.combined <- merge_list_check(c.gcpo.combined,
      control.gcpo,
      exclude = exc
    )
  }
  for (idx in seq_along(c.gcpo)) {
    c.gcpo.combined <- merge_list_check(c.gcpo.combined,
      c.gcpo[[idx]],
      exclude = exc
    )
  }

  c.gcpo.combined
}


#' @title Log Gaussian Cox process (LGCP) inference using INLA
#'
#' @description
#' This function performs inference on a LGCP observed via points residing
#' possibly multiple dimensions. These dimensions are defined via the left hand
#' side of the formula provided via the model parameter. The left hand side
#' determines the intensity function that is assumed to drive the LGCP. This may
#' include effects that lead to a thinning (filtering) of the point process. By
#' default, the log intensity is assumed to be a linear combination of the
#' effects defined by the formula's RHS.
#'
#' More sophisticated models, e.g.
#' non-linear thinning, can be achieved by using the predictor argument. The
#' latter requires multiple runs of INLA for improving the required
#' approximation of the predictor. In many applications the LGCP is only
#' observed through subsets of the dimensions the process is living in. For
#' example, spatial point realizations may only be known in sub-areas of the
#' modelled space. These observed subsets of the LGCP domain are called samplers
#' and can be provided via the respective parameter. If samplers is NULL it is
#' assumed that all of the LGCP's dimensions have been observed completely.
#'
#' @export
#' @inheritParams bru_obs
#' @inheritParams bru
#' @param \dots Further arguments passed on to [bru_obs()].
#' @details
#' The `E` and `weights` arguments are evaluated in the data context, like for
#' [bru_obs()].
#' @return An [bru()] object
#' @examples
#' \donttest{
#' if (bru_safe_inla() &&
#'   require(ggplot2, quietly = TRUE) &&
#'   require(fmesher, quietly = TRUE) &&
#'   requireNamespace("sn", quietly = TRUE)) {
#'   # Load the Gorilla data
#'   data <- gorillas_sf
#'
#'   # Plot the Gorilla nests, the mesh and the survey boundary
#'   ggplot() +
#'     fmesher::geom_fm(data = data$mesh) +
#'     gg(data$boundary, fill = "blue", alpha = 0.2) +
#'     gg(data$nests, col = "red", alpha = 0.2)
#'
#'   # Define SPDE prior
#'   matern <- INLA::inla.spde2.pcmatern(
#'     data$mesh,
#'     prior.sigma = c(0.1, 0.01),
#'     prior.range = c(0.1, 0.01)
#'   )
#'
#'   # Define domain of the LGCP as well as the model components (spatial SPDE
#'   # effect and Intercept)
#'   cmp <- geometry ~ field(geometry, model = matern) + Intercept(1)
#'
#'   # Fit the model (with int.strategy="eb" to make the example take less time)
#'   fit <- lgcp(cmp, data$nests,
#'     samplers = data$boundary,
#'     domain = list(geometry = data$mesh),
#'     options = list(control.inla = list(int.strategy = "eb"))
#'   )
#'
#'   # Predict the spatial intensity surface
#'   lambda <- predict(
#'     fit,
#'     fmesher::fm_pixels(data$mesh, mask = data$boundary),
#'     ~ exp(field + Intercept)
#'   )
#'
#'   # Plot the intensity
#'   ggplot() +
#'     gg(lambda, geom = "tile") +
#'     gg(data$nests, col = "red", alpha = 0.2)
#' }
#' }
#'
lgcp <- function(components,
                 data,
                 domain = NULL,
                 samplers = NULL,
                 ips = NULL,
                 formula = . ~ .,
                 E = NULL,
                 weights = NULL,
                 ...,
                 options = list(),
                 .envir = parent.frame()) {
  # If formula response missing, copy from components (for formula input)
  if (inherits(components, "formula")) {
    response <- extract_response(components)
    formula <- auto_response(formula, response)
  }
  lik <- bru_obs(
    family = "cp",
    formula = formula, data = data,
    domain = domain, samplers = samplers, ips = ips,
    ...,
    E = {{ E }},
    weights = {{ weights }},
    options = options,
    .envir = .envir
  )
  bru(components, lik, options = options, .envir = .envir)
}


expand_to_dataframe <- function(x, data = NULL) {
  if (is.null(data)) {
    data <- data.frame(matrix(nrow = NROW(x), ncol = 0))
  }
  only_x <- setdiff(names(x), names(data))
  if (length(only_x) < length(names(x))) {
    x <- x[!(names(x) %in% names(data))]
  }
  if (inherits(x, "Spatial")) {
    bru_safe_sp(force = TRUE)
    if (inherits(x, "SpatialPixels") &&
      !inherits(x, "SpatialPixelsDataFrame")) {
      result <- sp::SpatialPixelsDataFrame(x, data = data)
    } else if (inherits(x, "SpatialGrid") &&
      !inherits(x, "SpatialGridDataFrame")) {
      result <- sp::SpatialGridDataFrame(x, data = data)
    } else if (inherits(x, "SpatialLines") &&
      !inherits(x, "SpatialLinesDataFrame")) {
      result <- sp::SpatialLinesDataFrame(x, data = data)
    } else if (inherits(x, "SpatialPolygons") &&
      !inherits(x, "SpatialPolygonsDataFrame")) {
      result <- sp::SpatialPolygonsDataFrame(x, data = data)
    } else if (inherits(x, "SpatialPoints") &&
      !inherits(x, "SpatialPointsDataFrame")) {
      # Other classes inherit from SpatialPoints, so need to be handled first
      result <- sp::SpatialPointsDataFrame(x, data = data)
    } else {
      result <- sp::cbind.Spatial(x, data)
    }
  } else if (is.data.frame(x)) {
    result <- cbind(x, data)
  } else {
    result <- c(x, as.list(data))
  }
  result
}


#
#' Prediction from fitted bru model
#'
#' Takes a fitted `bru` object produced by the function [bru()] and produces
#' predictions given a new set of values for the model covariates or the
#' original values used for the model fit. The predictions can be based on any
#' R expression that is valid given these values/covariates and the joint
#' posterior of the estimated random effects.
#'
#' Mean value predictions are accompanied by the standard errors, upper and
#' lower 2.5% quantiles, the
#' median, variance, coefficient of variation as well as the variance and
#' minimum and maximum sample
#' value drawn in course of estimating the statistics.
#'
#' Internally, this method calls [generate.bru()] in order to draw samples from
#' the model.
#'
#' @export
#' @param object An object obtained by calling [bru()] or [lgcp()].
#' @param newdata A `data.frame` or `SpatialPointsDataFrame` of covariates
#'   needed for the prediction.
#' @param formula A formula where the right hand side defines an R expression
#' to evaluate for each generated sample. If `NULL`, the latent and
#' hyperparameter states are returned as named list elements.
#' See Details for more information.
#' @param n.samples Integer setting the number of samples to draw in order to
#' calculate the posterior statistics. The default is rather low but provides
#' a quick approximate result.
#' @param seed Random number generator seed passed on to `inla.posterior.sample`
#' @param probs A numeric vector of probabilities with values in `[0, 1]`,
#'   passed to `stats::quantile`
#' @param num.threads Specification of desired number of threads for parallel
#' computations. Default NULL, leaves it up to INLA.
#' When seed != 0, overridden to "1:1:1"
#' @param used Either `NULL` or a [bru_used()] object.
#'   Default, `NULL`, uses auto-detection of used variables in the formula.
#' @param drop logical; If `drop=FALSE`, and
#'   the prediction summary has the same number of rows as `newdata`, then the
#'   output is a joined object. Default `FALSE`.
#' @param \dots Additional arguments passed on to `inla.posterior.sample()`
#' @param include,exclude `r lifecycle::badge("deprecated")` If auto-detection
#' of used variables fails, use `used` instead.
#' @details
#' In addition to the component names (that give the effect of each component
#' evaluated for the input data), the suffix `_latent` variable name can be used
#' to directly access the latent state for a component, and the suffix function
#' `_eval` can be used to evaluate a component at other input values than the
#' expressions defined in the component definition itself, e.g.
#' `field_eval(cbind(x, y))` for a component that was defined with
#' `field(coordinates, ...)` (see also [bru_comp_eval()]).
#'
#' For "iid" models with `mapper = bm_index(n)`, `rnorm()` is used to
#' generate new realisations for indices greater than `n`, if accessed
#' via `<name>_eval(...)`.
#'
#' @return a `data.frame`, `sf`, or `Spatial*` object with predicted mean values
#'   and other summary statistics attached. Non-S4 object outputs have the class
#'   "bru_prediction" added at the front of the class list.
#' @examples
#' \donttest{
#' if (bru_safe_inla() &&
#'   requireNamespace("sn", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("sf", quietly = TRUE)) {
#'   # Load the Gorilla data
#'
#'   gorillas <- gorillas_sf
#'
#'   # Plot the Gorilla nests, the mesh and the survey boundary
#'
#'   ggplot() +
#'     gg(gorillas$mesh) +
#'     gg(gorillas$nests) +
#'     gg(gorillas$boundary, alpha = 0.1)
#'
#'   # Define SPDE prior
#'
#'   matern <- INLA::inla.spde2.pcmatern(
#'     gorillas$mesh,
#'     prior.sigma = c(0.1, 0.01),
#'     prior.range = c(0.01, 0.01)
#'   )
#'
#'   # Define domain of the LGCP as well as the model components (spatial SPDE
#'   # effect and Intercept)
#'
#'   cmp <- geometry ~ field(geometry, model = matern) + Intercept(1)
#'
#'   # Fit the model, with "eb" instead of full Bayes
#'   fit <- lgcp(
#'     cmp,
#'     data = gorillas$nests,
#'     samplers = gorillas$boundary,
#'     domain = list(geometry = gorillas$mesh),
#'     options = list(control.inla = list(int.strategy = "eb"))
#'   )
#'
#'   # Once we obtain a fitted model the predict function can serve various
#'   # purposes.
#'   # The most basic one is to determine posterior statistics of a univariate
#'   # random variable in the model, e.g. the intercept
#'
#'   icpt <- predict(fit, NULL, ~ c(Intercept = Intercept_latent))
#'   plot(icpt)
#'
#'   # The formula argument can take any expression that is valid within the
#'   # model, for instance a non-linear transformation of a random variable
#'
#'   exp.icpt <- predict(fit, NULL, ~ c(
#'     "Intercept" = Intercept_latent,
#'     "exp(Intercept)" = exp(Intercept_latent)
#'   ))
#'   plot(exp.icpt, bar = TRUE)
#'
#'   # The intercept is special in the sense that it does not depend on other
#'   # variables or covariates. However, this is not true for the smooth spatial
#'   # effects 'field'.
#'   # In order to predict 'field' we have to define where (in space) to
#'   # predict.
#'   # For this purpose, the second argument of the predict function can take
#'   # \code{data.frame} objects as well as sf (and legacy sp/Spatial) objects.
#'   # For instance, we might want to predict 'field' at the locations of the
#'   # mesh vertices. Using
#'
#'   vrt <- fmesher::fm_vertices(gorillas$mesh, format = "sf")
#'
#'   # we obtain these vertices as an sf object with POINT geometries
#'
#'   ggplot() +
#'     gg(gorillas$mesh) +
#'     gg(vrt, color = "red")
#'
#'   # Predicting 'field' at these locations works as follows
#'
#'   field <- predict(fit, vrt, ~field)
#'
#'   # Note that just like the input also the output will be a sf object with
#'   # points and that the predicted statistics are simply added as columns
#'
#'   class(field)
#'   head(vrt)
#'   head(field)
#'
#'   # Plotting the mean, for instance, at the mesh node is straight forward
#'
#'   ggplot() +
#'     gg(gorillas$mesh) +
#'     gg(field, aes(color = mean), size = 2)
#'
#'   # However, we are often interested in a spatial field and thus a linear
#'   # interpolation, which can be achieved by using the gg mechanism for meshes
#'
#'   ggplot() +
#'     gg(gorillas$mesh, color = field$mean)
#'
#'   # Alternatively, we can predict the spatial field at a grid of locations,
#'   # e.g. a sf object with a grid of points covering the relevant part of mesh
#'
#'   pxl <- fmesher::fm_pixels(gorillas$mesh,
#'     format = "sf",
#'     mask = gorillas$boundary
#'   )
#'   field2 <- predict(fit, pxl, ~field)
#'
#'   # This will give us a sf with the columns we are looking for
#'
#'   head(field2)
#'   ggplot() +
#'     gg(gorillas$boundary) +
#'     gg(data = field2, geom = "tile")
#' }
#' }
#'
predict.bru <- function(object,
                        newdata = NULL,
                        formula = NULL,
                        n.samples = 100,
                        seed = 0L,
                        probs = c(0.025, 0.5, 0.975),
                        num.threads = NULL,
                        used = NULL,
                        drop = FALSE,
                        ...,
                        include = deprecated(),
                        exclude = deprecated()) {
  object <- bru_check_object_bru(object)

  # Convert data into list, data.frame or a Spatial object if not provided as
  # such
  if (is.character(newdata)) {
    newdata <- as.list(setNames(newdata, newdata))
  } else if (inherits(newdata, "formula")) {
    stop(paste0(
      "Formula supplied as data to predict.bru(). ",
      "Please check your argument order/names."
    ))
  }

  if (inherits(newdata, "sf")) {
    # Make sure coordinates have unified XY/XYZ/XYM/XYZM structure:
    newdata <- fmesher::fm_zm(newdata)
  }

  vals <- generate(
    object,
    newdata = newdata,
    formula = formula,
    n.samples = n.samples,
    seed = seed,
    num.threads = num.threads,
    include = include,
    exclude = exclude,
    used = used,
    ...
  )

  # Summarise

  newdata <- expand_to_dataframe(newdata)
  if (is.data.frame(vals[[1]])) {
    vals.names <- names(vals[[1]])
    covar <- intersect(vals.names, names(newdata))
    estim <- setdiff(vals.names, covar)
    smy <- list()

    for (nm in estim) {
      smy[[nm]] <-
        bru_summarise(
          lapply(
            vals,
            function(v) v[[nm]]
          ),
          x = vals[[1]][, covar, drop = FALSE],
          probs = probs
        )
    }
    is.annot <- vapply(names(smy), function(v) all(smy[[v]]$sd == 0), TRUE)
    annot <- do.call(cbind, lapply(smy[is.annot], function(v) v[, 1]))
    smy <- smy[!is.annot]
    if (!is.null(annot)) {
      smy <- lapply(smy, function(v) cbind(data.frame(annot), v))
    }

    if (!drop) {
      smy <- lapply(
        smy,
        function(tmp) {
          if (NROW(newdata) == NROW(tmp)) {
            expand_to_dataframe(newdata, tmp)
          } else {
            tmp
          }
        }
      )
    }

    if (length(smy) == 1) smy <- smy[[1]]
  } else if (is.list(vals[[1]])) {
    vals.names <- names(vals[[1]])
    if (any(vals.names == "")) {
      bru_log_warn(
        "Some generated list elements are unnamed"
      )
    }
    smy <- list()
    for (nm in vals.names) {
      tmp <-
        bru_summarise(
          data = lapply(
            vals,
            function(v) v[[nm]]
          ),
          probs = probs
        )
      if (!drop &&
        (NROW(newdata) == NROW(tmp))) {
        smy[[nm]] <- expand_to_dataframe(newdata, tmp)
      } else {
        smy[[nm]] <- tmp
      }
    }
  } else {
    tmp <- bru_summarise(data = vals, probs = probs)
    if (!drop &&
      (NROW(newdata) == NROW(tmp))) {
      smy <- expand_to_dataframe(newdata, tmp)
    } else {
      smy <- tmp
    }
  }

  if (!isS4(smy)) {
    class(smy) <- c("bru_prediction", class(smy))
  }
  smy
}


bru_generate_check_used_deprecation <- function(
  include,
  exclude,
  env = rlang::caller_env(),
  user_env = rlang::caller_env(2)
) {
  if (lifecycle::is_present(include)) {
    bru_log_message(
      paste0(
        "The `include` argument to `generate.bru()` is deprecated. ",
        "If auto-detection doesn't work, use ",
        "`used = bru_used(effect = include)`."
      ),
      verbosity = 1L
    )
    lifecycle::deprecate_warn(
      "2.12.0.9003",
      "generate(include)",
      "generate(used)",
      "If auto-detection doesn't work, use `bru_used(effect = include)`",
      env = env,
      user_env = user_env
    )
  } else {
    include <- NULL
  }
  if (lifecycle::is_present(exclude)) {
    bru_log_message(
      paste0(
        "The `exclude` argument to `generate.bru()` is deprecated. ",
        "If auto-detection doesn't work, use ",
        "`used = bru_used(effect_exclude = exclude)`."
      ),
      verbosity = 1L
    )
    lifecycle::deprecate_warn(
      "2.12.0.9003",
      "generate(exclude)",
      "generate(used)",
      paste0(
        "If auto-detection doesn't work, ",
        "use `bru_used(effect_exclude = exclude)`"
      ),
      env = env,
      user_env = user_env
    )
  } else {
    exclude <- NULL
  }
  list(include = include, exclude = exclude)
}

#' Sampling based on bru posteriors
#'
#' @description
#' Takes a fitted `bru` object produced by the function [bru()] and produces
#' samples given a new set of values for the model covariates or the original
#' values used for the model fit. The samples can be based on any R expression
#' that is valid given these values/covariates and the joint
#' posterior of the estimated random effects.
#'
#' @export
#' @family sample generators
#' @param object A `bru` object obtained by calling [bru()].
#' @param newdata A `data.frame` or `SpatialPointsDataFrame` of covariates
#'   needed for sampling.
#' @param formula A formula where the right hand side defines an R expression
#' to evaluate for each generated sample. If `NULL`, the latent and
#' hyperparameter states are returned as named list elements.
#' See Details for more information.
#' @param n.samples Integer setting the number of samples to draw in order to
#' calculate the posterior statistics.
#' The default, 100, is rather low but provides a quick approximate result.
#' @param seed Random number generator seed passed on to
#'   `INLA::inla.posterior.sample`
#' @param num.threads Specification of desired number of threads for parallel
#' computations. Default NULL, leaves it up to INLA.
#' When seed != 0, overridden to "1:1:1"
#' @param used Either `NULL` or a [bru_used()] object.
#'   Default, `NULL`, uses auto-detection of used variables in the formula.
#' @param \dots additional, unused arguments.
#' @param include,exclude `r lifecycle::badge("deprecated")` If auto-detection
#' of used variables fails, use `used` instead.
#' @details
#' In addition to the component names (that give the effect of each component
#' evaluated for the input data), the suffix `_latent` variable name can be used
#' to directly access the latent state for a component, and the suffix function
#' `_eval` can be used to evaluate a component at other input values than the
#' expressions defined in the component definition itself, e.g.
#' `field_eval(cbind(x, y))` for a component that was defined with
#' `field(coordinates, ...)` (see also [bru_comp_eval()]).
#'
#' For "iid" models with `mapper = bm_index(n)`, `rnorm()` is used to
#' generate new realisations for indices greater than `n`, if accessed
#' via `<name>_eval(...)`.
#'
#' @return List of generated samples
#' @seealso [predict.bru]
#' @rdname generate

generate.bru <- function(object,
                         newdata = NULL,
                         formula = NULL,
                         n.samples = 100,
                         seed = 0L,
                         num.threads = NULL,
                         used = NULL,
                         ...,
                         include = deprecated(),
                         exclude = deprecated()) {
  object <- bru_check_object_bru(object)

  # Convert data into list, data.frame or a Spatial object if not provided as
  # such
  if (is.character(newdata)) {
    newdata <- as.list(setNames(newdata, newdata))
  } else if (inherits(newdata, "formula")) {
    stop(paste0(
      "Formula supplied as data to generate.bru(). ",
      "Please check your argument order/names."
    ))
  }

  state <- evaluate_state(
    object$bru_info$model,
    result = object,
    property = "sample",
    n = n.samples,
    seed = seed,
    num.threads = num.threads,
    ...
  )
  if (is.null(formula)) {
    state
  } else {
    # TODO: clarify the output format, and use the format parameter

    inc_exc <- bru_generate_check_used_deprecation(
      include,
      exclude
    )
    if (is.null(used)) {
      used <-
        bru_used(
          formula,
          effect = if (is.null(formula)) {
            character(0)
          } else {
            inc_exc$include
          },
          effect_exclude = inc_exc$exclude,
          latent = NULL
        )
    }
    used <- bru_used_update(used, labels = names(as_bru_comp_list(object)))

    pred <- new_bru_pred_expr(formula, used = used)

    if (inherits(newdata, "sf")) {
      # Make sure coordinates have unified XY/XYZ/XYM/XYZM structure:
      newdata <- fmesher::fm_zm(newdata)
    }

    vals <- evaluate_model(
      model = object$bru_info$model,
      state = state,
      data = newdata,
      predictor = pred
    )
    vals
  }
}


# Monte Carlo method for estimating posterior
#
# @export
# @param dfun A function returning a density for given x
# @param sfun A function providing samples from a posterior
# @param x Inital domain on which to perform the estimation. This will be
# adjusted as more samples are generated.
# @param samples An initial set of samples. Not required but will be used to
# estimate the inital domain \code{x} if \code{x} is \code{NULL}
# @param mcerr Monte Carlo error at which to stop the chain
# @param n Inital number of samples. This will be doubled for each iteration.
# @param discrete Set this to \code{TRUE} if the density is only defined for
#   integer \code{x}
# @param verbose Be verbose?

montecarlo.posterior <- function(dfun, sfun, x = NULL, samples = NULL,
                                 mcerr = 0.01, n = 100, discrete = FALSE,
                                 verbose = FALSE) {
  .Deprecated()
  xmaker <- function(hpd) {
    mid <- (hpd[2] + hpd[1]) / 2
    rg <- (hpd[2] - hpd[1]) / 2
    seq(mid - 1.2 * rg, mid + 1.2 * rg, length.out = 256)
  }
  xmaker2 <- function(hpd) {
    seq(hpd[1], hpd[2], length.out = 256)
  }

  initial.xmaker <- function(smp) {
    mid <- median(smp)
    rg <- (quantile(smp, 0.975) - quantile(smp, 0.25)) / 2
    seq(mid - 3 * rg, mid + 3 * rg, length.out = 256)
  }

  # Inital samples
  if (is.null(samples)) {
    samples <- sfun(n)
  }

  # Inital HPD
  if (is.null(x)) {
    x <- initial.xmaker(as.vector(unlist(samples)))
  }

  # Round x if needed
  if (discrete) x <- unique(round(x))

  # First density estimate
  lest <- dfun(x, samples)


  converged <- FALSE
  while (!converged) {
    # Compute last HPD interval
    xnew <- xmaker2(INLA::inla.hpdmarginal(0.999, list(x = x, y = lest)))

    # Map last estimate to the HPD interval
    if (discrete) xnew <- unique(round(xnew))
    lest <- INLA::inla.dmarginal(xnew, list(x = x, y = lest))
    x <- xnew

    # Sample new density
    n <- 2 * n
    samples <- sfun(n)
    est <- dfun(x, samples)

    # Compute Monte Carlo error
    # err = sd(est/sum(est)-lest/sum(lest))
    err <- max(((est - lest) / max(lest))^2)

    # Plot new density estimate versus old one (debugging)
    if (verbose) {
      cat(glue("hpd: ({min(x)}, {max(x)}), err = {err}, n = {n}"), "\n")
      # plot(x, lest, type = "l") ; lines(x, est, type = "l", col = "red")
    }

    # Convergence?
    if (err < mcerr) {
      converged <- TRUE
    } else {
      lest <- (est + lest) / 2
    }
  }

  marg <- list(x = x, y = est, samples = samples, mcerr = err)

  # Append some important statistics
  marg$quantiles <- INLA::inla.qmarginal(c(0.025, 0.5, 0.975), marg)
  marg$mean <- INLA::inla.emarginal(identity, marg)
  marg$sd <- sqrt(INLA::inla.emarginal(function(x) x^2, marg) - marg$mean^2)
  marg$cv <- marg$sd / marg$mean
  marg$mce <- err

  marg
}


#' Summarise and annotate data
#'
#' @export
#' @param data A list of samples, each either numeric or a `data.frame`
#' @param probs A numeric vector of probabilities with values in `[0, 1]`,
#'   passed to `stats::quantile`
#' @param x A `data.frame` of data columns that should be added to the summary
#'   data frame
#' @param cbind.only If TRUE, only `cbind` the samples and return a matrix where
#'   each column is a sample
#' @param max_moment integer, at least 2. Determines the largest moment
#'   order information to include in the output. If `max_moment > 2`,
#'   includes "skew" (skewness, `E[(x-m)^3/s^3]`), and
#'   if `max_moment > 3`, includes
#'   "ekurtosis" (excess kurtosis, `E[(x-m)^4/s^4] - 3`). Default 2.
#'   Note that the Monte Carlo variability of the `ekurtois` estimate may be
#'   large.
#' @return A `data.frame` or `Spatial[Points/Pixels]DataFrame` with summary
#'   statistics, "mean", "sd", `paste0("q", probs)`, "mean.mc_std_err",
#'   "sd.mc_std_err"
#'
#' @examples
#' bru_summarise(matrix(rexp(10000), 10, 1000), max_moment = 4, probs = NULL)
#'
#' @keywords internal
bru_summarise <- function(data, probs = c(0.025, 0.5, 0.975),
                          x = NULL, cbind.only = FALSE,
                          max_moment = 2) {
  if (is.list(data)) {
    data <- do.call(cbind, data)
  }
  N <- NCOL(data)
  if (cbind.only) {
    smy <- data.frame(data)
    colnames(smy) <- paste0("sample.", seq_len(N))
  } else {
    if (length(probs) == 0) {
      smy <- data.frame(
        rowMeans(data, na.rm = TRUE),
        apply(data, MARGIN = 1, sd, na.rm = TRUE)
      )
      qs_names <- NULL
    } else if (length(probs) == 1) {
      smy <- data.frame(
        rowMeans(data, na.rm = TRUE),
        apply(data, MARGIN = 1, sd, na.rm = TRUE),
        as.matrix(apply(data,
          MARGIN = 1, quantile, probs = probs, na.rm = TRUE,
          names = FALSE
        ))
      )
      qs_names <- paste0("q", probs)
    } else {
      smy <- data.frame(
        rowMeans(data, na.rm = TRUE),
        apply(data, MARGIN = 1, sd, na.rm = TRUE),
        t(apply(data,
          MARGIN = 1, quantile, probs = probs, na.rm = TRUE,
          names = FALSE
        ))
      )
      qs_names <- paste0("q", probs)
    }
    colnames(smy) <- c("mean", "sd", qs_names)
    # For backwards compatibility, add a median column:
    if (any(qs_names == "q0.5")) {
      smy[["median"]] <- smy[["q0.5"]]
    }

    # Get 3rd and 4rt central moments
    # Use 1/N normalisation of the sample sd
    skew <- rowMeans(
      ((data - smy$mean) / smy$sd)^3 * (N / (N - 1))^3,
      na.rm = TRUE
    )
    if (max_moment >= 3) {
      smy[["skew"]] <- skew
    }
    # eK + 3 >= skew^2 + 1
    # eK >= skew^2 - 2
    # Use 1/N normalisation of the sample sd
    ekurtosis <- pmax(
      skew^2 - 2,
      rowMeans(
        ((data - smy$mean) / smy$sd)^4 * (N / (N - 1))^4 - 3,
        na.rm = TRUE
      )
    )

    # Add Monte Carlo standard errors
    # Var(s) \approx (eK + 2) \sigma^2 / (4 n):
    # +2 replaced by 3-(n-3)/(n-1) = 2n/(n-1), from Rao 1973, p438
    sd.mc_std_err <-
      sqrt(pmax(0, ekurtosis + 2 * N / (N - 1))) *
        smy[["sd"]] / sqrt(4 * N)
    # Include sd MC error in estimate of mean MC error:
    smy[["mean.mc_std_err"]] <- (smy[["sd"]] + sd.mc_std_err * 2) / sqrt(N)
    smy[["sd.mc_std_err"]] <- sd.mc_std_err
    if (max_moment >= 3) {
      smy[["skew"]] <- skew
    }
    if (max_moment >= 4) {
      smy[["ekurtosis"]] <- ekurtosis
    }
  }
  if (!is.null(x)) {
    smy <- expand_to_dataframe(x, smy)
  }
  smy
}


lin_predictor <- function(lin, state) {
  do.call(
    c,
    lapply(
      lin,
      function(x) {
        as.vector(ibm_eval(x, state = state))
      }
    )
  )
}
nonlin_predictor <- function(param, state) {
  do.call(
    c,
    lapply(
      seq_along(param[["lhoods"]]),
      function(lh_idx) {
        as.vector(
          evaluate_model(
            model = param[["model"]],
            data = param[["lhoods"]][[lh_idx]][["data"]],
            data_extra = param[["lhoods"]][[lh_idx]][["data_extra"]],
            input = param[["input"]][[lh_idx]],
            state = list(state),
            comp_simple = param[["comp_simple"]][[lh_idx]],
            predictor = bru_pred_expr(param[["lhoods"]][[lh_idx]],
              format = "expr"
            ),
            format = "matrix",
            n_pred = bru_response_size(param[["lhoods"]][[lh_idx]])
          )
        )
      }
    )
  )
}
scale_state <- function(state0, state1, scaling_factor) {
  new_state <- lapply(
    names(state0),
    function(idx) {
      state0[[idx]] + (state1[[idx]] - state0[[idx]]) * scaling_factor
    }
  )
  names(new_state) <- names(state0)
  new_state
}

line_search_opt_target <- function(x, param) {
  (x - 1)^2 * param[1] + 2 * (x - 1) * x^2 * param[2] + x^4 * param[3]
}

line_search_opt_target_exact <- function(x, param, nonlin_param) {
  state <- scale_state(param$state0, param$state1, x)
  nonlin <- nonlin_predictor(
    param = nonlin_param,
    state = state
  )
  sum((nonlin - param$lin)^2 * param$weights)
}


# @title FUNCTION_TITLE
# @description FUNCTION_DESCRIPTION
# @param model PARAM_DESCRIPTION
# @param lhoods PARAM_DESCRIPTION
# @param lin PARAM_DESCRIPTION
# @param state0 PARAM_DESCRIPTION
# @param state PARAM_DESCRIPTION
# @param A PARAM_DESCRIPTION
# @param weights PARAM_DESCRIPTION
# @param options PARAM_DESCRIPTION
# @return OUTPUT_DESCRIPTION
# @details DETAILS
# @examples
# \dontrun{
# if (interactive()) {
#  #EXAMPLE1
#  }
# }
# @export
# @rdname bru_line_search

bru_line_search <- function(model,
                            lhoods,
                            lin,
                            state0,
                            state,
                            input,
                            comp_lin,
                            comp_simple,
                            weights = 1,
                            options) {
  # Metrics ----
  pred_scalprod <- function(delta1, delta2) {
    if (!all(is.finite(delta1)) || !all(is.finite(delta2))) {
      Inf
    } else {
      sum(delta1 * delta2 * weights)
    }
  }
  pred_norm2 <- function(delta) {
    pred_scalprod(delta, delta)
  }
  pred_norm <- function(delta) {
    pred_scalprod(delta, delta)^0.5
  }

  # Is line search enabled? ----
  if (length(options$bru_method$search) == 0) {
    return(
      list(
        active = FALSE,
        step_scaling = 1,
        state = state
      )
    )
  }

  # Initialise ----
  if (is.null(weights)) {
    bru_log_warn(
      paste0(
        "NULL weights detected for line search. Using weights = 1 instead.",
        "\n\tThis is a bug in the inlabru package. Please notify the developer."
      )
    )
    weights <- 1
  }

  fact <- options$bru_method$factor

  nonlin_param <- list(
    model = model,
    lhoods = lhoods,
    input = input,
    comp_simple = comp_simple
  )

  state1 <- state
  lin_pred0 <- lin_predictor(lin, state0)
  lin_pred1 <- lin_predictor(lin, state1)
  nonlin_pred <- nonlin_predictor(
    param = nonlin_param,
    state = state1
  )

  if (length(lin_pred1) != length(nonlin_pred)) {
    bru_log_warn(
      paste0(
        "Please notify the inlabru package developer:",
        "\n\tThe line search linear and nonlinear predictors have ",
        "different lengths.",
        "\n\tThis should not happen!"
      )
    )
  }

  step_scaling <- 1

  if (isTRUE(options[["bru_debug"]])) {
    df_debug <- data.frame(
      idx = seq_along(lin_pred0),
      lin0 = lin_pred0,
      lin1 = lin_pred1,
      nonlin1 = nonlin_pred,
      weights = weights
    )
  }

  norm01 <- pred_norm(lin_pred1 - lin_pred0)
  norm1 <- pred_norm(nonlin_pred - lin_pred1)

  do_finite <-
    any(c("all", "finite") %in% options$bru_method$search)
  do_contract <-
    any(c("all", "contract") %in% options$bru_method$search)
  do_expand <-
    any(c("all", "expand") %in% options$bru_method$search)
  do_optimise <-
    any(c("all", "optimise") %in% options$bru_method$search)
  maximum_step <- options$bru_method$max_step

  finite_active <- 0
  contract_active <- 0
  expand_active <- 0

  # Contraction methods ----
  if (do_finite || do_contract) {
    nonfin <- !all(is.finite(nonlin_pred))
    norm0 <- pred_norm(nonlin_pred - lin_pred0)
    norm1 <- pred_norm(nonlin_pred - lin_pred1)

    while ((do_finite && nonfin) ||
      (do_contract && ((norm0 > norm01 * fact) || (norm1 > norm01 * fact)))) {
      if (do_finite && nonfin) {
        finite_active <- finite_active - 1
      } else {
        contract_active <- contract_active - 1
      }
      step_scaling <- step_scaling / fact
      state <- scale_state(state0, state1, step_scaling)
      nonlin_pred <- nonlin_predictor(
        param = nonlin_param,
        state = state
      )
      nonfin <- !all(is.finite(nonlin_pred))
      norm0 <- pred_norm(nonlin_pred - lin_pred0)
      norm1 <- pred_norm(nonlin_pred - lin_pred1)

      bru_log_message(
        glue(
          "iinla: Step rescaling: {signif(100 * step_scaling, 4)}%, Contract",
          " (",
          "norm0 = {signif(norm0, 4)}, ",
          "norm1 = {signif(norm1, 4)}, ",
          "norm01 = {signif(norm01, 4)}",
          ")"
        ),
        verbosity = 3
      )

      if (step_scaling < 2^-10) {
        break
      }
    }
  }

  # Expansion method ----
  if (do_expand &&
    finite_active == 0 &&
    contract_active == 0) {
    overstep <- FALSE
    norm0 <- pred_norm(nonlin_pred - lin_pred0)
    norm1 <- pred_norm(nonlin_pred - lin_pred1)
    while (!overstep && (norm0 < norm01)) {
      expand_active <- expand_active + 1
      step_scaling <- step_scaling * fact
      state <- scale_state(state0, state1, step_scaling)

      nonlin_pred <- nonlin_predictor(
        param = nonlin_param,
        state = state
      )
      norm1_prev <- norm1
      norm0 <- pred_norm(nonlin_pred - lin_pred0)
      norm1 <- pred_norm(nonlin_pred - lin_pred1)
      overstep <-
        (norm1 > norm1_prev) || !is.finite(norm1)

      bru_log_message(
        glue(
          "iinla: Step rescaling: {signif(100 * step_scaling, 3)}%, Expand",
          " (",
          "norm0 = {signif(norm0, 4)}, ",
          "norm1 = {signif(norm1, 4)}, ",
          "norm01 = {signif(norm01, 4)}",
          ")"
        ),
        verbosity = 3
      )

      if (step_scaling > 2^10) {
        break
      }
    }
    # Overstep ----
    if (((step_scaling > 1) && overstep) || !is.finite(norm1)) {
      expand_active <- expand_active - 1
      step_scaling <- step_scaling / fact
      state <- scale_state(state0, state1, step_scaling)
      nonlin_pred <- nonlin_predictor(
        param = nonlin_param,
        state = state
      )
      norm0 <- pred_norm(nonlin_pred - lin_pred0)
      norm1 <- pred_norm(nonlin_pred - lin_pred1)

      bru_log_message(
        glue(
          "iinla: Step rescaling: {signif(100 * step_scaling, 3)}%, Overstep",
          " (",
          "norm0 = {signif(norm0, 4)}, ",
          "norm1 = {signif(norm1, 4)}, ",
          "norm01 = {signif(norm01, 4)}",
          ")"
        ),
        verbosity = 3
      )
    }
  }

  # Optimisation method ----
  if (do_optimise) {
    lin_pred <- lin_predictor(lin, state)
    delta_lin <- (lin_pred1 - lin_pred0)
    delta_nonlin <- (nonlin_pred - lin_pred) / step_scaling^2
    alpha <-
      optimise(
        line_search_opt_target,
        step_scaling * c(1 / fact^2, fact),
        param = c(
          pred_norm2(delta_lin),
          pred_scalprod(delta_lin, delta_nonlin),
          pred_norm2(delta_nonlin)
        )
      )
    step_scaling_opt_approx <- alpha$minimum

    state_opt <- scale_state(state0, state1, step_scaling_opt_approx)
    nonlin_pred_opt <- nonlin_predictor(
      param = nonlin_param,
      state = state_opt
    )
    norm0_opt <- pred_norm(nonlin_pred_opt - lin_pred0)
    norm1_opt <- pred_norm(nonlin_pred_opt - lin_pred1)

    bru_log_message(
      glue(
        "iinla: Step rescaling: {signif(100 * step_scaling_opt_approx, 4)}%,",
        " Approx Optimisation",
        " (",
        "norm0 = {signif(norm0_opt, 4)}, ",
        "norm1 = {signif(norm1_opt, 4)}, ",
        "norm01 = {signif(norm01, 4)}",
        ")"
      ),
      verbosity = 3
    )

    if (norm1_opt > norm01) {
      bru_log_message(
        glue(
          "iinla: norm1_opt > |delta|: ",
          "{signif(norm1_opt, 4)}",
          " > ",
          "{signif(norm01, 4)}"
        ),
        verbosity = 3
      )
    }

    if ((norm1_opt > norm01) ||
      identical(options$bru_method$line_opt_method, "full")) {
      alpha <-
        optimise(
          line_search_opt_target_exact,
          step_scaling * c(0, fact),
          param = list(
            lin = lin_pred1,
            state0 = state0,
            state1 = state1,
            weights = weights
          ),
          nonlin_param = nonlin_param
        )

      step_scaling_opt <- alpha$minimum
      state_opt <- scale_state(state0, state1, step_scaling_opt)
      nonlin_pred_opt <- nonlin_predictor(
        param = nonlin_param,
        state = state_opt
      )
      norm0_opt <- pred_norm(nonlin_pred_opt - lin_pred0)
      norm1_opt <- pred_norm(nonlin_pred_opt - lin_pred1)

      bru_log_message(
        glue(
          "iinla: Step rescaling: {signif(100 * step_scaling_opt, 4)}%,",
          " Optimisation",
          " (",
          "norm0 = {signif(norm0_opt, 4)}, ",
          "norm1 = {signif(norm1_opt, 4)}, ",
          "norm01 = {signif(norm01, 4)}",
          ")"
        ),
        verbosity = 3
      )

      if (norm1_opt > norm01) {
        bru_log_message(
          glue(
            "iinla: norm1_opt > |delta|: ",
            "{signif(norm1_opt, 4)}",
            " > ",
            "{signif(norm01, 4)}"
          ),
          verbosity = 3
        )
      }
    } else {
      step_scaling_opt <- step_scaling_opt_approx
    }

    if (norm1_opt < norm1) {
      step_scaling <- step_scaling_opt
      state <- state_opt
      nonlin_pred <- nonlin_pred_opt
      norm0 <- norm0_opt
      norm1 <- norm1_opt
    } else {
      bru_log_message(
        paste0("iinla: Optimisation did not improve on previous solution."),
        verbosity = 3
      )
    }
  }

  bru_log_message(
    glue("
      iinla: |lin1-lin0| = {signif(pred_norm(lin_pred1 - lin_pred0), 4)}
        <eta-lin1,delta>/|delta| = {signif(rel1, 4)}
        |eta-lin0 - delta <delta,eta-lin0>/<delta,delta>| = {signif(rel2, 4)}",
      rel1 = pred_scalprod(nonlin_pred - lin_pred1, lin_pred1 - lin_pred0) /
        pred_norm(lin_pred1 - lin_pred0),
      rel2 = pred_norm(nonlin_pred - lin_pred0 - (lin_pred1 - lin_pred0) *
        pred_scalprod(lin_pred1 - lin_pred0, nonlin_pred - lin_pred0) /
        pred_norm2(lin_pred1 - lin_pred0))
    ),
    verbosity = 4
  )

  # Prevent long steps ----
  if (step_scaling > maximum_step) {
    step_scaling <- maximum_step
    state <- scale_state(state0, state1, step_scaling)
    nonlin_pred <- nonlin_predictor(
      param = nonlin_param,
      state = state
    )
    norm0 <- pred_norm(nonlin_pred - lin_pred0)
    norm1 <- pred_norm(nonlin_pred - lin_pred1)

    bru_log_message(
      glue(
        "iinla: Step rescaling: ",
        "{signif(100 * step_scaling, 3)}%, ",
        "Maximum step length",
        " (",
        "norm0 = {signif(norm0, 4)}, ",
        "norm1 = {signif(norm1, 4)}, ",
        "norm01 = {signif(norm01, 4)}",
        ")"
      ),
      verbosity = 3
    )
  }

  active <-
    (finite_active + contract_active + expand_active != 0) ||
      (abs(step_scaling - 1) > 0.001)

  if (active && (options$bru_verbose <= 2)) {
    bru_log_message(
      glue(
        "iinla: Step rescaling: ",
        "{signif(100 * step_scaling, 3)}%",
        " (",
        "norm0 = {signif(norm0, 4)}, ",
        "norm1 = {signif(norm1, 4)}, ",
        "norm01 = {signif(norm01, 4)}",
        ")"
      ),
      verbosity = 2
    )
  }

  if (isTRUE(options[["bru_debug"]])) {
    df_debug$nonlinopt <- nonlin_pred

    requireNamespace("ggplot2")
    requireNamespace("patchwork")
    pl1 <- ggplot2::ggplot(df_debug) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        (.data$lin0 - .data$lin1),
        col = "start"
      )) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        (.data$lin1 - .data$lin1),
        col = "inla"
      )) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        (.data$nonlin1 - .data$lin1),
        col = "full"
      )) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        (.data$nonlinopt - .data$lin1),
        col = "opt"
      )) +
      ggplot2::geom_abline(slope = 0, intercept = 0) +
      ggplot2::scale_color_discrete(
        breaks = c("start", "inla", "full", "opt")
      ) +
      ggplot2::ylab("eta - lin1")
    pl2 <- ggplot2::ggplot(df_debug) +
      ggplot2::geom_point(ggplot2::aes(
        .data$idx,
        (.data$lin0 - .data$lin1) * .data$weights^0.5,
        col = "start"
      )) +
      ggplot2::geom_point(ggplot2::aes(
        .data$idx,
        (.data$lin1 - .data$lin1) * .data$weights^0.5,
        col = "inla"
      )) +
      ggplot2::geom_point(ggplot2::aes(
        .data$idx,
        (.data$nonlin1 - .data$lin1) * .data$weights^0.5,
        col = "full"
      )) +
      ggplot2::geom_point(ggplot2::aes(
        .data$idx,
        (.data$nonlinopt - .data$lin1) * .data$weights^0.5,
        col = "opt"
      )) +
      ggplot2::geom_abline(slope = 0, intercept = 0) +
      ggplot2::scale_color_discrete(
        breaks = c("start", "inla", "full", "opt")
      ) +
      ggplot2::ylab("(eta - lin1) * sqrt(w)")
    pl3 <- ggplot2::ggplot(df_debug) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        .data$lin0,
        col = "start"
      )) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        .data$lin1,
        col = "inla"
      )) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        .data$nonlin1,
        col = "full"
      )) +
      ggplot2::geom_point(ggplot2::aes(.data$idx,
        .data$nonlinopt,
        col = "opt"
      )) +
      ggplot2::geom_ribbon(
        ggplot2::aes(
          as.numeric(.data$idx),
          ymin = .data$lin1 - 2 * .data$weights^-0.5,
          ymax = .data$lin1 + 2 * .data$weights^-0.5
        ),
        alpha = 0.1
      ) +
      ggplot2::geom_abline(slope = 0, intercept = 0) +
      ggplot2::scale_color_discrete(
        breaks = c("start", "inla", "full", "opt")
      ) +
      ggplot2::ylab("eta")
    pl4 <- ggplot2::ggplot(data.frame(
      idx = seq_along(unlist(state0)),
      state0 = unlist(state0),
      state1 = unlist(state1)[names(unlist(state0))],
      state = unlist(state)[names(unlist(state0))]
    )) +
      ggplot2::geom_point(
        ggplot2::aes(.data$idx, .data$state0, col = "start")
      ) +
      ggplot2::geom_point(
        ggplot2::aes(.data$idx, .data$state1, col = "inla")
      ) +
      ggplot2::geom_point(
        ggplot2::aes(.data$idx, .data$state1, col = "full")
      ) +
      ggplot2::geom_point(
        ggplot2::aes(.data$idx, .data$state, col = "opt")
      ) +
      ggplot2::scale_color_discrete(
        breaks = c("start", "inla", "full", "opt")
      ) +
      ggplot2::ylab("state")
    pl1234 <-
      ((pl1 | pl2) / (pl3 | pl4)) +
        patchwork::plot_layout(guides = "collect") &
        ggplot2::theme(legend.position = "right")

    # Compute deviations in the delta = lin1-lin0 direction and the orthogonal
    # direction
    # delta = lin1-lin0
    delta <- lin_pred1 - lin_pred0
    delta_norm <- norm01

    # <eta-lin0, delta>/|delta|
    along_opt <- pred_scalprod(nonlin_pred - lin_pred0, delta) / delta_norm
    # |eta-lin0 - delta <delta,eta-lin0>/<delta,delta>|
    ortho_opt <- pred_norm(
      nonlin_pred - lin_pred0 - delta * along_opt / delta_norm
    )

    step_scaling_range <- seq(0, step_scaling * fact, length.out = 20)
    along <- ortho <- distance <- numeric(length(step_scaling_range))
    for (iii in seq_along(step_scaling_range)) {
      step_scaling_ <- step_scaling_range[iii]
      state_ <- scale_state(state0, state1, step_scaling_)
      nonlin_pred_ <- nonlin_predictor(
        param = nonlin_param,
        state = state_
      )
      # <eta-lin0, delta>/|delta|
      along[iii] <- pred_scalprod(nonlin_pred_ - lin_pred0, delta) / delta_norm
      # |eta-lin0 - delta <delta,eta-lin0>/<delta,delta>|
      ortho[iii] <- pred_norm(
        nonlin_pred_ - lin_pred0 - delta * along[iii] / delta_norm
      )
      distance[iii] <-
        line_search_opt_target_exact(
          step_scaling_,
          param = list(
            lin = lin_pred1,
            state0 = state0,
            state1 = state1,
            weights = weights
          ),
          nonlin_param = nonlin_param
        )
    }


    pl0 <-
      ggplot2::ggplot(
        data =
          data.frame(
            along = along,
            ortho = ortho,
            distance = distance^0.5,
            what = "eta"
          ),
        mapping = ggplot2::aes(.data$along, .data$ortho, col = .data$what)
      ) +
      ggplot2::geom_path() +
      ggplot2::geom_point(ggplot2::aes(size = .data$distance)) +
      ggplot2::geom_point(
        data = data.frame(
          along = c(0, along_opt, delta_norm),
          ortho = c(0, ortho_opt, 0),
          what = c("eta0", "eta_opt", "eta1")
        )
      )

    # jarl-ignore browser: Special debug use
    # jarl-ignore undesirable_function: Special debug use
    browser()
  }

  list(
    active = active,
    step_scaling = step_scaling,
    state = state
  )
}

latent_names <- function(state) {
  nm <- lapply(
    names(state),
    function(x) {
      if (length(state[[x]]) == 1) {
        x
      } else {
        glue("{x}.{seq_along(state[[x]])}")
      }
    }
  )
  names(nm) <- names(state)
  nm
}
tidy_state <- function(state, value_name = "value") {
  df <- data.frame(
    effect = rep(names(state), lengths(state)),
    index = unlist(lapply(lengths(state), function(x) seq_len(x))),
    THEVALUE = unlist(state)
  )
  nm <- names(df)
  nm[nm == "THEVALUE"] <- value_name
  names(df) <- nm
  rownames(df) <- NULL
  df
}
tidy_states <- function(states, value_name = "value", id_name = "iteration") {
  df <- lapply(states, function(x) tidy_state(x, value_name = value_name))
  id <- rep(seq_along(states), each = nrow(df[[1]]))
  df <- do.call(rbind, df)
  df[[id_name]] <- id
  df
}


bru_timer_do <- function(timer, task = NULL, iteration = NA_integer_) {
  if (is.null(timer)) {
    timer <- data.frame(
      Task = character(0),
      Iteration = integer(0),
      Time = as.difftime(numeric(0), units = "secs"),
      System = as.difftime(numeric(0), units = "secs"),
      Elapsed = as.difftime(numeric(0), units = "secs")
    )
  }
  tag <- list(
    time = proc.time(),
    task = task,
    iter = iteration
  )
  old_tag <- attr(timer, "bru_timer_tag", exact = TRUE)
  if (!is.null(old_tag)) {
    ptm <- tag$time - old_tag$time
    if (!is.na(ptm[4])) {
      ptm[1] <- ptm[1] + ptm[4]
    }
    if (!is.na(ptm[5])) {
      ptm[2] <- ptm[2] + ptm[5]
    }

    if (is.null(old_tag$task)) {
      old_tag$task <- "Unknown"
    }
    if (is.null(old_tag$iter)) {
      old_tag$iter <- NA_integer_
    }
    timer <- rbind(
      timer,
      data.frame(
        Task = old_tag$task,
        Iteration = old_tag$iter,
        Time = as.difftime(unname(ptm[1]), units = "secs"),
        System = as.difftime(unname(ptm[2]), units = "secs"),
        Elapsed = as.difftime(unname(ptm[3]), units = "secs")
      )
    )
  }
  attr(timer, "bru_timer_tag") <- if (is.null(task)) {
    NULL
  } else {
    tag
  }
  timer
}
bru_timer_done <- function(timer) {
  bru_timer_do(timer, task = NULL)
}


#' Iterated INLA
#'
#' This is an internal wrapper for iterated runs of `INLA::inla`.
#' For nonlinear models, a linearisation is done with
#' `bru_compute_linearisation`, with a line search method between each
#' iteration. The `INLA::inla.stack` information is setup by [bru_make_stack()].
#'
#' @export
#' @param model A [bru_model] object
#' @param lhoods A list of likelihood objects from [bru_obs()]
#' @param inputs Optional pre-computed  list of per-likelihood component
#'   evaluations, from [bru_input.bru_obs_list()].
#' @param initial A previous `bru` result or a list of named latent variable
#' initial states (missing elements are set to zero), to be used as starting
#' point, or `NULL`. If non-null, overrides `options$bru_initial`
#' @param options A `bru_options` object.
#' @return An `iinla` object that inherits from `INLA::inla`, with an
#' added field `bru_iinla` with elements
#' \describe{
#' \item{log}{The diagnostic log messages produced by the run}
#' \item{states}{The list of linearisation points, one for each inla run}
#' \item{inla_stack}{The `inla.stack` object from the final inla run}
#' \item{track}{A list of convergence tracking vectors}
#' }
#' If an inla run is aborted by an error, the returned object also contains
#' an element `error` with the error object.
#' @keywords internal

iinla <- function(model, lhoods, inputs = NULL, initial = NULL, options) {
  options <- bru_call_options(options)
  bru_options_set_local(options, .reset = TRUE)

  timings <- bru_timer_do(NULL, "Preprocess", 1L)

  inla.options <- bru_options_inla(options)

  bru_log_bookmark("iinla")
  orig_timings <- NULL
  original_log <- character(0) # Updated further below
  # Local utility method for collecting information object:
  collect_misc_info <- function(...) {
    timings <- bru_timer_done(timings)
    if (is.null(orig_track)) {
      track_df <- list()
      for (label in names(states[[1]])) {
        if (length(states[[1]][[label]]) > 0) {
          track_df[[label]] <-
            data.frame(
              effect = label,
              index = seq_along(states[[1]][[label]]),
              iteration = 0L,
              mode = NA_real_,
              sd = NA_real_,
              new_linearisation = states[[1]][[label]]
            )
        }
      }
      orig_track <- do.call(rbind, track_df)
    }

    if (is.null(orig_inla_track)) {
      orig_inla_track <- tibble::tibble(
        iteration = integer(0),
        iteration_part = numeric(0),
        f = numeric(0),
        nfunc = integer(0),
        nfunc_total = integer(0)
      )
      nfunc_offset <- 0L
    } else {
      nfunc_offset <- max(orig_inla_track$nfunc_total)
    }

    track_names <- if (length(track) > 0) {
      names(track[[1]])
    } else {
      c(
        "effect",
        "index",
        "iteration",
        "mode",
        "sd",
        "new_linearisation"
      )
    }

    track <-
      if (is.null(orig_track) ||
        setequal(names(orig_track), track_names)) {
        do.call(dplyr::bind_rows, c(list(orig_track), track))
      } else {
        track <- do.call(dplyr::bind_rows, track)
        original_names <- names(orig_track)
        new_names <- names(track)
        for (nn in setdiff(new_names, original_names)) {
          orig_track[[nn]] <- NA
        }
        for (nn in setdiff(original_names, new_names)) {
          track[[nn]] <- NA
        }
        dplyr::bind_rows(orig_track, track)
      }

    inla_track <- {
      offsets <- cumsum(vapply(
        seq_along(inla_track),
        function(k) {
          if (nrow(inla_track[[k]]) > 0) {
            max(inla_track[[k]]$nfunc)
          } else {
            0L
          }
        },
        0L
      ))
      offsets <- nfunc_offset + c(0, offsets)
      for (k in seq_along(inla_track)) {
        if (nrow(inla_track[[k]]) > 0) {
          inla_track[[k]]$nfunc_total <- inla_track[[k]]$nfunc + offsets[k]
        }
      }
      ## In case the number of theta values changes during the iterations,
      ## e.g.
      ## bru_rerun(..., options = list(control.mode = list(fixed = TRUE)))
      orig_n <- NCOL(orig_inla_track[["theta"]])
      if (length(inla_track) > 0) {
        new_n <- vapply(inla_track, function(x) NCOL(x[["theta"]]), 0L)
      } else {
        new_n <- orig_n
      }
      if (max(new_n) > orig_n) {
        if (is.null(orig_inla_track[["theta"]])) {
          orig_inla_track[["theta"]] <- matrix(NA_real_,
            nrow = NROW(orig_inla_track),
            ncol = max(new_n)
          )
        } else {
          orig_inla_track[["theta"]] <-
            cbind(
              orig_inla_track[["theta"]],
              matrix(NA_real_,
                nrow = NROW(orig_inla_track),
                ncol = max(new_n) - orig_n
              )
            )
        }
      }
      orig_n <- NCOL(orig_inla_track[["theta"]])
      if (any(new_n != orig_n)) {
        inla_track <- lapply(
          inla_track,
          function(x) {
            if (NCOL(x[["theta"]]) < orig_n) {
              if (is.null(x[["theta"]])) {
                x[["theta"]] <-
                  matrix(NA_real_,
                    nrow = NROW(x),
                    ncol = orig_n
                  )
              } else {
                x[["theta"]] <- cbind(
                  x[["theta"]],
                  matrix(NA_real_,
                    nrow = NROW(x),
                    ncol = orig_n - NCOL(x[["theta"]])
                  )
                )
              }
            }
            x
          }
        )
      }
      inla_track <- dplyr::bind_rows(inla_track)
      inla_track$iteration_part <-
        (seq_len(NROW(inla_track)) - 1L) / NROW(inla_track)
      dplyr::bind_rows(orig_inla_track, inla_track)
    }

    list(
      log = c(original_log, bru_log()["iinla"]),
      states = states,
      inla_stack = stk,
      track = track,
      inla_track = inla_track,
      timings = {
        iteration_offset <- if (is.null(orig_timings)) {
          0L
        } else {
          max(c(0L, orig_timings$Iteration), na.rm = TRUE)
        }
        rbind(
          orig_timings,
          data.frame(
            Task = timings$Task,
            Iteration = timings$Iteration + iteration_offset,
            Time = timings$Time,
            System = timings$System,
            Elapsed = timings$Elapsed
          )
        )
      },
      ...
    )
  }

  if (!is.null(inla.options[["offset"]])) {
    stop(paste0(
      "An offset option was specified which may interfere with the ",
      "inlabru model construction.\n",
      "Please use an explicit constant component instead; ",
      "e.g. ~ myoffset(value, model = 'const')"
    ))
  }

  # Initialise required local options
  inla.options <- modifyList(
    inla.options,
    list(
      control.mode = list(),
      control.predictor = list(compute = TRUE),
      control.compute = list()
    )
  )

  # Extract the family of each likelihood
  family <- bru_obs_inla_family(lhoods)

  # Extract the control.family information for each likelihood
  inla.options[["control.family"]] <-
    bru_obs_control_family(lhoods, inla.options[["control.family"]])

  # Combine the control.gcpo information from all the likelihoods
  control.gcpo.combined <-
    bru_obs_control_gcpo(
      lhoods,
      inla.options[["control.compute"]][["control.gcpo"]]
    )
  inla.options[["control.compute"]][["control.gcpo"]] <- control.gcpo.combined

  initial <-
    if (is.null(initial)) {
      options[["bru_initial"]]
    } else {
      initial
    }
  result <- NULL
  if (is.null(initial) || inherits(initial, "bru")) {
    if (!is.null(initial[["bru_iinla"]][["states"]])) {
      states <- initial[["bru_iinla"]][["states"]]
      # The last linearisation point will be used again:
      states <- c(states, states[length(states)])
    } else {
      # Set old result
      states <- evaluate_state(model,
        lhoods = lhoods,
        result = initial,
        property = "joint_mode"
      )
    }
    if (inherits(initial, "bru")) {
      result <- initial
    }
  } else if (is.list(initial)) {
    comp_lst <- as_bru_comp_list(model)
    state <- initial[intersect(names(comp_lst), names(initial))]
    for (lab in names(comp_lst)) {
      if (is.null(state[[lab]])) {
        state[[lab]] <- rep(0, ibm_n(comp_lst[[lab]][["mapper"]]))
      } else if (length(state[[lab]]) == 1) {
        state[[lab]] <-
          rep(
            state[[lab]],
            ibm_n(comp_lst[[lab]][["mapper"]])
          )
      }
    }
    states <- list(state)
    result <- NULL
  } else {
    stop("Unknown previous result information class")
  }
  old.result <- result

  # Preserve old log output
  original_log <- bru_log(
    if (is.null(old.result)) {
      character(0)
    } else if (is.null(old.result[["bru_iinla"]])) {
      character(0)
    } else if (is.null(old.result[["bru_iinla"]][["log"]])) {
      character(0)
    } else {
      old.result[["bru_iinla"]][["log"]]
    }
  )
  bru_log_message(
    "iinla: Start",
    verbosity = 3
  )

  # Track variables
  track <- list()
  if (is.null(old.result[["bru_iinla"]][["track"]])) {
    orig_track <- NULL
    track_size <- 0
  } else {
    orig_track <- old.result[["bru_iinla"]][["track"]]
    track_size <- max(orig_track[["iteration"]])
  }
  inla_track <- list()
  if (is.null(old.result[["bru_iinla"]][["inla_track"]]) ||
    (NROW(old.result[["bru_iinla"]][["inla_track"]]) == 0L)) {
    orig_inla_track <- NULL
    inla_track_size <- 0L
  } else {
    orig_inla_track <- old.result[["bru_iinla"]][["inla_track"]]
    inla_track_size <- max(orig_inla_track[["iteration"]])
  }

  # Preserve old timings
  if (!is.null(old.result[["bru_iinla"]][["timings"]])) {
    orig_timings <- old.result[["bru_iinla"]][["timings"]]
  }

  if (is.null(inputs)) {
    bru_log_message(
      "iinla: Evaluate component inputs",
      verbosity = 3
    )
    inputs <- bru_input(model, lhoods = lhoods)
  }

  timings <- bru_timer_do(timings, "Linearise", 1L)
  bru_log_message(
    "iinla: Evaluate component linearisations",
    verbosity = 3
  )
  comp_lin <- ibm_linear(
    model,
    input = inputs,
    state = states[[length(states)]],
    inla_f = TRUE
  )
  bru_log_message(
    "iinla: Evaluate component simplifications",
    verbosity = 3
  )
  comp_simple <- ibm_simplify(
    model,
    input = inputs,
    inla_f = TRUE
  )
  bru_log_message(
    "iinla: Evaluate predictor linearisation",
    verbosity = 3
  )
  lin <- bru_compute_linearisation(
    model,
    lhoods = lhoods,
    input = inputs,
    state = states[[length(states)]],
    comp_simple = comp_simple,
    options = options
  )

  do_line_search <- (length(options[["bru_method"]][["search"]]) > 0)
  if (do_line_search) {
    # Always compute linearisation (above)
  }

  bru_log_message(
    "iinla: Construct inla stack",
    verbosity = 3
  )
  # Initial stack
  idx <- bru_index(model, used = bru_used(lhoods))
  stk <- bru_make_stack(lhoods, lin, idx)

  if (utils::packageVersion("INLA") <= "24.06.02") {
    stk.data <- INLA::inla.stack.data(stk)
  } else {
    stk.data <- INLA::inla.stack.data(stk, .response.name = "BRU.response")
  }
  inla.options$control.predictor$A <- INLA::inla.stack.A(stk)

  N_response <- sum(bru_response_size(lhoods))
  N_predictor <- NROW(inla.options$control.predictor$A)
  if (N_response != N_predictor) {
    msg <- glue(
      "The total number of response values (N={N_response})",
      " and predictor values (N={N_predictor})",
      " do not match.\n",
      "  This is likely due to a mistake in the component or predictor",
      " constructions.",
      .trim = FALSE
    )
    if ((N_response > 1L) && (N_predictor == 1L)) {
      msg <- paste0(
        msg,
        "\nPerhaps you only have components with scalar inputs?"
      )
    }
    bru_log_message(msg, verbosity = 1)
    stop(msg)
  }

  bru_log_message(
    "iinla: Model initialisation completed",
    verbosity = 3
  )

  k <- 1L
  interrupt <- FALSE
  line_search <- list(
    active = FALSE,
    step_scaling = 1,
    state = states[[length(states)]]
  )

  if ((!is.null(result) && !is.null(result$mode))) {
    previous_x <- result$mode$x
  } else {
    # TODO: construct from the initial linearisation state instead
    previous_x <- 0
  }

  track_df <- NULL
  do_final_integration <- (options$bru_max_iter == 1)
  do_final_theta_no_restart <- FALSE
  while (!interrupt) {
    if (k > 1L) {
      timings <- bru_timer_do(timings, "Preprocess", k)
    }

    if ((k >= options$bru_max_iter) && !do_final_integration) {
      do_final_integration <- TRUE
      bru_log_message(
        "iinla: Maximum iterations reached, running final INLA integration.",
        verbosity = 1
      )
    }

    # When running multiple times propagate theta
    if ((k > 1) || (!is.null(result) && !is.null(result$mode))) {
      inla.options[["control.mode"]]$restart <- TRUE
      inla.options[["control.mode"]]$theta <- result$mode$theta
      inla.options[["control.mode"]]$x <- (
        previous_x +
          (result$mode$x - previous_x) * line_search[["step_scaling"]]
      )
      result_indexing <- inla_result_latent_idx(result)
      # Reset the predictor values, to avoid spurious iteration effects.
      inla.options[["control.mode"]]$x[result_indexing$APredictor] <- 0
      inla.options[["control.mode"]]$x[result_indexing$Predictor] <- 0

      inla.options[["control.inla"]]$use.directions <-
        result$misc$opt.directions

      # Further settings, from inla.rerun
      inla.options[["control.inla"]]$optimise.strategy <- "plain"

      inla.options[["control.inla"]]$step.factor <- 1
      inla.options[["control.inla"]]$tolerance.step <- 1e-10

      if (identical(
        result$.args$control.inla[["stencil"]],
        INLA::inla.set.control.inla.default()$stencil
      )) {
        inla.options$control.inla$stencil <- 9
      }

      h.def <- INLA::inla.set.control.inla.default()$h
      if (identical(
        result$.args$control.inla[["h"]], h.def
      )) {
        inla.options$control.inla$h <-
          0.2 * result$.args$control.inla[["h"]]
      } else {
        inla.options$control.inla$h <-
          max(0.1 * h.def, 0.75 * result$.args$control.inla[["h"]])
      }

      tol.def <- INLA::inla.set.control.inla.default()$tolerance
      if (identical(result$.args$control.inla[["tolerance"]], tol.def)) {
        inla.options$control.inla$tolerance <-
          result$.args$control.inla[["tolerance"]] * 0.01
      } else {
        inla.options$control.inla$tolerance <-
          max(tol.def^2, result$.args$control.inla[["tolerance"]] * 0.1)
      }
    }
    if ((!is.null(result) && !is.null(result$mode))) {
      previous_x <- result$mode$x
    }

    bru_log_message(
      glue("iinla: Iteration {k} [max: {options$bru_max_iter}]"),
      verbosity = 1L
    )

    # Return previous result if inla crashes, e.g. when connection to server is
    # lost
    old.result <- result
    result <- NULL

    inla.formula <- update.formula(model$formula, BRU.response ~ .)
    inla.data <-
      c(
        stk.data,
        do.call(c, c(
          lapply(
            as_bru_comp_list(model),
            function(xx) as.list(bru_comp_env_extra(xx))
          ),
          use.names = FALSE
        ))
      )
    #        list.data(model$formula))

    inla.options.merged <-
      modifyList(
        inla.options,
        list(
          formula = inla.formula,
          data = inla.data,
          family = family,
          E = stk.data[["BRU.E"]],
          Ntrials = stk.data[["BRU.Ntrials"]],
          weights = stk.data[["BRU.weights"]],
          scale = stk.data[["BRU.scale"]],
          offset = stk.data[["BRU.offset"]],
          control.predictor = list(link = stk.data[["BRU.link"]])
        )
      )
    if (do_final_integration) {
      if (do_final_theta_no_restart) {
        # Compute the minimal amount required
        inla.options.merged <-
          modifyList(
            inla.options.merged,
            list(
              control.mode = list(restart = FALSE)
            )
          )
      }
      inla.options.merged[["control.compute"]][["control.gcpo"]] <-
        control.gcpo.combined
    } else {
      # Compute the minimal amount required
      # Note: configs = TRUE only because latent indexing into mode$x is
      #   defined in configs$contents
      inla.options.merged <-
        modifyList(
          inla.options.merged,
          list(
            control.inla = list(
              control.vb = list(enable = FALSE),
              strategy = "gaussian",
              int.strategy = "eb"
            ),
            control.compute = list(
              config = TRUE,
              dic = FALSE,
              waic = FALSE
            ),
            # Required for line search weights:
            control.predictor = list(compute = TRUE)
          )
        )
      inla.options.merged[["control.compute"]][["control.gcpo"]] <- NULL
    }

    timings <- bru_timer_do(timings, "Run inla()", k)

    result <- fm_try_callstack(
      do.call(
        INLA::inla,
        inla.options.merged,
        envir = environment(as_bru_comp_list(model))
      )
    )

    timings <- bru_timer_do(timings, "Postprocess", k)

    if (inherits(result, "try-error")) {
      bru_log_warn(
        paste0("iinla: Problem in inla:\n", result)
      )
      bru_log_message(
        paste0(
          "iinla: Giving up and returning last successfully obtained result ",
          "for diagnostic purposes."
        ),
        verbosity = 1L,
        verbose = TRUE
      )
      if (is.null(old.result)) {
        old.result <- list()
      }
      if (!inherits(old.result, "iinla")) {
        class(old.result) <- c("iinla", class(old.result))
      }

      old.result[["bru_iinla"]] <- collect_misc_info()
      old.result$error <- result
      return(old.result)
    }

    # Extract values tracked for estimating convergence
    # Note: The number of fixed effects may be zero, and strong
    # non-linearities that don't necessarily affect the fixed
    # effects may appear in the random effects, so we need to
    # track all of them.
    result_mode <- evaluate_state(
      model,
      result,
      property = "joint_mode"
    )[[1]]
    result_sd <- evaluate_state(
      model,
      result,
      property = "sd",
      internal_hyperpar = TRUE
    )[[1]]
    track_df <- list()
    for (label in names(result_mode)) {
      track_df[[label]] <-
        data.frame(
          effect = label,
          index = seq_along(result_mode[[label]]),
          iteration = track_size + k,
          mode = result_mode[[label]],
          sd = result_sd[[label]]
        )
    }
    # label <- "bru_offset"
    # track_df[[label]] <-
    #   data.frame(
    #     effect = label,
    #     index = seq_along(stk.data[["BRU.offset"]]),
    #     iteration = track_size + k,
    #     mode = stk.data[["BRU.offset"]],
    #     sd = NA
    #   )
    label <- "log.posterior.mode"
    track_df[[label]] <-
      data.frame(
        effect = label,
        index = 1L,
        iteration = track_size + k,
        mode = result[["misc"]][["configs"]][["max.log.posterior"]],
        sd = Inf
      )
    track[[k]] <- do.call(rbind, track_df)

    if (is.null(result[["misc"]][["opt.trace"]])) {
      inla_track[[k]] <-
        tibble::tibble(
          iteration = integer(0),
          f = numeric(0),
          nfunc = integer(0)
        )
    } else {
      inla_track[[k]] <-
        tibble::tibble(
          iteration = inla_track_size + k,
          f = result[["misc"]][["opt.trace"]][["f"]],
          nfunc = result[["misc"]][["opt.trace"]][["nfunc"]],
          theta = result[["misc"]][["opt.trace"]][["theta"]]
        )
    }

    # Only update the linearisation state after the non-final "eb" iterations:
    if (!do_final_integration) {
      # Update stack given current result
      state0 <- states[[length(states)]]
      state <- evaluate_state(model,
        result = result,
        property = "joint_mode"
      )[[1]]
      if ((options$bru_max_iter > 1)) {
        if (do_line_search) {
          timings <- bru_timer_do(timings, "Line search", k + 1L)
          line_weights <-
            extract_property(
              result = result,
              property = "predictor_sd",
              internal_hyperpar = FALSE
            )^{
              -2
            }
          line_search <- bru_line_search(
            model = model,
            lhoods = lhoods,
            lin = lin,
            state0 = state0,
            state = state,
            input = inputs,
            comp_lin = comp_lin,
            comp_simple = comp_simple,
            weights = line_weights,
            options = options
          )
          state <- line_search[["state"]]
        }
        timings <- bru_timer_do(timings, "Linearise", k + 1L)

        bru_log_message(
          "iinla: Evaluate component linearisations",
          verbosity = 3
        )
        comp_lin <- ibm_linear(
          model,
          input = inputs,
          state = state,
          inla_f = TRUE
        )
        bru_log_message(
          "iinla: Evaluate predictor linearisation",
          verbosity = 3
        )
        lin <- bru_compute_linearisation(
          model,
          lhoods = lhoods,
          input = inputs,
          state = state,
          comp_simple = comp_simple,
          options = options
        )

        stk <- bru_make_stack(lhoods, lin, idx)

        if (utils::packageVersion("INLA") <= "24.06.02") {
          stk.data <- INLA::inla.stack.data(stk)
        } else {
          stk.data <- INLA::inla.stack.data(stk,
            .response.name = "BRU.response"
          )
        }
        inla.options$control.predictor$A <- INLA::inla.stack.A(stk)

        N_response <- sum(bru_response_size(lhoods))
        N_predictor <- NROW(inla.options$control.predictor$A)
        if (N_response != N_predictor) {
          msg <- glue(
            "The total number of response values (N={N_response})",
            " and predictor values (N={N_predictor})",
            " do not match.\n",
            "  This is likely due to a mistake in the component or predictor",
            " constructions.",
            .trim = FALSE
          )
          if ((N_response > 1L) && (N_predictor == 1L)) {
            msg <- paste0(
              msg,
              "\nPerhaps you only have components with scalar inputs?"
            )
          }
          bru_log_message(msg, verbosity = 1L)
          stop(msg)
        }
      }
      # Store the state
      states <- c(states, list(state))
    }

    # Stopping criteria
    if (do_final_integration || (k >= options$bru_max_iter)) {
      interrupt <- TRUE
    } else if (!interrupt && (k > 1)) {
      max.dev <- options$bru_method$rel_tol
      dev <- abs(track[[k - 1]]$mode - track[[k]]$mode) / track[[k]]$sd
      ## do.call(c, lapply(by(do.call(rbind, track),
      ##   as.factor(do.call(rbind, track)$effect),
      ##   identity),
      ##   function(X) { abs(X$mean[k-1] - X$mean[k])/X$sd[k] }))
      bru_log_message(
        glue(
          "iinla: Max deviation from previous: ",
          "{signif(100 * max(dev), 3)}",
          "% of SD, and line search is ",
          if (line_search[["active"]]) "active" else "inactive",
          "\n       [stop if: < {100 * max.dev}% and line search inactive]"
        ),
        verbosity = 1L
      )
      do_final_integration <- all(dev < max.dev) && (!line_search[["active"]])
      if (do_final_integration) {
        do_final_theta_no_restart <- TRUE
        bru_log_message(
          "iinla: Convergence criterion met.",
          "\n       Running final INLA integration step with known theta mode.",
          verbosity = 1L
        )
      }
    }
    track[[k]][["new_linearisation"]] <- NA_real_
    for (label in names(states[[1]])) {
      track[[k]][["new_linearisation"]][track[[k]][["effect"]] == label] <-
        unlist(states[[length(states)]][[label]])
    }

    k <- k + 1L
  }

  bru_log_message("iinla: Computation completed", verbosity = 3)

  timings <- bru_timer_done(timings)

  result[["bru_iinla"]] <- collect_misc_info()
  class(result) <- c("iinla", class(result))
  result
}


auto_intercept <- function(components) {
  if (!inherits(components, "formula")) {
    if (inherits(components, "bru_comp_list")) {
      return(components)
    }
    stop("`components` must be a formula to auto-add an intercept component")
  }
  env <- environment(components)

  tm <- terms(components)
  # Check for -1/+0 and +/- Intercept/NULL
  # Required modification:
  # IVar ITerm AutoI Update
  #  T    T     1     -1
  #  T    T     0     -1
  #  T    F     1     -IVar-1
  #  T    F     0     -IVar-1
  #  F    T     1     Impossible
  #  F    T     0     Impossible
  #  F    F     1     +Intercept-1
  #  F    F     0     -1

  # Convert list(var1,var2) call to vector of variable names.
  # ([-1] removes "list"!)
  var_names <- as.character(attr(tm, "variables"))[-1]
  # Locate Intercept, if present
  inter_var <- grep("^Intercept[\\($]*", var_names)
  inter_term <- grep("^Intercept[\\($]*", attr(tm, "term.labels"))
  if (length(inter_var) > 0) {
    if (length(inter_term) > 0) {
      components <- update.formula(components, . ~ . + 1)
    } else {
      components <- update.formula(
        components,
        as.formula(glue(
          ". ~ . - {var_names[inter_var]} -1"
        ))
      )
    }
  } else if (attr(tm, "intercept")) {
    components <- update.formula(components, . ~ . + Intercept(1) + 1)
  } else {
    components <- update.formula(components, . ~ . + 1)
  }
  environment(components) <- env
  components
}


# If missing RHS, set to the sum of the component names
auto_additive_formula <- function(formula, components) {
  if (as.character(formula)[length(as.character(formula))] != ".") {
    return(formula)
  }
  stopifnot(inherits(components, "bru_comp_list"))

  env <- environment(formula)
  formula <- update.formula(
    formula,
    glue(". ~ {glue_collapse(names(components), sep = ' + ')}")
  )
  environment(formula) <- env
  formula
}

# Extract the LHS of a formula, as response ~ .
extract_response <- function(formula) {
  if (inherits(formula, "formula") &&
    (length(as.character(formula)) == 3)) {
    as.formula(glue("{as.character(formula)[2]} ~ ."))
  } else {
    . ~ .
  }
}
# If response missing, add one
auto_response <- function(formula, response) {
  if (!inherits(formula, "formula")) {
    return(formula)
  }
  if ((length(as.character(formula)) == 3) &&
    (as.character(formula)[2] != ".")) {
    # Already has response; do nothing
    return(formula)
  }
  env <- environment(formula)
  if (as.character(formula)[length(as.character(formula))] == ".") {
    # No formula RHS
    formula <- response
  } else {
    formula <- update.formula(formula, response)
  }
  environment(formula) <- env
  formula
}


# Returns a formula's environment as a list. Removes all variable that are of
# type inla, function or formula. Also removes all variables that are not
# variables of the formula.
list.data <- function(formula) {
  # Formula environment as list
  elist <- as.list(environment(formula))

  # Remove previous inla results. For some reason these slow down the next INLA
  # call.
  elist <- elist[unlist(lapply(elist, function(x) !inherits(x, "inla")))]

  # Remove functions. This can cause problems as well.
  # elist <- elist[unlist(lapply(elist, function(x) !is.function(x)))]

  # Remove formulae. This can cause problems as well.
  # elist <- elist[unlist(lapply(elist, function(x) !inherits(x, "formula")))]

  # The formula expression is too general for this to be reliable:
  # Keep only purse formula variables
  # elist <- elist[names(elist) %in% all.vars(formula)]

  elist
}


#' @describeIn bru Takes a fitted `bru` object produced by [bru()] or [lgcp()]
#' and creates various summaries from it, including the summary output from
#' the corresponding `INLA::sumary()` method.
#' The \dots arguments are passed on to component summary functions, see
#' [summary.bru_comp()].
#'
#' @export
#' @method summary bru
#' @param object An object obtained from a [bru()] or [lgcp()] call
#' @param verbose logical; If `TRUE`, include more details of the
#' component definitions. If `FALSE`, only show basic component
#' definition information. Default: `FALSE`
#'
summary.bru <- function(object, verbose = FALSE, ...) {
  object <- bru_check_object_bru(object)

  result <- list(
    bru_info = summary(object[["bru_info"]], verbose = verbose, ...)
  )

  if (inherits(object, "inla")) {
    result$inla <- NextMethod("summary", object)
    result[["inla"]][["call"]] <- NULL
  }

  class(result) <- "summary_bru"
  result
}


#' @export
#' @param x An object to be printed
#' @rdname bru

print.summary_bru <- function(x, ...) {
  print(x$bru_info, ...)
  print(x$inla, ...)
  invisible(x)
}

#' @export
#' @describeIn bru Print a summary of a `bru` object.

print.bru <- function(x, ...) {
  print(summary(x, ...), ...)
  invisible(x)
}

test_that("bru: inla copy feature", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123L)

  df1 <- data.frame(x = cos(1:100))
  df2 <- data.frame(x = sin(1:100))
  df1 <- within(df1, {
    y <- 1 + exp(2 * x) + rnorm(length(x), mean = 0, sd = 0.1)
  })
  df2 <- within(df2, {
    y <- 1 + (6 * x) + rnorm(length(x), mean = 0, sd = 0.1)
  })
  cmp <- ~
    +1 +
      myLin1(x,
        model = "rw1",
        mapper = bru_mapper(fm_mesh_1d(seq(-1, 1, length.out = 100)),
          indexed = FALSE
        )
      ) +
      myLin2(x, copy = "myLin1", fixed = FALSE)
  cmps <- bru_comp_list(cmp)

  fit <- bru(
    cmp,
    bru_obs(
      y ~ Intercept + exp(myLin1),
      family = "gaussian",
      data = df1
    ),
    bru_obs(
      y ~ Intercept + (myLin2),
      family = "gaussian",
      data = df2
    ),
    options = list(control.inla = list(int.strategy = "eb"))
  )


  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    1,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.hyperpar["Beta for myLin2", "mean"],
    3,
    tolerance = midtol
  )

  skip_if_not_installed("sn")
  pr <- predict(
    fit,
    data.frame(x = c(0.5, 1)),
    ~myLin2,
    n.samples = 500,
    seed = 1L
  )

  expect_equal(pr[, "mean"], c(3, 6), tolerance = midtol)
})


# test_that("bru: Non-copy option", {
#   skip_on_cran()
#   local_bru_safe_inla()
#
#   # Seed influences data as well as predict()!
#   withr::local_seed(123L)
#
#   df1 <- data.frame(x = cos(1:100))
#   df2 <- data.frame(x = sin(1:100))
#   df1 <- within(df1, eta <- 1 + 2 * x)
#   df2 <- within(df2, eta <- 1 + 6 * x)
#   df1 <- within(df1, y <- rnorm(length(eta), mean = eta, sd = 0.1))
#   df2 <- within(df2, y <- rpois(length(x), lambda = exp(eta)))
#
#   cmp <- ~
#     0 + I1(1) + I2(1) +
#     myLin1(x,
#            model = "rw1",
#            mapper = bru_mapper(fm_mesh_1d(seq(-1, 1, length.out = 100)),
#                                indexed = FALSE
#            ),
#            scale.model = TRUE
#     ) +
#         myLin2(x, copy = "myLin1", fixed = FALSE, initial = 1)
# #    myLin2(1)
#   cmps <- bru_comp_list(cmp)
#
#   fit <- bru(
#     cmp,
#     bru_obs(y ~ I1 + myLin1, family = "gaussian", data = df1),
# #    bru_obs(y ~ I2 + myLin2*myLin1, family = "poisson", data = df2),
#     bru_obs(y ~ I2 + myLin2, family = "poisson", data = df2),
#     options = list(control.inla = list(int.strategy = "eb"),
#                    bru_initial = list(myLin2 = 1, myLin1 = rnorm(100)))
#   )
#
#
#   expect_equal(
#     fit$summary.fixed["Intercept", "mean"],
#     1,
#     tolerance = midtol
#   )
#   expect_equal(
#     fit$summary.hyperpar["Beta for myLin2", "mean"],
#     3,
#     tolerance = midtol
#   )
#
#   skip_if_not_installed("sn")
#   pr <- predict(
#     fit,
#     data.frame(x = c(0.5, 1)),
#     ~myLin2,
#     n.samples = 500,
#     seed = 1L
#   )
#
#   expect_equal(pr[, "mean"], c(3, 6), tolerance = midtol)
# })


test_that("Component copy feature", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123L)

  mydata <- data.frame(
    x0 = 1,
    x1 = rep(1:20, times = 2) / 10,
    x2 = rep(1:10, each = 4) / 10
  )
  mydata <- within(mydata, {
    y <- rpois(
      nrow(mydata),
      exp(x1^0.5 + x2^0.5 * 2 + x0 * 3)
    )
  })

  inlaform <- y ~ 0 + x0 +
    f(x1, model = "rw2", values = sort(unique(mydata$x1)), scale.model = TRUE) +
    f(x2, copy = "x1", fixed = FALSE)
  fit <- INLA::inla(
    formula = inlaform,
    data = mydata, family = "poisson",
    inla.mode = bru_options_get("inla.mode"),
    control.compute = list(config = TRUE),
    control.inla = list(int.strategy = "eb")
  )

  cmp <- y ~ 0 + x0 +
    x1(x1, model = "rw2", scale.model = TRUE) +
    x2(x2, copy = "x1", fixed = FALSE)
  fit_bru <- bru(
    cmp,
    family = "poisson",
    data = mydata,
    options = list(control.inla = list(int.strategy = "eb"))
  )

  expect_equal(
    fit_bru$summary.hyperpar$x1$mean,
    fit$summary.hyperpar$x1$mean,
    tolerance = lowtol
  )
  expect_equal(
    fit_bru$summary.hyperpar$x2$mean,
    fit$summary.hyperpar$x2$mean,
    tolerance = lowtol
  )
  expect_equal(
    fit_bru$summary.hyperpar$x1$sd,
    fit$summary.hyperpar$x1$sd,
    tolerance = hitol
  )
  expect_equal(
    fit_bru$summary.hyperpar$x2$sd,
    fit$summary.hyperpar$x2$sd,
    tolerance = hitol
  )
})

test_that("Component copy feature with group", {
  skip_on_cran()
  local_bru_safe_inla()

  # Seed influences data as well as predict()!
  withr::local_seed(123L)

  n <- c(16, 8)
  mydata <- data.frame(
    x1 = rep(seq_len(n[1]), times = n[2]),
    x2 = rep(seq_len(n[2]), each = n[1])
  )
  mydata <- rbind(mydata, mydata)
  mydata <- within(mydata, {
    Intercept <- 1
    z <- rep(c(1, 2), times = length(x1) / 2)
    z2 <- rep(c(1, 2), each = length(x1) / 2)
  })
  mydata <- within(mydata, {
    y <- rnorm(prod(n), x1^1.0 / n[1] * 4 + x2^1.0 / n[1] * 4 * 10, 1)
  })

  inlaform <- y ~ -1 + Intercept +
    f(
      x1,
      model = "rw1",
      values = seq_len(n[1]),
      scale.model = TRUE,
      group = z
    ) +
    f(x2,
      copy = "x1",
      fixed = FALSE,
      group = z2
    )
  fit <- INLA::inla(
    formula = inlaform, data = mydata, family = "normal",
    control.inla = list(int.strategy = "eb"),
    control.compute = list(config = TRUE),
    inla.mode = bru_options_get("inla.mode")
  )

  cmp <- ~ -1 + Intercept +
    x1(x1, model = "rw1", scale.model = TRUE, group = z) +
    x2(x2, copy = "x1", fixed = FALSE, group = z2)
  fit_bru <- bru(
    cmp,
    formula = y ~ Intercept + x1 + x2,
    family = "normal", data = mydata,
    options = list(
      bru_max_iter = 1,
      control.inla = list(int.strategy = "eb")
    )
  )

  # cmp2 <- ~ -1 + Intercept +
  #   x1(x1, model = "rw1", scale.model = TRUE, group = z) +
  #   x1copy(x2, copy = "x1", fixed = TRUE, group = z2) +
  #   beta(1, model = "linear")
  # fit_bru2 <- bru(
  #   cmp2,
  #   formula = y ~ Intercept + x1 + beta * exp(x1copy),
  #   family = "normal",
  #   data = mydata,
  #   options = list(
  #     bru_max_iter = 5,
  #     bru_initial = list(beta = 1),
  #     control.inla = list(int.strategy = "eb")
  #   )
  # )

  expect_equal(
    fit_bru$summary.hyperpar$mean,
    fit$summary.hyperpar$mean,
    tolerance = hitol
  )
  expect_equal(
    fit_bru$summary.hyperpar$sd,
    fit$summary.hyperpar$sd,
    tolerance = hitol
  )
})

test_that("Additivity: Additive predictor with nonlinear components", {
  skip_on_cran()
  local_bru_safe_inla()

  withr::local_seed(12345L)

  df <- tibble::tibble(
    u = 2,
    x = qexp(pnorm(u)),
    y = rpois(10, exp(x))
  )

  # Linear
  cmp <- ~ x(1, mean.linear = 0, prec.linear = 1) + 0
  form <- y ~ x
  fit <- bru(
    components = cmp,
    bru_obs(
      formula = form,
      family = "poisson",
      data = df
    )
  )

  # The latent variable should match x
  expect_equal(
    fit$summary.fixed["x", "mean"],
    3.821398,
    tolerance = midtol
  )

  # Nonlinear
  cmp <- ~ x(
    1,
    mean.linear = 0,
    prec.linear = 1,
    marginal = bm_marginal(qexp, pexp, dexp, rate = 1)
  ) + 0
  form <- y ~ x
  fit <- bru(
    components = cmp,
    bru_obs(
      formula = form,
      family = "poisson",
      data = df
    )
  )

  # The latent variable should match u
  expect_equal(
    fit$summary.fixed["x", "mean"],
    2.007725,
    tolerance = midtol
  )
})

test_that("Joint gcpo", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not_installed("sf")

  ## partition
  cvpart <- cv_hex(
    gorillas_sf$boundary,
    cellsize = 0.5,
    n_group = 3,
    resolution = c(95, 80)
  )
  cvpart$block_ID <- seq_len(nrow(cvpart))
  cvpart$group <- NULL
  nblock <- nrow(cvpart)

  ## split
  gorillas_nests_major <-
    gorillas_sf$nests[gorillas_sf$nests$group == "major", ]
  gorillas_nests_minor <-
    gorillas_sf$nests[gorillas_sf$nests$group == "minor", ]

  # blocks for each nest
  a <- sf::st_intersects(gorillas_nests_major, cvpart)
  if (!all(vapply(a, function(x) (length(x) == 1), logical(1)))) {
    stop("Point in none or multiple polygons")
  }
  gorillas_nests_major$.block <- unlist(a)

  b <- sf::st_intersects(gorillas_nests_minor, cvpart)
  if (!all(vapply(b, function(x) (length(x) == 1), logical(1)))) {
    stop("Point in none or multiple polygons")
  }
  gorillas_nests_minor$.block <- unlist(b)

  # model
  matern <- INLA::inla.spde2.pcmatern(
    gorillas_sf$mesh,
    prior.range = c(0.1, 0.01),
    prior.sigma = c(1, 0.01)
  )

  cmp <- ~
    Common(geometry, model = matern) +
      Difference(geometry, model = matern) +
      Intercept(1)

  fml.major <- geometry ~ Intercept + Common + Difference / 2
  fml.minor <- geometry ~ Intercept + Common - Difference / 2

  lik_major <- bru_obs("cp",
    formula = fml.major,
    samplers = cvpart,
    data = gorillas_nests_major,
    domain = list(geometry = gorillas_sf$mesh),
    control.gcpo = list(
      enable = TRUE,
      type.cv = "joint",
      num.level.sets = -1
    ),
    tag = "major"
  )
  lik_minor <- bru_obs("cp",
    formula = fml.minor,
    samplers = cvpart,
    data = gorillas_nests_minor,
    domain = list(geometry = gorillas_sf$mesh),
    control.gcpo = list(
      enable = TRUE,
      type.cv = "joint",
      num.level.sets = -1
    ),
    tag = "minor"
  )

  c_g <- bru_obs_control_gcpo(
    c(lik_major, lik_minor),
    control.gcpo = bru_options("control.gcpo")
  )
  expect_length(
    c_g$friends,
    length(bru_index(lik_major)) + length(bru_index(lik_minor))
  )
})

test_that("fm_pixels sp vs sf", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if_not(bru_safe_sp())
  skip_if_not_installed("sn")
  withr::local_options(lifecycle_verbosity = "quiet")

  mesh <- fm_mesh_2d_inla(cbind(0, 0),
    offset = 10, max.edge = 1,
    crs = fm_CRS("longlat_globe")
  )

  withr::local_seed(12345L)
  mydata <- sp::SpatialPointsDataFrame(
    mesh$loc,
    data = data.frame(y = rnorm(mesh$n) + 10),
    proj4string = fm_CRS("longlat_globe")
  )

  fit <- bru(
    ~ 0 +
      Intercept(1) +
      field(
        sp::coordinates,
        model = INLA::inla.spde2.pcmatern(
          mesh,
          prior.range = c(1, 0.01),
          prior.sigma = c(1, 0.01)
        )
      ),
    bru_obs(
      y ~ .,
      family = "gaussian",
      data = mydata
    )
  )

  system.time({
    withr::local_seed(1234L)
    surface1 <- fm_pixels(mesh, dims = c(5, 5), mask = TRUE, format = "sp")
    density1 <- predict(fit,
      surface1,
      ~ exp(field_eval(sp::coordinates(.data.)) + Intercept),
      n.samples = 10,
      seed = 12345L
    )
  })

  system.time({
    withr::local_seed(1234L)
    surface2 <- fm_pixels(mesh, dims = c(5, 5), mask = TRUE, format = "sf")
    density2 <- predict(fit,
      surface2,
      ~ exp(field_eval(geometry) + Intercept),
      n.samples = 10,
      seed = 12345L
    )
  })

  expect_equal(
    density1$mean,
    density2$mean
  )

  # withr::local_seed(1234L)
  # A1<-generate(fit,n.samples=10,seed=12345L,num.threads="1:1")
  #
  # withr::local_seed(1234L)
  # A2<-generate(fit,n.samples=10,seed=12345L,num.threads="1:1:1")
  #
  # expect_equal(INLA::inla.getOption("num.threads"), "1:1:1")
  #
  # expect_equal(A1[[1]]$field, A2[[1]]$field)
  # expect_equal(A1[[2]]$field, A2[[2]]$field)
  # expect_equal(A1[[3]]$field, A2[[3]]$field)
  # expect_equal(A1[[4]]$field, A2[[4]]$field)
  # expect_equal(A1[[5]]$field, A2[[5]]$field)
})

test_that("mdata", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if(utils::packageVersion("INLA") <= "24.06.02")
  set.seed(12345L)

  sim.poisson <- function(prob, m) {
    stopifnot(length(prob) == length(m) && length(prob) > 0)
    n <- length(m)
    y <- numeric(n)
    event <- (runif(n) < prob)
    idx.zero <- which(event)
    idx.non.zero <- which(!event)
    y[idx.zero] <- 0
    y[idx.non.zero] <- rpois(length(idx.non.zero), lambda = m[idx.non.zero])
    y
  }
  ## chose link-function to use for the zero-inflation probability
  link.simple <- "logit"
  inv.link <- INLA::inla.link.invlogit
  ## link.simple <- "probit"
  ## inv.link <- inla.link.invprobit
  ## link.simple <- "cloglog"
  ## inv.link <- inla.link.invcloglog
  n <- 1000
  z <- rnorm(n, sd = 0.3)
  x <- rnorm(n, sd = 0.2)
  xx <- rnorm(n, sd = 0.3)
  zz <- rnorm(n, sd = 0.2)
  E <- runif(n, min = 0.8, max = 1 / 0.8)
  beta <- c(1, 1.1, 2.1, 0, -2, 1.2, 2.2, 0)
  eta2 <- beta[1] + beta[2] * xx + beta[3] * zz + beta[4] * xx * zz
  eta1 <- beta[5] + beta[6] * x + beta[7] * z + beta[8] * x * z
  prob <- inv.link(eta1)
  m <- E * exp(eta2)
  ok <- FALSE
  while (!ok) {
    y <- sim.poisson(prob, m)
    ok <- !all(y == 0)
  }
  ## head(data.frame(y, E, x, z, xx, zz))
  suppressWarnings(
    r <- INLA::inla(
      INLA::inla.mdata(
        cbind(y, E),
        cbind(1, x, z, x * z)
      ) ~ 1 + xx + zz + xx * zz,
      family = "0poisson",
      data = data.frame(y, E, x, z, xx, zz),
      control.fixed = list(prec = 1, prec.intercept = 1),
      control.compute = list(cpo = TRUE),
      control.family = list(
        link.simple = link.simple,
        hyper = list(
          beta1 = list(param = c(0, 1)),
          beta2 = list(param = c(0, 1)),
          beta3 = list(param = c(0, 1)),
          beta4 = list(param = c(0, 1)),
          beta5 = list(param = c(0, 1))
        )
      )
    )
  )
  #  y2 <- INLA::inla.mdata(cbind(y, -1, E/2),
  #                         cbind(1, x, z, x*z))
  #  r2 <- INLA::inla(
  #    y2 ~ 1 + xx + zz + xx*zz,
  #    family = "0poisson",
  #    data = data.frame(y, E, x, z, xx, zz),
  #    control.fixed = list(prec = 1, prec.intercept = 1),
  #    control.compute = list(cpo = TRUE),
  #    control.family = list(
  #      list(link.simple = link.simple,
  #                          hyper = list(beta1 = list(param = c(0, 1)),
  #                                       beta2 = list(param = c(0, 1)),
  #                                       beta3 = list(param = c(0, 1)),
  #                                       beta4 = list(param = c(0, 1)),
  #                                       beta5 = list(param = c(0, 1)))))
  #  )
  #

  r_bru <- bru(
    ~ 0 + fix(
      ~ 1 + xx + zz + xx * zz,
      model = "fixed",
      hyper = list(prec = list(initial = 0, fixed = TRUE))
    ),
    bru_obs(
      INLA::inla.mdata(cbind(y, E), cbind(1, x, z, x * z)) ~ .,
      family = "0poisson",
      data = data.frame(y, E, x, z, xx, zz),
      control.family = list(
        link.simple = link.simple,
        hyper = list(
          beta1 = list(param = c(0, 1)),
          beta2 = list(param = c(0, 1)),
          beta3 = list(param = c(0, 1)),
          beta4 = list(param = c(0, 1)),
          beta5 = list(param = c(0, 1))
        )
      )
    ),
    options = list(control.compute = list(cpo = TRUE))
  )

  expect_equal(
    r_bru$summary.fixed$mean,
    r$summary.random$fix$mean,
    tolerance = lowtol
  )
  expect_equal(
    r_bru$summary.fixed$sd,
    r$summary.random$fix$sd,
    tolerance = lowtol
  )


  if (FALSE) {
    rr <- inla(
      inla.mdata(cbind(y, E), cbind(1, xx, zz, xx * zz)) ~ 1 + x + z + x * z,
      family = "0poissonS",
      data = data.frame(y, E, x, z, xx, zz),
      control.fixed = list(prec = 1, prec.intercept = 1),
      control.compute = list(cpo = TRUE),
      ## in this case we need to define link.simple as the main link
      control.family = list(
        control.link = list(model = link.simple),
        hyper = list(
          beta1 = list(param = c(0, 1)),
          beta2 = list(param = c(0, 1)),
          beta3 = list(param = c(0, 1)),
          beta4 = list(param = c(0, 1)),
          beta5 = list(param = c(0, 1))
        )
      )
    )
    summary(r)
    summary(rr)
    res <- cbind(
      "beta" = beta,
      "0poisson" = c(r$summary.fixed$mean, r$summary.hyperpar$mean),
      "0poissonS" = c(rr$summary.hyperpar$mean, rr$summary.fixed$mean)
    )
    res <- cbind(res,
      diff = (res[, 2] - beta),
      diffS = (res[, 3] - beta),
      "diff/sd" = (res[, 2] - beta) /
        c(r$summary.fixed$sd, r$summary.hyperpar$sd),
      "diffS/sd" = (res[, 3] - beta) /
        c(rr$summary.hyperpar$sd, rr$summary.fixed$sd)
    )
    mm <- nrow(res) %/% 2
    rownames(res) <-
      c(paste0("beta", 1:mm, ".poisson"), paste0("beta", 1:mm, ".prob"))
    print(round(dig = 2, res))
  }
})


test_that("surv", {
  skip_on_cran()
  local_bru_safe_inla()
  skip_if(utils::packageVersion("INLA") <= "24.06.26")

  set.seed(12345L)
  df <- data.frame(
    time = 1:4,
    event = c(1, 1, 0, 1),
    xx = rnorm(4),
    Intercept = 1
  )

  r_inla <- INLA::inla(
    INLA::inla.surv(time, event) ~ 0 +
      f(Intercept, model = "linear", prec.linear = 0) +
      f(xx, model = "linear"),
    family = "weibullsurv",
    data = df
  )

  r_bru <- bru(
    INLA::inla.surv(time, event) ~ 0 + Intercept(1, prec.linear = 0) + xx(xx),
    family = "weibullsurv",
    data = df
  )

  expect_equal(
    r_bru$summary.fixed[, "mean"],
    r_inla$summary.fixed[, "mean"],
    tolerance = lowtol
  )

  expect_equal(
    r_bru$summary.fixed[, "sd"],
    r_inla$summary.fixed[, "sd"],
    tolerance = midtol
  )

  expect_equal(
    r_bru$summary.hyperpar[, "mean"],
    r_inla$summary.hyperpar[, "mean"],
    tolerance = lowtol
  )

  expect_equal(
    r_bru$summary.hyperpar[, "sd"],
    r_inla$summary.hyperpar[, "sd"],
    tolerance = midtol
  )

  # Multi-family
  stk1.est <- INLA::inla.stack(
    data = list(.dummy = rep(1, 4), link = 1),
    A = list(1),
    effects = list(list(Intercept = 1, xx = df$xx)),
    responses = list(with(data = df, expr = INLA::inla.surv(time, event)))
  )
  stk1.pred <- INLA::inla.stack(
    data = list(.dummy = rep(1, 4), link = 1),
    A = list(1),
    effects = list(list(Intercept = 1, xx = df$xx)),
    responses = list(with(data = df, expr = INLA::inla.surv(time * NA, event)))
  )
  stk2 <- INLA::inla.stack(
    data = list(.dummy = rep(1, 4), link = 2),
    A = list(1),
    effects = list(list(Intercept = 1, xx = df$xx)),
    responses = list(with(data = df, expr = INLA::inla.surv(time, event)))
  )
  stk <- INLA::inla.stack(
    INLA::inla.stack(stk1.est, stk1.pred),
    stk2,
    multi.family = TRUE
  )


  stk_data <- INLA::inla.stack.data(stk, .response.name = "response")
  r2_inla <- INLA::inla(
    response ~ 0 + f(Intercept, model = "linear", prec.linear = 0) +
      f(xx, model = "linear"),
    family = c("exponentialsurv", "weibullsurv"),
    data = stk_data,
    control.predictor = list(A = INLA::inla.stack.A(stk), link = stk_data$link)
  )

  r2_bru <- bru(
    ~ 0 + Intercept(Intercept, prec.linear = 0) + xx(xx),
    bru_obs(
      INLA::inla.surv(time, event) ~ .,
      family = "exponentialsurv",
      data = rbind(
        df,
        data.frame(time = NA, event = df$event, xx = df$xx, Intercept = 1)
      )
    ),
    bru_obs(
      INLA::inla.surv(time, event) ~ .,
      family = "weibullsurv",
      data = df,
    )
  )

  expect_equal(
    r2_bru$summary.fixed[, c("mean", "sd")],
    r2_inla$summary.fixed[, c("mean", "sd")],
    tolerance = lowtol
  )

  expect_equal(
    r2_bru$summary.hyperpar[, c("mean", "sd")],
    r2_inla$summary.hyperpar[, c("mean", "sd")],
    tolerance = midtol
  )
})

# Test for data input with mismatching input/output sizes (e.g. for regional
# integration models, etc)

test_that("Component construction: default mesh/mapping, data is list", {
  skip_on_cran()
  local_bru_safe_inla()

  lik <- bru_obs("gaussian",
    formula = y ~ .,
    data = list(x = c(1, 1.5, 2, NA, 3, 4), y = 11:15),
    used = bru_used(effect = "effect"),
    is_rowwise = FALSE
  )

  cmp1 <- bru_comp_list(~ effect(c(1, 1.5, 2, 3, 4), model = "iid") - 1)
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(unique(lik$data$x), na.last = NA)
  )

  cmp1 <- bru_comp_list(~ effect(x, model = "rw2") - 1)
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(unique(lik$data$x), na.last = NA)
  )

  mesh1 <- fm_mesh_1d(
    sort(unique(lik$data$x), na.last = NA)
  )
  expect_error(
    bru_comp_list(
      ~ effect(x, model = "rw2", mapper = mesh1) - 1
    ),
    regexp = "Unknown mapper"
  )

  cmp1 <- bru_comp_list(
    ~ effect(x,
      model = "rw2",
      mapper = bru_mapper(mesh1, indexed = FALSE)
    ) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    sort(unique(lik$data$x), na.last = NA)
  )

  cmp1 <- bru_comp_list(
    ~ effect(x,
      model = "rw2",
      mapper = bru_mapper(mesh1, indexed = TRUE)
    ) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    seq_along(sort(unique(lik$data$x), na.last = NA))
  )
})


test_that("Component construction: unsafe intercepts, data is list", {
  skip_on_cran()
  local_bru_safe_inla()

  cmp <- bru_comp_list(~ something_unknown - 1)
  lik <- bru_obs(
    formula = response ~ ., data = list(response = 1:5),
    is_rowwise = FALSE
  )
  lik <- bru_used_update(lik, labels = names(cmp))
  expect_error(
    object = {
      model <- bru_model(cmp, bru_obs_list(list(lik)))
    },
    paste0(
      "The input evaluation 'something_unknown' for 'something_unknown' ",
      "failed.\nPerhaps the data object doesn't contain the needed variables?"
    )
  )
})


test_that("Component construction: separate response_data input", {
  skip_on_cran()
  local_bru_safe_inla()

  lik1 <- bru_obs("gaussian",
    formula = y ~ c(sum(effect), sum(effect^2)),
    data = list(x = c(1, 1.5, 2, 3, 4)),
    response_data = data.frame(y = c(11, 12)),
    is_rowwise = FALSE
  )

  lik2 <- bru_obs("gaussian",
    formula = y ~ c(sum(effect), sum(effect^2)),
    data = data.frame(x = c(1, 1.5, 2, 3, 4)),
    response_data = data.frame(y = c(11, 12)),
    is_rowwise = FALSE
  )


  cmp1 <- bru_comp_list(~ effect(x, model = "iid") - 1)

  fit1 <- bru(components = cmp1, lik1)
  fit2 <- bru(components = cmp1, lik2)


  cmp1 <- bru_comp_list(~ effect(x, model = "rw2") - 1)
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik2)))
  expect_equal(ibm_values(cmp2$effect$mapper, multi = 1)$main, lik2$data$x)

  mesh1 <- fm_mesh_1d(lik1$data$x)
  expect_error(
    bru_comp_list(
      ~ effect(x, model = "rw2", mapper = mesh1) - 1
    ),
    regexp = "Unknown mapper"
  )

  cmp1 <- bru_comp_list(
    ~ effect(x,
      model = "rw2",
      mapper = bru_mapper(mesh1, indexed = FALSE)
    ) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik1)))
  expect_equal(ibm_values(cmp2$effect$mapper, multi = 1)$main, lik1$data$x)

  cmp1 <- bru_comp_list(
    ~ effect(x,
      model = "rw2",
      mapper = bru_mapper(mesh1, indexed = TRUE)
    ) - 1
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik1)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    seq_along(lik1$data$x)
  )
  cmp2 <- add_mappers(cmp1, lhoods = bru_obs_list(list(lik2)))
  expect_equal(
    ibm_values(cmp2$effect$mapper, multi = 1)$main,
    seq_along(lik2$data$x)
  )
})

test_data <- function() {
  data(Poisson2_1D, package = "inlabru", envir = environment())
  x <- seq(0, 55, length = 50)
  mesh1D <- fm_mesh_1d(x, boundary = "free", degree = 2)
  matern <- INLA::inla.spde2.pcmatern(
    mesh1D,
    prior.range = c(150, 0.75),
    prior.sigma = c(0.1, 0.75),
    constr = TRUE
  )
  mdl <- x ~ spde1D(main = x, model = matern) + Intercept(1)
  fit <- lgcp(
    mdl,
    pts2,
    domain = list(x = mesh1D),
    options = list(
      control.inla = list(int.strategy = "eb")
    )
  )
  list(
    mesh1D = mesh1D,
    fit = fit
  )
}


test_that("1D LGCP fitting", {
  skip_on_cran()
  local_bru_safe_inla()

  result <- test_data()
  mesh1D <- result$mesh1D
  fit <- result$fit

  # Needed for reproducible predict
  withr::local_seed(123L)

  expect_s3_class(fit, "bru")

  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.60974,
    tolerance = midtol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "sd"],
    0.10916,
    tolerance = midtol
  )

  expect_snapshot_value(
    fit$summary.random$spde1D$mean[c(1, 27, 50)],
    tolerance = midtol,
    style = "deparse"
  )
  expect_snapshot_value(
    fit$summary.random$spde1D$sd[c(2, 32, 29)],
    tolerance = midtol,
    style = "deparse"
  )

  skip_if_not_installed("sn")

  # predicted intensity integral
  ips <- fm_int(mesh1D, name = "x")
  Lambda <- predict(
    fit,
    ips,
    ~ sum(weight * exp(spde1D + Intercept)),
    n.samples = 100,
    seed = 4354
  )
  expect_equal(Lambda$mean, 130.25, tolerance = hitol)
  expect_equal(Lambda$sd, 12.28, tolerance = 1)
})


test_data_discrete <- function() {
  data(Poisson2_1D, package = "inlabru", envir = environment())
  xx <- ceiling(pts2$x)
  data <- data.frame(x = xx)
  x <- seq(1, 55, length = 55)
  mesh1D <- fm_mesh_1d(x, boundary = "free")
  matern <- INLA::inla.spde2.pcmatern(mesh1D,
    prior.range = c(0.01, 0.01),
    prior.sigma = c(1, 0.01),
    constr = TRUE
  )
  mdl <- x ~ spde1D(main = x, model = matern) + Intercept(1)
  fit <- lgcp(mdl,
    data = data,
    domain = list(x = x),
    options = list(
      control.inla = list(int.strategy = "eb")
    )
  )
  list(
    mesh1D = mesh1D,
    fit = fit
  )
}


test_that("1D LGCP fitting, discrete point domain", {
  skip_on_cran()
  local_bru_safe_inla()

  result <- test_data_discrete()
  mesh1D <- result$mesh1D
  fit <- result$fit

  # Needed for reproducible predict
  withr::local_seed(123L)

  expect_s3_class(fit, "bru")

  expect_equal(
    fit$summary.fixed["Intercept", "mean"],
    0.605155,
    tolerance = hitol
  )
  expect_equal(
    fit$summary.fixed["Intercept", "sd"],
    0.11224,
    tolerance = hitol
  )

  expect_snapshot_value(
    fit$summary.random$spde1D$mean[c(1, 27, 50)],
    tolerance = hitol,
    style = "deparse"
  )
  expect_snapshot_value(
    fit$summary.random$spde1D$sd[c(2, 32, 29)],
    tolerance = hitol,
    style = "deparse"
  )

  skip_if_not_installed("sn")

  # predicted intensity integral
  ips <- data.frame(x = 1:55, weight = 1)
  Lambda <- predict(
    fit,
    ips,
    ~ sum(weight * exp(spde1D + Intercept)),
    n.samples = 100,
    seed = 4354
  )
  expect_equal(Lambda$mean, 130.127, tolerance = hitol)
  expect_equal(Lambda$sd, 11.09, tolerance = 1)
})


test_that("1D LGCP fitting, compressed format", {
  skip_on_cran()
  local_bru_safe_inla()
  bru_options_set_local(
    control.inla = list(int.strategy = "eb")
  )

  data(Poisson2_1D, package = "inlabru", envir = environment())
  x <- seq(0, 55, length.out = 50)
  mesh1D <- fm_mesh_1d(x, boundary = "free")
  matern <- INLA::inla.spde2.pcmatern(
    mesh1D,
    prior.range = c(1, 0.01),
    prior.sigma = c(0.1, 0.75),
    constr = TRUE
  )

  mdl <- ~ spde1D(main = x, model = matern) + Intercept(1)

  fit1 <- bru(
    mdl,
    bru_obs(
      formula = x ~ .,
      family = "cp",
      data = pts2,
      domain = list(x = mesh1D),
      options = list(bru_compress_cp = FALSE)
    )
  )
  fit2 <- bru(
    mdl,
    bru_obs(
      formula = x ~ .,
      family = "cp",
      data = pts2,
      domain = list(x = mesh1D),
      options = list(bru_compress_cp = TRUE)
    )
  )

  # newdata <- data.frame(x = seq(0, 55, by = 1))
  # pred0 <- data.frame(x = newdata$x, mean = lambda2_1D(newdata$x))
  # pred1 <- predict(fit1, newdata, formula = ~ exp(spde1D + Intercept),
  # n.samples = 1000)
  # pred2 <- predict(fit2, newdata, formula = ~ exp(spde1D + Intercept),
  # n.samples = 1000)
  # library(ggplot2)
  # ggplot() +
  #   geom_line(aes(x = x, y = mean), color = "black", data = pred0) +
  #   geom_line(aes(x = x, y = mean), color = "red", data = pred1) +
  #   geom_ribbon(aes(x = x, ymin = q0.025, ymax = q0.975), fill = "red", alpha
  #   = 0.2, data = pred1) +
  #   geom_line(aes(x = x, y = mean), color = "blue", data = pred2) +
  #   geom_ribbon(aes(x = x, ymin = q0.025, ymax = q0.975), fill = "blue", alpha
  #   = 0.2, data = pred2)
  #
  # pred0 <- data.frame(x = newdata$x, mean = log(lambda2_1D(newdata$x)))
  # pred1 <- predict(fit1, newdata, formula = ~ (spde1D + Intercept), n.samples
  # = 1000)
  # pred2 <- predict(fit2, newdata, formula = ~ (spde1D + Intercept), n.samples
  # = 1000)
  # ggplot() +
  #   geom_line(aes(x = x, y = mean), color = "black", data = pred0) +
  #   geom_line(aes(x = x, y = mean), color = "red", data = pred1) +
  #   geom_ribbon(aes(x = x, ymin = q0.025, ymax = q0.975), fill = "red", alpha
  #   = 0.2, data = pred1) +
  #   geom_line(aes(x = x, y = mean), color = "blue", data = pred2) +
  #   geom_ribbon(aes(x = x, ymin = q0.025, ymax = q0.975), fill = "blue", alpha
  #   = 0.2, data = pred2)

  expect_equal(
    fit1$summary.hyperpar,
    fit2$summary.hyperpar,
    tolerance = hitol
  )
  expect_equal(
    fit1$summary.fixed,
    fit2$summary.fixed,
    tolerance = midtol
  )
  expect_equal(
    fit1$summary.random,
    fit2$summary.random,
    tolerance = midtol
  )

  fit3 <- bru(
    mdl,
    bru_obs(
      formula = x ~ spde1D + Intercept,
      family = "cp",
      data = pts2,
      domain = list(x = mesh1D),
      options = list(bru_compress_cp = FALSE)
    )
  )
  fit4 <- bru(
    mdl,
    bru_obs(
      formula = x ~ spde1D + Intercept,
      family = "cp",
      data = pts2,
      domain = list(x = mesh1D),
      options = list(bru_compress_cp = TRUE)
    )
  )

  expect_equal(
    fit3$summary.hyperpar,
    fit4$summary.hyperpar,
    tolerance = hitol
  )
  expect_equal(
    fit3$summary.fixed,
    fit4$summary.fixed,
    tolerance = midtol
  )
  expect_equal(
    fit3$summary.random,
    fit4$summary.random,
    tolerance = midtol
  )
})

#' @name toygroups
#' @title Simulated 1D animal group locations and group sizes
#' @docType data
#' @description
#'
#' This data set serves to teach the concept of modelling species that gather in
#' groups and where the grouping behaviour depends on space.
#'
#' @usage data(toygroups)
#'
#' @format The data are a list that contains these elements:
#'  \describe{
#'    \item{`groups`:}{ A `data.frame` of group locations `x` and size `size`}
#'    \item{`df.size`:}{ IGNORE THIS }
#'    \item{`df.intensity`:}{ A `data.frame` with Poisson process
#'      intensity `d.lambda` at locations `x`}
#'    \item{`df.rate`:}{ A `data.frame` the locations `x` and associated `rate`
#'      which parameterized the exponential distribution from which the group
#'      sizes were drawn.}
#'  }
#'
#' @examples
#' \donttest{
#' if (require(ggplot2, quietly = TRUE)) {
#'   # Load the data
#'
#'   data("toygroups", package = "inlabru")
#'
#'   # The data set is a simulation of animal groups residing in a 1D space.
#'   # Their locations in x-space are sampled from a Cox process with
#'   # intensity
#'
#'   ggplot(toygroups$df.intensity) +
#'     geom_line(aes(x = x, y = g.lambda))
#'
#'   # Adding the simulated group locations to this plot we obtain
#'
#'   ggplot(toygroups$df.intensity) +
#'     geom_line(aes(x = x, y = g.lambda)) +
#'     geom_point(data = toygroups$groups, aes(x, y = 0), pch = "|")
#'
#'   # Each group has a size mark attached to it.
#'   # These group sizes are sampled from an exponential distribution
#'   # for which the rate parameter depends on the x-coordinate
#'
#'   ggplot(toygroups$groups) +
#'     geom_point(aes(x = x, y = size))
#'
#'   ggplot(toygroups$df.rate) +
#'     geom_line(aes(x, rate))
#' }
#' }
#'
"toygroups"

#' @title Transformation tools
#' @description Tools for transforming between N(0,1) variables and other
#' distributions in predictor expressions
#' @param qfun A quantile function object, such as `qexp`
#' @param x Values to be transformed
#' @param \dots Distribution parameters passed on to the `qfun` and `pfun`
#'   functions
#' @param tail.split. For x-values larger than `tail.split.`, upper quantile
#'   calculations are used internally, and for smaller values lower quantile
#'   calculations are used. This can avoid lack of accuracy in the distribution
#'   tails. If `NULL`, forward calculations split at 0, and inverse calculations
#'   use lower tails only, potentially losing accuracy in the upper tails.
#' @return * For `bru_forward_transformation`, a numeric vector
#' @export
#' @rdname bru_transformation
#' @name bru_transformation
#' @examples
#' u <- rnorm(5, 0, 1)
#' y <- bru_forward_transformation(qexp, u, rate = 2)
#' v <- bru_inverse_transformation(pexp, y, rate = 2)
#' rbind(u, y, v)
#'
bru_forward_transformation <- function(qfun, x, ..., tail.split. = 0) {
  if (is.null(tail.split.)) {
    # By default, split at 0
    upper <- x >= 0
  } else {
    upper <- x >= tail.split.
  }
  param <- list(...)
  param_upper <- param_lower <- param
  for (k in seq_along(param)) {
    if (length(param[[k]]) == length(x)) {
      param_upper[[k]] <- param[[k]][upper]
      param_lower[[k]] <- param[[k]][!upper]
    }
  }
  res <- numeric(length(x))
  if (sum(upper) > 0) {
    res[upper] <-
      do.call(
        qfun,
        c(
          list(
            pnorm(
              x[upper],
              lower.tail = FALSE,
              log.p = TRUE
            )
          ),
          param_upper,
          list(
            lower.tail = FALSE,
            log.p = TRUE
          )
        )
      )
  }
  if (sum(!upper) > 0) {
    res[!upper] <-
      do.call(
        qfun,
        c(
          list(
            pnorm(
              x[!upper],
              lower.tail = TRUE,
              log.p = TRUE
            )
          ),
          param_lower,
          list(
            lower.tail = TRUE,
            log.p = TRUE
          )
        )
      )
  }
  res
}

#' @param pfun A CDF function object, such as `pexp`
#' @param x Values to be transformed
#' @return * For `bru_inverse_transformation`, a numeric vector
#' @export
#' @rdname bru_transformation

bru_inverse_transformation <- function(pfun, x, ..., tail.split. = NULL) {
  if (is.null(tail.split.)) {
    # By default, upper = FALSE
    upper <- logical(length(x))
  } else {
    upper <- x >= tail.split.
  }
  param <- list(...)
  param_upper <- param_lower <- param
  for (k in seq_along(param)) {
    if (length(param[[k]]) == length(x)) {
      param_upper[[k]] <- param[[k]][upper]
      param_lower[[k]] <- param[[k]][!upper]
    }
  }
  res <- numeric(length(x))
  if (sum(upper) > 0) {
    res[upper] <-
      qnorm(
        do.call(
          pfun,
          c(
            list(x[upper]),
            param_upper,
            list(
              lower.tail = FALSE,
              log.p = TRUE
            )
          )
        ),
        lower.tail = FALSE,
        log.p = TRUE
      )
  }
  if (sum(!upper) > 0) {
    res[!upper] <-
      qnorm(
        do.call(
          pfun,
          c(
            list(x[!upper]),
            param_lower,
            list(
              lower.tail = TRUE,
              log.p = TRUE
            )
          )
        ),
        lower.tail = TRUE,
        log.p = TRUE
      )
  }
  res
}


# p = 0.5 + 0.5 * sign(q) * (1 - exp(-abs(q)))
#   = 0.5 + 0.5 * sign(q) - 0.5 * sign(q) * exp(-abs(q))
# 2p = 1 + sign(q) - sign(q) * exp(-abs(q))
# 2p - sign(2p-1) -1 = -sign(2p-1) * exp(-abs(q))
# -(2p-1-sign(2p-1))/sign(2p-1) = exp(-abs(q))
# -(2p-1)/sign(2p-1)+1 = exp(-abs(q))
# 1 - |2p-1| = exp(-abs(q))
# -log(1 - |2p-1|) = abs(q)
# -sign(2p-1) * log(1 - |2p-1|) = q
# -sign(2p-1) * log1p(- |2p-1|) = q
#
# p <= 0.5: log(2p) = log(2) + log(p)
# p >= 0.5: -log(2-2p) = log(2) + log(1-p) = q

qlaplace <- function(p, rate = 1, lower.tail = TRUE, log.p = FALSE) {
  q <- numeric(length(p))
  if (log.p) {
    upper <- p > log(1 / 2)
    q[upper] <- -log(2) - log1p(-exp(p[upper]))
    q[!upper] <- log(2) + p[!upper]
  } else {
    upper <- p > 1 / 2
    q[upper] <- -log(2) - log1p(-p[upper])
    q[!upper] <- log(2) + log(p[!upper])
  }
  if (!lower.tail) {
    q <- -q
  }
  q / rate
}


plaplace <- function(q, rate = 1, lower.tail = TRUE, log.p = FALSE) {
  p <- numeric(length(q))
  if (lower.tail) {
    q <- q * rate
  } else {
    q <- -q * rate
  }
  if (log.p) {
    upper <- q > 0
    p[upper] <- log1p(-exp(-q[upper]) / 2)
    p[!upper] <- q[!upper] - log(2)
  } else {
    upper <- q > 0
    p[upper] <- 1 - exp(-q[upper]) / 2
    p[!upper] <- exp(q[!upper]) / 2
  }
  p
}

#' @include deprecated.R
#' @include mappers.R

## _repeat ####

#' @title Mapper for repeating a mapper
#' @param mapper The mapper to be repeated.
#' @param n_rep The number of times to repeat the mapper. If a vector, the
#'   non-interleaved repeats are combined into a single repeat mapping,
#'   and combined with interleaved repeats via a [bm_sum()] of mappers.
#' @param interleaved logical; if `TRUE`, the repeated mapping columns are
#' interleaved; `(x1[1], x2[1], ..., x1[2], x2[2], ...)`.
#' If `FALSE` (default), the repeated
#' mapping columns are contiguous, `(x1[1], x1[2], ..., x2[1], x2[2], ...)`,
#' and the Jacobian is a `cbind()` of the Jacobians of the repeated mappers.
#'
#' If `n_rep` is a vector, `interleaved` should either be a single logical,
#' or a vector of the same length. Each element applies to the corresponding
#' `n_rep` repetition specification.
#' @export
#' @description Defines a repeated-space mapper that sums the contributions for
#'   each copy. The `ibm_n()` method returns `ibm_n(mapper) * n_rep`, and
#'   `ibm_values()` returns `seq_len(ibm_n(mapper))`.
#' @returns A `bm_repeat` or `bm_sum` object, or the original
#' input `mapper`.
#' @rdname bm_repeat
#' @inheritParams bru_mapper_generics
#' @inheritParams bm_scale
#' @inheritParams bm_multi
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' (m0 <- bm_index(3))
#' (m <- bm_repeat(m0, 5))
#' ibm_n(m)
#' ibm_values(m)
#' ibm_jacobian(m, 1:3)
#' ibm_eval(m, 1:3, seq_len(ibm_n(m)))
#'
#' # Interleaving and grouping
#' (m <- bm_repeat(m0, c(2, 1, 2), c(TRUE, FALSE, FALSE)))
#' ibm_n(m)
#' ibm_values(m)
#' ibm_jacobian(m, 1:3)
#' ibm_eval(m, 1:3, seq_len(ibm_n(m)))
#'
bm_repeat <- function(mapper, n_rep, interleaved = FALSE) {
  stopifnot((length(n_rep) > 0L) && sum(n_rep) > 0L)
  if ((length(n_rep) == 1L) || !any(interleaved)) {
    n_rep <- sum(n_rep)
    if (n_rep == 1L) {
      return(mapper)
    }
    mapper <- list(
      mapper = mapper,
      n_rep = n_rep,
      is_linear = ibm_is_linear(mapper),
      interleaved = interleaved
    )
    return(bru_mapper_define(mapper, new_class = "bm_repeat"))
  }

  # Combine non-interleaved repeats
  if (length(interleaved) == 1L) {
    interleaved <- rep(interleaved, length(n_rep))
  }
  mappers <- list()
  m_idx <- 0L
  n_rep_cumulative <- 0L
  for (k in seq_along(n_rep)) {
    if (interleaved[k]) {
      if (n_rep_cumulative > 0L) {
        m_idx <- m_idx + 1L
        mappers[[m_idx]] <- bm_repeat(
          mapper,
          n_rep = n_rep_cumulative,
          interleaved = FALSE
        )
        n_rep_cumulative <- 0L
      }
      m_idx <- m_idx + 1L
      mappers[[m_idx]] <- bm_repeat(
        mapper,
        n_rep = n_rep[k],
        interleaved = TRUE
      )
    } else {
      n_rep_cumulative <- n_rep_cumulative + n_rep[k]
    }
  }
  if (n_rep_cumulative > 0L) {
    m_idx <- m_idx + 1L
    mappers[[m_idx]] <- bm_repeat(
      mapper,
      n_rep = n_rep_cumulative,
      interleaved = FALSE
    )
    n_rep_cumulative <- 0L
  }

  if (m_idx == 1L) {
    return(mappers[[1]])
  }
  mapper_ <- bm_sum(mappers, single_input = TRUE)
  mapper_
}

#' @export
#' @rdname bm_repeat
bru_mapper_repeat <- function(...) {
  bm_repeat(...)
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_repeat <- function(mapper, ...) {
  ibm_n(mapper[["mapper"]], ...) * mapper[["n_rep"]]
}
#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_repeat <- function(mapper, ...) {
  ibm_n_output(mapper[["mapper"]], ...)
}

#' @export
#' @rdname ibm_values
#'
ibm_values.bm_repeat <- function(mapper, ...) {
  seq_len(ibm_n(mapper, ...))
}


bm_repeat_sub_lin <- function(mapper, input, state,
                              ...) {
  # We need all the sub_lin objects even for linear mappers
  idx <- bm_repeat_indexing(
    n_map = ibm_n(mapper[["mapper"]]),
    n_rep = mapper[["n_rep"]],
    interleaved = mapper[["interleaved"]]
  )
  sub_lin <-
    lapply(
      seq_len(mapper[["n_rep"]]),
      function(x) {
        state_subset <- state[idx$offsets[x] + idx$index]
        ibm_linear(
          mapper[["mapper"]],
          input = input,
          state = state_subset
        )
      }
    )
  sub_lin
}


#' @describeIn ibm_jacobian The input should take the format of the
#'   repeated submapper.
#' @export
#'
ibm_jacobian.bm_repeat <- function(mapper, input, state = NULL,
                                   inla_f = FALSE,
                                   multi = FALSE,
                                   ...,
                                   sub_lin = NULL) {
  if (is.null(sub_lin)) {
    sub_lin <- bm_repeat_sub_lin(mapper, input, state)
  }
  A <- lapply(sub_lin, function(x) x[["jacobian"]])

  if (multi) {
    return(A)
  }

  # Combine the matrices (A1, A2, A3, ...) -> permuted cbind(A1, A2, A3, ...)
  A <- do.call(cbind, A)
  if (isTRUE(mapper[["interleaved"]])) {
    A <- A %*% bm_repeat_indexing_matrix(
      n_map = ibm_n(mapper[["mapper"]]),
      n_rep = mapper[["n_rep"]]
    )
  }
  A
}


#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_repeat <- function(mapper, input, state,
                               multi = FALSE,
                               ...,
                               sub_lin = NULL) {
  if (is.null(sub_lin)) {
    sub_lin <- bm_repeat_sub_lin(mapper, input, state)
  }
  val <- lapply(sub_lin, function(x) x[["offset"]])
  if (multi) {
    return(val)
  }

  # Combine the vectors (b1, b2, b3, ...) -> b1 + b2 + b3 + ...
  val <- as.vector(Matrix::rowSums(do.call(cbind, val)))
  val
}


#' @export
#' @rdname ibm_linear
#'
ibm_linear.bm_repeat <- function(mapper, input, state,
                                 ...) {
  sub_lin <-
    bm_repeat_sub_lin(
      mapper, input, state,
      ...
    )
  eval2 <- ibm_eval2(
    mapper,
    input = input,
    state = state,
    multi = FALSE,
    ...,
    sub_lin = sub_lin
  )
  bm_taylor(
    offset = eval2$offset,
    jacobian = eval2$jacobian,
    state0 = state,
    values_mapper = mapper
  )
}


#' @describeIn ibm_invalid_output
#' Passes on the input to the corresponding method.
#' @export
#'
ibm_invalid_output.bm_repeat <- function(mapper, input, state,
                                         ...) {
  idx <- bm_repeat_indexing(
    n_map = ibm_n(mapper[["mapper"]]),
    n_rep = mapper[["n_rep"]],
    interleaved = mapper[["interleaved"]]
  )
  ibm_invalid_output(
    mapper[["mapper"]],
    input = input,
    state = state[idx$offsets[1] + idx$index]
  )
}

# Interleaving ####

#' @title Re-indexing matrix for repeated mappers
#' @name bm_repeat_indexing
#' @rdname bm_repeat_indexing
#' @description Helper methods for regular and interleaved state vector
#'   indexing, meant for use in [bm_repeat] mappers.
#' @keywords internal
NULL

#' @describeIn bm_repeat_indexing Construct the offsets and within-block
#' index vectors for a repeated mapper, with or without interleaving.
#' @param n_map The block mapper size
#' @param n_rep The number of times the block is repeated
#' @param interleaved logical; if `TRUE`, the state vector indexing is
#'   interleaved. Default is `FALSE`, for blockwise indexing.
#' @returns `bm_repeat_indexing`: Returns a list with a vector of offsets,
#'   `offsets`, and a vector of relative index values, `index`; block `k`
#'   indexing is given by `offsets[k] + index`.
#' @export
#' @examples
#' (idx <- bm_repeat_indexing(3, 2, FALSE))
#' (idx <- bm_repeat_indexing(3, 2, TRUE))
bm_repeat_indexing <- function(n_map,
                               n_rep,
                               interleaved = FALSE) {
  if (isTRUE(interleaved)) {
    list(
      offsets = seq_len(n_rep) - 1L,
      index = n_rep * (seq_len(n_map) - 1L) + 1L
    )
  } else {
    list(
      offsets = n_map * (seq_len(n_rep) - 1L),
      index = seq_len(n_map)
    )
  }
}
#' @describeIn bm_repeat_indexing Creates a sparse matrix `A` such that
#'   `z <- A %*% x` constructs a blockwise version `z = c(x1, x2, ..., xn)` of
#'   an interleaved state vector `x = c(x1[1], x2[1], ..., x1[2], x2[2], ...)`.
#'   Each block is of size `n_map`, and there are `n_rep` blocks. The reverse
#'   operation, taking a blockwise `z = c(x1, x2, ..., xn)` to an interleaved
#'   vector is `x <- Matrix::t(A) %*% z`.
#' @returns `bm_repeat_indexing_matrix`: A `sparseMatrix` object.
#' @export
#' @examples
#' (A <- bm_repeat_indexing_matrix(3, 2))
#' (x_interleaved <- 1:6)
#' (x_blockwise <- as.vector(A %*% x_interleaved))
#' (x_recovered <- as.vector(Matrix::t(A) %*% x_blockwise))
#'
bm_repeat_indexing_matrix <- function(n_map, n_rep) {
  Matrix::sparseMatrix(
    i = rep((seq_len(n_rep) - 1L) * n_map, times = n_map) +
      rep(seq_len(n_map), each = n_rep),
    j = seq_len(n_map * n_rep),
    x = 1,
    dims = c(1L, 1L) * n_map * n_rep
  )
}

#' Build an inla data stack from linearisation information
#'
#' Combine linearisation for multiple likelihoods
#'
#' @param \dots Arguments passed on to other methods
#' @export
#' @keywords internal
#' @rdname bru_make_stack
bru_make_stack <- function(...) {
  UseMethod("bru_make_stack")
}

#' @param lhood A `bru_obs` object
#' @param lin Linearisation information
#' * For `.bru_obs`, a `bm_taylor` object
#' * For `.bru_obs_list`, a list of `bm_taylor` objects
#' @param idx Output from [bru_index.bru_model()]
#' @param family_index integer specifying the family sequence index of the
#'   observation model
#' @export
#' @rdname bru_make_stack
bru_make_stack.bru_obs <- function(lhood, lin, idx, ..., family_index = 1L) {
  stopifnot(inherits(lin, c("bm_taylor", "bru_mapper_taylor")))
  stopifnot(!is.null(lin[["offset"]]))
  stopifnot(is.null(lin[["jacobian"]]) || is.list(lin[["jacobian"]]))
  stopifnot(is.null(lin[["state0"]]))

  if (is.null(lhood[["response_data"]])) {
    BRU.response <- lhood$data[[lhood$response]]
  } else {
    BRU.response <- lhood$response_data[[lhood$response]]
  }
  nms <- names(lin$jacobian)
  if (utils::packageVersion("INLA") <= "24.06.02") {
    INLA::inla.stack(
      list(
        BRU.response = BRU.response,
        BRU.E = lhood[["response_data"]][["BRU_E"]],
        BRU.Ntrials = lhood[["response_data"]][["BRU_Ntrials"]],
        BRU.weights = lhood[["response_data"]][["BRU_weights"]],
        BRU.scale = lhood[["response_data"]][["BRU_scale"]],
        BRU.offset = as.vector(lin$offset),
        BRU.link = family_index
      ),
      A = lapply(nms, function(nm) {
        lin$jacobian[[nm]][, idx[["inla_subset"]][[nm]], drop = FALSE]
      }),
      effects = idx[["idx_inla"]][nms],
      remove.unused = FALSE
    )
  } else {
    INLA::inla.stack(
      list(
        BRU.E = lhood[["response_data"]][["BRU_E"]],
        BRU.Ntrials = lhood[["response_data"]][["BRU_Ntrials"]],
        BRU.weights = lhood[["response_data"]][["BRU_weights"]],
        BRU.scale = lhood[["response_data"]][["BRU_scale"]],
        BRU.offset = as.vector(lin$offset),
        BRU.link = family_index
      ),
      A = lapply(nms, function(nm) {
        lin$jacobian[[nm]][, idx[["inla_subset"]][[nm]], drop = FALSE]
      }),
      effects = idx[["idx_inla"]][nms],
      responses = list(BRU.response),
      remove.unused = FALSE
    )
  }
}


#' @param lhoods A `bru_obs_list` object
#' @export
#' @rdname bru_make_stack
bru_make_stack.bru_obs_list <- function(lhoods, lin, idx, ...) {
  stks <-
    lapply(
      seq_along(lhoods),
      function(lh_idx) {
        bru_make_stack(
          lhood = lhoods[[lh_idx]],
          lin = lin[[lh_idx]],
          idx = idx,
          family_index = lh_idx
        )
      }
    )

  # TODO: Either fix the check in inla that removes n&values information from
  # model="linear" input in INLA::f(), making it unable to handle all-NA linear
  # effect specifications (arguably, that check is wrong, since it doesn't need
  # that information for linear components), or remove all-but one unused value
  # above or in bru_make_stack.bru_obs, and keep using remove.unused=FALSE
  # here.
  stk <-
    do.call(
      bru_inla.stack.mjoin,
      c(stks, list(compress = TRUE, remove.unused = FALSE))
    )

  stk
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo=FALSE,include=FALSE------------------------------------------

## ----echo=FALSE,results="asis"------------------------------------------------
files <- list.files(".", "\\.Rmd$")
for (file in files) {
  lines <- readLines(file, n = 10)
  title_idx <- grep("^title: ", lines)
  if (length(title_idx) > 0) {
    title <- sub("^title: ", "", lines[title_idx[1]])
    title <- sub('^"', "", title)
    title <- sub('"$', "", title)
    cat("* [",
      title,
      "](https://inlabru-org.github.io/inlabru/articles/",
      sub("\\.Rmd", ".html", file),
      ")\n",
      sep = ""
    )
  }
}

## ----echo=FALSE,results="asis"------------------------------------------------
files <- list.files("articles", "\\.Rmd$")
for (file in file.path("articles", files)) {
  lines <- readLines(file, n = 10)
  title_idx <- grep("^title: ", lines)
  if (length(title_idx) > 0) {
    title <- sub("^title: ", "", lines[title_idx[1]])
    title <- sub('^"', "", title)
    title <- sub('"$', "", title)
    cat("* [",
      title,
      "](https://inlabru-org.github.io/inlabru/articles/",
      sub("\\.Rmd", ".html", basename(file)),
      ")\n",
      sep = ""
    )
  }
}

#' @include deprecated.R
#' @include mappers.R

## _sum ####

#' @title Mapper for adding multiple mappers
#' @param mappers A list of `bru_mapper` objects.
#' @param single_input logical. If `TRUE`, the input is passed to all
#' sub-mappers. Otherwise, the input should be a list, data.frame, or matrix.
#' If the `mappers` list has named entries, the `input` can reference their
#' corresponding sub-mapper using its name.
#' @export
#' @description Defines a mapper that adds the effects of each submapper.
#'   The `ibm_n()` method returns the sum of `ibm_n(mappers[[k]])`, and
#'   `ibm_values()` returns `seq_len(ibm_n(mapper))`.
#' @returns A `bm_sum` object.
#' @rdname bm_sum
#' @seealso [bru_mapper], [bru_mapper_generics]
#' @family mappers
#' @examples
#' (m <- bm_sum(list(a = bm_index(3), b = bm_index(2))))
#' ibm_n(m)
#' ibm_values(m)
#' ibm_jacobian(m, list(a = 1:3, b = c(1, 1, 2)))
#' ibm_eval(
#'   m,
#'   list(a = 1:3, b = c(1, 1, 2)),
#'   seq_len(ibm_n(m))
#' )
#'
bm_sum <- function(mappers, single_input = FALSE) {
  mappers <- as_bm_list(mappers)
  if (is.null(names(mappers))) {
    names(mappers) <- as.character(seq_along(mappers))
  }
  mapper <- list(
    mappers = mappers,
    n_multi = lapply(mappers, ibm_n),
    values_multi = lapply(mappers, ibm_values),
    is_linear_multi = lapply(mappers, ibm_is_linear),
    is_rowwise_multi = lapply(mappers, ibm_is_rowwise),
    single_input = single_input
  )
  mapper[["n"]] <- sum(unlist(mapper[["n_multi"]]))
  mapper[["values"]] <- seq_len(mapper[["n"]])
  mapper[["is_linear"]] <- all(unlist(mapper[["is_linear_multi"]]))
  mapper[["is_rowwise"]] <- all(unlist(mapper[["is_rowwise_multi"]]))
  bru_mapper_define(mapper, new_class = "bm_sum")
}

#' @export
#' @param \dots Arguments passed on to [bm_sum()]
#' @rdname bm_sum
bru_mapper_sum <- function(...) {
  bm_sum(...)
}

#' @export
#' @rdname ibm_n
#'
ibm_n.bm_sum <- function(mapper,
                         inla_f = FALSE,
                         multi = FALSE,
                         ...) {
  if (multi) {
    mapper[["n_multi"]]
  } else {
    mapper[["n"]]
  }
}


bm_sum_prepare_input <- function(mapper, input) {
  if (mapper[["single_input"]]) {
    return(input)
  }
  if (is.matrix(input)) {
    input_names <- colnames(input)
    input <- as.data.frame(input)
  } else {
    input_names <- names(input)
  }
  if (!is.null(input_names)) {
    return(input)
  }
  if (is.data.frame(input)) {
    names(input) <- names(mapper[["mappers"]])[seq_len(ncol(input))]
  } else {
    if (is.null(input)) {
      input <- rep(list(NULL), length(mapper[["mappers"]]))
    }
    # input should now be a list
    names(input) <- names(mapper[["mappers"]])[seq_along(input)]
  }
  input
}


#' @export
#' @rdname ibm_n_output
#'
ibm_n_output.bm_sum <- function(mapper, input, state = NULL, ...) {
  input <- bm_sum_prepare_input(mapper, input)
  # Assume that the first mapper fully handles the output size
  nm <- names(mapper[["mappers"]])[1]
  ibm_n_output(
    mapper[["mappers"]][[nm]],
    if (mapper[["single_input"]]) {
      input
    } else {
      input[[nm]]
    },
    state = state[seq_len(ibm_n(mapper[["mappers"]][[nm]]))],
    ...
  )
}

#' @export
#' @rdname ibm_values
#'
ibm_values.bm_sum <- function(mapper,
                              inla_f = FALSE,
                              multi = FALSE,
                              ...) {
  if (multi) {
    mapper[["values_multi"]]
  } else {
    mapper[["values"]]
  }
}

#' @export
#' @rdname ibm_is_linear
#'
ibm_is_linear.bm_sum <- function(mapper,
                                 multi = FALSE,
                                 ...) {
  if (multi) {
    mapper[["is_linear_multi"]]
  } else {
    mapper[["is_linear"]]
  }
}


bm_sum_sub_lin <- function(mapper, input, state, ...) {
  input <- bm_sum_prepare_input(mapper, input)

  indexing <- names(mapper[["mappers"]])
  n_multi <- unlist(mapper[["n_multi"]])
  n_offset <- c(0, cumsum(n_multi))
  names(n_offset) <- c(names(n_multi), "THEEND")
  sub_lin <-
    lapply(
      indexing,
      function(x) {
        state_subset <- state[n_offset[x] + seq_len(mapper[["n_multi"]][[x]])]
        ibm_linear(
          mapper[["mappers"]][[x]],
          input =
            if (mapper[["single_input"]]) {
              input
            } else {
              input[[x]]
            },
          state = state_subset
        )
      }
    )
  sub_lin
}


#' @describeIn ibm_jacobian
#' Accepts a list with
#' named entries, or a list with unnamed but ordered elements.
#' The names must match the sub-mappers, see [ibm_names.bm_sum()].
#' Each list element should take a format accepted by the corresponding
#' sub-mapper. In case each element is a vector, the input can be given as a
#' data.frame with named columns, a matrix with named columns, or a matrix
#' with unnamed but ordered columns.
#' @export
#'
ibm_jacobian.bm_sum <- function(mapper, input, state = NULL,
                                inla_f = FALSE,
                                multi = FALSE,
                                ...,
                                sub_lin = NULL) {
  if (is.null(sub_lin)) {
    sub_lin <- bm_sum_sub_lin(mapper, input, state)
  }
  A <- lapply(sub_lin, function(x) x[["jacobian"]])

  if (multi) {
    return(A)
  }

  # Combine the matrices (A1, A2, A3, ...) -> cbind(A1, A2, A3, ...)
  A <- do.call(cbind, A)
  A
}


#' @export
#' @rdname ibm_eval
#'
ibm_eval.bm_sum <- function(mapper, input, state,
                            multi = FALSE,
                            ...,
                            sub_lin = NULL) {
  if (is.null(sub_lin)) {
    sub_lin <- bm_sum_sub_lin(mapper, input, state)
  }
  val <- lapply(sub_lin, function(x) x[["offset"]])
  if (multi) {
    return(val)
  }

  # Combine the vectors (b1, b2, b3, ...) -> b1 + b2 + b3 + ...
  val <- as.vector(Matrix::rowSums(do.call(cbind, val)))
  val
}


#' @export
#' @rdname ibm_linear
#'
ibm_linear.bm_sum <- function(mapper, input, state,
                              ...) {
  sub_lin <-
    bm_sum_sub_lin(
      mapper, input, state,
      ...
    )
  eval2 <- ibm_eval2(
    mapper,
    input = input,
    state = state,
    multi = FALSE,
    ...,
    sub_lin = sub_lin
  )
  bru_mapper_taylor(
    offset = eval2$offset,
    jacobian = eval2$jacobian,
    state0 = state,
    values_mapper = mapper
  )
}


#' @describeIn ibm_invalid_output
#' Accepts a list with
#' named entries, or a list with unnamed but ordered elements.
#' The names must match the sub-mappers, see [ibm_names.bm_sum()].
#' Each list element should take a format accepted by the corresponding
#' sub-mapper. In case each element is a vector, the input can be given as a
#' data.frame with named columns, a matrix with named columns, or a matrix
#' with unnamed but ordered columns.
#' @export
#'
ibm_invalid_output.bm_sum <- function(mapper, input, state,
                                      multi = FALSE,
                                      ...) {
  input <- bm_sum_prepare_input(mapper, input)
  invalid <-
    lapply(
      names(mapper[["mappers"]]),
      function(x) {
        if (mapper[["single_input"]]) {
          inp <- input
        } else {
          inp <- input[[x]]
        }
        ibm_invalid_output(
          mapper[["mappers"]][[x]],
          input = inp,
          multi = FALSE
        )
      }
    )

  if (multi) {
    return(invalid)
  }

  # Combine the vectors (v1, v2, v3) -> (v1 | v2 | v3)
  invalid_ <- invalid[[1]]
  for (k in seq_len(length(invalid) - 1L) + 1L) {
    invalid_ <- invalid_ | invalid[[k]]
  }
  invalid_
}

#' @return
#' * `[`-indexing a `bm_sum` extracts a subset
#'   `bm_sum` object (for drop `FALSE`) or an individual sub-mapper
#'   (for drop `TRUE`, and `i` identifies a single element)
#' @export
#' @param x object from which to extract element(s)
#' @param i indices specifying element(s) to extract
#' @param drop logical;
#' For `[.bm_sum`, whether to extract an individual mapper when
#' `i` identifies a single element. If `FALSE`, a list of sub-mappers is
#' returned (suitable e.g. for creating a new `bm_sum` object).
#' Default: `TRUE`
#' @rdname bm_sum
`[.bm_sum` <- function(x, i, drop = TRUE) {
  if (is.logical(i)) {
    i <- which(i)
  }
  mapper <- x[["mappers"]][i]
  if (drop) {
    if (length(mapper) == 1) {
      mapper <- mapper[[1]]
    } else if (length(mapper) == 0) {
      mapper <- NULL
    }
  }
  mapper
}

#' @export
#' @rdname bm_sum
`[.bru_mapper_sum` <- function(x, i, drop = TRUE) {
  if (is.logical(i)) {
    i <- which(i)
  }
  mapper <- x[["mappers"]][i]
  if (drop) {
    if (length(mapper) == 1) {
      mapper <- mapper[[1]]
    } else if (length(mapper) == 0) {
      mapper <- NULL
    }
  }
  mapper
}

#' @describeIn ibm_names
#' Returns the names from the sub-mappers list
#' @export
#'
`ibm_names.bm_sum` <- function(mapper) {
  names(mapper[["mappers"]])
}

#' @export
#' @rdname ibm_names
#'
`ibm_names<-.bm_sum` <- function(mapper, value) {
  names(mapper[["mappers"]]) <- value
  names(mapper[["n_multi"]]) <- value
  names(mapper[["values_multi"]]) <- value
  mapper
}

#' @export
#' @rdname ibm_names
#'
`ibm_names<-.bru_mapper_sum` <- function(mapper, value) {
  names(mapper[["mappers"]]) <- value
  names(mapper[["n_multi"]]) <- value
  names(mapper[["values_multi"]]) <- value
  mapper
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @title Convert node-block indices into groups of node indices
#' @description Converts a vector of block indices into a list of
#' node index vectors.
#' @param block An integer vector of 1-based block indices, one for each node.
#' @param max_block Precomputed maximal value of `block`. Must be at least
#' as large as `max(block)`. Block indices larger than `max_block` are ignored.
#' @param per_node Logical; if `FALSE`, the function returns a list with
#' one element for each block. If `TRUE`, the function returns a list with
#' one element for each node.
#' @returns If `per_node` is `false`, a list of integer vectors, where each
#' vector contains the 1-based node indices of nodes belonging to the
#' corresponding block. If `per_node` is `true`, a list of integer vectors,
#' where each vector contains the 1-based node indices of nodes belonging to
#' the same block as the corresponding node; `b[i] <- a[block[j]]`
#' @keywords internal
inlabru_group_cv_block_conversion <- function(block, max_block, per_node) {
    .Call(`_inlabru_inlabru_group_cv_block_conversion`, block, max_block, per_node)
}

#' @title Extract predictor or component index information
#'
#' @description
#' `r lifecycle::badge("experimental")`
#' Extract the index vector for a [bru_obs()] predictor,
#' or the whole or a subset of a full [bru()] predictor.
#'
#' @param object A [bru()] or [bru_obs()] output object
#' @param \dots Arguments passed on to sub-methods.
#' @export
bru_index <- function(object, ...) {
  UseMethod("bru_index")
}

#' @describeIn bru_index Extract the index vector for the predictor vector for a
#'   [bru_obs()] sub-model. The indices are relative to the sub-model, and need
#'   to be appropriately offset to be used in the full model predictor.
#' @param what `character` or `NULL`; One of `NULL`, "all", "observed", and
#'   "missing". If `NULL` (default) or "all", gives the index vector for the
#'   full sub-model predictor. If "observed", gives the index vector for the
#'   observed part (response is not `NA`). If "missing", gives the index vector
#'   for the missing part (response is `NA`) of the model.
#' @examplesIf bru_safe_inla()
#' fit <- bru(
#'   ~ 0 + x,
#'   bru_obs(
#'     y ~ .,
#'     data = data.frame(x = 1:3, y = 1:3 + rnorm(3)),
#'     tag = "A"
#'   ),
#'   bru_obs(
#'     y ~ .,
#'     data = data.frame(x = 1:4, y = c(NA, NA, 3:4) + rnorm(4)),
#'     tag = "B"
#'   ),
#'   options = list(bru_run = FALSE) # We only need the model structure
#' )
#' bru_index(fit)
#' bru_index(fit, "A")
#' bru_index(fit, "B")
#' bru_index(fit, c("B", "A"))
#' bru_index(fit, what = "missing")
#' @export
#' @returns * `bru_index(bru_obs)`: An `integer` vector.
bru_index.bru_obs <- function(object, what = NULL, ...) {
  what <- match.arg(what, c("all", "observed", "missing"))
  size <- bru_response_size(object)
  idx <- seq_len(size)
  if (identical(what, "all")) {
    return(idx)
  }
  resp <- object[["response_data"]][[object[["response"]]]]
  if (is.vector(resp)) {
    miss <- is.na(resp)
  } else {
    # Should work for both inla.surv and inla.mdata objects:
    miss <- is.na(resp[[1]])
  }
  if (identical(what, "observed")) {
    return(idx[!miss])
  }
  idx[miss]
}

#' @describeIn bru_index Extract the index vector for "APredictor" for one or
#'  more specified observation [bru_obs()] sub-models. Accepts any combination
#'  of `tag` and `what`.
#' @param tag `character` or `integer`; Either a character vector identifying
#'  the tags of one or more of the [bru_obs()] observation models, or an
#'  integer vector identifying models by their [bru()] specification order. If
#'  `NULL` (default) computes indices for all sub-models.
#' @export
#' @returns * `bru_index(bru_obs_list)`: An `integer` vector.
bru_index.bru_obs_list <- function(object, tag = NULL, what = NULL, ...) {
  if (is.null(tag)) {
    tag <- seq_along(object)
  }
  if (length(tag) == 0L) {
    return(integer(0))
  }
  size <- bru_response_size(object)
  off <- c(0L, cumsum(size))
  if (is.character(tag)) {
    ok_tags <- tag %in% names(object)
    if (!all(ok_tags)) {
      stop(glue(
        "Invalid tag(s) {{{glue_collapse(tag[!ok_tags], sep = ', ')}}} for ",
        "bru object with tags {{",
        "{glue_collapse(names(object), sep = ', ')}",
        "}}"
      ))
    }
    unlist(lapply(tag, function(x) {
      x_idx <- which(names(object) %in% x)
      unlist(lapply(x_idx, function(x_idx_) {
        off[x_idx_] + bru_index(object[[x_idx_]], what = what)
      }))
    }))
  } else {
    ok_tags <- (tag >= 1L) &
      (tag <= length(object))
    if (!all(ok_tags)) {
      stop(glue(
        "Invalid tag indices {{{glue_collapse(tag[!ok_tags], sep = ', ')}}}",
        " for bru object with tag indices {{1, ..., ",
        "{length(object)}}}"
      ))
    }
    unlist(lapply(tag, function(x) {
      off[x] + bru_index(object[[x]], what = what)
    }))
  }
}

#' @describeIn bru_index Extract the index vector for "APredictor" for one or
#'   more specified observation [bru_obs()] sub-models. Accepts any combination
#'   of `tag` and `what`.
#' @export
#' @returns * `bru_index(bru)`: An `integer` vector.
bru_index.bru <- function(object, tag = NULL, what = NULL, ...) {
  bru_index(object[["bru_info"]], tag = tag, what = what, ...)
}
#' @describeIn bru_index Extract the index vector for "APredictor" for one or
#'   more specified observation [bru_obs()] sub-models. Accepts any combination
#'   of `tag` and `what`.
#' @export
#' @returns * `bru_index(bru_info)`: An `integer` vector.
bru_index.bru_info <- function(object, tag = NULL, what = NULL, ...) {
  bru_index(object[["lhoods"]], tag = tag, what = what, ...)
}

#' @export
#' @param object A [component].
#' @param inla_f logical; when `TRUE`, must result in
#' values compatible with `INLA::f(...)`
#' an specification and corresponding `INLA::inla.stack(...)` constructions.
#' @returns * `bru_index(bru_comp)`:
#'   A list of indices into the latent variables compatible with the
#'   component mapper.
#' @author Fabian E. Bachl \email{bachlfab@@gmail.com},
#'   Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname bru_index

bru_index.bru_comp <- function(object, inla_f, ...) {
  idx <- ibm_values(object[["mapper"]], inla_f = inla_f, multi = TRUE)
  if (is.null(idx)) {
    return(list())
  }
  nms <- names(idx)
  present <- match(nms, c("main", "group", "replicate"))
  names(idx) <-
    glue("{object[['label']]}{suffix}",
      suffix = c("", ".group", ".repl")[present]
    )
  idx
}


#' @export
#' @rdname bru_index
#' @returns * `bru_index(bru_comp_list)`:
#'   A list of list of indices into the latent variables compatible with each
#'   component mapper.
bru_index.bru_comp_list <- function(object, inla_f, ...) {
  lapply(object, function(x) bru_index(x, inla_f = inla_f, ...))
}


#' @describeIn bru_index Compute all index values for a [bru_model()] object.
#' Computes the index value matrices for included components according to the
#' `used` argument.
#'
#' @param used A [bru_used()] object
#' @return * `bru_index(bru_model)`: A named list of `idx_full` and `idx_inla`,
#' named list of indices, and `inla_subset`, and `inla_subset`,
#' a named list of logical subset specifications for extracting the `INLA::f()`
#' compatible index subsets.
#' @export
bru_index.bru_model <- function(object, used, ...) {
  stopifnot(inherits(object, "bru_model"))
  included <- union(used[["effect"]], used[["latent"]])

  comp_lst <- as_bru_comp_list(object)[included]
  list(
    idx_full = bru_index(comp_lst, inla_f = FALSE),
    idx_inla = bru_index(comp_lst, inla_f = TRUE),
    inla_subset = inla_subset_eval(comp_lst)
  )
}


#' @export
#' @describeIn bru_index `r lifecycle::badge("deprecated")` since "2.12.0.9023".
#' Use `bru_index()` instead.
index_eval <- function(...) {
  lifecycle::deprecate_warn(
    when = "2.12.0.9023",
    "index_eval()",
    "bru_index()"
  )
  bru_index(...)
}

#' Obtain inla index subset information
#'
#' Subsets for `INLA::f()` compatible indexing
#'
#' @export
#' @param \dots Passed on to submethods.
#' @rdname inla_subset_eval
inla_subset_eval <- function(...) {
  UseMethod("inla_subset_eval")
}

#' @export
#' @keywords internal
#' @param components A component list.
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#' @rdname inla_subset_eval

inla_subset_eval.bru_comp_list <- function(components, ...) {
  lapply(components, function(x) ibm_inla_subset(x[["mapper"]], multi = TRUE))
}

# Convert data frame to SpatialLinesDataFrame
#
#
# @keywords internal
# @export
# @param data A SpatialGridDataFrame or SpatialPixelDataFrame
# @param where Spatial*DataFrame of locations for which to fill in values from
# \code{data}. If NULL, use \code{data} to determine the locations.
# @return Spatial object

sfill <- function(data, where = NULL) {
  check_spatstat("spatstat.geom")
  bru_safe_sp(force = TRUE)

  if (is.null(where)) {
    where <- data
  }
  vallist <- list()
  for (k in seq_len(ncol(data@data))) {
    dpoints <- sp::SpatialPoints(data)
    vals <- data@data[, k]
    dpoints <- dpoints[!is.na(vals), ]
    vals <- vals[!is.na(vals)]

    data.ow <- spatstat.geom::owin(
      range(sp::coordinates(dpoints)[, 1]),
      range(sp::coordinates(dpoints)[, 2])
    )
    data.ppp <- spatstat.geom::as.ppp(sp::coordinates(dpoints), data.ow)
    where.ow <- spatstat.geom::owin(
      range(sp::coordinates(where)[, 1]),
      range(sp::coordinates(where)[, 2])
    )
    where.ppp <- spatstat.geom::as.ppp(sp::coordinates(where), where.ow)

    nn <- spatstat.geom::nncross(where.ppp, data.ppp)[, "which"]
    vallist[[k]] <- vals[nn]
  }
  ret <- data.frame(do.call(data.frame, vallist))
  colnames(ret) <- colnames(data@data)
  ret <- sp::SpatialPixelsDataFrame(where, data = ret)
  ret
}


#' Convert data frame to SpatialLinesDataFrame
#'
#' A line in 2D space is defined by a start and an end point, each associated
#' with 2D coordinates. This function takes a `data.frame` as input and assumes
#' that each row defines a line in space. In order to do so, the data frame must
#' have at least four columns and the `start.cols` and `end.cols` parameters
#' must be used to point out the names of the columns that define the start and
#' end coordinates of the line. The data is then converted to a
#' `SpatialLinesDataFrame` `DF`. If a coordinate reference system `crs` is
#' provided it is attached to `DF`. If also `to.crs` is provided, the coordinate
#' system of `DF` is transformed accordingly. Additional columns of the input
#' data, e.g. covariates, are retained and attached to `DF`.
#'
#' @export
#' @param data A data.frame
#' @param start.cols Character array pointing out the columns of `data` that
#'   hold the start points of the lines
#' @param end.cols Character array pointing out the columns of `data` that hold
#'   the end points of the lines
#' @param crs Coordinate reference system of the original `data`
#' @param to.crs Coordinate reference system for the output.
#' @param format Format of the output object. Either `"sp"` (default) or `"sf"`
#' @return `sp::SpatialLinesDataFrame` or [sf::sf]
#' @keywords internal
#'
#' @examples
#' \donttest{
#' # Create a data frame defining three lines
#' lns <- data.frame(
#'   xs = c(1, 2, 3), ys = c(1, 1, 1), # start points
#'   xe = c(2, 3, 4), ye = c(2, 2, 2)
#' ) # end points
#'
#'
#' # Conversion to sf without CRS
#' spl <- sline(lns,
#'   start.cols = c("xs", "ys"),
#'   end.cols = c("xe", "ye"),
#'   format = "sf"
#' )
#'
#' if (require(ggplot2, quietly = TRUE)) {
#'   # Plot the lines
#'   ggplot() +
#'     gg(spl)
#' }
#' }
#'
sline <- function(data, start.cols, end.cols, crs = fm_crs(), to.crs = NULL,
                  format = c("sp", "sf")) {
  format <- match.arg(format)

  sp <- as.data.frame(data[, start.cols])
  ep <- as.data.frame(data[, end.cols])

  colnames(sp) <- c("x", "y")
  colnames(ep) <- c("x", "y")

  if (identical(format, "sp")) {
    bru_safe_sp(force = TRUE)

    lilist <- lapply(seq_len(nrow(sp)), function(k) {
      sp::Lines(list(sp::Line(rbind(sp[k, ], ep[k, ]))), ID = k)
    })
    spl <- sp::SpatialLines(lilist, proj4string = fm_CRS(crs))

    df <- data[, setdiff(names(data), c(start.cols, end.cols))]
    rownames(df) <- seq_len(nrow(df))

    slines <- sp::SpatialLinesDataFrame(spl, data = df)
  } else {
    lin <- lapply(seq_len(nrow(sp)), function(k) {
      sf::st_linestring(as.matrix(rbind(
        sp[k, , drop = FALSE],
        ep[k, , drop = FALSE]
      )))
    })
    lin <- sf::st_sfc(lin, crs = fm_crs(crs))
    df <- data[, setdiff(names(data), c(start.cols, end.cols))]
    rownames(df) <- seq_len(nrow(df))
    slines <- sf::st_sf(geometry = lin, df)
  }

  # If requested, change CRS
  if (!is.null(to.crs)) slines <- fm_transform(slines, to.crs)

  slines
}


#' Convert a data.frame of boundary points into a SpatialPolgonsDataFrame
#'
#' A polygon can be described as a sequence of points defining the polygon's
#' boundary. When given such a sequence (anti clockwise!) this function creates
#' a `SpatialPolygonsDataFrame` or `sf` holding the polygon decribed. By
#' default, the first two columns of `data` are assumed to define the x and y
#' coordinates of the points. This behaviour can be changed using the `cols`
#' parameter, which points out the names of the columns holding the coordinates.
#' The coordinate reference system of the resulting spatial polygon can be set
#' via the `crs` parameter. Posterior conversion to a different CRS is supported
#' using the `to.crs` parameter.
#'
#' @keywords internal
#' @export
#' @param data A data.frame of points describing the boundary of the polygon
#' (unique points, no holes)
#' @param cols Column names of the x and y coordinates within the data
#' @param crs Coordinate reference system of the points
#' @param to.crs Coordinate reference system for the `SpatialLines`/`sf` ouput.
#' @param format Format of the output object. Either "sp" (default) or "sf"
#' @return `sp::SpatialPolygonsDataFrame` or [sf::sf]
#'
#' @examples
#' \donttest{
#' # Create data frame of boundary points (anti clockwise!)
#' pts <- data.frame(
#'   x = c(1, 2, 1.7, 1.3),
#'   y = c(1, 1, 2, 2)
#' )
#'
#' # Convert to sf
#' pol <- spoly(pts, format = "sf")
#'
#' if (require(ggplot2, quietly = TRUE)) {
#'   # Plot it!
#'   ggplot() +
#'     gg(pol)
#' }
#' }
#'
spoly <- function(data,
                  cols = colnames(data)[1:2],
                  crs = fm_crs(),
                  to.crs = NULL,
                  format = c("sp", "sf")) {
  format <- match.arg(format)
  if (identical(format, "sp")) {
    bru_safe_sp(force = TRUE)

    po <- sp::Polygon(data[, cols], hole = FALSE)
    pos <- sp::Polygons(list(po), ID = "tmp")
    predpoly <- sp::SpatialPolygons(list(pos), proj4string = fm_CRS(crs))
    df <- data.frame(weight = 1)
    rownames(df) <- "tmp"
    pol <- sp::SpatialPolygonsDataFrame(predpoly, data = df)
  } else {
    po <- sf::st_polygon(
      list(as.matrix(data[c(seq_len(NROW(data)), 1L), cols]))
    )
    po <- sf::st_sfc(po, crs = fm_crs(crs))
    pol <- sf::st_sf(geometry = po, weight = 1, check_ring_dir = FALSE)
  }

  # If requested, change CRS
  if (!is.null(to.crs)) pol <- fm_transform(pol, to.crs)
  pol
}

#' @include deprecated.R

#' @describeIn inlabru-deprecated
#' Plot a map using extent of a spatial object
#'
#' `r lifecycle::badge("deprecated")` This function is deprecated as `ggmap`
#' isn't supported.
#'
#' Used `ggmap::get_map()` to query map services like Google Maps for a region
#' centered around the spatial object provided. Then calls `ggmap()` to plot the
#' map.
#'
#' This function required the `ggmap` package.
#' @export
gmap <- function(...) {
  lifecycle::deprecate_stop(
    when = "2.11.1.9002",
    what = "gmap()",
    details = "ggmap isn't supported anymore."
  )
  # data <- fm_transform(data, crs = fm_crs("longlat_globe"))
  # df <- cbind(sp::coordinates(data), data@data)
  #
  # # Figure out a sensible bounding box (range of data plus 30%)
  # extend <- 2.5
  # crange <- apply(sp::coordinates(data), MARGIN = 2, range)
  # lonlim <- (extend * (crange[, 1] - mean(crange[, 1]))) + mean(crange[, 1])
  # latlim <- (extend * (crange[, 2] - mean(crange[, 2]))) + mean(crange[, 2])
  #
  # # Create map
  # requireNamespace("ggmap")
  # myMap <-
  #   ggmap::get_map(
  #     c(lonlim[1], latlim[1], lonlim[2], latlim[2]),
  #     ...
  #   )
  #
  # # Return map
  # ggmap::ggmap(myMap)
}

#' ggplot2 geomes for inlabru related objects
#'
#' @description
#' gg is a generic function for generating geomes from various kinds of spatial
#' objects, e.g. Spatial* data, meshes, Raster objects and inla/inlabru
#' predictions. The function invokes particular methods which depend on the
#' [class] of the first argument.
#'
#' @name gg
#' @export
#' @param data an object for which to generate a geom.
#' @param \dots Arguments passed on to the geom method.
#' @return The form of the value returned by gg depends on the class of its
#'   argument. See the documentation of the particular methods for details of
#'   what is produced by that method.
#'
#' @family geomes
#'
#' @examples
#' if (require("ggplot2", quietly = TRUE)) {
#'   # Load Gorilla data
#'
#'   gorillas <- inlabru::gorillas_sf
#'
#'   # Invoke ggplot and add geomes for the Gorilla nests and the survey
#'   # boundary
#'
#'   ggplot() +
#'     gg(gorillas$boundary) +
#'     gg(gorillas$nests)
#' }
gg <- function(data, ...) {
  UseMethod("gg")
}


#' @describeIn inlabru-deprecated
#' `r lifecycle::badge("deprecated")` This function is deprecated as `ggmap`
#' isn't supported.
#'
#' ggplot geom for spatial data
#'
#' gm is a wrapper for the [gg] method. It will take the first argument and
#' transform its coordinate system to latitude and longitude. Thereafter, [gg]
#' is called using the transformed data and the arguments provided via `...`. gm
#' is intended to replace gg whenever the data is supposed to be plotted over a
#' spatial map generated by [gmap], which only works if the coordinate system is
#' latitude/longitude.
gm <- function(...) {
  lifecycle::deprecate_warn(
    when = "2.11.1.9002",
    what = "gm()",
    with = "gg()",
    details = "ggmap isn't supported anymore."
  )
  gg(..., crs = fm_CRS("+proj=longlat"))
}


#' Geom for matrix
#'
#' Creates a tile geom for plotting a matrix
#'
#' Requires the `ggplot2` package.
#'
#' @name gg.matrix
#' @export
#' @param data A `matrix` object.
#' @param mapping a set of aesthetic mappings created by `aes`. These are passed
#'   on to `geom_tile`.
#' @param \dots Arguments passed on to `geom_tile`.
#' @return A `geom_tile` with reversed y scale.
#' @family geomes
#'
#' @importFrom rlang .data
#' @examples
#' if (require("ggplot2", quietly = TRUE)) {
#'   A <- matrix(runif(100), nrow = 10)
#'   ggplot() +
#'     gg(A)
#' }
gg.matrix <- function(data, mapping = NULL, ...) {
  requireNamespace("ggplot2")
  A <- as.matrix(data)
  grd <- expand.grid(row = seq_len(nrow(A)), column = seq_len(ncol(A)))
  df <- data.frame(value = as.vector(A), grd)

  dmap <- ggplot2::aes(x = .data$column, y = .data$row, fill = .data$value)

  if (!is.null(mapping)) {
    dmap <- modifyList(dmap, mapping)
  }

  ggp <- c(
    ggplot2::geom_tile(dmap, data = df, ...),
    ggplot2::scale_y_reverse()
  )
}


#' @describeIn gg.bru_prediction
#' This geom constructor will simply call [gg.bru_prediction()] for the data
#' provided.
#'
#' @export
#' @param \dots Arguments passed on to [gg.bru_prediction()].

gg.data.frame <- function(...) {
  gg.bru_prediction(...)
}


#' Geom for predictions
#'
#' @description
#'
#' This geom serves to visualize `prediction` objects which usually results from
#' a call to [predict.bru()]. Predictions objects provide summary statistics
#' (mean, median, sd, ...) for one or more random variables. For single
#' variables (or if requested so by setting `bar = TRUE`), a boxplot-style geom
#' is constructed to show the statistics. For multivariate predictions the mean
#' of each variable (y-axis) is plotted against the row number of the variable
#' in the prediction data frame (x-axis) using `geom_line`. In addition, a
#' `geom_ribbon` is used to show the confidence interval.
#'
#' Note: `gg.bru_prediction` also understands the format of INLA-style posterior
#' summaries, e.g. `fit$summary.fixed` for an inla object `fit`
#'
#' Requires the `ggplot2` package.
#'
#' @name gg.bru_prediction
#' @export
#' @importFrom utils modifyList
#' @param data A prediction object, usually the result of a [predict.bru()]
#'   call.
#' @param mapping a set of aesthetic mappings created by `aes`. These are passed
#'   on to `geom_line()`.
#'   If "fill" is present, it is passed to `geom_ribbon()`.
#' @param ribbon If TRUE, plot a ribbon around the line based on the smallest
#'   and largest quantiles present in the data, found by matching names starting
#'   with `q` and followed by a numerical value.  `inla()`-style
#'   `numeric+"quant"` names are converted to inlabru style before matching.
#' @param alpha The ribbons numeric alpha (transparency) level in `[0,1]`.
#' @param bar If TRUE plot boxplot-style summary for each variable.
#' @param \dots Arguments passed on to `geom_line`.
#' @return Concatenation of a `geom_line` value and optionally a `geom_ribbon`
#'   value.
#' @family geomes
#' @examples
#' \donttest{
#' if (bru_safe_inla() &&
#'   requireNamespace("sn", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE) &&
#'   require("patchwork", quietly = TRUE)) {
#'   # Generate some data
#'
#'   input.df <- data.frame(x = cos(1:10))
#'   input.df <- within(
#'     input.df,
#'     {
#'       y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
#'     }
#'   )
#'
#'   # Fit a model with fixed effect 'x' and intercept 'Intercept'
#'
#'   fit <- bru(y ~ x, family = "gaussian", data = input.df)
#'
#'   # Predict posterior statistics of 'x'
#'
#'   xpost <- predict(fit, NULL, formula = ~x_latent)
#'
#'   # The statistics include mean, standard deviation, the 2.5% quantile, the
#'   # median, the 97.5% quantile, minimum and maximum sample drawn from the
#'   # posterior as well as the coefficient of variation and the variance.
#'
#'   xpost
#'
#'   # For a single variable like 'x' the default plotting method invoked by
#'   # gg() will show these statistics in a fashion similar to a box plot:
#'   ggplot() +
#'     gg(xpost)
#'
#'
#'   # The predict function can also be used to simultaneously estimate
#'   # posteriors of multiple variables:
#'
#'   xipost <- predict(fit,
#'     newdata = NULL,
#'     formula = ~ c(
#'       Intercept = Intercept_latent,
#'       x = x_latent
#'     )
#'   )
#'   xipost
#'
#'   # If we still want a plot in the previous style we have to set the bar
#'   # parameter to TRUE
#'
#'   p1 <- ggplot() +
#'     gg(xipost, bar = TRUE)
#'   p1
#'
#'   # Note that gg also understands the posterior estimates generated while
#'   # running INLA
#'
#'   p2 <- ggplot() +
#'     gg(fit$summary.fixed, bar = TRUE)
#'   (p1 / p2)
#'
#'   # By default, if the prediction has more than one row, gg will plot the
#'   # column mean' against the row index. This is for instance useful for
#'   # predicting and plotting function but not very meaningful given the above
#'   # example:
#'
#'   ggplot() +
#'     gg(xipost)
#'
#'   # For ease of use we can also type
#'
#'   plot(xipost)
#'
#'   # This type of plot will show a ribbon around the mean, which visualizes
#'   # the upper and lower quantiles mentioned above (2.5% and 97.5%).
#'   # Plotting the ribbon can be turned of using the \code{ribbon} parameter
#'
#'   ggplot() +
#'     gg(xipost, ribbon = FALSE)
#'
#'   # Much like the other geomes produced by gg we can adjust the plot using
#'   # ggplot2 style commands, for instance
#'
#'   ggplot() +
#'     gg(xipost) +
#'     gg(xipost, mapping = aes(y = median), ribbon = FALSE, color = "red")
#' }
#' }
#'
gg.bru_prediction <- function(data,
                              mapping = NULL,
                              ribbon = TRUE,
                              alpha = NULL,
                              bar = FALSE,
                              ...) {
  if (inherits(data, c("Spatial", "SpatRaster", "RasterLayer", "sf"))) {
    return(NextMethod())
  }

  if (is.null(alpha)) {
    alpha <- 0.3
  }

  if (is.list(data) && !is.data.frame(data)) {
    data <- lapply(names(data), function(x) {
      if (NROW(data[[x]]) == 1) {
        rownames(data[[x]]) <- x
      } else {
        rownames(data[[x]]) <- glue("{x}.{seq_len(NROW(data[[x]]))}")
      }
      data[[x]]
    })
    data <- dplyr::bind_rows(data)
  }

  requireNamespace("ggplot2")
  # Convert from old and inla style names
  if (("median" %in% names(data)) &&
    !("0.5quant" %in% names(data)) &&
    !("q0.5" %in% names(data))) {
    names(data)[names(data) == "median"] <- "q0.5"
  }
  new_quant_names <- list(
    q0.025 = "0.025quant",
    q0.5 = "0.5quant",
    q0.975 = "0.975quant"
  )
  for (quant_name in names(new_quant_names[new_quant_names %in% names(data)])) {
    names(data)[names(data) == new_quant_names[[quant_name]]] <- quant_name
  }
  # Find quantile levels
  quant_names <- names(data)[grepl("^q[01]\\.?[0-9]*%?$", names(data))]
  if (length(quant_names) > 0) {
    perc <- grepl("%$", quant_names)
    quant_probs <- as.numeric(
      sub("%$", "", sub("^q", "", quant_names))
    )
    quant_probs[perc] <- quant_probs[perc] / 100
    quant_names <- quant_names[order(quant_probs)]
    quant_probs <- sort(quant_probs)
    lqname <- quant_names[1]
    uqname <- quant_names[length(quant_names)]
  }

  if (bar || (nrow(data) == 1)) {
    sz <- 10 # bar width
    med_sz <- 4 # median marker size

    data <- cbind(
      data.frame(data),
      data.frame(
        variable = rownames(data),
        summary = data$mean,
        sdmax = data$mean + data$sd,
        sdmin = data$mean - data$sd
      )
    )

    geom <- c(
      ggplot2::geom_point(
        data = data,
        mapping = ggplot2::aes(
          x = .data$variable,
          y = .data$summary,
          color = .data$variable
        ),
        shape = 95, size = 0
      )
    )
    if (length(quant_names) > 0) {
      geom <- c(
        geom,
        ggplot2::geom_segment(
          data = data,
          mapping = ggplot2::aes(
            y = .data[[lqname]],
            yend = .data[[uqname]],
            x = .data$variable,
            xend = .data$variable,
            color = .data$variable
          ),
          linewidth = sz
        )
      )
    }

    # Mean and median
    geom <- c(
      geom,
      ggplot2::geom_point(
        data = data,
        mapping = ggplot2::aes(
          x = .data$variable,
          y = .data$mean
        ),
        color = "black", shape = 20, size = med_sz
      )
    )
    if ("q0.5" %in% quant_names) {
      geom <- c(
        geom,
        ggplot2::geom_point(
          data = data,
          mapping = ggplot2::aes(
            x = .data$variable,
            y = .data$q0.5
          ),
          color = "black", shape = 3, size = sz * 2 / 3
        )
      )
    }
    geom <- c(
      geom,
      ggplot2::coord_flip()
    )
  } else {
    if ("pdf" %in% names(data)) {
      y.str <- "pdf"
      ribbon <- FALSE
    } else if ("mean" %in% names(data)) {
      y.str <- "mean"
    } else if ("q0.5" %in% names(data)) {
      y.str <- "q0.5"
    } else if ("median" %in% names(data)) {
      y.str <- "q0.5"
    } else {
      stop(paste0(
        "Prediction has neither mean nor median/q0.5 or pdf as column. ",
        "Don't know what to plot."
      ))
    }

    line.map <- ggplot2::aes(
      x = .data[[names(data)[1]]],
      y = .data[[y.str]]
    )

    if (!is.null(mapping)) {
      line.map <- utils::modifyList(line.map, mapping)
    }

    geom <- ggplot2::geom_line(data = data, line.map, ...)

    if (ribbon && length(quant_names) > 0) {
      ribbon.map <- ggplot2::aes(
        x = .data[[names(data)[1]]],
        ymin = .data[[lqname]],
        ymax = .data[[uqname]]
      )
      if (!is.null(mapping)) {
        ribbon.map <- utils::modifyList(ribbon.map, mapping)
      }
      geom <- c(
        geom,
        ggplot2::geom_ribbon(
          data = data,
          ribbon.map,
          alpha = alpha,
          colour = NA
        )
      )
    }
  }
  geom
}

#' @rdname gg.bru_prediction
#' @export
gg.prediction <- function(data, ...) {
  gg.bru_prediction(data = data, ...)
}

#' @title Geoms for sp Spatial objects
#'
#' @description
#' Methods for plotting sp spatial objects with `ggplot2`.
#'
#' @name gg.Spatial
#' @rdname gg.Spatial
NULL

#' @describeIn gg.Spatial Geom for `SpatialPoints` objects.
#' This function coerces the `SpatialPoints` into a `data.frame` and uses
#' `geom_point` to plot the points. Requires the `ggplot2` package.
#'
#' @export
#' @param data A `Spatial*` object.
#' @param mapping Aesthetic mappings created by `aes` used to update the default
#'   mapping. Uness specified otherwise below, the default mapping is
#'   ```{r eval = FALSE}
#'   ggplot2::aes(
#'     x = .data[[sp::coordnames(data)[1]]],
#'     y = .data[[sp::coordnames(data)[2]]]
#'   )
#'   ```
#' @param crs A `sp::CRS` object defining the coordinate system to project the
#'   data to before plotting.
#' @param \dots Arguments passed on to `geom_*`.
#' @return A `geom_point`, `geom_segment`, `geom_sf`, `geom_tile`,
#' or a list of `ggplot` geomes
#' @family geomes
#'
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   bru_safe_sp() &&
#'   require("sp")) {
#'   # Load Gorilla data
#'
#'   gorillas <- inlabru::gorillas_sf
#'
#'   gcov <- gorillas_sf_gcov()
#'   elev <- terra::as.data.frame(gcov$elevation, xy = TRUE)
#'   elev <- sf::as_Spatial(sf::st_as_sf(elev, coords = c("x", "y")))
#'
#'   # Turn elevation covariate into SpatialGridDataFrame
#'   elev <- sp::SpatialPixelsDataFrame(elev, data = as.data.frame(elev))
#'
#'   # Plot Gorilla elevation covariate provided as SpatialPixelsDataFrame.
#'   # The same syntax applies to SpatialGridDataFrame objects.
#'
#'   ggplot() +
#'     gg(elev)
#'
#'   # Add Gorilla survey boundary and nest sightings
#'
#'   ggplot() +
#'     gg(elev) +
#'     gg(gorillas$boundary, alpha = 0.0, col = "red") +
#'     gg(gorillas$nests)
#'
#'   # Load pantropical dolphin data
#'
#'   mexdolphin <- inlabru::mexdolphin_sp()
#'
#'   # Plot the pantropical survey boundary, ship transects, and dolphin
#'   # sightings
#'
#'   ggplot() +
#'     gg(mexdolphin$ppoly) + # survey boundary as SpatialPolygon
#'     gg(mexdolphin$samplers) + # ship transects as SpatialLines
#'     gg(mexdolphin$points) # dolphin sightings as SpatialPoints
#'
#'   # Change color
#'
#'   ggplot() +
#'     gg(mexdolphin$ppoly, color = "green") + # survey boundary; SpatialPolygon
#'     gg(mexdolphin$samplers, color = "red") + # ship transects; SpatialLines
#'     gg(mexdolphin$points, color = "blue") # dolphin sightings; SpatialPoints
#'
#'
#'   # Visualize data annotations: line width by segment number
#'
#'   names(mexdolphin$samplers) # 'seg' holds the segment number
#'   ggplot() +
#'     gg(mexdolphin$samplers, aes(color = seg))
#'
#'   # Visualize data annotations: point size by dolphin group size
#'
#'   names(mexdolphin$points) # 'size' holds the group size
#'   ggplot() +
#'     gg(mexdolphin$points, aes(size = size))
#' }
#' }
#'
gg.SpatialPoints <- function(data, mapping = NULL, crs = NULL, ...) {
  bru_safe_sp(force = TRUE)
  requireNamespace("ggplot2")
  if (!is.null(crs)) {
    data <- fm_transform(data, crs)
  }

  df <- as.data.frame(data)
  cnames <- sp::coordnames(data)
  if (is.null(cnames)) {
    cnames <- c("x", "y")
  }

  dmap <- ggplot2::aes(
    x = .data[[cnames[1]]],
    y = .data[[cnames[2]]]
  )

  if (!is.null(mapping)) {
    dmap <- modifyList(dmap, mapping)
  }

  ggplot2::geom_point(data = df, mapping = dmap, ...)
}

#' Geom helper for sf objects
#'
#' This function uses `geom_sf()`, unless overridden by the geom argument.
#' Requires the `ggplot2` package.
#'
#' @name gg.sf
#' @export
#' @param data An `sf` object.
#' @param mapping Default mapping is `ggplot2::aes(geometry = ...)`,
#' where the geometry name is obtained from `attr(data, "sf_column")`.
#' This is merged with the user supplied mapping.
#' @param geom Either "sf" (default) or "tile". For "tile", uses
#' `geom_tile(..., stat = "sf_coordinates")`, intended for converting point data
#' to grid tiles with the `fill` aesthetic, which is by default set to the first
#' data column.
#' @param \dots Arguments passed on to `geom_sf` or `geom_tile`.
#' @return A ggplot return value
#' @family geomes
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("tidyterra", quietly = TRUE)) {
#'   # Load Gorilla data
#'
#'   gorillas <- inlabru::gorillas_sf
#'   gorillas$gcov <- gorillas_sf_gcov()
#'
#'   # Plot Gorilla elevation covariate provided as terra::rast.
#'
#'   ggplot() +
#'     gg(gorillas$gcov$elevation)
#'
#'   # Add Gorilla survey boundary and nest sightings
#'
#'   ggplot() +
#'     gg(gorillas$gcov$elevation) +
#'     gg(gorillas$boundary, alpha = 0) +
#'     gg(gorillas$nests)
#'
#'   # Load pantropical dolphin data
#'
#'   mexdolphin <- inlabru::mexdolphin_sf
#'
#'   # Plot the pantropical survey boundary, ship transects and dolphin
#'   # sightings
#'
#'   ggplot() +
#'     gg(mexdolphin$ppoly, alpha = 0.5) + # survey boundary
#'     gg(mexdolphin$samplers) + # ship transects
#'     gg(mexdolphin$points) # dolphin sightings
#'
#'   # Change color
#'
#'   ggplot() +
#'     gg(mexdolphin$ppoly, color = "green", alpha = 0.5) + # survey boundary
#'     gg(mexdolphin$samplers, color = "red") + # ship transects
#'     gg(mexdolphin$points, color = "blue") # dolphin sightings
#'
#'
#'   # Visualize data annotations: line width by segment number
#'
#'   names(mexdolphin$samplers) # 'seg' holds the segment number
#'   ggplot() +
#'     gg(mexdolphin$samplers, aes(color = seg))
#'
#'   # Visualize data annotations: point size by dolphin group size
#'
#'   names(mexdolphin$points) # 'size' holds the group size
#'   ggplot() +
#'     gg(mexdolphin$points, aes(size = size))
#' }
#' }
#'
gg.sf <- function(data, mapping = NULL, ..., geom = "sf") {
  requireNamespace("ggplot2")

  sf_column <- attr(data, "sf_column")
  geom <- match.arg(geom, c("sf", "tile"))

  if (identical(geom, "sf")) {
    dmap <- ggplot2::aes(
      geometry = .data[[sf_column]]
    )
    if (!is.null(mapping)) {
      dmap <- modifyList(dmap, mapping)
    }
    result <- ggplot2::geom_sf(data = data, mapping = dmap, ...)
  } else {
    data_column <- setdiff(names(data), sf_column)
    if (length(data_column) > 0) {
      data_column <- data_column[1]
      dmap <- ggplot2::aes(
        geometry = .data[[sf_column]],
        fill = .data[[data_column]]
      )
    } else {
      dmap <- ggplot2::aes(
        geometry = .data[[sf_column]]
      )
    }
    if (!is.null(mapping)) {
      dmap <- modifyList(dmap, mapping)
    }
    result <-
      c(
        ggplot2::geom_tile(
          data = data,
          mapping = dmap,
          stat = "sf_coordinates",
          ...
        ),
        if (fmesher::fm_crs_is_null(fm_crs(data))) {
          ggplot2::coord_sf(default = TRUE)
        } else {
          ggplot2::coord_sf(default = TRUE, crs = fm_crs(data))
        }
      )
  }

  result
}

#' @describeIn gg.Spatial Geom for SpatialLines objects.
#'
#' Extracts start and end points of the lines and calls `geom_segment` to plot
#' lines between them.
#'
#' `mapping`: Aesthetic mappings created by `ggplot2::aes` or
#'   `ggplot2::aes_` used to update the default mapping. The default mapping is
#'   ```{r eval = FALSE}
#'   ggplot2::aes(
#'     x = .data[[sp::coordnames(data)[1]]],
#'     y = .data[[sp::coordnames(data)[2]]],
#'     xend = .data[[paste0("end.", sp::coordnames(data)[1])]],
#'     yend = .data[[paste0("end.", sp::coordnames(data)[2])]])
#'   ```
#' @export

gg.SpatialLines <- function(data, mapping = NULL, crs = NULL, ...) {
  bru_safe_sp(force = TRUE)
  requireNamespace("ggplot2")
  if (!is.null(crs)) {
    data <- fm_transform(data, crs)
  }

  qq <- sp::coordinates(data)
  cnames <- sp::coordnames(data)
  if (is.null(cnames)) {
    cnames <- c("x", "y")
  }
  sp <- do.call(rbind, lapply(
    qq,
    function(k) do.call(rbind, lapply(k, function(x) x[1:(nrow(x) - 1), ]))
  ))
  ep <- do.call(rbind, lapply(
    qq,
    function(k) do.call(rbind, lapply(k, function(x) x[2:(nrow(x)), ]))
  ))
  colnames(sp) <- cnames
  colnames(ep) <- glue("end.{cnames}")
  if (!inherits(data, "SpatialLinesDataFrame")) {
    df <- data.frame(cbind(sp, ep))
  } else {
    df <- data.frame(
      cbind(sp, ep),
      data@data[
        rep(
          seq_len(NROW(data)),
          times = vapply(
            qq,
            FUN = function(x) {
              NROW(x[[1]]) - 1
            },
            0
          )
        ), ,
        drop = FALSE
      ]
    )
  }
  dmap <- ggplot2::aes(
    x = .data[[cnames[1]]],
    y = .data[[cnames[2]]],
    xend = .data[[glue("end.{cnames[1]}")]],
    yend = .data[[glue("end.{cnames[2]}")]]
  )

  if (!is.null(mapping)) {
    dmap <- modifyList(dmap, mapping)
  }

  ggplot2::geom_segment(data = df, mapping = dmap, ...)
}

#' @describeIn gg.Spatial Geom for SpatialPolygons objects.
#' Uses the `ggplot2::fortify()` function to turn the `SpatialPolygons` objects
#' into a `data.frame`. Then calls `geom_polygon` to plot the polygons.
#'
#' Unless specified by the user,
#' the argument `alpha = 0.2` (alpha level for polygon filling) is added.
#'
#' Up to version `2.10.0`, the `ggpolypath` package was used to ensure proper
#' plotting for polygons, since the `ggplot2::geom_polygon` function doesn't
#' always handle geometries with holes properly. After `2.10.0`, the object is
#' converted to `sf` format and passed on to [gg.sf()] instead, as `ggplot2`
#' version `3.4.4` deprecated the internally used `ggplot2::fortify()` method
#' for `SpatialPolygons/DataFrame` objects.
#' @export

gg.SpatialPolygons <- function(data, mapping = NULL, crs = NULL, ...) {
  bru_safe_sp(force = TRUE)
  data <- sf::st_as_sf(data)
  if (!is.null(crs)) {
    data <- fm_transform(data, crs, passthrough = TRUE)
  }
  arg <- list(...)
  if ("alpha" %in% names(arg)) {
    gg(data, mapping = mapping, ...)
  } else {
    gg(data, mapping = mapping, ..., alpha = 0.2)
  }
}

#' @describeIn gg.Spatial Geom for SpatialGridDataFrame objects
#'
#' Coerces input `SpatialGridDataFrame` to `SpatialPixelsDataFrame` and calls
#' [gg.SpatialPixelsDataFrame()] to plot it.
#'
#' @export

gg.SpatialGridDataFrame <- function(data, ...) {
  bru_safe_sp(force = TRUE)
  data <- as(data, "SpatialPixelsDataFrame")
  gg(data, ...)
}

#' @describeIn gg.Spatial Geom for SpatialPixelsDataFrame objects.
#'
#' Coerces `SpatialPixelsDataFrame` input to `data.frame` and uses
#' `geom_tile` to plot it.
#'
#' `mapping`: Aesthetic mappings created by `aes` used to update the default
#'   mapping. The default mapping is
#'   ```{r eval=FALSE}
#'   ggplot2::aes(
#'     x = .data[[sp::coordnames(data)[1]]],
#'     y = .data[[sp::coordnames(data)[2]]],
#'     fill = .data[[names(data)[[1]]]]
#'   )
#'   ```
#'
#' @export
#' @param mask A `sp::SpatialPolygons` object defining the region that is
#'   plotted.
gg.SpatialPixelsDataFrame <- function(data,
                                      mapping = NULL,
                                      crs = NULL,
                                      mask = NULL, ...) {
  bru_safe_sp(force = TRUE)
  if (!is.null(crs)) {
    data <- fm_transform(data, crs)
  }
  if (!is.null(mask)) {
    data <- data[as.vector(!is.na(sp::over(data, mask))), ]
  }

  df <- as.data.frame(data)

  if (length(names(data)) == 0) {
    dmap <- ggplot2::aes(
      x = .data[[sp::coordnames(data)[1]]],
      y = .data[[sp::coordnames(data)[2]]]
    )
  } else {
    dmap <- ggplot2::aes(
      x = .data[[sp::coordnames(data)[1]]],
      y = .data[[sp::coordnames(data)[2]]],
      fill = .data[[names(data)[[1]]]]
    )
  }

  if (!is.null(mapping)) {
    dmap <- modifyList(dmap, mapping)
  }
  gm <- ggplot2::geom_tile(data = df, mapping = dmap, ...)

  gm
}


#' @describeIn gg.Spatial Geom for SpatialPixels objects
#'
#' Converts the input to `SpatialPoints` and calls
#' [gg.SpatialPoints()` to plot it.
#'
#' @export
#' @examples
#' if (require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   bru_safe_sp()) {
#'   # Load Gorilla data
#'
#'   gcov <- gorillas_sf_gcov()
#'   elev <- terra::as.data.frame(gcov$elevation, xy = TRUE)
#'   pxl <- sf::as_Spatial(sf::st_as_sf(elev, coords = c("x", "y")))
#'
#'   # Turn elevation covariate into SpatialPixels
#'   pxl <- sp::SpatialPixels(pxl)
#'
#'   # Plot the pixel centers
#'   ggplot() +
#'     gg(pxl, size = 0.1)
#' }
#'
gg.SpatialPixels <- function(data, ...) {
  bru_safe_sp(force = TRUE)
  gg(sp::SpatialPoints(data), ...)
}


#' Geom wrapper for SpatRaster objects
#'
#' Convenience wrapper function for `tidyterra::geom_spatraster()`.
#' Requires the `ggplot2` and `tidyterra` packages.
#'
#' @export
#' @param data A SpatRaster object.
#' @param \dots Arguments passed on to `geom_spatraster`.
#' @return The output from `geom_spatraster.
#' @family geomes
#' @examples
#' if (require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("tidyterra", quietly = TRUE)) {
#'   # Load Gorilla covariates
#'
#'   gcov <- gorillas_sf_gcov()
#'
#'   # Plot the pixel centers
#'   ggplot() +
#'     gg(gcov$elevation)
#' }
gg.SpatRaster <- function(data, ...) {
  if (!requireNamespace("tidyterra", quietly = TRUE)) {
    stop(paste0(
      "gg.SpatRaster requires the 'tidyterra' package, ",
      "which is not installed.\n",
      "Please install it and try again!"
    ))
  }
  if (".mask" %in% names(data)) {
    requireNamespace("terra")
    data <-
      terra::mask(
        data,
        mask = data$.mask,
        maskvalues = FALSE,
        updatevalue = NA
      )
  }
  tidyterra::geom_spatraster(data = data, ...)
}


#' Geom for fm_mesh_2d objects
#'
#' @description
#' This function extracts the graph of an [fmesher::fm_mesh_2d] object and uses
#' `geom_line` to visualize the graph's edges. Alternatively, if the `color`
#' argument is provided, interpolates the colors across for a set of
#' SpatialPixels covering the mesh area and calls [gg.SpatialPixelsDataFrame()]
#' to plot the interpolation.
#' Requires the `ggplot2` package.
#'
#' Also see the [fmesher::geom_fm()] method.
#'
#' @export
#'
#' @param data An `fm_mesh_2d` object.
#' @param color A vector of scalar values to fill the mesh with colors.
#' The length of the vector mus correspond to the number of mesh vertices.
#' The alternative name `colour` is also recognised.
#' @param alpha A vector of scalar values setting the alpha value of the colors
#'   provided.
#' @param edge.color Color of the regular mesh edges.
#' @param edge.linewidth Line width for the regular mesh edges. Default 0.25
#' @param interior If TRUE, plot the interior boundaries of the mesh.
#' @param int.color Color used to plot the interior constraint edges.
#' @param int.linewidth Line width for the interior constraint edges. Default
#'   0.5
#' @param exterior If TRUE, plot the exterior boundaries of the mesh.
#' @param ext.color Color used to plot the exterior boundary edges.
#' @param ext.linewidth Line width for the exterior boundary edges. Default 1
#' @param crs A CRS object supported by [fmesher::fm_transform()] defining the
#'   coordinate system to project the mesh to before plotting.
#' @param nx Number of pixels in x direction (when plotting using the color
#'   parameter).
#' @param ny Number of pixels in y direction (when plotting using the color
#'   parameter).
#' @param mask A `SpatialPolygon` or `sf` polygon defining the region that is
#'   plotted.
#' @param \dots ignored arguments (S3 generic compatibility).
#' @return `geom_line` return values or, if the color argument is used, the
#'   values of [gg.SpatialPixelsDataFrame()].
#' @family geomes
#' @examples
#' \donttest{
#' if (require(fmesher, quietly = TRUE) &&
#'   require(ggplot2, quietly = TRUE)) {
#'   # Load Gorilla data
#'   gorillas <- inlabru::gorillas_sf
#'
#'   # Plot mesh using default edge colors
#'
#'   ggplot() +
#'     gg(gorillas$mesh)
#'
#'   # Don't show interior and exterior boundaries
#'
#'   ggplot() +
#'     gg(gorillas$mesh, interior = FALSE, exterior = FALSE)
#'
#'   # Change the edge colors
#'
#'   ggplot() +
#'     gg(gorillas$mesh,
#'       edge.color = "green",
#'       int.color = "black",
#'       ext.color = "blue"
#'     )
#'
#'   # Use the x-coordinate of the vertices to colorize the triangles and
#'   # mask the plotted area by the survey boundary, i.e. only plot the inside
#'
#'   xcoord <- gorillas$mesh$loc[, 1]
#'   ggplot() +
#'     gg(gorillas$mesh, color = (xcoord - 580), mask = gorillas$boundary) +
#'     gg(gorillas$boundary, alpha = 0)
#' }
#' }
#'
gg.fm_mesh_2d <- function(data,
                          color = NULL,
                          alpha = NULL,
                          edge.color = "grey",
                          edge.linewidth = 0.25,
                          interior = TRUE,
                          int.color = "blue",
                          int.linewidth = 0.5,
                          exterior = TRUE,
                          ext.color = "black",
                          ext.linewidth = 1,
                          crs = NULL,
                          mask = NULL,
                          nx = 500, ny = 500,
                          ...) {
  requireNamespace("ggplot2")
  if (is.null(color) && ("colour" %in% names(list(...)))) {
    color <- list(...)[["colour"]]
  }
  if (!is.null(color)) {
    if (inherits(mask, "Spatial")) {
      # For backwards compatibility with old plotting code, use Spatial
      # format, and mask only in the gg.Spatial* method.
      px <- fm_pixels(data, dims = c(nx, ny), format = "sp")
      A <- fm_basis(data, px)
      px$color <- as.vector(A %*% color)
      if (!is.null(alpha)) {
        px$alpha <- as.vector(A %*% alpha)
        gg <- gg(px, mask = mask, alpha = "alpha")
      } else {
        gg <- gg(px, mask = mask)
      }
    } else {
      px <- fm_pixels(data, dims = c(nx, ny), mask = mask, format = "sf")
      proj <- fm_evaluator(data, px)
      px$color <- fm_evaluate(proj, field = color)
      if (!is.null(alpha)) {
        if (length(alpha) == 1) {
          px$alpha <- alpha
        } else {
          px$alpha <- fm_evaluate(proj, field = alpha)
        }
        gg <- gg(px,
          ggplot2::aes(fill = .data[["color"]]),
          alpha = px[["alpha"]],
          geom = "tile"
        )
      } else {
        gg <- gg(px,
          ggplot2::aes(fill = .data[["color"]]),
          geom = "tile"
        )
      }
    }

    return(gg)
  }

  if (data$manifold == "S2") {
    stop("Geom not implemented for spherical meshes (manifold = S2)")
  }
  if (!is.null(crs)) {
    data <- fm_transform(data, crs = crs)
  }

  df <- rbind(
    data.frame(
      a = data$loc[data$graph$tv[, 1], c(1, 2)],
      b = data$loc[data$graph$tv[, 2], c(1, 2)]
    ),
    data.frame(
      a = data$loc[data$graph$tv[, 2], c(1, 2)],
      b = data$loc[data$graph$tv[, 3], c(1, 2)]
    ),
    data.frame(
      a = data$loc[data$graph$tv[, 1], c(1, 2)],
      b = data$loc[data$graph$tv[, 3], c(1, 2)]
    )
  )

  colnames(df) <- c("x", "y", "xend", "yend")
  mp <- ggplot2::aes(
    x = .data$x,
    y = .data$y,
    xend = .data$xend,
    yend = .data$yend
  )
  msh <- ggplot2::geom_segment(
    data = df,
    mapping = mp,
    color = edge.color,
    linewidth = edge.linewidth
  )

  # Outer boundary
  if (exterior) {
    df <- data.frame(
      data$loc[data$segm$bnd$idx[, 1], 1:2],
      data$loc[data$segm$bnd$idx[, 2], 1:2]
    )
    colnames(df) <- c("x", "y", "xend", "yend")
    bnd <- ggplot2::geom_segment(
      data = df,
      mapping = mp,
      color = ext.color,
      linewidth = ext.linewidth
    )
  } else {
    bnd <- NULL
  }

  if (interior) {
    # Interior boundary
    df <- data.frame(
      data$loc[data$segm$int$idx[, 1], 1:2],
      data$loc[data$segm$int$idx[, 2], 1:2]
    )
    colnames(df) <- c("x", "y", "xend", "yend")
    if (nrow(df) == 0) {
      int <- NULL
    } else {
      int <- ggplot2::geom_segment(
        data = df,
        mapping = mp,
        color = int.color,
        linewidth = int.linewidth
      )
    }
  } else {
    int <- NULL
  }

  # Return combined geomes
  c(msh, bnd, int)
}


#' Geom for fm_mesh_1d objects
#'
#' This function generates a `geom_point` object showing the knots (vertices)
#' of a 1D mesh.
#' Requires the `ggplot2` package.
#'
#' @export
#' @param data An [fmesher::fm_mesh_1d] object.
#' @param mapping aesthetic mappings created by `aes`. These are passed on to
#'   `geom_point`.
#' @param y Single or vector numeric defining the y-coordinates of the mesh
#'   knots to plot.
#' @param shape Shape of the knot markers.
#' @param \dots parameters passed on to `geom_point`.
#' @return An object generated by `geom_point`.
#' @family geomes
#'
#' @examples
#' \donttest{
#' if (require("fmesher", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE)) {
#'   # Create a 1D mesh
#'
#'   mesh <- fmesher::fm_mesh_1d(seq(0, 10, by = 0.5))
#'
#'   # Plot it
#'
#'   ggplot() +
#'     gg(mesh)
#'
#'   # Plot it using a different shape and size for the mesh nodes
#'
#'   ggplot() +
#'     gg(mesh, shape = "|", size = 5)
#' }
#' }
#'
gg.fm_mesh_1d <- function(data,
                          mapping = ggplot2::aes(.data[["x"]], .data[["y"]]),
                          y = 0, shape = 4, ...) {
  df <- data.frame(x = data$loc, y = y)
  ggplot2::geom_point(data = df, mapping = mapping, shape = shape, ...)
}


#' Geom for RasterLayer objects
#'
#' This function takes a RasterLayer object, converts it into a
#' `SpatialPixelsDataFrame` and uses `geom_tile` to plot the data.
#'
#' This function requires the `raster` and `ggplot2` packages.
#'
#' @export
#' @param data A RasterLayer object.
#' @param mapping aesthetic mappings created by `aes`. These are passed on to
#'   `geom_tile`.
#' @param \dots Arguments passed on to `geom_tile`.
#' @return An object returned by `geom_tile`
#' @family geomes
#'
#' @examples
#' \dontrun{
#' # Some features require the raster and spatstat.data packages.
#' if (require("spatstat.data", quietly = TRUE) &&
#'   require("raster", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE)) {
#'   # Load Gorilla data
#'   data("gorillas", package = "spatstat.data", envir = environment())
#'
#'   # Convert elevation covariate to RasterLayer
#'
#'   elev <- as(gorillas.extra$elevation, "RasterLayer")
#'
#'   # Plot the elevation
#'
#'   ggplot() +
#'     gg(elev)
#' }
#' }
gg.RasterLayer <- function(data,
                           mapping = ggplot2::aes(
                             x = .data[["x"]],
                             y = .data[["y"]],
                             fill = .data[["layer"]]
                           ),
                           ...) {
  requireNamespace("ggplot2")
  requireNamespace("raster")
  spdf <- as(data, "SpatialPixelsDataFrame")
  df <- as.data.frame(spdf)
  # head(r.df)
  # g <- ggplot(r.df, ggplot2::aes(x=x, y=y)) +
  #   geom_tile(ggplot2::aes(fill = layer)) +
  #   coord_equal()
  ggplot2::geom_tile(data = df, mapping = mapping, ...)
}

#' Plot method for posterior marginals estimated by bru
#'
#' @description
#'
#' From version `2.11.0`, `plot.bru(x, ...)` calls `plot.inla(x, ...)`
#' from the `INLA` package, unless the first argument after `x` is a
#' `character`, in which case the pre-`2.11.0` behaviour is used, calling
#' `plotmarginal.inla(x, ...)` instead.
#'
#' Requires the `ggplot2` package.
#'
#' @method plot bru
#' @export
#' @param x a fitted [bru()] model.
#' @param \dots Options passed on to other methods.
#'
#' @examples
#' \dontrun{
#' if (require("ggplot2", quietly = TRUE)) {
#'   # Generate some data and fit a simple model
#'   input.df <- data.frame(x = cos(1:10)) |>
#'     dplyr::mutate(
#'       y = 5 + 2 * x + rnorm(length(x), mean = 0, sd = 0.1)
#'     )
#'   fit <- bru(y ~ x, family = "gaussian", data = input.df)
#'   summary(fit)
#'
#'   # Plot the posterior density of the model's x-effect
#'   plot(fit, "x")
#' }
#' }
#'
plot.bru <- function(x, ...) {
  ll <- list(...)
  if (length(ll) == 0) {
    NextMethod("plot")
  } else if (is.character(ll[[1]])) {
    plotmarginal.inla(x, ...)
  } else {
    NextMethod("plot")
  }
}

#' @describeIn gg.bru_prediction
#' Generates a base ggplot2 using `ggplot()` and adds a geom for input `x` using
#' [gg.bru_prediction()]. Returns a ggplot object.
#'
#' @param x a prediction object.
#' @param y Ignored argument but required for S3 compatibility.
#' @export
#' @method plot bru_prediction

plot.bru_prediction <- function(x, y = NULL, ...) {
  requireNamespace("ggplot2")
  ggplot2::ggplot() +
    gg(x, ...)
}

#' @describeIn gg.bru_prediction Identical to [gg.bru_prediction()].
#' @export
#' @method plot prediction
plot.prediction <- function(x, y = NULL, ...) {
  requireNamespace("ggplot2")
  ggplot2::ggplot() +
    gg(x, ...)
}


#' @title Multiple ggplots on a page.
#'
#' @description
#' `r lifecycle::badge("deprecated")` in favour of the `patchwork` package;
#' see the example below.
#'
#' Renders multiple ggplots on a single page.
#'
#' @param \dots Comma-separated `ggplot` objects.
#' @param plotlist A list of `ggplot` objects - an alternative to the
#'   comma-separated argument above.
#' @param cols Number of columns of plots on the page.
#' @param layout A matrix specifying the layout. If present, `cols` is ignored.
#'   If the layout is something like `matrix(c(1,2,3,3), nrow=2, byrow=TRUE)`,
#'   then plot 1 will go in the upper left, 2 will go in the upper right, and 3
#'   will go all the way across the bottom.
#'
#' @author David L. Borchers \email{dlb@@st-andrews.ac.uk}
#'
#' @source
#' <http://www.cookbook-r.com/Graphs/Multiple_graphs_on_one_page_(ggplot2)/>
#'
#' @examples
#' if (require("ggplot2", quietly = TRUE)) {
#'   df <- data.frame(x = 1:10, y = cos(1:10), z = sin(1:10))
#'   pl1 <- ggplot(data = df) +
#'     geom_line(mapping = aes(x, y), color = "red")
#'   pl2 <- ggplot(data = df) +
#'     geom_line(mapping = aes(x, z), color = "blue")
#'   pl3 <- ggplot(data = df) +
#'     geom_path(mapping = aes(y, z), color = "magenta")
#'
#'   multiplot(
#'     pl1, pl2, pl3,
#'     layout = rbind(c(1, 2), c(3, 3))
#'   )
#'
#'   # Recommended alternative using the patchwork package:
#'   if (require("patchwork")) {
#'     (pl1 | pl2) / pl3
#'   }
#' }
#' @export
#
multiplot <- function(..., plotlist = NULL, cols = 1, layout = NULL) {
  lifecycle::deprecate_soft(
    "2.13.0.9026",
    "multiplot()",
    details = "Please use the 'patchwork' package for combining ggplots."
  )

  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)

  numPlots <- length(plots)

  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout)) {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots / cols)),
      ncol = cols, nrow = ceiling(numPlots / cols)
    )
  }

  if (numPlots == 1) {
    print(plots[[1]])
  } else {
    # Set up the page
    grid::grid.newpage()
    grid::pushViewport(grid::viewport(
      layout = grid::grid.layout(nrow(layout), ncol(layout))
    ))

    # Make each plot, in the correct location
    for (i in 1:numPlots) {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))

      print(plots[[i]], vp = grid::viewport(
        layout.pos.row = matchidx$row,
        layout.pos.col = matchidx$col
      ))
    }
  }
}


# Default color palette for plots using \link{gg}
#
# @export
# @param
# @return Color values as returned by
#   `RColorBrewer::brewer.pal(9, "YlOrRd"`

bru.pal <- function() {
  # library(RColorBrewer)
  # brewer.pal(9, "YlOrRd")
  pal <- c(
    "#FFFFCC",
    "#FFEDA0",
    "#FED976",
    "#FEB24C",
    "#FD8D3C",
    "#FC4E2A",
    "#E31A1C",
    "#BD0026",
    "#800026"
  )
}

#' @importFrom rlang .data
#' @importFrom rlang .env

#' @title Plot inlabru convergence diagnostics
#'
#' @description
#' Draws four panels of convergence diagnostics for an iterated INLA method
#' estimation
#'
#' @param x a [bru] object, typically a result from [bru()] for a nonlinear
#' predictor model
#' @param from,to integer values for the range of iterations to plot.
#' Default `from = 1` (start from the first iteration) and `to = NULL` (end at
#' the last iteration).
#' Set `from = 0` to include the initial linearisation point in the track plot.
#' @param type `r lifecycle::badge("experimental")` character; "bru" (default)
#'   for iterative nonlinear inlabru convergence diagnostics plots, or "inla"
#'   for INLA optimiser trace plots.
#' @return A ggplot object with four panels of convergence diagnostics:
#' - `Tracks`: Mode and linearisation values for each effect
#' - `Mode - Lin`: Difference between mode and linearisation values for each
#'   effect
#' - `|Change| / sd`: Absolute change in mode and linearisation values
#' divided by the standard deviation for each effect
#' - `Change & sd`: Absolute change in mode and linearisation values
#' and standard deviation for each effect
#'
#' For multidimensional components, only the overall average, maximum, and
#' minimum values are shown.
#' @seealso [bru()]
#'
#' @details Requires the "dplyr", "ggplot2", and "patchwork"
#' packages to be installed.
#' @export
#' @examples
#' \dontrun{
#' fit <- bru(...)
#' bru_convergence_plot(fit)
#' }
bru_convergence_plot <- function(x, from = 1, to = NULL, type = NULL) {
  stopifnot(inherits(x, "bru"))
  x <- bru_check_object_bru(x)

  type <- match.arg(type, c("bru", "inla"))
  if (identical(type, "bru")) {
    make_bru_track_plots(x, from = from, to = to)[["default"]]
  } else if (identical(type, "inla")) {
    make_inla_track_plots(x, from = from, to = to)[["default"]]
  }
}


make_bru_track_plots <- function(fit, from = 1, to = NULL) {
  needed <- c("dplyr", "ggplot2", "patchwork")
  are_installed <-
    vapply(
      needed,
      function(x) {
        requireNamespace(x, quietly = TRUE)
      },
      TRUE
    )
  if (!all(are_installed)) {
    stop(
      paste0(
        "Needed package(s) ",
        paste0("'", needed[!are_installed], "'", collapse = ", "),
        " not installed, but are needed by make_track_plots()"
      )
    )
  }

  track_data <- fit$bru_iinla$track
  if (!is.null(from)) {
    stopifnot(is.numeric(from))
    stopifnot(from >= 0)
  } else {
    from <- min(track_data$iteration)
  }
  if (!is.null(to)) {
    stopifnot(is.numeric(to))
    stopifnot(to >= 0)
  } else {
    to <- max(track_data$iteration)
  }

  track_data <- track_data |>
    dplyr::group_by(.data$effect, .data$iteration) |>
    dplyr::mutate(alpha_value = 1 / sqrt(max(.data$index))) |>
    dplyr::ungroup()

  alpha_value_lin_scale <- 0.8

  lty_ <- factor(c("Mode", "Lin", "Mode-Lin", "SD"),
    levels = c("Mode", "Lin", "Mode-Lin", "SD")
  )
  names(lty_) <- levels(lty_)
  col_ <- factor(c("Max", "Mean", "Min"),
    levels = c("Max", "Mean", "Min")
  )
  names(col_) <- levels(col_)

  sc <- ggplot2::scale_linetype_discrete(
    name = "Quantity",
    breaks = names(lty_),
    labels = labels(lty_),
    drop = FALSE,
    na.translate = FALSE
  )
  # Colour blind friendliness, see https://davidmathlogic.com/colorblind/
  sc_minmax <- list(
    ggplot2::scale_color_discrete(
      name = "Aspect",
      type = c("#DC3220", "#000000", "#005AB5"),
      breaks = names(col_),
      labels = labels(col_),
      drop = FALSE,
      na.translate = FALSE
    ),
    ggplot2::scale_fill_discrete(
      name = "Aspect",
      type = c("#DC3220", "#000000", "#005AB5"),
      breaks = names(col_),
      labels = labels(col_),
      drop = FALSE,
      na.translate = FALSE
    )
  )

  pl_theme_abs <-
    list(
      ggplot2::facet_wrap(ggplot2::vars(.data$effect), scales = "free_y"),
      sc,
      sc_minmax #
      #      ggplot2::guides(color = "none")
    )
  pl_theme_rel <-
    list(
      ggplot2::facet_wrap(ggplot2::vars(.data$effect), scales = "free_y"),
      sc,
      sc_minmax # ,
      #      ggplot2::guides(color = "none")
    )
  pl_theme_norm <-
    list(
      ggplot2::facet_wrap(ggplot2::vars(.data$effect), scales = "free_y"),
      sc,
      sc_minmax # ,
      #      ggplot2::guides(color = "none")
    )

  pl_tracks <-
    ggplot2::ggplot(
      track_data |>
        dplyr::filter(.data$iteration >= from, .data$iteration <= to) |>
        dplyr::group_by(
          .data$effect,
          .data$iteration
        ) |>
        dplyr::summarise(
          MaxMode = max(.data$mode),
          MeanMode = mean(.data$mode),
          MinMode = min(.data$mode),
          MaxLin = max(.data$new_linearisation),
          MeanLin = mean(.data$new_linearisation),
          MinLin = min(.data$new_linearisation),
          .groups = "drop"
        )
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxMode,
        lty = lty_["Mode"],
        color = col_["Max"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinMode,
        lty = lty_["Mode"],
        color = col_["Min"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanMode,
        lty = lty_["Mode"],
        color = col_["Mean"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxLin,
        lty = lty_["Lin"],
        col = "Max"
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinLin,
        lty = lty_["Lin"],
        col = "Min"
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanLin,
        lty = lty_["Lin"],
        col = "Mean"
      ),
      na.rm = TRUE
    ) +
    ggplot2::ylab("Value") +
    pl_theme_abs +
    ggplot2::ggtitle("Tracks")

  pl_mode_lin <-
    ggplot2::ggplot(
      track_data |>
        dplyr::filter(
          .data$iteration >= max(1, from),
          .data$iteration <= to
        ) |>
        dplyr::filter(
          is.finite(.data$sd),
          is.finite(.data$new_linearisation)
        ) |>
        dplyr::group_by(
          .data$effect,
          .data$iteration
        ) |>
        dplyr::summarise(
          MaxSD = max(.data$sd),
          MeanSD = mean(.data$sd),
          MinSD = min(.data$sd),
          MaxDiff = max(.data$mode - .data$new_linearisation),
          MeanDiff = mean(.data$mode - .data$new_linearisation),
          MinDiff = min(.data$mode - .data$new_linearisation),
          .groups = "drop"
        )
    ) +
    ggplot2::geom_ribbon(
      ggplot2::aes(
        .data$iteration,
        ymin = -.data$MaxSD,
        ymax = .data$MaxSD,
        fill = col_["Max"],
        lty = lty_["SD"]
      ),
      alpha = 0.1,
      na.rm = TRUE
    ) +
    ggplot2::geom_ribbon(
      ggplot2::aes(
        .data$iteration,
        ymin = -.data$MeanSD,
        ymax = .data$MeanSD,
        fill = col_["Mean"],
        lty = lty_["SD"]
      ),
      alpha = 0.1,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxDiff,
        col = col_["Max"],
        lty = lty_["Mode-Lin"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinDiff,
        col = col_["Min"],
        lty = lty_["Mode-Lin"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanDiff,
        col = col_["Mean"],
        lty = lty_["Mode-Lin"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxSD,
        col = col_["Max"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinSD,
        col = col_["Min"],
        lty = lty_["SD"]
      ),
      alpha = 0.5
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanSD,
        col = col_["Mean"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        -.data$MaxSD,
        col = col_["Max"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        -.data$MinSD,
        col = col_["Min"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        -.data$MeanSD,
        col = col_["Mean"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::guides(alpha = "none", fill = "none") +
    pl_theme_rel +
    ggplot2::ggtitle("Mode - Lin")

  track_data_prev <- dplyr::filter(
    track_data,
    .data$iteration < max(.data$iteration)
  )
  track_data_prev$iteration <- track_data_prev$iteration + 1
  track_data <- dplyr::left_join(track_data, track_data_prev,
    by = c("effect", "iteration", "index"),
    suffix = c("", ".prev")
  )

  pl_relative_change <-
    ggplot2::ggplot(
      track_data |>
        dplyr::filter(
          .data$iteration >= max(1, from),
          .data$iteration <= to
        ) |>
        dplyr::group_by(
          .data$effect,
          .data$iteration
        ) |>
        dplyr::mutate(sd = dplyr::if_else(is.finite(.data$sd),
          .data$sd, 1.0
        )) |>
        dplyr::summarise(
          MaxMode = max(abs(.data$mode - .data$mode.prev) / .data$sd),
          MeanMode = mean(abs(.data$mode - .data$mode.prev) / .data$sd),
          RMSMode = mean((.data$mode - .data$mode.prev)^2 / .data$sd^2)^0.5,
          MaxLin = max(abs(.data$new_linearisation -
            .data$new_linearisation.prev) / .data$sd),
          MeanLin = mean(abs(.data$new_linearisation -
            .data$new_linearisation.prev) / .data$sd),
          RMSLin = mean((.data$new_linearisation -
            .data$new_linearisation.prev)^2 / .data$sd^2)^0.5,
          .groups = "drop"
        ) |>
        dplyr::mutate(
          MaxMode = dplyr::if_else(.data$MaxMode > 0, .data$MaxMode, NA),
          MeanMode = dplyr::if_else(.data$MeanMode > 0, .data$MeanMode, NA),
          RMSMode = dplyr::if_else(.data$RMSMode > 0, .data$RMSMode, NA),
          MaxLin = dplyr::if_else(.data$MaxLin > 0, .data$MaxLin, NA),
          MeanLin = dplyr::if_else(.data$MeanLin > 0, .data$MeanLin, NA),
          RMSLin = dplyr::if_else(.data$RMSLin > 0, .data$RMSLin, NA),
        )
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxMode,
        lty = lty_["Mode"],
        col = col_["Max"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanMode,
        lty = lty_["Mode"],
        col = col_["Mean"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxLin,
        lty = lty_["Lin"],
        col = col_["Max"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanLin,
        lty = lty_["Lin"],
        col = col_["Mean"]
      ),
      na.rm = TRUE
    ) +
    pl_theme_norm +
    ggplot2::scale_y_log10() +
    ggplot2::ggtitle("|Change| / sd") +
    ggplot2::geom_hline(yintercept = fit$bru_info$options$bru_method$rel_tol)

  pl_change <-
    ggplot2::ggplot(
      track_data |>
        dplyr::filter(
          .data$iteration >= max(1, from),
          .data$iteration <= to
        ) |>
        dplyr::filter(is.finite(.data$sd)) |>
        dplyr::group_by(
          .data$effect,
          .data$iteration
        ) |>
        dplyr::summarise(
          MaxSD = max(.data$sd),
          MeanSD = mean(.data$sd),
          MinSD = min(.data$sd),
          MaxMode = max(.data$mode - .data$mode.prev),
          MeanMode = mean(.data$mode - .data$mode.prev),
          MinMode = min(.data$mode - .data$mode.prev),
          MaxLin = max(.data$new_linearisation -
            .data$new_linearisation.prev),
          MeanLin = mean(.data$new_linearisation -
            .data$new_linearisation.prev),
          MinLin = min(.data$new_linearisation -
            .data$new_linearisation.prev),
          .groups = "drop"
        )
    ) +
    ggplot2::geom_ribbon(
      ggplot2::aes(
        .data$iteration,
        ymin = -.data$MaxSD,
        ymax = .data$MaxSD,
        fill = col_["Max"],
        lty = lty_["SD"]
      ),
      alpha = 0.1,
      na.rm = TRUE
    ) +
    ggplot2::geom_ribbon(
      ggplot2::aes(
        .data$iteration,
        ymin = -.data$MeanSD,
        ymax = .data$MeanSD,
        fill = col_["Mean"],
        lty = lty_["SD"]
      ),
      alpha = 0.1,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxMode,
        col = col_["Max"],
        lty = lty_["Mode"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinMode,
        col = col_["Min"],
        lty = lty_["Mode"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanMode,
        col = col_["Mean"],
        lty = lty_["Mode"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxLin,
        col = col_["Max"],
        lty = lty_["Lin"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinLin,
        col = col_["Min"],
        lty = lty_["Lin"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanLin,
        col = col_["Mean"],
        lty = lty_["Lin"]
      ),
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MaxSD,
        col = col_["Max"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MinSD,
        col = col_["Min"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        .data$MeanSD,
        col = col_["Mean"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        -.data$MaxSD,
        col = col_["Max"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        -.data$MinSD,
        col = col_["Min"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::geom_line(
      ggplot2::aes(
        .data$iteration,
        -.data$MeanSD,
        col = col_["Mean"],
        lty = lty_["SD"]
      ),
      alpha = 0.5,
      na.rm = TRUE
    ) +
    ggplot2::guides(alpha = "none", fill = "none") +
    pl_theme_rel +
    ggplot2::ggtitle("Change & sd")

  pl_combined <-
    ((
      pl_tracks +
        ggplot2::geom_line(
          ggplot2::aes(.data$iteration, NA_real_, lty = lty_["Mode-Lin"]),
          na.rm = TRUE
        ) +
        ggplot2::geom_line(
          ggplot2::aes(.data$iteration, NA_real_, lty = lty_["SD"]),
          na.rm = TRUE
        ) |
        pl_mode_lin + ggplot2::guides(linetype = "none", color = "none")
    ) /
      (
        pl_relative_change +
          ggplot2::guides(linetype = "none", color = "none") |
          pl_change + ggplot2::guides(linetype = "none", color = "none")
      )
    ) +
      patchwork::plot_layout(guides = "collect") &
      ggplot2::theme(legend.position = "bottom")


  list(
    tracks = pl_tracks,
    mode_lin = pl_mode_lin,
    relative_change = pl_relative_change,
    change = pl_change,
    default = pl_combined
  )
}

make_inla_track_plots <- function(fit, from = 1, to = NULL) {
  needed <- c("dplyr", "ggplot2", "patchwork")
  are_installed <-
    vapply(
      needed,
      function(x) {
        requireNamespace(x, quietly = TRUE)
      },
      TRUE
    )
  if (!all(are_installed)) {
    stop(
      paste0(
        "Needed package(s) ",
        paste0("'", needed[!are_installed], "'", collapse = ", "),
        " not installed, but are needed by make_track_plots()"
      )
    )
  }

  track_data <- fit$bru_iinla$inla_track
  if (!is.null(from)) {
    stopifnot(is.numeric(from))
    stopifnot(from >= 0)
  } else {
    from <- min(track_data$iteration)
  }
  if (!is.null(to)) {
    stopifnot(is.numeric(to))
    stopifnot(to >= 0)
  } else {
    to <- max(track_data$iteration)
  }

  colnames(track_data$theta) <- paste0("theta", seq_len(ncol(track_data$theta)))
  rownames(track_data$theta) <- NULL

  alpha_value_lin_scale <- 0.8

  pl_trace1a <-
    ggplot2::ggplot(
      track_data |>
        dplyr::filter(.data$iteration >= from, .data$iteration <= to),
      ggplot2::aes(
        .data$nfunc_total,
        .data$f,
        col = as.factor(.data$iteration)
      )
    ) +
    ggplot2::geom_line() +
    ggplot2::geom_point() +
    ggplot2::ggtitle("target function value")

  tracks <- cbind(
    track_data |>
      dplyr::select(.data$iteration, .data$f, .data$nfunc, .data$nfunc_total),
    tibble::as_tibble(track_data$theta)
  )
  tracks <- tracks |>
    tidyr::pivot_longer(
      cols = -c(.data$iteration, .data$f, .data$nfunc, .data$nfunc_total),
      names_to = "theta",
      values_to = "value"
    )
  pl_trace1b <-
    ggplot2::ggplot(
      tracks |> dplyr::filter(.data$iteration >= from, .data$iteration <= to),
      ggplot2::aes(
        .data$nfunc_total,
        .data$value,
        col = .data$theta,
        lty = .data$theta
      )
    ) +
    ggplot2::geom_line() +
    ggplot2::geom_point() +
    ggplot2::ggtitle("theta values")

  pl_trace1b <-
    ggplot2::ggplot(
      tracks,
      ggplot2::aes(
        .data$nfunc_total,
        .data$value,
        col = .data$theta,
        lty = .data$theta
      )
    ) +
    ggplot2::geom_line() +
    ggplot2::geom_point() +
    ggplot2::ggtitle("theta values")

  pl_trace_combined <-
    (
      pl_trace1a | pl_trace1b
      #        ggplot2::guides(linetype = "none", color = "none")
    ) +
      patchwork::plot_layout(guides = "collect") &
      ggplot2::theme(legend.position = "bottom")


  tr <- track_data |>
    dplyr::select(.data$iteration, .data$f, .data$nfunc, .data$nfunc_total)
  th <- tibble::as_tibble(track_data$theta)
  tr1 <- cbind(tr[-nrow(tr), ], theta = th[-nrow(tr), ])
  tr2 <- cbind(tr[-nrow(tr), ], theta = th[-1, ])

  track1 <- tr1 |> tidyr::pivot_longer(
    cols = -c(.data$iteration, .data$f, .data$nfunc, .data$nfunc_total),
    names_to = "theta",
    values_to = "value"
  )
  track2 <- tr2 |> tidyr::pivot_longer(
    cols = -c(.data$iteration, .data$f, .data$nfunc, .data$nfunc_total),
    names_to = "theta",
    values_to = "value"
  )
  track1$value_next <- track2$value

  pl_trace2 <-
    ggplot2::ggplot(
      track1 |> dplyr::filter(.data$iteration >= from, .data$iteration <= to)
    ) +
    ggplot2::geom_path(ggplot2::aes(.data$value, .data$value_next)) +
    ggplot2::facet_wrap(~ .data$theta) +
    ggplot2::ggtitle("theta")

  list(
    trace1a = pl_trace1a,
    trace1b = pl_trace1b,
    trace2 = pl_trace2,
    default = pl_trace_combined
  )
}


#' @title Plot inlabru iteration timings
#'
#' @description
#' Draws the time per iteration for preprocessing (including linearisation),
#' `inla()` calls, and
#' line search. Iteration `0` is the time used for defining the model structure.
#'
#' @param x a [bru] object, typically a result from [bru()] for a nonlinear
#' predictor model
#'
#' @details Requires the "ggplot2" package to be installed.
#' @export
#' @examples
#' \dontrun{
#' fit <- bru(...)
#' bru_timings_plot(fit)
#' }
bru_timings_plot <- function(x) {
  needed <- c("ggplot2")
  are_installed <-
    vapply(
      needed,
      function(x) {
        requireNamespace(x, quietly = TRUE)
      },
      TRUE
    )
  if (!all(are_installed)) {
    stop(
      paste0(
        "Needed package(s) ",
        paste0("'", needed[!are_installed], "'", collapse = ", "),
        " not installed, but are needed by bru_timings_plot()"
      )
    )
  }

  timings <- bru_timings(x)

  ggplot2::ggplot(
    tidyr::pivot_longer(
      dplyr::rename(timings, CPU = .data$Time),
      cols = c("CPU", "System", "Elapsed"),
      names_to = "Time",
      values_to = "Value"
    )
  ) +
    ggplot2::geom_point(ggplot2::aes(
      .data$Iteration,
      .data$Value,
      col = .data$Task,
      shape = .data$Task
    )) +
    ggplot2::geom_line(ggplot2::aes(
      .data$Iteration,
      .data$Value,
      col = .data$Task,
      linetype = .data$Time
    )) +
    ggplot2::scale_y_continuous() +
    ggplot2::ylab("Time (s)") +
    ggplot2::guides(
      color = ggplot2::guide_legend(title = "Task"),
      linetype = ggplot2::guide_legend(title = "Time")
    )
}

# @title Make rectangular \code{SpatialPolygonsDataFrame}.
#
# @description
# Makes a rectangular \code{SpatialPolygonsDataFrame} using its bottom left
# coordinates, its width and its height.
#
# @param start Bottom left coordinates (x,y).
# @param width Width of base of the rectangle.
# @param height Height of the rectangle.
#
# @return A \code{SpatialPolygonsDataFrame}.
#
# @export

makepoly <- function(start, width, height) {
  poly <- matrix(
    c(
      start[1], start[2],
      start[1], start[2] + height,
      start[1] + width, start[2] + height,
      start[1] + width, start[2],
      start[1], start[2]
    ),
    ncol = 2, byrow = TRUE
  )
  sp::Polygon(poly)
}

#' @title Create a plot sample.
#'
#' @description
#' Creates a plot sample on a regular grid with a random start location.
#'
#' @param spdf A `SpatialPointsDataFrame` defining the points that are to be
#'   sampled by the plot sample.
#' @param boundary A `SpatialPolygonsDataFrame` defining the survey boundary
#'   within which the  points occur.
#' @param x.ppn The proportion of the x=axis that is to be included in the
#'   plots.
#' @param y.ppn The proportion of the y=axis that is to be included in the
#'   plots.
#' @param nx The number of plots in the x-dimension.
#' @param ny The number of plots in the y-dimension.
#'
#' @return A list with three components:
#'  \describe{
#'    \item{`plots`}{A `SpatialPolygonsDataFrame` object containing the plots
#'      that were sampled.}
#'    \item{`dets`}{A `SpatialPointsDataFrame` object containing the locations
#'      of the points within the plots.}
#'    \item{`counts`}{A `dataframe` containing the following columns
#'    \describe{
#'    \item{`x`}{The x-coordinates of the centres of the plots within the
#'      boundary.}
#'    \item{`y`}{The y-coordinates of the centres of the plots within the
#'      boundary.}
#'    \item{`n`}{The numbers of points in each plot.}
#'    \item{`area`}{The areas of the plots within the boundary}
#'    }}
#'  }.
#'
#' @examples
#' \donttest{
#' # Some features require the raster package
#' if (bru_safe_sp() &&
#'   require("sp") &&
#'   require("raster", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("sf", quietly = TRUE)) {
#'   gorillas <- gorillas_sp()
#'   plotpts <- plotsample(gorillas$nests, gorillas$boundary,
#'     x.ppn = 0.4, y.ppn = 0.4, nx = 5, ny = 5
#'   )
#'   ggplot() +
#'     gg(plotpts$plots) +
#'     gg(plotpts$dets, pch = "+", cex = 2) +
#'     gg(gorillas$boundary)
#' }
#' }
#'
#' @export
#'
plotsample <- function(spdf,
                       boundary,
                       x.ppn = 0.25,
                       y.ppn = 0.25,
                       nx = 5,
                       ny = 5) {
  if (x.ppn <= 0 ||
    x.ppn >= 1) {
    stop("'x.ppn' must greater than 0 and less than 1")
  }
  if (y.ppn <= 0 ||
    y.ppn >= 1) {
    stop("'y.ppn' must greater than 0 and less than 1")
  }

  srange <- raster::extent(boundary)
  xrange <- srange[1:2]
  yrange <- srange[3:4]
  nxtot <- round(nx / x.ppn)
  nytot <- round(ny / y.ppn)
  width <- diff(xrange) / nxtot
  height <- diff(yrange) / nytot
  dx <- diff(xrange) / nx
  dy <- diff(yrange) / ny
  startx <- runif(
    1,
    xrange[1] - 0.99999 * width,
    xrange[1] + dx - 0.99999 * width
  )
  starty <- runif(
    1,
    yrange[1] - 0.99999 * height,
    yrange[1] + dy - 0.99999 * height
  )
  xs <- startx + (0:nx) * dx
  ys <- starty + (0:ny) * dy

  nxs <- length(xs)
  nys <- length(ys)
  starts <- data.frame(x = rep(xs, nys), y = rep(ys, rep(nxs, nys)))
  nplots <- dim(starts)[1]

  polys <- vector("list", nplots)
  for (i in 1:nplots) {
    polys[[i]] <- sp::Polygons(
      list(makepoly(
        as.numeric(starts[i, ]),
        width,
        height
      )),
      i
    )
  }
  plots <- sp::SpatialPolygons(polys, proj4string = fm_CRS(spdf))
  # remove bits of plot outside boundary
  plots <- raster::intersect(boundary, plots)
  # extract only those nests inside the polygons (neat!)
  dets <- spdf[plots, ]

  list(plots = plots, dets = dets)
}


#' @title Convert a plot sample of points into one of counts.
#'
#' @description
#' Converts a plot sample with locations of each point within each plot, into a
#' plot sample with only the count within each plot.
#'
#' @param plots A `SpatialPolygonsDataFrame` object containing the plots that
#'   were sampled.
#' @param dets A `SpatialPointsDataFrame` object containing the locations of the
#'   points within the plots.
#'
#' @return A `SpatialPolygonsDataFrame` with counts in each plot contained in
#'   slot `@data$n`.
#'
#' @examples
#' \donttest{
#' # Some features require the raster package
#' if (bru_safe_sp() &&
#'   require("sp") &&
#'   require("raster", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("sf", quietly = TRUE) &&
#'   require("patchwork", quietly = TRUE)) {
#'   gorillas <- gorillas_sp()
#'   plotpts <- plotsample(gorillas$nests, gorillas$boundary,
#'     x.ppn = 0.4, y.ppn = 0.4, nx = 5, ny = 5
#'   )
#'   p1 <- ggplot() +
#'     gg(plotpts$plots) +
#'     gg(plotpts$dets) +
#'     gg(gorillas$boundary)
#'   countdata <- point2count(plotpts$plots, plotpts$dets)
#'   x <- sp::coordinates(countdata)[, 1]
#'   y <- sp::coordinates(countdata)[, 2]
#'   count <- countdata@data$n
#'   p2 <- ggplot() +
#'     gg(gorillas$boundary) +
#'     gg(plotpts$plots) +
#'     geom_text(aes(label = count, x = x, y = y))
#'   (p1 | p2)
#' }
#' }
#'
#' @export
#'
point2count <- function(plots, dets) {
  np <- length(plots)
  x <- y <- plotarea <- count <- numeric(length = np)
  for (i in 1:np) {
    polylist <- list(sp::Polygons(list(plots@polygons[[i]]@Polygons[[1]]), 1))
    spoly <- sp::SpatialPolygons(
      polylist,
      proj4string = sp::CRS(as.character(sp::proj4string(plots)))
    )
    count[i] <- dim(dets[spoly, ])[1]

    plotarea[i] <- plots@polygons[[i]]@area
    xs <- unique(plots@polygons[[i]]@Polygons[[1]]@coords[, 1])
    ys <- unique(plots@polygons[[i]]@Polygons[[1]]@coords[, 2])
    x[i] <- min(xs) + abs(diff(range(xs)) / 2)
    y[i] <- min(ys) + abs(diff(range(ys)) / 2)
  }

  # make a data frame of it
  countdf <- data.frame(n = count, area = plotarea, x = x, y = y)
  # make SpatialPointsDataFrame of it
  plotcounts <- sp::SpatialPointsDataFrame(
    coords = data.frame(x = x, y = y),
    data = data.frame(n = count, area = plotarea),
    proj4string = sp::CRS(as.character(sp::proj4string(plots)))
  )
  plotcounts
}

#' @name Poisson1_1D
#' @title 1-Dimensional Homogeneous Poisson example.
#' @docType data
#' @description Point data and count data, together with intensity function and
#'   expected counts for a homogeneous 1-dimensional Poisson process example.
#'
#' @aliases lambda1_1D E_nc1 pts1 countdata1
#'
#' @usage data(Poisson1_1D)
#'
#' @format The data contain the following `R` objects:
#'  \describe{
#'    \item{`lambda1_1D`}{ A function defining the intensity function of a
#'    nonhomogeneous Poisson process. Note that this function is only defined on
#'    the interval (0,55).}
#'    \item{`E_nc1`}{ The expected counts of the gridded data.}
#'    \item{`pts1`}{ The locations of the observed points (a data frame with
#'      one column, named `x`).}
#'    \item{`countdata1`}{ A data frame with three columns, containing the
#'      count data:}
#'    \describe{
#'      \item{`x`}{ The grid cell midpoint.}
#'      \item{`count`}{ The number of detections in the cell.}
#'      \item{`exposure`}{ The width of the cell.}
#'    }
#'  }
#'
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE)) {
#'   data(Poisson1_1D)
#'   ggplot(countdata1) +
#'     geom_point(data = countdata1, aes(x = x, y = count), col = "blue") +
#'     ylim(0, max(countdata1$count)) +
#'     geom_point(data = pts1, aes(x = x), y = 0.2, shape = "|", cex = 4) +
#'     geom_point(
#'       data = countdata1, aes(x = x), y = 0, shape = "+",
#'       col = "blue", cex = 4
#'     ) +
#'     xlab(expression(bold(s))) +
#'     ylab("count")
#' }
#' }
NULL

bru_check_object_bru <- function(object,
                                 new_version = getNamespaceVersion("inlabru")) {
  object <-
    bru_info_upgrade(
      object,
      new_version = new_version
    )
  object
}

# Used for upgrading from versions <= 2.7.0.9017 to >= 2.7.0.2021
bru_used_upgrade_2.7.0.9017_to_2.7.0.9021 <- function(lhoods, labels) {
  for (k in seq_along(lhoods)) {
    if (is.null(lhoods[[k]][["used"]]) &&
      is.null(lhoods[[k]][["used_components"]])) {
      used <- bru_used(
        NULL,
        effect = lhoods[[k]][["include_components"]],
        latent = lhoods[[k]][["include_latent"]],
        effect_exclude = lhoods[[k]][["exclude_components"]],
      )

      lhoods[[k]][["include_components"]] <- NULL
      lhoods[[k]][["exclude_components"]] <- NULL
      lhoods[[k]][["include_latent"]] <- NULL
    } else if (!is.null(lhoods[[k]][["used_components"]])) {
      used <- bru_used(
        NULL,
        effect = lhoods[[k]][["effect"]],
        latent = lhoods[[k]][["latent"]]
      )
      lhoods[[k]][["used_components"]] <- NULL
    } else {
      used <- bru_used(lhoods[[k]])
    }

    used <- bru_used_update(used, labels = labels)
    lhoods[[k]][["used"]] <- used
  }
  lhoods
}


bru_info_upgrade_functions <- function() {
  list(
    "2.1.14.901" = function(object) {
      # Ensure INLA_version stored
      if (is.null(object[["INLA_version"]])) {
        object[["INLA_version"]] <- "0.0.0"
      }
      # Check for component$mapper as a bm_multi
      for (k in seq_along(object[["model"]][["effects"]])) {
        cmp <- object[["model"]][["effects"]][[k]]
        cmp[["mapper"]] <-
          bm_multi(
            list(
              main = cmp$main$mapper,
              group = cmp$group$mapper,
              replicate = cmp$replicate$mapper
            ),
            simplify = TRUE
          )
        object[["model"]][["effects"]][[k]] <- cmp
      }
      object
    },
    "2.5.3.9003" = function(object) {
      # Check that likelihoods store 'weights'
      for (k in seq_along(object[["lhoods"]])) {
        lhood <- object[["lhoods"]][[k]]
        if (is.null(lhood[["weights"]])) {
          lhood[["weights"]] <- 1
          object[["lhoods"]][[k]] <- lhood
        }
      }
      object
    },
    "2.5.3.9005" = function(object) {
      # Make sure component$mapper is a bm_pipe
      for (k in seq_along(object[["model"]][["effects"]])) {
        cmp <- object[["model"]][["effects"]][[k]]
        # Convert offset mappers to const mappers
        if (inherits(cmp$main$mapper, c("bm_offset", "bru_mapper_offset"))) {
          cmp$main$mapper <- bm_const()
        }
        cmp[["mapper"]] <-
          bm_pipe(
            list(
              mapper = bm_multi(
                list(
                  main = cmp$main$mapper,
                  group = cmp$group$mapper,
                  replicate = cmp$replicate$mapper
                ),
                simplify = TRUE
              ),
              scale = bm_scale()
            )
          )
        object[["model"]][["effects"]][[k]] <- cmp
      }
      object
    },
    "2.6.0.9000" = function(object) {
      # Make sure component$mapper is a bm_pipe
      for (k in seq_along(object[["model"]][["effects"]])) {
        cmp <- object[["model"]][["effects"]][[k]]
        cmp[["mapper"]] <-
          bm_pipe(
            list(
              mapper = bm_multi(
                list(
                  main = cmp$main$mapper,
                  group = cmp$group$mapper,
                  replicate = cmp$replicate$mapper
                ),
                simplify = TRUE
              ),
              scale = bm_scale()
            )
          )
        object[["model"]][["effects"]][[k]] <- cmp
      }
      object
    },
    "2.7.0.9010" = function(object) {
      # Make sure component$group/replicate$input isn't NULL
      for (k in seq_along(object[["model"]][["effects"]])) {
        cmp <- object[["model"]][["effects"]][[k]]
        if (identical(deparse(cmp[["group"]][["input"]][["input"]]), "NULL")) {
          cmp[["group"]][["input"]][["input"]] <- expression(1L)
        }
        if (
          identical(
            deparse(cmp[["replicate"]][["input"]][["input"]]),
            "NULL"
          )
        ) {
          cmp[["replicate"]][["input"]][["input"]] <- expression(1L)
        }
        object[["model"]][["effects"]][[k]] <- cmp
      }
      object
    },
    "2.7.0.9016" = function(object) {
      # Convert old style include_component/allow_latent to intermediate
      # include_component/include_latent format

      # Update include/exclude information to limit it to existing components
      for (k in seq_along(object[["lhoods"]])) {
        if (isTRUE(object[["lhoods"]][[k]][["allow_latent"]])) {
          # Force inclusion of all components
          object[["lhoods"]][[k]][["include_latent"]] <- NULL
        } else {
          if (is.null(object[["lhoods"]][[k]][["include_latent"]])) {
            object[["lhoods"]][[k]][["include_latent"]] <- character(0)
          }
        }
      }
      object[["lhoods"]] <-
        bru_used_upgrade_2.7.0.9017_to_2.7.0.9021(
          object[["lhoods"]],
          labels = names(object[["model"]][["effects"]])
        )
      object
    },
    "2.7.0.9017" = function(object) {
      # Convert old style include_component/allow_latent to new
      # bru_used format

      # Update include/exclude information to limit it to existing components
      for (k in seq_along(object[["lhoods"]])) {
        if (is.null(object[["lhoods"]][[k]][["used_components"]])) {
          if (isTRUE(object[["lhoods"]][[k]][["allow_latent"]])) {
            # Force inclusion of all components
            object[["lhoods"]][[k]][["include_latent"]] <- NULL
          } else {
            if (is.null(object[["lhoods"]][[k]][["include_latent"]])) {
              object[["lhoods"]][[k]][["include_latent"]] <- character(0)
            }
          }
        }
      }
      object[["lhoods"]] <-
        bru_used_upgrade_2.7.0.9017_to_2.7.0.9021(
          object[["lhoods"]],
          labels = names(object[["model"]][["effects"]])
        )
      object
    },
    "2.7.0.9021" = function(object) {
      # Make sure 'used' components format is properly stored

      if (is.null(object[["lhoods"]][["is_additive"]])) {
        object[["lhoods"]][["is_additive"]] <-
          object[["lhoods"]][["linear"]]
      }

      object[["lhoods"]] <-
        bru_used_upgrade_2.7.0.9017_to_2.7.0.9021(
          object[["lhoods"]],
          labels = names(object[["model"]][["effects"]])
        )
      object
    },
    "2.10.1.9007" = function(object_full) {
      # Update timings info to difftime format

      warning(
        glue(
          "
           From 2.10.1.9007, elapsed time is in Elapsed, Time is CPU time.
             Copying old elapsed time to both Elapsed and Time,
             setting System to zero;
             Do not over-interpret."
        ),
        immediate. = TRUE
      )
      conversion <- as.difftime(
        object_full[["bru_timings"]][["Time"]],
        units = "secs"
      )
      object_full[["bru_timings"]][["Time"]] <- conversion
      object_full[["bru_timings"]][["System"]] <-
        as.difftime(rep(0.0, length(conversion)), units = "secs")
      object_full[["bru_timings"]][["Elapsed"]] <- conversion

      object_full
    },
    "2.10.1.9012" = function(object_full) {
      # Update log format to new format

      object_full[["bru_iinla"]][["log"]] <-
        bru_log_new(
          object_full[["bru_iinla"]][["log"]][["log"]],
          bookmarks = object_full[["bru_iinla"]][["log"]][["bookmarks"]]
        )

      object_full
    },
    "2.12.0.9014" = function(object) {
      # Add bru_obs class name, to support old bru_like objects.
      # Technically for 2.12.0.9017, but the upgrade code needs it for 2014
      if (!is.null(object[["lhoods"]])) {
        object[["lhoods"]] <-
          lapply(object[["lhoods"]], function(x) {
            class(x) <- c("bru_obs", class(x))
            x
          })
        class(object[["lhoods"]]) <- c("bru_obs_list", "list")
      }

      # Update is_additive/linear storage
      object[["lhoods"]][["is_additive"]] <-
        object[["lhoods"]][["linear"]]

      # Update predictor expression
      # object[["lhoods"]] <-
      #   bru_used_update(
      #     object[["lhoods"]],
      #     labels = names(object[["model"]][["effects"]])
      #   )

      object
    },
    "2.12.0.9017" = function(object) {
      # Update bru_like class names to bru_obs
      if (!is.null(object[["lhoods"]])) {
        object[["lhoods"]] <-
          lapply(object[["lhoods"]], function(x) {
            class(x) <- "bru_obs"
            x
          })
        class(object[["lhoods"]]) <- c("bru_obs_list", "list")
      }

      eff <- object[["model"]][["effects"]]
      for (k in seq_along(object[["model"]][["effects"]])) {
        class(eff[[k]][["main"]]) <- "bru_subcomp"
        class(eff[[k]][["group"]]) <- "bru_subcomp"
        class(eff[[k]][["replicate"]]) <- "bru_subcomp"
        class(eff[[k]]) <- "bru_comp"
      }
      class(eff) <- c("bru_comp_list", "list")
      object[["model"]][["effects"]] <- eff

      object
    },
    "2.13.0.9012" = function(object) {
      # Update bru_input input&layer to quosures
      quosure_update <- function(x, env) {
        if (is.null(x)) {
          return(NULL)
        }
        rlang::as_quosure(x, env = env)
      }
      eff <- object[["model"]][["effects"]]
      for (k in seq_along(object[["model"]][["effects"]])) {
        for (part in c("main", "group", "replicate")) {
          if (!is.null(eff[[k]][[part]][["input"]])) {
            for (input_part in c("input", "layer")) {
              eff[[k]][[part]][["input"]][[input_part]] <-
                quosure_update(
                  eff[[k]][[part]][["input"]][[input_part]],
                  eff[[k]][["env_extra"]]
                )
            }
          }
        }
        part <- "weights"
        if (!is.null(eff[[k]][[part]])) {
          for (input_part in c("input", "layer")) {
            eff[[k]][[part]][[input_part]] <-
              quosure_update(
                eff[[k]][[part]][[input_part]],
                eff[[k]][["env_extra"]]
              )
          }
        }
      }
      object[["model"]][["effects"]] <- eff

      object
    },
    "2.13.0.9015" = function(object) {
      # Update lhoods to streamlined storage format
      if (!is.null(object[["lhoods"]])) {
        object[["lhoods"]] <-
          lapply(object[["lhoods"]], function(x) {
            for (what in c("E", "Ntrials", "weights", "scale")) {
              BRU_what <- paste0("BRU_", what)
              if (is.null(x[["response_data"]][[BRU_what]])) {
                x[["response_data"]][[BRU_what]] <- x[[what]]
              }
              x[[what]] <- NULL
            }
            x
          })
        class(object[["lhoods"]]) <- c("bru_obs_list", "list")
      }

      object
    },
    "2.13.0.9017" = function(object) {
      # Update lhoods to new bru_pred_expr object storage format
      if (!is.null(object[["lhoods"]])) {
        lhood_upgrader <- function(x) {
          pred_text <- rlang::expr_deparse(x[["expr"]])
          pred_text <- gsub(
            pattern = "^expression\\(",
            replacement = "(",
            x = pred_text,
            fixed = FALSE
          )
          pred_expr <- x[["expr"]]
          x[["pred_expr"]] <- structure(
            list(
              pred_text = pred_text,
              pred_expr = rlang::parse_expr(pred_text),
              is_additive = x[["is_additive"]],
              is_rowwise = !x[["allow_combine"]],
              is_linear = x[["linear"]],
              used = x[["used"]],
              .envir = x[[".envir"]],
              resp_text = as.character(x[["formula"]])[2]
            ),
            class = "bru_pred_expr"
          )
          x
        }
        object[["lhoods"]] <-
          lapply(object[["lhoods"]], lhood_upgrader)
        class(object[["lhoods"]]) <- c("bru_obs_list", "list")
      }

      object
    },
    "2.13.0.9018" = function(object) {
      # Update component input storage to mapper-attached storage
      if (!is.null(object[["model"]][["effects"]])) {
        object[["model"]][["effects"]] <-
          lapply(object[["model"]][["effects"]], function(x) {
            for (subcomp in c("main", "group", "replicate")) {
              if (
                !is.null(x[[subcomp]][["input"]]) &&
                  inherits(x[[subcomp]][["input"]], "bru_input") &&
                  is.null(x[[subcomp]][["mapper"]][[".input"]])
              ) {
                x[[subcomp]][["mapper"]][[".input"]] <- x[[subcomp]][["input"]]
                x[[subcomp]][["input"]] <- NULL
              }
            }
            if (
              !is.null(x[["marginal"]]) &&
                is.null(x[["marginal"]][[".input"]])
            ) {
              x[["marginal"]][[".input"]] <- new_bru_input(NULL)
            }
            if (
              !is.null(x[["weights"]]) &&
                inherits(x[["weights"]], "bru_input")
            ) {
              weights_input <- x[["weights"]]
              x[["weights"]] <- bm_scale()
              x[["weights"]][[".input"]] <- weights_input
            }
            x <- bru_comp_update_mapper(x)
            x <- bru_compat_pre_2_14_bru_comp(x)
            x
          })
        class(object[["model"]][["effects"]]) <- c("bru_comp_list", "list")
      }

      object
    },
    "2.13.0.9036" = function(object) {
      # Update component input storage to mapper-attached storage
      if (!is.null(object[["model"]][["effects"]])) {
        object[["model"]][["effects"]] <-
          lapply(object[["model"]][["effects"]], function(x) {
            x <- bru_comp_update_mapper(x)
            x <- bru_compat_pre_2_14_bru_comp(x)
            x
          })
        class(object[["model"]][["effects"]]) <- c("bru_comp_list", "list")
      }

      object
    },
    "2.14.0.9002" = function(object) {
      # Update bm_factor mappers to proper levels for cases where previously
      # incorrect data reading resulted in an incorrect (but matching) mapper
      the_inputs <- bru_input(
        object[["lhoods"]],
        object[["model"]][["effects"]]
      )
      if (!is.null(object[["model"]][["effects"]])) {
        object[["model"]][["effects"]] <-
          lapply(object[["model"]][["effects"]], function(x) {
            lab <- x[["label"]]
            mapper <- x[["main"]][["mapper"]]
            if (inherits(mapper, "bm_factor")) {
              inp_levels <- lapply(the_inputs, function(xx) {
                if (is.factor(xx[[lab]][["core"]][["main"]])) {
                  levels(xx[[lab]][["core"]][["main"]])
                } else {
                  NULL
                }
              })
              inp_levels_non_null <-
                vapply(inp_levels, function(x) !is.null(x), logical(1))
              if (any(inp_levels_non_null)) {
                inp_levels <- unique(inp_levels[inp_levels_non_null])
                if (length(unique(inp_levels)) > 1) {
                  warning(glue::glue(
                    "Inconsistent input factor levels for {lab}.
                   Levels found: {paste(unique(inp_levels), collapse = '; ')}.
                   Using levels from first input with non-null levels."
                  ))
                }
                inp_levels <- inp_levels[[1]]
                if (length(inp_levels) != length(mapper$levels)) {
                  warning(glue::glue(
                    "Number of levels for {lab} in mapper ",
                    "({length(mapper$levels)}) does ",
                    "not match number of levels ",
                    "in input ({length(inp_levels)}). ",
                    "Using levels from input."
                  ))
                }
                mapper[["levels"]] <- inp_levels
                mapper[["n"]] <- length(inp_levels)
                if (identical(mapper[["factor_mapping"]], "contrast")) {
                  mapper[["n"]] <- mapper[["n"]] - 1L
                }
                x[["main"]][["mapper"]] <- mapper
              }
            }
            x <- bru_comp_update_mapper(x)
            x
          })
        class(object[["model"]][["effects"]]) <- c("bru_comp_list", "list")
      }

      object
    }
  )
}
bru_info_upgrade <- function(object,
                             new_version = getNamespaceVersion("inlabru")) {
  object_full <- object
  object <- object[["bru_info"]]
  msg <- NULL
  if (!is.list(object)) {
    msg <- "Not a list"
  } else if (!inherits(object, "bru_info")) {
    msg <- "Not a bru_info object"
  } else if (is.null(object[["inlabru_version"]])) {
    msg <- "`inlabru_version` is missing"
  }
  if (!is.null(msg)) {
    stop(glue(
      "bru_info part of the object can't be converted to `bru_info`; {msg}"
    ))
  }

  old_ver <- object[["inlabru_version"]]
  if (utils::compareVersion(new_version, old_ver) > 0) {
    warning(
      glue("Old bru_info object version {old_ver} detected.
            Attempting upgrade to version {new_version}.")
    )

    upgrade_functions <- bru_info_upgrade_functions()

    message("Detected bru_info version ", old_ver)
    for (ver in names(upgrade_functions)) {
      if (utils::compareVersion(ver, old_ver) > 0) {
        message("Upgrading bru_info to ", ver)
        if ("object_full" %in% names(formals(upgrade_functions[[ver]]))) {
          object_full <- upgrade_functions[[ver]](object_full = object_full)
        } else {
          object <- upgrade_functions[[ver]](object)
        }
        object[["inlabru_version"]] <- ver
        object_full[["bru_info"]] <- object
      }
    }

    object[["inlabru_version"]] <- new_version
    object_full[["bru_info"]] <- object
    message(glue("Upgraded bru_info to {new_version}"))
  }
  object_full
}

#' @name Poisson2_1D
#' @title 1-Dimensional NonHomogeneous Poisson example.
#' @docType data
#' @description Point data and count data, together with intensity function and
#'   expected counts for a unimodal nonhomogeneous 1-dimensional Poisson process
#'   example.
#'
#' @aliases lambda2_1D cov2_1D E_nc2 pts2 countdata2
#'
#' @usage data(Poisson2_1D)
#'
#' @format The data contain the following `R` objects:
#'  \describe{
#'    \item{`lambda2_1D`:}{ A function defining the intensity function of a
#'    nonhomogeneous Poisson process. Note that this function is only defined on
#'    the interval (0,55).}
#'    \item{`cov2_1D`:}{ A function that gives what we will call a
#'    'habitat suitability' covariate in 1D space.}
#'    \item{`E_nc2`}{ The expected counts of the gridded data.}
#'    \item{`pts2`}{ The locations of the observed points (a data frame with
#'      one column, named `x`).}
#'    \item{`countdata2`}{ A data frame with three columns, containing the
#'      count data:}
#'    \describe{
#'      \item{`x`}{ The grid cell midpoint.}
#'      \item{`count`}{ The number of detections in the cell.}
#'      \item{`exposure`}{ The width of the cell.}
#'    }
#'  }
#'
#' @examples
#' \donttest{
#' if (require("ggplot2", quietly = TRUE) &&
#'   require("patchwork", quietly = TRUE)) {
#'   data(Poisson2_1D)
#'   p1 <- ggplot(countdata2) +
#'     geom_point(data = countdata2, aes(x = x, y = count), col = "blue") +
#'     ylim(0, max(countdata2$count, E_nc2)) +
#'     geom_point(
#'       data = countdata2, aes(x = x), y = 0, shape = "+",
#'       col = "blue", cex = 4
#'     ) +
#'     geom_point(
#'       data = data.frame(x = countdata2$x, y = E_nc2), aes(x = x),
#'       y = E_nc2, shape = "_", cex = 5
#'     ) +
#'     xlab(expression(bold(s))) +
#'     ylab("count")
#'   ss <- seq(0, 55, length.out = 200)
#'   lambda <- lambda2_1D(ss)
#'   p2 <- ggplot() +
#'     geom_line(
#'       data = data.frame(x = ss, y = lambda),
#'       aes(x = x, y = y), col = "blue"
#'     ) +
#'     ylim(0, max(lambda)) +
#'     geom_point(data = pts2, aes(x = x), y = 0.2, shape = "|", cex = 4) +
#'     xlab(expression(bold(s))) +
#'     ylab(expression(lambda(bold(s))))
#'   (p1 / p2)
#' }
#' }
NULL

#' 1D LGCP bin count simulation and comparison with data
#'
#' A common procedure of analyzing the distribution of 1D points is to chose a
#' binning and plot the data's histogram with respect to this binning. This
#' function compares the counts that the histogram calculates to simulations
#' from a 1D log Gaussian Cox process conditioned on the number of data samples.
#' For each bin this results in a median number of counts as well as a
#' confidence interval. If the LGCP is a plausible model for the observed points
#' then most of the histogram counts (number of points within a bin) should be
#' within the confidence intervals. Note that a proper comparison  is a multiple
#' testing problem which the function does not solve for you.
#'
#'
#' @export
#' @importFrom graphics hist
#' @param result A result object from a [bru()] or [lgcp()] call
#' @param predictor A formula describing the prediction of a 1D LGCP via
#'   [predict()].
#' @param observations A vector of observed values
#' @param breaks A vector of bin boundaries
#' @param nint Number of integration intervals per bin. Increase this if the
#'   bins are wide and the LGCP is not smooth.
#' @param probs numeric vector of probabilities with values in `[0,1]`
#' @param \dots arguments passed on to [predict.bru()]
#' @return An `data.frame` with a ggplot attribute `ggp`
#' @importFrom rlang .data
#' @examples
#' \dontrun{
#' if (require(ggplot2) && require(fmesher) && bru_safe_inla()) {
#'   # Load a point pattern
#'   data(Poisson2_1D)
#'
#'   # Take a look at the point (and frequency) data
#'
#'   ggplot(pts2) +
#'     geom_histogram(
#'       aes(x = x),
#'       binwidth = 55 / 20,
#'       boundary = 0,
#'       fill = NA,
#'       color = "black"
#'     ) +
#'     geom_point(aes(x), y = 0, pch = "|", cex = 4) +
#'     coord_fixed(ratio = 1)
#'
#'   # Fit an LGCP model
#'   x <- seq(0, 55, length.out = 50)
#'   mesh1D <- fmesher::fm_mesh_1d(x, boundary = "free")
#'   matern <- INLA::inla.spde2.pcmatern(mesh1D,
#'     prior.range = c(1, 0.01),
#'     prior.sigma = c(1, 0.01),
#'     constr = TRUE
#'   )
#'   mdl <- x ~ spde1D(x, model = matern) + Intercept(1)
#'   fit.spde <- lgcp(mdl, pts2, domain = list(x = mesh1D))
#'
#'   # Calculate bin statistics
#'   bc <- bincount(
#'     result = fit.spde,
#'     observations = pts2,
#'     breaks = seq(0, max(pts2), length.out = 12),
#'     predictor = x ~ exp(spde1D + Intercept)
#'   )
#'
#'   # Plot them!
#'   attributes(bc)$ggp
#' }
#' }
#'
bincount <- function(
  result,
  predictor,
  observations,
  breaks,
  nint = 20,
  probs = c(0.025, 0.5, 0.975),
  ...
) {
  # Sort probabilities
  probs <- sort(probs)

  # Filter out observations outside bins
  observations <- observations[
    (observations >= min(breaks)) &
      (observations <= max(breaks))
  ]

  # Number of ...
  nobs <- length(observations) # observations
  nbins <- length(breaks) - 1 # bins

  # Mid points of bins
  mid <- breaks[1:(length(breaks) - 1)] + diff(breaks) / 2

  # Create integration points
  nsub <- nint
  u <- rep(
    (seq_len(nsub) - 1L) / (nsub + 1),
    nbins
  )
  domain_breaks <-
    c(
      breaks[rep(seq_len(nbins), each = nsub)] *
        (1 - u) +
        breaks[rep(seq_len(nbins) + 1, each = nsub)] * u,
      breaks[length(breaks)]
    )
  points <- fmesher::fm_int(
    samplers = cbind(
      breaks[1:(length(breaks) - 1)],
      breaks[2:(length(breaks))]
    ),
    domain = fmesher::fm_mesh_1d(domain_breaks),
    int.args = list(
      method = "stable",
      nsub1 = 1L
    ),
    name = as.character(predictor)[2]
  )

  # Sampler
  smp <- generate(
    result,
    newdata = points,
    formula = predictor,
    ...,
    format = "matrix"
  )

  # Integrate per bin
  qq <- matrix(0.0, nrow = nbins, ncol(smp))
  for (k in seq_len(ncol(smp))) {
    qq[, k] <- fmesher::fm_block_logsumexp_eval(
      points$.block,
      weights = points$weight,
      values = log(smp[, k]),
      log = TRUE
    )
    # Normalize bin probabilities
    qq[, k] <- qq[, k] - max(qq[, k])
    qq[, k] <- exp(qq[, k] - log(sum(exp(qq[, k]))))
  }

  # For each bin calculate predictive interval
  pint <- list()
  for (k in 1:nbins) {
    xx <- 0:(nobs + 1L)
    cdff <- function(p) pbinom(xx, size = nobs, prob = p)
    zz <- apply(qq[k, , drop = FALSE], MARGIN = 2, cdff)
    zz <- rowMeans(zz)
    pint[[k]] <- vapply(probs, function(pr) xx[sum(zz < pr) + 1L], 0.0)
  }
  pint <- data.frame(do.call(rbind, pint))
  colnames(pint) <- paste0("q", probs)
  mxname <- paste0("q", max(probs))
  miname <- paste0("q", min(probs))
  mdname <- paste0("q", probs[1 + floor(length(probs) / 2)])

  # Append more information
  pint <- cbind(
    data.frame(
      mid = mid,
      width = diff(breaks),
      counts = hist(observations, breaks = breaks, plot = FALSE)$counts
    ),
    pint
  )
  pint$inside <-
    (pint$counts >= pint[[miname]]) &
      (pint$counts <= pint[[mxname]])

  ggp <- ggplot2::ggplot() +
    ggplot2::geom_crossbar(
      data = pint,
      mapping = ggplot2::aes(
        x = .data[["mid"]],
        y = .data[[mdname]],
        ymin = .data[[miname]],
        ymax = .data[[mxname]],
        fill = .data[["inside"]],
        color = .data[["inside"]]
      ),
      show.legend = FALSE
    ) +
    ggplot2::geom_point(
      data = pint,
      mapping = ggplot2::aes(
        x = .data[["mid"]],
        y = .data[[mdname]]
      ),
      shape = 95,
      size = 3,
      color = "blue"
    ) +
    ggplot2::geom_point(
      data = pint,
      mapping = ggplot2::aes(
        x = .data[["mid"]],
        y = .data[["counts"]]
      ),
      shape = 20,
      size = 2
    ) +
    ggplot2::xlab(all.vars(update.formula(predictor, . ~ 0))) +
    ggplot2::ylab("count")

  attr(pint, "ggp") <- ggp

  # Return
  pint
}


#' Variance and correlations measures for prediction components
#'
#' Calculates local and integrated variance and correlation measures as
#' introduced by Yuan et al. (2017).
#'
#' @export
#' @param joint A joint `prediction` of two latent model components.
#' @param prediction1 A `prediction` of the first component.
#' @param prediction2 A `prediction` of the second component.
#' @param samplers An `sf` or `SpatialPolygon` object describing the area for
#'   which to compute the cumulative variance measure.
#' @param mesh The [fmesher::fm_mesh_2d] for which the prediction was performed
#'   (required for cumulative Vmeasure).
#' @return Variance and correlations measures.
#'
#' @references
#' Y. Yuan, F. E. Bachl, F. Lindgren, D. L. Brochers, J. B. Illian, S. T.
#' Buckland, H. Rue, T. Gerrodette. 2017. Point process models for
#' spatio-temporal distance sampling data from a large-scale survey of blue
#' whales. <https://arxiv.org/abs/1604.06013>
#'
#' @examples
#' \donttest{
#' if (bru_safe_inla() &&
#'   require("ggplot2", quietly = TRUE) &&
#'   require("patchwork", quietly = TRUE) &&
#'   requireNamespace("sn", quietly = TRUE) &&
#'   bru_safe_terra(quietly = TRUE) &&
#'   require("sf", quietly = TRUE) &&
#'   require("RColorBrewer", quietly = TRUE) &&
#'   require("dplyr", quietly = TRUE)) {
#'   # Load Gorilla data
#'
#'   gorillas <- gorillas_sf
#'   gorillas$gcov <- gorillas_sf_gcov()
#'
#'   # Use RColorBrewer
#'
#'   # Fit a model with two components:
#'   # 1) A spatial smooth SPDE
#'   # 2) A spatial covariate effect (vegetation)
#'
#'   pcmatern <- INLA::inla.spde2.pcmatern(
#'     gorillas$mesh,
#'     prior.sigma = c(0.1, 0.01),
#'     prior.range = c(0.01, 0.01)
#'   )
#'
#'   cmp <- geometry ~ 0 +
#'     vegetation(gorillas$gcov$vegetation, model = "factor_contrast") +
#'     spde(geometry, model = pcmatern) +
#'     Intercept(1)
#'
#'   fit <- lgcp(
#'     cmp,
#'     gorillas$nests,
#'     samplers = gorillas$boundary,
#'     domain = list(geometry = gorillas$mesh),
#'     options = list(control.inla = list(int.strategy = "eb"))
#'   )
#'
#'   # Predict SPDE and vegetation at a grid covering the domain of interest
#'   pred_loc <- fmesher::fm_pixels(
#'     gorillas$mesh,
#'     mask = gorillas$boundary,
#'     dims = c(200, 200),
#'     format = "sf"
#'   )
#'   pred <- predict(
#'     fit,
#'     pred_loc,
#'     ~ list(
#'       joint = spde + vegetation,
#'       field = spde,
#'       veg = vegetation
#'     )
#'   )
#'
#'   pred_collect <-
#'     rbind(
#'       pred$joint |> dplyr::mutate(component = "joint"),
#'       pred$field |> dplyr::mutate(component = "field"),
#'       pred$veg |> dplyr::mutate(component = "veg")
#'     ) |>
#'     dplyr::mutate(var = sd^2)
#'
#'   # Plot component mean
#'
#'   ggplot(pred_collect) +
#'     geom_tile(aes(geometry = geometry, fill = mean),
#'       stat = "sf_coordinates"
#'     ) +
#'     coord_equal() +
#'     facet_wrap(~component, nrow = 1) +
#'     theme(legend.position = "bottom")
#'
#'   # Plot component variance
#'
#'   ggplot(pred_collect) +
#'     geom_tile(aes(geometry = geometry, fill = var),
#'       stat = "sf_coordinates"
#'     ) +
#'     coord_equal() +
#'     facet_wrap(~component, nrow = 1) +
#'     theme(legend.position = "bottom")
#'
#'   # Calculate variance and correlation measure
#'
#'   vm <- devel.cvmeasure(pred$joint, pred$field, pred$veg)
#'   # Compute nominal relative variance contributions; note that these can be
#'   # greater than 100%!
#'   vm <- dplyr::mutate(
#'     vm,
#'     var1_rel = if_else(var1 <= 0, NA, var1 / var.joint),
#'     var2_rel = if_else(var2 <= 0, NA, var2 / var.joint)
#'   )
#'   lprange <- range(vm$var.joint, vm$var1, vm$var2)
#'   vm <- tidyr::pivot_longer(vm,
#'     cols = c(var.joint, var1, var2, cov, cor, var1_rel, var2_rel),
#'     names_to = "component",
#'     values_to = "value"
#'   )
#'
#'   # Variance contribution of the components
#'
#'   csc <- scale_fill_gradientn(
#'     colours = brewer.pal(9, "YlOrRd"),
#'     limits = lprange
#'   )
#'
#'   vm_ <- dplyr::filter(vm, component %in% c("var.joint", "var1", "var2"))
#'   ggplot(vm_) +
#'     geom_tile(aes(geometry = geometry, fill = value),
#'       stat = "sf_coordinates"
#'     ) +
#'     csc +
#'     coord_equal() +
#'     facet_wrap(~component, nrow = 1) +
#'     theme(legend.position = "bottom")
#'
#'   # Relative variance contribution of the components
#'   # When bo
#'
#'   vm_ <- dplyr::filter(vm, component %in% c("var1_rel", "var2_rel"))
#'   ggplot(vm_) +
#'     geom_tile(aes(geometry = geometry, fill = value),
#'       stat = "sf_coordinates"
#'     ) +
#'     scale_fill_gradientn(
#'       colours = brewer.pal(9, "YlOrRd"),
#'       trans = "log"
#'     ) +
#'     coord_equal() +
#'     facet_wrap(~component, nrow = 1)
#'
#'   # Where both relative contributions are larger than 1, the posterior
#'   # correlations are strongly negative.
#'
#'   # Covariance and correlation of field and vegetation
#'
#'   vm_cov <- dplyr::filter(vm, component %in% "cov")
#'   vm_cor <- dplyr::filter(vm, component %in% "cor")
#'   (ggplot(vm_cov) +
#'     geom_tile(aes(geometry = geometry, fill = value),
#'       stat = "sf_coordinates"
#'     ) +
#'     scale_fill_gradientn(
#'       colours = rev(brewer.pal(9, "RdBu")),
#'       limits = c(-1, 1) * max(abs(vm_cov$value), na.rm = TRUE)
#'     ) +
#'     coord_equal() +
#'     theme(legend.position = "bottom") +
#'     ggtitle("Covariances") |
#'     ggplot(vm_cor) +
#'       geom_tile(aes(geometry = geometry, fill = value),
#'         stat = "sf_coordinates"
#'       ) +
#'       scale_fill_gradientn(
#'         colours = rev(brewer.pal(9, "RdBu")),
#'         limits = c(-1, 1) * max(abs(vm_cor$value), na.rm = TRUE)
#'       ) +
#'       coord_equal() +
#'       theme(legend.position = "bottom") +
#'       ggtitle("Correlations")
#'   )
#'
#'   # Variance and correlation integrated over space
#'
#'   vrt <- fmesher::fm_vertices(gorillas$mesh, format = "sf")
#'   pred_vrt <- predict(
#'     fit,
#'     vrt,
#'     ~ list(
#'       joint = spde + vegetation,
#'       field = spde,
#'       veg = vegetation
#'     )
#'   )
#'
#'   vm.int <- devel.cvmeasure(
#'     pred_vrt$joint,
#'     pred_vrt$field,
#'     pred_vrt$veg,
#'     samplers = fmesher::fm_int(gorillas$mesh, gorillas$boundary),
#'     mesh = gorillas$mesh
#'   )
#'   vm.int
#' }
#' }
#'
devel.cvmeasure <- function(
  joint,
  prediction1,
  prediction2,
  samplers = NULL,
  mesh = NULL
) {
  # Covariance
  joint$cov <- (joint[["sd"]]^2 -
    prediction1[["sd"]]^2 -
    prediction2[["sd"]]^2) /
    2

  # Correlation
  corr <- function(joint, a, b) {
    ((joint - a - b) / (2 * sqrt(a * b)))
  }
  cor <- corr(joint[["sd"]]^2, prediction1[["sd"]]^2, prediction2[["sd"]]^2)
  if (any((cor > 1) | (cor < (-1)), na.rm = TRUE)) {
    cor[(cor > 1) | (cor < (-1))] <- NA
  }
  joint$cor <- cor

  if (!is.null(samplers)) {
    #
    # PRESENTED TO YOU BY HACKY McHACKERSON
    #

    wips <- fm_int(mesh, samplers)
    A <- fm_basis(mesh, loc = wips)

    weights <- wips$weight
    weights <- weights / sum(weights)
    vj <- sum(as.vector(A %*% joint[["sd"]]^2) * weights)
    v1 <- sum(as.vector(A %*% prediction1[["sd"]]^2) * weights)
    v2 <- sum(as.vector(A %*% prediction2[["sd"]]^2) * weights)
    cr <- corr(vj, v1, v2)
    ret <- data.frame(var.joint = vj, var1 = v1, var2 = v2, cor = cr)
  } else {
    tmp <- attributes(joint)
    ret <- joint[, c("cov", "cor")]
    ret$var.joint <- joint[["sd"]]^2
    ret$var1 <- prediction1[["sd"]]^2
    ret$var2 <- prediction2[["sd"]]^2
    attr(ret, "misc") <- tmp[["misc"]]
    attr(ret, "type") <- tmp[["type"]]
  }

  ret
}

#' @include 0_inlabru_envir.R

#' @title Get access to the internal environment
#' @details The environment is defined in `0_inlabru_envir.R` which is loaded
#'   first.
#' @keywords internal
bru_env_get <- function() {
  pkg_envir <- parent.env(environment())
  envir <- get0(".inlabru_envir", envir = pkg_envir)
  if (!is.environment(envir)) {
    stop(
      "Something went wrong: cannot find internal .inlabru_envir environment."
    )
  }
  envir
}


# Documentation would clash with base .onLoad documentation
# @title Initialise log storage and global options
# @param libname a character string giving the library directory where the
#   package defining the namespace was found.
# @param pkgname a character string giving the name of the package.
# @aliases namespace_hooks
# @keywords internal
# @rdname namespace_hooks
.onLoad <- function(libname, pkgname) {
  bru_log_reset()
  bru_log_message("inlabru loaded", allow_verbose = FALSE)
  bru_log_message("Clear override options", allow_verbose = FALSE)
  bru_options_reset()
  # For Matrix coercion deprecation testing: 1=warn, 2=stop, NA=something else
  #  options(Matrix.warnDeprecatedCoerce = 2)
}


# inlabru log methods ----

#' @title Clear log contents
#' @description
#' Clears the log contents up to
#' a given `offset` or `bookmark`. Default: clear the entire log.
#' When `x` is NULL, the global `inlabru` log is updated, and `invisible(NULL)`
#' is returned. Otherwise the updated object is returned (invisibly).
#' @param x A `bru_log` object, or in some cases, and object that can be
#' converted/extracted to a `bru_log` object. `NULL` denotes the global
#' `inlabru` log object.
#' @inheritParams bru_log_offset
#' @returns Returns (invisibly) the modified `bru_log` object, or `NULL` (when
#'   `x` is `NULL`)
#' @examples
#' \dontrun{
#' if (interactive()) {
#'   bru_log_reset()
#' }
#' }
#' @export
#' @family inlabru log methods

bru_log_reset <- function(x = NULL, bookmark = NULL, offset = NULL) {
  offset <- bru_log_offset(x = x, bookmark = bookmark, offset = offset)
  if (is.null(x)) {
    # Clear log up to the given offset
    envir <- bru_env_get()
    log_length <- length(envir[["log"]])
    if (offset >= log_length) {
      envir[["log"]] <- bru_log_new()
      return(invisible(NULL))
    }
    if (offset == 0L) {
      return(invisible(NULL))
    }
    index <- offset + seq_len(log_length - offset)
    envir[["log"]] <- envir[["log"]][index]
    return(invisible(NULL))
  }
  # Clear log up to the given offset
  log_length <- length(x)
  if (offset >= log_length) {
    x <- bru_log_new()
    return(invisible(x))
  }
  if (offset == 0L) {
    return(invisible(x))
  }
  index <- offset + seq_len(log_length - offset)
  x <- x[index]
  invisible(x)
}


#' @export
#' @title Create a `bru_log` object
#' @description
#' Create a `bru_log` object, by default empty.
#' @param x An optional character vector of log messages, or `data.frame`
#' with columns `message`, `timestamp`, and `verbosity`.
#' @param bookmarks An optional `integer` vector of named bookmarks
#' message in `x`.
#' @family inlabru log methods
#' @examples
#' x <- bru_log_new()
#' x <- bru_log_message("Test message", x = x)
#' print(x)
bru_log_new <- function(x = NULL, bookmarks = NULL) {
  if (is.null(x)) {
    x <- character(0)
  }
  if (!is.data.frame(x)) {
    x <- data.frame(
      message = as.character(x)
    )
  }
  if (is.null(x[["timestamp"]])) {
    ts <- Sys.time()
    is.na(ts) <- TRUE
    ts <- rep(ts, nrow(x))
    x[["timestamp"]] <- ts
  }
  if (is.null(x[["verbosity"]])) {
    x[["verbosity"]] <- integer(nrow(x))
  }
  bookmarks <- if (is.null(bookmarks)) {
    integer(0)
  } else {
    storage.mode(bookmarks) <- "integer"
    bookmarks
  }
  x <- structure(
    list(
      log = x,
      bookmarks = bookmarks
    ),
    class = "bru_log"
  )
  x
}


#' @title Methods for `bru_log` bookmarks
#' @description
#' Methods for `bru_log` bookmarks.
#'
#' @param bookmark character; The label for a bookmark with a stored offset.
#' @param offset integer; a position offset in the log, with `0L` pointing at
#' the start of the log. If negative, denotes the point `abs(offset)` elements
#' from tail of the log. When `bookmark` is non-NULL, the `offset` applies a
#' shift (forwards or backwards) to the bookmark list.
#' @param x A `bru_log` object. If `NULL`, the global `inlabru` log is used.
#' @returns `bru_log_bookmark()`: Returns the modified `bru_log` object if `x`
#'   is non-NULL.
#' @export
#' @describeIn bru_log_bookmark
#' Set a log bookmark. If `offset` is `NULL` (the default),
#' the bookmark will point to the current end of the log.
#' @family inlabru log methods
bru_log_bookmark <- function(bookmark = "", offset = NULL, x = NULL) {
  if (nchar(bookmark) == 0) {
    stop("Bookmark labels must have at least one character.")
  }
  offset <- bru_log_offset(x = x, bookmark = NULL, offset = offset)
  if (is.null(x)) {
    envir <- bru_env_get()
    envir[["log"]][["bookmarks"]] <- c(
      envir[["log"]][["bookmarks"]],
      structure(offset, names = bookmark)
    )
    return(invisible(NULL))
  }
  x[["bookmarks"]] <- c(
    x[["bookmarks"]],
    structure(offset, names = bookmark)
  )
  invisible(x)
}

#' @export
#' @describeIn bru_log_bookmark Return a integer vector with named elements
#'   being bookmarks into the global `inlabru` log with associated log position
#'   offsets.
#' @returns `bru_log_bookmarks()`: Returns the bookmark vector associated with
#'   `x`
bru_log_bookmarks <- function(x = NULL) {
  if (is.null(x)) {
    bru_env_get()[["log"]][["bookmarks"]]
  } else {
    x[["bookmarks"]]
  }
}

#' @title Position methods for `bru_log` objects
#' @export
#' @description
#' Position methods for `bru_log` objects.
#'
#' @describeIn bru_log_offset
#' Utility function for computing log position offsets.
#' @family inlabru log methods
#' @inheritParams bru_log_bookmark
bru_log_offset <- function(x = NULL,
                           bookmark = NULL,
                           offset = NULL) {
  if (is.null(x)) {
    log_length <- length(bru_env_get()[["log"]])
  } else {
    log_length <- length(x)
  }
  if (is.null(bookmark)) {
    if (is.null(offset) || (offset > log_length)) {
      return(log_length)
    }
    if (offset < 0L) {
      offset <- max(0L, log_length + offset)
    }
    return(offset)
  }
  marks <- c(
    "STARTOFLOG" = 0L,
    bru_log_bookmarks(x = x),
    "ENDOFLOG" = log_length
  )
  if (bookmark == "") {
    bookmark <- names(marks)[length(marks) - 1L]
  }
  found <- names(marks) %in% bookmark
  if (any(found)) {
    which_found <- max(which(found))
    found <- marks[[which_found]]
  } else {
    warning(glue("Log bookmark '{bookmark}' not found; assuming start of log."))
    marks <- c(0L, log_length)
    which_found <- 1L
  }
  which_found <- max(1L, min(length(marks), which_found + offset))
  marks[[which_found]]
}

#' @export
#' @describeIn bru_log_offset Utility function for computing index vectors
#' for `bru_log` objects.
#' @param i indices specifying elements to extract. If `character`, denotes
#' the sequence between bookmark `i` and the next bookmark (or the end of the
#' log if `i` is the last bookmark)
#' @param verbosity integer value for limiting the highest verbosity level being
#'   returned.
bru_log_index <- function(x = NULL, i, verbosity = NULL) {
  if (is.null(x)) {
    x <- bru_env_get()[["log"]]
  }
  log_length <- length(x)
  if (is.null(i)) {
    i <- seq_len(log_length)
  } else if (is.character(i)) {
    offset0 <- bru_log_offset(x, bookmark = i, offset = 0L)
    offset1 <- bru_log_offset(x, bookmark = i, offset = 1L)
    i <- offset0 + seq_len(offset1 - offset0)
  } else if (is.logical(i)) {
    stopifnot(length(i) == log_length)
    i <- which(i)
  }
  if (any(i < 0L)) {
    stopifnot(all(i < 0L))
    i <- setdiff(seq_len(log_length), -i)
  }
  if ((length(i) > 0) && !is.null(verbosity)) {
    i <- i[x[["log"]][["verbosity"]][i] <= verbosity]
  }
  i
}


#' @title Access methods for `bru_log` objects
#' @description Access method for `bru_log` objects.
#' Note: Up to version `2.8.0`, `bru_log()` was a deprecated alias for
#' `bru_log_message()`. When running on `2.8.0` or earlier, use `bru_log_get()`
#' to access the global log, and `cat(fit$bru_iinla$log, sep = "\n")` to print a
#' stored estimation object log.
#' After version `2.8.0`, use `bru_log()` to access the global log, and
#' `bru_log(fit)` to access a stored estimation log.
#' @param x An object that is, contains, or can be converted to,
#' a `bru_log` object. If `NULL`, refers to the global `inlabru` log.
#' @param verbosity integer value for limiting the highest verbosity level being
#'   returned.
#' @return `bru_log` A `bru_log` object, containing a
#' character vector of log messages, and potentially a vector of bookmarks.
#' @export
#' @family inlabru log methods
#' @describeIn bru_log Extract stored log messages. If non-`NULL`, the
#'   `verbosity` argument determines the maximum verbosity level of the messages
#'   to extract.
bru_log <- function(x = NULL, verbosity = NULL) {
  if (is.null(x)) {
    x <- bru_env_get()[["log"]]
  }
  if (inherits(x, "bru_log")) {
    if (!is.null(verbosity)) {
      i <- bru_log_index(x, i = NULL, verbosity = verbosity)
      x <- x[i]
    }
    return(x)
  }
  UseMethod("bru_log")
}

#' @rdname bru_log
#' @export
bru_log.character <- function(x, verbosity = NULL) {
  y <- bru_log_new(x = x)
  if (!is.null(verbosity)) {
    i <- bru_log_index(y, i = NULL, verbosity = verbosity)
    y <- y[i]
  }
  y
}

#' @rdname bru_log
#' @export
bru_log.bru_log <- function(x, verbosity = NULL) {
  if (!is.null(verbosity)) {
    i <- bru_log_index(x, i = NULL, verbosity = verbosity)
    x <- x[i]
  }
  x
}

#' @rdname bru_log
#' @export
bru_log.iinla <- function(x, verbosity = NULL) {
  if (is.null(x[["log"]])) {
    return(bru_log_new(character(0)))
  }
  bru_log(x[["log"]], verbosity = verbosity)
}

#' @rdname bru_log
#' @export
bru_log.bru <- function(x, verbosity = NULL) {
  x <- bru_check_object_bru(x)

  if (is.null(x[["bru_info"]][["log"]])) {
    result <- bru_log_new(character(0))
  } else {
    result <- bru_log(x[["bru_info"]][["log"]], verbosity = verbosity)
  }

  if (is.null(x[["bru_iinla"]][["log"]])) {
    return(result)
  }
  c(result, bru_log(x[["bru_iinla"]][["log"]], verbosity = verbosity))
}

#' @describeIn bru_log Format a `bru_log` object for printing.
#' If `verbosity` is `TRUE`, include the verbosity level of each message.
#' @param \dots further arguments passed to or from other methods.
#' @param timestamp If `TRUE`, include the timestamp of each message. Default
#'   `TRUE`.
#' @export
#' @examples
#' bru_log(verbosity = 2L)
#' format(bru_log())
#'
format.bru_log <- function(x, ..., timestamp = TRUE, verbosity = FALSE) {
  msg <- x[["log"]][["message"]]
  if (timestamp) {
    msg <- glue("{x[['log']][['timestamp']]}: {msg}")
  }
  if (verbosity) {
    msg <- glue("{msg} (level {x[['log']][['verbosity']]})")
  }
  msg
}

#' @describeIn bru_log Print a `bru_log` object with `cat(x, sep = "\n")`.
#' If `verbosity` is `TRUE`, include the verbosity level of each message.
#' @export
#' @examples
#' bru_log(verbosity = 2L)
#' print(bru_log(), timestamp = TRUE, verbosity = TRUE)
#'
print.bru_log <- function(x, ..., timestamp = TRUE, verbosity = FALSE) {
  cat(
    format(x,
      ...,
      timestamp = timestamp,
      verbosity = verbosity
    ),
    sep = "\n"
  )
  invisible(x)
}


#' @export
#' @describeIn bru_log Convert `bru_log` object to a plain `character` vector
as.character.bru_log <- function(x, ...) {
  x[["log"]][["message"]]
}


#' @export
#' @param i indices specifying elements to extract. If `character`, denotes
#' the sequence between bookmark `i` and the next bookmark (or the end of the
#' log if `i` is the last bookmark)
#' @describeIn bru_log Extract a subset of a `bru_log` object
`[.bru_log` <- function(x, i) {
  i <- bru_log_index(x, i)
  if (length(i) == 0L) {
    marks <- integer(0)
  } else {
    marks <- x[["bookmarks"]]
    marks <- marks[(marks >= min(i) - 1L) & (marks <= max(i))]
    for (k in seq_along(marks)) {
      marks[k] <- sum(i <= marks[k])
    }
  }
  bru_log_new(
    x = x[["log"]][i, , drop = FALSE],
    bookmarks = marks
  )
}

#' @export
#' @describeIn bru_log Concatenate several `bru_log` or `character` objects
#' into a `bru_log` object.
`c.bru_log` <- function(...) {
  obj <- list(...)
  is_ch <- vapply(
    obj,
    function(x) inherits(x, "character"),
    TRUE
  )
  if (any(is_ch)) {
    for (k in which(is_ch)) {
      obj[[k]] <- bru_log_new(x = obj[[k]])
    }
  }
  stopifnot(all(vapply(
    obj,
    function(x) inherits(x, "bru_log"),
    TRUE
  )))
  offset <- 0L
  for (k in seq_along(obj)) {
    obj[[k]][["bookmarks"]] <- obj[[k]][["bookmarks"]] + offset
    offset <- offset + length(obj[[k]])
  }
  bru_log_new(
    x = do.call(rbind, lapply(obj, function(x) x[["log"]])),
    bookmarks = do.call(c, lapply(obj, function(x) x[["bookmarks"]]))
  )
}

#' @export
#' @describeIn bru_log Obtain the number of log entries
#' into a `bru_log` object.
`length.bru_log` <- function(x) {
  NROW(x[["log"]])
}


#' @title Add a log message
#' @description Adds a log message.
#' @param \dots For `bru_log_message()`, zero or more objects passed on to
#'   [`base::.makeMessage()`]. For `bru_log_abort()` and `bru_log_warn()`,
#'   passed on to `rlang::abort()` and `rlang::warn()`.
#' @param domain Domain for translations, passed on to [`base::.makeMessage()`]
#' @param appendLF logical; whether to add a newline to the message. Only
#'   used for verbose output.
#' @param verbosity numeric value describing the verbosity level of the message
#' @param allow_verbose Whether to allow verbose output. Must be set to FALSE
#' until the options object has been initialised.
#' @param verbose logical, numeric, or `NULL`; local override for verbose
#' output. If `NULL`, the global option `bru_verbose` or default value is used.
#' If `FALSE`, no messages are printed. If `TRUE`, messages with `verbosity`
#' \eqn{\le 1}{<=1}
#' are printed. If numeric, messages with `verbosity` \eqn{\le}{<=} `verbose`
#' are printed.
#' @param verbose_store Same as `verbose`, but controlling what messages are
#'   stored in the global log object. Can be controlled via the
#'   `bru_verbose_store` with [bru_options_set()].
#' @param x A `bru_log` object. If `NULL`, refers to the global `inlabru` log.
#' @return
#' `bru_log_message` returns `invisible(x)`, where `x` is the updated `bru_log`
#' object, or `NULL`.
#' @family inlabru log methods
#' @examples
#' if (interactive()) {
#'   code_runner <- function() {
#'     bru_options_set_local(
#'       # Show messages up to and including level 2 (default 0)
#'       bru_verbose = 2,
#'       # Store messages to an including level 3 (default Inf, storing all)
#'       bru_verbose_store = 3
#'     )
#'
#'     bru_log_bookmark("bookmark 1")
#'     bru_log_message("Test message 1", verbosity = 1)
#'     bru_log_message("Test message 2", verbosity = 2)
#'     bru_log_bookmark("bookmark 2")
#'     bru_log_message("Test message 3", verbosity = 3)
#'     bru_log_message("Test message 4", verbosity = 4)
#'
#'     invisible()
#'   }
#'   message("Run code")
#'   code_runner()
#'   message("Check log from bookmark 1")
#'   print(bru_log()["bookmark 1"])
#'   message("Check log from bookmark 2")
#'   print(bru_log()["bookmark 2"])
#' }
#' @export

bru_log_message <- function(..., domain = NULL, appendLF = TRUE,
                            verbosity = 1L,
                            allow_verbose = TRUE, verbose = NULL,
                            verbose_store = NULL,
                            x = NULL) {
  new_x <- data.frame(
    message = .makeMessage(...,
      domain = domain,
      appendLF = FALSE
    ),
    timestamp = Sys.time(),
    verbosity = as.integer(verbosity)
  )
  if (allow_verbose) {
    if ((!is.null(verbose) && (verbose >= verbosity)) ||
      (is.null(verbose) &&
        bru_options_get("bru_verbose", include_default = TRUE) >= verbosity)) {
      message(new_x[["message"]], domain = NA, appendLF = appendLF)
    }
  }
  if ((!is.null(verbose_store) && (verbose_store >= verbosity)) ||
    !allow_verbose ||
    (is.null(verbose_store) &&
      (bru_options_get("bru_verbose_store",
        include_default = TRUE
      ) >= verbosity))) {
    if (is.null(x)) {
      envir <- bru_env_get()
      envir[["log"]][["log"]] <- rbind(envir[["log"]][["log"]], new_x)
    } else {
      x[["log"]] <- rbind(x[["log"]], new_x)
    }
  }
  if (is.null(x)) {
    invisible(NULL)
  } else {
    invisible(x)
  }
}


#' @describeIn bru_log_message Store a log message and throw an error.
#' @param msg character; passed to [`base::.makeMessage()`]
#' @param call The calling environment.
#' @param .frame The throwing context, for when `.internal` is `TRUE`
#' @export
bru_log_abort <- function(
  msg,
  ...,
  domain = NULL,
  appendLF = TRUE,
  verbosity = 1L,
  allow_verbose = TRUE,
  verbose = FALSE,
  verbose_store = NULL,
  call = rlang::caller_env(),
  .frame = rlang::caller_env()
) {
  bru_log_message(
    msg,
    domain = domain,
    appendLF = appendLF,
    verbosity = verbosity,
    allow_verbose = allow_verbose,
    verbose = verbose,
    verbose_store = verbose_store
  )
  rlang::abort(msg, call = call, .frame = .frame, ...)
}

#' @describeIn bru_log_message Store a log message and throw a warning.
#' @export
bru_log_warn <- function(
  msg,
  ...,
  domain = NULL,
  appendLF = TRUE,
  verbosity = 1L,
  allow_verbose = TRUE,
  verbose = FALSE,
  verbose_store = NULL,
  call = rlang::caller_env(),
  .frame = rlang::caller_env()
) {
  bru_log_message(
    msg,
    domain = domain,
    appendLF = appendLF,
    verbosity = verbosity,
    allow_verbose = allow_verbose,
    verbose = verbose,
    verbose_store = verbose_store
  )
  ## Need to solve inla swallowing warnings before this works:
  ## rlang::warn(msg, call = call, .frame = .frame, ...)
  warning(msg, immediate. = TRUE)
}


# Options methods ----

#' @title Create or update an options objects
#' @description Create a new options object, or merge information from several
#'   objects.
#'
#'   The `_get`, `_set`, and `_reset` functions operate on a global
#'   package options override object. In many cases, setting options in
#'   specific calls to [bru()] is recommended instead.
#' @param \dots A collection of named options, optionally including one or more
#'   [`bru_options`] objects. Options specified later override the previous
#'   options.
#' @return `bru_options()` returns a `bru_options` object.
#'
#' @examples
#' \dontrun{
#' if (interactive()) {
#'   # Combine global and user options:
#'   options1 <- bru_options(bru_options_get(), bru_verbose = TRUE)
#'   # Create a proto-options object in two equivalent ways:
#'   options2 <- as.bru_options(bru_verbose = TRUE)
#'   options2 <- as.bru_options(list(bru_verbose = TRUE))
#'   # Combine options objects:
#'   options3 <- bru_options(options1, options2)
#' }
#' }
#' @export
#' @rdname bru_options

bru_options <- function(...) {
  new_bru_options <- function() {
    options <- list()
    class(options) <- c("bru_options", "list")
    options
  }

  input_options <- list(...)

  options <- new_bru_options()
  for (k in seq_along(input_options)) {
    if (inherits(input_options[[k]], "bru_options")) {
      options <- modifyList(options, input_options[[k]], keep.null = TRUE)
    } else {
      options <- modifyList(options, input_options[k], keep.null = TRUE)
    }
  }

  options
}

#' @param x An object to be converted to an `bru_options` object.
#' @return For `as.bru_options()`, `NULL` or no input returns an empty
#' `bru_options` object, a `list` is converted via `bru_options(...)`,
#' and `bru_options` input is passed through. Other types of input generates
#' an error.
#'
#' @export
#' @describeIn bru_options Coerces inputs to a `bru_options` object.

as.bru_options <- function(x = NULL) {
  if (inherits(x, "bru_options")) {
    x
  } else if (is.null(x)) {
    bru_options()
  } else if (is.list(x)) {
    do.call(bru_options, x)
  } else {
    stop("Unable coerce object to 'bru_options'")
  }
}

#' @describeIn bru_options Returns the default options.
#' @return `bru_options_default()` returns an `bru_options` object containing
#'   default options.
#' @export
#' @section Valid options:

bru_options_default <- function() {
  bru_options(
    # inlabru options
    #' For `bru_options` and `bru_options_set`, recognised options are:
    #' \describe{
    #' \item{bru_verbose}{
    #' numeric or logical; if `TRUE`, log messages with `verbosity` \eqn{\le}
    #' 1 are printed by [bru_log_message()]. If numeric, messages with
    #' `verbosity` \eqn{\le} `bru_verbose` are printed.
    #' For line search details, set `bru_verbose=2` or `3`.
    #' Default: 0, to not print any messages.}
    bru_verbose = 0,
    #' \item{bru_verbose_store}{
    #' numeric or logical, as for `bru_verbose`, but controls what messages
    #' are stored in the global log object.
    #' Default: `Inf`, to store all messages.}
    bru_verbose_store = Inf,
    #' \item{bru_max_iter}{maximum number of inla iterations, default 10.
    #'  Also see `bru_method$rel_tol` and related options below.}
    bru_max_iter = 10,
    #' \item{bru_run}{
    #'   If `TRUE`, run inference. Otherwise only return configuration needed to
    #'   run inference.}
    bru_run = TRUE,
    #' \item{bru_initial}{
    #'   An `inla` object returned from previous calls of `INLA::inla`, [bru()]
    #'   or [lgcp()], or a list of named vectors of starting values for the
    #'   latent variables. This will be used as a starting point for further
    #'   improvement of the approximate posterior.}
    #' \item{bru_int_args}{
    #' List of arguments passed all the way to the integration method
    #' [fmesher::fm_int()] for 'cp' family models;
    #' \describe{
    bru_int_args = list(
      #' \item{method}{
      #' "stable" or "direct". For "stable" (default) integration points are
      #' aggregated to mesh vertices.}
      method = "stable",
      #' \item{nsub1}{Number of integration points per knot interval in 1D.
      #'   Default 30.}
      nsub1 = 30,
      #' \item{nsub2}{Number of integration points along a triangle edge for 2D.
      #'   Default 9.}
      nsub2 = 9
      #' \item{nsub}{
      #'   Deprecated parameter that overrides `nsub1` and `nsub2` if set.
      #'   Default `<not set>`.}
      #' }
      #' }
    ),
    #' \item{bru_method}{
    #' List of arguments controlling the iterative inlabru method:
    #' \describe{
    bru_method = list(
      #' \item{taylor}{'pandemic'
      #' (default, from version 2.1.15).}
      taylor = "pandemic",
      #' \item{search}{Either 'all' (default), to use all available line search
      #' methods, or one or more of
      #' \describe{
      #' \item{'finite'}{(reduce step size until predictor is finite)}
      #' \item{'contract'}{(decrease step size until trust hypersphere reached)}
      #' \item{'expand'}{(increase step size until no improvement)}
      #' \item{'optimise'}{(fast approximate error norm minimisation)}
      #' }
      #' To disable line search, set to an empty vector. Line search is not
      #' available for `taylor="legacy"`.}
      search = "all",
      #' \item{factor}{
      #' Numeric, \eqn{> 1} determining the line search step scaling multiplier.
      #' Default \eqn{(1 + \sqrt{5})/2}{(1+sqrt(5))/2}.}
      factor = (1 + sqrt(5)) / 2,
      #' \item{rel_tol}{
      #' Stop the iterations when the largest change in linearisation point (the
      #' conditional latent state mode) in relation to the estimated posterior
      #' standard deviation is less than `rel_tol`. Default 0.1 (ten percent).}
      rel_tol = 0.1,
      #' \item{max_step}{
      #' The largest allowed line search step factor. Factor 1 is the full INLA
      #' step. Default is 2.}
      max_step = 2,
      #' \item{line_opt_method}{
      #' Which method to use for the line search optimisation step. Default
      #' "onestep", using a quadratic approximation based on the value and
      #' gradient at zero, and the value at the current best step length guess.
      #' The method "full" does line optimisation on the full nonlinear
      #' predictor; this is slow and intended for debugging purposes only.}
      line_opt_method = "onestep"
      #' }
      #' }
    ),
    #' \item{bru_compress_cp}{logical; when `TRUE`, compress the
    #' \eqn{\sum_{i=1}^n \eta_i}{sum_i=1^n eta_i}
    #' part of the Poisson process likelihood (`family = "cp"`) into
    #' either a single term,
    #' with \eqn{y=n}{y=n}, and predictor `mean(eta)`, or a blockwise version of
    #' this. Default: `FALSE` (was `TRUE` prior to version `2.13.0.9031`.)}
    bru_compress_cp = FALSE,
    #' \item{bru_debug}{
    #' logical; when `TRUE`, activate temporary debug features for package
    #' development. Default: `FALSE`}
    bru_debug = FALSE,
    #' \item{bru_compat_pre_2_14_enable}{
    #' logical; when `TRUE`, enable compatibility features for inlabru versions
    #' prior to 2.14. Set to `FALSE` to test external package compatibility
    #' updates. Default: `TRUE` before version 2.14, and will be set to `FALSE`
    #' by default in a later version.}
    bru_compat_pre_2_14_enable = TRUE,
    # inla options
    #' \item{`inla()` options}{
    #' All options not starting with `bru_` are passed on to `inla()`, sometimes
    #' after altering according to the needs of the inlabru method.
    #'
    #' Warning:
    #'   Due to how inlabru currently constructs the `inla()` call, the `mean`,
    #'   `prec`, `mean.intercept`, and `prec.intercept` settings in
    #'   `control.fixed` will have no effect. Until a more elegant alternative
    #'   has been implemented, use explicit `mean.linear` and `prec.linear`
    #'   specifications in each `model="linear"` component instead.
    #'
    #'   The following `inla()` options have inlabru specific defaults:
    #'   \describe{
    #'     \item{`E`}{Default `1`.}
    E = 1,
    #'     \item{`Ntrials`}{Default `1L`.}
    Ntrials = 1L,
    #'     \item{`control.compute`}{
    #'       Default `list(config = TRUE, control.gcpo = list())`.}
    control.compute = list(
      config = TRUE,
      control.gcpo = list()
    ),
    #'     \item{`control.inla`}{Default `list(int.strategy = "auto")`.}
    control.inla = list(int.strategy = "auto"),
    #'     \item{`control.fixed`}{
    #'       Default `list(expand.factor.strategy = "inla")`.}
    control.fixed = list(expand.factor.strategy = "inla")
    #'   }
    #' }
    #' }
  )
}


bru_options_deprecated <- function(args) {
  handle_args <- function(args, replacements) {
    names_args <- names(args)
    deprecated_args <- replacements[names(replacements) %in% names_args]
    depr_list <- vapply(deprecated_args, is.list, TRUE)
    if (any(depr_list)) {
      for (k in which(depr_list)) {
        args[[names(deprecated_args[k])]] <-
          handle_args(args[[names(deprecated_args[k])]], deprecated_args[[k]])
      }
    }
    deprecated_args <- deprecated_args[!depr_list]
    if (any(nzchar(names(deprecated_args)) == 0)) {
      warning(glue(
        "Ignoring deprecated global options ",
        glue_collapse(
          glue("'{depr}'",
            depr =
              names(deprecated_args)[nzchar(names(deprecated_args)) == 0]
          ),
          sep = ", "
        ),
        ".",
      ))
      names_args <- setdiff(
        names_args,
        names(deprecated_args)[nzchar(names(deprecated_args)) == 0]
      )
      deprecated_args <- deprecated_args[nzchar(names(deprecated_args)) > 0]
      args <- args[names_args]
    }
    if (length(deprecated_args) > 0) {
      warning(paste0(
        "Converting deprecated global option(s) ",
        glue_collapse(glue("'{names(deprecated_args)}'"), sep = ", "),
        " to new option(s) ",
        glue_collapse(glue("'{deprecated_args}'"), sep = ", "),
        "."
      ))
      names_args <- names(args)
      names_args[names_args %in% names(deprecated_args)] <-
        deprecated_args[names_args[names_args %in% names(deprecated_args)]]
      names(args) <- names_args
    }
    args
  }

  stopifnot(inherits(args, "bru_options"))
  cl <- class(args)
  deprecated_args <- list(
    mesh = "",
    run = "bru_run",
    max.iter = "bru_max_iter",
    result = "bru_initial",
    bru_result = "bru_initial",
    int.args = "bru_int_args",
    bru_method = list(stop_at_max_rel_deviation = "rel_tol")
  )
  args <- handle_args(args, deprecated_args)
  class(args) <- cl
  args
}


#' Additional bru options
#'
#' Construct a `bru_options` object including the default and global options,
#' and converting deprecated option names.
#'
#' @param \dots Options passed on to [as.bru_options()]
#'
#' @export
#' @returns A `bru_options` object
#'
#' @author Finn Lindgren \email{finn.lindgren@@gmail.com}
#'
#' @examples
#' \donttest{
#'
#' opts <- bru_call_options()
#'
#' # Print them:
#' opts
#' }
#'
#' @keywords internal
bru_call_options <- function(...) {
  opt <- as.bru_options(...)
  opt <- bru_options_deprecated(opt)
  opt <- bru_options(bru_options_get(), opt)
  bru_options_check(opt)
  opt
}


# Extract bru options
bru_options_bru <- function(options) {
  stopifnot(inherits(options, "bru_options"))
  cl <- class(options)
  options <- options[grepl("^bru_", names(options))]
  class(options) <- cl
  options
}

# Extract non-bru options
bru_options_inla <- function(options) {
  stopifnot(inherits(options, "bru_options"))
  cl <- class(options)
  options <- options[!grepl("^bru_", names(options))]
  class(options) <- cl
  options
}


#' @describeIn bru_options Checks for valid contents of a `bru_options`
#' object, and produces warnings for invalid options.
#' @param options An `bru_options` object to be checked
#' @param ignore_null Ignore missing or NULL options.
#' @return `bru_options_check()` returns a `logical`; `TRUE` if the object
#'   contains valid options for use by other functions
#' @examples
#' \dontrun{
#' if (interactive()) {
#'   bru_options_check(bru_options(bru_max_iter = "text"))
#' }
#' }
#' @export

bru_options_check <- function(options, ignore_null = TRUE) {
  options <- as.bru_options(options)
  ok <- TRUE
  are_null <- vapply(options, is.null, TRUE)
  if (!ignore_null) {
    disallowed_null <-
      intersect(
        names(options),
        c("bru_max_iter")
      )
    disallowed_null <- disallowed_null[are_null[disallowed_null]]
    if (length(disallowed_null) > 0) {
      warning(paste0(
        glue_collapse(glue("'{disallowed_null}'"), sep = ", "),
        " should not be set to NULL."
      ))
    }
  }
  for (name in names(options)[!are_null]) {
    # Check valid max_iter
    opt <- options[[name]]
    if (name == "bru_max_iter") {
      if (!is.numeric(opt) || opt <= 0) {
        ok <- FALSE
        warning("'bru_max_iter' should be a positive integer.")
      }
    }
  }

  ok
}


#' @param name Either `NULL`, or single option name string, or character vector
#'  or list with option names,
#'   Default: NULL
#' @param include_default logical; If `TRUE`, the default options are included
#'   together with the global override options. Default: `TRUE`
#' @return `bru_options_get` returns either an [`bru_options`] object, for
#'   `name == NULL`, the contents of single option, if `name` is a options name
#'   string, or a named list of option contents, if `name` is a list of option
#'   name strings.
#'
#' @examples
#' bru_options_get("bru_verbose")
#' @export
#' @describeIn bru_options Used to access global package options.

bru_options_get <- function(name = NULL, include_default = TRUE) {
  if (include_default) {
    default <- bru_options_default()
  } else {
    default <- bru_options()
  }
  global <- bru_env_get()$options
  options <- bru_options(default, global)
  if (is.null(name)) {
    return(options)
  }
  if (is.list(name)) {
    mget(unlist(name), as.environment(options))
  } else {
    options[[name]]
  }
}


#' @describeIn bru_options Used to set global package options.
#' @return `bru_options_set()` returns a copy of the global override options,
#' invisibly (as `bru_options_get(include_default = FALSE)`).
#' @seealso [bru_options()], [bru_options_default()], [bru_options_get()]
#' @param .reset For `bru_options_set`, logical indicating if the global
#'   override options list should be emptied before setting the new option(s).
#'
#' @examples
#' \dontrun{
#' if (interactive()) {
#'   bru_options_set(
#'     bru_verbose = TRUE,
#'     verbose = TRUE
#'   )
#' }
#' }
#' @export

bru_options_set <- function(..., .reset = FALSE) {
  envir <- bru_env_get()
  if (.reset) {
    envir$options <- bru_options(...)
  } else {
    envir$options <- bru_options(envir$options, ...)
  }
  invisible(bru_options_get(include_default = FALSE))
}

#' @describeIn bru_options Clears the global option overrides.
#' @export

bru_options_reset <- function() {
  envir <- bru_env_get()
  envir$options <- bru_options()
  invisible(bru_options_get(include_default = FALSE))
}


#' @describeIn bru_options Sets local option overrides, that are
#'   automatically reset using `withr::defer()`.
#' @param .envir The environment in which to set the options.
#'   Default: `parent.frame()`
#' @examples
#' my_fun <- function(val) {
#'   bru_options_set_local(bru_verbose = val)
#'   bru_options_get("bru_verbose")
#' }
#' # Inside the function, the bru_verbose option is changed.
#' # Outside the function, the bru_verbose option is unchanged.
#' print(my_fun(TRUE))
#' print(bru_options_get("bru_verbose"))
#' print(my_fun(FALSE))
#' print(bru_options_get("bru_verbose"))
#' @export
bru_options_set_local <- function(...,
                                  .reset = FALSE,
                                  .envir = parent.frame()) {
  original_options <- bru_options_get(include_default = FALSE)
  withr::defer(bru_options_set(original_options, .reset = TRUE), envir = .envir)
  invisible(bru_options_set(..., .reset = .reset))
}


#' @title Print inlabru options
#' @param object A [bru_options] object to be summarised
#' @param x A `summary_bru_options` object to be printed
#' @param legend logical; If `TRUE`, include explanatory text, Default: `TRUE`
#' @param include_global logical; If `TRUE`, include global override options
#' @param include_default logical; If `TRUE`, include default options
#' @param \dots Further parameters, currently ignored
#'
#' @examples
#' if (interactive()) {
#'   options <- bru_options(verbose = TRUE)
#'
#'   # Don't print options only set in default:
#'   print(options, include_default = FALSE)
#'
#'   # Only include options set in the object:
#'   print(options, include_default = FALSE, include_global = FALSE)
#' }
#' @method summary bru_options
#' @export
#' @rdname summary.bru_options
summary.bru_options <- function(object,
                                legend = TRUE,
                                include_global = TRUE,
                                include_default = TRUE,
                                ...) {
  traverse <- function(combined, default, global, object) {
    result <- list()
    for (name in sort(names(combined))) {
      if (is.list(combined[[name]])) {
        result[[name]] <- list(
          is_list = TRUE,
          value = traverse(
            combined[[name]], default[[name]],
            global[[name]], object[[name]]
          )
        )
      } else {
        result[[name]] <- list(
          is_list = FALSE,
          value = if (is.null(combined[[name]])) {
            "NULL"
          } else {
            combined[[name]]
          },
          origin =
            if (
              !is.null(default[[name]]) &&
                identical(default[[name]], combined[[name]])
            ) {
              "default"
            } else if (
              !is.null(global[[name]]) &&
                identical(global[[name]], combined[[name]])
            ) {
              "global"
            } else if (
              !is.null(object[[name]]) &&
                identical(object[[name]], combined[[name]])
            ) {
              "user"
            } else {
              "unknown"
            }
        )
      }
    }
    result
  }
  if (include_default) {
    default <- bru_options_default()
  } else {
    default <- bru_options()
  }
  if (include_global) {
    global <- bru_options_get(include_default = FALSE)
  } else {
    global <- bru_options()
  }
  combined <- bru_options(default, global, object)

  if (legend) {
    legend <- c(
      "default = value from the default options",
      "global  = value from the global override object",
      "user    = value from the user override object"
    )
  } else {
    legend <- NULL
  }
  result <- structure(
    list(
      legend = legend,
      value = traverse(combined, default, global, object)
    ),
    class = "summary_bru_options"
  )
  result
}

#' @export
#' @rdname summary.bru_options
print.summary_bru_options <- function(x, ...) {
  traverse <- function(tree, prefix = "") {
    for (name in sort(names(tree))) {
      if (tree[[name]]$is_list) {
        cat(glue("{prefix}{name} ="), "\n")
        traverse(
          tree[[name]]$value,
          prefix = glue("{prefix}\t")
        )
      } else {
        cat(glue(
          "{prefix}{name} =",
          "\t{tree[[name]]$value}",
          "\t({tree[[name]]$origin})"
        ), "\n")
      }
    }
  }

  if (!is.null(x[["legend"]])) {
    cat("Legend:\n")
    cat(glue_collapse("  {x[['legend']]}", sep = "\n"))
  }
  cat("Options for inlabru:\n")
  traverse(x[["value"]], prefix = "  ")
  invisible(x)
}

test_that("bru_log", {
  bru_options_set_local(
    # Show messages up to and including level 2 (default 0)
    bru_verbose = 2,
    # Store messages to an including level 3 (default Inf, storing all)
    bru_verbose_store = 3
  )

  suppressMessages({
    bru_log_bookmark("bookmark 1")
    bru_log_message("Test message 1", verbosity = 1)
    bru_log_message("Test message 2", verbosity = 2)
    bru_log_bookmark("bookmark 2")
    bru_log_message("Test message 3", verbosity = 3)
    bru_log_message("Test message 4", verbosity = 4)
  })

  expect_length(bru_log()["bookmark 1"], 2L)
  expect_length(bru_log()["bookmark 2"], 1L)
})

#' Render objects using RGL
#'
#' `glplot()` is a generic function for renders various kinds of spatial
#' objects, i.e. `Spatial*` data and `fm_mesh_2d` objects. The function invokes
#' particular methods which depend on the class of the first argument.
#'
#' @name glplot
#' @export
#' @param object an object used to select a method.
#' @param \dots further arguments passed to or from other methods.
#'
#' @examples
#' \donttest{
#' if (interactive() &&
#'   require("rgl", quietly = TRUE) &&
#'   require("sphereplot", quietly = TRUE) &&
#'   bru_safe_sp() &&
#'   require("sp")) {
#'   # Show the globe:
#'   globe()
#'
#'   # Load pantropoical dolphin data
#'   mexdolphin <- inlabru::mexdolphin_sp()
#'
#'   # Add mesh, ship transects and dolphin sightings stored
#'   # as fm_mesh_2d, SpatialLines and SpatialPoints objects, respectively
#'
#'   glplot(mexdolphin$mesh, alpha = 0.2)
#'   glplot(mexdolphin$samplers, lwd = 5)
#'   glplot(mexdolphin$points, size = 10)
#' }
#' }
#'
glplot <- function(object, ...) {
  UseMethod("glplot")
}

#' @describeIn glplot This function will calculate the cartesian coordinates of
#'   the points provided and use points3d() in order to render them.
#'
#' @export
#'
#' @param add If TRUE, add the points to an existing plot. If FALSE, create new
#'   plot.
#' @param color vector of R color characters. See material3d() for details.


glplot.SpatialPoints <- function(object, add = TRUE, color = "red", ...) {
  if (length(sp::coordnames(object)) < 3) {
    ll <- data.frame(object)
    ll$TMP.ZCOORD <- 0
    sp::coordinates(ll) <- c(sp::coordnames(object), "TMP.ZCOORD")
    sp::proj4string(ll) <- fm_CRS(object)
    object <- ll
  }

  object <- fm_transform(object, crs = fm_crs("sphere"))
  cc <- sp::coordinates(object)
  requireNamespace("rgl")
  rgl::points3d(
    x = cc[, 1],
    y = cc[, 2],
    z = cc[, 3],
    add = add,
    color = color,
    ...
  )
}

#' @describeIn glplot This function will calculate a cartesian representation of
#'   the lines provided and use `lines3d()` in order to render them.
#'
#' @export

glplot.SpatialLines <- function(object, add = TRUE, ...) {
  qq <- sp::coordinates(object)
  sp <- do.call(rbind, lapply(qq, function(k) {
    do.call(rbind, lapply(k, function(x) {
      x[1:(nrow(x) - 1), ]
    }))
  }))
  ep <- do.call(rbind, lapply(qq, function(k) {
    do.call(rbind, lapply(k, function(x) {
      x[2:(nrow(x)), ]
    }))
  }))
  sp <- data.frame(x = sp[, 1], y = sp[, 2], z = 0)
  ep <- data.frame(x = ep[, 1], y = ep[, 2], z = 0)

  sp::coordinates(sp) <- c("x", "y", "z")
  sp::coordinates(ep) <- c("x", "y", "z")
  sp::proj4string(sp) <- fm_CRS(object)
  sp::proj4string(ep) <- fm_CRS(object)

  sp <- fm_transform(sp, crs = fm_crs("sphere"))
  ep <- fm_transform(ep, crs = fm_crs("sphere"))

  cs <- sp::coordinates(sp)
  ce <- sp::coordinates(ep)
  na <- matrix(NA, ncol = 3, nrow = nrow(cs))

  mm <- matrix(t(cbind(cs, ce, na)),
    ncol = 3,
    nrow = 3 * nrow(ce),
    byrow = TRUE
  )

  requireNamespace("rgl")

  rgl::lines3d(mm, add = add, ...)
}


#' @describeIn glplot This function transforms the mesh to 3D cartesian
#'   coordinates and uses [fmesher::plot_rgl()] to plot the result.
#'
#' @export
#'
#' @param col Color specification. A single named color, a vector of scalar
#'   values, or a matrix of RGB values.
#' @param \dots Parameters passed on to `plot_rgl.fm_mesh_2d()`

glplot.fm_mesh_2d <- function(object, add = TRUE, col = NULL, ...) {
  if (object$manifold == "S2") {
    # mesh$loc = mesh$loc
  } else {
    object <- fm_transform(object, crs = fm_crs("sphere"), passthrough = TRUE)
  }

  if (is.null(col)) {
    fmesher::plot_rgl(object, add = add, ...)
  } else {
    fmesher::plot_rgl(object, add = add, col = col, ...)
  }
}

#' @describeIn glplot Visualize a globe using RGL
#'
#' Creates a textured sphere and lon/lat coordinate annotations.
#' This function requires the `rgl` and `sphereplot` packages.
#'
#' @export
#' @param R Radius of the globe
#' @param R.grid Radius of the annotation sphere.
#' @param specular Light color of specular effect.
#' @param axes If TRUE, plot x, y and z axes.
#' @param box If TRUE, plot a box around the globe.
#' @param xlab,ylab,zlab Axes labels
globe <- function(R = 1,
                  R.grid = 1.05,
                  specular = "black",
                  axes = FALSE,
                  box = FALSE,
                  xlab = "", ylab = "", zlab = "") {
  # coordinates for texture
  n.smp <- 50
  lat <- matrix(-asin(seq(-1, 1, length.out = n.smp)),
    n.smp,
    n.smp,
    byrow = TRUE
  )
  long <- matrix(seq(-180, 180, length.out = n.smp) * pi / 180, n.smp, n.smp)
  x <- R * cos(lat) * cos(long)
  y <- R * cos(lat) * sin(long)
  z <- R * sin(lat)

  # globe and texture
  requireNamespace("rgl")
  rgl::persp3d(x, y, z,
    col = "white",
    texture = system.file("misc", "Lambert_ocean.png", package = "inlabru"),
    specular = "black",
    axes = axes,
    box = box,
    xlab = xlab,
    ylab = ylab,
    zlab = zlab,
    normal_x = x,
    normal_y = y,
    normal_z = z
  )

  # spheric grid
  requireNamespace("sphereplot")
  sphereplot::rgl.sphgrid(longtype = "D", add = TRUE, radius = R.grid)
}

---
title: "Prediction scores"
author: "Finn Lindgren"
date: "Corrections and code updates 2025-04-09"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Prediction scores}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
header-includes:
  - \newcommand{\bm}[1]{\boldsymbol{#1}}
  - \newcommand{\wh}[1]{\widehat{#1}}
  - \newcommand{\ol}[1]{\overline{#1}}
  - \newcommand{\wt}[1]{\widetilde{#1}}
  - \newcommand{\pP}{\mathbb{P}}
  - \newcommand{\pE}{\mathbb{E}}
  - \newcommand{\pVar}{\mathbb{V}}
  - \newcommand{\pCov}{\mathbb{C}}
  - \newcommand{\mmd}{\mathrm{d}}
  - \newcommand{\md}{\,\mmd}
  - \newcommand{\e}{\mathrm{e}}
  - \newcommand{\RR}{\mathbb{R}}
  - \DeclareMathOperator*{\argmin}{arg\,min}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(inlabru)
inla_available <- inlabru::bru_safe_inla(
  multicore = interactive(),
  quietly = TRUE
)
```


```{r setup, include=TRUE, eval=inla_available}
library(dplyr)
library(scoringRules)
```

## Proper posterior prediction scores

A prediction **score** $S(F,y)$ evaluates some measure of **closeness** between
a prediction distribution identified by $F$, and an observed value $y$.

### A basic score, and motivating remarks

A common score is the **Squared Error**,
$$
S_\text{SE}(F,y) = [y - \pE_F(Y)]^2,
$$
and we would like predictions to have low values for $S_\text{SE}(F,y)$, indicating a "good"
prediction, in that specific sense that puts a penalty on the squared deviation of the prediction mean from the true observed value.  We can imagine constructing other such scoring functions
that penalise other aspects of the prediction.

A score where "lower is better" is called negatively oriented, and a score where "higher is better"
is called positively oriented.  One can always turn one type of score into the other
by changing the sign, so to simplify the presentation, we'll make all scores be negatively oriented, like the squared error.

We often care about the prediction uncertainty and not just the mean (or at least we _should_ care!).
Just adding a prediction variance penalty to the squared error wouldn't be useful,
as we could then construct a new, "better", prediction by reducing the stated
prediction variance to zero. This would understate the real prediction uncertainty,
so wouldn't be a fair scoring approach for comparing different prediction models.
In the next section, we make this fairness idea more precise.

### Proper and strictly proper scores

The expected value under a distribution identified by $G$ is denoted
$S(F,G):=\pE_{Y\sim F}[S(F,Y)]$.   For a negatively oriented score, we seek scoring
functions that are **fair**, in the sense that one cannot, on average, make a better
prediction that that which generated the data.  This requires $S(F,G)\geq S(G,G)$
for all predictive distributions $F$ and any distribution $G$.
Such scores are call **proper**.  If in addition, equality of the score expectations
only hold when $F=G$, the score is **strictly proper**.

Non-strict proper scores ignore some aspect of the prediction, typically by only being
sensitive to some summary information, such as the mean, median, and/or variance.

It's notable that proper scores retain their properness under affine transformations,
with just potential changes in whether they are positively or negatively oriented.
If $S(F,y)$ is a proper score, then
$$
S'(F,y) = a + b S(F,y),\quad a,b\in\mathbb{R},
$$
is also a proper score, with the same orientation if $b>0$ and the opposite
orientation if $b<0$. The degenerate case $b=0$ gives the score $a$ to all predictions,
which is technically a proper score (you cannot to better than an ideal prediction),
but a useless one (an ideal prediction is no better than any other prediction).

### Examples

* log-score: $S_\text{log}(F,y) = -\log\{p_F(y)\}$, where $p_F(y)$ is
  a predictive pdf or pmf for $y$, is a strictly proper score.
* Squared Error: $S_\text{SE}(F,y) = [y - \pE_F(Y)]^2$ is a proper score
* Brier score:
  - Binary events: $$S_\text{Brier}(F,z) = [z - \pP_F(Z = 1)]^2,$$ where $Z\in\{0,1\}$ is a binary
    event indicator, is a strictly proper score for the event prediction, but
    non-strict with respect to any underlying outcome $y$ generating the event indicator,
    e.g. via $z=I(y=0)$.
  - Class indicator events: if $y\in\{1,\dots,K\}$ is a class category outcome, the
    Brier score can be generalised to
    $$S_\text{MultiBrier}(F,y) = \sum_{k=1}^K [I(y = k) - \pP_F(Y=k)]^2$$
    which can be seen as the Squared Error for the Multinomial prediction model
    for $z_k=I(y=k)$, with $\{z_1,\dots,z_K\}\sim\text{Multinomial}(1,\{p_1,\dots,p_K\})$,
    where $p_k=\pP_F(Y=k)$.
    Sometimes,
    the sum is normalised by $1/K$. This generalised Brier score is a proper score.
* Dawid-Sebastiani: 
  $$S_\text{DS}(F,y)=[y - \pE_F(Y)]^2 / \pVar_F(Y) + \log[\pVar_F(Y)]$$
  is a proper score.
  It's derived from the strictly proper log-score of a Gaussian prediction,
  but it's also a non-strict proper score for other distributions. It has
  the advantage that it only involves the predictive mean and variance, making
  it computable also in cases when log-densities are hard to obtain.
  Since it's based on the symmetric Gaussian distribution, it tends to be affected
  by skewness, so should be applied with care in such cases.
* Absolute (Median) Error: $S_\text{AE}(F,y)=|y - \text{median}_F|$ is a proper score,
  with expectation minimised when the medians of $F$ and $G$ match.
  Note that $|y-\pE_F(Y)|$, the absolute error with respect to the _expectation_,
  is _not_ a proper score!  Another way of expressing this is that if $|y-m_F|$
  is a proper score _with respect to the median_, i.e. it is proper when $m_F$
  is taken to be the median of $F$, and not some other point prediction. Unfortunately,
  the predictive median is not readily available, and typically requires sampling
  from the observation level predictive distribution, or estimating the median
  of the a mixture distribution.
  In the applied literature, this distinction is often overlooked, and the predictive
  mean is inserted into both the SE and AE scores, making the resulting AE score
  comparisons less clear than they could be. 
* CRPS (Continuous Ranked Probability Score):
  
  $$
  S_\text{CRPS}(F,y)=
    \int_{-\infty}^\infty [\pP_F(Y \leq x) - I(y \leq x)]^2 \md x
  $$
  This is a strictly proper score, related to the absolute error of point predictions.
    
Other scores include the Interval score that is minimised for short prediction intervals with the intended coverage probability, and the Quantile score, that generalises the Absolute Median Error
to other quantiles than the median.

### Improper scores

We've seen that some scores are _strictly_ proper, and others are only proper scores,
sensitive to specific aspects of the predictive distribution, such as mean, median, and/or variance.

In contrast, *improper* scores do not fulfil the fairness idea. Such scores
include the the aforementioned penalised squared error,
$[y-\pE(Y)]^2+\pVar_F(Y)$, but also the probability/density function, $p_F(y)$.
The latter might come as a surprise, as the log-score _is_ proper.

### Mean error/score

Up to this point, we only considered individual scores.  When summarising
predictions $\{F_i\}$ for a collection of observations $\{y_i\}$, we usually
compute the mean score,
$$
S(\{F_i\},\{y_i\}) = \frac{1}{N}\sum_{i=1}^N S(F_i,y_i) .
$$

When comparing two different prediction models $F$ and $F'$, the scores are dependent
with respect to the observations $y_i$. This means that in order to more easily
handle the score variability in the comparison, we should treat it as a paired
sample problem.
The pairwise score differences are given by
$$
S(F_i,F'_i,y_i) = S(F_i,y_i) - S(F'_i,y_i) .
$$
It's also much more reasonable to make conditional independence assumptions about
these differences, than for the plain score values $S(F_i,y_i)$;
$$
\pVar_{\{y_i\}\sim G}\left[\frac{1}{N}\sum_{i=1}^N S(F_i,y_i)\right] =
\frac{1}{N^2}\sum_{i=1}^N\sum_{j=1}^N\pCov_{\{y_i\}\sim G}\left[S(F_i,y_i),S(F_j,y_j)\right]
$$
but
$$
\pVar_{\{y_i\}\sim G}\left[\frac{1}{N}\sum_{i=1}^N S(F_i,y_i) - S(F'_i,y_i)\right] \approx
\frac{1}{N^2}\sum_{i=1}^N\pVar_{y_i\sim G_i}\left[ S(F_i,y_i) - S(F'_i,y_i)\right].
$$

Note that taking the average of prediction scores, or averages of prediction score differences,
is quite different from assessing summary statistics of the collection of predictions,
since the scores are individual for each observation; we're not assessing the
collective value distribution, as that might be misleading. 
For example, consider a spatial model where he estimated procession has
an empirical distribution of the predictive means that matches that of the
observed data. Scores based on the marginal empirical distribution would
not be able to detect if the _location_ of the values is maximally different
to the actual locations, whereas averages of individual scores would be sensitive to
this.

## Poisson model example

Consider a model with Poisson outcomes $y$, conditionally on a
log-linear predictor $\lambda=\exp(\eta)$, where $\eta$ is some
linear expression in latent variables.

The posterior predictive distributions are Poisson mixture distributions
across the posterior distribution of $\lambda$, $p(\lambda|\text{data})$.

For illustration purposes, consider a simple Poisson model with a single
covariate $x$ and a linear predictor $\eta = \beta_0 + \beta_1 x$.
```{r, eval=inla_available}
df_pois <- tibble::tibble(
  x = rnorm(50),
  y = rpois(length(x), exp(2 + 1 * x))
)
fit_pois <- bru(
  y ~ Intercept(1) + x,
  family = "poisson",
  data = df_pois
)
newdata_pois <- tibble::tibble(
  x = rnorm(100),
  y = rpois(100, exp(2 + 1 * x))
)
```

### Moment scores

For the Squared Error and Dawid-Sebastiani scores, we'll need
the posterior expectation and variance:
$$
  \pE(Y|\text{data}) = \pE[\pE(Y|\lambda,\text{data}) | \text{data}] = \pE(\lambda | \text{data})
$$
and
$$
  \pVar(Y|\text{data}) = \pE[\pVar(Y|\lambda,\text{data}) | \text{data}] +
    \pVar[\pE(Y | \lambda, \text{data}) | \text{data}]
     = \pE[\lambda | \text{data}] + \pVar[\lambda | \text{data}]
$$
i.e. the sum of the posterior mean and variance for lambda.

The SE and DS scores are therefore relatively easy to compute after estimating
a model with `inlabru`. You just need
to estimate the posterior mean and variance with `predict()` for each
test data point. If `Intercept+x` is the expression for the linear predictor,
and `newdata`
holds the covariate information for the prediction points, run
```{r, eval=inla_available}
pred_pois <- predict(
  fit_pois,
  newdata_pois,
  formula = ~ exp(Intercept + x),
  n.samples = 2000
)
post_E <- pred_pois$mean
post_Var <- pred_pois$mean + pred_pois$sd^2
scores <- data.frame(
  SE = (newdata_pois$y - post_E)^2,
  DS = (newdata_pois$y - post_E)^2 / post_Var + log(post_Var)
)
summary(scores)
```

### Log-Probability and log-density scores

The full log-score can actually also be estimated/computed in a
similar way. We seek, for a fixed observation y,
  $\log[\pP(Y = y | \text{data})]$
The probability is
$$
  \pP(Y = y | \text{data}) = \pE[\pP(Y = y | \lambda, \text{data}) | \text{data}],
$$
so we can estimate it using `predict()`:
```{r, eval=inla_available}
pred_pois <- predict(fit_pois,
  newdata_pois,
  formula = ~ dpois(y, lambda = exp(Intercept + x)),
  n.samples = 2000
)
log_score <- -log(pred_pois$mean)
summary(log_score)
```
to estimate the log_score (increase `n.samples` if needed for
sufficiently small Monte Carlo error). Approximate Monte Carlo confidence
intervals for the scores can be obtained via the Monte Carlo standard errors:
```{r, eval=inla_available}
head(
  data.frame(
    log_S = log_score,
    lower = -log(pred_pois$mean + 2 * pred_pois$mean.mc_std_err),
    upper = -log(pred_pois$mean - 2 * pred_pois$mean.mc_std_err)
  )
)
```

### CRPS

Yet another option would be to use the CRPS, which
for each prediction value $y_i$ would be
$$
S_\text{CRPS}(F_i,y_i) = \sum_{k=0}^\infty [\pP(Y_i \leq k |\text{data}) - I(y_i \leq k)]^2
$$
For this, one would first need to get $\pP(Y \leq k | \text{data})$ from a
predict call with `ppois(k, lambda = exp(eta))`, for
a vector $k=0,1\dots,K$, for each $y_i$, for some sufficiently large $K>\max_i{y_i}$
for the remainder to be negligible.
However, to avoid repeated `predict()` calls for each $y_i$,
the storage requirements is of order $(K+1) \times N\times N_\text{samples}$.
To avoid that, one option would be to reformulate the estimator into a recursive
estimator, so that batches of simulations could be used to iteratively compute the estimator.

A basic estimator can proceed as follows:

Define $K\geq K_0=\max_i(y_i)$ sufficiently large for the posterior predictive probability
above $K$ to be negligible. Perhaps a value like $K=K_0+4\sqrt{K_0}$ might be sufficient.
You can check afterwards, and change if needed.

1. Simulate samples from $\lambda^{(j)}\sim p(\lambda|\text{data})$ using `generate()` (size $N\times N_\text{samples}$).
2. For each $i=1,\dots,N$, use the samples to estimate the residuals
$r_{ik}=\pP(Y\leq k|\text{data})-I(y_i\leq k)$,
  for $k\in 0,1,2,\dots,K$, with

  $$
  \wh{r}_{ik} = \frac{1}{N_\text{samples}} \sum_{j=1}^{N_\text{samples}}
  \{
  \pP(Y\leq k|\lambda^{(j)}_i)
  -
  I(y_i\leq k)
  \} .
  $$
3. Compute

  $$
\wh{S}_\text{CRPS}(F_i,y_i) = \sum_{k=0}^{K} \wh{r}_{ik}^2
  $$

```{r, eval=inla_available}
# some large value, so that 1-F(K) is small
max_K <- ceiling(max(newdata_pois$y) + 4 * sqrt(max(newdata_pois$y)))
k <- seq(0, max_K)
kk <- rep(k, times = length(newdata_pois$y))
i <- seq_along(newdata_pois$y)
pred_pois <- generate(fit_pois, newdata_pois,
  formula = ~ {
    lambda <- exp(Intercept + x)
    ppois(kk, lambda = rep(lambda, each = length(k)))
  },
  n.samples = 2000
)
results <- data.frame(
  i = rep(i, each = length(k)),
  k = kk,
  Fpred = rowMeans(pred_pois),
  residuals =
    rowMeans(pred_pois) - (rep(newdata_pois$y, each = length(k)) <= kk)
)
# Check that the cutoff point K has nearly probability mass 1 below it,
# for all i:
min(results |> dplyr::filter(k == max_K) |> pull(Fpred))

crps_scores <-
  (results |>
    group_by(i) |>
    summarise(crps = sum(residuals^2), .groups = "drop") |>
    pull(crps))
summary(crps_scores)
```


## Normal model example

Consider a model with Normal outcomes $y\sim N(\eta,\sigma^2_y)$, conditionally on a
linear predictor $\lambda=\eta$, where $\eta$ is some
linear expression in latent variables.

The posterior predictive distributions are Normal mixture distributions
across the posterior distribution of $\eta$ and $\sigma^2_y$,
$p(\lambda,\sigma^2_y|\text{data})$.

For illustration purposes, consider a simple Poisson model with a single
covariate $x$ and a linear predictor $\eta = \beta_0 + \beta_1 x$.
```{r, eval=inla_available}
true_sigma_y <- 0.1
df_normal <- tibble::tibble(
  x = rnorm(50),
  y = rnorm(length(x), 2 + 1 * x, sd = true_sigma_y)
)
fit_normal <- bru(
  y ~ Intercept(1) + x,
  family = "normal",
  data = df_normal
)
newdata_normal <- tibble::tibble(
  x = rnorm(100),
  y = rnorm(100, 2 + 1 * x, sd = true_sigma_y)
)
```

### Moment scores and log-score

For the Squared Error and Dawid-Sebastiani scores, we'll need
the posterior expectation and variance:
$$
  \pE(Y|\text{data}) = \pE[\pE(Y|\eta, \sigma_y^2,\text{data}) | \text{data}] = \pE(\eta | \text{data})
$$
and
$$
  \pVar(Y|\text{data}) = \pE[\pVar(Y|\eta,\sigma_y^2,\text{data}) | \text{data}] +
    \pVar[\pE(Y | \eta,\sigma_y^2,\text{data}) | \text{data}]
     = \pE[\sigma_y^2 | \text{data}] + \pVar[\eta | \text{data}]
$$
i.e. the sum of the posterior mean of $\sigma_y^2$ and the variance of $\eta$.

The SE and DS scores are therefore relatively easy to compute after estimating
a model with `inlabru`, as well as the log-score. You just need
to estimate the posterior mean and variance with `predict()` for each
test data point. If `Intercept+x` is the expression for the linear predictor,
and `newdata`
holds the covariate information for the prediction points, run
```{r, eval=inla_available}
pred_normal <- predict(fit_normal, newdata_normal,
  formula = ~ {
    E <- Intercept + x
    V <- 1 / Precision_for_the_Gaussian_observations
    list(
      eta = E,
      sigma_y_2 = V,
      dens = dnorm(y, mean = E, sd = sqrt(V))
    )
  },
  n.samples = 2000
)
post_E <- pred_normal$eta$mean
post_Var <- pred_normal$sigma_y_2$mean + pred_normal$eta$sd^2
scores_normal <- data.frame(
  SE = (newdata_normal$y - post_E)^2,
  DS = (newdata_normal$y - post_E)^2 / post_Var + log(post_Var),
  LS = -log(pred_normal$dens$mean)
)
summary(scores_normal)
```
Here, we've again used the negated log-score, so that small values are "better"
for all three scores.
Since the DS score is based on the log-score for Normal predictions, the
_conditional_ DS and LS scores differ only by a constant offset and scaling.
However, since the full posterior predictions are Normal mixtures, the DS and
LS scores for the full posterior predictive distribution are not equivalent.

As for the Poisson case, we can estimate the log-score uncertainty:

```{r, eval=inla_available}
head(
  data.frame(
    log_S = scores_normal$LS,
    lower = -log(pred_normal$dens$mean + 2 * pred_normal$dens$mean.mc_std_err),
    upper = -log(pred_normal$dens$mean - 2 * pred_normal$dens$mean.mc_std_err)
  )
)
```



## Posterior expectation of conditional scores

In some cases, one might be tempted to consider posterior distribution properties
of conditional predictive scores,
e.g. the posterior expectation $\pE_{\lambda|\text{data}}[S(F_\lambda, y)]$ for
$S(F_\lambda, y)$ under the posterior distribution
for $\lambda$ in the Poisson model.

For Squared Error,
$$
\begin{aligned}
\pE_{\lambda|\text{data}}[(y - \lambda)^2] &= 
\pE_{\lambda|\text{data}}[\{y - \pE(\lambda|\text{data}) + \pE(\lambda|\text{data}) - \lambda\}^2] \\
&=
[y - \pE(\lambda|\text{data})]^2 +
\pE_{\lambda|\text{data}}(\{\pE(\lambda|\text{data}) - \lambda\}^2] \\
&=
[y - \pE(\lambda|\text{data})]^2 +
\pVar(\lambda|\text{data}) .
\end{aligned}
$$
It's noteworthy that this is similar to the _improper_ score $[y - \pE(y|\text{data})]^2 +
\pVar(y|\text{data})$, and also in this new case, one can have a model with artificially small posterior variance with a smaller expected score, making this type of construction problematic to interpret.

However, in some cases it does provide alternative approaches for how to compute the
proper scores for the full posterior predictive distributions.
If $\lambda_{ij}$, $j=1,\dots,J$ are samples from the posterior distribution, one score estimator
is
$$
\wh{S}(F_i,y_i) = \left(y_i - \frac{1}{J}\sum_{j=1}^J \lambda_{ij}\right)^2 ,
$$
with the averaging over the samples inside the quadratic expression, and we can
use
```{r, eval=inla_available}
pred_pois <- predict(fit_pois, newdata_pois, formula = ~ exp(Intercept + x))
scores <- (newdata_pois$y - pred_pois$mean)^2
```

If we instead take advantage of the new expression above, we have
$$
[y - \pE(\lambda|\text{data})]^2 
=
\pE_{\lambda|\text{data}}[(y - \lambda)^2]
-
\pVar(\lambda|\text{data})
$$
so the score can be estimated by
```{r, eval=inla_available}
pred_pois <- predict(fit_pois, newdata_pois,
  formula = ~ {
    lambda <- exp(Intercept + x)
    list(
      cond_scores = (y - lambda)^2,
      lambda = lambda
    )
  },
  n.samples = 2000
)
scores <- pred_pois$cond_scores$mean - pred_pois$lambda$sd^2
summary(scores)
```
For this particular case, this approach is unlikely to be an improvement or more
accurate than the basic estimator. Remarkably, this estimator of Squared Error isn't even guaranteed to be positive! However, for other scores there may
potentially be practical benefits.

### An alternative estimator for CRPS
For the CRPS score, there are closed form expressions available for some distributions,
conditionally on their parameters, but not for the full predictive mixture distribution.
We take a similar approach as for SE,
and let $F$ and $F_\lambda$ denote the unconditional and conditional 
cumulative distribution functions for the posterior predictive distribution.
Then $F(x)=\pE_{\lambda|\text{data}}[F_\lambda(x)]$ for all $x$, and
$$
\begin{aligned}
S_\text{CRPS}(F,y) &=
\int_{-\infty}^\infty [F(x) - I(y\leq x)]^2 \md x 
\\
&=
\pE_{\lambda|\text{data}}\left[
\int_{-\infty}^\infty [F_\lambda(x) - I(y\leq x)]^2 \md x 
\right] -
\int_{-\infty}^\infty \left\{\pE_{\lambda|\text{data}}\left[F_\lambda(x)^2\right] 
- \pE_{\lambda|\text{data}}[F_\lambda(x)]^2\right\} \md x 
\\
&=
\pE_{\lambda|\text{data}}\left[
S_\text{CRPS}(F_\lambda,y)
\right] -
\int_{-\infty}^\infty \pVar_{\lambda|\text{data}}[F_\lambda(x)] \md x .
\end{aligned}
$$
Note that we didn't need to use any particular model properties here, so this holds for any
predictive model with mixture structure, when $\lambda$ is the collection
of model parameters.  We also note the resemblance to the alternative expression for
the Squared Error; this is because CRPS can be seen as the integral over all
Brier scores for predicting event indicators of the form $z=I(y\leq x)$, with probability $F(x)$.

In the Poisson case, we can now estimate the CRPS scores like this, that makes
the code a bit easier than the previous version that needed `generate()`.
We can use the Poisson CRPS implementation
`crps_pois()` from the [`scoringRules`](https://cran.r-project.org/package=scoringRules) package.

However, it can be shown that the two approaches have nearly identical Monte Carlo variance,
so the previous version is likely preferable as it doesn't require knowing
a closed form CRPS expression.
```{r, eval=inla_available}
max_K <- 100 # some large value, so that 1-F(K) is small
pred_pois <- predict(fit_pois, newdata_pois,
  formula = ~ {
    lambda <- exp(Intercept + x)
    list(
      crps = scoringRules::crps_pois(y, lambda),
      F = do.call(
        c,
        lapply(
          seq_along(y),
          function(i) {
            ppois(seq(0, max_K), lambda = lambda[i])
          }
        )
      )
    )
  },
  n.samples = 2000
)
crps_score <-
  pred_pois$crps$mean -
  (pred_pois$F |>
    mutate(i = rep(seq_along(newdata_pois$y), each = max_K + 1)) |>
    group_by(i) |>
    summarise(F_var = sum(sd^2), .groups = "drop") |>
    pull(F_var))
summary(crps_score)
```
Formulas and functions for Poisson CRPS, as well as for other distributions, can be found via
the ["Evaluating Probabilistic Forecasts with scoringRules"](https://cran.r-project.org/package=scoringRules/vignettes/article.pdf)
vignette of the `scoringRules` package, and the references in `?scoringRules::crps.numeric`.

---
title: "Nonlinear model approximation"
output:
  rmarkdown::html_vignette:
  rmarkdown::pdf_document:
vignette: >
  %\VignetteIndexEntry{Nonlinear model approximation}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
header-includes:
  - \newcommand{\bm}[1]{\boldsymbol{#1}}
  - \newcommand{\wt}[1]{\widetilde{#1}}
  - \newcommand{\ol}[1]{\overline{#1}}
  - \newcommand{\wh}[1]{\widehat{#1}}
  - \DeclareMathOperator*{\argmax}{arg\,max}
  - \newcommand{\proper}[1]{\mathsf{#1}}
  - \newcommand{\pExp}{\proper{Exp}}
  - \newcommand{\pN}{\proper{N}}
  - \newcommand{\pPo}{\proper{Po}}
  - \newcommand{\pGa}{\proper{Ga}}
  - \newcommand{\pE}{\proper{E}}
  - \newcommand{\pP}{\proper{P}}
  - \newcommand{\pVar}{\proper{Var}}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "png",
  dev.args = list(type = "cairo-png"),
  fig.width = 7,
  fig.height = 5
)
```

```{r setup, include = FALSE}
library(inlabru)
library(ggplot2)
```

See the `method` vignette for details of the general `inlabru` method.
Here, we'll take a look at a small toy example to see the difference between
a few posterior approximation methods.

Note: This vignette is not completed yet.

## A small toy problem

Hierarchical model:
$$
\begin{aligned}
\lambda &\sim \pExp(\gamma) \\
(y_i|\lambda) &\sim \pPo(\lambda), \text{ independent across $i=1,\dots,n$}
\end{aligned}
$$
With $\ol{y}=\frac{1}{n}\sum_{i=1}^n y_i$, the posterior density is
$$
\begin{aligned}
p(\lambda | \{y_i\}) &\propto p(\lambda, y_1,\dots,y_n) \\
&\propto \exp(-\gamma\lambda) \exp(-n\lambda) \lambda^{n\ol{y}} \\
&= \exp\{-(\gamma+n)\lambda\} \lambda^{n\ol{y}},
\end{aligned}
$$
which is the density of a $\pGa(\alpha = 1+n\ol{y}, \beta = \gamma+n)$ distribution.

## Latent Gaussian predictor version

Introducing a latent Gaussian variable $u\sim\pN(0,1)$, the model can be
reformulated as
$$
\begin{aligned}
\lambda(u) &=-\ln\{1-\Phi(u)\}/\gamma \\
(y_i|u) &\sim \pPo(\lambda(u))
\end{aligned}
$$
We will need some derivatives of $\lambda$ with respect to $u$:
$$
\begin{aligned}
\frac{\partial\lambda(u)}{\partial u} &= \frac{1}{\gamma}\frac{\phi(u)}{1-\Phi(u)}
  = \lambda'(u) \\
\frac{\partial^2\lambda(u)}{\partial u^2} &=
-
\frac{1}{\gamma}\frac{\phi(u)}{1-\Phi(u)}
\left(
u + \frac{\phi(u)}{1-\Phi(u)}
\right)
=
-\lambda'(u)\left\{u-\gamma\lambda'(u)\right\}
\\
\frac{\partial\ln\lambda(u)}{\partial u} &=
  \frac{1}{\lambda(u)} \frac{\partial\lambda(u)}{\partial u}
  =\frac{1}{-\ln\{1-\Phi(u)\}} \frac{\phi(u)}{1-\Phi(u)}
  = \frac{\lambda'(u)}{\lambda(u)} \\
\frac{\partial^2\ln\lambda(u)}{\partial u^2} &=
  \frac{1}{\lambda(u)} \frac{\partial^2\lambda(u)}{\partial u^2}
  -\frac{1}{\lambda(u)^2} \left(\frac{\partial\lambda(u)}{\partial u}\right)^2
  = \frac{-\lambda'(u)\{u - \gamma\lambda'(u)\}}{\lambda(u)}
  - \frac{\lambda'(u)^2}{\lambda(u)^2}\\
  &= -\frac{\lambda'(u)}{\lambda(u)}\left\{
  u - \gamma\lambda'(u) +\frac{\lambda'(u)}{\lambda(u)}
  \right\}
\end{aligned}
$$
```{r}
lambda <- function(u, gamma) {
  -pnorm(u, lower.tail = FALSE, log.p = TRUE) / gamma
}
lambda_inv <- function(lambda, gamma) {
  qnorm(-lambda * gamma, lower.tail = FALSE, log.p = TRUE)
}
D1lambda <- function(u, gamma) {
  exp(dnorm(u, log = TRUE) - pnorm(u, lower.tail = FALSE, log.p = TRUE)) / gamma
}
D2lambda <- function(u, gamma) {
  D1L <- D1lambda(u, gamma)
  -D1L * (u - gamma * D1L)
}
D1log_lambda <- function(u, gamma) {
  D1lambda(u, gamma) / lambda(u, gamma)
}
D2log_lambda <- function(u, gamma) {
  D1logL <- D1log_lambda(u, gamma)
  -D1logL * (u - gamma * D1lambda(u, gamma = gamma) + D1logL)
}
```

## Latent Gaussian posterior approximations

A basic approximation of the posterior distribution for $\lambda$ given $y$
can be defined as a deterministic transformation of a Gaussian distribution
obtained from a 2nd order Taylor approximation of $\ln p(u|\{y_i\})$ at the posterior
mode $u_0$ of $p(u|\{y_i\})$. The needed derivatives are
$$
\begin{aligned}
\frac{\partial\ln p(u|\{y_i\})}{\partial u} &= \frac{\partial\ln\phi(u)}{\partial u}
- n\lambda'(u) + n\ol{y}\frac{\lambda'(u)}{\lambda(u)}
= -u + n\frac{\lambda'(u)}{\lambda(u)}\left\{
\ol{y} - \lambda(u)
\right\}
\\
\frac{\partial^2\ln p(u|\{y_i\})}{\partial u^2}
&=
-1
-
n
\frac{\lambda'(u)}{\lambda(u)}\left\{
u - \gamma\lambda'(u) + \frac{\lambda'(u)}{\lambda(u)}
\right\}
\left\{
\ol{y} - \lambda(u)
\right\}
-
n
\frac{\lambda'(u)^2}{\lambda(u)}
\end{aligned}
$$
At the mode $u_0$, the first order derivative is zero, and
$$
\begin{aligned}
\left.\frac{\partial^2\ln p(u|\{y_i\})}{\partial u^2}\right|_{u=u_0}
&=
-1
-
\left\{
u_0 - \gamma\lambda'(u_0) + \frac{\lambda'(u_0)}{\lambda(u_0)}
\right\}
u_0
-
n
\frac{\lambda'(u_0)^2}{\lambda(u_0)} .
\end{aligned}
$$
The quadratic approximation of the log-posterior density at the mode $u_0$ is then
$$
\ln \breve{p}(u|\{y_i\}) = \text{const} - \frac{(u-u_0)^2}{2}
\left[
-
\left.\frac{\partial^2\ln p(u|\{y_i\})}{\partial u^2}\right|_{u=u_0}
\right]
$$
In `inlabru`, the approximation first linearises $\ln \lambda(u)$ at $u_0$ before applying
the Taylor approximation of $\ln p(u|\{y_i\})$. The linearised log-predictor is
$$
\ln \ol{\lambda}(u) = \ln \lambda(u_0) + \frac{\lambda'(u_0)}{\lambda(u_0)}(u - u_0)
$$
so that
$$
\ol{\lambda}'(u) = \frac{\lambda'(u_0)}{\lambda(u_0)} \ol{\lambda}(u)
$$
and the second order derivative of the linearised log-posterior density is
$$
\begin{aligned}
\left.\frac{\partial^2\ln \ol{p}(u|\{y_i\})}{\partial u^2}\right|_{u=u_0}
&=
-1
-
n
\frac{\lambda'(u_0)^2}{\lambda(u_0)} .
\end{aligned}
$$

```{r}
log_p <- function(u, y, gamma) {
  L <- lambda(u, gamma)
  n <- length(y)
  dnorm(u, log = TRUE) - n * L + n * mean(y) * log(L) - sum(lgamma(y + 1))
}
D1log_p <- function(u, y, gamma) {
  n <- length(y)
  -u + n * D1log_lambda(u, gamma) * (mean(y) - lambda(u, gamma))
}
D2log_p <- function(u, y, gamma) {
  n <- length(y)
  -1 +
    n * D2log_lambda(u, gamma) * (mean(y) - lambda(u, gamma)) -
    n * D1log_lambda(u, gamma) * D1lambda(u, gamma)
}
```

```{r}
g <- 1
y <- c(0, 1, 2)
y <- c(0, 0, 0, 0, 0)
y <- rpois(5, 5)
mu_quad <- uniroot(
  D1log_p,
  lambda_inv((1 + sum(y)) / (g + length(y)) * c(1 / 10, 10), gamma = g),
  y = y, gamma = g
)$root
sd_quad <- (-D2log_p(mu_quad, y = y, gamma = g))^-0.5
sd_lin <- (1 + length(y) * D1log_lambda(mu_quad, gamma = g)^2 *
  lambda(mu_quad, gamma = g))^-0.5
lambda0 <- lambda(mu_quad, gamma = g)
```

### Posterior densities


```{r}
ggplot() +
  xlim(
    lambda(mu_quad - 4 * sd_quad, gamma = g),
    lambda(mu_quad + 4 * sd_quad, gamma = g)
  ) +
  labs(x = "lambda", y = "Density") +
  geom_function(
    fun = function(x) {
      exp(log_p(lambda_inv(x, gamma = g), y = y, gamma = g)) /
        D1lambda(lambda_inv(x, gamma = g), gamma = g) / (
          exp(log_p(lambda_inv(lambda0, gamma = g), y = y, gamma = g)) /
            D1lambda(lambda_inv(lambda0, gamma = g), gamma = g)
        ) *
        dgamma(lambda0, shape = 1 + sum(y), rate = g + length(y))
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = dgamma,
    args = list(shape = 1 + sum(y), rate = g + length(y)),
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      dnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_quad) /
        D1lambda(lambda_inv(x, gamma = g), gamma = g)
    },
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      dnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_lin) /
        D1lambda(lambda_inv(x, gamma = g), gamma = g)
    },
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(mapping = aes(
    xintercept = (1 + sum(y)) / (g + length(y)),
    lty = "Bayes mean"
  )) +
  geom_vline(mapping = aes(xintercept = lambda0, lty = "Bayes mode")) +
  geom_vline(mapping = aes(xintercept = mean(y), lty = "Plain mean"))
```
```{r}
ggplot() +
  xlim(
    lambda_inv(lambda0, gamma = g) - 4 * sd_quad,
    lambda_inv(lambda0, gamma = g) + 4 * sd_quad
  ) +
  xlab("u") +
  ylab("Density") +
  geom_function(
    fun = function(x) {
      exp(log_p(x, y = y, gamma = g) -
        log_p(lambda_inv(lambda0, gamma = g), y = y, gamma = g)) *
        (dgamma(lambda0, shape = 1 + sum(y), rate = g + length(y)) *
          D1lambda(lambda_inv(lambda0, gamma = g), gamma = g))
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      dgamma(lambda(x, gamma = g), shape = 1 + sum(y), rate = g + length(y)) *
        D1lambda(x, gamma = g)
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = dnorm,
    args = list(mean = mu_quad, sd = sd_quad),
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = dnorm,
    args = list(mean = mu_quad, sd = sd_lin),
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(
    mapping = aes(
      xintercept = lambda_inv((1 + sum(y)) / (g + length(y)),
        gamma = g
      ),
      lty = "Bayes mean"
    )
  ) +
  geom_vline(
    mapping = aes(
      xintercept = lambda_inv(lambda0, gamma = g),
      lty = "Bayes mode"
    )
  ) +
  geom_vline(
    mapping = aes(
      xintercept = lambda_inv(mean(y), gamma = g),
      lty = "Plain mean"
    )
  )
```

### Posterior CDFs

```{r}
ggplot() +
  xlim(
    lambda(mu_quad - 4 * sd_quad, gamma = g),
    lambda(mu_quad + 4 * sd_quad, gamma = g)
  ) +
  xlab("lambda") +
  ylab("CDF") +
  geom_function(
    fun = pgamma,
    args = list(shape = 1 + sum(y), rate = g + length(y)),
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      pnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_quad)
    },
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = function(x) {
      pnorm(lambda_inv(x, gamma = g), mean = mu_quad, sd = sd_lin)
    },
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(mapping = aes(
    xintercept = (1 + sum(y)) / (g + length(y)),
    lty = "Bayes mean"
  )) +
  geom_vline(mapping = aes(xintercept = lambda0, lty = "Bayes mode")) +
  geom_vline(mapping = aes(xintercept = mean(y), lty = "Plain mean"))
```
```{r}
ggplot() +
  xlim(
    lambda_inv(lambda0, gamma = g) - 4 * sd_quad,
    lambda_inv(lambda0, gamma = g) + 4 * sd_quad
  ) +
  xlab("u") +
  ylab("CDF") +
  geom_function(
    fun = function(x) {
      pgamma(lambda(x, gamma = g), shape = 1 + sum(y), rate = g + length(y))
    },
    mapping = aes(col = "Theory"),
    n = 1000
  ) +
  geom_function(
    fun = pnorm,
    args = list(mean = mu_quad, sd = sd_quad),
    mapping = aes(col = "Quadratic"),
    n = 1000
  ) +
  geom_function(
    fun = pnorm,
    args = list(mean = mu_quad, sd = sd_lin),
    mapping = aes(col = "Linearised"),
    n = 1000
  ) +
  geom_vline(mapping = aes(
    xintercept = lambda_inv((1 + sum(y)) / (g + length(y)),
      gamma = g
    ),
    lty = "Bayes mean"
  )) +
  geom_vline(mapping = aes(
    xintercept = lambda_inv(lambda0, gamma = g),
    lty = "Bayes mode"
  )) +
  geom_vline(mapping = aes(
    xintercept = lambda_inv(mean(y), gamma = g),
    lty = "Plain mean"
  ))
```

---
title: "Classes and methods"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Classes and methods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
suppressPackageStartupMessages(library(inlabru))
suppressPackageStartupMessages(library(fmesher))
suppressPackageStartupMessages(library(tibble))

convert_fun_link <- function(x, webref) {
  pkg_name <- sub(
    pattern = "^(([^:]*)::)?([^(]+)(\\(.*\\))(<-)?$",
    replacement = "\\2",
    x = x
  )
  if (identical(pkg_name, "")) {
    pkg_name <- "inlabru"
  }
  nm <- sub(
    pattern = "^(([^:]*)::)?([^(]+)(\\(.*\\))(<-)?$",
    replacement = "\\3",
    x = x
  )
  args <- sub(
    pattern = "^(([^:]*)::)?([^(]+)(\\(.*\\))(<-)?$",
    replacement = "\\4\\5",
    x = x
  )
  help_name <- tryCatch(
    {
      basename(utils::help((nm), package = (pkg_name)))
    },
    error = function(e) {
      nm
    }
  )
  paste0(
    '<a href="', webref[[pkg_name]], help_name, '.html">Y</a>',
  )
}

convert_fun_links <- function(df) {
  webref <- list(
    inlabru = "https://inlabru-org.github.io/inlabru/reference/",
    fmesher = "https://inlabru-org.github.io/fmesher/reference/"
  )
  df$Fun <- vapply(
    df[["Fun"]],
    function(x) convert_fun_link(x, webref),
    ""
  )
  df
}

convert_NAs <- function(df) {
  for (k in colnames(df)) {
    df[[k]] <- vapply(
      df[[k]],
      function(x) if (is.na(x)) "" else paste0("`", x, "`"),
      ""
    )
  }
  df
}
```

```{r, echo=FALSE}
special_classes <- c(
  "bru",
  "bru_comp",
  "bru_comp_list",
  "bru_obs",
  "bru_obs_list",
  "bru_predictor",
  "bru_model",
  "bru_info",
  "bru_mapper",
  "bm_list"
)

bm_classes <- names(rlang::pkg_env("inlabru"))
bm_classes <- bm_classes[grepl("^bm_", bm_classes)]
all_classes <- unique(c(special_classes, bm_classes))
# TODO: further filtering

names(all_classes) <- all_classes
methods_list <- c(
  #  "predict",
  #  "generate",
  "gg"
)
names(methods_list) <- methods_list
```

```{r, echo=FALSE}
class_methods <- lapply(all_classes, function(x) {
  y <- unclass(methods(class = x))
  if (NROW(y) == 0) {
    NULL
  } else {
    y <- cbind(class = x, attr(y, "info"))
    rownames(y) <- NULL
    y
  }
})
method_classes <- lapply(methods_list, function(x) {
  y <- unclass(methods(generic.function = x))
  if (NROW(y) == 0) {
    NULL
  } else {
    y <- cbind(class = gsub(paste0("^", x, "\\."), "", y), attr(y, "info"))
    rownames(y) <- NULL
    y
  }
})
```

```{r, echo=FALSE}
info <- dplyr::bind_rows(
  dplyr::bind_rows(class_methods),
  dplyr::bind_rows(method_classes)
) |> # dplyr::filter(from == "inlabru") |>
  dplyr::group_by(generic, class) |>
  dplyr::summarise(generic = generic[1], class = class[1], .groups = "drop") |>
  dplyr::select(Generic = generic, Class = class) |>
  dplyr::mutate(Support = "Y") |>
  dplyr::arrange(Class, Generic)

info_sep <- list(
  gg = info |>
    dplyr::filter(Generic == "gg"),
  bm = info |>
    dplyr::filter(Class %in% bm_classes),
  special = info |>
    dplyr::filter(Class %in% special_classes)
)

info_sep_wide <- lapply(
  info_sep,
  function(x) {
    x |>
      tidyr::pivot_wider(names_from = Class, values_from = Support) |>
      convert_NAs() |>
      dplyr::arrange(Generic)
  }
)
```

```{css,echo=FALSE}
th.rotate {
  /* Something you can count on */
  height: 140px;
  white-space: nowrap;
}

th.rotate > div {
  transform: 
    /* Magic Numbers */
    translate(-5px, 51px)
    /* 60 is really 360 - 60 */
    rotate(300deg);
  width: 30px;
}

th.rotate > div > span {
  border-bottom: 1px solid #ccc;
  padding: 5px 10px;
}
```

### General classes
```{r, echo = FALSE,results="asis"}
x <- knitr::kable(
  info_sep_wide$special,
  caption = "General classes and methods in `inlabru`",
  #  col.names = c("Method", "Class"),
  row.names = FALSE,
  format = "html"
)

gsub('<th style="text-align:(left|center|right);"> ([^(>| )]+) </th>',
  '<th class="rotate"><div><span>\\2</span></div></th>',
  x = x
) |>
  cat(sep = "\n")
```

### Bru mapper classes
```{r, echo = FALSE,results="asis"}
x <- knitr::kable(
  info_sep_wide$bm,
  caption = "`bru_mapper` classes and methods in `inlabru`",
  #  col.names = c("Method", "Class"),
  row.names = FALSE,
  format = "html"
)

gsub('<th style="text-align:(left|center|right);"> ([^(>| )]+) </th>',
  '<th class="rotate"><div><span>\\2</span></div></th>',
  x = x
) |>
  cat(sep = "\n")
```

---
title: "Articles list"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Articles list}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo=FALSE,include=FALSE}
```

## Package vignettes

These package vignettes are available in the R package and on
[`https://inlabru-org.github.io/inlabru/`](https://inlabru-org.github.io/inlabru/)

```{r, echo=FALSE,results="asis"}
files <- list.files(".", "\\.Rmd$")
for (file in files) {
  lines <- readLines(file, n = 10)
  title_idx <- grep("^title: ", lines)
  if (length(title_idx) > 0) {
    title <- sub("^title: ", "", lines[title_idx[1]])
    title <- sub('^"', "", title)
    title <- sub('"$', "", title)
    cat("* [",
      title,
      "](https://inlabru-org.github.io/inlabru/articles/",
      sub("\\.Rmd", ".html", file),
      ")\n",
      sep = ""
    )
  }
}
```

## Package examples

These potentially long running examples/tutorials are available on
[`https://inlabru-org.github.io/inlabru/`](https://inlabru-org.github.io/inlabru/)

```{r, echo=FALSE,results="asis"}
files <- list.files("articles", "\\.Rmd$")
for (file in file.path("articles", files)) {
  lines <- readLines(file, n = 10)
  title_idx <- grep("^title: ", lines)
  if (length(title_idx) > 0) {
    title <- sub("^title: ", "", lines[title_idx[1]])
    title <- sub('^"', "", title)
    title <- sub('"$', "", title)
    cat("* [",
      title,
      "](https://inlabru-org.github.io/inlabru/articles/",
      sub("\\.Rmd", ".html", basename(file)),
      ")\n",
      sep = ""
    )
  }
}
```

---
title: "Iterative linearised INLA method"
author: "Finn Lindgren and Man Ho Suen"
output:
  - rmarkdown::html_vignette
  - rmarkdown::pdf_document
vignette: >
  %\VignetteIndexEntry{Iterative linearised INLA method}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
header-includes:
  - \require{boldsymbol}
  - \newcommand{\bm}[1]{\boldsymbol{#1}}
  - \newcommand{\wt}[1]{\widetilde{#1}}
  - \newcommand{\ol}[1]{\overline{#1}}
  - \newcommand{\wh}[1]{\widehat{#1}}
  - \newcommand{\OO}{\mathcal{O}}
  - \DeclareMathOperator*{\argmax}{arg\,max}
  - \DeclareMathOperator*{\tr}{tr}
  - \newcommand{\pN}{\mathsf{N}}
  - \newcommand{\pCov}{\mathsf{Cov}}
  - \newcommand{\pE}{\mathsf{E}}
  - \newcommand{\pKL}[2]{\mathsf{KL}\left(#1\,\middle\|\,#2\right)}
  - \newcommand{\mmd}{\mathrm{d}}
  - \newcommand{\md}{\,\mmd}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(inlabru)
library(ggplot2)
```

## The INLA method for linear predictors

The INLA method is used to compute fast approximative posterior distribution
for Bayesian generalised additive models. The hierarchical structure of such a 
model with latent Gaussian components $\bm{u}$, covariance parameters
$\bm{\theta}$, and measured response variables $\bm{y}$, can be written as
$$
\begin{aligned}
\bm{\theta} &\sim p(\bm{\theta}) \\
\bm{u}|\bm{\theta} &\sim \mathcal{N}\!\left(\bm{\mu}_u, \bm{Q}(\bm{\theta})^{-1}\right) \\
\bm{\eta}(\bm{u}) &= \bm{A}\bm{u} \\
\bm{y}|\bm{u},\bm{\theta} & \sim p(\bm{y}|\bm{\eta}(\bm{u}),\bm{\theta})
\end{aligned}
$$
where typically each linear predictor element, $\eta_i(\bm{u})$, is linked to a
location parameter of the distribution for observation $y_i$, for each $i$,
via a (non-linear) link function $g^{-1}(\cdot)$. In the R-INLA implementation,
the observations are assumed to be conditionally independent, given $\bm{\eta}$
and $\bm{\theta}$.

## Approximate INLA for non-linear predictors

The premise for the inlabru method for non-linear predictors is to build on
the existing implementation, and only add a linearisation step. The properties of
the resulting approximation will depend on the nature of the non-linearity.

Let $\wt{\bm{\eta}}(\bm{u})$ be a non-linear predictor, i.e. a deterministic function of $\bm{u}$, $$ \wt{\bm{\eta}} (\bm{u}) = \textsf{fcn} (\bm{u}), $$
and let
$\ol{\bm{\eta}}(\bm{u})$ be the 1st order Taylor approximation at $\bm{u}_0$,
$$
\ol{\bm{\eta}}(\bm{u})
= \wt{\bm{\eta}}(\bm{u}_0) + \bm{B}(\bm{u} - \bm{u}_0)
= \left[\wt{\bm{\eta}}(\bm{u}_0) - \bm{B}\bm{u}_0\right] + \bm{B}\bm{u}
,
$$
where $\bm{B}$ is the derivative matrix for the non-linear predictor, evaluated
at $\bm{u}_0$.
Hence, we define
$$
\begin{aligned}
\bm{y} | \bm{u}, {\bm{\theta}} &\overset{d}{=} \bm{y} | \wt{\bm{\eta}}(\bm{u}), {\bm{\theta}} \\
&\sim p (\bm{y} | g^{-1}[\wt{\bm{\eta}}(\bm{u})], {\bm{\theta}})\\
\end{aligned}
$$
The non-linear observation model
$p(\bm{y}|g^{-1}[\wt{\bm{\eta}}(\bm{u})],\bm{\theta})$
is approximated by replacing the non-linear predictor with its linearisation,
so that the linearised model is defined by

$$
\ol{p}(\bm{y}|\bm{u},\bm{\theta})
=
p(\bm{y}|\ol{\bm{\eta}}(\bm{u}),\bm{\theta})
=
p(\bm{y}|g^{-1}[\ol{\bm{\eta}}(\bm{u})],\bm{\theta})
\approx
p(\bm{y}|g^{-1}[\wt{\bm{\eta}}(\bm{u})],\bm{\theta})
=
p(\bm{y}|\wt{\bm{\eta}}(\bm{u}),\bm{\theta})
=
\wt{p}(\bm{y}|\bm{u},\bm{\theta})
$$
The non-linear model posterior is factorised as
$$
\wt{p}(\bm{\theta},\bm{u}|\bm{y}) = \wt{p}(\bm{\theta}|\bm{y})\wt{p}(\bm{u}|\bm{y},\bm{\theta}),
$$
and the linear model approximation is factorised as
$$
\ol{p}(\bm{\theta},\bm{u}|\bm{y}) = \ol{p}(\bm{\theta}|\bm{y})\ol{p}(\bm{u}|\bm{y},\bm{\theta}).
$$

### Fixed point iteration

The remaining step of the approximation is how to choose the linearisation
point $\bm{u}_*$.
For a given linearisation point $\bm{v}$, INLA will compute the posterior mode for $\bm{\theta}$,
$$
\wh{\bm{\theta}}_{\bm{v}} = \argmax_{\bm{\theta}} \ol{p}_\bm{v} ( {\bm{\theta}} | \bm{y} ),
$$
and the joint conditional posterior mode for $\bm{u}$,
$$
\wh{\bm{u}}_{\bm{v}} = \argmax_{\bm{u}} \ol{p}_\bm{v} ( \bm{u} | \bm{y}, \wh{\bm{\theta}}_{\bm{v}} ) .
$$

Define the Bayesian estimation functional^[
Potential other choices for $f(\cdot)$ include the posterior expectation
$\ol{E}(\bm{u}|\bm{y})$ and the marginal conditional modes, 
$$
\left\{\argmax_{u_i} \ol{p}_{\bm{v}}(u_i|\bm{y}),\,i=1,\dots,n\right\},
$$
which was used in `inlabru` up to version 2.1.15, which caused problems for some nonlinear models.
From version 2.2.3, the joint conditional mode is used,
$$
\argmax_{\bm{u}} \ol{p}_{\bm{v}}(\bm{u}|\bm{y},\wh{\bm{\theta}}_{\bm{v}}),
$$
where $\wh{\bm{\theta}}=\argmax_{\bm{\theta}} \ol{p}_{\bm{v}}(\bm{\theta}|\bm{y})$.
]
$$
f(\ol{p}_{\bm{v}}) = (\wh{\bm{\theta}}_{\bm{v}},\wh{\bm{u}}_{\bm{v}})
$$
and let $f(p)=(\wh{\bm{\theta}},\wh{\bm{u}})$ denote the corresponding posterior modes for the true posterior distribution,
$$
\begin{aligned}
    \wh{{\bm{\theta}}} &= \argmax_{\bm{\theta}} p ( {\bm{\theta}} | \bm{y} ), \\
    \wh{\bm{u}} &= \argmax_{\bm{u}} p (\bm{u} | \bm{y}, \wh{{\bm{\theta}}}).
\end{aligned}
$$

The fixed point $(\bm{\theta}_*,\bm{u}_*)=f(\ol{p}_{\bm{u}_*})$ should ideally be close to $(\wh{\bm{\theta}},\wh{\bm{u}})$, i.e. close to the true marginal/conditional posterior mode. We can achieve this for the conditional
latent mode, so that $\bm{u}_*=\argmax_{\bm{u}} p (\bm{u} | \bm{y}, \wh{\bm{\theta}}_{\bm{u}_*})$.

We therefore seek the latent vector $\bm{u}_*$ that generates the fixed point of the
functional, so that $(\bm{\theta}_*,\bm{u}_*)=f(\ol{p}_{\bm{u}_*})$.


One key to the fixed point iteration is that the observation model is linked to
$\bm{u}$ only through the non-linear predictor $\wt{\bm{\eta}}(\bm{u})$, since
this leads to a simplified line search method below.

0. Let $\bm{u}_0$ be an initial linearisation point for the latent variables 
   obtained from the initial INLA call. Iterate the following steps for $k=0,1,2,...$
1. Compute the predictor linearisation at $\bm{u}_0$.
2. Compute the linearised INLA posterior $\ol{p}_{\bm{u}_0}(\bm{\theta}|\bm{y})$.
3. Let $(\bm{\theta}_1,\bm{u}_1)=(\wh{\bm{\theta}}_{\bm{u}_0},\wh{\bm{u}}_{\bm{u}_0})=f(\ol{p}_{\bm{u}_0})$ be the initial candidate for new
   linearisation point.
4. Let $\bm{v}_\alpha=(1-\alpha)\bm{u}_1+\alpha\bm{u}_0$, and find the value
   $\alpha$ minimises $\|\wt{\eta}(\bm{v}_\alpha)-\ol{\eta}(\bm{u}_1)\|$.
5. Set the new linearisation point $\bm{u}_0$ equal to $\bm{v}_\alpha$ and repeat from step 1,
   unless the iteration has converged to a given tolerance.

A potential improvement of step 4 might be to also take into account the prior
distribution for $\bm{u}$ as a minimisation penalty, to avoid moving further than
would be indicated by a full likelihood optimisation.

#### Line search

In step 4, we would ideally want $\alpha$ to be  
$$
\argmax_{\alpha} \left[\ln p(\bm{u}|\bm{y},\bm{\theta}_1)\right]_{\bm{u}=\bm{v}_\alpha}.
$$
However, since this requires access to the internal likelihood and prior density
evaluation code, we instead use a simpler alternative. We consider norms of the form $\|\wt{\eta}(\bm{v}_\alpha)-\ol{\eta}(\bm{u}_1)\|$
that only depend on the nonlinear and linearised predictor expressions, and other known quantities, given $\bm{u}_0$, such as the current INLA estimate of the component wise predictor variances.

Let $\sigma_i^2 = \mathrm{Var}_{\bm{u}\sim \ol{p}(\bm{u}|\bm{y},\bm{\theta}_1)}(\ol{\bm{\eta}}_i(\bm{u}))$
be the current estimate of the posterior variance for each predictor element $i$.
We then define an inner product on the space of predictor vectors as
$$
\langle \bm{a},\bm{b} \rangle_V
=
\sum_i \frac{a_i b_i}{\sigma_i^2} .
$$
The squared norm for the difference between the predictor vectors $\wt{\bm{\eta}}(\bm{v}_\alpha)$
and $\ol{\bm{\eta}}(\bm{u}_1)$,with respect to this inner product, is defined as
$$
\| \wt{\bm{\eta}}(\bm{v}_\alpha) - \ol{\bm{\eta}}(\bm{u}_1)\|^2_V
=
\sum_i \frac{|\wt{\bm{\eta}}_i(\bm{v}_\alpha)-\ol{\bm{\eta}}_i(\bm{u}_1)|^2}{\sigma_i^2} .
$$
Using this norm as the target loss function for the line search avoids many potentially
expensive evaluations of the true posterior conditional log-density. We evaluate
$\wt{\bm{\eta}}_1=\wt{\bm{\eta}}(\bm{u}_1)$ and make use of the linearised predictor
information. Let $\wt{\bm{\eta}}_\alpha=\wt{\bm{\eta}}(\bm{v}_\alpha)$ and $\ol{\bm{\eta}}_\alpha=\ol{\bm{\eta}}(\bm{v}_\alpha)=(1-\alpha)\wt{\bm{\eta}}(\bm{u}_0)+\alpha\ol{\bm{\eta}}(\bm{u}_1)$.
In other words, $\alpha=0$ corresponds to the previous linear predictor, and $\alpha=1$
is the current estimate from INLA.
An exact line search would minimise $\|\wt{\bm{\eta}}_\alpha-\ol{\bm{\eta}}_1\|$. 
Instead, we define a quadratic approximation to the
non-linear predictor as a function of $\alpha$,
$$
\breve{\bm{\eta}}_\alpha =
\ol{\bm{\eta}}_\alpha + \alpha^2 (\wt{\bm{\eta}}_1 - \ol{\bm{\eta}}_1)
$$
and minimise the quartic polynomial in $\alpha$,
$$
\begin{aligned}
\|\breve{\bm{\eta}}_\alpha-\ol{\bm{\eta}}_1\|^2
&=
\| (\alpha-1)(\ol{\bm{\eta}}_1 - \ol{\bm{\eta}}_0) + \alpha^2 (\wt{\bm{\eta}}_1 - \ol{\bm{\eta}}_1) \|^2
.
\end{aligned}
$$
If initial expansion and contraction steps are carried out, leading to an initial
guess of $\alpha=\gamma^k$, where $\gamma>1$ is a scaling factor (see `?bru_options`, `bru_method$factor`) and $k$ is the
(signed) number of expansions and contractions, the quadratic expression is replaced by
$$
\begin{aligned}
\|\breve{\bm{\eta}}_\alpha-\ol{\bm{\eta}}_1\|^2
&=
\| (\alpha-1)(\ol{\bm{\eta}}_1 - \ol{\bm{\eta}}_0) + \frac{\alpha^2}{\gamma^{2k}} (\wt{\bm{\eta}}_{\gamma^k} - \ol{\bm{\eta}}_{\gamma^k}) \|^2
,
\end{aligned}
$$
which is minimised on the interval $\alpha\in[\gamma^{k-1},\gamma^{k+1}]$.

## Posterior non-linearity checks

Whereas the inlabru optimisation method leads to an estimate where $\| \wt{\bm{\eta}} (\bm{u}_*) - \ol{\bm{\eta}}(\bm{u}_*)\|=0$ for a specific $\bm{u}_*$, the overall posterior approximation accuracy depends on the degree of nonlinearity in the vicinity of $\bm{u}_*$.
There are two main options for evaluating this nonlinearity, using sampling from the approximate posterior distribution.
The first option is 
$$
\begin{aligned}
\sum_i \frac{E_{\bm{u}\sim \ol{p}(\bm{u}|\bm{y})}\left[
|\ol{\bm{\eta}}_i(\bm{u})-\wt{\bm{\eta}}_i(\bm{u})|^2\right]}{\mathrm{Var}_{\bm{u}\sim \ol{p}(\bm{u}|\bm{y})}(\ol{\bm{\eta}}_i(\bm{u}))} ,
\end{aligned}
$$ 
which is the posterior expectation of the component-wise variance-normalised squared deviation between the non-linear and linearised predictor. Note that the normalising variance includes the variability induced by the posterior uncertainty for $\bm{\theta}$, whereas the $\|\cdot\|_V$ norm used for the line search used only the posterior mode.
Another option is 
$$
E_{(\bm{u},\bm{\theta})\sim \ol{p}(\bm{u},\bm{\theta}|\bm{y})} \left[\ln \frac{\ol{p}(\bm{u} |\bm{y},{\bm{\theta}})}{\wt{p}(\bm{u}|\bm{y},{\bm{\theta}})}\right]
=
E_{\bm{\theta}\sim \ol{p}(\bm{\theta}|\bm{y})} \left\{
E_{\bm{u}\sim \ol{p}(\bm{u}|\bm{y},\bm{\theta})}
\left[\ln \frac{\ol{p}(\bm{u} |\bm{y},{\bm{\theta}})}{\wt{p}(\bm{u}|\bm{y},{\bm{\theta}})}\right]
\right\}
$$ 
which is the Kullback–Leibler divergence for the conditional posterior densities, $\pKL{\ol{p}}{\wt{p}}$, integrated over the approximate posterior
distribution for $\bm{\theta}$.  Implementing this would require access to the likelihood and prior distribution details.  The next section explores this in more detail.

#### Accuracy

We wish to assess how accurate the approximation is. Thus, we compare $\wt{p}(\bm{u} | \bm{y}, \bm{\theta} )$ and $\ol{p}(\bm{u} |\bm{y},\bm{\theta})$. 
With Bayes' theorem, 
$$
\begin{aligned}
    p(\bm{u}|\bm{y},{\bm{\theta}}) &= \frac{p(\bm{u},\bm{y}|{\bm{\theta}})}{p(\bm{y}|{\bm{\theta}})} \\
    &= \frac{p(\bm{y}|\bm{u},{\bm{\theta}}) p(\bm{u}|{\bm{\theta}})}{p(\bm{y}|{\bm{\theta}})},
\end{aligned}
$$
where $p(\bm{u}|\bm{\theta})$ is a Gaussian density and $p(\bm{y}|\bm{\theta})$ is a scaling factor that doesn't depend on $\bm{u}$. We can therefore focus on the behaviour of $\ln p(\bm{y}|\bm{\theta},\bm{u})$ for the exact and linearised observation models.

Recall that the observation likelihood only depends on $\bm{u}$ through $\bm{\eta}$.
Using a Taylor expansion with respect to $\bm{\eta}$ and $\bm{\eta}^*=\wt{\bm{\eta}}(\bm{u}_*)$,
$$
\begin{aligned}
    \ln p(\bm{y}|\bm{\eta},\bm{\theta}) &=
    \ln p (\bm{y}|{\bm{\theta}},\bm{\eta}^*))  \\ 
    &\qquad + \sum_i \left.\frac{\partial}{\partial\eta_i} \ln p (\bm{y} | {\bm{\theta}}, \bm{\eta}) \right|_{\bm{\eta}^*}\cdot (\eta_i - \eta^*_i) \\
    &\qquad + \frac{1}{2}\sum_{i,j} \left.\frac{\partial^2}{\partial\eta_i\partial\eta_j} \ln p (\bm{y} | {\bm{\theta}}, \bm{\eta}) \right|_{\bm{\eta}^*}\cdot (\eta_i - \eta^*_i) (\eta_j - \eta^*_j) + \OO(\|\bm{\eta}-\bm{\eta}^*\|^3),
\end{aligned}
$$
Similarly, for each component of $\wt{\bm{\eta}}$,
$$
\begin{aligned}
\wt{\eta}_i(\bm{u}) &= \eta^*_i 
+\left[\left.\nabla_{u}\wt{\eta}_i(\bm{u})\right|_{\bm{u}_*}\right]^\top (\bm{u} - \bm{u}_*)
\\&\quad
+\frac{1}{2}(\bm{u} - \bm{u}_*)^\top\left[\left.\nabla_{u}\nabla_{u}^\top\wt{\eta}_i(\bm{u})\right|_{\bm{u}_*}\right] (\bm{u} - \bm{u}_*) + \OO(\|\bm{u}-\bm{u}^*\|^3)
\\&= \eta_i^* + b_i(\bm{u}) + h_i(\bm{u}) + \OO(\|\bm{u}-\bm{u}_*\|^3)
\\&= \ol{\eta}_i(\bm{u}) + h_i(\bm{u}) + \OO(\|\bm{u}-\bm{u}_*\|^3)
\end{aligned}
$$

where $\nabla_u\nabla_u^\top$ is the Hessian with respect to $\bm{u}$, $b_i$ are linear in $\bm{u}$, and $h_i$ are quadratic in $\bm{u}$.
Combining the two expansions and taking the difference between the full and linearised log-likelihoods, we get
$$
\begin{aligned}
    \ln \wt{p}(\bm{y}|\bm{u},\bm{\theta}) -
    \ln \ol{p}(\bm{y}|\bm{u},\bm{\theta})
    &=
    \sum_i \left.\frac{\partial}{\partial\eta_i} \ln p (\bm{y} | {\bm{\theta}}, \bm{\eta}) \right|_{\bm{\eta}^*}\cdot h_i(\bm{u}) + \OO(\|\bm{u}-\bm{u}_*\|^3)
\end{aligned}
$$
Note that the log-likelihood Hessian difference contribution only involves third order $\bm{u}$ terms
and higher, so the expression above includes all terms up to second order.

Let
$$
g_i^*=\left.\frac{\partial}{\partial\eta_i} \ln p (\bm{y} | {\bm{\theta}}, \bm{\eta}) \right|_{\bm{\eta}^*}
$$
and
$$
\bm{H}^*_i = \left.\nabla_{u}\nabla_{u}^\top\wt{\eta}_i(\bm{u})\right|_{\bm{u}_*} .
$$
and form the sum of their products, $\bm{G}=\sum_i g_i^*\bm{H}_i^*$.
Then
$$
\begin{aligned}
    \ln \wt{p}(\bm{y}|\bm{u},\bm{\theta}) -
    \ln \ol{p}(\bm{y}|\bm{u},\bm{\theta})
    &=
    \frac{1}{2}
    \sum_i g_i^* (\bm{u}-\bm{u}_*)^\top \bm{H}_i^* (\bm{u}-\bm{u}_*) + \OO(\|\bm{u}-\bm{u}_*\|^3)
    \\&=
    \frac{1}{2}
    (\bm{u}-\bm{u}_*)^\top \bm{G} (\bm{u}-\bm{u}_*) + \OO(\|\bm{u}-\bm{u}_*\|^3).
\end{aligned}
$$
With $\bm{m}=\pE_\ol{p}(\bm{u}|\bm{y},\bm{\theta})$ and $\bm{Q}^{-1}=\pCov_\ol{p}(\bm{u},\bm{u}|\bm{y},\bm{\theta})$, we obtain
$$
\begin{aligned}
\pE_{\ol{p}}\left[ \nabla_{\bm{u}}  \left\{\ln \wt{p}(\bm{y}|\bm{u},\bm{\theta}) -
    \ln \ol{p}(\bm{y}|\bm{u},\bm{\theta})\right\}\right]
 &\approx
    \bm{G}(\bm{m}-\bm{u}_*) ,
    \\
\pE_{\ol{p}}\left[ \nabla_{\bm{u}}\nabla_{\bm{u}}^\top  \left\{\ln \wt{p}(\bm{y}|\bm{u},\bm{\theta}) -
    \ln \ol{p}(\bm{y}|\bm{u},\bm{\theta})\right\}\right]
 &\approx
    \bm{G} ,
    \\
\pE_{\ol{p}}\left[    \ln \wt{p}(\bm{y}|\bm{u},\bm{\theta}) -
    \ln \ol{p}(\bm{y}|\bm{u},\bm{\theta})\right]
 &\approx
    \frac{1}{2}
    \tr(\bm{G}\bm{Q}^{-1}) + \frac{1}{2} (\bm{m}-\bm{u}_*)\bm{G}(\bm{m}-\bm{u}_*)^\top .
\end{aligned}
$$
For each $\bm{\theta}$ configuration in the INLA output, we can extract both $\bm{m}$ and the sparse precision matrix $\bm{Q}$ for the Gaussian approximation. The non-sparsity structure of $\bm{G}$ is contained in the non-sparsity of $\bm{Q}$, which allows the use of Takahashi recursion (`inla.qinv(Q)`) to compute the corresponding $\bm{Q}^{-1}$ values needed to evaluate the trace $\tr(\bm{G}\bm{Q}^{-1})$.
Thus, to implement a numerical approximation of this error analysis only needs special access to the log-likelihood derivatives $g_i^*$, as $H_i^*$ can in principle be evaluated numerically.

For a given $\bm{\theta}$,
$$
\begin{aligned}
\pKL{\ol{p}}{\wt{p}} &=
E_{\ol{p}}\left[\ln\frac{\ol{p}(\bm{u}|\bm{y},\bm{\theta})}{\wt{p}(\bm{u}|\bm{y},\bm{\theta})}\right]
\\&=
E_{\ol{p}}\left[
\ln\frac{\ol{p}(\bm{y}|\bm{u},\bm{\theta})}{\wt{p}(\bm{y}|\bm{u},\bm{\theta})}
\right]
-
\ln\frac{\ol{p}(\bm{y}|\bm{\theta})}{\wt{p}(\bm{y}|\bm{\theta})} .
\end{aligned}
$$
The first term was approximated above. The second term can also be approximated using the derived quantities (to be continued...).

Summary: The form of the observation likelihood discrepancy shows that, given a linearised posterior $\pN(\bm{m},\bm{Q}^{-1})$, a Gaussian approximation to the  nonlinear model posterior, $\pN(\wt{\bm{m}},\wt{\bm{Q}}^{-1})$, can be obtained from $\wt{\bm{Q}}=\bm{Q}-\bm{G}$ and $\wt{\bm{Q}}\wt{\bm{m}}=\bm{Q}\bm{m}-\bm{G}\bm{u}_*$. The K-L divergence becomes
$$
\begin{aligned}
\pKL{\ol{p}}{\wt{p}} &\approx
\frac{1}{2}
\left[
\ln\det(\bm{Q})-
\ln\det(\bm{Q}-\bm{G})
-\tr\left(\bm{G}\bm{Q}^{-1}\right)
+
(\bm{m}-\bm{u}_*)^\top\bm{G}(\bm{Q}-\bm{G})^{-1}\bm{G}(\bm{m}-\bm{u}_*)
\right] .
\end{aligned}
$$
When the INLA posterior has mean $\bm{m}=\bm{u}_*$, e.g. for models with additive Gaussian observation noise, and $\bm{\theta}=\wh{\bm{\theta}}_{\bm{u}_*}$, the last term vanishes.

Note: by implementing the K-L divergence accuracy metric, a by-product would be
improved posterior estimates based on $\wt{\bm{m}}$ and $\wt{\bm{Q}}$.

### Well-posedness and initialisation
On a side note, one might be concerned about initialisation at, or convergence to, a saddle point. Although it is not implemented in inlabru, we want to talk about the technicality how we define the initial linearisation point $u_0$.

Generally speaking, any values of $\bm{u}_0$ work except the case that the gradient evaluated at $\bm{u}_0$ is $\bm{0}$ because the linearisation point will never move away if the prior mean is also $\bm{0}$. In general, this tends to be a saddle point problem. In some cases the problem can be handled by changing the predictor parameterisation or just changing the initialisation point using the `bru_initial` option. However, for true saddle point problems, it indicates that the predictor parameterisation may lead to a multimodal posterior distribution or is ill-posed in some other way. This is a more fundamental problem that cannot be fixed by changing the initialisation point.

In these examples, where $\beta$ and $\bm{u}$ are latent Gaussian components, the predictors 1, 3, and 4 would typically be safe, but predictor 2 is fundamentally non-identifiable.
$$ 
\begin{aligned}
\bm{\eta}_1 &= \bm{u}, \\
\bm{\eta}_2 &= \beta \bm{u}, \\
\bm{\eta}_3 &= e^\beta \bm{u}, \\
\bm{\eta}_4 &= F_\beta^{-1} ( \Phi(z_\beta)) \bm{u}, \quad z_{\beta} \sim \mathsf{N}(0,1) .
\end{aligned}
$$
Note that for $\bm{\eta}_3$ and $\bm{\eta}_4$, the partial derivatives with respect to $\beta$ are zero for $\bm{u}=\bm{0}$. However, the first inlabru iteration will give a non-zero estimate of $\bm{u}$, so that subsequent iteration will involve both $\beta$ and $\bm{u}$.

---
title: "Devel: Customised model components with the bru_mapper system"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Devel: Customised model components with the bru_mapper system}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
header-includes:
  - \require{boldsymbol}
  - \newcommand{\bm}[1]{\boldsymbol{#1}}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

(Vignette under construction!)

## Mapper system introduction

```{r setup, eval = TRUE, echo=FALSE, include=FALSE}
library(inlabru)
```

Each inlabru latent model component generates an _effect_, given the latent _state_ vector.
The purpose of the mapper system is to define the link from state vectors to effect vectors.
For an ordinary model component, named $c$, this link can be represented as a matrix-vector product
$$
\eta_c(u_c) = A_c(\text{input}) u_c,
$$
where $u_c$ is the latent state vector, $\eta_c(u_c)$ is the resulting component effect vector, and
$A_c(\text{input})$ is a _component model matrix_. This matrix depends on the component
inputs (covariates, index information, etc) from `main`, `group`, `replicate`, and `weights`
in the component definition.

The wider scope of component definitions is discussed in the [component vignette](https://inlabru-org.github.io/inlabru/articles/component.html). Here, we
will focus on the mapper system itself, how the basic building blocks are used to
construct the built-in mappers, and finally how to construct new mappers.

The in-package documentation for the mapper methods is contained in four parts:
```{r, eval=FALSE}
?bru_mapper # Mapper constructors, with links to predefined mappers
?bru_mapper_generics # Generic and default methods
?bru_get_mapper # Mapper extraction methods
```

Regular users normally at most need some of the methods from `?bru_mapper`
and links to predefined mappers, and sometimes `?bru_mapper_generics`.
The `bru_mapper_define()` and `bru_get_mapper()` methods are needed for those
implementing their own mapper class.

## Mappers

The main purpose of each mapper class is to allow evaluating a component
effect, from given `input` and latent `state`, by calling
```{r, eval=FALSE}
ibm_eval(mapper, input, state)
```
for the `mapper` associated with the component definition.

### Basic mappers

Basic mappers take covariate vectors or matrices as input and numeric vectors as
state.

Constructors:

* `bm_const()`
* `bm_linear()`
* `bm_index(n)`
* `bm_factor(values, factor_mapping)`
* `bm_matrix()`
* `bm_harmonics(order, scaling, intercept, interval)`

#### const

The `const` mapper defines a mapping from an empty state vector. This is used to
define offset components, $\eta(u)_j = z_j$, where $z$ is a fixed input vector,
indexed by $j$.

#### linear

The `linear` mapper defines a mapper from a length 1 state vector. This is used to
define ordinary "fixed" covariate effects, linear in both the covariate and in the state
variable; $\eta(u)_j = z_j u$, where $u$ is the state, and $z$ is the
input covariate vector.

#### index

The `index(n)` mapper defines a direct mapping from a length `n` state
variable. This is used for structured and unstructured "random effect" component
effects; $\eta(u)_j = u_{z_j}$, where $u$ is the state vector, and $z$ is the
input index vector.

#### factor

The `factor(values, factor_mapping)` mapper defines a direct mapping between
a state vector of length equal to the number of factor levels of `values`
(for `factor_mapping = "full"`), or one less (for `factor_mapping = "contrast"`).
Each factor level is represented as a dummy 0/1 variable, or equivalently,
the input is used as a label index into the state vector; $\eta(u)_j = u_{z_j}$.
In the `"contrast"` case, the first factor level is removed from the model,
and the corresponding effect defined to be zero.

#### matrix

The `matrix` mapper defines a direct matrix multiplication between
a pre-computed model matrix $A_\text{input}$ with the state
vector; $\eta(u) = A_\text{input} u$.
This is used e.g. for linear model formula input, that is converted to a
component model matrix before handing it over to the mapper system.

#### harmonics

The `harmonics(order, scaling, intercept, interval)` mapper defines a sum of
weighted harmonics on a real interval $(L, U)$, each frequency
additionally scaled by factors determined by `scale`;
$\eta(u_j) = \sum_{k=1}^{1+2p} A_{j,k} u_k$, where $A_{j,k}=s_k$ for $k=1$,
$A_{j,2k}=s_{k+1}\cos[2\pi k (z_j - L)/(U-L)]$ for $k=1,2,\dots,p$, and
$A_{j,2k+1}=s_{k+1}\sin[2\pi k (z_j - L)/(U-L)]$ for $k=1,2,\dots,p$,
where $p=$`order`.  When `intercept` is `TRUE`, the state vector has length
$1+2p$.
When `intercept` is `FALSE`, the first, constant, column is
removed, and the state vector has length $2p$.

### Transformation mappers

Transformation mappers are mappers that would normally be combined with other
mappers, as steps in a sequence of transformations, or as individual
transformation mappers.

* `bm_shift()`
* `bm_scale()`
* `bm_marginal(qfun, pfun, ...)`
* `bm_aggregate(rescale)`
* `bm_logsumexp(rescale)`
* `bm_logitaverage()`

#### shift

The `shift` mapper adds the `input` vector to the `state` vector.
This can be used as a stage of a `pipe` mapper.

```{r, eval=FALSE}
mapper <- bm_shift()
ibm_eval(mapper,
  input = ...,
  state = ...
)
```

#### scale

The `scale` mapper multiplies the `input` vector with the `state` vector.
This is used for INLA models to implement the `weights` component scaling,
as the final stage of a `pipe` mapper, see below.

```{r, eval=FALSE}
mapper <- bm_scale()
ibm_eval(mapper,
  input = ...,
  state = ...
)
```

#### marginal

The `marginal` mapper transforms the `state` vector from standardised Gaussian marginal distribution
to the distribution defined by a quantile function.
This can be used to define components with latent marginal `N(0,1)` distribution but non-Gaussian
effect.

```{r, eval=FALSE}
mapper <- bm_marginal(
  qfun = ..., pfun = ..., dfun = ..., ..., inverse = ...
)
ibm_eval(mapper,
  input = NULL, # If a list is given, it overrides the parameter specification
  state = ...
)

# Examples:
mapper <- bm_marginal(qfun = qexp, rate = 1 / 8)
mapper <- bm_marginal(
  qfun = qexp, pfun = pexp, dfun = dexp, rate = 1 / 8
)
```

#### aggregate

The `aggregate` mapper aggregates the `state` vector elements by blockwise
summation after scaling the elements by weights.
This can be used for summation, integration, and averaging (for `rescale=TRUE`).

```{r, eval=FALSE}
mapper <- bm_aggregate(rescale = ...)
ibm_eval(mapper,
  input = list(block = ..., weights = ...),
  state = ...
)
```

#### logsumexp

The `logsumexp` mapper aggregates the `exp(state)` vector elements by blockwise
summation after scaling the elements by weights, and takes the logarithm. The
implementation takes care to avoid numerical overflow.
This can be used for summation, integration, and averaging (for `rescale=TRUE`).

```{r, eval=FALSE}
mapper <- bm_logsumexp(rescale = ...)
ibm_eval(mapper,
  input = list(block = ..., weights = ...),
  state = ...
)
```

#### logitaverage

The `logitaverage` mapper aggregates `plogis(state)` vector elements by blockwise
summation after scaling the elements by weights, and takes `qlogis()` of the results.
The implementation takes care to avoid numerical overflow.
This can be used for averaging probabilities on the logit scale.
However, note that while aggregation of Poisson variables gives a new Poisson
variable, aggregation of Binomial variables with different probability
parameters does not give a new Binomial variable, so this type of calculation
should be used with care.

```{r, eval=FALSE}
mapper <- bm_logitaverage()
ibm_eval(mapper,
  input = list(block = ..., weights = ...),
  state = ...
)
```

### Compound mappers

Compound mappers define collections or chains of mappings,
and can take various forms of input. The state vector is normally a numeric vector,
but can in some cases be a list of vectors.

* `bm_collect(mappers, hidden)`
* `bm_multi(mappers)`
* `bm_pipe(mappers)`
* `bm_sum(mappers, single_input)`
* `bm_repeat(mapper, n_rep, interleaved)`

#### collect

The `collect` mapper defines a compound mapper where the Jacobian is
block-diagonal, and each block is defined by a separate sub-mapper.
When `hidden = TRUE`, it can be used for INLA models where the user-visible
part of the latent state is only a part of the internal state, such as for
`"bym"` models.

```{r, eval=FALSE}
mapper <- bm_collect(list(name1 = ..., name2 = ..., ...),
  hidden = FALSE
)
ibm_eval(mapper,
  input = list(name1 = ..., name2 = ..., ...),
  state = ...
)
# If hidden = TRUE, the inla_f=TRUE argument "hides" all but the first mapper:
ibm_eval(mapper,
  input = name1_input,
  state = ...,
  inla_f = TRUE
)
```

#### multi

The `multi` mapper defines a compound mapper where the Jacobian is
the row-wise Kronecker product between the sub-mapper Jacobians.
This can be used for INLA models to construct the internal mapper
for `main`, `group`, and `replicate`, but also for user-defined models
where a single `rgeneric` or `cgeneric` model is defined on e.g.
a space-time product domain.

```{r, eval=FALSE}
mapper <- bm_multi(list(name1 = ..., name2 = ..., ...))
ibm_eval(mapper,
  input = list(name1 = ..., name2 = ..., ...),
  state = ...
)
```

#### pipe

Pipe mappers chain multiple mappers into a sequence, where the evaluation
result of each mapper is given as the state vector for the next mapper.

```{r, eval=FALSE}
mapper <- bm_pipe(list(name1 = ..., name2 = ..., ...))
ibm_eval(mapper,
  input = list(name1 = ..., name2 = ..., ...),
  state = ...
)
```

#### sum (from version `2.12.0.9001`)

Sum mappers take a list of mappers and adds their effects, based on a state
vector with the individual mapper states combined in sequence.
The `ibm_` method inputs are passed on to the sub-mapper methods.
When `single_input` is `TRUE`, the same input is passed to all sub-mappers.
Otherwise, the `input` must be a `list`, `data.frame`, or `matrix`, whose
elements/columns are passed on to the respective sub-mappers.

```{r, eval=FALSE}
mapper <- bm_sum(mappers = ..., single_input = ...)
ibm_eval(mapper,
  input = ...,
  state = ...
)
```

#### repeat

Repeat mappers take a single mapper and repeats it,
corresponding to a state vector
of length `sum(n_rep) * ibm_n(mapper, ...)`. The `ibm_` method inputs are
passed on to the sub-mapper methods.

If `interleaved` is `TRUE` (from version `2.12.0.9001`),
the repeated mappers are interleaved, e.g. if
the state vectors for two individual mappers were `x1` and `x2`, interleaved
mappers have state vector `c(x1[1], x2[1], ..., x1[2], x2[2], ...)`, whereas
ordinary non-interleaved mappers have state vector `c(x1, x2, ...)`.

For version `2.12.0.9001`, `n_rep` and `interleaved` may be vectors, for grouped
repetition and interleaving. Non-interleaved repeats are combined, so that
the resulting internal mapper structure is a compact as possible.

```{r, eval=FALSE}
mapper <- bm_repeat(mapper = ..., n_rep = ..., interleaved = ...)
ibm_eval(mapper,
  input = ...,
  state = ...
)
```

#### The core model component mapper

All model component mappers are currently defined as a `pipe` mapper containing
a `multi` mapper followed by a `scale` mapper:
```{r, eval=FALSE}
mapper <-
  bm_pipe(
    list(
      mapper = bm_multi(list(main = ..., group = ..., replicate = ...)),
      scale = bm_scale()
    )
  )
ibm_eval(mapper,
  input = list(
    mapper = list(main = ..., group = ..., replicate = ...),
    scale = ...
  ),
  state = ...
)
```

### Object mappers

For object with predefined mapper conversion methods, use `bru_mapper(object)`
to obtain a suitable mapper object.
Current predefined object mappers are defined for objects of these classes:

* `fm_mesh_2d`; Use `bru_mapper(object)` or `bm_fmesher(object)`
* `fm_mesh_1d`; Use `bru_mapper(object, indexed)` or,
  if `indexed=TRUE`, `bm_fmesher(object)`.
* Any object supporting the `fmesher::fm_dof()` and `fmesher::fm_basis()` methods;
  Use `bm_fmesher(object)`. If `fmesher` in the future gains a common
  base class for such objects, a corresponding `bru_mapper()` conversion method
  will be added.

More details are given below.

### Model object mappers

For inla model objects with predefined mappers, use `bru_get_mapper(object)`
Current predefined model object mappers are defined for models of these classes:

* `inla.spde`; this includes `inla.spde2.matern()` and `inla.spde2.pcmatern()`
models. The conversion method calls `bru_mapper()` in the appropriate way for
the mesh type used for the model
* `inla.rgeneric`; This relies on the rgeneric definition including a `"mapper"`
callback argument. A less invasive method, that works for both rgeneric and cgeneric
models, is to define a `bru_get_mapper()` method for the class, which needs to
include a unique class identifier, e.g. add
`class(object) <- c("my_unique_model_class", class(object))` and define
```{r, eval=FALSE}
bru_get_mapper.my_unique_model_class <- function(model, ...) {
  ...
}
```
and register it in the namespace.

### Special mappers

* `bm_taylor()`

## Mapper methods

```{r, eval=FALSE}
ibm_n(mapper, inla_f, ...)
ibm_n_output(mapper, input, ...)
ibm_values(mapper, inla_f, ...)
ibm_jacobian(mapper, input, state, ...)
ibm_eval(mapper, input, state, ...)
ibm_names(mapper, ...)
ibm_inla_subset(mapper, ...)
```

An example where the `inla_f` argument matters is the `bm_collect` class,
when the `hidden=TRUE` argument is used to indicate that only the first mapper
should be used for the `INLA::f()` inputs, e.g. for `"bym2"` models. For
`inla_f=FALSE` (the default), the `ibm_n` and `ibm_values` methods return the
total number of latent variables, but with `inla_f=TRUE` we get the values
needed by `INLA::f()` instead:
```{r eval=TRUE}
mapper <- bm_collect(
  list(
    a = bm_index(3),
    b = bm_index(2)
  ),
  hidden = TRUE
)
ibm_n(mapper)
ibm_values(mapper)
ibm_n(mapper, inla_f = TRUE)
ibm_values(mapper, inla_f = TRUE)
```
```{r eval=TRUE}
ibm_n(mapper, multi = TRUE)
ibm_values(mapper, multi = TRUE)
ibm_n(mapper, inla_f = TRUE, multi = TRUE)
ibm_values(mapper, inla_f = TRUE, multi = TRUE)
```

For the `bm_multi` class, the `multi` argument provides access to the
inner layers of the multi-mapper:
```{r eval=TRUE}
mapper <- bm_multi(list(
  a = bm_index(3),
  b = bm_index(2)
))
ibm_n(mapper)
ibm_n(mapper, multi = TRUE)
ibm_values(mapper)
ibm_values(mapper, multi = TRUE)
```

```{r eval=TRUE}
ibm_n(mapper, inla_f = TRUE)
ibm_n(mapper, multi = TRUE, inla_f = TRUE)
ibm_values(mapper, inla_f = TRUE)
ibm_values(mapper, multi = TRUE, inla_f = TRUE)
```

The default `ibm_inla_subset` method determines the inla subset by comparing
the full values information from `ibm_values(mapper, inla_f = FALSE)` to the
inla specific values information from `ibm_values(mapper, inla_f = TRUE)`,
to determine the logical vector identifying the inla values subset.  Custom mappers
should normally not need to specialise this method.

## Mappers for `fmesher` objects

For component models referenced by a character label (e.g. `"iid"`, `"bym2"`, `"rw2"` etc),
inlabru will construct default mappers, that in most cases replicate the default `INLA::f()`
behaviour.

All `fmesher` objects with defined `fmesher::fm_dof()` and `fmesher::fm_basis()` methods
are supported by the `bm_fmesher()` constructor:
```{r, eval=FALSE}
bm_fmesher(mesh)
```
This defines a mapper that gives `fm_dof(mesh)` as output from `ibm_n()`, and
`seq_len(fm_dof(mesh))` as output from `ibm_values()` (a so-called "indexed mapper").

For the `fm_mesh_1d` and `fm_mesh_2d` classes, the default mappers can also be
constructed by `bru_mapper` method:
```{r, eval=FALSE}
bru_mapper(mesh)
```

For 1d meshes (`fm_mesh_1d`), one can specify a non-indexed mapper
with the `indexed = FALSE` argument, that gives `mesh$mid` or `mesh$loc` as output
```{r, eval=FALSE}
# If ibm_values() should return
#   mesh$mid (if available, e.g. for "rw2" models with degree=2 meshes)
# or
#   mesh$loc
bru_mapper(mesh, indexed = FALSE)
# If ibm_values() should return seq_len(fm_dof(mesh))
#   (e.g. for inla.spde2.pcmatern() models)
bru_mapper(mesh, indexed = TRUE)
```

## Customised mappers

A mapper object should store enough information in order for the `ibm_*` methods
to work.
The simplest case of  a customised mapper is to just attached a new class
label to the front of the S3 `class()` information of an existing mapper,
to obtain a class to override some of the standard `ibm_*` method
implementations.  More commonly, one would instead start with a basic `list()`,
that might _contain_ an existing mapper object, and then add methods that
know how to use that information.  In all these cases, the `bru_mapper_define()`
method should be used to properly set the class information:
```{r, eval=FALSE}
bru_mapper_define(mapper, new_class)
```
Implementations can avoid having to define `ibm_n` and `ibm_values` methods,
by instead computing and storing `n`, `n_inla`, `values`, and `values_inla`
in the mapper object during construction:
```{r, eval=FALSE}
bru_mapper_define(
  mapper = list(n = 10, values = 1:10),
  new_class = "my_bm_class_name"
)
```
The default `ibm_n` and `ibm_values`
methods checks if these values are available, and return the appropriate values
depending on the `inla_f` argument. When `*_inla` values are requested but not
available, the methods fall back to the non-inla versions. If the needed information
is not found, the default methods give an error message.

Note: Before version 2.6.0, `ibm_amatrix` was used instead of `ibm_jacobian`.
Version 2.6.0 still supports this, by having the default `ibm_jacobian` method
call `ibm_amatrix`, but will give a deprecation message later,
and it may be removed in a future version.

## Example

Let's build a mapper class `bm_p_quadratic`
for a model component that takes input covariates with values $a_{ij}$,
$i=1,\dots,m$ and $j=1,\dots,p$,
in a `matrix` or `data.frame` and evaluates a full quadratic expression
$$
\eta_i = x_0 + \sum_{j=1}^p a_{ij} x_j +
  \frac{1}{2}\sum_{j=1}^p\sum_{k=1}^j \gamma_{j,k} a_{ij}a_{ik} x_{j,k},
$$
where the latent component vector is $\bm{x}=[x_0,x_1,\dots,x_p,x_{1,1},\dots,x_{p,1},x_{2,2}\dots,x_{p,p}]$,
and
$$
\gamma_{j,k} = \begin{cases}
1, & j = k,\\
2, & j > k.
\end{cases}
$$

We start with the constructor method. Like the `bm_matrix`, we require the user to supply
a vector of covariate labels, and store them as a character vector. We also
include a `min_degree` parameter to control if an intercept (`min_degree <= 0`)
and linear terms (`min_degree <= 1`) should be included in the model.
```{r, eval=TRUE}
bm_p_quadratic <- function(labels, min_degree = 0, ...) {
  if (is.factor(labels)) {
    mapper <- list(
      labels = levels(labels),
      min_degree = min_degree
    )
  } else {
    mapper <- list(
      labels = as.character(labels),
      min_degree = min_degree
    )
  }
  bru_mapper_define(mapper, new_class = "bm_p_quadratic")
}
```
The `ibm_n` method can compute the value of $n$ from $n=1+p+\frac{p(p+1)}{2}$: 
```{r, eval=TRUE}
ibm_n.bm_p_quadratic <- function(mapper, ...) {
  p <- length(mapper$labels)
  (mapper$min_degree <= 0) + (mapper$min_degree <= 1) * p + p * (p + 1) / 2
}
```
For `ibm_values`, the default method is sufficient, returning the vector $[1,\dots,n]$
with $n$ obtained by `ibm_n(mapper)`. However, for clearer result naming, we can use a
character vector, at least for INLA `f()` models that allow it (we could have an option argument
to the `bm_p_quadratic` constructor to control this):
```{r, eval=TRUE}
ibm_values.bm_p_quadratic <- function(mapper, ...) {
  p <- length(mapper$labels)
  n <- ibm_n(mapper)
  jk <- expand.grid(seq_len(p), seq_len(p))
  jk <- jk[jk[, 2] <= jk[, 1], , drop = FALSE]
  c(
    if (mapper$min_degree <= 0) "Intercept" else NULL,
    if (mapper$min_degree <= 1) mapper$labels else NULL,
    paste0(mapper$labels[jk[, 1]], ":", mapper$labels[jk[, 2]])
  )
}
```
While the approach to `ibm_n` and `ibm_values` above works, it is wasteful to
recompute the `n` and `values` information for each call. Instead we can use that
the default method will check if `n` or `values` are stored in the mapper object,
and then return them. If we change the `.` in the method definitions to `_`,
we can end the constructor method like this, making subsequent method calls faster:
```{r, eval=FALSE}
bm_p_quadratic <- function(labels, min_degree = 0, ...) {
  ...
  mapper <- bru_mapper_define(mapper, new_class = "bm_p_quadratic")
  mapper$n <- ibm_n_bm_p_quadratic(mapper)
  mapper$values <- ibm_values_bm_p_quadratic(mapper)
  mapper
}
```
Note that the order matters, since our `ibm_values_bm_p_quadratic`
function calls `ibm_n(mapper)`.

We can now define the main part of the mapper interface that computes the
model matrix linking the latent variables to the component effect. It's required
that `NULL` input should return a $0$-by-$n$ matrix.
```{r, eval=TRUE}
ibm_jacobian.bm_p_quadratic <- function(mapper, input, ...) {
  if (is.null(input)) {
    return(Matrix::Matrix(0, 0, ibm_n(mapper)))
  }
  p <- length(mapper$labels)
  n <- ibm_n(mapper)
  N <- NROW(input)
  A <- list()
  in_ <- as(input, "Matrix")
  idx <- 0
  if (mapper$min_degree <= 0) {
    idx <- idx + 1
    A[[idx]] <- Matrix::Matrix(1, N)
  }
  if (mapper$min_degree <= 1) {
    idx <- idx + 1
    A[[idx]] <- in_
  }
  for (k in seq_len(p)) {
    idx <- idx + 1
    A[[idx]] <- in_[, seq(k, p, by = 1), drop = FALSE] * in_[, k]
    A[[idx]][, k] <- A[[idx]][, k] / 2
  }
  A <- do.call(cbind, A)
  colnames(A) <- as.character(ibm_values(mapper))
  A
}
```
If the output of a mapper is of a different size than `NROW(input)`,
the mapper should define a `ibm_n_output(mapper, input, ...)` method to return
the output size for a given input.

---
title: "Defining model components"
author: "Andy Seaton"
date: "Generated on `r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Defining model components}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  dev = "png",
  dev.args = list(type = "cairo-png"),
  fig.width = 7,
  fig.height = 5
)
```

```{r results="hide",warning=FALSE,message=FALSE,echo=FALSE}
options(width = 100) # sets width of output window
```

(Note: vignette under construction!)

## Basic component features

Model components are defined using a formula syntax that is similar to `R-INLA` but has some differences.  The basic syntax is
```{r eval=FALSE}
~ my_component_name(
  main = ...,
  model = ...
)
```

`my_component_name` is a user-chosen label for the model component.  This label is used in model summaries, to label relevant parts of a fitted model object, and to access model components when sampling from a model using `generate()` and `predict()`.  

The `main` argument defines the input data for the component.  For example, an intercept-like component has a vector of ones as an input.  The linear effect of a covariate has the vector of covariate values as an input.  A 2-dimensional SPDE effect takes a 2-column matrix of coordinate locations as an input.   This argument can be a general `R` expression, more details on this are below.  The `main` argument doesn't not need to be named. Other arguments should normally be named, to avoid confusion.

The type of model component is specified using the `model` component, (see `?component`, and `?INLA::inla.list.models()$latent`).

Each component type has an associated `bru_mapper` method that takes `main` as an input and constructs the component design matrix.  Users can also specify their own mapper methods (see `?bru_mapper`).

This syntax replaces the `INLA::f()` function that has the disadvantage that there is no clear separation between the name of the covariate and the label of the effect, and the user often has to do a substantial amount of pre-processing of data to construct relevant inputs.  The documentation for defining model components can be viewed at `?component`. 

The rest of the vignette goes into more detail about defining model components and highlights some advantages of the syntax.

## What is a component design matrix?

A linear additive predictor of a latent Gaussian model can be written as
$$
\begin{equation}
\eta(u)_i = u_0 + \sum_{k=1}^K a_{ik} u_{ik} ,
\end{equation}
$$
where $u$ is a multivariate Gaussian vector, $a_k$ are input information such as covariates or weights for random effects and $i = 1, \ldots, n$.  This can also be written as $\eta(u) = Au$, where $A$ is the model design matrix with $i$-th row $\left[1, a_{i1}, \ldots, a_{iK}\right]$.

We can also conceptally think of the predictor as the sum of $D$ model components, so that we partition $A = \left[A^{(1)} \cdots A^{(D)}\right]$, and each component has an associated **component design matrix** $A^{(d)}$. 

For example, if the component is an intercept parameter, then $A^{(d)} = \left[1, \ldots, 1\right]^\intercal$. If the component is the linear effect of a covariate $z$ then $A^{(d)} = \left[z_1, \ldots, z_n\right]^\intercal$. For more complicated effects, such as SPDE models, the component design matrix maps latent Gaussian parameters to the predictor (also known as the "projector" matrix in this context).  The the construction of each $A^{(d)}$ is handled automatically by `bru_mapper` methods, that define general (linear) **component effects** $\eta^{(d)}(u^{(d)}) = A^{(d)} u^{(d)}$.

Each linear predictor is defined by
$$
\eta(u^{(1)},\dots,u^{(D)}) = \sum_{d=1}^D \eta^{(d)}(u^{(d)}) = \sum_{d=1}^D A^{(d)} u^{(d)} .
$$
Non-linear predictors are defined by R expressions where the component label denotes the corresponding component effect.

### Mapper methods

Each component type has an associated method for converting the information given in the component definition into a component design matrix. The full model design matrix is then used internally in a call `INLA::inla()` to fit the model.  

The advantage of specifying mapper methods is that it supports automatic 'stack building'.  A key feature of `inlabru` is that the full model stack is constructed automatically from the component definitions. The building blocks of the stack are built using `bru_mapper` methods. 

#### Mapper example:  2D SPDE

The mapper for the 2D SPDE effect takes as an input a 2-column matrix of coordinates that represent the locations are which to evaluate the effect.  The parameters of the SPDE component are defined at mesh nodes that may not be the same as the locations at which the effect should be evaluated.  

The appropriate weights required to evaluate the effect at the observation locations can be constructed using `fm_evaluator()`.  The mapper for this model component takes the information in the component definition, in this case the minimum information required is an SPDE model object, and the 2-column matrix that `main` is evaluated to. The mapper then calls `fm_evaluator()` with appropriate arguments extracted from this information.

(NOTE: Deliberately not going into huge detail here; the [bru_mapper](bru_mapper.html) vignette will have more details.)

### Defining `main`, `group`, and `replicate`.

The arguments `main`, `group`, and `replicate` can all take a general `R` expression as an input.  This expression is then evaluated in an environment that consists of the named variables in the data (note: for `sp` objects this does _not_ include the column names from the `@coords` slot, but does include the columns in `@data`).  

If the names are not found in the data then the global environment is searched for objects of that name.

For example, suppose the data has columns named `x` and `y`, then a 2D SPDE model component could be specified as
```{r eval=FALSE}
~ my_spde_effect(
  cbind(x, y),
  model = spde_model
)
```
The expression `cbind(x,y)` is internally evaluated in an environment that contains the columns of the data, which includes the variables `x` and `y`.

The full data object can be accessed using the `.data.` key-word.
An equivalent way to define the same component is
```{r eval=FALSE}
get_xy <- function(df) {
  cbind(df$x, df$y)
}
~ my_spde_effect(
  get_xy(.data.),
  model = spde_model
)
```
This keyword allows code to be written that works with arbitrarily named input data, rather than hardcoding with a specific name of a dataset that may change in future.

If the objects required to evaluate the R expression cannot be found in the data, then the global environment is searched.  This allows users to access objects in the global environment, such as other data-structures that may be of a different dimension to the response data.  This avoids the need to pre-process everything into a single `data.frame`.

The functionality of allowing general `R` expressions can be used to extend the types of data that can be passed to the `bru()`, `bru_obs()` and `lgcp()` functions.  It is the basis for the support of spatial data structures such as `sp` objects, and there is also experimental support to allow users to pass data as a list.  `inlabru` is thus readily extendible, given appropriate functions to extract the relevant information for each component, and associated mappers that convert this information into a component design matrix.

In addition to the three main inputs, the optional `weights` argument also takes an R expression, and
the result is used to scale the component. This can be used for spatially varying coefficient models,
where the `weights` argument provides the covariate values.

### `inlabru`-specific component types

In addition to the majority of latent models that can be defined using `INLA::f()` function (see `INLA::inla.list.models()$latent)`), `inlabru` also has the following models:  `'linear'`, `'fixed'`, `'offset'`, `'factor_full'` and `'factor_contrast'`).

<!--
(NOTE: where should we put some documentation for these things? Some is in `?component` but not all.  I could add to that?
And can we implement an `inlabru.list.models()` function?)
-->

## Shortcuts

There are a few shortcuts to defining model components.  They are for

1. Parameters that are the same for all predictor evaluations (intercept-like parameters).
2. Using a covariate stored in an `sp` `Spatial*` or `terra` `SpatRaster` object.
3. Defining linear effects using an `lm`-style syntax.
4. Behaviour for if `main`, `group` or `replicate` is a function given with no arguments.

### Intercept-like components

The syntax
```{r eval=FALSE}
~ my_intercept(1)
```
can be used as a shortcut for
```{r eval=FALSE}
~ my_intercept(
  main = rep(1, n),
  model = "linear"
)
```
where `n` is the length of the predictor vector.  Note that this shortcut makes an assumption about the approriate length of the predictor.  For many models this can be easily deduced by inspecting input data, but this is not always the case, for example if the response and covariate data are of different dimensions or for joint likelihood models with shared components.

### Spatial covariates

If `main`, `group`, or `replicate`, is the name of an `sf`, `SpatRaster`, or `Spatial*` object stored in the global `R` environment, then `inlabru` attempts to do something intelligent by extracting the covariate information at the locations of the data passed to `bru()` or `bru_obs()`.  This **requires that this data is a `sf` or `SpatialPoints*` object**. `inlabru` does this by calling the `inlabru::eval_spatial()` method, which supports several covariate storage types.

The shortcut
```{r eval=FALSE}
~ my_sp_effect(
  main = a_spatial_object,
  model = "linear"
)
```
internally calls `eval_spatial()` which equivalent to
```{r, eval=FALSE}
~ my_sp_effect(
  main = eval_spatial(a_spatial_object, .data.),
  model = "linear"
)
```
Note that this requires `a_spatial_object` to be stored in the global `R` environment
(or in the environment associated with the model definition code) so it is findable when the expression is internally evaluated by `inlabru`.  Also note that the `eval_spatial()` function by default extracts the first column of the raster object.  For more general situations, one can either specify the optional `main_layer` argument to extract another named or indexed column, or directly
use `main = eval_spatial(a_spatial_object, .data., layer = some_layer)`.

Note that this assumes that the data is either `sf`, so that `st_geometry(.data.)` retrieves point locations, or `sp` where `coordinates(.data.)` retrieves coordinates.  This might not be a sensible thing for all models!  For example, if the input data is a `SpatialPolygonsDataFrame` then `coordinates(.data.)` returns the centroid of each polygon.  As more specific input type support is developed, and support for `sp` is gradually deprecated in favour of `sf` and `terra`,
these special cases may be given more precise meaning.  For example, the `a_spatial_object` above may be an `sf` or `sp` polygon object with data columns, which is interpreted as a spatially piecewise constant
covariate.

### `lm`-style fixed effect and interaction syntax

Since `inlabru` version 2.5.0, a feature has been added to allow users to specify linear fixed effects using a formula as input.  This uses the `model = 'fixed'` component type.  The basic component input is a model matrix. Alternatively, one can supply a formula specification, which is then used to generate a model matrix automatically, with `MatrixModels::model.Matrix(formula, data = .data.)`. If you want a different kind of model matrix construction, replace `~ x1 + x2` by some other R code that generates the needed matrix, using the `.data.` object as input.

Example syntax:
```{r eval=FALSE}
~ my_fixed_effects(
  main = ~ x1:x2 + x3 * x4,
  model = "fixed"
)
```
which is equvalent to
```{r eval=FALSE}
~ my_fixed_effects(
  main = MatrixModels::model.Matrix(~ x1:x2 + x3 * x4, .data.),
  model = "fixed"
)
```
where the data has columns named `x1`, `x2`, `x3`, and `x4`.

This allows users to define interactions in a concise way, by utilising the functionality already supported by the `MatrixModels` package.
The formula is interpreted in the conventional way, `x1:x2` is the interaction of covariates `x1` and `x2`, not including their individual fixed effects, and `x3 * x4` is the interaction of `x3` and `x4` inclusive of the individual fixed effects `x3` and `x4`.
Note that for implementation technical reasons, the estimated parameters appear in
`'summary.random'` instead of the normal `'summary.fixed'`
part of the `inla`/`bru` output object.

The alternative to using this shortcut would be for the user to define and name individual components for each term in the formula.

### A function given with no arguments

If `main`, `group`, or `replicate` are given as a function with no covariates, then this function is applied to the data.  For example,
```{r eval=FALSE}
~ a_component(
  main = a_function,
  model = ...
)
```
is equivalent to
```{r eval=FALSE}
~ a_component(
  main = a_function(.data.),
  model = ...
)
```

### Non-linear predictors

`inlabru` supports non-linear predictors, where $\tilde{\eta}(u,v)$ is a non-linear function of $\eta_u$ and $\eta_v$.  It is important to note that the mapping each component effect vector $\eta_u=A^{(u)} u$ happens **before** the non-linear function is applied.  So, for example, if $\tilde{\eta}(u,v) = \exp(\eta_u + \eta_v)$ then this is evaluated as $\exp(A^{(u)} u + A^{(v)} v)$.

# 2D LGCP fitting

    WAoAAAACAAQFAgACAwAAAAAOAAAABkAMzarvzFUeQAC35LPWQOU/+W5HC7kV0UAR0MD1iAO3
    QAzLB4wnOLRAD22P8nzFDg==

---

    WAoAAAACAAQFAgACAwAAAAAOAAAABj/jj+60Uw0mP74yXKqnqtI/y/rFb3uoNz+mqJJQha2e
    P8red6Fef+Y/z11PkiF1aA==

# 2D LGCP fitting (sf)

    WAoAAAACAAQFAgACAwAAAAAOAAAABb/jCNZYO5mDQAtPqPCJIW8/ytHux+g7Kb/JKJJKBb1R
    v++BgRD8VYA=

---

    WAoAAAACAAQFAgACAwAAAAAOAAAABT/8SYedodjqP+OA5s8ws6U/6LBi1gMiLD/pfrKl1tWw
    P/R7efxhv4A=

---

    WAoAAAACAAQFAgACAwAAAAAOAAAABL/W/TEYi4bWv4nkmJPX5Ia/ypmHVl02HL/fyfyioJmg
    

---

    WAoAAAACAAQFAQACAwAAAAAOAAAABD/lvtZhqfQWP+tl+nXWUrw/9LjvMOUXAEABA9yVDC5v
    

# 2D LGCP fitting (sp)

    WAoAAAACAAQFAgACAwAAAAAOAAAABb/jCNZYO5mDQAtPqPCJIW8/ytHux+g7Kb/JKJJKBb1R
    v++BgRD8VYA=

---

    WAoAAAACAAQFAgACAwAAAAAOAAAABT/8SYedodjqP+OA5s8ws6U/6LBi1gMiLD/pfrKl1tWw
    P/R7efxhv4A=

---

    WAoAAAACAAQFAgACAwAAAAAOAAAABL/W/TEYi4bWv4nkmJPX5Ia/ypmHVl02HL/fyfyioJmg
    

---

    WAoAAAACAAQFAQACAwAAAAAOAAAABD/lvtZhqfQWP+tl+nXWUrw/9LjvMOUXAEABA9yVDC5v
    

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapper_collect.R
\name{bm_collect}
\alias{bm_collect}
\alias{bru_mapper_collect}
\alias{[.bm_collect}
\alias{[.bru_mapper_collect}
\title{Mapper for concatenated variables}
\usage{
bm_collect(mappers, hidden = FALSE)

bru_mapper_collect(...)

\method{[}{bm_collect}(x, i, drop = TRUE)

\method{[}{bru_mapper_collect}(x, i, drop = TRUE)
}
\arguments{
\item{mappers}{A list of \code{bru_mapper} objects}

\item{hidden}{\code{logical}, set to \code{TRUE} to flag that the mapper is to be used
as a first level input mapper for \code{INLA::f()} in a model that requires making
only the first mapper visible to \code{INLA::f()} and \code{INLA::inla.stack()}, such
as for "bym2" models, as activated by the \code{inla_f} argument to \code{ibm_n},
\code{ibm_values}, and \code{ibm_jacobian}. Set to \code{FALSE} to always access the full
mapper, e.g. for \code{rgeneric} models}

\item{...}{Arguments passed on to \code{\link[=bm_scale]{bm_scale()}}}

\item{x}{object from which to extract element(s)}

\item{i}{indices specifying element(s) to extract}

\item{drop}{logical;
For \verb{[.bm_collect}, whether to extract an individual mapper when
\code{i} identifies a single element. If \code{FALSE}, a list of sub-mappers is
returned (suitable e.g. for creating a new \code{bm_collect} object).
Default: \code{TRUE}}
}
\value{
\itemize{
\item \code{[}-indexing a \code{bm_collect} extracts a subset
\code{bm_collect} object (for drop \code{FALSE}) or an individual sub-mapper
(for drop \code{TRUE}, and \code{i} identifies a single element)
}
}
\description{
Constructs a concatenated collection mapping
}
\examples{
(m <- bm_collect(list(
  a = bm_index(2),
  b = bm_index(3)
), hidden = FALSE))
ibm_eval2(m, list(a = c(1, 2), b = c(1, 3, 2)), 1:5)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bru_mapper}
\alias{bru_mapper}
\alias{bru_mapper_define}
\title{Constructors for \code{bru_mapper} objects}
\usage{
bru_mapper(...)

bru_mapper_define(mapper, new_class = NULL, remove_class = "list", ...)
}
\arguments{
\item{\dots}{Arguments passed on to sub-methods, or used for special
purposes, see details for each function below.}

\item{mapper}{For \code{bru_mapper_define}, a prototype mapper object, see
Details.}

\item{new_class}{If non-\code{NULL}, this is added at the front of the class
definition}

\item{remove_class}{If non-\code{NULL}, this class or classes is removed from the
class definition before adding the \code{new_class} names. Default is \code{"list"}.}
}
\value{
A \code{bru_mapper} object.
}
\description{
Constructors for \code{bru_mapper} objects
}
\section{Functions}{
\itemize{
\item \code{bru_mapper()}: Generic mapper S3 constructor, used for constructing mappers for special
objects. See below for details of the default constructor
\code{\link[=bru_mapper_define]{bru_mapper_define()}} that can be used to define new mapper clesses in user
code. To extracting mappers for latent component models, see
\code{\link[=bru_get_mapper]{bru_get_mapper()}}.

\item \code{bru_mapper_define()}: Adds the \code{new_class} and \code{"bru_mapper"} class names to
the inheritance list for the input \code{mapper} object, unless the object
already inherits from these.

To register mapper classes and methods in scripts, use \code{.S3method()}
to register the methods, e.g.
\code{.S3method("ibm_jacobian", "my_mapper_class", ibm_jacobian.my_mapper_class)}.

In packages with \code{Suggests: inlabru}, add method information for delayed
registration, e.g.:

\if{html}{\out{<div class="sourceCode">}}\preformatted{#' @rawNamespace S3method(inlabru::bru_get_mapper, inla_rspde)
#' @rawNamespace S3method(inlabru::ibm_n, bru_mapper_inla_rspde)
#' @rawNamespace S3method(inlabru::ibm_values, bru_mapper_inla_rspde)
#' @rawNamespace S3method(inlabru::ibm_jacobian, bru_mapper_inla_rspde)
}\if{html}{\out{</div>}}

or before each method, use \verb{@exportS3Method}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{#' @exportS3Method inlabru::bru_get_mapper
}\if{html}{\out{</div>}}

etc., which semi-automates it.

}}
\examples{
mapper <- bm_index(5)
ibm_jacobian(mapper, input = c(1, 3, 4, 5, 2))
}
\seealso{
\link{bru_mapper_generics} for generic methods,
the individual mapper pages for special method implementations, and
\link{bru_get_mapper} for hooks to extract mappers from latent model object
class objects.

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_harmonics}
\alias{bm_harmonics}
\alias{bru_mapper_harmonics}
\title{Mapper for cos/sin functions}
\usage{
bm_harmonics(order = 1, scaling = 1, intercept = TRUE, interval = c(0, 1))

bru_mapper_harmonics(...)
}
\arguments{
\item{order}{For \code{bm_harmonics}, specifies the maximum \code{cos}/\code{sin}
order. (Default 1)}

\item{scaling}{For \code{bm_harmonics}, specifies an optional vector of
scaling factors of length \code{intercept + order}, or a common single scalar.}

\item{intercept}{logical; For \code{bm_harmonics}, if \code{TRUE}, the first
basis function is a constant. (Default \code{TRUE})}

\item{interval}{numeric length-2 vector specifying a domain interval.
Default \code{c(0, 1)}.}

\item{\dots}{Arguments passed on to \code{\link[=bm_harmonics]{bm_harmonics()}}}
}
\description{
Constructs a mapper for \code{cos}/\code{sin} functions of orders 1 (if
\code{intercept} is \code{TRUE}, otherwise 0) through \code{order}. The total number of
basis functions is \code{intercept + 2 * order}.

Optionally, each order can be given a non-unit scaling, via the \code{scaling}
vector, of length \code{intercept + order}. This can be used to
give an effective spectral prior. For example, let

\if{html}{\out{<div class="sourceCode">}}\preformatted{scaling = 1 / (1 + (0:4)^2)
x <- seq(0, 1, length.out = 11)
bmh1 = bm_harmonics(order = 4, interval = c(0, 1))
u1 <- ibm_eval(
  bmh1,
  input = x,
  state = rnorm(9, sd = rep(scaling, c(1, 2, 2, 2, 2)))
)
}\if{html}{\out{</div>}}

Then, with

\if{html}{\out{<div class="sourceCode">}}\preformatted{bmh2 = bm_harmonics(order = 4, scaling = scaling)
u2 = ibm_eval(bmh2, input = x, state = rnorm(9))
}\if{html}{\out{</div>}}

the stochastic properties of \code{u1} and \code{u2} will be the same, with
\code{scaling^2} determining the variance for each frequency contribution.

The period for the first order harmonics is shifted and scaled to match
\code{interval}.
}
\examples{
m <- bm_harmonics(2)
ibm_eval2(m, input = c(0, pi / 4, pi / 2, 3 * pi / 4), 1:5)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R
\name{bru_comp}
\alias{bru_comp}
\alias{component}
\alias{bru_component}
\alias{bru_comp.character}
\title{Latent model component construction}
\usage{
bru_comp(...)

bru_component(...)

\method{bru_comp}{character}(
  object,
  main = NULL,
  weights = NULL,
  ...,
  model = NULL,
  mapper = NULL,
  main_layer = NULL,
  main_selector = NULL,
  n = NULL,
  values = NULL,
  season.length = NULL,
  nrow = NULL,
  ncol = NULL,
  copy = NULL,
  weights_layer = NULL,
  weights_selector = NULL,
  group = 1L,
  group_mapper = NULL,
  group_layer = NULL,
  group_selector = NULL,
  ngroup = NULL,
  control.group = NULL,
  replicate = 1L,
  replicate_mapper = NULL,
  replicate_layer = NULL,
  replicate_selector = NULL,
  nrep = NULL,
  marginal = NULL,
  A.msk = deprecated(),
  .envir = parent.frame(),
  envir_extra = NULL
)
}
\arguments{
\item{\dots}{Parameters passed on to other methods}

\item{object}{A character label for the component}

\item{main}{\code{main} takes an R expression that evaluates to where the latent variables
should be evaluated (coordinates, indices, continuous scalar (for rw2 etc)).
Arguments starting with weights, group, replicate behave similarly to main,
but for the corresponding features of \code{INLA::f()}. \code{main} supports \code{tidy}
evaluation expressions, with \code{.data} and \code{.env} pronouns.}

\item{weights, weights_layer, weights_selector}{Optional specification of effect scaling weights.
Same syntax as for \code{main}.}

\item{model}{Either one of "const" (same as "offset"), "factor_full",
"factor_contrast", "linear",
"fixed", or a model name or
object accepted by INLA's \code{f} function. If set to NULL, then "linear" is used
for vector inputs, and "fixed" for matrix input (converted internally to
an iid model with fixed precision)}

\item{mapper}{Information about how to do the mapping from the values evaluated in \code{main},
and to the latent variables. Auto-detects spde model objects in model and
extracts the mesh object to use as the mapper, and auto-generates mappers
for indexed models. (Default: NULL, for auto-determination)}

\item{main_layer, main_selector}{The \verb{_layer} input should evaluate to a numeric index or character name or
vector of which
layer/variable to extract from a covariate data object given in \code{main}.
(Default: NULL if \verb{_selector} is given. Otherwise the effect component name,
if it exists in the covariate object, and otherwise the first column of
the covariate data frame)

The \verb{_selector} value should be a character name of a variable
whose contents determines which layer to extract from a covariate for each
data point. (Default: NULL)}

\item{n}{The number of latent variables in the model. Should be auto-detected
for most or all models. Default: NULL, for auto-detection. Models with
matrix input for \code{Cmatrix} or \code{graph} will use the matrix size.
An error is given if it realises it can't figure it out by itself.}

\item{values}{Specifies for what covariate/index values INLA should build
the latent model. Normally generated internally based on the mapping details.
(Default: NULL, for auto-determination)}

\item{season.length}{Passed on to \code{INLA::f()} for model \code{"seasonal"}
(TODO: check if this parameter is still fully handled)}

\item{nrow, ncol}{Number of rows and columns for \code{model} types
"rw2d", "rw2diid", and "matern2d". Default is \code{NULL}.}

\item{copy}{character; label of other component that this component should
be a copy of. If the \code{fixed = FALSE}, a scaling constant is estimated, via a
hyperparameter. If \code{fixed = TRUE}, the component scaling is fixed, by
default to 1; for fixed scaling, it's more efficient to express the scaling
in the predictor expression instead of making a copy component.}

\item{group, group_mapper, group_layer, group_selector, ngroup}{Optional specification of kronecker/group model indexing.}

\item{control.group}{\code{list} of kronecker/group model parameters, currently
passed directly on to \code{INLA::f}}

\item{replicate, replicate_mapper, replicate_layer, replicate_selector, nrep}{Optional specification of indices for an independent
replication model. Same syntax as for \code{main}}

\item{marginal}{May specify a \code{bm_marginal()} mapper,
that is applied before scaling by \code{weights}.}

\item{A.msk}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} and has no effect.}

\item{.envir}{Evaluation environment. Can later be accessed via
\code{\link[=bru_comp_env]{bru_comp_env()}}. Default: \code{parent.frame()}}

\item{envir_extra}{Environment for storing \code{INLA::f()} argument values.
If NULL, an new environment is created. Can be accessed via
\code{\link[=bru_comp_env_extra]{bru_comp_env_extra()}}}
}
\description{
Similar to \code{glm()}, \code{gam()} and \code{inla()}, \code{\link[=bru]{bru()}} models can be constructed
via a formula-like syntax, where each latent effect is specified. However, in
addition to the parts of the syntax compatible with \code{INLA::inla}, \code{bru}
components offer additional functionality which facilitates modelling, and
the predictor expression can be specified separately, allowing more complex
and non-linear predictors to be defined. The formula syntax is just a way to
allow all model components to be defined in a single line of code, but the
definitions can optionally be split up into separate component definitions.
See Details for more information.

The \code{bru_comp} methods all rely on the \code{\link[=bru_comp.character]{bru_comp.character()}}
method, that defines a model component with a given label/name. The user
usually doesn't need to call these methods directly, but can instead supply a
formula expression that can be interpreted by the
\code{\link[=bru_comp_list.formula]{bru_comp_list.formula()}} method, called inside \code{\link[=bru]{bru()}}.
}
\details{
As shorthand, \code{\link[=bru]{bru()}} will understand basic additive formulae describing
fixed effect models. For instance, the components specification \code{y ~ x} will
define the linear combination of an effect named \code{x} and an intercept to the
response \code{y} with respect to the likelihood family stated when calling
\code{\link[=bru]{bru()}}. Mathematically, the linear predictor \eqn{\eta}{eta} would be
written as

\deqn{\eta = \beta x + c,}{eta = beta * x + c,}

where:

\describe{
\item{\eqn{c}{c}}{is the \emph{intercept}}
\item{\eqn{x}{x}}{is a \emph{covariate}}
\item{\eqn{\beta}{beta}}{is a \emph{latent variable} associated with \eqn{x}{x}
and} \item{\eqn{\psi = \beta x}{psi = beta * x}}{ is called the \emph{effect} of
\eqn{x}{x}}
}

A problem that arises when using this kind of R formula is that it does not
clearly reflect the mathematical
formula. For instance, when providing the formula to inla, the resulting
object will refer to the random
effect \eqn{\psi = \beta * x } as \code{x}.
Hence, it is not clear when \code{x} refers to the covariate
or the effect of the covariate.

The \code{bru_comp.character} method is inlabru's equivalent to
\code{INLA}'s \code{f()} function but adds functionality that is unique to inlabru.

The \code{main}, \code{weights}, \code{group}, \code{replicate}, and \verb{*_layer} arguments
support \code{tidy} evaluation expressions, with \code{.data} and \code{.env} pronouns.
}
\section{Functions}{
\itemize{
\item \code{bru_component()}: Backwards compatibility alias for \code{bru_comp()}

}}
\section{Naming random effects}{


In INLA, the \code{f()} notation is used to define more complex models, but
a simple linear effect model can also be expressed as

\itemize{\item{\code{formula = y ~ f(x, model = "linear")},}}

where \code{f()} is the inla specific function to set up random effects of all
kinds. The underlying predictor would again be \eqn{\eta = \beta * x + c} but
the result of fitting the model would state \code{x} as the random effect's name.
bru allows rewriting this formula in order to explicitly state the name of
the random effect and the name of the associated covariate. This is achieved
by replacing \code{f} with an arbitrary name that we wish to assign to the effect,
e.g.

\itemize{\item{\code{components = y ~ psi(x, model = "linear")}.}}

Being able to discriminate between \eqn{x} and \eqn{\psi} is relevant because
of two functionalities bru offers. The formula parameters of both \code{\link[=bru]{bru()}} and
the prediction method \link{predict.bru} are interpreted in the mathematical
sense. For instance, \code{predict} may be used to analyze the analytical
combination of the covariate \eqn{x} and the intercept using

\itemize{\item{\verb{predict(fit, data.frame(x=2)), ~ exp(psi + Intercept)}.}}

which corresponds to the mathematical expression \ifelse{html}{\out{e
<sup>&#946; + c</sup>}}{\eqn{e^{x \beta + c}}}.

On the other hand, predict may be used to only look at a transformation of
the latent variable \eqn{\beta_\psi}

\itemize{\item{\code{predict(fit, NULL, ~ exp(psi_latent))}.}}

which corresponds to the mathematical expression \ifelse{html}{\out{e
<sup>&#946;</sup>}}{\eqn{e^{\beta}}}.
}

\examples{
# As an example, let us create a linear component. Here, the component is
# called "myLinearEffectOfX" while the covariate the component acts on is
# called "x". Note that a list of components is returned because the
# formula may define multiple components

cmp <- bru_comp_list(~ myLinearEffectOfX(main = x, model = "linear"))
summary(cmp)
# Equivalent shortcuts:
cmp <- bru_comp_list(~ myLinearEffectOfX(x, model = "linear"))
cmp <- bru_comp_list(~ myLinearEffectOfX(x))
# Individual component
cmp <- bru_comp("myLinearEffectOfX", main = x, model = "linear")
summary(cmp)
\donttest{
if (bru_safe_inla()) {
  # As an example, let us create a linear component. Here, the component is
  # called "myEffectOfX" while the covariate the component acts on is called
  # "x":

  cmp <- bru_comp("myEffectOfX", main = x, model = "linear")
  summary(cmp)

  # A more complicated component:
  cmp <- bru_comp("myEffectOfX",
    main = x,
    model = INLA::inla.spde2.matern(fmesher::fm_mesh_1d(1:10))
  )

  # Compound fixed effect component, where x and z are in the input data.
  # The formula will be passed on to MatrixModels::model.Matrix:
  cmp <- bru_comp("eff", ~ -1 + x:z, model = "fixed")
  summary(cmp)
}
}

}
\seealso{
\code{\link[=bru_input]{bru_input()}}, \code{\link[=summary.bru_comp]{summary.bru_comp()}}

Other component constructors: 
\code{\link{bru_comp_list}()}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com} and
Finn Lindgren \email{Finn.Lindgren@gmail.com}
}
\concept{component constructors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_log}
\alias{bru_log}
\alias{bru_log.character}
\alias{bru_log.bru_log}
\alias{bru_log.iinla}
\alias{bru_log.bru}
\alias{format.bru_log}
\alias{print.bru_log}
\alias{as.character.bru_log}
\alias{[.bru_log}
\alias{c.bru_log}
\alias{length.bru_log}
\title{Access methods for \code{bru_log} objects}
\usage{
bru_log(x = NULL, verbosity = NULL)

\method{bru_log}{character}(x, verbosity = NULL)

\method{bru_log}{bru_log}(x, verbosity = NULL)

\method{bru_log}{iinla}(x, verbosity = NULL)

\method{bru_log}{bru}(x, verbosity = NULL)

\method{format}{bru_log}(x, ..., timestamp = TRUE, verbosity = FALSE)

\method{print}{bru_log}(x, ..., timestamp = TRUE, verbosity = FALSE)

\method{as.character}{bru_log}(x, ...)

\method{[}{bru_log}(x, i)

\method{c}{bru_log}(...)

\method{length}{bru_log}(x)
}
\arguments{
\item{x}{An object that is, contains, or can be converted to,
a \code{bru_log} object. If \code{NULL}, refers to the global \code{inlabru} log.}

\item{verbosity}{integer value for limiting the highest verbosity level being
returned.}

\item{\dots}{further arguments passed to or from other methods.}

\item{timestamp}{If \code{TRUE}, include the timestamp of each message. Default
\code{TRUE}.}

\item{i}{indices specifying elements to extract. If \code{character}, denotes
the sequence between bookmark \code{i} and the next bookmark (or the end of the
log if \code{i} is the last bookmark)}
}
\value{
\code{bru_log} A \code{bru_log} object, containing a
character vector of log messages, and potentially a vector of bookmarks.
}
\description{
Access method for \code{bru_log} objects.
Note: Up to version \verb{2.8.0}, \code{bru_log()} was a deprecated alias for
\code{bru_log_message()}. When running on \verb{2.8.0} or earlier, use \code{bru_log_get()}
to access the global log, and \code{cat(fit$bru_iinla$log, sep = "\\n")} to print a
stored estimation object log.
After version \verb{2.8.0}, use \code{bru_log()} to access the global log, and
\code{bru_log(fit)} to access a stored estimation log.
}
\section{Methods (by generic)}{
\itemize{
\item \code{format(bru_log)}: Format a \code{bru_log} object for printing.
If \code{verbosity} is \code{TRUE}, include the verbosity level of each message.

\item \code{print(bru_log)}: Print a \code{bru_log} object with \code{cat(x, sep = "\\n")}.
If \code{verbosity} is \code{TRUE}, include the verbosity level of each message.

\item \code{as.character(bru_log)}: Convert \code{bru_log} object to a plain \code{character} vector

\item \code{[}: Extract a subset of a \code{bru_log} object

\item \code{c(bru_log)}: Concatenate several \code{bru_log} or \code{character} objects
into a \code{bru_log} object.

\item \code{length(bru_log)}: Obtain the number of log entries
into a \code{bru_log} object.

}}
\section{Functions}{
\itemize{
\item \code{bru_log()}: Extract stored log messages. If non-\code{NULL}, the
\code{verbosity} argument determines the maximum verbosity level of the messages
to extract.

}}
\examples{
bru_log(verbosity = 2L)
format(bru_log())

bru_log(verbosity = 2L)
print(bru_log(), timestamp = TRUE, verbosity = TRUE)

}
\seealso{
Other inlabru log methods: 
\code{\link{bru_log_bookmark}()},
\code{\link{bru_log_message}()},
\code{\link{bru_log_new}()},
\code{\link{bru_log_offset}()},
\code{\link{bru_log_reset}()}
}
\concept{inlabru log methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexagon_tiling.R
\name{cv_hex}
\alias{cv_hex}
\title{Hexagon tiling of region}
\usage{
cv_hex(samplers, cellsize = 0.5, n_group = 3, ...)
}
\arguments{
\item{samplers}{A sf object containing region for which
partitions to be created}

\item{cellsize}{hexagon cellsize, see \code{\link[sf:st_make_grid]{sf::st_make_grid()}} description}

\item{n_group}{number of cv folds.}

\item{\dots}{Passed on to \code{fm_nonconvex_hull()}, e.g. \code{resolution}}
}
\value{
a hexagonal partition covering the \code{samplers} object, with a \code{group}
integer variable indicating the fold assignment, 1, 2, ..., \code{n_group}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Both the interface and function name
may change in future versions.

Partitions the region based on the hexagon tiling

Input:
}
\examples{
if (interactive()) {
  bnd <- gorillas_sf$boundary
  hex_cv <- cv_hex(bnd, cellsize = 0.5, n_group = 3, resolution = 100)
  plot(hex_cv)

  chess <- cv_partition(bnd, resolution = 0.5, chess = TRUE)
  plot(chess$white)
}

}
\author{
Man Ho Suen
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_sp.R
\name{spoly}
\alias{spoly}
\title{Convert a data.frame of boundary points into a SpatialPolgonsDataFrame}
\usage{
spoly(
  data,
  cols = colnames(data)[1:2],
  crs = fm_crs(),
  to.crs = NULL,
  format = c("sp", "sf")
)
}
\arguments{
\item{data}{A data.frame of points describing the boundary of the polygon
(unique points, no holes)}

\item{cols}{Column names of the x and y coordinates within the data}

\item{crs}{Coordinate reference system of the points}

\item{to.crs}{Coordinate reference system for the \code{SpatialLines}/\code{sf} ouput.}

\item{format}{Format of the output object. Either "sp" (default) or "sf"}
}
\value{
\code{sp::SpatialPolygonsDataFrame} or \link[sf:sf]{sf::sf}
}
\description{
A polygon can be described as a sequence of points defining the polygon's
boundary. When given such a sequence (anti clockwise!) this function creates
a \code{SpatialPolygonsDataFrame} or \code{sf} holding the polygon decribed. By
default, the first two columns of \code{data} are assumed to define the x and y
coordinates of the points. This behaviour can be changed using the \code{cols}
parameter, which points out the names of the columns holding the coordinates.
The coordinate reference system of the resulting spatial polygon can be set
via the \code{crs} parameter. Posterior conversion to a different CRS is supported
using the \code{to.crs} parameter.
}
\examples{
\donttest{
# Create data frame of boundary points (anti clockwise!)
pts <- data.frame(
  x = c(1, 2, 1.7, 1.3),
  y = c(1, 1, 2, 2)
)

# Convert to sf
pol <- spoly(pts, format = "sf")

if (require(ggplot2, quietly = TRUE)) {
  # Plot it!
  ggplot() +
    gg(pol)
}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{evaluate_model}
\alias{evaluate_model}
\alias{evaluate_state}
\title{Evaluate or sample from a posterior result given a model and locations}
\usage{
evaluate_model(
  model,
  state,
  data = NULL,
  data_extra = NULL,
  input = NULL,
  comp_simple = NULL,
  predictor = NULL,
  format = NULL,
  used = NULL,
  n_pred = NULL,
  ...
)

evaluate_state(
  model,
  result,
  property = "mode",
  n = 1,
  seed = 0L,
  num.threads = NULL,
  internal_hyperpar = FALSE,
  ...
)
}
\arguments{
\item{model}{A \link{bru} model}

\item{state}{list of state lists, as generated by \code{\link[=evaluate_state]{evaluate_state()}}}

\item{data}{A \code{list}, \code{data.frame}, or \code{Spatial*DataFrame}, with coordinates
and covariates needed to evaluate the predictor.}

\item{data_extra}{Additional data for the predictor evaluation}

\item{input}{Precomputed inputs list for the components}

\item{comp_simple}{Precomputed \link{bm_list} of simplified mappers for the
components}

\item{predictor}{A formula or a \link{bru_pred_expr} expression to be evaluated
given the posterior or for each sample thereof. The default (\code{NULL})
returns a \code{data.frame} containing the sampled effects. In case of a formula
the right hand side is used for evaluation.}

\item{format}{character; determines the storage format of predictor output.
Available options:
\itemize{
\item \code{"auto"} If the first evaluated result is a vector or single-column matrix,
the "matrix" format is used, otherwise "list".
\item \code{"matrix"} A matrix where each column contains the evaluated predictor
expression for a state.
\item \code{"list"} A list where each element contains the evaluated predictor
expression for a state.
}}

\item{used}{A \code{\link[=bru_used]{bru_used()}} object, or NULL (default)}

\item{n_pred}{integer. If provided, scalar predictor results are expanded to
vectors of length \code{n_pred}.}

\item{\dots}{Additional arguments passed on to \code{inla.posterior.sample}}

\item{result}{A \code{bru} object from \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}}}

\item{property}{Property of the model components to obtain value from.
Default: "mode". Other options are "mean", "0.025quant", "0.975quant",
"sd" and "sample". In case of "sample" you will obtain samples from the
posterior (see \code{n} parameter). If \code{result} is \code{NULL}, all-zero vectors are
returned for each component.}

\item{n}{Number of samples to draw.}

\item{seed}{If seed != 0L, the random seed}

\item{num.threads}{Specification of desired number of threads for parallel
computations. Default NULL, leaves it up to INLA.
When seed != 0, overridden to "1:1:1"}

\item{internal_hyperpar}{logical; If \code{TRUE}, return hyperparameter properties
on the internal scale. Currently ignored when \code{property="sample"}.
Default is \code{FALSE}.}
}
\description{
Evaluate or sample from a posterior result given a model and locations
}
\details{
\itemize{
\item \code{evaluate_model} is a wrapper to evaluate model state, A-matrices,
effects, and predictor, all in one call.
}

\itemize{
\item \code{evaluate_state} evaluates model state properties or samples
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{multiplot}
\alias{multiplot}
\title{Multiple ggplots on a page.}
\source{
\url{http://www.cookbook-r.com/Graphs/Multiple_graphs_on_one_page_(ggplot2)/}
}
\usage{
multiplot(..., plotlist = NULL, cols = 1, layout = NULL)
}
\arguments{
\item{\dots}{Comma-separated \code{ggplot} objects.}

\item{plotlist}{A list of \code{ggplot} objects - an alternative to the
comma-separated argument above.}

\item{cols}{Number of columns of plots on the page.}

\item{layout}{A matrix specifying the layout. If present, \code{cols} is ignored.
If the layout is something like \code{matrix(c(1,2,3,3), nrow=2, byrow=TRUE)},
then plot 1 will go in the upper left, 2 will go in the upper right, and 3
will go all the way across the bottom.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} in favour of the \code{patchwork} package;
see the example below.

Renders multiple ggplots on a single page.
}
\examples{
if (require("ggplot2", quietly = TRUE)) {
  df <- data.frame(x = 1:10, y = cos(1:10), z = sin(1:10))
  pl1 <- ggplot(data = df) +
    geom_line(mapping = aes(x, y), color = "red")
  pl2 <- ggplot(data = df) +
    geom_line(mapping = aes(x, z), color = "blue")
  pl3 <- ggplot(data = df) +
    geom_path(mapping = aes(y, z), color = "magenta")

  multiplot(
    pl1, pl2, pl3,
    layout = rbind(c(1, 2), c(3, 3))
  )

  # Recommended alternative using the patchwork package:
  if (require("patchwork")) {
    (pl1 | pl2) / pl3
  }
}
}
\author{
David L. Borchers \email{dlb@st-andrews.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{inlabru_group_cv_block_conversion}
\alias{inlabru_group_cv_block_conversion}
\title{Convert node-block indices into groups of node indices}
\usage{
inlabru_group_cv_block_conversion(block, max_block, per_node)
}
\arguments{
\item{block}{An integer vector of 1-based block indices, one for each node.}

\item{max_block}{Precomputed maximal value of \code{block}. Must be at least
as large as \code{max(block)}. Block indices larger than \code{max_block} are ignored.}

\item{per_node}{Logical; if \code{FALSE}, the function returns a list with
one element for each block. If \code{TRUE}, the function returns a list with
one element for each node.}
}
\value{
If \code{per_node} is \code{false}, a list of integer vectors, where each
vector contains the 1-based node indices of nodes belonging to the
corresponding block. If \code{per_node} is \code{true}, a list of integer vectors,
where each vector contains the 1-based node indices of nodes belonging to
the same block as the corresponding node; \code{b[i] <- a[block[j]]}
}
\description{
Converts a vector of block indices into a list of
node index vectors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_log_message}
\alias{bru_log_message}
\alias{bru_log_abort}
\alias{bru_log_warn}
\title{Add a log message}
\usage{
bru_log_message(
  ...,
  domain = NULL,
  appendLF = TRUE,
  verbosity = 1L,
  allow_verbose = TRUE,
  verbose = NULL,
  verbose_store = NULL,
  x = NULL
)

bru_log_abort(
  msg,
  ...,
  domain = NULL,
  appendLF = TRUE,
  verbosity = 1L,
  allow_verbose = TRUE,
  verbose = FALSE,
  verbose_store = NULL,
  call = rlang::caller_env(),
  .frame = rlang::caller_env()
)

bru_log_warn(
  msg,
  ...,
  domain = NULL,
  appendLF = TRUE,
  verbosity = 1L,
  allow_verbose = TRUE,
  verbose = FALSE,
  verbose_store = NULL,
  call = rlang::caller_env(),
  .frame = rlang::caller_env()
)
}
\arguments{
\item{\dots}{For \code{bru_log_message()}, zero or more objects passed on to
\code{\link[base:message]{base::.makeMessage()}}. For \code{bru_log_abort()} and \code{bru_log_warn()},
passed on to \code{rlang::abort()} and \code{rlang::warn()}.}

\item{domain}{Domain for translations, passed on to \code{\link[base:message]{base::.makeMessage()}}}

\item{appendLF}{logical; whether to add a newline to the message. Only
used for verbose output.}

\item{verbosity}{numeric value describing the verbosity level of the message}

\item{allow_verbose}{Whether to allow verbose output. Must be set to FALSE
until the options object has been initialised.}

\item{verbose}{logical, numeric, or \code{NULL}; local override for verbose
output. If \code{NULL}, the global option \code{bru_verbose} or default value is used.
If \code{FALSE}, no messages are printed. If \code{TRUE}, messages with \code{verbosity}
\eqn{\le 1}{<=1}
are printed. If numeric, messages with \code{verbosity} \eqn{\le}{<=} \code{verbose}
are printed.}

\item{verbose_store}{Same as \code{verbose}, but controlling what messages are
stored in the global log object. Can be controlled via the
\code{bru_verbose_store} with \code{\link[=bru_options_set]{bru_options_set()}}.}

\item{x}{A \code{bru_log} object. If \code{NULL}, refers to the global \code{inlabru} log.}

\item{msg}{character; passed to \code{\link[base:message]{base::.makeMessage()}}}

\item{call}{The calling environment.}

\item{.frame}{The throwing context, for when \code{.internal} is \code{TRUE}}
}
\value{
\code{bru_log_message} returns \code{invisible(x)}, where \code{x} is the updated \code{bru_log}
object, or \code{NULL}.
}
\description{
Adds a log message.
}
\section{Functions}{
\itemize{
\item \code{bru_log_abort()}: Store a log message and throw an error.

\item \code{bru_log_warn()}: Store a log message and throw a warning.

}}
\examples{
if (interactive()) {
  code_runner <- function() {
    bru_options_set_local(
      # Show messages up to and including level 2 (default 0)
      bru_verbose = 2,
      # Store messages to an including level 3 (default Inf, storing all)
      bru_verbose_store = 3
    )

    bru_log_bookmark("bookmark 1")
    bru_log_message("Test message 1", verbosity = 1)
    bru_log_message("Test message 2", verbosity = 2)
    bru_log_bookmark("bookmark 2")
    bru_log_message("Test message 3", verbosity = 3)
    bru_log_message("Test message 4", verbosity = 4)

    invisible()
  }
  message("Run code")
  code_runner()
  message("Check log from bookmark 1")
  print(bru_log()["bookmark 1"])
  message("Check log from bookmark 2")
  print(bru_log()["bookmark 2"])
}
}
\seealso{
Other inlabru log methods: 
\code{\link{bru_log}()},
\code{\link{bru_log_bookmark}()},
\code{\link{bru_log_new}()},
\code{\link{bru_log_offset}()},
\code{\link{bru_log_reset}()}
}
\concept{inlabru log methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_is.R
\name{bru_is_linear}
\alias{bru_is_linear}
\alias{bru_is_linear.bru}
\alias{bru_is_linear.bru_info}
\alias{bru_is_linear.bru_model}
\alias{bru_is_linear.bru_obs}
\alias{bru_is_linear.bru_obs_list}
\alias{bru_is_linear.bru_pred_expr}
\alias{bru_is_linear.bru_comp_list}
\alias{bru_is_linear.bru_comp}
\alias{bru_is_linear.bru_mapper}
\title{Check for predictor linearity}
\usage{
bru_is_linear(x, ...)

\method{bru_is_linear}{bru}(x, ...)

\method{bru_is_linear}{bru_info}(x, ...)

\method{bru_is_linear}{bru_model}(x, ...)

\method{bru_is_linear}{bru_obs}(x, ...)

\method{bru_is_linear}{bru_obs_list}(x, ...)

\method{bru_is_linear}{bru_pred_expr}(x, ...)

\method{bru_is_linear}{bru_comp_list}(x, ...)

\method{bru_is_linear}{bru_comp}(x, ...)

\method{bru_is_linear}{bru_mapper}(x, ...)
}
\arguments{
\item{x}{A object containing a predictor definition}

\item{\dots}{Arguments passed on recursively.}
}
\value{
\code{TRUE} if the expression is detected to be linear, \code{FALSE}
otherwise.
}
\description{
Checks if a predictor expression is linear (or affine)
}
\examples{
bru_is_linear(new_bru_pred_expr(~ x + y))
bru_is_linear(new_bru_pred_expr(~ x * y))
bru_is_linear(bm_scale())
bru_is_linear(bm_logsumexp())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_conversion.R
\name{as_bru_comp}
\alias{as_bru_comp}
\alias{as_bru_comp_list}
\alias{as_bru_comp.bru_comp}
\alias{as_bru_comp_list.bru_comp_list}
\alias{as_bru_comp_list.bru_comp}
\alias{as_bru_comp_list.bru}
\alias{as_bru_comp_list.bru_info}
\alias{as_bru_comp_list.bru_model}
\alias{as_bru_comp_list.list}
\alias{as_bru_comp_list.formula}
\title{Conversion methods for \code{bru_comp} and \code{bru_comp_list} objects}
\usage{
as_bru_comp(x, ...)

as_bru_comp_list(x, ...)

\method{as_bru_comp}{bru_comp}(x, ...)

\method{as_bru_comp_list}{bru_comp_list}(x, ...)

\method{as_bru_comp_list}{bru_comp}(x, ...)

\method{as_bru_comp_list}{bru}(x, ...)

\method{as_bru_comp_list}{bru_info}(x, ...)

\method{as_bru_comp_list}{bru_model}(x, ...)

\method{as_bru_comp_list}{list}(x, ...)

\method{as_bru_comp_list}{formula}(x, ...)
}
\arguments{
\item{x}{An object to convert to \link{bru_comp} or \link{bru_comp_list}}

\item{\dots}{Additional arguments passed on to \code{\link[=bru_comp_list]{bru_comp_list()}}.}
}
\value{
An object of class \link{bru_comp_list}.
}
\description{
Methods for converting to \code{bru_comp} and \code{bru_comp_list}
objects.
}
\section{Functions}{
\itemize{
\item \code{as_bru_comp_list(bru)}: Extract the component list from a \code{\link[=bru]{bru()}} object.

\item \code{as_bru_comp_list(bru_info)}: Extract the component list from a \code{\link[=bru_info]{bru_info()}}
object.

\item \code{as_bru_comp_list(bru_model)}: Extract the component list from a \code{\link[=bru_model]{bru_model()}}
object.

}}
\seealso{
\code{\link[=as_bru_obs]{as_bru_obs()}}, \code{\link[=as_bru_obs_list]{as_bru_obs_list()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_sp.R
\name{sline}
\alias{sline}
\title{Convert data frame to SpatialLinesDataFrame}
\usage{
sline(
  data,
  start.cols,
  end.cols,
  crs = fm_crs(),
  to.crs = NULL,
  format = c("sp", "sf")
)
}
\arguments{
\item{data}{A data.frame}

\item{start.cols}{Character array pointing out the columns of \code{data} that
hold the start points of the lines}

\item{end.cols}{Character array pointing out the columns of \code{data} that hold
the end points of the lines}

\item{crs}{Coordinate reference system of the original \code{data}}

\item{to.crs}{Coordinate reference system for the output.}

\item{format}{Format of the output object. Either \code{"sp"} (default) or \code{"sf"}}
}
\value{
\code{sp::SpatialLinesDataFrame} or \link[sf:sf]{sf::sf}
}
\description{
A line in 2D space is defined by a start and an end point, each associated
with 2D coordinates. This function takes a \code{data.frame} as input and assumes
that each row defines a line in space. In order to do so, the data frame must
have at least four columns and the \code{start.cols} and \code{end.cols} parameters
must be used to point out the names of the columns that define the start and
end coordinates of the line. The data is then converted to a
\code{SpatialLinesDataFrame} \code{DF}. If a coordinate reference system \code{crs} is
provided it is attached to \code{DF}. If also \code{to.crs} is provided, the coordinate
system of \code{DF} is transformed accordingly. Additional columns of the input
data, e.g. covariates, are retained and attached to \code{DF}.
}
\examples{
\donttest{
# Create a data frame defining three lines
lns <- data.frame(
  xs = c(1, 2, 3), ys = c(1, 1, 1), # start points
  xe = c(2, 3, 4), ye = c(2, 2, 2)
) # end points


# Conversion to sf without CRS
spl <- sline(lns,
  start.cols = c("xs", "ys"),
  end.cols = c("xe", "ye"),
  format = "sf"
)

if (require(ggplot2, quietly = TRUE)) {
  # Plot the lines
  ggplot() +
    gg(spl)
}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_testthat.R
\name{local_testthat}
\alias{local_testthat}
\alias{local_bru_testthat_assign}
\alias{local_bru_testthat_tolerances}
\alias{local_bru_options_set}
\alias{local_basic_intercept_testdata}
\alias{local_basic_fixed_effect_testdata}
\alias{local_bru_safe_inla}
\alias{local_bru_testthat_setup}
\title{Unit test helpers}
\usage{
local_bru_testthat_assign(x, values, envir = parent.frame())

local_bru_testthat_tolerances(
  tolerances = c(1e-04, 0.01, 0.1),
  envir = parent.frame()
)

local_bru_options_set(..., .reset = FALSE, envir = parent.frame())

local_basic_intercept_testdata()

local_basic_fixed_effect_testdata()

local_bru_safe_inla(multicore = FALSE, quietly = TRUE, envir = parent.frame())

local_bru_testthat_setup(envir = parent.frame())
}
\arguments{
\item{x}{character; Name of variable to assign to}

\item{values}{the object to assign to \code{x}}

\item{envir}{environment for exit handlers}

\item{tolerances}{numeric vector of length 3; \verb{[lowtol, midtol, hitol]}}

\item{.reset}{For \code{local_bru_options_set}, logical indicating if the global
override options list should be emptied before setting the new option(s).}

\item{multicore}{logical; if \code{TRUE}, multiple cores are allowed, and the INLA
\code{num.threads} option is not checked or altered. Default: \code{FALSE}, multicore
not allowed (used for examples and unit tests).}

\item{quietly}{logical; if \code{TRUE}, prints diagnostic messages. A message is
always printed if the INLA \code{num.threads} option is altered, regardless of
the \code{quietly} argument. Default: TRUE.}
}
\value{
\code{local_bru_options_set()} returns a copy of the global override
options (not including the defaults), invisibly.
}
\description{
Local helper functions for package unit tests
}
\section{Functions}{
\itemize{
\item \code{local_bru_testthat_assign()}: Assign local variable. Useful for easy cleanup
of global workspace with \code{withr::deferred_run()} when running tests
interactively.

\item \code{local_bru_testthat_tolerances()}: Assign test tolerances
Assign local tolerance variables. Useful for easy cleanup
of global workspace with \code{withr::deferred_run()} when running tests
interactively.

\item \code{local_bru_options_set()}: Wrapper for \code{\link[=bru_options_set_local]{bru_options_set_local()}},
to locally override the global package options.

\item \code{local_bru_safe_inla()}: Tests should set num.threads = "1:1:1" to ensure
within-system repeatability by calling \code{local_bru_safe_inla()}; see also
\code{\link[=bru_safe_inla]{bru_safe_inla()}}

\item \code{local_bru_testthat_setup()}: Initialise environment for tests.
Assigns tolerance variables.
To be called either at the top of a testfile, or inside tests.
Does \emph{not} call \code{\link[=local_bru_safe_inla]{local_bru_safe_inla()}}, since that may invoke a skip and
should be called inside each test that relies on INLA.

}}
\seealso{
\code{\link[=bru_options_set_local]{bru_options_set_local()}}, \code{\link[=bru_options_default]{bru_options_default()}},
\code{\link[=bru_options_get]{bru_options_get()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_response_size}
\alias{bru_response_size}
\alias{bru_response_size.default}
\alias{bru_response_size.list}
\alias{bru_response_size.inla.surv}
\alias{bru_response_size.bru_obs}
\alias{bru_response_size.bru_obs_list}
\alias{bru_response_size.bru_info}
\alias{bru_response_size.bru}
\title{Response size queries}
\usage{
bru_response_size(object)

\method{bru_response_size}{default}(object)

\method{bru_response_size}{list}(object)

\method{bru_response_size}{inla.surv}(object)

\method{bru_response_size}{bru_obs}(object)

\method{bru_response_size}{bru_obs_list}(object)

\method{bru_response_size}{bru_info}(object)

\method{bru_response_size}{bru}(object)
}
\arguments{
\item{object}{An object from which to extract response size(s).}
}
\value{
An \code{integer} vector.
}
\description{
Extract the number of response values from \code{bru} and related objects.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_response_size(default)}: Extract the number of observations from an
object supporting \code{NROW()}.

\item \code{bru_response_size(list)}: Extract the number of observations from an
object supporting \code{NROW(object[[1]])}.

\item \code{bru_response_size(inla.surv)}: Extract the number of observations from an
\code{inla.surv} object.

\item \code{bru_response_size(bru_obs)}: Extract the number of observations from a
\code{bru_obs} object.

\item \code{bru_response_size(bru_obs_list)}: Extract the number of observations from a
\code{bru_obs_list} object, as a vector with one value per observation model.

\item \code{bru_response_size(bru_info)}: Extract the number of observations from a
\code{bru_info} object, as a vector with one value per observation model.

\item \code{bru_response_size(bru)}: Extract the number of observations from a
\code{bru} object, as a vector with one value per observation model.

}}
\examples{
bru_response_size(
  bru_obs(y ~ 1, data = data.frame(y = rnorm(10)), family = "gaussian")
)
}
\seealso{
\code{\link[=like]{like()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_reparam}
\alias{bm_reparam}
\title{Mapper for reparameterising mapper states}
\usage{
bm_reparam(mapper, B)
}
\arguments{
\item{mapper}{A \code{bru_mapper} object}

\item{B}{a square or rectangular basis conversion matrix}
}
\description{
Creates a mapper for handling basis conversions. Functionally
equivalent to \code{bm_pipe(list(bm_matrix(ncol(B)), mapper))}, but with an
internally stored matrix input \code{B} for efficiency, and allowing the mapper
\code{input} format to be identical to that of the original \code{mapper}.
}
\examples{
# 2->2 reparameterisation; (u1,u2) -> (u1, u1 + u2)
(m <- bm_reparam(bm_index(2), B = matrix(c(1, 1, 0, 1), 2, 2)))

# 2->3 reparameterisation; (u1,u2) -> (u1, u2, u1+u2)
# This is an example of a low-rank representation of a higher-dimensional
# state vector.
(m <- bm_reparam(bm_index(3), B = cbind(c(1, 0, 1), c(0, 1, 1))))

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru}
\alias{bru}
\alias{bru_rerun}
\alias{summary.bru}
\alias{print.summary_bru}
\alias{print.bru}
\title{Convenient model fitting using (iterated) INLA}
\usage{
bru(components = ~Intercept(1), ..., options = list(), .envir = parent.frame())

bru_rerun(result, options = list())

\method{summary}{bru}(object, verbose = FALSE, ...)

\method{print}{summary_bru}(x, ...)

\method{print}{bru}(x, ...)
}
\arguments{
\item{components}{Latent component definitions, either as a \code{\link[=bru_comp_list]{bru_comp_list()}}
object, or a \code{formula}-like specification.
Also used to define a default linear additive predictor.  See
\code{\link[=bru_comp]{bru_comp()}} for details.}

\item{\dots}{Observation models, each constructed by a calling \code{\link[=bru_obs]{bru_obs()}},
or \code{\link[=bru_obs_list]{bru_obs_list()}}.

Alternatively, for backwards compatibility, may be named parameters that
can be passed to a single \code{\link[=bru_obs]{bru_obs()}} call. These arguments will be
evaluated before calling \code{\link[=bru_obs]{bru_obs()}}, in order to detect if they already
are \code{bru_obs} objects. This means that special arguments that are only
available in the context of \code{data} or \code{response_data} (such as \code{Ntrials})
will only work properly in direct calls to \code{\link[=bru_obs]{bru_obs()}}.}

\item{options}{A \link{bru_options} options object or a list of options passed
on to \code{\link[=bru_options]{bru_options()}}}

\item{.envir}{Environment for component evaluation (for when a non-formula
specification is used)}

\item{result}{A previous estimation object of class \code{bru}}

\item{object}{An object obtained from a \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}} call}

\item{verbose}{logical; If \code{TRUE}, include more details of the
component definitions. If \code{FALSE}, only show basic component
definition information. Default: \code{FALSE}}

\item{x}{An object to be printed}
}
\value{
bru returns an object of class "bru". A \code{bru} object inherits
from \code{INLA::inla} (see the inla documentation for its properties) and
adds additional information stored in the \code{bru_info} field.
}
\description{
This method is a wrapper for \code{INLA::inla} and provides
multiple enhancements.

\itemize{
\item
Easy usage of spatial covariates and automatic construction of inla
projection matrices for (spatial) SPDE models. This feature is
accessible via the \code{components} parameter. Practical examples on how to
use spatial data by means of the components parameter can also be found
by looking at the \code{\link[=lgcp]{lgcp()}} function's documentation.
\item
Constructing multiple observation models is straightforward. See
\code{\link[=bru_obs]{bru_obs()}} for more information on how to provide additional
models to \code{bru} using the \code{...} parameter list.
\item
Support for non-linear predictors. See example below.
\item
Log Gaussian Cox process (LGCP) inference is
available by using the \code{"cp"} family or (even easier) by using the
\code{\link[=lgcp]{lgcp()}} function.
}
}
\section{Methods (by generic)}{
\itemize{
\item \code{summary(bru)}: Takes a fitted \code{bru} object produced by \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}}
and creates various summaries from it, including the summary output from
the corresponding \code{INLA::sumary()} method.
The \dots arguments are passed on to component summary functions, see
\code{\link[=summary.bru_comp]{summary.bru_comp()}}.

\item \code{print(bru)}: Print a summary of a \code{bru} object.

}}
\section{Functions}{
\itemize{
\item \code{bru_rerun()}: Continue the optimisation from a previously computed estimate. The estimation
\code{options} list can be given new values to override the original settings.

To rerun with a subset of the data (e.g. for cross validation or prior
sampling), use \code{\link[=bru_set_missing]{bru_set_missing()}} to set all or part of the response data
to \code{NA} before calling \code{bru_rerun()}.

}}
\examples{
\donttest{
if (bru_safe_inla()) {
  # Simulate some covariates x and observations y
  input.df <- data.frame(x = cos(1:10))
  input.df <- within(input.df, {
    y <- 5 + 2 * x + rnorm(10, mean = 0, sd = 0.1)
  })

  # Fit a Gaussian likelihood model
  fit <- bru(y ~ x + Intercept(1), family = "gaussian", data = input.df)

  # Obtain summary
  fit$summary.fixed
}


if (bru_safe_inla()) {
  # Alternatively, we can use the bru_obs() function to construct the
  # likelihood:

  lik <- bru_obs(
    family = "gaussian",
    formula = y ~ x + Intercept,
    data = input.df
  )
  fit <- bru(~ x + Intercept(1), lik)
  fit$summary.fixed
}

# An important addition to the INLA methodology is bru's ability to use
# non-linear predictors. Such a predictor can be formulated via bru_obs()'s
# \code{formula} parameter. The z(1) notation is needed to ensure that
# the z component should be interpreted as single latent variable and not
# a covariate:

if (bru_safe_inla()) {
  z <- 2
  input.df <- within(input.df, {
    y <- 5 + exp(z) * x + rnorm(10, mean = 0, sd = 0.1)
  })
  lik <- bru_obs(
    family = "gaussian", data = input.df,
    formula = y ~ exp(z) * x + Intercept
  )
  fit <- bru(~ z(1) + Intercept(1), lik)

  # Check the result (z posterior should be around 2)
  fit$summary.fixed
}
}

}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchical_basis.R
\name{make_hierarchical_mesh_basis}
\alias{make_hierarchical_mesh_basis}
\alias{inla.spde2.pcmatern_B}
\title{Make hierarchical mesh basis functions}
\usage{
make_hierarchical_mesh_basis(mesh, forward = TRUE, method = NULL, alpha = 1)

inla.spde2.pcmatern_B(mesh, ..., B)
}
\arguments{
\item{mesh}{A \code{fmesher} object.}

\item{forward}{logical; if \code{TRUE} (default), the basis functions are
constructed in forward order; otherwise in reverse order.}

\item{method}{Character; one of \code{"laplace"} (default), \code{"graphdistance"} or
\code{"graphlaplace"}. The method used to construct the basis functions. See
Details.}

\item{alpha}{Numeric; only used if \code{method = "laplace"}. The approximate
Laplacian power used to construct the basis functions. Default is
\code{(1 + fm_manifold_dim(mesh)) / 2}, which gives an approximately linear
function along the shortest path. Must be in the range \verb{[1, 2]}. See
Details.}
}
\value{
A sparse matrix of class \code{dgCMatrix}, where each column
corresponds to a basis function.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Construct hierarchical
basis functions. This is highly experimental and may change or be removed
in future versions.
}
\details{
The hierarchical basis functions are constructed by iteratively
adding basis functions centered at the point furthest away from the
previously added basis functions. The radius of each basis function is
equal to the distance to the previously added basis functions at the time
of addition. Three methods are available to construct the basis functions:
\describe{
\item{\code{laplace}}{The basis functions are constructed using a combination
of
two operators, the Laplacian and the squared Laplacian (the biharmonic
operator). This approximates the Laplacian raised to the power \code{alpha}. The
default value for \code{alpha} gives an approximately linear function along the
shortest path.one based on the standard FEM stiffness matrix, and one based
on the graph Laplacian. The \code{alpha} parameter controls the weight between
the two operators, with \code{alpha = 1} using only the Laplacian and \code{alpha = 2}
using only the squared Laplacian.}
\item{\code{graphdistance}}{The basis functions are constructed using a simple
neighbour distance-based approach, where the basis function value
decreases by equal amounts for each traversed edge.}
\item{\code{graphlaplace}}{The basis functions are constructed using the
graph Laplacian only.}
}
}
\section{Functions}{
\itemize{
\item \code{inla.spde2.pcmatern_B()}: Construct a pcmatern model with basis change

}}
\examples{
m <- fmesher::fmexample$mesh
B <- make_hierarchical_mesh_basis(m, method = "laplace2", alpha = 1.5)

\donttest{
m <- fmesher::fm_subdivide(fmesher::fmexample$mesh, 1)
B <- list()
for (methodB in c(
  "laplace", "laplace2",
  "graphdistance",
  "graphlaplace", "graphlaplace2"
)) {
  B[[methodB]] <- make_hierarchical_mesh_basis(
    m,
    method = methodB, alpha = 1.5
  )
}
if (require("ggplot2") && require("patchwork")) {
  print(
    ggplot(
      data =
        data.frame(
          idx = seq_len(ncol(B$laplace)),
          nnz = as.vector(Matrix::colSums(0 != B$laplace)),
          laplace = as.vector(Matrix::colSums(B$laplace)),
          laplace2 = as.vector(Matrix::colSums(B$laplace2)),
          distance = as.vector(Matrix::colSums(B$graphdistance)),
          graph = as.vector(Matrix::colSums(B$graphlaplace)),
          graph2 = as.vector(Matrix::colSums(B$graphlaplace2))
        )
    ) +
      geom_point(aes(x = idx, y = nnz, color = "nnz")) +
      geom_point(aes(x = idx, y = laplace, color = "laplace")) +
      geom_point(aes(x = idx, y = laplace2, color = "laplace2")) +
      geom_point(aes(x = idx, y = distance, color = "graphdistance")) +
      geom_point(aes(x = idx, y = graph, color = "graphlaplace")) +
      scale_y_log10() +
      scale_x_log10()
  )

  idx <- seq_len(fmesher::fm_dof(m))
  idx <- seq_len(10)
  method0 <- "laplace"
  theta <- qr.solve(B[[method0]][, idx, drop = FALSE], m$loc[, 1])
  print(
    ggplot() +
      gg(m,
        col = B[[method0]][, idx, drop = FALSE] \%*\% theta - m$loc[, 1],
        nx = 60, ny = 60
      ) +
      fmesher::geom_fm(
        data = m,
        color = ggplot2::alpha("black", 0.1),
        alpha = 0
      ) +
      scale_fill_distiller(palette = "RdBu")
  )

  ev <- fmesher::fm_evaluator(m, dims = c(60, 60))
  df <- NULL
  for (method in c(
    "laplace", "laplace2",
    "graphdistance",
    "graphlaplace", "graphlaplace2"
  )) {
    fun_cumulative <- 1e-16
    for (k in 1:28) {
      fun_orig <- as.vector(B[[method]][, k, drop = FALSE])
      fun_cumulative <- fun_cumulative + fun_orig
      fun_norm <- fun_orig / fun_cumulative
      df <- rbind(
        df,
        data.frame(
          x = ev$lattice$loc[, 1],
          y = ev$lattice$loc[, 2],
          orig = as.vector(fmesher::fm_evaluate(ev, fun_orig)),
          norm = as.vector(fmesher::fm_evaluate(ev, fun_norm)),
          index = k,
          fun = paste0(
            "fun",
            formatC(k, width = 2, format = "d", flag = "0")
          ),
          method = method
        )
      )
    }
  }
  df$norm[df$fun == "fun01"] <- NA
  df$orig[df$orig == 0] <- NA
  df$norm[df$norm == 0] <- NA
  print(
    ggplot(data = df[df$index <= 4, ]) +
      geom_tile(aes(x, y, fill = orig),
        na.rm = TRUE
      ) +
      fmesher::geom_fm(
        data = m,
        color = ggplot2::alpha("black", 0.1),
        alpha = 0
      ) +
      geom_contour(aes(x, y, z = orig),
        breaks = seq_len(29 - 15) / (30 - 15), #* 2-0.5,
        na.rm = TRUE, col = "black", alpha = 0.5
      ) +
      geom_contour(aes(x, y, z = norm),
        breaks = seq_len(29 - 15) / (30 - 15), #* 2-0.5,
        na.rm = TRUE, col = "red", alpha = 0.5
      ) +
      scale_fill_distiller(palette = "RdBu", limits = c(0, 1)) + #* 2-0.5) +
      facet_wrap(vars(fun, method))
  )
}
}

}
\seealso{
\code{\link[=bru_get_mapper.inla_model_reparam]{bru_get_mapper.inla_model_reparam()}}
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.matrix}
\alias{gg.matrix}
\title{Geom for matrix}
\usage{
\method{gg}{matrix}(data, mapping = NULL, ...)
}
\arguments{
\item{data}{A \code{matrix} object.}

\item{mapping}{a set of aesthetic mappings created by \code{aes}. These are passed
on to \code{geom_tile}.}

\item{\dots}{Arguments passed on to \code{geom_tile}.}
}
\value{
A \code{geom_tile} with reversed y scale.
}
\description{
Creates a tile geom for plotting a matrix
}
\details{
Requires the \code{ggplot2} package.
}
\examples{
if (require("ggplot2", quietly = TRUE)) {
  A <- matrix(runif(100), nrow = 10)
  ggplot() +
    gg(A)
}
}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R, R/hierarchical_basis.R
\name{bru_get_mapper}
\alias{bru_get_mapper}
\alias{bru_get_mapper.inla.spde}
\alias{bru_get_mapper.inla.rgeneric}
\alias{bru_get_mapper.inla.cgeneric}
\alias{bru_get_mapper_safely}
\alias{bru_get_mapper.inla_model_reparam}
\title{Extract mapper information from INLA model component objects}
\usage{
bru_get_mapper(model, ...)

\method{bru_get_mapper}{inla.spde}(model, ...)

\method{bru_get_mapper}{inla.rgeneric}(model, ...)

\method{bru_get_mapper}{inla.cgeneric}(model, ...)

bru_get_mapper_safely(model, ...)

\method{bru_get_mapper}{inla_model_reparam}(model, ...)
}
\arguments{
\item{model}{A model component object}

\item{\dots}{Arguments passed on to other methods}
}
\value{
A \link{bru_mapper} object defined by the model component
}
\description{
The component definitions will automatically attempt to extract mapper
information from any model object by calling the generic \code{bru_get_mapper}.
Any class method implementation should return a \link{bru_mapper} object suitable
for the given latent model.
}
\details{
Before implementing your own \code{bru_get_mapper} method, check if there
is already a general method available that handles your model class, such as
\code{\link[=bru_get_mapper.inla.spde]{bru_get_mapper.inla.spde()}}, \code{\link[=bru_get_mapper.inla.rgeneric]{bru_get_mapper.inla.rgeneric()}}, and
\code{\link[=bru_get_mapper.inla.cgeneric]{bru_get_mapper.inla.cgeneric()}}.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_get_mapper(inla.spde)}: Extract an indexed mapper for
the \code{model$mesh} object contained in the model object,
which is assumed to be of a class supporting relevant \code{fmesher} methods.

\item \code{bru_get_mapper(inla.rgeneric)}: Returns the mapper given by a pre-computed mapper,
or an index mapper mapping the size of the model graph.

The easiest method to define a mapper for an \code{inla.rgeneric} model is to
store the mapper in the object.
Alternatively, define your model using a subclass and define a
\verb{bru_get_mapper.<subclass>} method that should return the
corresponding \code{bru_mapper} object.

The order of precedence for the mapper construction when calling
\code{bru_get_mapper(model)} has the following precedence:
\enumerate{
\item \verb{bru_get_mapper.<subclass>}, if \code{model} has a subclass, otherwise
\item \code{model[["mapper"]]} if that is \code{NULL}, and otherwise
\item \code{\link[=bm_index]{bm_index()}} using the size of the graph returned by \code{model[["f"]][["n"]]}
}

\item \code{bru_get_mapper(inla.cgeneric)}: Works the same as the method of \code{inla.rgeneric}

\item \code{bru_get_mapper(inla_model_reparam)}: Reparameterised inla model mapper, see
\code{\link[=inla.spde2.pcmatern_B]{inla.spde2.pcmatern_B()}}
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

}}
\section{Functions}{
\itemize{
\item \code{bru_get_mapper_safely()}: Tries to call the \code{bru_get_mapper},
and returns \code{NULL} if it fails (e.g. due to no available class method).
If the call succeeds and returns non-\code{NULL}, it checks that the object
inherits from the \code{bru_mapper} class, and gives an error if it does not.

}}
\examples{
if (bru_safe_inla()) {
  library(INLA)
  mesh <- fmesher::fm_rcdt_2d_inla(globe = 2)
  spde <- inla.spde2.pcmatern(mesh,
    prior.range = c(1, 0.5),
    prior.sigma = c(1, 0.5)
  )
  mapper <- bru_get_mapper(spde)
  ibm_n(mapper)
}
}
\seealso{
\link{bru_mapper} for mapper constructor methods, and
the individual mappers for specific implementation details.

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{summary.bru_obs}
\alias{summary.bru_obs}
\alias{summary.bru_obs_list}
\alias{print.summary_bru_obs}
\alias{print.summary_bru_obs_list}
\alias{print.bru_obs}
\alias{print.bru_obs_list}
\title{Summary and print methods for observation models}
\usage{
\method{summary}{bru_obs}(object, verbose = TRUE, ...)

\method{summary}{bru_obs_list}(object, verbose = TRUE, ...)

\method{print}{summary_bru_obs}(x, ...)

\method{print}{summary_bru_obs_list}(x, ...)

\method{print}{bru_obs}(x, ...)

\method{print}{bru_obs_list}(x, ...)
}
\arguments{
\item{object}{Object to operate on}

\item{verbose}{logical; If \code{TRUE}, include more details of the
component definitions. If \code{FALSE}, only show basic component
definition information. Default: \code{TRUE}}

\item{\dots}{Arguments passed on to other \code{summary} methods}

\item{x}{Object to be printed}
}
\description{
Summary and print methods for observation models
}
\examples{
obs <- bru_obs(y ~ ., data = data.frame(y = rnorm(10)))
summary(obs)
print(obs)

}
\seealso{
\code{\link[=bru_obs]{bru_obs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bru_mapper_generics}
\alias{bru_mapper_generics}
\title{Generic methods for bru_mapper objects}
\description{
A \code{bru_mapper} sub-class implementation must provide an
\code{\link[=ibm_jacobian]{ibm_jacobian()}} method. If the model size 'n' and definition
values 'values' are stored in the object itself, default methods \code{\link[=ibm_n]{ibm_n()}}
and \code{\link[=ibm_values]{ibm_values()}} are
available. Otherwise the
\code{\link[=ibm_n]{ibm_n()}} and \code{\link[=ibm_values]{ibm_values()}} methods also need to be provided.
}
\seealso{
\link{bru_mapper} for constructor methods, and
\link{bru_get_mapper} for hooks to extract mappers from latent model object
class objects.

Other mapper methods: 
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_aggregate}
\alias{bm_aggregate}
\alias{bru_mapper_aggregate}
\title{Mapper for aggregation}
\usage{
bm_aggregate(rescale = FALSE, n_block = NULL, type = NULL)

bru_mapper_aggregate(...)
}
\arguments{
\item{rescale}{logical; For \code{bm_aggregate} and \code{bm_logsumexp},
specifies if the blockwise sums should be normalised by the blockwise weight
sums or not:
\itemize{
\item \code{FALSE}: (default) Straight weighted sum, no rescaling.
\item \code{TRUE}: Divide by the sum of the weight values within each block.
This is useful for integration averages, when the given weights are plain
integration weights. If the weights are \code{NULL} or all ones, this is
the same as dividing by the number of entries in each block.
}}

\item{n_block}{Predetermined number of output blocks. If \code{NULL}, overrides
the maximum block index in the inputs. The priority order is \code{input$n_block},
the mapper definition \code{n_block}, then \code{max(input$block)}.}

\item{type}{character; if non-NULL, overrides the \code{rescale} argument, and
constructs an aggregation mapper of the given type instead. Supported
values are "sum", "average" (for regular \code{\link[=bm_aggregate]{bm_aggregate()}}),
"logsumexp", "logaverageexp" (for \code{\link[=bm_logsumexp]{bm_logsumexp()}}), and
"logitaverage" (for \code{\link[=bm_logitaverage]{bm_logitaverage()}}).}

\item{\dots}{Arguments passed on to \code{\link[=bm_aggregate]{bm_aggregate()}}}
}
\description{
Constructs a mapper
that aggregates elements of the input state, so it can be used e.g.
for weighted summation or integration over blocks of values.
}
\examples{
m <- bm_aggregate()
ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4), 11:14)
ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4, n_block = 3), 11:14)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_log_new}
\alias{bru_log_new}
\title{Create a \code{bru_log} object}
\usage{
bru_log_new(x = NULL, bookmarks = NULL)
}
\arguments{
\item{x}{An optional character vector of log messages, or \code{data.frame}
with columns \code{message}, \code{timestamp}, and \code{verbosity}.}

\item{bookmarks}{An optional \code{integer} vector of named bookmarks
message in \code{x}.}
}
\description{
Create a \code{bru_log} object, by default empty.
}
\examples{
x <- bru_log_new()
x <- bru_log_message("Test message", x = x)
print(x)
}
\seealso{
Other inlabru log methods: 
\code{\link{bru_log}()},
\code{\link{bru_log_bookmark}()},
\code{\link{bru_log_message}()},
\code{\link{bru_log_offset}()},
\code{\link{bru_log_reset}()}
}
\concept{inlabru log methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_input.R
\name{bru_input_text}
\alias{bru_input_text}
\alias{bru_input_text.bru_input}
\alias{bru_input_text.bru_comp}
\alias{bru_input_text.bru_comp_list}
\alias{bru_input_text.bru_mapper}
\alias{bru_input_text.bm_pipe}
\alias{bru_input_text.bm_multi}
\alias{bru_input_text.bm_collect}
\alias{bru_input_text.bm_repeat}
\alias{bru_input_text.bm_sum}
\title{Extract input definitions as text}
\usage{
bru_input_text(...)

\method{bru_input_text}{bru_input}(x, .envir = parent.frame(), ...)

\method{bru_input_text}{bru_comp}(x, ..., label = x$label)

\method{bru_input_text}{bru_comp_list}(x, ...)

\method{bru_input_text}{bru_mapper}(x, ..., label = "<unknown>")

\method{bru_input_text}{bm_pipe}(x, ..., label = "<unknown>")

\method{bru_input_text}{bm_multi}(x, ..., label = "<unknown>")

\method{bru_input_text}{bm_collect}(x, ..., label = "<unknown>")

\method{bru_input_text}{bm_repeat}(x, ..., label = "<unknown>")

\method{bru_input_text}{bm_sum}(x, ..., label = "<unknown>")
}
\arguments{
\item{...}{Passed on to sub-methods.}

\item{x}{A \code{bru_input} object, or other object for recursive evaluation.}

\item{.envir}{environment in which to evaluate the input expression. Default
is \code{parent.frame()}}

\item{label}{character; optional label used to identify the object in
informational messages.}
}
\value{
\itemize{
\item \code{bru_input(bru_comp)}: A character vector of mapper input values.
}

\itemize{
\item \code{bru_input_text(bru_comp_list)}: A list of mapper input definition
text strings, with one entry for each component.
}
}
\description{
Extract input definitions from the \link{bru_input} information stored in bru
model components.  Primarily intended for diagnostic use, as the output does
not currently contain all information, such as the layer and selector
information for spatial covariates.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_input_text(bru_input)}: Extract the input definitions from a \code{bru_input}
object, as a named character vector

\item \code{bru_input_text(bru_comp)}: Extract the input definition as a named character
vector.

\item \code{bru_input_text(bru_mapper)}: Extract the input definitions associated with a
\code{bru_mapper}.

}}
\examples{
(inp <- new_bru_input(x, "LABEL"))
bru_input_text(inp)
if (bru_safe_inla()) {
  bru_input_text(bru_comp("x", cos(y)))
}

}
\seealso{
\code{\link[=bru_input]{bru_input()}}

\link{ibm_input}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_is.R
\name{bru_is_additive}
\alias{bru_is_additive}
\alias{bru_is_additive.character}
\alias{bru_is_additive.expression}
\alias{bru_is_additive.formula}
\alias{bru_is_additive.bru_pred_expr}
\alias{bru_is_additive.bru_obs}
\alias{bru_is_additive.bru_obs_list}
\title{Check for predictor expression additivity}
\usage{
bru_is_additive(x, ...)

\method{bru_is_additive}{character}(x, ..., verbose = FALSE)

\method{bru_is_additive}{expression}(x, ..., verbose = FALSE)

\method{bru_is_additive}{formula}(x, ..., verbose = FALSE)

\method{bru_is_additive}{bru_pred_expr}(x, ...)

\method{bru_is_additive}{bru_obs}(x, ...)

\method{bru_is_additive}{bru_obs_list}(x, ...)
}
\arguments{
\item{x}{A predictor \code{expression}, \code{formula}, or parse information
\code{data.frame}.}

\item{\dots}{Arguments passed on recursively.}

\item{verbose}{logical; if \code{TRUE}, print diagnostic parsing information.}
}
\value{
\code{TRUE} if the expression is detected to be additive, \code{FALSE}
otherwise.
}
\description{
Checks if a predictor expression is additive or not
}
\examples{
bru_is_additive(~ x + y)
bru_is_additive(~ x * y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapper_sum.R
\name{bm_sum}
\alias{bm_sum}
\alias{bru_mapper_sum}
\alias{[.bm_sum}
\alias{[.bru_mapper_sum}
\title{Mapper for adding multiple mappers}
\usage{
bm_sum(mappers, single_input = FALSE)

bru_mapper_sum(...)

\method{[}{bm_sum}(x, i, drop = TRUE)

\method{[}{bru_mapper_sum}(x, i, drop = TRUE)
}
\arguments{
\item{mappers}{A list of \code{bru_mapper} objects.}

\item{single_input}{logical. If \code{TRUE}, the input is passed to all
sub-mappers. Otherwise, the input should be a list, data.frame, or matrix.
If the \code{mappers} list has named entries, the \code{input} can reference their
corresponding sub-mapper using its name.}

\item{\dots}{Arguments passed on to \code{\link[=bm_sum]{bm_sum()}}}

\item{x}{object from which to extract element(s)}

\item{i}{indices specifying element(s) to extract}

\item{drop}{logical;
For \verb{[.bm_sum}, whether to extract an individual mapper when
\code{i} identifies a single element. If \code{FALSE}, a list of sub-mappers is
returned (suitable e.g. for creating a new \code{bm_sum} object).
Default: \code{TRUE}}
}
\value{
A \code{bm_sum} object.

\itemize{
\item \code{[}-indexing a \code{bm_sum} extracts a subset
\code{bm_sum} object (for drop \code{FALSE}) or an individual sub-mapper
(for drop \code{TRUE}, and \code{i} identifies a single element)
}
}
\description{
Defines a mapper that adds the effects of each submapper.
The \code{ibm_n()} method returns the sum of \code{ibm_n(mappers[[k]])}, and
\code{ibm_values()} returns \code{seq_len(ibm_n(mapper))}.
}
\examples{
(m <- bm_sum(list(a = bm_index(3), b = bm_index(2))))
ibm_n(m)
ibm_values(m)
ibm_jacobian(m, list(a = 1:3, b = c(1, 1, 2)))
ibm_eval(
  m,
  list(a = 1:3, b = c(1, 1, 2)),
  seq_len(ibm_n(m))
)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatstat.R
\name{spatial.to.ppp}
\alias{spatial.to.ppp}
\title{Convert SpatialPoints and boundary polygon to spatstat ppp object}
\usage{
spatial.to.ppp(points, samplers)
}
\arguments{
\item{points}{A \code{SpatialPoints[DataFrame]} object describing the point
pattern.}

\item{samplers}{A \code{SpatialPolygons[DataFrame]} object describing the
observation window.}
}
\value{
A spatstat \code{spatstat} \code{ppp} object
}
\description{
Spatstat point pattern objects consist of points and an observation windows.
This function uses a \code{SpatialPoints} object and a \code{SpatialPolygon} object to
generate the points and the window. Lastly, the \code{ppp()} function is called to
create the \code{ppp} object.
}
\examples{
\donttest{
if (require("spatstat.geom") &&
  bru_safe_sp() &&
  require("sp") &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE)) {
  # Load Gorilla data

  gorillas <- gorillas_sp()

  # Use nest locations and survey boundary to create a spatstat ppp object

  gp <- spatial.to.ppp(gorillas$nests, gorillas$boundary)
  class(gp)

  # Plot it

  plot(gp)
}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spde.R
\name{materncov.bands}
\alias{materncov.bands}
\title{Matern correlation or covariance function approximate credible bands.}
\usage{
materncov.bands(
  manifold,
  dist,
  log.range,
  log.variance = NULL,
  alpha = 2,
  quantile = 0.95,
  n = 64,
  S1.L = NULL
)
}
\arguments{
\item{manifold}{Either "R1", "S1", "R2", or "S2", from
\code{fm_manifold(mesh)}, or an object supported by
\code{\link[fmesher:fm_manifold]{fmesher::fm_manifold()}}.}

\item{dist}{A vector of distances at which to calculate the
covariances/correlations}

\item{log.range}{A scalar or a list (mean, sd), such as produced by
\code{inla.spde.result(...)$summary.log.range.nominal[[1]][c("mean","sd")]}}

\item{log.variance}{Either \code{NULL}, a scalar, or vector of the
same type as for log.range. When \code{NULL}, the correlations
are calculated instead of the covariances.}

\item{alpha}{The SPDE operator order. Default 2.}

\item{quantile}{The target credible probability. Default 0.95.}

\item{n}{The number of parameter combinations to use for the
approximation. Default 64.}

\item{S1.L}{For \code{manifold} \code{"S1"}, give the length of the
cyclic interval}
}
\value{
A list with estimated covariance or correlation (when \code{log.variance}
is \code{NULL}) functions:
\item{lower}{An approximate lower bound for the \code{quantile} credible region}
\item{median}{The function for for the approximate median parameters
quantile}
\item{upper}{An approximate upper bound for the \code{quantile} credible region}
}
\description{
Evaluate the covariance function for an inla.spde
object. Calculates the posterior distribution of the range,
log(range), variance, or log(variance) parameter of a model's
SPDE component. Can also calculate Matern correlation or covariance
function.
}
\details{
Uses a Gaussian assumption for the internal model parameters, and finds a
region in parameter space with approximately \code{quantile} probability.
}
\author{
Finn Lindgren \email{Finn.Lindgren@ed.ac.uk}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{expand_labels}
\alias{expand_labels}
\title{Expand labels}
\usage{
expand_labels(labels, expand, suffix)
}
\arguments{
\item{labels}{character vector; original labels}

\item{expand}{character vector; subset of labels to expand}

\item{suffix}{character; the suffix to add to the labels selected by \code{expand}}
}
\value{
a vector of labels with suffix appended to the selected labels
}
\description{
Expand labels
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.Spatial}
\alias{gg.Spatial}
\alias{gg.SpatialPoints}
\alias{gg.SpatialLines}
\alias{gg.SpatialPolygons}
\alias{gg.SpatialGridDataFrame}
\alias{gg.SpatialPixelsDataFrame}
\alias{gg.SpatialPixels}
\title{Geoms for sp Spatial objects}
\usage{
\method{gg}{SpatialPoints}(data, mapping = NULL, crs = NULL, ...)

\method{gg}{SpatialLines}(data, mapping = NULL, crs = NULL, ...)

\method{gg}{SpatialPolygons}(data, mapping = NULL, crs = NULL, ...)

\method{gg}{SpatialGridDataFrame}(data, ...)

\method{gg}{SpatialPixelsDataFrame}(data, mapping = NULL, crs = NULL, mask = NULL, ...)

\method{gg}{SpatialPixels}(data, ...)
}
\arguments{
\item{data}{A \verb{Spatial*} object.}

\item{mapping}{Aesthetic mappings created by \code{aes} used to update the default
mapping. Uness specified otherwise below, the default mapping is

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ggplot2::aes(
  x = .data[[sp::coordnames(data)[1]]],
  y = .data[[sp::coordnames(data)[2]]]
)
}\if{html}{\out{</div>}}}

\item{crs}{A \code{sp::CRS} object defining the coordinate system to project the
data to before plotting.}

\item{\dots}{Arguments passed on to \verb{geom_*}.}

\item{mask}{A \code{sp::SpatialPolygons} object defining the region that is
plotted.}
}
\value{
A \code{geom_point}, \code{geom_segment}, \code{geom_sf}, \code{geom_tile},
or a list of \code{ggplot} geomes
}
\description{
Methods for plotting sp spatial objects with \code{ggplot2}.
}
\section{Functions}{
\itemize{
\item \code{gg(SpatialPoints)}: Geom for \code{SpatialPoints} objects.
This function coerces the \code{SpatialPoints} into a \code{data.frame} and uses
\code{geom_point} to plot the points. Requires the \code{ggplot2} package.

\item \code{gg(SpatialLines)}: Geom for SpatialLines objects.

Extracts start and end points of the lines and calls \code{geom_segment} to plot
lines between them.

\code{mapping}: Aesthetic mappings created by \code{ggplot2::aes} or
\code{ggplot2::aes_} used to update the default mapping. The default mapping is

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ggplot2::aes(
  x = .data[[sp::coordnames(data)[1]]],
  y = .data[[sp::coordnames(data)[2]]],
  xend = .data[[paste0("end.", sp::coordnames(data)[1])]],
  yend = .data[[paste0("end.", sp::coordnames(data)[2])]])
}\if{html}{\out{</div>}}

\item \code{gg(SpatialPolygons)}: Geom for SpatialPolygons objects.
Uses the \code{ggplot2::fortify()} function to turn the \code{SpatialPolygons} objects
into a \code{data.frame}. Then calls \code{geom_polygon} to plot the polygons.

Unless specified by the user,
the argument \code{alpha = 0.2} (alpha level for polygon filling) is added.

Up to version \verb{2.10.0}, the \code{ggpolypath} package was used to ensure proper
plotting for polygons, since the \code{ggplot2::geom_polygon} function doesn't
always handle geometries with holes properly. After \verb{2.10.0}, the object is
converted to \code{sf} format and passed on to \code{\link[=gg.sf]{gg.sf()}} instead, as \code{ggplot2}
version \verb{3.4.4} deprecated the internally used \code{ggplot2::fortify()} method
for \code{SpatialPolygons/DataFrame} objects.

\item \code{gg(SpatialGridDataFrame)}: Geom for SpatialGridDataFrame objects

Coerces input \code{SpatialGridDataFrame} to \code{SpatialPixelsDataFrame} and calls
\code{\link[=gg.SpatialPixelsDataFrame]{gg.SpatialPixelsDataFrame()}} to plot it.

\item \code{gg(SpatialPixelsDataFrame)}: Geom for SpatialPixelsDataFrame objects.

Coerces \code{SpatialPixelsDataFrame} input to \code{data.frame} and uses
\code{geom_tile} to plot it.

\code{mapping}: Aesthetic mappings created by \code{aes} used to update the default
mapping. The default mapping is

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ggplot2::aes(
  x = .data[[sp::coordnames(data)[1]]],
  y = .data[[sp::coordnames(data)[2]]],
  fill = .data[[names(data)[[1]]]]
)
}\if{html}{\out{</div>}}

\item \code{gg(SpatialPixels)}: Geom for SpatialPixels objects

Converts the input to \code{SpatialPoints} and calls
[gg.SpatialPoints()` to plot it.

}}
\examples{
\donttest{
if (require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  bru_safe_sp() &&
  require("sp")) {
  # Load Gorilla data

  gorillas <- inlabru::gorillas_sf

  gcov <- gorillas_sf_gcov()
  elev <- terra::as.data.frame(gcov$elevation, xy = TRUE)
  elev <- sf::as_Spatial(sf::st_as_sf(elev, coords = c("x", "y")))

  # Turn elevation covariate into SpatialGridDataFrame
  elev <- sp::SpatialPixelsDataFrame(elev, data = as.data.frame(elev))

  # Plot Gorilla elevation covariate provided as SpatialPixelsDataFrame.
  # The same syntax applies to SpatialGridDataFrame objects.

  ggplot() +
    gg(elev)

  # Add Gorilla survey boundary and nest sightings

  ggplot() +
    gg(elev) +
    gg(gorillas$boundary, alpha = 0.0, col = "red") +
    gg(gorillas$nests)

  # Load pantropical dolphin data

  mexdolphin <- inlabru::mexdolphin_sp()

  # Plot the pantropical survey boundary, ship transects, and dolphin
  # sightings

  ggplot() +
    gg(mexdolphin$ppoly) + # survey boundary as SpatialPolygon
    gg(mexdolphin$samplers) + # ship transects as SpatialLines
    gg(mexdolphin$points) # dolphin sightings as SpatialPoints

  # Change color

  ggplot() +
    gg(mexdolphin$ppoly, color = "green") + # survey boundary; SpatialPolygon
    gg(mexdolphin$samplers, color = "red") + # ship transects; SpatialLines
    gg(mexdolphin$points, color = "blue") # dolphin sightings; SpatialPoints


  # Visualize data annotations: line width by segment number

  names(mexdolphin$samplers) # 'seg' holds the segment number
  ggplot() +
    gg(mexdolphin$samplers, aes(color = seg))

  # Visualize data annotations: point size by dolphin group size

  names(mexdolphin$points) # 'size' holds the group size
  ggplot() +
    gg(mexdolphin$points, aes(size = size))
}
}

if (require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  bru_safe_sp()) {
  # Load Gorilla data

  gcov <- gorillas_sf_gcov()
  elev <- terra::as.data.frame(gcov$elevation, xy = TRUE)
  pxl <- sf::as_Spatial(sf::st_as_sf(elev, coords = c("x", "y")))

  # Turn elevation covariate into SpatialPixels
  pxl <- sp::SpatialPixels(pxl)

  # Plot the pixel centers
  ggplot() +
    gg(pxl, size = 0.1)
}

}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_index.R
\name{inla_subset_eval}
\alias{inla_subset_eval}
\alias{inla_subset_eval.bru_comp_list}
\title{Obtain inla index subset information}
\usage{
inla_subset_eval(...)

\method{inla_subset_eval}{bru_comp_list}(components, ...)
}
\arguments{
\item{\dots}{Passed on to submethods.}

\item{components}{A component list.}
}
\description{
Subsets for \code{INLA::f()} compatible indexing
}
\author{
Finn Lindgren \email{finn.lindgren@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_input.R
\name{bru_input}
\alias{bru_input}
\alias{new_bru_input}
\alias{bru_input.bru_input}
\alias{bru_input.bru_comp}
\alias{bru_input.bru_comp_list}
\alias{bru_input.bru_model}
\alias{bru_input.bru_obs}
\alias{bru_input.bru_obs_list}
\alias{bru_input.bru_mapper}
\alias{bru_input.bm_pipe}
\alias{bru_input.bm_multi}
\alias{bru_input.bm_collect}
\alias{bru_input.bm_repeat}
\alias{bru_input.bm_sum}
\title{Store and evaluate component inputs}
\usage{
new_bru_input(input, label = NULL, layer = NULL, selector = NULL, ...)

bru_input(...)

\method{bru_input}{bru_input}(
  x,
  data = NULL,
  mask = NULL,
  null.on.fail = FALSE,
  .envir = parent.frame(),
  ...
)

\method{bru_input}{bru_comp}(x, ..., label = x$label)

\method{bru_input}{bru_comp_list}(x, ...)

\method{bru_input}{bru_model}(x, lhoods, ...)

\method{bru_input}{bru_obs}(x, components, ...)

\method{bru_input}{bru_obs_list}(x, components, ...)

\method{bru_input}{bru_mapper}(x, ..., label = "<unknown>")

\method{bru_input}{bm_pipe}(x, ..., label = "<unknown>")

\method{bru_input}{bm_multi}(x, ..., label = "<unknown>")

\method{bru_input}{bm_collect}(x, ..., label = "<unknown>")

\method{bru_input}{bm_repeat}(x, ..., label = "<unknown>")

\method{bru_input}{bm_sum}(x, ..., label = "<unknown>")
}
\arguments{
\item{input}{An expression to be evaluated.}

\item{label}{character; optional label used to identify the object in
informational messages.}

\item{layer}{Optional expression that evaluates layer names or indices for
use with \code{\link[=eval_spatial]{eval_spatial()}}}

\item{selector}{character or integer; optional selector for use with
use with \code{\link[=eval_spatial]{eval_spatial()}}}

\item{\dots}{Passed on to sub-methods.}

\item{x}{A \code{bru_input} object, or other object for recursive evaluation.}

\item{data}{A \code{data.frame}, \code{tibble}, \code{sf}, \code{list}, or \verb{Spatial*} object of
covariates and/or point locations.}

\item{mask}{A \code{bru_data_mask} object, constructed internally.}

\item{null.on.fail}{logical; if \code{TRUE}, return \code{NULL} if the input evaluation
fails. If \code{FALSE} (default), return a vector of 1s and issue a warning
(only for deprecated use of \code{coordinates} without having loaded the \code{sp}
package), or stop with an error.}

\item{.envir}{environment in which to evaluate the input expression. Default
is \code{parent.frame()}}

\item{lhoods}{A \code{bru_obs_list} object containing all observations models.}

\item{components}{A \code{bru_comp_list} object containing all components
defined in the model.}
}
\value{
\itemize{
\item \code{bru_input(bru_comp)}: A list of mapper input values, formatted
for the full component mapper (of type \link{bm_pipe})
}

\itemize{
\item \code{bru_input(bru_comp_list)}: A list of mapper input values,
with one entry for each component.
}

\itemize{
\item \code{bru_input(bru_model)}: A list of mapper input values,
with one entry for each observation model, each containing a list
of inputs for the components used by the corresponding observation model.
}

\itemize{
\item \code{bru_input(bru_obs_list)}: A list of mapper input values,
with one entry for each observation model, each containing a list
of inputs for the components used by the corresponding observation model.
}

\itemize{
\item \code{bru_input(bru_obs_list)}: A list of mapper input values,
with one entry for each observation model, each containing a list
of inputs for the components used by the corresponding observation model.
}
}
\description{
Store and evaluate component inputs, including support for
non-standard-evaluation with \code{rlang}, automatic transfer to
function calls, \code{\link[=eval_spatial]{eval_spatial()}}, and \code{\link[MatrixModels:model.Matrix]{MatrixModels::model.Matrix()}}.

These functions are normally only called internally, but users may find
it useful to call them directly to experiment, and to make advanced
model definitions using \link{ibm_input_set}/\code{\link[=ibm_input_new]{ibm_input_new()}}.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_input(bru_input)}: Attempts to evaluate a component input (e.g. \code{main},
\code{group}, \code{replicate}, or \code{weight}), and process the results:
\enumerate{
\item If \code{rlang::eval_tidy()} failed, return NULL or map everything to 1
(see the \code{null.on.fail} argument). This should normally not
happen, unless the component use logic is incorrect,
(via \code{used})
leading to missing columns for a certain likelihood in a
multi-\code{bru_obs()} model.
\item If we obtain a function, apply the function to the data object
\item If we obtain an object supported by \code{\link[=eval_spatial]{eval_spatial()}}, extract the values
of that data frame at the point locations
\item If we obtain a formula, call \code{ModelMatrix::model.Matrix()}
\item Else we obtain a vector and return as-is. This happens when input
references a column of the data points, or some other complete expression
}

\item \code{bru_input(bru_model)}: Computes the component inputs for included components
for each observation model.

\item \code{bru_input(bru_mapper)}: Evaluate the input associated with a \code{bru_mapper}.

\item \code{bru_input(bm_pipe)}: Evaluate the inputs for each sub-mapper in a \code{bm_pipe}
object.

\item \code{bru_input(bm_multi)}: Evaluate the inputs for each sub-mapper in a \code{bm_multi}
object.

\item \code{bru_input(bm_collect)}: Evaluate the inputs for each sub-mapper in a
\code{bm_collect} object.

\item \code{bru_input(bm_repeat)}: Evaluate the inputs for the sub-mapper in a \code{bm_repeat}
object.

\item \code{bru_input(bm_sum)}: Evaluate the inputs for each sub-mapper in a \code{bm_sum}
object.

}}
\section{Functions}{
\itemize{
\item \code{new_bru_input()}: Create a \code{bru_input} object.

\item \code{bru_input()}: Evaluate inputs

}}
\section{Simple covariates and the input parameters}{


It is not unusual for a random effect act on a transformation of a covariate.
In other frameworks this would mean that the transformed covariate would have
to be calculated in advance and added to the data frame that is usually
provided via the \code{data} parameter. inlabru provides the option to do this
transformation automatically. For instance, one might be interested in the
effect of a covariate \eqn{x^2}. In inla and other frameworks this would
require to add a column \code{xsquared} to the input data frame and use the
formula

\itemize{\item{\code{formula = y ~ f(xsquared, model = "linear")},}}

In inlabru this can be achieved in several ways of using the \code{main} parameter
(\code{map} in version 2.1.13 and earlier), which does not need to be named.

\itemize{
\item{\code{components = y ~ psi(x^2, model = "linear")}}
\item{\code{components = y ~ psi(main = x^2, model = "linear")}}
\item{\code{components = y ~ psi(mySquareFun(x), model = "linear")},}
\item{\code{components = y ~ psi(myOtherSquareFun, model = "linear")},}
}

In the first example inlabru will interpret the map parameter as an
expression to be evaluated within the data provided. Since \eqn{x} is a known
covariate it will know how to calculate it. The second example is an
expression as well but it uses a function called \code{mySquareFun}. This function
is defined by user but has to be accessible within the work space when
setting up the components. The third example provides the function
\code{myOtherSquareFun}. In this case, inlabru will call the function as
\code{myOtherSquareFun(.data.)}, where \code{.data.} is the data provided via the
\code{\link[=bru_obs]{bru_obs()}} \code{data} parameter. The function needs to know what parts of the
data to use to construct the needed output. For example,

\if{html}{\out{<div class="sourceCode">}}\preformatted{myOtherSquareFun <- function(data) \{
  data[ ,"x"]^2
\}
}\if{html}{\out{</div>}}

Interactions can be handled by a formula input and \code{model = "fixed"}:
\code{components = y ~ 0 + name(~ 1 + x:z, model = "fixed")}
}

\section{Spatial Covariates}{


When fitting spatial models it is common to work with covariates that depend
on space, e.g. sea surface temperature or elevation. Although it is
straightforward to add this data to the input data frame or write a covariate
function like in the previous section there is an even more convenient way in
inlabru. Spatial covariates are often stored as \code{SpatialPixelsDataFrame},
\code{SpatialPixelsDataFrame} or \code{RasterLayer} objects. These can be provided
directly via the input expressions if they are supported by \code{\link[=eval_spatial]{eval_spatial()}},
and the \code{\link[=bru_obs]{bru_obs()}} data is an \code{sf} or \code{SpatialPointsDataFrame} object.
\code{inlabru} will then automatically evaluate and/or interpolate the covariate
at your data locations when using code like

\if{html}{\out{<div class="sourceCode">}}\preformatted{components = y ~ psi(mySpatialPixels, model = "linear")
}\if{html}{\out{</div>}}

For more precise control, use the the \code{layer} and \code{selector} arguments (see
\code{\link[=bru_comp]{bru_comp()}}), or call \code{\link[=eval_spatial]{eval_spatial()}} directly, e.g.:

\if{html}{\out{<div class="sourceCode">}}\preformatted{components = y ~ psi(eval_spatial(mySpatialPixels, where = .data.),
                     model = "linear")
}\if{html}{\out{</div>}}
}

\section{Coordinates}{


A common spatial modelling component when using inla are SPDE models. An
important feature of inlabru is that it will automatically calculate the so
called A-matrix (a component model matrix) which maps SPDE values at the mesh
vertices to values at the data locations. For this purpose, the input can be
set to \code{coordinates}, which is the \code{sp} package function that extracts point
coordinates from the \code{SpatialPointsDataFrame} that was provided as input to
\code{\link[=bru_obs]{bru_obs()}}. The code for this would look as follows:

\if{html}{\out{<div class="sourceCode">}}\preformatted{components = y ~ field(coordinates, model = inla.spde2.matern(...))
}\if{html}{\out{</div>}}

Since \code{coordinates} is a function from the \code{sp} package, this results in
evaluation of \code{sp::coordinates(.data.)}, which loses any CRS information
from the data object.

For \code{sf} data with a geometry column (by default named \code{geometry}), use

\if{html}{\out{<div class="sourceCode">}}\preformatted{components = y ~ field(geometry, model = inla.spde2.matern(...))
}\if{html}{\out{</div>}}

Since the CRS information is part of the geometry column of the \code{sf} object,
this retains CRS information, so this is more robust, and allows the model
to be built on a different CRS than the observation data.
}

\examples{
(inp <- new_bru_input(x, "LABEL"))
bru_input(inp, data.frame(x = 1:3))

}
\seealso{
\code{\link[=bru_input_text]{bru_input_text()}}

\code{\link[=summary.bru_input]{summary.bru_input()}}, \code{\link[=bru_comp]{bru_comp()}}

\link{ibm_input}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}, Finn Lindgren
\email{finn.lindgren@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deltaIC.R
\name{deltaIC}
\alias{deltaIC}
\title{Summarise DIC from \code{lgcp} estimates.}
\usage{
deltaIC(..., criterion = "DIC")
}
\arguments{
\item{\dots}{Comma-separated objects inheriting from class \code{inla} and
obtained from a run of \code{INLA::inla()}, \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}}}

\item{criterion}{character vector.
If it includes 'DIC', computes DIC differences;
'WAIC' is also allowed, but note that plain WAIC values from inla are not
well-defined for point process models. Default 'WAIC'}
}
\value{
A data frame with each row containing the Model name,
DIC and Delta.DIC.
}
\description{
Calculates DIC differences and produces an ordered summary.
}
\examples{
\donttest{
if (bru_safe_inla()) {
  # Generate some data
  input.df <- data.frame(idx = 1:10) |>
    dplyr::mutate(
      x = cos(idx),
      y = rpois(length(x), 5 + 2 * x + rnorm(length(x), mean = 0, sd = 0.1))
    )

  # Fit two models
  fit1 <- bru(
    y ~ x,
    family = "poisson",
    data = input.df,
    options = list(control.compute = list(dic = TRUE))
  )
  fit2 <- bru(
    y ~ x + rand(idx, model = "iid"),
    family = "poisson",
    data = input.df,
    options = list(control.compute = list(dic = TRUE))
  )

  # Compare DIC

  deltaIC(fit1, fit2)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R
\name{bru_formula_to_bru_obs_code}
\alias{bru_formula_to_bru_obs_code}
\title{Convert components to R code}
\usage{
bru_formula_to_bru_obs_code(components, add = "")
}
\arguments{
\item{components}{A \link{formula} describing latent model components.}
}
\description{
Convert components to R code
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_eval_in_data_context}
\alias{bru_eval_in_data_context}
\title{Evaluate expressions in data contexts}
\usage{
bru_eval_in_data_context(
  input,
  data = NULL,
  default = NULL,
  .envir = parent.frame()
)
}
\arguments{
\item{input}{An expression to be evaluated}

\item{data}{list of data objects in priority order. Named elements will be
available as whole objects \code{.name.} as well as pronouns \code{.name} in the
evaluation.}

\item{default}{Value used if the expression is evaluated as NULL. Default
NULL}

\item{.envir}{The evaluation environment}
}
\value{
The result of expression evaluation
}
\description{
Evaluate an expression in a series of data contexts, also making
the objects directly available as names surrounded by ".", stopping when
the expression evaluation completes with no error, as well as \code{tidy}
evaluation pronouns, e.g. \code{.data}.

This is an internal inlabru method, not intended for general use.
}
\examples{
# The A values come from the 'data' element, and the B values come from
# the 'response_data' element, as that is listed first.
bru_eval_in_data_context(
  list(A = .data$x, B = x),
  list(
    response_data = tibble::tibble(x = 1:5),
    data = tibble::tibble(x = 1:10)
  )
)
# Both A and B come from the 'data' element, as 'x' is found there,
# and the data objects are listed in order of precedence.
bru_eval_in_data_context(
  list(A = .data$x, B = x),
  list(
    data = tibble::tibble(x = 1:10),
    response_data = tibble::tibble(x = 1:5)
  )
)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/used.R
\name{bru_used_update}
\alias{bru_used_update}
\alias{bru_used_update.bru_obs_list}
\alias{bru_used_update.bru_obs}
\alias{bru_used_update.bru_pred_expr}
\alias{bru_used_update.bru_used}
\title{Update used_component information objects}
\usage{
bru_used_update(x, labels, ...)

\method{bru_used_update}{bru_obs_list}(x, labels, ...)

\method{bru_used_update}{bru_obs}(x, labels, ...)

\method{bru_used_update}{bru_pred_expr}(x, labels, ...)

\method{bru_used_update}{bru_used}(x, labels, ...)
}
\arguments{
\item{x}{Object to be updated}

\item{labels}{character vector of component labels}

\item{\dots}{Unused}
}
\value{
An updated version of \code{x}
}
\description{
Merge available component labels information with used components
information.
}
\seealso{
Other bru_used: 
\code{\link{bru_used}()},
\code{\link{bru_used_vars}()},
\code{\link{new_bru_used}()}
}
\concept{bru_used}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapper_repeat.R
\name{bm_repeat}
\alias{bm_repeat}
\alias{bru_mapper_repeat}
\title{Mapper for repeating a mapper}
\usage{
bm_repeat(mapper, n_rep, interleaved = FALSE)

bru_mapper_repeat(...)
}
\arguments{
\item{mapper}{The mapper to be repeated.}

\item{n_rep}{The number of times to repeat the mapper. If a vector, the
non-interleaved repeats are combined into a single repeat mapping,
and combined with interleaved repeats via a \code{\link[=bm_sum]{bm_sum()}} of mappers.}

\item{interleaved}{logical; if \code{TRUE}, the repeated mapping columns are
interleaved; \verb{(x1[1], x2[1], ..., x1[2], x2[2], ...)}.
If \code{FALSE} (default), the repeated
mapping columns are contiguous, \verb{(x1[1], x1[2], ..., x2[1], x2[2], ...)},
and the Jacobian is a \code{cbind()} of the Jacobians of the repeated mappers.

If \code{n_rep} is a vector, \code{interleaved} should either be a single logical,
or a vector of the same length. Each element applies to the corresponding
\code{n_rep} repetition specification.}

\item{...}{Arguments passed on to \code{\link[=bm_scale]{bm_scale()}}}
}
\value{
A \code{bm_repeat} or \code{bm_sum} object, or the original
input \code{mapper}.
}
\description{
Defines a repeated-space mapper that sums the contributions for
each copy. The \code{ibm_n()} method returns \code{ibm_n(mapper) * n_rep}, and
\code{ibm_values()} returns \code{seq_len(ibm_n(mapper))}.
}
\examples{
(m0 <- bm_index(3))
(m <- bm_repeat(m0, 5))
ibm_n(m)
ibm_values(m)
ibm_jacobian(m, 1:3)
ibm_eval(m, 1:3, seq_len(ibm_n(m)))

# Interleaving and grouping
(m <- bm_repeat(m0, c(2, 1, 2), c(TRUE, FALSE, FALSE)))
ibm_n(m)
ibm_values(m)
ibm_jacobian(m, 1:3)
ibm_eval(m, 1:3, seq_len(ibm_n(m)))

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{bru_comp_eval}
\alias{bru_comp_eval}
\alias{bru_component_eval}
\title{Evaluate component values in predictor expressions}
\usage{
bru_comp_eval(
  main,
  group = NULL,
  replicate = NULL,
  weights = NULL,
  .state = NULL
)
}
\arguments{
\item{main, group, replicate, weights}{Specification of where to evaluate a
component. The four inputs are passed on to the joint \code{bru_mapper} for the
component, as

\if{html}{\out{<div class="sourceCode">}}\preformatted{list(mapper = list(
       main = main,
       group = group,
       replicate = replicate),
     scale = weights)
}\if{html}{\out{</div>}}

NOTE: If you have model component with the same name as a data variable you
want to supply as input to \code{name_eval()}, you need to use
\code{.data.[["myvar"]]} to access it. Otherwise, it will try to use the other
component effect as input, which is ill-defined.}

\item{.state}{The internal component state. Normally supplied automatically
by the internal methods for evaluating inlabru predictor expressions.}
}
\value{
A vector of values for a component
}
\description{
In predictor expressions, \code{name_eval(...)} can be used to evaluate
the effect of a component called "name".
}
\examples{
if (bru_safe_inla() &&
  require("sf", quietly = TRUE) &&
  requireNamespace("sn", quietly = TRUE)) {
  mesh <- fmesher::fm_mesh_2d_inla(
    cbind(0, 0),
    offset = 2,
    max.edge = 2.5
  )
  spde <- INLA::inla.spde2.pcmatern(
    mesh,
    prior.range = c(1, NA),
    prior.sigma = c(0.2, NA)
  )
  set.seed(12345L)
  data <- sf::st_as_sf(
    data.frame(
      x = runif(50),
      y = runif(50),
      z = rnorm(50)
    ),
    coords = c("x", "y")
  )
  fit <- bru(
    z ~ -1 + field(geometry, model = spde),
    family = "gaussian", data = data,
    options = list(control.inla = list(int.strategy = "eb"))
  )
  pred <- generate(
    fit,
    newdata = data.frame(A = 0.5, B = 0.5),
    formula = ~ field_eval(cbind(A, B)),
    n.samples = 1L
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{as_bru_mapper}
\alias{as_bru_mapper}
\alias{as_bru_mapper.bru_mapper}
\alias{as_bru_mapper.bru_comp}
\alias{as_bru_mapper.bru_subcomp}
\title{Methods for mapper extraction}
\usage{
as_bru_mapper(x)

\method{as_bru_mapper}{bru_mapper}(x)

\method{as_bru_mapper}{bru_comp}(x)

\method{as_bru_mapper}{bru_subcomp}(x)
}
\arguments{
\item{x}{Object to convert/extract}
}
\value{
A \code{bru_mapper} object
}
\description{
Extract a mapper from another object
}
\examples{
# Extract a mapper from a `bru_subcomp` object
as_bru_mapper(bru_comp("x", x, mapper = bm_index(4))$main)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformation.R
\name{bru_transformation}
\alias{bru_transformation}
\alias{bru_forward_transformation}
\alias{bru_inverse_transformation}
\title{Transformation tools}
\usage{
bru_forward_transformation(qfun, x, ..., tail.split. = 0)

bru_inverse_transformation(pfun, x, ..., tail.split. = NULL)
}
\arguments{
\item{qfun}{A quantile function object, such as \code{qexp}}

\item{x}{Values to be transformed}

\item{\dots}{Distribution parameters passed on to the \code{qfun} and \code{pfun}
functions}

\item{tail.split.}{For x-values larger than \code{tail.split.}, upper quantile
calculations are used internally, and for smaller values lower quantile
calculations are used. This can avoid lack of accuracy in the distribution
tails. If \code{NULL}, forward calculations split at 0, and inverse calculations
use lower tails only, potentially losing accuracy in the upper tails.}

\item{pfun}{A CDF function object, such as \code{pexp}}
}
\value{
\itemize{
\item For \code{bru_forward_transformation}, a numeric vector
}

\itemize{
\item For \code{bru_inverse_transformation}, a numeric vector
}
}
\description{
Tools for transforming between N(0,1) variables and other
distributions in predictor expressions
}
\examples{
u <- rnorm(5, 0, 1)
y <- bru_forward_transformation(qexp, u, rate = 2)
v <- bru_inverse_transformation(pexp, y, rate = 2)
rbind(u, y, v)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{ibm_eval2}
\alias{ibm_eval2}
\alias{ibm_eval2.default}
\alias{ibm_eval2.bm_pipe}
\title{Evaluate a mapper and its Jacobian}
\usage{
ibm_eval2(mapper, input, state = NULL, ...)

\method{ibm_eval2}{default}(mapper, input, state, ...)

\method{ibm_eval2}{bm_pipe}(mapper, input, state = NULL, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{...}{Arguments passed on to other methods}
}
\description{
Implementations must return a list with elements \code{offset} and \code{jacobian}.
The \code{input} contents must
be in a format accepted by \code{\link[=ibm_jacobian]{ibm_jacobian()}}
for the mapper.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_eval2(default)}: Calls \code{jacobian <- ibm_jacobian(...)} and
\code{offset <- ibm_eval(..., jacobian = jacobian)}
and returns a list with elements \code{offset} and \code{jacobian}, as needed by
\code{\link[=ibm_linear.default]{ibm_linear.default()}} and similar methods. Mapper classes can implement
their own \code{ibm_eval2} method if joint construction of evaluation and Jacobian
is more efficient than separate or sequential construction.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R
\name{ibm_n}
\alias{ibm_n}
\alias{ibm_n.default}
\alias{ibm_n.bm_fmesher}
\alias{ibm_n.bm_fm_mesh_1d}
\alias{ibm_n.bm_taylor}
\alias{ibm_n.bm_linear}
\alias{ibm_n.bm_matrix}
\alias{ibm_n.bm_factor}
\alias{ibm_n.bm_const}
\alias{ibm_n.bm_shift}
\alias{ibm_n.bm_scale}
\alias{ibm_n.bm_aggregate}
\alias{ibm_n.bm_marginal}
\alias{ibm_n.bm_pipe}
\alias{ibm_n.bm_multi}
\alias{ibm_n.bm_harmonics}
\alias{ibm_n.bm_reparam}
\alias{ibm_n.bm_collect}
\alias{ibm_n.bm_repeat}
\alias{ibm_n.bm_sum}
\title{Size of the latent vector of a mapping}
\usage{
ibm_n(mapper, inla_f = FALSE, ...)

\method{ibm_n}{default}(mapper, inla_f = FALSE, ...)

\method{ibm_n}{bm_fmesher}(mapper, ...)

\method{ibm_n}{bm_fm_mesh_1d}(mapper, ...)

\method{ibm_n}{bm_taylor}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_n}{bm_linear}(mapper, ...)

\method{ibm_n}{bm_matrix}(mapper, ...)

\method{ibm_n}{bm_factor}(mapper, ...)

\method{ibm_n}{bm_const}(mapper, ...)

\method{ibm_n}{bm_shift}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_n}{bm_scale}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_n}{bm_aggregate}(mapper, ..., input = NULL, state = NULL, n_state = NULL)

\method{ibm_n}{bm_marginal}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_n}{bm_pipe}(mapper, ..., input = NULL, state = NULL)

\method{ibm_n}{bm_multi}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_n}{bm_harmonics}(mapper, inla_f = FALSE, ...)

\method{ibm_n}{bm_reparam}(mapper, ...)

\method{ibm_n}{bm_collect}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_n}{bm_repeat}(mapper, ...)

\method{ibm_n}{bm_sum}(mapper, inla_f = FALSE, multi = FALSE, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{\dots}{Arguments passed on to other methods}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{n_state}{integer giving the length of the state vector for mappers
that have state dependent output size.}

\item{input}{Data input for the mapper.}
}
\description{
Implementations must return the size of the latent vector
being mapped to.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_n(default)}: Returns a non-null element 'n' from the
mapper object, and gives an error if it doesn't exist. If \code{inla_f=TRUE},
first checks for a 'n_inla' element.

\item \code{ibm_n(bm_fmesher)}: Returns the \code{\link[fmesher:fm_dof]{fmesher::fm_dof()}} value of the mesh being
mapped.

\item \code{ibm_n(bm_fm_mesh_1d)}: Returns the \code{\link[fmesher:fm_dof]{fmesher::fm_dof()}} value of the mesh being
mapped.

\item \code{ibm_n(bm_linear)}: Returns \code{1L}

\item \code{ibm_n(bm_matrix)}: Returns the number of columns in the matrix mapper.

\item \code{ibm_n(bm_factor)}: Returns the number of levels when \code{factor_mapping} is
"full", and the number of levels minus one if \code{factor_mapping} is
"contrast".

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_multi}
\alias{bm_multi}
\alias{bru_mapper_multi}
\alias{[.bm_multi}
\alias{[.bru_mapper_multi}
\title{Mapper for tensor product domains}
\usage{
bm_multi(mappers, simplify = FALSE)

bru_mapper_multi(...)

\method{[}{bm_multi}(x, i, drop = TRUE)

\method{[}{bru_mapper_multi}(x, i, drop = TRUE)
}
\arguments{
\item{mappers}{A list of \code{bru_mapper} objects}

\item{simplify}{logical; If \code{TRUE}, removes trivial submappers.
Currently only sub-mappers of class \code{\link[=bm_index]{bm_index()}} with \code{ibm_n() == 1L} are
removed, and only if the mappers are named (to avoid ordering mismatches).
Default: \code{FALSE}}

\item{\dots}{Arguments passed on to \code{\link[=bm_multi]{bm_multi()}}}

\item{x}{object from which to extract element(s)}

\item{i}{indices specifying element(s) to extract}

\item{drop}{logical;
For \verb{[.bm_multi}, whether to extract an individual mapper when
\code{i} identifies a single element. If \code{FALSE}, a list of sub-mappers is
returned (suitable e.g. for creating a new \code{bm_multi} object).
Default: \code{TRUE}}
}
\value{
\itemize{
\item \code{[}-indexing a \code{bm_multi} extracts a subset
\code{bm_multi} object (for drop \code{FALSE}) or an individual sub-mapper
(for drop \code{TRUE}, and \code{i} identifies a single element)
}
}
\description{
Constructs a row-wise Kronecker product mapping of linear/affine
mappers. Any offset in sub-mappers is added into a combined offset. Only
linear/affine sub-mappers are allowed.
}
\examples{
(m <- bm_multi(list(a = bm_index(2), b = bm_index(3))))
ibm_eval2(m, list(a = c(1, 2, 1), b = c(1, 3, 2)), 1:6)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R
\name{ibm_jacobian}
\alias{ibm_jacobian}
\alias{ibm_jacobian.default}
\alias{ibm_jacobian.bm_fmesher}
\alias{ibm_jacobian.bm_fm_mesh_1d}
\alias{ibm_jacobian.bm_index}
\alias{ibm_jacobian.bm_taylor}
\alias{ibm_jacobian.bm_linear}
\alias{ibm_jacobian.bm_matrix}
\alias{ibm_jacobian.bm_factor}
\alias{ibm_jacobian.bm_const}
\alias{ibm_jacobian.bm_shift}
\alias{ibm_jacobian.bm_scale}
\alias{ibm_jacobian.bm_aggregate}
\alias{ibm_jacobian.bm_logsumexp}
\alias{ibm_jacobian.bm_logitaverage}
\alias{ibm_jacobian.bm_marginal}
\alias{ibm_jacobian.bm_pipe}
\alias{ibm_jacobian.bm_multi}
\alias{ibm_jacobian.bm_harmonics}
\alias{ibm_jacobian.bm_reparam}
\alias{ibm_jacobian.bm_collect}
\alias{ibm_jacobian.bm_repeat}
\alias{ibm_jacobian.bm_sum}
\title{Jacobian of a mapper}
\usage{
ibm_jacobian(mapper, input, state = NULL, inla_f = FALSE, ...)

\method{ibm_jacobian}{default}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_fmesher}(mapper, input, ...)

\method{ibm_jacobian}{bm_fm_mesh_1d}(mapper, input, ...)

\method{ibm_jacobian}{bm_index}(mapper, input, state, ...)

\method{ibm_jacobian}{bm_taylor}(mapper, ..., multi = FALSE)

\method{ibm_jacobian}{bm_linear}(mapper, input, ...)

\method{ibm_jacobian}{bm_matrix}(mapper, input, state = NULL, inla_f = FALSE, ...)

\method{ibm_jacobian}{bm_factor}(mapper, input, ...)

\method{ibm_jacobian}{bm_const}(mapper, input, ...)

\method{ibm_jacobian}{bm_shift}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_scale}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_aggregate}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_logsumexp}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_logitaverage}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_marginal}(mapper, input, state = NULL, ..., reverse = FALSE)

\method{ibm_jacobian}{bm_pipe}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_multi}(
  mapper,
  input,
  state = NULL,
  inla_f = FALSE,
  multi = FALSE,
  ...,
  sub_A = NULL
)

\method{ibm_jacobian}{bm_harmonics}(mapper, input, state = NULL, inla_f = FALSE, ...)

\method{ibm_jacobian}{bm_reparam}(mapper, input, state = NULL, ...)

\method{ibm_jacobian}{bm_collect}(
  mapper,
  input,
  state = NULL,
  inla_f = FALSE,
  multi = FALSE,
  ...,
  sub_lin = NULL
)

\method{ibm_jacobian}{bm_repeat}(
  mapper,
  input,
  state = NULL,
  inla_f = FALSE,
  multi = FALSE,
  ...,
  sub_lin = NULL
)

\method{ibm_jacobian}{bm_sum}(
  mapper,
  input,
  state = NULL,
  inla_f = FALSE,
  multi = FALSE,
  ...,
  sub_lin = NULL
)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{...}{Arguments passed on to other methods}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}

\item{reverse}{logical; control \code{bm_marginal} evaluation. Default
\code{FALSE}. When \code{TRUE}, reverses the direction of the mapping, see details
for \code{marginal} mappers.}

\item{sub_A}{Internal; precomputed Jacobian matrices.}

\item{sub_lin}{Internal, optional pre-computed sub-mapper information}
}
\description{
Implementations must return a (sparse) matrix of size
\code{ibm_n_output(mapper, input, inla_f)}
by \code{ibm_n(mapper, inla_f = FALSE)}. The \code{inla_f=TRUE} argument should
only affect the allowed type of input format.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_jacobian(default)}: Mapper classes must implement their own \code{ibm_jacobian} method.

\item \code{ibm_jacobian(bm_fmesher)}: Returns the \code{\link[fmesher:fm_basis]{fmesher::fm_basis()}} matrix of the
mesh being mapped.

\item \code{ibm_jacobian(bm_fm_mesh_1d)}: Returns the \code{\link[fmesher:fm_basis]{fmesher::fm_basis()}} matrix of the
mesh being mapped.

\item \code{ibm_jacobian(bm_matrix)}: Accepts \code{input} as a matrix, \code{Matrix}, \code{Spatial},
or \code{sfc_POINT} object.

\item \code{ibm_jacobian(bm_shift)}: \code{input} NULL values are interpreted as no shift.

\item \code{ibm_jacobian(bm_scale)}: \code{input} NULL values
are interpreted as no scaling.

\item \code{ibm_jacobian(bm_aggregate)}: \code{input} should be a list with elements \code{block}
and \code{weights}. \code{block}
should be a vector of the same length as the \code{state}, or \code{NULL}, with \code{NULL}
equivalent to all-1.
If \code{weights} is \code{NULL}, it's interpreted as all-1.

\item \code{ibm_jacobian(bm_logsumexp)}: \code{input} should be a list with elements
\code{block} and \code{weights}. \code{block} should be a vector of the same length as the
\code{state}, or \code{NULL}, with \code{NULL} equivalent to all-1.
If \code{weights} is \code{NULL}, it's interpreted as all-1.

\item \code{ibm_jacobian(bm_logitaverage)}: \code{input} should be a list with elements
\code{block} and \code{weights}. \code{block} should be a vector of the same length as the
\code{state}, or \code{NULL}, with \code{NULL} equivalent to all-1.
If \code{weights} is \code{NULL}, it's interpreted as all-1.

\item \code{ibm_jacobian(bm_marginal)}: Non-NULL \code{input} values are interpreted
as a parameter list for \code{qfun}, overriding that of the mapper itself.

\item \code{ibm_jacobian(bm_multi)}: Accepts a list with
named entries, or a list with unnamed but ordered elements.
The names must match the sub-mappers, see \code{\link[=ibm_names.bm_multi]{ibm_names.bm_multi()}}.
Each list element should take a format accepted by the corresponding
sub-mapper. In case each element is a vector, the input can be given as a
data.frame with named columns, a matrix with named columns, or a matrix
with unnamed but ordered columns.

\item \code{ibm_jacobian(bm_collect)}: Accepts a list with
named entries, or a list with unnamed but ordered elements.
The names must match the sub-mappers, see \code{\link[=ibm_names.bm_collect]{ibm_names.bm_collect()}}.
Each list element should take a format accepted by the corresponding
sub-mapper. In case each element is a vector, the input can be given as a
data.frame with named columns, a matrix with named columns, or a matrix
with unnamed but ordered columns. When \code{inla_f=TRUE} and \code{hidden=TRUE} in
the mapper definition, the input format should instead match that of
the first, non-hidden, sub-mapper.

\item \code{ibm_jacobian(bm_repeat)}: The input should take the format of the
repeated submapper.

\item \code{ibm_jacobian(bm_sum)}: Accepts a list with
named entries, or a list with unnamed but ordered elements.
The names must match the sub-mappers, see \code{\link[=ibm_names.bm_sum]{ibm_names.bm_sum()}}.
Each list element should take a format accepted by the corresponding
sub-mapper. In case each element is a vector, the input can be given as a
data.frame with named columns, a matrix with named columns, or a matrix
with unnamed but ordered columns.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_list}
\alias{bm_list}
\alias{as_bm_list}
\alias{as_bm_list.list}
\alias{as_bm_list.bm_list}
\alias{as_bm_list.bru_comp_list}
\alias{c.bru_mapper}
\alias{c.bm_list}
\alias{[.bm_list}
\title{Methods for mapper lists}
\usage{
as_bm_list(x)

\method{as_bm_list}{list}(x)

\method{as_bm_list}{bm_list}(x)

\method{as_bm_list}{bru_comp_list}(x)

\method{c}{bru_mapper}(...)

\method{c}{bm_list}(...)

\method{[}{bm_list}(x, i)
}
\arguments{
\item{x}{\code{bm_list} object from which to extract element(s)}

\item{\dots}{Objects to be combined.}

\item{i}{indices specifying elements to extract}
}
\value{
A \code{bm_list} object
}
\description{
\code{bru_mapper} lists can be combined into \code{bm_list} lists.
}
\section{Methods (by generic)}{
\itemize{
\item \code{c(bm_list)}: The \code{...} arguments should be \code{bm_list}
objects.

\item \code{[}: Extract sub-list

}}
\section{Functions}{
\itemize{
\item \code{c(bru_mapper)}: The \code{...} arguments should be \code{bru_mapper}
objects.

}}
\examples{
m <- c(A = bm_const(), B = bm_scale())
str(m)
str(m[2])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{parse_inclusion}
\alias{parse_inclusion}
\title{Parse inclusion of component labels in a predictor expression}
\usage{
parse_inclusion(thenames, include = NULL, exclude = NULL)
}
\arguments{
\item{thenames}{Set of labels to restrict}

\item{include}{Character vector of component labels that are needed by the
predictor expression; Default: NULL (include all components that are not
explicitly excluded)}

\item{exclude}{Character vector of component labels that are not used by the
predictor expression. The exclusion list is applied to the list
as determined by the \code{include} parameter; Default: NULL (do not remove
any components from the inclusion list)}
}
\description{
Parse inclusion of component labels in a predictor expression
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{bru_fill_missing}
\alias{bru_fill_missing}
\title{Fill in missing values in Spatial grids}
\usage{
bru_fill_missing(
  data,
  where,
  values,
  layer = NULL,
  selector = NULL,
  batch_size = deprecated()
)
}
\arguments{
\item{data}{A SpatialPointsDataFrame, SpatialPixelsDataFrame,
SpatialGridDataFrame, SpatRaster, Raster, or sf object
containing data to use for filling}

\item{where}{A, matrix, data.frame, or SpatialPoints or
SpatialPointsDataFrame, or sf object, containing the locations of the
evaluated values}

\item{values}{A vector of values to be filled in where \code{is.na(values)} is
\code{TRUE}}

\item{layer, selector}{Specifies what data column or columns from which to
extract data, see \code{\link[=bru_comp]{bru_comp()}} for details.}

\item{batch_size}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} due to improved
algorithm.
Size of nearest-neighbour calculation blocks, to limit the
memory and computational complexity.}
}
\value{
An infilled vector of values
}
\description{
Computes nearest-available-value imputation for missing values in space
}
\examples{
\dontrun{
if (require("sf", quietly = TRUE)) {
  points <-
    sf::st_as_sf(
      data.frame(
        x = 1:3,
        y = 4:6,
        val = c(NA, NA, NA)
      ),
      coords = c("x", "y")
    )
  input_coord <- expand.grid(x = 0:7, y = 0:7)
  input <-
    sf::st_as_sf(
      cbind(input_coord, val = as.vector(input_coord$y)),
      coords = c("x", "y")
    )
  points$val <- bru_fill_missing(input, points, points$val)
  print(points)

  # To fill in missing values in a grid:
  print(input$val[c(3, 30)])
  input$val[c(3, 30)] <- NA # Introduce missing values
  input$val <- bru_fill_missing(input, input, input$val)
  print(input$val[c(3, 30)])
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/used.R
\name{bru_used}
\alias{bru_used}
\alias{bru_used.NULL}
\alias{bru_used.default}
\alias{bru_used.bru}
\alias{bru_used.bru_info}
\alias{bru_used.list}
\alias{bru_used.bru_obs}
\alias{bru_used.bru_pred_expr}
\alias{bru_used.bru_used}
\alias{format.bru_used}
\alias{summary.bru_used}
\alias{print.bru_used}
\title{List components used in a model}
\usage{
bru_used(x = NULL, ...)

\method{bru_used}{`NULL`}(x = NULL, ...)

\method{bru_used}{default}(x, ...)

\method{bru_used}{bru}(x, ..., join = TRUE)

\method{bru_used}{bru_info}(x, ..., join = TRUE)

\method{bru_used}{list}(x, ..., join = TRUE)

\method{bru_used}{bru_obs}(x, ...)

\method{bru_used}{bru_pred_expr}(x, ...)

\method{bru_used}{bru_used}(x, labels = NULL, ...)

\method{format}{bru_used}(x, ...)

\method{summary}{bru_used}(object, ...)

\method{print}{bru_used}(x, ...)
}
\arguments{
\item{x}{An object that contains information about used components}

\item{\dots}{Parameters passed on to the other methods}

\item{join}{Whether to join list output into a single object; Default
may depend on the input object class}

\item{labels}{character; component labels passed on to
\code{\link[=bru_used_update]{bru_used_update()}}}
}
\value{
A \code{bru_used} object (a list with elements \code{effect}
and \code{latent}), or a list of such objects
(for methods with \code{join = FALSE})
}
\description{
Create or extract information about which components are used by a model, or
its individual observation models. If a non-NULL \code{labels} argument is
supplied, also calls \code{\link[=bru_used_update]{bru_used_update()}} on the \code{bru_used} objects.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_used(`NULL`)}: Create a \code{bru_used} object by calling \code{\link[=new_bru_used]{new_bru_used()}}.

\item \code{bru_used(default)}: Create a \code{bru_used} object by calling \code{\link[=new_bru_used]{new_bru_used()}}

\item \code{bru_used(bru)}: Extract the \code{bru_used} information for the collection
of observation models used in a \code{bru} object.

\item \code{bru_used(bru_info)}: Extract the \code{bru_used} information for the collection
of observation models used in a \code{bru_info} object.

\item \code{bru_used(list)}: Extract the \code{bru_used} information for each element
of a list, and optionally join into a single \code{bru_used} object.

\item \code{bru_used(bru_obs)}: Extract the \code{bru_used} information for the observation
model predictor used in a \code{bru} observation model \code{bru_obs} object.

\item \code{bru_used(bru_pred_expr)}: Extract the \code{bru_used} information for an observation
model predictor \code{bru_pred_expr} object.

\item \code{bru_used(bru_used)}: Convenience method that takes
an existing \code{bru_used} object and calls \code{\link[=bru_used_update]{bru_used_update()}}
if \code{labels} is non-NULL.

}}
\section{Methods (by generic)}{
\itemize{
\item \code{format(bru_used)}: Text formatting method for \code{bru_used} objects.

\item \code{summary(bru_used)}: Summary method for \code{bru_used} objects.

\item \code{print(bru_used)}: Print method for \code{bru_used} objects.

}}
\examples{
(used <- new_bru_used(~.))
bru_used(used, labels = c("a", "c"))
(used <- new_bru_used(~ a + b + c_latent + d_latent))
bru_used(used, labels = c("a", "c"))
(used <- new_bru_used(expression(a + b + c_latent + d_latent)))
bru_used(used, labels = c("a", "c"))

}
\seealso{
Other bru_used: 
\code{\link{bru_used_update}()},
\code{\link{bru_used_vars}()},
\code{\link{new_bru_used}()}
}
\concept{bru_used}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inla.R
\name{bru_standardise_names}
\alias{bru_standardise_names}
\title{Standardise inla hyperparameter names}
\usage{
bru_standardise_names(x)
}
\arguments{
\item{x}{character vector; names to be standardised}
}
\value{
A character vector with standardised names
}
\description{
The inla hyperparameter output uses parameter names that can include
whitespace and special characters. This function replaces those characters
with underscores.
}
\examples{
bru_standardise_names("Precision for the Gaussian observations")
}
\seealso{
\code{\link[=bru_names]{bru_names()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_log_offset}
\alias{bru_log_offset}
\alias{bru_log_index}
\title{Position methods for \code{bru_log} objects}
\usage{
bru_log_offset(x = NULL, bookmark = NULL, offset = NULL)

bru_log_index(x = NULL, i, verbosity = NULL)
}
\arguments{
\item{x}{A \code{bru_log} object. If \code{NULL}, the global \code{inlabru} log is used.}

\item{bookmark}{character; The label for a bookmark with a stored offset.}

\item{offset}{integer; a position offset in the log, with \code{0L} pointing at
the start of the log. If negative, denotes the point \code{abs(offset)} elements
from tail of the log. When \code{bookmark} is non-NULL, the \code{offset} applies a
shift (forwards or backwards) to the bookmark list.}

\item{i}{indices specifying elements to extract. If \code{character}, denotes
the sequence between bookmark \code{i} and the next bookmark (or the end of the
log if \code{i} is the last bookmark)}

\item{verbosity}{integer value for limiting the highest verbosity level being
returned.}
}
\description{
Position methods for \code{bru_log} objects.
}
\section{Functions}{
\itemize{
\item \code{bru_log_offset()}: Utility function for computing log position offsets.

\item \code{bru_log_index()}: Utility function for computing index vectors
for \code{bru_log} objects.

}}
\seealso{
Other inlabru log methods: 
\code{\link{bru_log}()},
\code{\link{bru_log_bookmark}()},
\code{\link{bru_log_message}()},
\code{\link{bru_log_new}()},
\code{\link{bru_log_reset}()}
}
\concept{inlabru log methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pred_expr.R
\name{bru_pred_expr}
\alias{bru_pred_expr}
\alias{new_bru_pred_expr}
\alias{bru_pred_expr.bru_obs}
\alias{bru_pred_expr.bru_obs_list}
\alias{bru_pred_expr.bru_info}
\alias{bru_pred_expr.bru}
\alias{bru_pred_expr.bru_pred_expr}
\alias{format.bru_pred_expr}
\alias{print.bru_pred_expr}
\title{Predictor expression handler object}
\usage{
new_bru_pred_expr(
  x,
  ...,
  used = NULL,
  is_rowwise = NULL,
  .envir = parent.frame()
)

bru_pred_expr(x, ...)

\method{bru_pred_expr}{bru_obs}(x, ...)

\method{bru_pred_expr}{bru_obs_list}(x, ...)

\method{bru_pred_expr}{bru_info}(x, ...)

\method{bru_pred_expr}{bru}(x, ...)

\method{bru_pred_expr}{bru_pred_expr}(x, ..., format = "object")

\method{format}{bru_pred_expr}(x, ...)

\method{print}{bru_pred_expr}(x, ...)
}
\arguments{
\item{x}{For creation, a formula or character string giving the predictor
expression. If the character string is \code{"."}, it is taken to mean an
additive expression including all latent variables. For access, an object
containing a \code{bru_pred_expr} object.}

\item{\dots}{Passed on to submethods}

\item{used}{An optional \code{bru_used} object, overriding the automated
detection. Default: \code{NULL}}

\item{is_rowwise}{logical; whether the predictor can be assumed to use the
input data rowwise, so that blockwise evaluation is possible. If \code{NULL}, it
needs to be set before use. Default: \code{NULL}}

\item{.envir}{The environment in which to evaluate the expression.}

\item{format}{Character; one of
\itemize{
\item \code{"object"} (the default),
\item \code{"text"} (plain text),
\item \code{"quo"} (an \code{rlang} quosure),
\item \code{"expr"} (an \code{rlang} expression),
\item \code{"formula"} (the original formula input if available, otherwise
constructed from the expression),
\item \code{"formula_text"} (a text version of the "formula"),
\item \code{"text_raw"} (plain text, without substituting \code{BRU_EXPRESSION}), or
\item \code{"resp_text"} (plain text of the response side of the formula).
}}
}
\description{
Create and access predictor expression information, keeping track of what
latent variables are used, whether the predictor is additive/linear/rowwise.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_pred_expr(bru_obs)}: Accessor for the \code{bru_pred_expr} object stored
inside a \code{bru_obs} object.

\item \code{bru_pred_expr(bru_obs_list)}: Accessor for the \code{bru_pred_expr} objects stored
inside a \code{bru_obs_list} object.

\item \code{bru_pred_expr(bru_info)}: Accessor for the \code{bru_pred_expr} object stored
inside a \code{bru_info} object.

\item \code{bru_pred_expr(bru)}: Accessor for the \code{bru_pred_expr} object stored
inside a \code{bru} object.

\item \code{bru_pred_expr(bru_pred_expr)}: Access a \code{bru_pred_expr} object or convert it to
text or expression object.

}}
\section{Functions}{
\itemize{
\item \code{new_bru_pred_expr()}: Create a \code{bru_pred_expr} object from a formula or
character string.

\item \code{bru_pred_expr()}: Accessor generic for \code{bru_pred_expr} objects,
including ones stored inside other objects.

}}
\examples{
(new_bru_pred_expr(~ x + z + Intercept))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{bru_safe_sp}
\alias{bru_safe_sp}
\title{Check for potential \code{sp} version compatibility issues}
\usage{
bru_safe_sp(quietly = FALSE, force = FALSE, minimum_version = "2.1")
}
\arguments{
\item{quietly}{logical; if \code{TRUE}, prints diagnostic messages. Default
\code{FALSE}}

\item{force}{logical; If \code{rgdal} is unavailable
and evolution status is less that \code{2L}, return \code{FALSE} if \code{force} is \code{FALSE}.
If \code{force} is \code{TRUE}, return \code{TRUE} if the package configuration is safe,
potentially after forcing the evolution status to \code{2L}.
Default \code{FALSE}}

\item{minimum_version}{character; the minimum required version.
Default 2.1 (should always match the requirement in the package
DESCRIPTION)}
}
\value{
Returns (invisibly) \code{FALSE} if a potential issue is detected, and
give a message if \code{quietly} is \code{FALSE}. Otherwise returns \code{TRUE}
}
\description{
Loads the sp package with \code{requireNamespace("sp", quietly = TRUE)}, and
checks and optionally sets the \code{sp} evolution status flag if \code{rgdal} is
unavailable.
}
\examples{
\dontrun{
if (bru_safe_sp()) {
  # Run sp dependent calculations
}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_fm_mesh_old}
\alias{bm_fm_mesh_old}
\alias{ibm_n.bm_fm_mesh_2d}
\alias{ibm_values.bm_fm_mesh_2d}
\alias{ibm_jacobian.bm_fm_mesh_2d}
\alias{ibm_n.bm_inla_mesh_2d}
\alias{ibm_values.bm_inla_mesh_2d}
\alias{ibm_jacobian.bm_inla_mesh_2d}
\alias{ibm_n.bm_inla_mesh_1d}
\alias{ibm_values.bm_inla_mesh_1d}
\alias{ibm_jacobian.bm_inla_mesh_1d}
\title{Deprecated methods}
\usage{
\method{ibm_n}{bm_fm_mesh_2d}(mapper, ...)

\method{ibm_values}{bm_fm_mesh_2d}(mapper, ...)

\method{ibm_jacobian}{bm_fm_mesh_2d}(mapper, input, ...)

\method{ibm_n}{bm_inla_mesh_2d}(mapper, ...)

\method{ibm_values}{bm_inla_mesh_2d}(mapper, ...)

\method{ibm_jacobian}{bm_inla_mesh_2d}(mapper, input, ...)

\method{ibm_n}{bm_inla_mesh_1d}(mapper, ...)

\method{ibm_values}{bm_inla_mesh_1d}(mapper, ...)

\method{ibm_jacobian}{bm_inla_mesh_1d}(mapper, input, ...)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Deprecated methods only used for old stored
\code{bru_mapper_fm_mesh_1d} and \code{bru_mapper_fm_mesh_2d} objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_input.R
\name{format.bru_input}
\alias{format.bru_input}
\alias{summary.bru_input}
\alias{print.bru_input}
\title{Summarise component inputs}
\usage{
\method{format}{bru_input}(x, verbose = TRUE, ..., label.override = NULL, type = NULL)

\method{summary}{bru_input}(object, verbose = TRUE, ..., label.override = NULL)

\method{print}{bru_input}(x, verbose = TRUE, ..., label.override = NULL)
}
\arguments{
\item{x}{Object to be printed}

\item{verbose}{logical; If \code{TRUE}, includes more details of the
component definitions. When \code{FALSE}, only show basic component
definition information.  Default \code{TRUE}.}

\item{\dots}{Passed on to other summary methods.}

\item{label.override}{character; If not \code{NULL}, use this label instead of
the object's label.}

\item{type}{character; if non-NULL, added to the output'; \code{label = type(input)}.}

\item{object}{Object to be summarised.}
}
\description{
Summarise component inputs
}
\seealso{
\code{\link[=bru_input]{bru_input()}}, \code{\link[=bru_comp]{bru_comp()}}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}

Finn Lindgren \email{finn.lindgren@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_inla_formula}
\alias{bru_inla_formula}
\alias{bru_inla_formula.bru}
\alias{bru_inla_formula.bru_info}
\alias{bru_inla_formula.bru_comp_list}
\title{Extract INLA formula}
\usage{
bru_inla_formula(x, ...)

\method{bru_inla_formula}{bru}(x, ...)

\method{bru_inla_formula}{bru_info}(x, ...)

\method{bru_inla_formula}{bru_comp_list}(x, ...)
}
\arguments{
\item{x}{An object containing information about an INLA formula}

\item{\dots}{Additional arguments passed on to submethods}
}
\description{
Extracts the INLA formula used in a \code{bru} model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.SpatRaster}
\alias{gg.SpatRaster}
\title{Geom wrapper for SpatRaster objects}
\usage{
\method{gg}{SpatRaster}(data, ...)
}
\arguments{
\item{data}{A SpatRaster object.}

\item{\dots}{Arguments passed on to \code{geom_spatraster}.}
}
\value{
The output from `geom_spatraster.
}
\description{
Convenience wrapper function for \code{tidyterra::geom_spatraster()}.
Requires the \code{ggplot2} and \code{tidyterra} packages.
}
\examples{
if (require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("tidyterra", quietly = TRUE)) {
  # Load Gorilla covariates

  gcov <- gorillas_sf_gcov()

  # Plot the pixel centers
  ggplot() +
    gg(gcov$elevation)
}
}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_log_reset}
\alias{bru_log_reset}
\title{Clear log contents}
\usage{
bru_log_reset(x = NULL, bookmark = NULL, offset = NULL)
}
\arguments{
\item{x}{A \code{bru_log} object, or in some cases, and object that can be
converted/extracted to a \code{bru_log} object. \code{NULL} denotes the global
\code{inlabru} log object.}

\item{bookmark}{character; The label for a bookmark with a stored offset.}

\item{offset}{integer; a position offset in the log, with \code{0L} pointing at
the start of the log. If negative, denotes the point \code{abs(offset)} elements
from tail of the log. When \code{bookmark} is non-NULL, the \code{offset} applies a
shift (forwards or backwards) to the bookmark list.}
}
\value{
Returns (invisibly) the modified \code{bru_log} object, or \code{NULL} (when
\code{x} is \code{NULL})
}
\description{
Clears the log contents up to
a given \code{offset} or \code{bookmark}. Default: clear the entire log.
When \code{x} is NULL, the global \code{inlabru} log is updated, and \code{invisible(NULL)}
is returned. Otherwise the updated object is returned (invisibly).
}
\examples{
\dontrun{
if (interactive()) {
  bru_log_reset()
}
}
}
\seealso{
Other inlabru log methods: 
\code{\link{bru_log}()},
\code{\link{bru_log_bookmark}()},
\code{\link{bru_log_message}()},
\code{\link{bru_log_new}()},
\code{\link{bru_log_offset}()}
}
\concept{inlabru log methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sampling.R
\name{sample.lgcp}
\alias{sample.lgcp}
\title{Sample from an inhomogeneous Poisson process}
\usage{
sample.lgcp(
  mesh,
  loglambda,
  strategy = NULL,
  R = NULL,
  samplers = NULL,
  ignore.CRS = FALSE
)
}
\arguments{
\item{mesh}{An \link[fmesher:fm_mesh_1d]{fmesher::fm_mesh_1d} or \link[fmesher:fm_mesh_2d]{fmesher::fm_mesh_2d} object}

\item{loglambda}{vector or matrix; A vector of log intensities at the mesh
vertices (for higher order basis functions, e.g. for \code{fm_mesh_1d} meshes,
\code{loglambda} should be given as \code{mesh$m} basis function weights rather than
the values at the \code{mesh$n} vertices) A single scalar is expanded to a
vector of the appropriate length. If a matrix is supplied, one process
sample for each column is produced.}

\item{strategy}{Only relevant for 2D meshes. One of \code{'triangulated'},
\code{'rectangle'}, \code{'sliced-spherical'}, \code{'spherical'}. The \code{'rectangle'}
method is only valid for CRS-less flat 2D meshes.
If \code{NULL} or \code{'auto'}, the the likely fastest method is chosen;
\code{'rectangle'} for flat 2D meshes with no CRS,
\code{'sliced-spherical'} for CRS \code{'longlat'} meshes, and
\code{'triangulated'} for all other meshes.}

\item{R}{Numerical value only applicable to spherical and geographical
meshes. It is interpreted as \code{R} is the equivalent Earth radius, in km,
used to scale the lambda intensity. For CRS enabled meshes, the default is
6371. For CRS-less spherical meshes, the default is 1.}

\item{samplers}{A \code{SpatialPolygonsDataFrame} or \code{fm_mesh_2d} object.
Simulated points that fall outside these polygons are discarded.}

\item{ignore.CRS}{logical; if \code{TRUE}, ignore any CRS information in the mesh.
Default \code{FALSE}. This affects \code{R} and the permitted values for \code{strategy}.}
}
\value{
A \code{data.frame} (1D case),
or \code{sf} (2D flat and 3D spherical surface cases,
or 2D/2.5D surface cases with multiple samples).
For multiple samples, the \code{data.frame} output has a
column \code{'sample'} giving the index for each sample.
object of point locations.

For the old (pre version \verb{2.13.0.9034}) \code{sp} output format, use \code{as(ret, "Spatial")} or \code{sf::as_Spatial(ret)}.
}
\description{
This function provides point samples from one- and two-dimensional
inhomogeneous Poisson processes. The log intensity has to be provided via its
values at the nodes of an \code{fm_mesh_1d} or \code{fm_mesh_2d} object. In between
mesh nodes the log intensity is assumed to be linear.
}
\details{
For 2D processes on a sphere the \code{R} parameter can be used to adjust to
sphere's radius implied by the mesh. If the intensity is very high the
standard \code{strategy} "spherical" can cause memory issues. Using the
"sliced-spherical" strategy can help in this case.
\itemize{
\item For crs-less meshes on R2: Lambda is interpreted in the raw coordinate
system. Output has an NA CRS.
\item For crs-less meshes on S2: Lambda with raw units, after scaling the mesh
to radius \code{R}, if specified.
Output is given on the same domain as the mesh, with an NA CRS.
\item For crs meshes on R2: Lambda is interpreted as per km^2, after scaling the
globe to the Earth radius 6371 km, or \code{R}, if specified. Output given in
the same CRS as the mesh.
\item For crs meshes on S2: Lambda is interpreted as per km^2, after scaling the
globe to the Earth radius 6371 km, or \code{R}, if specified. Output given in
the same CRS as the mesh.
}
}
\examples{
\donttest{
# The INLA package is required
if (bru_safe_inla()) {
  vertices <- seq(0, 3, by = 0.1)
  mesh <- fmesher::fm_mesh_1d(vertices)
  loglambda <- 5 - 0.5 * vertices
  pts <- sample.lgcp(mesh, loglambda)
  pts$y <- 0
  plot(vertices, exp(loglambda), type = "l", ylim = c(0, 150))
  points(pts, pch = "|")
}
}

\donttest{
# The INLA package is required
if (bru_safe_inla() &&
  require(ggplot2, quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE)) {
  gorillas <- gorillas_sf
  pts <- sample.lgcp(gorillas$mesh,
    loglambda = 1.5,
    samplers = gorillas$boundary
  )
  ggplot() +
    gg(gorillas$mesh) +
    gg(pts)

  pts <- sample.lgcp(gorillas$mesh,
    loglambda = base::scale(gorillas$mesh$loc) * 2,
    samplers = gorillas$boundary,
    ignore.CRS = TRUE
  )
  ggplot() +
    gg(gorillas$mesh) +
    gg(pts, aes(color = as.factor(sample))) +
    labs(color = "Sample")
}
}

}
\author{
Daniel Simpson \email{dp.simpson@gmail.com} (base rectangle and
spherical algorithms), Fabian E. Bachl \email{bachlfab@gmail.com}
(inclusion in inlabru, sliced spherical sampling), Finn Lindgren
\email{finn.lindgren@gmail.com} (extended CRS support, triangulated
sampling)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R
\name{bru_comp_env_extra}
\alias{bru_comp_env_extra}
\alias{bru_comp_env_extra<-}
\alias{bru_comp_env}
\alias{bru_comp_env<-}
\title{Get/set component environment data}
\usage{
bru_comp_env_extra(x, name = NULL)

bru_comp_env_extra(x, name) <- value

bru_comp_env(x, name = NULL)

bru_comp_env(x, name) <- value
}
\arguments{
\item{x}{A \code{bru_comp} object}

\item{name}{The name of the variable to get or set. Names
prefixed by \code{"BRU_"} are reserved for internal use and should not be set
by users or external package authors.}

\item{value}{The value to set for \code{name} in \code{env_extra} (for the setter
function).}
}
\value{
For the getter, either the entire environment or the
value of \code{name} in the environment. For the setters, the modified
\code{bru_comp} object.
}
\description{
Get or set data in the component's \code{env_extra} and \code{env}
environments. If \code{name} is \code{NULL}, the entire environment is returned.
}
\section{Functions}{
\itemize{
\item \code{bru_comp_env_extra(x, name) <- value}: Set data in the component's \code{env_extra}
environment

\item \code{bru_comp_env()}: Get the component's \code{env} environment, or an
element from that environment.
Note that in most cases this environment is the global R environment.

\item \code{bru_comp_env(x, name) <- value}: Set data in the component's \code{env} environment.
Note that in most cases this environment is the global R environment, so
modifying it is normally not recommended.

}}
\examples{
if (bru_safe_inla()) {
  cmp <- bru_comp("x", x)
  as.list(bru_comp_env_extra(cmp))

  cmp <- bru_comp("x", x, model = "fixed")
  as.list(bru_comp_env_extra(cmp))

  bru_comp_env(cmp)
}

}
\seealso{
bru_comp
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_set_missing}
\alias{bru_set_missing}
\alias{bru_set_missing<-}
\alias{bru_set_missing.bru}
\alias{bru_set_missing.bru_info}
\alias{bru_set_missing.bru_obs_list}
\alias{bru_set_missing.bru_obs}
\alias{bru_set_missing.default}
\alias{bru_set_missing.data.frame}
\alias{bru_set_missing.inla.surv}
\title{Set missing values in observation models}
\usage{
bru_set_missing(object, keep = FALSE, ...)

bru_set_missing(x, ...) <- value

\method{bru_set_missing}{bru}(object, keep = FALSE, ...)

\method{bru_set_missing}{bru_info}(object, keep = FALSE, ...)

\method{bru_set_missing}{bru_obs_list}(object, keep = FALSE, ...)

\method{bru_set_missing}{bru_obs}(object, keep = FALSE, ...)

\method{bru_set_missing}{default}(object, keep = FALSE, ...)

\method{bru_set_missing}{data.frame}(object, keep = FALSE, ...)

\method{bru_set_missing}{inla.surv}(object, keep = FALSE, ...)
}
\arguments{
\item{object}{A \code{bru}, \code{bru_obs} or \code{bru_obs_list} object}

\item{keep}{For \code{bru_obs}, a single logical or an integer vector;
If \code{TRUE}, keep all the response data, if \code{FALSE} (default),
set all of it to \code{NA}. An integer vector determines which elements
to keep (for positive values) or to set as missing (negative values).

For \code{bru} and \code{bru_obs_list}, a logical scalar or vector, or a list, see
Details.}

\item{\dots}{Additional arguments passed on to the \code{bru_obs} and individual
data class methods.  Currently unused.}

\item{x}{Object on which to apply \code{bru_set_missing()}}

\item{value}{Value to be passed as \code{keep} to \code{bru_set_missing()}}
}
\description{
Set all or parts of the observation model response data
to \code{NA}, for example for use in cross validation (with \code{\link[=bru_rerun]{bru_rerun()}})
or prior sampling (with \code{\link[=bru_rerun]{bru_rerun()}} and \code{\link[=generate]{generate()}}, but see
"Prior sampling caveats" below).
}
\details{
For \code{bru} and \code{bru_obs_list},
\itemize{
\item{\code{keep} must be either a single logical, which is expanded to a list,}
\item{a logical vector, which is converted to a list,}
\item{an unnamed list of the same length as the number of observation
models, with elements compatible with the \code{bru_obs} method, or}
\item{a named list with elements compatible with the \code{bru_obs} method,
and only the named \code{bru_obs} models are acted upon, i.e. the elements
not present in the list are treated as \code{keep = TRUE}.}
}

E.g.: \code{keep = list(b = FALSE)} sets all observations in model \code{b} to missing,
and does not change model \code{a}.

E.g.: \code{keep = list(a = 1:4, b = -(3:5))} keeps only observations \code{1:4} of
model \code{a}, marking the rest as missing, and sets observations \code{3:5} of model
\code{b} to missing.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_set_missing(default)}: From \verb{> 2.13.0}, set missing values in any
object supporting \code{\link[base:NA]{base::is.na<-()}} for positive and negative indices.

\item \code{bru_set_missing(data.frame)}: From \verb{> 2.13.0}, handles \code{data.frame}, tibbles,
including \code{inla.mdata}.

\item \code{bru_set_missing(inla.surv)}: From \verb{> 2.13.0}, handles \code{inla.surv}.

}}
\section{Functions}{
\itemize{
\item \code{bru_set_missing(x, ...) <- value}: Setter method for \code{bru_set_missing()}

}}
\section{Prior sampling caveats}{

Note that prior sampling requires special care for hyperparameters, as the
prior modes are not typically useful; in the future, we plan to have a
dedicated method that samples from the hyperparameters, and then uses

\if{html}{\out{<div class="sourceCode">}}\preformatted{bru_rerun(
  bru_set_missing(...),
  options = list(
    control.mode = list(
      theta = theta_sample,
      fixed = TRUE)
    )
  )
}\if{html}{\out{</div>}}

for each sample.
}

\examples{
obs <- c(
  A = bru_obs(y_A ~ ., data = data.frame(y_A = 1:6)),
  B = bru_obs(y_B ~ ., data = data.frame(y_B = 11:15))
)
bru_response_size(obs)
lapply(
  bru_set_missing(obs, keep = FALSE),
  function(x) {
    x[["response_data"]][[x[["response"]]]]
  }
)
lapply(
  bru_set_missing(obs, keep = list(B = FALSE)),
  function(x) {
    x[["response_data"]][[x[["response"]]]]
  }
)
lapply(
  bru_set_missing(obs, keep = list(1:4, -(3:5))),
  function(x) {
    x[["response_data"]][[x[["response"]]]]
  }
)

\dontshow{if (bru_safe_inla()) withAutoprint(\{ # examplesIf}
(obs <- INLA::inla.mdata(y = 1:4, X = matrix(1:8, 4, 2)))
bru_set_missing(obs, keep = c(1, 4))
bru_set_missing(obs) <- -(1:2)
obs
\dontshow{\}) # examplesIf}
\dontshow{if (bru_safe_inla()) withAutoprint(\{ # examplesIf}
(obs <- INLA::inla.surv(time = 1:4, event = c(1, 0, 1, 0)))
bru_set_missing(obs, keep = c(1, 4))

(obs <- INLA::inla.surv(
  time = 1:4,
  event = c(1, 0, 1, 0),
  cure = matrix(1:8, 4, 2)
))
bru_set_missing(obs, keep = c(1, 4))

(obs <- INLA::inla.surv(
  time = 1:4,
  event = c(1, 0, 1, 0),
  subject = c(1, 1, 2, 1)
))
bru_set_missing(obs, keep = c(1, 4))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_conversion.R
\name{as_bru_obs}
\alias{as_bru_obs}
\alias{as_bru_obs_list}
\alias{as_bru_obs.bru_obs}
\alias{as_bru_obs_list.bru_obs}
\alias{as_bru_obs_list.list}
\alias{as_bru_obs_list.bru_obs_list}
\alias{as_bru_obs_list.bru}
\alias{as_bru_obs_list.bru_info}
\alias{as_bru_obs_list.bru_model}
\title{Conversion methods for \code{bru_obs} and \code{bru_obs_list} objects}
\usage{
as_bru_obs(x, ...)

as_bru_obs_list(x, .tag = NULL)

\method{as_bru_obs}{bru_obs}(x, ...)

\method{as_bru_obs_list}{bru_obs}(x, .tag = NULL)

\method{as_bru_obs_list}{list}(x, .tag = NULL)

\method{as_bru_obs_list}{bru_obs_list}(x, .tag = NULL)

\method{as_bru_obs_list}{bru}(x, .tag = NULL)

\method{as_bru_obs_list}{bru_info}(x, .tag = NULL)

\method{as_bru_obs_list}{bru_model}(x, .tag = NULL)
}
\arguments{
\item{x}{An object to convert to \link{bru_obs} or \link{bru_obs_list}}

\item{\dots}{Additional arguments passed to sub-methods.}

\item{.tag}{character; optional name for the single observation model in the
returned \link{bru_obs_list} object. Default is \code{NULL}, which results in
automatic naming based on the \code{tag} attribute of \code{x}, if present.}
}
\value{
An object of class \link{bru_obs} or \link{bru_obs_list}.
}
\description{
Methods for converting to \code{bru_obs} and \code{bru_obs_list}
objects.
}
\seealso{
\code{\link[=as_bru_comp_list]{as_bru_comp_list()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.toygroups.R
\docType{data}
\name{toygroups}
\alias{toygroups}
\title{Simulated 1D animal group locations and group sizes}
\format{
The data are a list that contains these elements:
\describe{
\item{\code{groups}:}{ A \code{data.frame} of group locations \code{x} and size \code{size}}
\item{\code{df.size}:}{ IGNORE THIS }
\item{\code{df.intensity}:}{ A \code{data.frame} with Poisson process
intensity \code{d.lambda} at locations \code{x}}
\item{\code{df.rate}:}{ A \code{data.frame} the locations \code{x} and associated \code{rate}
which parameterized the exponential distribution from which the group
sizes were drawn.}
}
}
\usage{
data(toygroups)
}
\description{
This data set serves to teach the concept of modelling species that gather in
groups and where the grouping behaviour depends on space.
}
\examples{
\donttest{
if (require(ggplot2, quietly = TRUE)) {
  # Load the data

  data("toygroups", package = "inlabru")

  # The data set is a simulation of animal groups residing in a 1D space.
  # Their locations in x-space are sampled from a Cox process with
  # intensity

  ggplot(toygroups$df.intensity) +
    geom_line(aes(x = x, y = g.lambda))

  # Adding the simulated group locations to this plot we obtain

  ggplot(toygroups$df.intensity) +
    geom_line(aes(x = x, y = g.lambda)) +
    geom_point(data = toygroups$groups, aes(x, y = 0), pch = "|")

  # Each group has a size mark attached to it.
  # These group sizes are sampled from an exponential distribution
  # for which the rate parameter depends on the x-coordinate

  ggplot(toygroups$groups) +
    geom_point(aes(x = x, y = size))

  ggplot(toygroups$df.rate) +
    geom_line(aes(x, rate))
}
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{bru_safe_inla}
\alias{bru_safe_inla}
\title{Load INLA safely for examples and tests}
\usage{
bru_safe_inla(multicore = NULL, quietly = FALSE, minimum_version = "23.1.31")
}
\arguments{
\item{multicore}{logical; if \code{TRUE}, multiple cores are allowed, and the
INLA \code{num.threads} option is not checked or altered.
If \code{FALSE}, forces \code{num.threads="1:1:1"}. Default: NULL, checks
if running in testthat or non-interactively, in which case sets
\code{multicore=FALSE}, otherwise \code{TRUE}.}

\item{quietly}{logical; if \code{FALSE} and \code{multicore} is \code{FALSE},
prints a message if the \code{num.threads} option
isn't already "1.1:1" to alert the user to the change.
Default: FALSE.}

\item{minimum_version}{character; the minimum required INLA version.
Default 23.1.31 (should always match the requirement in the package
DESCRIPTION)}
}
\value{
logical; \code{TRUE} if INLA was loaded safely, otherwise FALSE
}
\description{
Loads the INLA package with \code{requireNamespace("INLA", quietly = TRUE)}, and
optionally checks and sets the multicore \code{num.threads} INLA option.
}
\examples{
\dontrun{
if (bru_safe_inla()) {
  # Run inla dependent calculations
}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.fm_mesh_2d}
\alias{gg.fm_mesh_2d}
\title{Geom for fm_mesh_2d objects}
\usage{
\method{gg}{fm_mesh_2d}(
  data,
  color = NULL,
  alpha = NULL,
  edge.color = "grey",
  edge.linewidth = 0.25,
  interior = TRUE,
  int.color = "blue",
  int.linewidth = 0.5,
  exterior = TRUE,
  ext.color = "black",
  ext.linewidth = 1,
  crs = NULL,
  mask = NULL,
  nx = 500,
  ny = 500,
  ...
)
}
\arguments{
\item{data}{An \code{fm_mesh_2d} object.}

\item{color}{A vector of scalar values to fill the mesh with colors.
The length of the vector mus correspond to the number of mesh vertices.
The alternative name \code{colour} is also recognised.}

\item{alpha}{A vector of scalar values setting the alpha value of the colors
provided.}

\item{edge.color}{Color of the regular mesh edges.}

\item{edge.linewidth}{Line width for the regular mesh edges. Default 0.25}

\item{interior}{If TRUE, plot the interior boundaries of the mesh.}

\item{int.color}{Color used to plot the interior constraint edges.}

\item{int.linewidth}{Line width for the interior constraint edges. Default
0.5}

\item{exterior}{If TRUE, plot the exterior boundaries of the mesh.}

\item{ext.color}{Color used to plot the exterior boundary edges.}

\item{ext.linewidth}{Line width for the exterior boundary edges. Default 1}

\item{crs}{A CRS object supported by \code{\link[fmesher:fm_transform]{fmesher::fm_transform()}} defining the
coordinate system to project the mesh to before plotting.}

\item{mask}{A \code{SpatialPolygon} or \code{sf} polygon defining the region that is
plotted.}

\item{nx}{Number of pixels in x direction (when plotting using the color
parameter).}

\item{ny}{Number of pixels in y direction (when plotting using the color
parameter).}

\item{\dots}{ignored arguments (S3 generic compatibility).}
}
\value{
\code{geom_line} return values or, if the color argument is used, the
values of \code{\link[=gg.SpatialPixelsDataFrame]{gg.SpatialPixelsDataFrame()}}.
}
\description{
This function extracts the graph of an \link[fmesher:fm_mesh_2d]{fmesher::fm_mesh_2d} object and uses
\code{geom_line} to visualize the graph's edges. Alternatively, if the \code{color}
argument is provided, interpolates the colors across for a set of
SpatialPixels covering the mesh area and calls \code{\link[=gg.SpatialPixelsDataFrame]{gg.SpatialPixelsDataFrame()}}
to plot the interpolation.
Requires the \code{ggplot2} package.

Also see the \code{\link[fmesher:geom_fm]{fmesher::geom_fm()}} method.
}
\examples{
\donttest{
if (require(fmesher, quietly = TRUE) &&
  require(ggplot2, quietly = TRUE)) {
  # Load Gorilla data
  gorillas <- inlabru::gorillas_sf

  # Plot mesh using default edge colors

  ggplot() +
    gg(gorillas$mesh)

  # Don't show interior and exterior boundaries

  ggplot() +
    gg(gorillas$mesh, interior = FALSE, exterior = FALSE)

  # Change the edge colors

  ggplot() +
    gg(gorillas$mesh,
      edge.color = "green",
      int.color = "black",
      ext.color = "blue"
    )

  # Use the x-coordinate of the vertices to colorize the triangles and
  # mask the plotted area by the survey boundary, i.e. only plot the inside

  xcoord <- gorillas$mesh$loc[, 1]
  ggplot() +
    gg(gorillas$mesh, color = (xcoord - 580), mask = gorillas$boundary) +
    gg(gorillas$boundary, alpha = 0)
}
}

}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nlinla.R
\name{bru_compute_linearisation}
\alias{bru_compute_linearisation}
\alias{bru_compute_linearisation.bru_comp}
\alias{bru_compute_linearisation.bru_obs}
\alias{bru_compute_linearisation.bru_obs_list}
\alias{bru_compute_linearisation.bru_model}
\title{Compute inlabru model linearisation information}
\usage{
bru_compute_linearisation(...)

\method{bru_compute_linearisation}{bru_comp}(
  cmp,
  model,
  lhood_expr,
  data,
  data_extra,
  input,
  state,
  comp_simple,
  effects,
  pred0,
  used,
  is_rowwise,
  eps,
  n_pred = NULL,
  ...
)

\method{bru_compute_linearisation}{bru_obs}(
  lhood,
  model,
  data,
  input,
  state,
  comp_simple,
  eps,
  ...
)

\method{bru_compute_linearisation}{bru_obs_list}(
  lhoods,
  model,
  input,
  state,
  comp_simple,
  eps = 1e-05,
  ...
)

\method{bru_compute_linearisation}{bru_model}(model, lhoods, input, state, comp_simple, ...)
}
\arguments{
\item{\dots}{Parameters passed on to other methods}

\item{cmp}{A \link{bru_comp} object}

\item{model}{A \code{bru_model} object}

\item{lhood_expr}{A predictor expression}

\item{data}{Input data}

\item{data_extra}{Additional data for the predictor}

\item{input}{Precomputed component inputs from \code{bru_input()}}

\item{state}{The state information, as a list of named vectors}

\item{comp_simple}{Component evaluation information
\itemize{
\item For \code{bru_comp}: A \link{bm_taylor} object
\item For \code{bru_obs}: A \link{bm_list} object
for the components in the likelihood
\item For \code{bru_obs_list}: A list of \link{bm_list} objects
}}

\item{effects}{\itemize{
\item For \code{bru_comp}:
Precomputed effect list for all components involved in the likelihood
expression
}}

\item{pred0}{Precomputed predictor for the given state}

\item{used}{A \code{\link[=bru_used]{bru_used()}} object for the predictor expression}

\item{is_rowwise}{logical; If \code{FALSE}, the predictor expression may
involve several rows of the input data to influence the same row.}

\item{eps}{The finite difference step size}

\item{n_pred}{The length of the predictor expression. If not \code{NULL}, scalar
predictor evaluations are expanded to vectors of length \code{n_pred}.}

\item{lhood}{A \code{bru_obs} object}

\item{lhoods}{A \code{bru_obs_list} object}
}
\description{
Compute inlabru model linearisation information
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.sf}
\alias{gg.sf}
\title{Geom helper for sf objects}
\usage{
\method{gg}{sf}(data, mapping = NULL, ..., geom = "sf")
}
\arguments{
\item{data}{An \code{sf} object.}

\item{mapping}{Default mapping is \code{ggplot2::aes(geometry = ...)},
where the geometry name is obtained from \code{attr(data, "sf_column")}.
This is merged with the user supplied mapping.}

\item{\dots}{Arguments passed on to \code{geom_sf} or \code{geom_tile}.}

\item{geom}{Either "sf" (default) or "tile". For "tile", uses
\code{geom_tile(..., stat = "sf_coordinates")}, intended for converting point data
to grid tiles with the \code{fill} aesthetic, which is by default set to the first
data column.}
}
\value{
A ggplot return value
}
\description{
This function uses \code{geom_sf()}, unless overridden by the geom argument.
Requires the \code{ggplot2} package.
}
\examples{
\donttest{
if (require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("tidyterra", quietly = TRUE)) {
  # Load Gorilla data

  gorillas <- inlabru::gorillas_sf
  gorillas$gcov <- gorillas_sf_gcov()

  # Plot Gorilla elevation covariate provided as terra::rast.

  ggplot() +
    gg(gorillas$gcov$elevation)

  # Add Gorilla survey boundary and nest sightings

  ggplot() +
    gg(gorillas$gcov$elevation) +
    gg(gorillas$boundary, alpha = 0) +
    gg(gorillas$nests)

  # Load pantropical dolphin data

  mexdolphin <- inlabru::mexdolphin_sf

  # Plot the pantropical survey boundary, ship transects and dolphin
  # sightings

  ggplot() +
    gg(mexdolphin$ppoly, alpha = 0.5) + # survey boundary
    gg(mexdolphin$samplers) + # ship transects
    gg(mexdolphin$points) # dolphin sightings

  # Change color

  ggplot() +
    gg(mexdolphin$ppoly, color = "green", alpha = 0.5) + # survey boundary
    gg(mexdolphin$samplers, color = "red") + # ship transects
    gg(mexdolphin$points, color = "blue") # dolphin sightings


  # Visualize data annotations: line width by segment number

  names(mexdolphin$samplers) # 'seg' holds the segment number
  ggplot() +
    gg(mexdolphin$samplers, aes(color = seg))

  # Visualize data annotations: point size by dolphin group size

  names(mexdolphin$points) # 'size' holds the group size
  ggplot() +
    gg(mexdolphin$points, aes(size = size))
}
}

}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_shift}
\alias{bm_shift}
\alias{bru_mapper_shift}
\title{Mapper for element-wise shifting}
\usage{
bm_shift(mapper = NULL)

bru_mapper_shift(...)
}
\arguments{
\item{mapper}{If non-NULL, a \code{bru_mapper} to be shifted.}

\item{\dots}{Arguments passed on to \code{\link[=bm_shift]{bm_shift()}}}
}
\description{
Create a standalone
shift mapper that can be used as part of a \code{bm_pipe}.
If \code{mapper} is non-null, the \code{bm_shift()} constructor
returns
\code{bm_pipe(list(mapper = mapper, shift = bm_shift()))}
}
\examples{
m <- bm_shift()
ibm_eval2(m, c(1, 2, 1, 2), 1:4)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_obs_methods}
\alias{bru_obs_methods}
\alias{bru_obs_inla_family}
\alias{bru_obs_inla_family.bru_obs}
\alias{bru_obs_inla_family.bru_obs_list}
\alias{bru_obs_control_family}
\alias{bru_obs_control_family.bru_obs}
\alias{bru_obs_control_family.bru_obs_list}
\alias{bru_obs_control_gcpo}
\alias{bru_obs_control_gcpo.bru_obs}
\alias{bru_obs_control_gcpo.bru_obs_list}
\title{Utility functions for bru observation model objects}
\usage{
bru_obs_inla_family(x, ...)

\method{bru_obs_inla_family}{bru_obs}(x, ...)

\method{bru_obs_inla_family}{bru_obs_list}(x, ...)

bru_obs_control_family(x, control.family = NULL, ...)

\method{bru_obs_control_family}{bru_obs}(x, control.family = NULL, ...)

\method{bru_obs_control_family}{bru_obs_list}(x, control.family = NULL, ...)

bru_obs_control_gcpo(x, ...)

\method{bru_obs_control_gcpo}{bru_obs}(
  x,
  index_offset,
  index_length = bru_response_size(x),
  force_weights,
  ...
)

\method{bru_obs_control_gcpo}{bru_obs_list}(x, control.gcpo = NULL, ...)
}
\arguments{
\item{x}{Object of \code{bru_obs} or \code{bru_obs_list} type}

\item{\dots}{Further arguments passed on to the submethods}

\item{control.family}{list of INLA \code{control.family} options to override}

\item{index_offset}{integer; offset to add to indices in \code{control.gcpo}}

\item{index_length}{integer; length of the response vector for the
observation model. Defaults to \code{bru_response_size(x)}.}

\item{force_weights}{logical; if \code{TRUE}, ensure that \code{control.gcpo$weights}
is populated. This is needed if any of the observation models in a
\code{bru_obs_list} has non-null \code{control.gcpo$weights}.}

\item{control.gcpo}{list of INLA \code{control.gcpo} default options}
}
\value{
\itemize{
\item \code{bru_obs_inla_family()} returns a string or vector of strings
}

\itemize{
\item \code{bru_obs_control_family()} returns a list with
\code{INLA::control.family} options, or a list of such lists, with one element
per observation model
}

\itemize{
\item \code{bru_obs_control_gcpo()} returns a list with
\code{INLA::control.gcpo} options, with predictor/response variable indices
unified for multi-observation models. If a \code{bru_obs} model has a NULL
\code{control.gcpo} argument, an empty list is returned for that model.
}
}
\description{
Utility functions for bru observation model objects
}
\seealso{
\code{\link[=summary.bru_obs]{summary.bru_obs()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_logitaverage}
\alias{bm_logitaverage}
\title{Mapper for logit-sum-inverse-logit aggregation}
\usage{
bm_logitaverage(n_block = NULL)
}
\arguments{
\item{n_block}{Predetermined number of output blocks. If \code{NULL}, overrides
the maximum block index in the inputs. The priority order is \code{input$n_block},
the mapper definition \code{n_block}, then \code{max(input$block)}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Constructs a mapper that averages elements of \code{plogis(state)}, with optional
non-negative weighting, and then takes the \code{qlogis()}.  Relies on the input
handling methods for \code{bm_aggregate}. To avoid numerical issues, it uses
\code{plogis(x, log.p = TRUE)}, \code{plogis(-x, log.p = TRUE)}, and the equivalent of
two applications of \code{\link[=bm_logsumexp]{bm_logsumexp()}} to evaluate
\eqn{\log(p_k)-\log(1-p_k)}{log(p[k])-log(1-p[k])}, where

\eqn{p_k=\sum_{i\in I_k} w_i / (1+e^{-\eta_i}) / \sum_{i\in I_k} w_i
}{p[k]=sum(w[i] * plogis(eta[i])) / sum(w[i])}
}
\examples{
m <- bm_logitaverage()
ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4), 11:14)
ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4, n_block = 3), 11:14)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_index.R
\name{bru_index}
\alias{bru_index}
\alias{bru_index.bru_obs}
\alias{bru_index.bru_obs_list}
\alias{bru_index.bru}
\alias{bru_index.bru_info}
\alias{bru_index.bru_comp}
\alias{bru_index.bru_comp_list}
\alias{bru_index.bru_model}
\alias{index_eval}
\title{Extract predictor or component index information}
\usage{
bru_index(object, ...)

\method{bru_index}{bru_obs}(object, what = NULL, ...)

\method{bru_index}{bru_obs_list}(object, tag = NULL, what = NULL, ...)

\method{bru_index}{bru}(object, tag = NULL, what = NULL, ...)

\method{bru_index}{bru_info}(object, tag = NULL, what = NULL, ...)

\method{bru_index}{bru_comp}(object, inla_f, ...)

\method{bru_index}{bru_comp_list}(object, inla_f, ...)

\method{bru_index}{bru_model}(object, used, ...)

index_eval(...)
}
\arguments{
\item{object}{A \link{component}.}

\item{\dots}{Arguments passed on to sub-methods.}

\item{what}{\code{character} or \code{NULL}; One of \code{NULL}, "all", "observed", and
"missing". If \code{NULL} (default) or "all", gives the index vector for the
full sub-model predictor. If "observed", gives the index vector for the
observed part (response is not \code{NA}). If "missing", gives the index vector
for the missing part (response is \code{NA}) of the model.}

\item{tag}{\code{character} or \code{integer}; Either a character vector identifying
the tags of one or more of the \code{\link[=bru_obs]{bru_obs()}} observation models, or an
integer vector identifying models by their \code{\link[=bru]{bru()}} specification order. If
\code{NULL} (default) computes indices for all sub-models.}

\item{inla_f}{logical; when \code{TRUE}, must result in
values compatible with \code{INLA::f(...)}
an specification and corresponding \code{INLA::inla.stack(...)} constructions.}

\item{used}{A \code{\link[=bru_used]{bru_used()}} object}
}
\value{
\itemize{
\item \code{bru_index(bru_obs)}: An \code{integer} vector.
}

\itemize{
\item \code{bru_index(bru_obs_list)}: An \code{integer} vector.
}

\itemize{
\item \code{bru_index(bru)}: An \code{integer} vector.
}

\itemize{
\item \code{bru_index(bru_info)}: An \code{integer} vector.
}

\itemize{
\item \code{bru_index(bru_comp)}:
A list of indices into the latent variables compatible with the
component mapper.
}

\itemize{
\item \code{bru_index(bru_comp_list)}:
A list of list of indices into the latent variables compatible with each
component mapper.
}

\itemize{
\item \code{bru_index(bru_model)}: A named list of \code{idx_full} and \code{idx_inla},
named list of indices, and \code{inla_subset}, and \code{inla_subset},
a named list of logical subset specifications for extracting the \code{INLA::f()}
compatible index subsets.
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Extract the index vector for a \code{\link[=bru_obs]{bru_obs()}} predictor,
or the whole or a subset of a full \code{\link[=bru]{bru()}} predictor.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_index(bru_obs)}: Extract the index vector for the predictor vector for a
\code{\link[=bru_obs]{bru_obs()}} sub-model. The indices are relative to the sub-model, and need
to be appropriately offset to be used in the full model predictor.

\item \code{bru_index(bru_obs_list)}: Extract the index vector for "APredictor" for one or
more specified observation \code{\link[=bru_obs]{bru_obs()}} sub-models. Accepts any combination
of \code{tag} and \code{what}.

\item \code{bru_index(bru)}: Extract the index vector for "APredictor" for one or
more specified observation \code{\link[=bru_obs]{bru_obs()}} sub-models. Accepts any combination
of \code{tag} and \code{what}.

\item \code{bru_index(bru_info)}: Extract the index vector for "APredictor" for one or
more specified observation \code{\link[=bru_obs]{bru_obs()}} sub-models. Accepts any combination
of \code{tag} and \code{what}.

\item \code{bru_index(bru_model)}: Compute all index values for a \code{\link[=bru_model]{bru_model()}} object.
Computes the index value matrices for included components according to the
\code{used} argument.

}}
\section{Functions}{
\itemize{
\item \code{index_eval()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} since "2.12.0.9023".
Use \code{bru_index()} instead.

}}
\examples{
\dontshow{if (bru_safe_inla()) withAutoprint(\{ # examplesIf}
fit <- bru(
  ~ 0 + x,
  bru_obs(
    y ~ .,
    data = data.frame(x = 1:3, y = 1:3 + rnorm(3)),
    tag = "A"
  ),
  bru_obs(
    y ~ .,
    data = data.frame(x = 1:4, y = c(NA, NA, 3:4) + rnorm(4)),
    tag = "B"
  ),
  options = list(bru_run = FALSE) # We only need the model structure
)
bru_index(fit)
bru_index(fit, "A")
bru_index(fit, "B")
bru_index(fit, c("B", "A"))
bru_index(fit, what = "missing")
\dontshow{\}) # examplesIf}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com},
Finn Lindgren \email{finn.lindgren@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R, R/model.R
\name{ibm_linear}
\alias{ibm_linear}
\alias{ibm_linear.default}
\alias{ibm_linear.bm_multi}
\alias{ibm_linear.bm_collect}
\alias{ibm_linear.bm_repeat}
\alias{ibm_linear.bm_sum}
\alias{ibm_linear.bru_comp}
\title{Compute a mapper linearisation}
\usage{
ibm_linear(mapper, input, state = NULL, ...)

\method{ibm_linear}{default}(mapper, input, state, ...)

\method{ibm_linear}{bm_multi}(mapper, input, state, inla_f = FALSE, ...)

\method{ibm_linear}{bm_collect}(mapper, input, state, inla_f = FALSE, ...)

\method{ibm_linear}{bm_repeat}(mapper, input, state, ...)

\method{ibm_linear}{bm_sum}(mapper, input, state, ...)

\method{ibm_linear}{bru_comp}(mapper, input, state = NULL, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{...}{Arguments passed on to other methods}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}
}
\value{
A \link{bm_taylor} object.
The \code{state0} information in the affine mapper indicates for which state
the \code{offset} was evaluated; The affine mapper output is defined as

\if{html}{\out{<div class="sourceCode">}}\preformatted{effect(input, state) =
  offset(input, state0) + jacobian(input, state0) \%*\% (state - state0)
}\if{html}{\out{</div>}}
}
\description{
Implementations must return a \link{bm_taylor} object
The linearisation information includes \code{offset}, \code{jacobian}, and \code{state0}.
The state information indicates for which state the \code{offset} was evaluated,
with \code{NULL} meaning all-zero.
The linearised mapper output is defined as

\if{html}{\out{<div class="sourceCode">}}\preformatted{effect(input, state) =
  offset(input, state0) + jacobian(input, state0) \%*\% (state - state0)
}\if{html}{\out{</div>}}

The default method calls \code{\link[=ibm_eval]{ibm_eval()}} and \code{\link[=ibm_jacobian]{ibm_jacobian()}} to generate
the needed information.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_linear(default)}: Calls \code{\link[=ibm_eval2]{ibm_eval2()}} and returns a \link{bm_taylor} object.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_mesh_B}
\alias{bm_mesh_B}
\alias{bru_mapper_mesh_B}
\title{Deprecated basis conversion mapper}
\usage{
bm_mesh_B(mesh, B)

bru_mapper_mesh_B(...)
}
\arguments{
\item{mesh}{Object supported by a \code{\link[=bru_mapper]{bru_mapper()}} method.}

\item{B}{a square or tall basis conversion matrix}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Old deprecated name for \code{\link[=bm_reparam]{bm_reparam()}}
}
\section{Functions}{
\itemize{
\item \code{bru_mapper_mesh_B()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Deprecated name for \code{bm_mesh_B}

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{ibm_simplify}
\alias{ibm_simplify}
\alias{ibm_simplify.default}
\alias{ibm_simplify.bm_pipe}
\title{Simplify a mapper}
\usage{
ibm_simplify(mapper, input = NULL, state = NULL, ...)

\method{ibm_simplify}{default}(mapper, input = NULL, state = NULL, ...)

\method{ibm_simplify}{bm_pipe}(
  mapper,
  input = NULL,
  state = NULL,
  inla_f = FALSE,
  ...,
  n_state = NULL
)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{...}{Arguments passed on to other methods}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{n_state}{integer giving the length of the state vector for mappers
that have state dependent output size.}
}
\value{
The original mapper is returned for non-linear mappers, and the
output of \code{\link[=ibm_linear]{ibm_linear()}} is returned for linear mappers.
}
\description{
Implementations must return a \link{bru_mapper} object.
The default method returns the result of \code{\link[=ibm_linear]{ibm_linear()}} for linear/affine
mappers, and the original \code{mapper} for non-linear mappers.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_simplify(default)}: Calls \code{\link[=ibm_linear]{ibm_linear()}} for linear mappers, and returns the original mapper
for non-linear mappers.

\item \code{ibm_simplify(bm_pipe)}: Constructs a simplified \code{pipe} mapper. For fully linear pipes, calls
\code{\link[=ibm_linear]{ibm_linear()}}.
For partially non-linear pipes, replaces each sequence of linear mappers with
a single \code{\link[=bm_taylor]{bm_taylor()}} mapper, while keeping the full list of
original mapper names, allowing the original \code{input} structure to be used
also with the simplified mappers, since the \code{taylor} mappers are not
dependent on inputs.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{eval_spatial}
\alias{eval_spatial}
\alias{eval_spatial.SpatialPolygonsDataFrame}
\alias{eval_spatial.SpatialPixelsDataFrame}
\alias{eval_spatial.SpatialGridDataFrame}
\alias{eval_spatial.sf}
\alias{eval_spatial.SpatRaster}
\alias{eval_spatial.stars}
\title{Evaluate spatial covariates}
\usage{
eval_spatial(data, where, layer = NULL, selector = NULL)

\method{eval_spatial}{SpatialPolygonsDataFrame}(data, where, layer = NULL, selector = NULL)

\method{eval_spatial}{SpatialPixelsDataFrame}(data, where, layer = NULL, selector = NULL)

\method{eval_spatial}{SpatialGridDataFrame}(data, where, layer = NULL, selector = NULL)

\method{eval_spatial}{sf}(data, where, layer = NULL, selector = NULL)

\method{eval_spatial}{SpatRaster}(data, where, layer = NULL, selector = NULL)

\method{eval_spatial}{stars}(data, where, layer = NULL, selector = NULL)
}
\arguments{
\item{data}{Spatial data}

\item{where}{Where to evaluate the data}

\item{layer}{Which \code{data} layer to extract (as integer or character).
May be a vector, specifying a separate layer for each \code{where} item.}

\item{selector}{The name of a variable in \code{where} specifying the \code{layer}
information.}
}
\description{
Evaluate spatial covariates
}
\section{Methods (by class)}{
\itemize{
\item \code{eval_spatial(SpatialPolygonsDataFrame)}: Compatibility wrapper for \code{eval_spatial.sf}

\item \code{eval_spatial(sf)}: Supports point-in-polygon information lookup.
Other combinations are untested.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.mexdolphin.R
\name{mexdolphin}
\alias{mexdolphin}
\title{Deprecated alias for sp version of the mexdolphin dataset}
\description{
Deprecated dataset name for the \code{sp} version
of \link{mexdolphin_sf}. Use \code{\link[=mexdolphin_sp]{mexdolphin_sp()}} instead.
}
\seealso{
\link{mexdolphin_sf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{generate}
\alias{generate}
\alias{generate.bru}
\title{Generate samples from fitted bru models}
\usage{
generate(object, ...)

\method{generate}{bru}(
  object,
  newdata = NULL,
  formula = NULL,
  n.samples = 100,
  seed = 0L,
  num.threads = NULL,
  used = NULL,
  ...,
  include = deprecated(),
  exclude = deprecated()
)
}
\arguments{
\item{object}{A \code{bru} object obtained by calling \code{\link[=bru]{bru()}}.}

\item{\dots}{additional, unused arguments.}

\item{newdata}{A \code{data.frame} or \code{SpatialPointsDataFrame} of covariates
needed for sampling.}

\item{formula}{A formula where the right hand side defines an R expression
to evaluate for each generated sample. If \code{NULL}, the latent and
hyperparameter states are returned as named list elements.
See Details for more information.}

\item{n.samples}{Integer setting the number of samples to draw in order to
calculate the posterior statistics.
The default, 100, is rather low but provides a quick approximate result.}

\item{seed}{Random number generator seed passed on to
\code{INLA::inla.posterior.sample}}

\item{num.threads}{Specification of desired number of threads for parallel
computations. Default NULL, leaves it up to INLA.
When seed != 0, overridden to "1:1:1"}

\item{used}{Either \code{NULL} or a \code{\link[=bru_used]{bru_used()}} object.
Default, \code{NULL}, uses auto-detection of used variables in the formula.}

\item{include, exclude}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} If auto-detection
of used variables fails, use \code{used} instead.}
}
\value{
The form of the value returned by \code{generate()} depends on the data
class and prediction formula. Normally, a data.frame is returned, or a list
of data.frames (if the prediction formula generates a list)

List of generated samples
}
\description{
Generic function for sampling for fitted models. The function invokes
particular methods which depend on the class of the first argument.

Takes a fitted \code{bru} object produced by the function \code{\link[=bru]{bru()}} and produces
samples given a new set of values for the model covariates or the original
values used for the model fit. The samples can be based on any R expression
that is valid given these values/covariates and the joint
posterior of the estimated random effects.
}
\details{
In addition to the component names (that give the effect of each component
evaluated for the input data), the suffix \verb{_latent} variable name can be used
to directly access the latent state for a component, and the suffix function
\verb{_eval} can be used to evaluate a component at other input values than the
expressions defined in the component definition itself, e.g.
\code{field_eval(cbind(x, y))} for a component that was defined with
\code{field(coordinates, ...)} (see also \code{\link[=bru_comp_eval]{bru_comp_eval()}}).

For "iid" models with \code{mapper = bm_index(n)}, \code{rnorm()} is used to
generate new realisations for indices greater than \code{n}, if accessed
via \verb{<name>_eval(...)}.
}
\examples{
\donttest{
if (
  bru_safe_inla() &&
    requireNamespace("sn", quietly = TRUE)
) {
  # Generate data for a simple linear model

  input.df <- data.frame(x = cos(1:10))
  input.df <- within(
    input.df,
    {
      y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
    }
  )

  # Fit the model

  fit <- bru(
    y ~ xeff(main = x, model = "linear"),
    family = "gaussian",
    data = input.df
  )
  summary(fit)

  # Generate samples for some predefined x

  df <- data.frame(x = seq(-4, 4, by = 0.1))
  smp <- generate(fit, df, ~ xeff + Intercept, n.samples = 10)

  # Plot the resulting realizations

  plot(df$x, smp[, 1], type = "l")
  for (k in 2:ncol(smp)) {
    points(df$x, smp[, k], type = "l")
  }

  # We can also draw samples form the joint posterior

  df <- data.frame(x = 1)
  smp <- generate(fit, df, ~ data.frame(xeff, Intercept), n.samples = 10)
  smp[[1]]

  # ... and plot them
  if (require(ggplot2, quietly = TRUE)) {
    plot(do.call(rbind, smp))
  }
}
}

}
\seealso{
\link{predict.bru}
}
\concept{sample generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.toypoints.R
\docType{data}
\name{toypoints}
\alias{toypoints}
\title{Simulated 2D point process data}
\format{
The data are a list that contains these elements:
\describe{
\item{\code{points}}{An \code{sf} object of point locations and and \code{z}
measurements}
\item{\code{mesh}}{An \code{fm_mesh_2d} object}
\item{\code{boundary}}{An \code{sf} polygon denting the region of interest}
\item{\code{pred_locs}}{A \code{sf} object with prediction point locations}
}
}
\usage{
data(toypoints)
}
\description{
This data set serves as an example for basic inlabru.
}
\examples{
if (require("ggplot2")) {
  ggplot() +
    fmesher::geom_fm(data = toypoints$mesh, alpha = 0) +
    geom_sf(data = toypoints$boundary, fill = "blue", alpha = 0.1) +
    geom_sf(data = toypoints$points, aes(color = z)) +
    scale_color_viridis_c()
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_fm_mesh_1d}
\alias{bm_fm_mesh_1d}
\alias{bru_mapper.fm_mesh_1d}
\title{Mapper for \code{fm_mesh_1d}}
\usage{
\method{bru_mapper}{fm_mesh_1d}(mesh, indexed = TRUE, ...)
}
\arguments{
\item{mesh}{An \code{fm_mesh_1d} object to use as a mapper}

\item{indexed}{logical; If \code{TRUE} (default), the \code{ibm_values()} output will
be the integer indexing sequence for the latent variables (needed for
\code{spde} models). If \code{FALSE}, points representative of the basis centres
are returned (useful for an interpolator for \code{rw2} models and similar,
for \code{fmesher} versions \verb{>= 0.3.0.9002}).}

\item{...}{Arguments passed on to \code{\link[=bm_fmesher]{bm_fmesher()}}}
}
\value{
A \code{bm_fm_mesh_1d} or \code{bm_fmesher} object. The the
general \code{\link[=bm_fmesher]{bm_fmesher()}} mapper handles all indexed \code{fmesher}
objects.
}
\description{
Create mapper for an \code{fm_mesh_1d} object
}
\examples{
m <- bru_mapper(fmesher::fm_mesh_1d(c(1:3, 5, 7)))
ibm_values(m)
ibm_eval(m, 1:7, 1:5)

m <- bru_mapper(fmesher::fm_mesh_1d(c(1:3, 5, 7)), indexed = FALSE)
ibm_values(m)
ibm_eval(m, 1:7, 1:5)

m <- bru_mapper(
  fmesher::fm_mesh_1d(c(1:3, 5, 7), degree = 2, boundary = "free"),
  indexed = FALSE
)
ibm_values(m)
ibm_eval(m, 1:7, 1:6)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.Poisson1_1D.R
\docType{data}
\name{Poisson1_1D}
\alias{Poisson1_1D}
\alias{lambda1_1D}
\alias{E_nc1}
\alias{pts1}
\alias{countdata1}
\title{1-Dimensional Homogeneous Poisson example.}
\format{
The data contain the following \code{R} objects:
\describe{
\item{\code{lambda1_1D}}{ A function defining the intensity function of a
nonhomogeneous Poisson process. Note that this function is only defined on
the interval (0,55).}
\item{\code{E_nc1}}{ The expected counts of the gridded data.}
\item{\code{pts1}}{ The locations of the observed points (a data frame with
one column, named \code{x}).}
\item{\code{countdata1}}{ A data frame with three columns, containing the
count data:}
\describe{
\item{\code{x}}{ The grid cell midpoint.}
\item{\code{count}}{ The number of detections in the cell.}
\item{\code{exposure}}{ The width of the cell.}
}
}
}
\usage{
data(Poisson1_1D)
}
\description{
Point data and count data, together with intensity function and
expected counts for a homogeneous 1-dimensional Poisson process example.
}
\examples{
\donttest{
if (require("ggplot2", quietly = TRUE)) {
  data(Poisson1_1D)
  ggplot(countdata1) +
    geom_point(data = countdata1, aes(x = x, y = count), col = "blue") +
    ylim(0, max(countdata1$count)) +
    geom_point(data = pts1, aes(x = x), y = 0.2, shape = "|", cex = 4) +
    geom_point(
      data = countdata1, aes(x = x), y = 0, shape = "+",
      col = "blue", cex = 4
    ) +
    xlab(expression(bold(s))) +
    ylab("count")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{predict.bru}
\alias{predict.bru}
\title{Prediction from fitted bru model}
\usage{
\method{predict}{bru}(
  object,
  newdata = NULL,
  formula = NULL,
  n.samples = 100,
  seed = 0L,
  probs = c(0.025, 0.5, 0.975),
  num.threads = NULL,
  used = NULL,
  drop = FALSE,
  ...,
  include = deprecated(),
  exclude = deprecated()
)
}
\arguments{
\item{object}{An object obtained by calling \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}}.}

\item{newdata}{A \code{data.frame} or \code{SpatialPointsDataFrame} of covariates
needed for the prediction.}

\item{formula}{A formula where the right hand side defines an R expression
to evaluate for each generated sample. If \code{NULL}, the latent and
hyperparameter states are returned as named list elements.
See Details for more information.}

\item{n.samples}{Integer setting the number of samples to draw in order to
calculate the posterior statistics. The default is rather low but provides
a quick approximate result.}

\item{seed}{Random number generator seed passed on to \code{inla.posterior.sample}}

\item{probs}{A numeric vector of probabilities with values in \verb{[0, 1]},
passed to \code{stats::quantile}}

\item{num.threads}{Specification of desired number of threads for parallel
computations. Default NULL, leaves it up to INLA.
When seed != 0, overridden to "1:1:1"}

\item{used}{Either \code{NULL} or a \code{\link[=bru_used]{bru_used()}} object.
Default, \code{NULL}, uses auto-detection of used variables in the formula.}

\item{drop}{logical; If \code{drop=FALSE}, and
the prediction summary has the same number of rows as \code{newdata}, then the
output is a joined object. Default \code{FALSE}.}

\item{\dots}{Additional arguments passed on to \code{inla.posterior.sample()}}

\item{include, exclude}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} If auto-detection
of used variables fails, use \code{used} instead.}
}
\value{
a \code{data.frame}, \code{sf}, or \verb{Spatial*} object with predicted mean values
and other summary statistics attached. Non-S4 object outputs have the class
"bru_prediction" added at the front of the class list.
}
\description{
Takes a fitted \code{bru} object produced by the function \code{\link[=bru]{bru()}} and produces
predictions given a new set of values for the model covariates or the
original values used for the model fit. The predictions can be based on any
R expression that is valid given these values/covariates and the joint
posterior of the estimated random effects.
}
\details{
Mean value predictions are accompanied by the standard errors, upper and
lower 2.5\% quantiles, the
median, variance, coefficient of variation as well as the variance and
minimum and maximum sample
value drawn in course of estimating the statistics.

Internally, this method calls \code{\link[=generate.bru]{generate.bru()}} in order to draw samples from
the model.

In addition to the component names (that give the effect of each component
evaluated for the input data), the suffix \verb{_latent} variable name can be used
to directly access the latent state for a component, and the suffix function
\verb{_eval} can be used to evaluate a component at other input values than the
expressions defined in the component definition itself, e.g.
\code{field_eval(cbind(x, y))} for a component that was defined with
\code{field(coordinates, ...)} (see also \code{\link[=bru_comp_eval]{bru_comp_eval()}}).

For "iid" models with \code{mapper = bm_index(n)}, \code{rnorm()} is used to
generate new realisations for indices greater than \code{n}, if accessed
via \verb{<name>_eval(...)}.
}
\examples{
\donttest{
if (bru_safe_inla() &&
  requireNamespace("sn", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE)) {
  # Load the Gorilla data

  gorillas <- gorillas_sf

  # Plot the Gorilla nests, the mesh and the survey boundary

  ggplot() +
    gg(gorillas$mesh) +
    gg(gorillas$nests) +
    gg(gorillas$boundary, alpha = 0.1)

  # Define SPDE prior

  matern <- INLA::inla.spde2.pcmatern(
    gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(0.01, 0.01)
  )

  # Define domain of the LGCP as well as the model components (spatial SPDE
  # effect and Intercept)

  cmp <- geometry ~ field(geometry, model = matern) + Intercept(1)

  # Fit the model, with "eb" instead of full Bayes
  fit <- lgcp(
    cmp,
    data = gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(control.inla = list(int.strategy = "eb"))
  )

  # Once we obtain a fitted model the predict function can serve various
  # purposes.
  # The most basic one is to determine posterior statistics of a univariate
  # random variable in the model, e.g. the intercept

  icpt <- predict(fit, NULL, ~ c(Intercept = Intercept_latent))
  plot(icpt)

  # The formula argument can take any expression that is valid within the
  # model, for instance a non-linear transformation of a random variable

  exp.icpt <- predict(fit, NULL, ~ c(
    "Intercept" = Intercept_latent,
    "exp(Intercept)" = exp(Intercept_latent)
  ))
  plot(exp.icpt, bar = TRUE)

  # The intercept is special in the sense that it does not depend on other
  # variables or covariates. However, this is not true for the smooth spatial
  # effects 'field'.
  # In order to predict 'field' we have to define where (in space) to
  # predict.
  # For this purpose, the second argument of the predict function can take
  # \code{data.frame} objects as well as sf (and legacy sp/Spatial) objects.
  # For instance, we might want to predict 'field' at the locations of the
  # mesh vertices. Using

  vrt <- fmesher::fm_vertices(gorillas$mesh, format = "sf")

  # we obtain these vertices as an sf object with POINT geometries

  ggplot() +
    gg(gorillas$mesh) +
    gg(vrt, color = "red")

  # Predicting 'field' at these locations works as follows

  field <- predict(fit, vrt, ~field)

  # Note that just like the input also the output will be a sf object with
  # points and that the predicted statistics are simply added as columns

  class(field)
  head(vrt)
  head(field)

  # Plotting the mean, for instance, at the mesh node is straight forward

  ggplot() +
    gg(gorillas$mesh) +
    gg(field, aes(color = mean), size = 2)

  # However, we are often interested in a spatial field and thus a linear
  # interpolation, which can be achieved by using the gg mechanism for meshes

  ggplot() +
    gg(gorillas$mesh, color = field$mean)

  # Alternatively, we can predict the spatial field at a grid of locations,
  # e.g. a sf object with a grid of points covering the relevant part of mesh

  pxl <- fmesher::fm_pixels(gorillas$mesh,
    format = "sf",
    mask = gorillas$boundary
  )
  field2 <- predict(fit, pxl, ~field)

  # This will give us a sf with the columns we are looking for

  head(field2)
  ggplot() +
    gg(gorillas$boundary) +
    gg(data = field2, geom = "tile")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.shrimp.R
\docType{data}
\name{shrimp}
\alias{shrimp}
\title{Blue and red shrimp in the Western Mediterranean Sea}
\format{
A list of objects:
\describe{
\item{\code{hauls}:}{ An \code{sf} object containing haul locations}
\item{\code{mesh}:}{ An \code{fm_mesh_2d} object containing a Delaunay triangulation
mesh (a type of discretization of continuous space) covering the haul
locations.}
\describe{
\item{\code{catch}}{Catch in Kg.}
\item{\code{landing}}{Landing in Kg.}
\item{\code{depth}}{Mean depth (in metres) of the fishery haul.}
}
}
}
\source{
Pennino, Maria Grazia. Personal communication.
}
\usage{
data(shrimp)
}
\description{
Blue and red shrimp in the Western Mediterranean Sea.
}
\examples{
\donttest{
if (require(ggplot2, quietly = TRUE)) {
  data(shrimp, package = "inlabru", envir = environment())
  ggplot() +
    fmesher::geom_fm(data = shrimp$mesh) +
    gg(shrimp$hauls, aes(col = catch)) +
    coord_sf(datum = fmesher::fm_crs(shrimp$hauls))
}
}
}
\references{
Pennino, M. G., Paradinas, I., Munoz, F., Illian, J.,Quilez-Lopez, A.,
Bellido, J.M., Conesa, D. Accounting for preferential sampling in species
distribution models. Ecology and Evolution, 9(1), p653-663, 2019
\doi{10.1002/ece3.4789}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_input.R, R/effect.R
\name{print.summary_bru_input}
\alias{print.summary_bru_input}
\alias{summary.bru_comp}
\alias{summary.bru_comp_list}
\alias{print.bru_comp}
\alias{print.bru_comp_list}
\alias{format.bru_subcomp}
\alias{summary.bru_subcomp}
\alias{print.bru_subcomp}
\alias{print.summary_component}
\alias{print.summary_bru_comp_list}
\alias{print.summary_bru_subcomp}
\title{Summarise components}
\usage{
\method{print}{summary_bru_input}(x, ...)

\method{summary}{bru_comp}(object, ..., depth = Inf, verbose = TRUE)

\method{summary}{bru_comp_list}(object, verbose = TRUE, ...)

\method{print}{bru_comp}(x, ...)

\method{print}{bru_comp_list}(x, ...)

\method{format}{bru_subcomp}(x, verbose = TRUE, ..., label.override = NULL)

\method{summary}{bru_subcomp}(object, verbose = TRUE, ..., label.override = NULL)

\method{print}{bru_subcomp}(x, verbose = TRUE, ..., label.override = NULL)

\method{print}{summary_component}(x, ...)

\method{print}{summary_bru_comp_list}(x, ...)

\method{print}{summary_bru_subcomp}(x, ...)
}
\arguments{
\item{x}{A summary object to be printed.}

\item{\dots}{Passed on to other summary methods.}

\item{object}{Object to be summarised.}

\item{depth}{The depth of which to expand the component mapper.
Default \code{Inf}, to traverse the entire mapper tree.}

\item{verbose}{logical; If \code{TRUE}, includes more details of the
component definitions. When \code{FALSE}, only show basic component
definition information.  Default \code{TRUE}.}

\item{label.override}{character; If not \code{NULL}, use this label instead of
the object's label.}
}
\description{
Summarise components
}
\seealso{
\code{\link[=bru_comp]{bru_comp()}}, \code{\link[=summary.bru_input]{summary.bru_input()}}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}

Finn Lindgren \email{finn.lindgren@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_input.R
\name{ibm_input}
\alias{ibm_input}
\alias{ibm_input_set}
\alias{ibm_input_new}
\alias{ibm_input_available}
\alias{ibm_input_get}
\alias{bm_autodetect}
\title{Interface between \code{bru_input} and \code{bru_mapper}}
\usage{
ibm_input_set(mapper, input)

ibm_input_new(mapper, ...)

ibm_input_available(mapper)

ibm_input_get(mapper)

bm_autodetect()
}
\arguments{
\item{mapper}{A \code{bru_mapper} object.}

\item{input}{A \code{bru_input} object, or \code{NULL} to remove any existing input.
Alternatively, an existing \code{bru_mapper} object with or without associated
input can be provided, in which case the input from that mapper
is copied.}

\item{\dots}{Passed on to \code{\link[=new_bru_input]{new_bru_input()}}.}
}
\description{
Associate \link{bru_input} objects with \link{bru_mapper} objects.
}
\section{Functions}{
\itemize{
\item \code{ibm_input_set()}: Add an existing \code{bru_input} to a \code{bru_mapper}.

\item \code{ibm_input_new()}: Create and add a \code{bru_input} to a \code{bru_mapper}.

\item \code{ibm_input_available()}: Check if a \code{bru_input} is associated with a
\code{bru_mapper}.

\item \code{ibm_input_get()}: Get the \code{bru_input} associated with a \code{bru_mapper}.

\item \code{bm_autodetect()}: Create a \code{bru_mapper} placeholder object of class
\code{bm_autodetect}. The main purpose of this class is to attach \link{bru_input}
information to it, which is later used to determine a suitable 'real'
mapper type.

}}
\examples{
(m <- bm_autodetect())
ibm_input_available(m)
(m <- ibm_input_new(m, cos(x)))
ibm_input_available(m)
ibm_input_set(bm_linear(), m)

}
\seealso{
\code{\link[=bru_input]{bru_input()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R
\name{ibm_eval}
\alias{ibm_eval}
\alias{ibm_eval.default}
\alias{ibm_eval.bm_taylor}
\alias{ibm_eval.bm_const}
\alias{ibm_eval.bm_shift}
\alias{ibm_eval.bm_scale}
\alias{ibm_eval.bm_aggregate}
\alias{ibm_eval.bm_logsumexp}
\alias{ibm_eval.bm_logitaverage}
\alias{ibm_eval.bm_marginal}
\alias{ibm_eval.bm_pipe}
\alias{ibm_eval.bm_multi}
\alias{ibm_eval.bm_reparam}
\alias{ibm_eval.bm_collect}
\alias{ibm_eval.bm_repeat}
\alias{ibm_eval.bm_sum}
\title{Evaluate a mapping}
\usage{
ibm_eval(mapper, input, state = NULL, ...)

\method{ibm_eval}{default}(mapper, input, state = NULL, ..., jacobian = NULL)

\method{ibm_eval}{bm_taylor}(mapper, input = NULL, state = NULL, ...)

\method{ibm_eval}{bm_const}(mapper, input, state = NULL, ...)

\method{ibm_eval}{bm_shift}(mapper, input, state = NULL, ...)

\method{ibm_eval}{bm_scale}(mapper, input, state = NULL, ...)

\method{ibm_eval}{bm_aggregate}(mapper, input, state = NULL, ...)

\method{ibm_eval}{bm_logsumexp}(mapper, input, state = NULL, log = TRUE, ...)

\method{ibm_eval}{bm_logitaverage}(mapper, input, state = NULL, logit = TRUE, ...)

\method{ibm_eval}{bm_marginal}(mapper, input, state = NULL, ..., reverse = FALSE)

\method{ibm_eval}{bm_pipe}(mapper, input, state = NULL, ...)

\method{ibm_eval}{bm_multi}(
  mapper,
  input,
  state = NULL,
  inla_f = FALSE,
  ...,
  jacobian = NULL,
  pre_A = deprecated()
)

\method{ibm_eval}{bm_reparam}(mapper, input, state = NULL, ..., jacobian = NULL)

\method{ibm_eval}{bm_collect}(
  mapper,
  input,
  state,
  inla_f = FALSE,
  multi = FALSE,
  ...,
  sub_lin = NULL
)

\method{ibm_eval}{bm_repeat}(mapper, input, state, multi = FALSE, ..., sub_lin = NULL)

\method{ibm_eval}{bm_sum}(mapper, input, state, multi = FALSE, ..., sub_lin = NULL)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{...}{Arguments passed on to other methods}

\item{jacobian}{For \code{ibm_eval()} methods, an optional pre-computed Jacobian, typically
supplied by internal methods that already have the Jacobian.}

\item{log}{logical; control \code{log} output. Default \code{TRUE}, see the
\code{ibm_eval()} details for \code{logsumexp} mappers.}

\item{logit}{logical; control \code{logit} output. Default \code{TRUE}, see the
\code{ibm_eval()} details for \code{logitaverage} mappers.}

\item{reverse}{logical; control \code{bm_marginal} evaluation. Default
\code{FALSE}. When \code{TRUE}, reverses the direction of the mapping, see details
for \code{marginal} mappers.}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{pre_A}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} in favour of \code{jacobian}.}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}

\item{sub_lin}{Internal, optional pre-computed sub-mapper information}
}
\description{
Implementations must return a vector of length \code{\link[=ibm_n_output]{ibm_n_output()}}.
The \code{input} contents must
be in a format accepted by \code{\link[=ibm_jacobian]{ibm_jacobian()}}
for the mapper.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_eval(default)}: Verifies that the mapper is linear
with \code{\link[=ibm_is_linear]{ibm_is_linear()}}, and then computes a linear mapping
as \code{ibm_jacobian(...) \%*\% state}.  When \code{state} is \code{NULL},
a zero vector of length \code{\link[=ibm_n_output]{ibm_n_output()}} is returned.

\item \code{ibm_eval(bm_taylor)}: Evaluates linearised
mapper information at the given \code{state}. The \code{input} argument is ignored,
so that the usual argument order
\code{ibm_eval(mapper, input, state)} syntax can be used, but also
\code{ibm_eval(mapper, state = state)}.  For a mapper with a named jacobian list,
the \code{state} argument must also be a named list.  If \code{state} is \code{NULL},
all-zero is assumed.

\item \code{ibm_eval(bm_const)}: Returns the input values, with NA replaced by 0.

\item \code{ibm_eval(bm_logsumexp)}: When \code{log} is \code{TRUE} (default), \code{ibm_eval()}
for \code{logsumexp} returns the log-sum-weight-exp value. If \code{FALSE}, the
\code{sum-weight-exp} value is returned.

\item \code{ibm_eval(bm_logitaverage)}: When \code{logit} is \code{TRUE} (default), \code{ibm_eval()}
for \code{logitaverage} returns the logit-sum-weight-inverse-logit value.
If \code{FALSE}, the \code{sum-weights=invere-logit} value is returned.

\item \code{ibm_eval(bm_marginal)}: When \code{xor(mapper[["inverse"]], reverse)} is
\code{FALSE}, \code{ibm_eval()}
for \code{marginal} returns \code{qfun(pnorm(x), param)}, evaluated in a numerically
stable way. Otherwise, evaluates the inverse \code{qnorm(pfun(x, param))} instead.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inla.R
\name{extract_property}
\alias{extract_property}
\title{Extract a summary property from all results of an inla result}
\usage{
extract_property(result, property, internal_hyperpar = FALSE)
}
\arguments{
\item{result}{an \code{inla} result object with a \code{bru_info} element}

\item{property}{character; "mean", "sd", "mode", or some other column
identifier for inla result \verb{$summary.fixed}, \verb{$summary.random$label}, and
\verb{$summary.hyperpar}, or "joint_mode". For "joint_mode", the joint latent mode
is extracted, and the joint hyperparameter mode, in the internal scale.
For "predictor_sd" the posterior standard deviations of the linear predictor
are returned.}

\item{internal_hyperpar}{logical; if \code{TRUE}, use internal scale for
hyperparamter properties. Default is \code{FALSE}, except when \code{property} is
"joint_mode" which forces \code{internal_hyperpar=TRUE}.}
}
\value{
named list for each estimated fixed effect coefficient,
random effect vector, and hyperparameter. The hyperparameter names are
standardised with \code{\link[=bru_standardise_names]{bru_standardise_names()}}
}
\description{
Extract a summary property from all results of an inla result
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spde.R
\name{spde.posterior}
\alias{spde.posterior}
\title{Posteriors of SPDE hyper parameters and Matern correlation or
covariance function.}
\usage{
spde.posterior(result, name, what = "range", quantile = 0.95)
}
\arguments{
\item{result}{An object inheriting from \code{inla}.}

\item{name}{Character stating the name of the SPDE effect, see
\code{names(result$summary.random)}.}

\item{what}{One of "range", "log.range", "variance", "log.variance",
"matern.correlation" or "matern.covariance".}

\item{quantile}{The target credible probability. Default 0.95.}
}
\value{
A \code{prediction} object.
}
\description{
Calculate posterior distribution of the range, log(range), variance, or
log(variance) parameter of a model's SPDE component. Can also plot Matern
correlation or covariance function. \code{inla.spde.result}.
}
\examples{
\donttest{
if (bru_safe_inla() && require(ggplot2, quietly = TRUE)) {
  # Load 1D Poisson process data

  data(Poisson2_1D, package = "inlabru")


  # Take a look at the point (and frequency) data

  ggplot(pts2) +
    geom_histogram(
      aes(x = x),
      binwidth = 55 / 20,
      boundary = 0,
      fill = NA,
      color = "black"
    ) +
    geom_point(aes(x), y = 0, pch = "|", cex = 4) +
    coord_fixed(ratio = 1)

  # Fit an LGCP model with  and SPDE component

  x <- seq(0, 55, length.out = 20)
  mesh1D <- fmesher::fm_mesh_1d(x, boundary = "free")
  constrained_matern <- INLA::inla.spde2.matern(mesh1D, constr = TRUE)
  mdl <- x ~ spde1D(x, model = constrained_matern) + Intercept(1)
  fit <- lgcp(mdl, data = pts2, domain = list(x = mesh1D))

  # Calculate and plot the posterior range

  range <- spde.posterior(fit, "spde1D", "range")
  plot(range)

  # Calculate and plot the posterior log range

  lrange <- spde.posterior(fit, "spde1D", "log.range")
  plot(lrange)

  # Calculate and plot the posterior variance

  variance <- spde.posterior(fit, "spde1D", "variance")
  plot(variance)

  # Calculate and plot the posterior log variance

  lvariance <- spde.posterior(fit, "spde1D", "log.variance")
  plot(lvariance)

  # Calculate and plot the posterior Matern correlation

  matcor <- spde.posterior(fit, "spde1D", "matern.correlation")
  plot(matcor)

  # Calculate and plot the posterior Matern covariance

  matcov <- spde.posterior(fit, "spde1D", "matern.covariance")
  plot(matcov)
}
}

}
\author{
Finn Lindgren \email{Finn.Lindgren@ed.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_marginal}
\alias{bm_marginal}
\alias{bru_mapper_marginal}
\title{Mapper for marginal distribution transformation}
\usage{
bm_marginal(qfun, pfun = NULL, dfun = NULL, dqfun = NULL, ..., inverse = FALSE)

bru_mapper_marginal(...)
}
\arguments{
\item{qfun}{A quantile function, supporting arguments \code{p}, \code{lower.tail}, and
\code{log.p}, like \code{\link[stats:Normal]{stats::qnorm()}}.}

\item{pfun}{A CDF, supporting arguments \code{q}, \code{lower.tail}, and \code{log.p},
like \code{\link[stats:Normal]{stats::pnorm()}}.  Only needed and used when
\code{xor(mapper[["inverse"]], reverse)} is \code{TRUE} in a method call.
Default \code{NULL}}

\item{dfun}{A pdf, supporting arguments \code{x} and \code{log},
like \code{\link[stats:Normal]{stats::dnorm()}}. If \code{NULL} (default), uses finite
differences on \code{qfun} or \code{pfun} instead.}

\item{dqfun}{A function evaluating the reciprocal of the derivative of
\code{qfun}, i.e. the density at a given CDF value. If \code{NULL} (default), uses
\code{dfun(qfun(...),...)} or finite differences on \code{qfun} or \code{pfun} instead.
Must support the same arguments as \code{qfun}.}

\item{\dots}{Arguments passed on to \code{\link[=bm_marginal]{bm_marginal()}}}

\item{inverse}{logical; If \code{FALSE} (default), \code{\link[=bm_marginal]{bm_marginal()}}
defines a mapping from standard Normal to a specified distribution.
If \code{TRUE}, it defines a mapping from the specified distribution to a standard
Normal.}
}
\description{
Constructs a mapper that transforms the marginal distribution \code{state} from
\eqn{\textrm{N}(0,1)}{N(0, 1)} to the distribution of a given (continuous)
quantile function. The \code{...} arguments are used as parameter arguments to
\code{qfun}, \code{pfun}, \code{dfun}, and \code{dqfun}.
}
\examples{
m <- bm_marginal(qexp, pexp, rate = 1 / 8)
(val <- ibm_eval(m, state = -5:5))
ibm_eval(m, state = val, reverse = TRUE)
m <- bm_marginal(qexp, pexp, dexp, rate = 1 / 8)
ibm_eval2(m, state = -3:3)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.RasterLayer}
\alias{gg.RasterLayer}
\title{Geom for RasterLayer objects}
\usage{
\method{gg}{RasterLayer}(
  data,
  mapping = ggplot2::aes(x = .data[["x"]], y = .data[["y"]], fill = .data[["layer"]]),
  ...
)
}
\arguments{
\item{data}{A RasterLayer object.}

\item{mapping}{aesthetic mappings created by \code{aes}. These are passed on to
\code{geom_tile}.}

\item{\dots}{Arguments passed on to \code{geom_tile}.}
}
\value{
An object returned by \code{geom_tile}
}
\description{
This function takes a RasterLayer object, converts it into a
\code{SpatialPixelsDataFrame} and uses \code{geom_tile} to plot the data.
}
\details{
This function requires the \code{raster} and \code{ggplot2} packages.
}
\examples{
\dontrun{
# Some features require the raster and spatstat.data packages.
if (require("spatstat.data", quietly = TRUE) &&
  require("raster", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE)) {
  # Load Gorilla data
  data("gorillas", package = "spatstat.data", envir = environment())

  # Convert elevation covariate to RasterLayer

  elev <- as(gorillas.extra$elevation, "RasterLayer")

  # Plot the elevation

  ggplot() +
    gg(elev)
}
}
}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.Poisson3_1D.R
\docType{data}
\name{Poisson3_1D}
\alias{Poisson3_1D}
\alias{lambda3_1D}
\alias{E_nc3a}
\alias{E_nc3b}
\alias{pts3}
\alias{countdata3a}
\alias{countdata3b}
\title{1-Dimensional NonHomogeneous Poisson example.}
\format{
The data contain the following \code{R} objects:
\describe{
\item{\code{lambda3_1D}}{A function defining the intensity function of a
nonhomogeneous Poisson process. Note that this function is only defined on
the interval (0,55).}
\item{\code{E_nc3a}}{The expected counts of gridded data for the wider bins
(10 bins).}
\item{\code{E_nc3b}}{The expected counts of gridded data for the wider bins
(20 bins).}
\item{\code{pts3}}{The locations of the observed points (a data frame with one
column, named \code{x}).}
\item{\code{countdata3a}}{A data frame with three columns, containing the
count data for the 10-interval case:}
\item{\code{countdata3b}}{A data frame with three columns, containing the
count data for the 20-interval case:}
\describe{
\item{\code{x}}{The grid cell midpoint.}
\item{\code{count}}{ The number of detections in the cell.}
\item{\code{exposure}}{ The width of the cell.}
}
}
}
\usage{
data(Poisson3_1D)
}
\description{
Point data and count data, together with intensity function and
expected counts for a multimodal nonhomogeneous 1-dimensional Poisson
process example. Counts are given for two different gridded data interval
widths.
}
\examples{
\donttest{
if (require("ggplot2", quietly = TRUE) &&
  require("patchwork", quietly = TRUE)) {
  data(Poisson3_1D)
  # first the plots for the 10-bin case:
  p1a <- ggplot(countdata3a) +
    geom_point(data = countdata3a, aes(x = x, y = count), col = "blue") +
    ylim(0, max(countdata3a$count, E_nc3a)) +
    geom_point(
      data = countdata3a, aes(x = x), y = 0, shape = "+",
      col = "blue", cex = 4
    ) +
    geom_point(
      data = data.frame(x = countdata3a$x, y = E_nc3a),
      aes(x = x), y = E_nc3a, shape = "_", cex = 5
    ) +
    xlab(expression(bold(s))) +
    ylab("count")
  ss <- seq(0, 55, length.out = 200)
  lambda <- lambda3_1D(ss)
  p2a <- ggplot() +
    geom_line(
      data = data.frame(x = ss, y = lambda), aes(x = x, y = y),
      col = "blue"
    ) +
    ylim(0, max(lambda)) +
    geom_point(data = pts3, aes(x = x), y = 0.2, shape = "|", cex = 4) +
    xlab(expression(bold(s))) +
    ylab(expression(lambda(bold(s))))
  (p1a / p2a)

  # Then the plots for the 20-bin case:
  p1a <- ggplot(countdata3b) +
    geom_point(data = countdata3b, aes(x = x, y = count), col = "blue") +
    ylim(0, max(countdata3b$count, E_nc3b)) +
    geom_point(
      data = countdata3b, aes(x = x), y = 0, shape = "+",
      col = "blue", cex = 4
    ) +
    geom_point(
      data = data.frame(x = countdata3b$x, y = E_nc3b),
      aes(x = x), y = E_nc3b, shape = "_", cex = 5
    ) +
    xlab(expression(bold(s))) +
    ylab("count")
  ss <- seq(0, 55, length.out = 200)
  lambda <- lambda3_1D(ss)
  p2a <- ggplot() +
    geom_line(
      data = data.frame(x = ss, y = lambda), aes(x = x, y = y),
      col = "blue"
    ) +
    ylim(0, max(lambda)) +
    geom_point(data = pts3, aes(x = x), y = 0.2, shape = "|", cex = 4) +
    xlab(expression(bold(s))) +
    ylab(expression(lambda(bold(s))))
  (p1a / p2a)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{summary.bru_options}
\alias{summary.bru_options}
\alias{print.summary_bru_options}
\title{Print inlabru options}
\usage{
\method{summary}{bru_options}(
  object,
  legend = TRUE,
  include_global = TRUE,
  include_default = TRUE,
  ...
)

\method{print}{summary_bru_options}(x, ...)
}
\arguments{
\item{object}{A \link{bru_options} object to be summarised}

\item{legend}{logical; If \code{TRUE}, include explanatory text, Default: \code{TRUE}}

\item{include_global}{logical; If \code{TRUE}, include global override options}

\item{include_default}{logical; If \code{TRUE}, include default options}

\item{\dots}{Further parameters, currently ignored}

\item{x}{A \code{summary_bru_options} object to be printed}
}
\description{
Print inlabru options
}
\examples{
if (interactive()) {
  options <- bru_options(verbose = TRUE)

  # Don't print options only set in default:
  print(options, include_default = FALSE)

  # Only include options set in the object:
  print(options, include_default = FALSE, include_global = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.mexdolphin.R
\docType{data}
\name{mexdolphin_sf}
\alias{mexdolphin_sf}
\alias{mexdolphin_sp}
\title{Pan-tropical spotted dolphins in the Gulf of Mexico}
\format{
A list of objects:
\describe{
\item{\code{points}:}{ An \code{sf} object containing the locations of
detected dolphin groups, with their size as an attribute.}
\item{\code{samplers}:}{ An \code{sf} object containing the transect lines
that were surveyed.}
\item{\code{mesh}:}{ An \code{fm_mesh_2d} object containing a Delaunay triangulation
mesh (a type of discretization of continuous space) covering the survey
region.}
\item{\code{ppoly}:}{ An \code{sf} object defining the boundary of the
survey region.}
\item{\code{simulated}:}{ A \code{sf} object containing the locations of a
\emph{simulated} population of dolphin groups. The population was simulated
from a \code{inlabru}
model fitted to the actual survey data. Note that the simulated data
do not have any associated
size information.}
}
}
\source{
Library \code{dsm}.
}
\usage{
mexdolphin_sf

mexdolphin_sp()
}
\description{
This a version of the \code{mexdolphins} dataset from the package
\code{dsm}, reformatted as point process data for use with \code{inlabru}, with the
parts stored in \code{sf} format. The data are from a combination of several
NOAA shipboard surveys conducted on pan-tropical spotted dolphins in the
Gulf of Mexico. 47 observations of groups of dolphins were detected. The
group size was recorded, as well as the Beaufort sea state at the time of
the observation. Transect width is 16 km, i.e. maximal detection distance 8
km (transect half-width 8 km).
}
\section{Functions}{
\itemize{
\item \code{mexdolphin_sp()}: Convert \code{mexdolphin_sf} to \code{sp} format. Replaces
the old \code{mexdolphin} dataset.

}}
\examples{
\donttest{
if (require("ggplot2", quietly = TRUE)) {
  data(mexdolphin_sf, package = "inlabru", envir = environment())
  ggplot() +
    gg(mexdolphin_sf$mesh) +
    gg(mexdolphin_sf$ppoly, color = "blue", alpha = 0, linewidth = 1) +
    gg(mexdolphin_sf$samplers) +
    gg(mexdolphin_sf$points, aes(size = size), color = "red") +
    scale_size_area()

  ggplot() +
    gg(mexdolphin_sf$mesh,
      color = mexdolphin_sf$lambda,
      mask = mexdolphin_sf$ppoly
    )
}
}
\donttest{
if (require("ggplot2", quietly = TRUE) &&
  require("sp", quietly = TRUE)) {
  mexdolphin <- mexdolphin_sp()
  ggplot() +
    gg(mexdolphin$mesh) +
    gg(mexdolphin$ppoly, color = "blue") +
    gg(mexdolphin$samplers) +
    gg(mexdolphin$points, aes(size = size), color = "red") +
    scale_size_area() +
    coord_equal()

  ggplot() +
    gg(mexdolphin$mesh,
      col = mexdolphin$lambda,
      mask = mexdolphin$ppoly
    ) +
    coord_equal()
}
}
}
\references{
Halpin, P.N., A.J. Read, E. Fujioka, B.D. Best, B. Donnelly, L.J. Hazen, C.
Kot, K. Urian, E. LaBrecque, A. Dimatteo, J. Cleary, C. Good, L.B. Crowder,
and K.D. Hyrenbach. 2009. OBIS-SEAMAP: The world data center for marine
mammal, sea bird, and sea turtle distributions. Oceanography 22(2):104-115

NOAA Southeast Fisheries Science Center. 1996. Report of a Cetacean Survey of
Oceanic and Selected Continental Shelf Waters of the Northern Gulf of Mexico
aboard NOAA Ship Oregon II (Cruise 220)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R, R/inla.R
\name{plot.bru}
\alias{plot.bru}
\alias{plotmarginal.inla}
\title{Plot method for posterior marginals estimated by bru}
\usage{
\method{plot}{bru}(x, ...)

plotmarginal.inla(
  result,
  varname = NULL,
  index = NULL,
  link = function(x) {
     x
 },
  add = FALSE,
  ggp = TRUE,
  lwd = 3,
  ...
)
}
\arguments{
\item{x}{a fitted \code{\link[=bru]{bru()}} model.}

\item{\dots}{Options passed on to other methods.}

\item{result}{an \code{inla} or \code{bru} result object}

\item{varname}{character; name of the variable to plot}

\item{index}{integer; index of the random effect to plot}

\item{link}{function; link function to apply to the variable}

\item{add}{logical; if \code{TRUE}, add to an existing plot}

\item{ggp}{logical; unused}

\item{lwd}{numeric; line width}
}
\description{
From version \verb{2.11.0}, \code{plot.bru(x, ...)} calls \code{plot.inla(x, ...)}
from the \code{INLA} package, unless the first argument after \code{x} is a
\code{character}, in which case the pre-\verb{2.11.0} behaviour is used, calling
\code{plotmarginal.inla(x, ...)} instead.

Requires the \code{ggplot2} package.
}
\examples{
\dontrun{
if (require("ggplot2", quietly = TRUE)) {
  # Generate some data and fit a simple model
  input.df <- data.frame(x = cos(1:10)) |>
    dplyr::mutate(
      y = 5 + 2 * x + rnorm(length(x), mean = 0, sd = 0.1)
    )
  fit <- bru(y ~ x, family = "gaussian", data = input.df)
  summary(fit)

  # Plot the posterior density of the model's x-effect
  plot(fit, "x")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_matrix}
\alias{bm_matrix}
\alias{bru_mapper_matrix}
\title{Mapper for matrix multiplication}
\usage{
bm_matrix(labels)

bru_mapper_matrix(...)
}
\arguments{
\item{labels}{Column labels for matrix mappings; Can be factor, character, or
a single integer specifying the number of columns for integer column
indexing.}

\item{\dots}{Arguments passed on to \code{\link[=bm_matrix]{bm_matrix()}}}
}
\description{
Create a matrix mapper, for a given number of columns
}
\examples{
m <- bm_matrix(labels = c("a", "b"))
ibm_values(m)
ibm_eval2(m, input = matrix(1:6, 3, 2), state = 2:3)

m <- bm_matrix(labels = 2L)
ibm_values(m)
ibm_eval2(m, input = matrix(1:6, 3, 2), state = 2:3)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.Poisson2_1D.R
\docType{data}
\name{Poisson2_1D}
\alias{Poisson2_1D}
\alias{lambda2_1D}
\alias{cov2_1D}
\alias{E_nc2}
\alias{pts2}
\alias{countdata2}
\title{1-Dimensional NonHomogeneous Poisson example.}
\format{
The data contain the following \code{R} objects:
\describe{
\item{\code{lambda2_1D}:}{ A function defining the intensity function of a
nonhomogeneous Poisson process. Note that this function is only defined on
the interval (0,55).}
\item{\code{cov2_1D}:}{ A function that gives what we will call a
'habitat suitability' covariate in 1D space.}
\item{\code{E_nc2}}{ The expected counts of the gridded data.}
\item{\code{pts2}}{ The locations of the observed points (a data frame with
one column, named \code{x}).}
\item{\code{countdata2}}{ A data frame with three columns, containing the
count data:}
\describe{
\item{\code{x}}{ The grid cell midpoint.}
\item{\code{count}}{ The number of detections in the cell.}
\item{\code{exposure}}{ The width of the cell.}
}
}
}
\usage{
data(Poisson2_1D)
}
\description{
Point data and count data, together with intensity function and
expected counts for a unimodal nonhomogeneous 1-dimensional Poisson process
example.
}
\examples{
\donttest{
if (require("ggplot2", quietly = TRUE) &&
  require("patchwork", quietly = TRUE)) {
  data(Poisson2_1D)
  p1 <- ggplot(countdata2) +
    geom_point(data = countdata2, aes(x = x, y = count), col = "blue") +
    ylim(0, max(countdata2$count, E_nc2)) +
    geom_point(
      data = countdata2, aes(x = x), y = 0, shape = "+",
      col = "blue", cex = 4
    ) +
    geom_point(
      data = data.frame(x = countdata2$x, y = E_nc2), aes(x = x),
      y = E_nc2, shape = "_", cex = 5
    ) +
    xlab(expression(bold(s))) +
    ylab("count")
  ss <- seq(0, 55, length.out = 200)
  lambda <- lambda2_1D(ss)
  p2 <- ggplot() +
    geom_line(
      data = data.frame(x = ss, y = lambda),
      aes(x = x, y = y), col = "blue"
    ) +
    ylim(0, max(lambda)) +
    geom_point(data = pts2, aes(x = x), y = 0.2, shape = "|", cex = 4) +
    xlab(expression(bold(s))) +
    ylab(expression(lambda(bold(s))))
  (p1 / p2)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R
\name{ibm_invalid_output}
\alias{ibm_invalid_output}
\alias{ibm_invalid_output.default}
\alias{ibm_invalid_output.bm_index}
\alias{ibm_invalid_output.bm_multi}
\alias{ibm_invalid_output.bm_collect}
\alias{ibm_invalid_output.bm_repeat}
\alias{ibm_invalid_output.bm_sum}
\title{Detect invalid input to a mapper}
\usage{
ibm_invalid_output(mapper, input, state, ...)

\method{ibm_invalid_output}{default}(mapper, input, state, ...)

\method{ibm_invalid_output}{bm_index}(mapper, input, state, ...)

\method{ibm_invalid_output}{bm_multi}(mapper, input, state, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_invalid_output}{bm_collect}(mapper, input, state, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_invalid_output}{bm_repeat}(mapper, input, state, ...)

\method{ibm_invalid_output}{bm_sum}(mapper, input, state, multi = FALSE, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{...}{Arguments passed on to other methods}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}
}
\description{
Implementations should return a logical vector of length
\code{ibm_n_output(mapper, input, state, ...)} indicating which, if any,
output elements of \code{ibm_eval(mapper, input, state, ...)} are known to be
invalid.
For for multi/collect mappers, a list, when given a \code{multi=TRUE} argument.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_invalid_output(default)}: Returns an all-\code{FALSE} logical vector.

\item \code{ibm_invalid_output(bm_index)}: Returns \code{TRUE} out-of-range input values.
Returns \code{FALSE} for in-range and \code{NA} input values (to support NA giving
zero effect).

\item \code{ibm_invalid_output(bm_multi)}: Accepts a list with
named entries, or a list with unnamed but ordered elements.
The names must match the sub-mappers, see \code{\link[=ibm_names.bm_multi]{ibm_names.bm_multi()}}.
Each list element should take a format accepted by the corresponding
sub-mapper. In case each element is a vector, the input can be given as a
data.frame with named columns, a matrix with named columns, or a matrix
with unnamed but ordered columns.

\item \code{ibm_invalid_output(bm_collect)}: Accepts a list with
named entries, or a list with unnamed but ordered elements.
The names must match the sub-mappers, see \code{\link[=ibm_names.bm_collect]{ibm_names.bm_collect()}}.
Each list element should take a format accepted by the corresponding
sub-mapper. In case each element is a vector, the input can be given as a
data.frame with named columns, a matrix with named columns, or a matrix
with unnamed but ordered columns.

\item \code{ibm_invalid_output(bm_repeat)}: Passes on the input to the corresponding method.

\item \code{ibm_invalid_output(bm_sum)}: Accepts a list with
named entries, or a list with unnamed but ordered elements.
The names must match the sub-mappers, see \code{\link[=ibm_names.bm_sum]{ibm_names.bm_sum()}}.
Each list element should take a format accepted by the corresponding
sub-mapper. In case each element is a vector, the input can be given as a
data.frame with named columns, a matrix with named columns, or a matrix
with unnamed but ordered columns.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/track_plotting.R
\name{bru_timings_plot}
\alias{bru_timings_plot}
\title{Plot inlabru iteration timings}
\usage{
bru_timings_plot(x)
}
\arguments{
\item{x}{a \link{bru} object, typically a result from \code{\link[=bru]{bru()}} for a nonlinear
predictor model}
}
\description{
Draws the time per iteration for preprocessing (including linearisation),
\code{inla()} calls, and
line search. Iteration \code{0} is the time used for defining the model structure.
}
\details{
Requires the "ggplot2" package to be installed.
}
\examples{
\dontrun{
fit <- bru(...)
bru_timings_plot(fit)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R
\name{add_mappers}
\alias{add_mappers}
\alias{add_mappers.bru_comp}
\alias{add_mappers.bru_comp_list}
\title{Add component input/latent mappers}
\usage{
add_mappers(...)

\method{add_mappers}{bru_comp}(component, lhoods, inputs = NULL, lh_data = NULL, ...)

\method{add_mappers}{bru_comp_list}(components, lhoods, inputs = NULL, lh_data = NULL, ...)
}
\arguments{
\item{\dots}{Parameters passed on to other methods}

\item{component}{A \link{component} object}

\item{lhoods}{A \link{bru_obs_list} object}

\item{inputs}{A precomputed list of inputs, as returned by
\code{\link[=bru_input.bru_comp_list]{bru_input.bru_comp_list()}}}

\item{lh_data}{A list of data object, one for each likelihood in \code{lhoods}
that is used to determine the mapper(s). If \code{NULL}, the data is
or inputs are used instead.}

\item{components}{A \code{bru_comp_list} object}
}
\value{
A \code{component} object with completed mapper information
}
\description{
Add missing mappers between input data and latent variables,
based on likelihood data

Equip component(s) with mappers for subcomponents that do not
have predefined mappers. When needed, the data in \code{lhoods} is used to
determine the appropriate mapper(s).
}
\examples{
\dontrun{
if (interactive() && bru_safe_inla()) {}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapper_repeat.R
\name{bm_repeat_indexing}
\alias{bm_repeat_indexing}
\alias{bm_repeat_indexing_matrix}
\title{Re-indexing matrix for repeated mappers}
\usage{
bm_repeat_indexing(n_map, n_rep, interleaved = FALSE)

bm_repeat_indexing_matrix(n_map, n_rep)
}
\arguments{
\item{n_map}{The block mapper size}

\item{n_rep}{The number of times the block is repeated}

\item{interleaved}{logical; if \code{TRUE}, the state vector indexing is
interleaved. Default is \code{FALSE}, for blockwise indexing.}
}
\value{
\code{bm_repeat_indexing}: Returns a list with a vector of offsets,
\code{offsets}, and a vector of relative index values, \code{index}; block \code{k}
indexing is given by \code{offsets[k] + index}.

\code{bm_repeat_indexing_matrix}: A \code{sparseMatrix} object.
}
\description{
Helper methods for regular and interleaved state vector
indexing, meant for use in \link{bm_repeat} mappers.
}
\section{Functions}{
\itemize{
\item \code{bm_repeat_indexing()}: Construct the offsets and within-block
index vectors for a repeated mapper, with or without interleaving.

\item \code{bm_repeat_indexing_matrix()}: Creates a sparse matrix \code{A} such that
\code{z <- A \%*\% x} constructs a blockwise version \code{z = c(x1, x2, ..., xn)} of
an interleaved state vector \code{x = c(x1[1], x2[1], ..., x1[2], x2[2], ...)}.
Each block is of size \code{n_map}, and there are \code{n_rep} blocks. The reverse
operation, taking a blockwise \code{z = c(x1, x2, ..., xn)} to an interleaved
vector is \code{x <- Matrix::t(A) \%*\% z}.

}}
\examples{
(idx <- bm_repeat_indexing(3, 2, FALSE))
(idx <- bm_repeat_indexing(3, 2, TRUE))
(A <- bm_repeat_indexing_matrix(3, 2))
(x_interleaved <- 1:6)
(x_blockwise <- as.vector(A \%*\% x_interleaved))
(x_recovered <- as.vector(Matrix::t(A) \%*\% x_blockwise))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{evaluate_predictor}
\alias{evaluate_predictor}
\title{Evaluate component effects or expressions}
\usage{
evaluate_predictor(
  model,
  state,
  data,
  data_extra,
  effects,
  predictor,
  used = NULL,
  format = "auto",
  n_pred = NULL
)
}
\arguments{
\item{state}{A list where each element is a list of named latent state
information, as produced by \code{\link[=evaluate_state]{evaluate_state()}}}

\item{data}{A \code{list}, \code{data.frame}, or \code{Spatial*DataFrame}, with coordinates
and covariates needed to evaluate the model.}

\item{data_extra}{Additional data for the predictor evaluation. Variables
with the same name as in \code{data} will be ignored, unless accessed via
\code{.data_extra.[["name"]]} or \code{.data_extra.$name}, or via pronouns;
see Details.}

\item{effects}{A list where each element is list of named evaluated effects,
each computed by \code{\link[=evaluate_effect_single_state.bru_comp_list]{evaluate_effect_single_state.bru_comp_list()}}}

\item{predictor}{Either a formula or \link{bru_pred_expr} expression}

\item{used}{A \code{\link[=bru_used]{bru_used()}} object, or NULL (default)}

\item{format}{character; determines the storage format of the output.
Available options:
\itemize{
\item \code{"auto"} If the first evaluated result is a vector or single-column matrix,
the "matrix" format is used, otherwise "list".
\item \code{"matrix"} A matrix where each column contains the evaluated predictor
expression for a state.
\item \code{"list"} A list where each column contains the evaluated predictor
expression for a state.
}

Default: "auto"}

\item{n_pred}{integer. If provided, scalar predictor results are expanded to
vectors of length \code{n_pred}.}
}
\value{
A list or matrix is returned, as specified by \code{format}
}
\description{
Evaluate component effects or expressions, based on a bru model and one or
several states of the latent variables and hyperparameters.
}
\details{
For each component, e.g. "name", the latent state values are
available as \code{name_latent}, and arbitrary evaluation can be done with
\code{name_eval(...)}, see \code{\link[=bru_comp_eval]{bru_comp_eval()}}.

The evaluation supports several \code{\link[rlang:as_data_mask]{rlang::as_data_pronoun()}} data masking
pronouns, to access variables from different data sources, and some of
these also have corresponding full objects, with an appended \code{.} in the
name. The full objects can be passed as arguments to functions.
\describe{
\item{.effect/.effect.}{refers to the \code{effects} vectors}
\item{.latent/.latent.}{refers to the latent state vectors}
\item{.data/.data.}{refers to the main \code{data} argument}
\item{.data_extra/.data_extra.}{refers to the \code{data_extra} argument}
\item{.env}{refers to the evaluation environment of the predictor}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_summarise}
\alias{bru_summarise}
\title{Summarise and annotate data}
\usage{
bru_summarise(
  data,
  probs = c(0.025, 0.5, 0.975),
  x = NULL,
  cbind.only = FALSE,
  max_moment = 2
)
}
\arguments{
\item{data}{A list of samples, each either numeric or a \code{data.frame}}

\item{probs}{A numeric vector of probabilities with values in \verb{[0, 1]},
passed to \code{stats::quantile}}

\item{x}{A \code{data.frame} of data columns that should be added to the summary
data frame}

\item{cbind.only}{If TRUE, only \code{cbind} the samples and return a matrix where
each column is a sample}

\item{max_moment}{integer, at least 2. Determines the largest moment
order information to include in the output. If \code{max_moment > 2},
includes "skew" (skewness, \code{E[(x-m)^3/s^3]}), and
if \code{max_moment > 3}, includes
"ekurtosis" (excess kurtosis, \code{E[(x-m)^4/s^4] - 3}). Default 2.
Note that the Monte Carlo variability of the \code{ekurtois} estimate may be
large.}
}
\value{
A \code{data.frame} or \verb{Spatial[Points/Pixels]DataFrame} with summary
statistics, "mean", "sd", \code{paste0("q", probs)}, "mean.mc_std_err",
"sd.mc_std_err"
}
\description{
Summarise and annotate data
}
\examples{
bru_summarise(matrix(rexp(10000), 10, 1000), max_moment = 4, probs = NULL)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotsample.R
\name{plotsample}
\alias{plotsample}
\title{Create a plot sample.}
\usage{
plotsample(spdf, boundary, x.ppn = 0.25, y.ppn = 0.25, nx = 5, ny = 5)
}
\arguments{
\item{spdf}{A \code{SpatialPointsDataFrame} defining the points that are to be
sampled by the plot sample.}

\item{boundary}{A \code{SpatialPolygonsDataFrame} defining the survey boundary
within which the  points occur.}

\item{x.ppn}{The proportion of the x=axis that is to be included in the
plots.}

\item{y.ppn}{The proportion of the y=axis that is to be included in the
plots.}

\item{nx}{The number of plots in the x-dimension.}

\item{ny}{The number of plots in the y-dimension.}
}
\value{
A list with three components:
\describe{
\item{\code{plots}}{A \code{SpatialPolygonsDataFrame} object containing the plots
that were sampled.}
\item{\code{dets}}{A \code{SpatialPointsDataFrame} object containing the locations
of the points within the plots.}
\item{\code{counts}}{A \code{dataframe} containing the following columns
\describe{
\item{\code{x}}{The x-coordinates of the centres of the plots within the
boundary.}
\item{\code{y}}{The y-coordinates of the centres of the plots within the
boundary.}
\item{\code{n}}{The numbers of points in each plot.}
\item{\code{area}}{The areas of the plots within the boundary}
}}
}.
}
\description{
Creates a plot sample on a regular grid with a random start location.
}
\examples{
\donttest{
# Some features require the raster package
if (bru_safe_sp() &&
  require("sp") &&
  require("raster", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE)) {
  gorillas <- gorillas_sp()
  plotpts <- plotsample(gorillas$nests, gorillas$boundary,
    x.ppn = 0.4, y.ppn = 0.4, nx = 5, ny = 5
  )
  ggplot() +
    gg(plotpts$plots) +
    gg(plotpts$dets, pch = "+", cex = 2) +
    gg(gorillas$boundary)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.gof.R
\name{devel.cvmeasure}
\alias{devel.cvmeasure}
\title{Variance and correlations measures for prediction components}
\usage{
devel.cvmeasure(joint, prediction1, prediction2, samplers = NULL, mesh = NULL)
}
\arguments{
\item{joint}{A joint \code{prediction} of two latent model components.}

\item{prediction1}{A \code{prediction} of the first component.}

\item{prediction2}{A \code{prediction} of the second component.}

\item{samplers}{An \code{sf} or \code{SpatialPolygon} object describing the area for
which to compute the cumulative variance measure.}

\item{mesh}{The \link[fmesher:fm_mesh_2d]{fmesher::fm_mesh_2d} for which the prediction was performed
(required for cumulative Vmeasure).}
}
\value{
Variance and correlations measures.
}
\description{
Calculates local and integrated variance and correlation measures as
introduced by Yuan et al. (2017).
}
\examples{
\donttest{
if (bru_safe_inla() &&
  require("ggplot2", quietly = TRUE) &&
  require("patchwork", quietly = TRUE) &&
  requireNamespace("sn", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE) &&
  require("RColorBrewer", quietly = TRUE) &&
  require("dplyr", quietly = TRUE)) {
  # Load Gorilla data

  gorillas <- gorillas_sf
  gorillas$gcov <- gorillas_sf_gcov()

  # Use RColorBrewer

  # Fit a model with two components:
  # 1) A spatial smooth SPDE
  # 2) A spatial covariate effect (vegetation)

  pcmatern <- INLA::inla.spde2.pcmatern(
    gorillas$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(0.01, 0.01)
  )

  cmp <- geometry ~ 0 +
    vegetation(gorillas$gcov$vegetation, model = "factor_contrast") +
    spde(geometry, model = pcmatern) +
    Intercept(1)

  fit <- lgcp(
    cmp,
    gorillas$nests,
    samplers = gorillas$boundary,
    domain = list(geometry = gorillas$mesh),
    options = list(control.inla = list(int.strategy = "eb"))
  )

  # Predict SPDE and vegetation at a grid covering the domain of interest
  pred_loc <- fmesher::fm_pixels(
    gorillas$mesh,
    mask = gorillas$boundary,
    dims = c(200, 200),
    format = "sf"
  )
  pred <- predict(
    fit,
    pred_loc,
    ~ list(
      joint = spde + vegetation,
      field = spde,
      veg = vegetation
    )
  )

  pred_collect <-
    rbind(
      pred$joint |> dplyr::mutate(component = "joint"),
      pred$field |> dplyr::mutate(component = "field"),
      pred$veg |> dplyr::mutate(component = "veg")
    ) |>
    dplyr::mutate(var = sd^2)

  # Plot component mean

  ggplot(pred_collect) +
    geom_tile(aes(geometry = geometry, fill = mean),
      stat = "sf_coordinates"
    ) +
    coord_equal() +
    facet_wrap(~component, nrow = 1) +
    theme(legend.position = "bottom")

  # Plot component variance

  ggplot(pred_collect) +
    geom_tile(aes(geometry = geometry, fill = var),
      stat = "sf_coordinates"
    ) +
    coord_equal() +
    facet_wrap(~component, nrow = 1) +
    theme(legend.position = "bottom")

  # Calculate variance and correlation measure

  vm <- devel.cvmeasure(pred$joint, pred$field, pred$veg)
  # Compute nominal relative variance contributions; note that these can be
  # greater than 100\%!
  vm <- dplyr::mutate(
    vm,
    var1_rel = if_else(var1 <= 0, NA, var1 / var.joint),
    var2_rel = if_else(var2 <= 0, NA, var2 / var.joint)
  )
  lprange <- range(vm$var.joint, vm$var1, vm$var2)
  vm <- tidyr::pivot_longer(vm,
    cols = c(var.joint, var1, var2, cov, cor, var1_rel, var2_rel),
    names_to = "component",
    values_to = "value"
  )

  # Variance contribution of the components

  csc <- scale_fill_gradientn(
    colours = brewer.pal(9, "YlOrRd"),
    limits = lprange
  )

  vm_ <- dplyr::filter(vm, component \%in\% c("var.joint", "var1", "var2"))
  ggplot(vm_) +
    geom_tile(aes(geometry = geometry, fill = value),
      stat = "sf_coordinates"
    ) +
    csc +
    coord_equal() +
    facet_wrap(~component, nrow = 1) +
    theme(legend.position = "bottom")

  # Relative variance contribution of the components
  # When bo

  vm_ <- dplyr::filter(vm, component \%in\% c("var1_rel", "var2_rel"))
  ggplot(vm_) +
    geom_tile(aes(geometry = geometry, fill = value),
      stat = "sf_coordinates"
    ) +
    scale_fill_gradientn(
      colours = brewer.pal(9, "YlOrRd"),
      trans = "log"
    ) +
    coord_equal() +
    facet_wrap(~component, nrow = 1)

  # Where both relative contributions are larger than 1, the posterior
  # correlations are strongly negative.

  # Covariance and correlation of field and vegetation

  vm_cov <- dplyr::filter(vm, component \%in\% "cov")
  vm_cor <- dplyr::filter(vm, component \%in\% "cor")
  (ggplot(vm_cov) +
    geom_tile(aes(geometry = geometry, fill = value),
      stat = "sf_coordinates"
    ) +
    scale_fill_gradientn(
      colours = rev(brewer.pal(9, "RdBu")),
      limits = c(-1, 1) * max(abs(vm_cov$value), na.rm = TRUE)
    ) +
    coord_equal() +
    theme(legend.position = "bottom") +
    ggtitle("Covariances") |
    ggplot(vm_cor) +
      geom_tile(aes(geometry = geometry, fill = value),
        stat = "sf_coordinates"
      ) +
      scale_fill_gradientn(
        colours = rev(brewer.pal(9, "RdBu")),
        limits = c(-1, 1) * max(abs(vm_cor$value), na.rm = TRUE)
      ) +
      coord_equal() +
      theme(legend.position = "bottom") +
      ggtitle("Correlations")
  )

  # Variance and correlation integrated over space

  vrt <- fmesher::fm_vertices(gorillas$mesh, format = "sf")
  pred_vrt <- predict(
    fit,
    vrt,
    ~ list(
      joint = spde + vegetation,
      field = spde,
      veg = vegetation
    )
  )

  vm.int <- devel.cvmeasure(
    pred_vrt$joint,
    pred_vrt$field,
    pred_vrt$veg,
    samplers = fmesher::fm_int(gorillas$mesh, gorillas$boundary),
    mesh = gorillas$mesh
  )
  vm.int
}
}

}
\references{
Y. Yuan, F. E. Bachl, F. Lindgren, D. L. Brochers, J. B. Illian, S. T.
Buckland, H. Rue, T. Gerrodette. 2017. Point process models for
spatio-temporal distance sampling data from a large-scale survey of blue
whales. \url{https://arxiv.org/abs/1604.06013}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{bru_model}
\alias{bru_model}
\alias{summary.bru_model}
\alias{print.summary_bru_model}
\alias{print.bru_model}
\title{Create an inlabru model object from model components}
\usage{
bru_model(components, lhoods, options = list(), .envir = parent.frame())

\method{summary}{bru_model}(object, ...)

\method{print}{summary_bru_model}(x, ...)

\method{print}{bru_model}(x, ...)
}
\arguments{
\item{components}{A \link{bru_comp_list} object}

\item{lhoods}{Either a \code{\link[=bru_obs]{bru_obs()}} object, a \code{\link[=bru_obs_list]{bru_obs_list()}} object, or
a list of one or more \code{\link[=bru_obs]{bru_obs()}} or \code{\link[=bru_obs_list]{bru_obs_list()}} objects,
or a list of arguments for a call to \code{\link[=bru_obs]{bru_obs()}}.}

\item{options}{A \link{bru_options} options object or a list of options passed
on to \code{\link[=bru_options]{bru_options()}}}

\item{.envir}{The environment in which the components are evaluated.}

\item{object}{Object to operate on}

\item{\dots}{Arguments passed on to other methods}

\item{x}{An object to be printed}
}
\value{
A \link{bru_model} object
}
\description{
The \link{inlabru} syntax for model formulae is different from what
\code{INLA::inla} considers a valid.
In inla most of the effects are defined by adding an \code{f(...)} expression to
the formula.
In \link{inlabru} the \code{f} is replaced by an arbitrary (exceptions: \code{const} and
\code{offset}) string that will determine the label of the effect. See Details for
further information.
}
\details{
For instance

\code{y ~ f(myspde, ...)}

in INLA is equivalent to

\code{y ~ myspde(...)}

in inlabru.

A disadvantage of the inla way is that there is no clear separation between
the name of the covariate and the label of the effect. Furthermore, for some
models like SPDE it is much more natural to use spatial coordinates as
covariates rather than an index into the SPDE vertices. For this purpose
\link{inlabru} provides the \code{main} argument. For convenience, the \code{main} argument
can be used like the first argument of the f function, e.g., and is the first
argument of the component definition.
The \code{INLA} model formula

\code{y ~ f(temperature, model = 'linear')}

is equivalent to the \code{inlabru} component and formula definition

\code{y ~ temperature(temperature, model = 'linear')}
and
\code{y ~ temperature(main = temperature, model = 'linear')}
as well as
\code{y ~ temperature(model = 'linear')}
which sets \code{main = temperature}.

On the other hand, \code{main} can also be a function mapping, e.g the
\code{sp::coordinates()} function:

\code{y ~ mySPDE(coordinates, ...)}

This extracts the coordinates from the data object, and maps it to the latent
field via the information given in the \code{mapper}, which by default is
extracted from the \code{model} object, in the case of \code{spde} model objects.

Morevover, \code{main} can be any expression that evaluates within your data as an
environment.
For instance, if your data has columns 'a' and 'b', you can create a fixed
effect of 'sin(a+b)' by setting \code{map} in the following way:

\code{y ~ myEffect(sin(a+b))}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R, R/bru_input.R, R/ggplot.R,
%   R/utils.R
\name{inlabru-deprecated}
\alias{inlabru-deprecated}
\alias{input_eval}
\alias{evaluate_inputs}
\alias{gmap}
\alias{gm}
\alias{row_kron}
\title{Deprecated functions in inlabru}
\usage{
input_eval(...)

evaluate_inputs(...)

gmap(...)

gm(...)

row_kron(M1, M2, repl = NULL, n.repl = NULL, weights = NULL)
}
\arguments{
\item{\dots}{Usually passed on to other methods}

\item{M1}{A matrix that can be transformed into a sparse Matrix.}

\item{M2}{A matrix that can be transformed into a sparse Matrix.}

\item{repl}{An optional index vector.  For each entry, specifies which
replicate the row belongs to, in the sense used in
\code{INLA::inla.spde.make.A}}

\item{n.repl}{The maximum replicate index, in the sense used in
\code{INLA::inla.spde.make.A()}.}

\item{weights}{Optional scaling weights to be applied row-wise to the
resulting matrix.}
}
\description{
Some of these functions may still attempt to do their job, but will be
removed in a future version.
}
\section{Functions}{
\itemize{
\item \code{input_eval()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} from
\verb{2.12.0.9023}.
Use \code{bru_input()} instead.

\item \code{evaluate_inputs()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} since
version \verb{2.12.0.9023}. Use \code{\link[=bru_input]{bru_input()}} instead. Computes the component
inputs for included components for each observation model.

\item \code{gmap()}: Plot a map using extent of a spatial object

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This function is deprecated as \code{ggmap}
isn't supported.

Used \code{ggmap::get_map()} to query map services like Google Maps for a region
centered around the spatial object provided. Then calls \code{ggmap()} to plot the
map.

This function required the \code{ggmap} package.

\item \code{gm()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This function is deprecated as \code{ggmap}
isn't supported.

ggplot geom for spatial data

gm is a wrapper for the \link{gg} method. It will take the first argument and
transform its coordinate system to latitude and longitude. Thereafter, \link{gg}
is called using the transformed data and the arguments provided via \code{...}. gm
is intended to replace gg whenever the data is supposed to be plotted over a
spatial map generated by \link{gmap}, which only works if the coordinate system is
latitude/longitude.

\item \code{row_kron()}: Row-wise Kronecker products
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} in favour of \code{\link[fmesher:fm_row_kron]{fmesher::fm_row_kron()}}.

Takes two Matrices and computes the row-wise Kronecker product.  Optionally
applies row-wise weights and/or applies an additional 0/1 row-wise Kronecker
matrix product.

Returns a \code{Matrix::sparseMatrix} object.

}}
\author{
Finn Lindgren \email{finn.lindgren@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/used.R
\name{bru_used_vars}
\alias{bru_used_vars}
\alias{bru_used_vars.character}
\alias{bru_used_vars.expression}
\alias{bru_used_vars.quosure}
\alias{bru_used_vars.formula}
\title{Extract basic variable names from expression}
\usage{
bru_used_vars(x, functions = FALSE)

\method{bru_used_vars}{character}(x, functions = FALSE)

\method{bru_used_vars}{expression}(x, functions = FALSE)

\method{bru_used_vars}{quosure}(x, functions = FALSE)

\method{bru_used_vars}{formula}(x, functions = FALSE)
}
\arguments{
\item{x}{A \code{formula}, \code{expression}, or \code{character}}

\item{functions}{logical; if TRUE, include function names}
}
\value{
If successful, a character vector, otherwise \code{NULL}
}
\description{
Extracts the variable names from an R expression by pre- and post-processing
around \code{\link[=all.vars]{all.vars()}}.
First replaces \code{$} with \code{[[} indexing, so that internal column/variable names
are ignored, then calls \code{all.vars()}.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_used_vars(formula)}: Only the right-hand side is used.

}}
\examples{
bru_used_vars(~.)
bru_used_vars(~ a + b + c_latent + d_eval())
bru_used_vars(expression(a + b + c_latent + d_eval()))

bru_used_vars(~., functions = TRUE)
bru_used_vars(~ a + b + c_latent + d_eval(), functions = TRUE)
bru_used_vars(expression(a + b + c_latent + d_eval()), functions = TRUE)

bru_used_vars(a ~ b)
bru_used_vars(expression(a ~ b))

}
\seealso{
Other bru_used: 
\code{\link{bru_used}()},
\code{\link{bru_used_update}()},
\code{\link{new_bru_used}()}
}
\concept{bru_used}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R
\name{ibm_n_output}
\alias{ibm_n_output}
\alias{ibm_n_output.default}
\alias{ibm_n_output.bm_taylor}
\alias{ibm_n_output.bm_shift}
\alias{ibm_n_output.bm_scale}
\alias{ibm_n_output.bm_aggregate}
\alias{ibm_n_output.bm_marginal}
\alias{ibm_n_output.bm_pipe}
\alias{ibm_n_output.bm_multi}
\alias{ibm_n_output.bm_reparam}
\alias{ibm_n_output.bm_collect}
\alias{ibm_n_output.bm_repeat}
\alias{ibm_n_output.bm_sum}
\title{Output size of a mapping}
\usage{
ibm_n_output(mapper, input, state = NULL, inla_f = FALSE, ...)

\method{ibm_n_output}{default}(mapper, input, state = NULL, inla_f = FALSE, ...)

\method{ibm_n_output}{bm_taylor}(mapper, input, ...)

\method{ibm_n_output}{bm_shift}(mapper, input, state = NULL, ..., n_state = NULL)

\method{ibm_n_output}{bm_scale}(mapper, input, state = NULL, ..., n_state = NULL)

\method{ibm_n_output}{bm_aggregate}(mapper, input = NULL, ...)

\method{ibm_n_output}{bm_marginal}(mapper, input, state = NULL, ..., n_state = NULL)

\method{ibm_n_output}{bm_pipe}(mapper, input, state = NULL, ..., n_state = NULL)

\method{ibm_n_output}{bm_multi}(mapper, input, ...)

\method{ibm_n_output}{bm_reparam}(mapper, ...)

\method{ibm_n_output}{bm_collect}(mapper, input, state = NULL, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_n_output}{bm_repeat}(mapper, ...)

\method{ibm_n_output}{bm_sum}(mapper, input, state = NULL, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{...}{Arguments passed on to other methods}

\item{n_state}{integer giving the length of the state vector for mappers
that have state dependent output size.}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}
}
\description{
Implementations must return an integer denoting the
mapper output length.
The default implementation returns \code{NROW(input)}.
Mappers such as \code{bm_multi} and \code{bm_collect},
that can accept \code{list()} inputs require their own method implementations.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_n_output(default)}: Returns \code{NROW(input)}

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.gorillas.R
\name{gorillas}
\alias{gorillas}
\title{Deprecated alias for sp version of the gorillas dataset}
\description{
Deprecated dataset name for the \code{sp} version
of \link{gorillas_sf}. Use \code{\link[=gorillas_sp]{gorillas_sp()}} instead.
}
\seealso{
\link{gorillas_sf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.data.frame}
\alias{gg.data.frame}
\alias{gg.bru_prediction}
\alias{gg.prediction}
\alias{plot.bru_prediction}
\alias{plot.prediction}
\title{Geom for predictions}
\usage{
\method{gg}{data.frame}(...)

\method{gg}{bru_prediction}(data, mapping = NULL, ribbon = TRUE, alpha = NULL, bar = FALSE, ...)

\method{gg}{prediction}(data, ...)

\method{plot}{bru_prediction}(x, y = NULL, ...)

\method{plot}{prediction}(x, y = NULL, ...)
}
\arguments{
\item{\dots}{Arguments passed on to \code{geom_line}.}

\item{data}{A prediction object, usually the result of a \code{\link[=predict.bru]{predict.bru()}}
call.}

\item{mapping}{a set of aesthetic mappings created by \code{aes}. These are passed
on to \code{geom_line()}.
If "fill" is present, it is passed to \code{geom_ribbon()}.}

\item{ribbon}{If TRUE, plot a ribbon around the line based on the smallest
and largest quantiles present in the data, found by matching names starting
with \code{q} and followed by a numerical value.  \code{inla()}-style
\code{numeric+"quant"} names are converted to inlabru style before matching.}

\item{alpha}{The ribbons numeric alpha (transparency) level in \verb{[0,1]}.}

\item{bar}{If TRUE plot boxplot-style summary for each variable.}

\item{x}{a prediction object.}

\item{y}{Ignored argument but required for S3 compatibility.}
}
\value{
Concatenation of a \code{geom_line} value and optionally a \code{geom_ribbon}
value.
}
\description{
This geom serves to visualize \code{prediction} objects which usually results from
a call to \code{\link[=predict.bru]{predict.bru()}}. Predictions objects provide summary statistics
(mean, median, sd, ...) for one or more random variables. For single
variables (or if requested so by setting \code{bar = TRUE}), a boxplot-style geom
is constructed to show the statistics. For multivariate predictions the mean
of each variable (y-axis) is plotted against the row number of the variable
in the prediction data frame (x-axis) using \code{geom_line}. In addition, a
\code{geom_ribbon} is used to show the confidence interval.

Note: \code{gg.bru_prediction} also understands the format of INLA-style posterior
summaries, e.g. \code{fit$summary.fixed} for an inla object \code{fit}

Requires the \code{ggplot2} package.
}
\section{Functions}{
\itemize{
\item \code{gg(data.frame)}: This geom constructor will simply call \code{\link[=gg.bru_prediction]{gg.bru_prediction()}} for the data
provided.

\item \code{plot(bru_prediction)}: Generates a base ggplot2 using \code{ggplot()} and adds a geom for input \code{x} using
\code{\link[=gg.bru_prediction]{gg.bru_prediction()}}. Returns a ggplot object.

\item \code{plot(prediction)}: Identical to \code{\link[=gg.bru_prediction]{gg.bru_prediction()}}.

}}
\examples{
\donttest{
if (bru_safe_inla() &&
  requireNamespace("sn", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE) &&
  require("patchwork", quietly = TRUE)) {
  # Generate some data

  input.df <- data.frame(x = cos(1:10))
  input.df <- within(
    input.df,
    {
      y <- 5 + 2 * cos(1:10) + rnorm(10, mean = 0, sd = 0.1)
    }
  )

  # Fit a model with fixed effect 'x' and intercept 'Intercept'

  fit <- bru(y ~ x, family = "gaussian", data = input.df)

  # Predict posterior statistics of 'x'

  xpost <- predict(fit, NULL, formula = ~x_latent)

  # The statistics include mean, standard deviation, the 2.5\% quantile, the
  # median, the 97.5\% quantile, minimum and maximum sample drawn from the
  # posterior as well as the coefficient of variation and the variance.

  xpost

  # For a single variable like 'x' the default plotting method invoked by
  # gg() will show these statistics in a fashion similar to a box plot:
  ggplot() +
    gg(xpost)


  # The predict function can also be used to simultaneously estimate
  # posteriors of multiple variables:

  xipost <- predict(fit,
    newdata = NULL,
    formula = ~ c(
      Intercept = Intercept_latent,
      x = x_latent
    )
  )
  xipost

  # If we still want a plot in the previous style we have to set the bar
  # parameter to TRUE

  p1 <- ggplot() +
    gg(xipost, bar = TRUE)
  p1

  # Note that gg also understands the posterior estimates generated while
  # running INLA

  p2 <- ggplot() +
    gg(fit$summary.fixed, bar = TRUE)
  (p1 / p2)

  # By default, if the prediction has more than one row, gg will plot the
  # column mean' against the row index. This is for instance useful for
  # predicting and plotting function but not very meaningful given the above
  # example:

  ggplot() +
    gg(xipost)

  # For ease of use we can also type

  plot(xipost)

  # This type of plot will show a ribbon around the mean, which visualizes
  # the upper and lower quantiles mentioned above (2.5\% and 97.5\%).
  # Plotting the ribbon can be turned of using the \code{ribbon} parameter

  ggplot() +
    gg(xipost, ribbon = FALSE)

  # Much like the other geomes produced by gg we can adjust the plot using
  # ggplot2 style commands, for instance

  ggplot() +
    gg(xipost) +
    gg(xipost, mapping = aes(y = median), ribbon = FALSE, color = "red")
}
}

}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_fmesher}
\alias{bm_fmesher}
\alias{bru_mapper_fmesher}
\alias{bru_mapper.fm_mesh_2d}
\title{Mapper for general \code{fmesher} function space objects}
\usage{
bm_fmesher(mesh)

bru_mapper_fmesher(...)

\method{bru_mapper}{fm_mesh_2d}(mesh, ...)
}
\arguments{
\item{mesh}{An \code{fmesher} object to map, supported by
\link[fmesher:fm_basis]{fmesher::fm_basis}\verb{(mesh, input)} and
\link[fmesher:fm_dof]{fmesher::fm_dof}\code{(mesh)}.}

\item{\dots}{Arguments passed on to \code{\link[=bm_fmesher]{bm_fmesher()}}}
}
\value{
A \code{bm_fmesher} object.
}
\description{
Creates a mapper for general \code{fmesher} function space objects.
}
\details{
Handles indexed mapping for all \code{fmesher} classes that support
\code{fm_dof()} and \code{fm_basis()} methods. For non-indexed mapping of
\code{fm_mesh_1d} objects, use \code{bru_mapper(mesh, indexed = FALSE)} which
invokes the \code{\link[=bru_mapper.fm_mesh_1d]{bru_mapper.fm_mesh_1d()}} method.
}
\section{Functions}{
\itemize{
\item \code{bru_mapper(fm_mesh_2d)}: Equivalent to calling \code{\link[=bm_fmesher]{bm_fmesher()}}.
Note: Prior to version \verb{2.12.0.9021},
this returned a \code{bru_mapper_fm_mesh_2d} object. Also see the note for
\code{\link[=bru_mapper.fm_mesh_1d]{bru_mapper.fm_mesh_1d()}}.

}}
\examples{
m <- bm_fmesher(fmesher::fmexample$mesh)
ibm_n(m)
ibm_eval(m, as.matrix(expand.grid(-2:2, -2:2)), seq_len(ibm_n(m)))

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deltaIC.R
\name{lgcp_IC}
\alias{lgcp_IC}
\alias{delta_IC}
\title{Information criteria for point processes}
\usage{
lgcp_IC(fit, data, predictor, domain, samplers, n.samples = 2000L)

delta_IC(data)
}
\arguments{
\item{fit}{A fitted \code{bru} model object, from \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}}.}

\item{data}{A data frame containing the observed points, with columns for
coordinates and any covariates used in the model. The data frame may also
include a column named \code{.block} that indicates the block membership of each
point, if blockwise criteria are to be calculated.}

\item{predictor}{An expression that evaluates to the linear predictor of the
model, given the data. This should be an expression that can be evaluated
in the context of the data frame and model components, just as in the
\code{formula} argument of \code{\link[=bru_obs]{bru_obs()}} and \code{\link[=lgcp]{lgcp()}}.}

\item{domain, samplers}{Arguments defining the integration domain(s) and
sampling regions, as for \code{\link[=bru_obs]{bru_obs()}} and \code{\link[=lgcp]{lgcp()}}.}

\item{n.samples}{Integer. The number of posterior samples to draw for
estimating the criteria. Default is 2000.}
}
\value{
A \code{tibble} with columns for the \code{Criterion} name, \code{Method},
log pointwise predictive density (\code{lppd}), effective number of parameters
(\code{p_eff}), and the information criterion value (\code{IC})
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Calculate information criteria for point process models, using a fitted
model and optionally new data. The interface is still under development
and may change without warning in future versions.
}
\section{Functions}{
\itemize{
\item \code{delta_IC()}: Calculate IC differences based on joined output from
several \code{lgcp_IC} calls;

\if{html}{\out{<div class="sourceCode">}}\preformatted{delta_IC(rbind(cbind(lgcp_IC(...), Model = "A"),
               cbind(lgcp_IC(...), Model = "B")))
}\if{html}{\out{</div>}}

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_obs}
\alias{bru_obs}
\alias{like}
\alias{bru_obs_list}
\alias{bru_obs_list.list}
\alias{bru_obs_list.bru_obs}
\alias{bru_obs_list.bru_obs_list}
\alias{c.bru_obs}
\alias{c.bru_obs_list}
\alias{[.bru_obs_list}
\alias{like_list}
\alias{bru_like_list}
\title{Observation model construction for usage with \code{\link[=bru]{bru()}}}
\usage{
bru_obs(
  formula = . ~ .,
  family = "gaussian",
  data = NULL,
  response_data = NULL,
  data_extra = NULL,
  E = NULL,
  Ntrials = NULL,
  weights = NULL,
  scale = NULL,
  domain = NULL,
  samplers = NULL,
  ips = NULL,
  used = NULL,
  is_rowwise = NULL,
  aggregate = NULL,
  aggregate_input = NULL,
  control.family = NULL,
  control.gcpo = NULL,
  tag = NULL,
  options = list(),
  .envir = parent.frame(),
  include = deprecated(),
  exclude = deprecated(),
  include_latent = deprecated(),
  allow_combine = deprecated()
)

like(
  ...,
  E = NULL,
  Ntrials = NULL,
  weights = NULL,
  scale = NULL,
  options = list(),
  .envir = parent.frame()
)

bru_obs_list(..., .tag = NULL)

\method{bru_obs_list}{list}(object, ..., .tag = NULL)

\method{bru_obs_list}{bru_obs}(..., .tag = NULL)

\method{bru_obs_list}{bru_obs_list}(..., .tag = NULL)

\method{c}{bru_obs}(...)

\method{c}{bru_obs_list}(...)

\method{[}{bru_obs_list}(x, i)

like_list(...)

bru_like_list(...)
}
\arguments{
\item{formula}{a \code{formula} where the right hand side is a general R
expression defines the predictor used in the model.}

\item{family}{A string identifying a valid \code{INLA::inla} likelihood family.
The default is
\code{gaussian} with identity link. In addition to the likelihoods provided
by inla (see \code{names(INLA::inla.models()$likelihood)})
inlabru supports fitting latent Gaussian Cox
processes via \code{family = "cp"}.
As an alternative to \code{\link[=bru]{bru()}}, the \code{\link[=lgcp]{lgcp()}} function provides
a convenient interface to fitting Cox processes.}

\item{data}{Predictor expression-specific data, as a \code{data.frame}, \code{tibble},
or \code{sf}.  Since \verb{2.12.0.9023}, deprecated support for
\code{SpatialPoints[DataFrame]} objects.}

\item{response_data}{Observation/response-specific data for models that need
different size/format for inputs and response variables, as a \code{data.frame},
\code{tibble}, or \code{sf}. Since \verb{2.12.0.9023}, deprecated support for
\code{SpatialPoints[DataFrame]} objects.}

\item{data_extra}{object convertible with \code{as.list()} with additional
variables to be made available in predictor evaluations. Variables with
the same names as the data object will be ignored, unless accessed via
\code{.data_extra.[["name"]]} or \code{.data_extra.$name} in the formula.}

\item{E}{Exposure/effort parameter for family = 'poisson' passed on to
\code{INLA::inla}. Special case if family is 'cp': rescale all integration
weights by a scalar \code{E}. For sampler specific reweighting/effort, use a
\code{weight} column in the \code{samplers} object instead, see \code{\link[fmesher:fm_int]{fmesher::fm_int()}}.
Default taken from \code{options$E}, normally \code{1}.}

\item{Ntrials}{A vector containing the number of trials for the 'binomial'
likelihood. Default taken from \code{options$Ntrials}, normally \code{1}.}

\item{weights}{Fixed (optional) weights parameters of the likelihood, so the
log-likelihood\verb{[i]} is changed into \code{weights[i] * log_likelihood[i]}.
Default value is \code{1}. WARNING: The normalizing constant for the likelihood
is NOT recomputed, so ALL marginals (and the marginal likelihood) must be
interpreted with great care.

For \code{family = "cp"}, the weights are applied as \code{sum(weights * eta)} in
the point location contribution part of the log-likelihood, where \code{eta} is
the linear predictor, and do not affect the integration part of the
likelihood. This can be used to implement approximative methods for point
location uncertainty.}

\item{scale}{Fixed (optional) scale parameters of the precision for several
models, such as Gaussian and student-t response models.}

\item{domain, samplers, ips}{Arguments used for \code{family="cp"} and \verb{aggregate=}.
\describe{
\item{\code{domain}}{Named list of domain definitions, see \code{\link[fmesher:fm_int]{fmesher::fm_int()}}.}
\item{\code{samplers}}{Integration domain for \code{family="cp"} or subdomains for
\verb{aggregate=}, see \code{\link[fmesher:fm_int]{fmesher::fm_int()}}.}
\item{\code{ips}}{Integration points. Defaults
to \link[fmesher:fm_int]{fmesher::fm_int}\verb{(domain, samplers)}. If explicitly given,
overrides \code{domain} and \code{samplers}. \code{fmesher::new_fm_int()} (from fmesher
\verb{0.5.0.9013}) can be used for manually constructed integration schemes.}
}}

\item{used}{Either \code{NULL} (default) or a \code{\link[=bru_used]{bru_used()}} object. When,
\code{NULL}, the information about what effects and
latent vectors are made available to the predictor evaluation is defined by
\code{bru_used(formula)}, which will include all effects and latent vectors
used by the predictor expression.}

\item{is_rowwise, allow_combine}{logical; If \code{is_rowwise} is \code{FALSE}, the
predictor expression may involve several rows of the input data to
influence the same row. When \code{NULL}, it defaults to \code{TRUE}, unless
\code{response_data} is non-\code{NULL}, or \code{data} is a \code{list}, or the likelihood
construction requires it. The \code{allow_combine} argument is a deprecated and
inverted version, such that \code{is_rowwise = !allow_combine}.}

\item{aggregate}{character ("none" or a valid name for the \code{type} argument of
\code{\link[=bm_aggregate]{bm_aggregate()}}) or an aggregation \code{bru_mapper} object (\code{\link[=bm_aggregate]{bm_aggregate()}},
\code{\link[=bm_logsumexp]{bm_logsumexp()}}, or \code{\link[=bm_logitaverage]{bm_logitaverage()}}). Default \code{NULL}, interpreted as
"none". \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}, available from version
\verb{2.12.0.9013}.}

\item{aggregate_input}{\code{NULL} or an optional input list to the mapper
defined by non-NULL \code{aggregate}, overriding the default,

\if{html}{\out{<div class="sourceCode">}}\preformatted{list(block = .data.[[".block"]],
     weights = .data.[["weight"]],
     n_block = bru_response_size(.response_data.),
     block_response = ".block")
}\if{html}{\out{</div>}}

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}, available from version \verb{2.12.0.9013}.

From \verb{2.13.0.9016}, it will look for a \code{block_response} element in the
list, which should be the name of a response variable in \code{response_data} to
match the \code{block} information against, allowing character or factor
aggregation block information to be used by replacing \code{block} with
\code{match(block, block_response)}. If not supplied, the name
\code{".block"} is tried. If that isn't available, the \code{block} information must
be supplied directly as a numeric or integer vector, indexing into the rows
of the response variable. Having no \code{block_response} variable is equivalent
to having
\code{response_data$.block = seq_len(bru_response_size(response_data))}.}

\item{control.family}{A optional \code{list} of \code{INLA::control.family} options}

\item{control.gcpo}{A optional \code{list} of \code{INLA::control.gcpo} options}

\item{tag}{character; Name that can be used to identify the relevant parts
of INLA predictor vector output, via \code{\link[=bru_index]{bru_index()}}.}

\item{options}{A \link{bru_options} options object or a list of options passed
on to \code{\link[=bru_options]{bru_options()}}}

\item{.envir}{The evaluation environment to use for special arguments (\code{E},
\code{Ntrials}, \code{weights}, and \code{scale}) if not found in \code{response_data} or
\code{data}. Defaults to the calling environment.}

\item{include, exclude, include_latent}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}, use
\code{used} instead.}

\item{\dots}{For \code{bru_obs_list.bru_obs}, one or more \code{bru_obs} objects}

\item{.tag}{Optional name to assign to a single \code{bru_obs} object. Reserved
for internal use.}

\item{object}{A list of \code{bru_obs} and/or \code{bru_obs_list} objects}

\item{x}{\code{bru_obs_list} object from which to extract element(s)}

\item{i}{indices specifying elements to extract}
}
\value{
A likelihood configuration which can be used to parameterise \code{\link[=bru]{bru()}}.
}
\description{
Observation model construction for usage with \code{\link[=bru]{bru()}}.

Note: Prior to version \verb{2.12.0}, this function was called \code{like()}, and that
alias will remain for a while until examples etc have been updated and users
made aware of the change. The name change is to avoid issues with namespace
clashes, e.g. with \code{data.table::like()}, and also to signal that the function
defines observation models, not just likelihood functions.
}
\details{
The \code{E}, \code{Ntrials}, \code{weights}, and \code{scale} arguments are evaluated in the
data context, with values from \code{response_data} taking precedence over \code{data}.
}
\section{Methods (by generic)}{
\itemize{
\item \code{bru_obs_list(bru_obs)}: Combine one or more lists of \code{bru_obs} observation model objects
into a \code{bru_obs_list} object

\item \code{c(bru_obs)}: Combine several \code{bru_obs} objects into a \code{bru_obs_list} object

}}
\section{Functions}{
\itemize{
\item \code{like()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Legacy \code{like()}
method for \code{inlabru} prior to version \verb{2.12.0}. Use \code{\link[=bru_obs]{bru_obs()}} instead.

\item \code{bru_obs_list()}: Combine \code{bru_obs} observation model object into a \code{bru_obs_list} object

\item \code{bru_obs_list(list)}: Combine one or more lists of \code{bru_obs} observation model objects
into a \code{bru_obs_list} object

\item \code{bru_obs_list(bru_obs_list)}: Combine one or more \code{bru_obs_list} objects into a \code{bru_obs_list} object

\item \code{c(bru_obs_list)}: Combine several \code{bru_obs_list} objects into a \code{bru_obs_list} object

\item \code{like_list()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Backwards compatibility for versions \verb{<= 2.12.0}. For later versions, use
\code{as_bru_obs_list()}, \code{bru_obs_list()}, or \code{c()}.

\item \code{bru_like_list()}: \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Backwards compatibility for versions \verb{<= 2.12.0.9017}. For later versions,
use \code{as_bru_obs_list()}, \code{bru_obs_list()} or \code{c()}.

}}
\examples{
\donttest{
if (bru_safe_inla() &&
  require(ggplot2, quietly = TRUE) &&
  require(patchwork, quietly = TRUE)) {
  # The 'bru_obs()' (previously 'like()') function's main purpose is to set
  # up observation models, both for single- and multi-likelihood models.
  # The following example generates some random covariates which are observed
  # through two different random effect models with different likelihoods

  # Generate the data

  set.seed(123)

  n1 <- 200
  n2 <- 10

  x1 <- runif(n1)
  x2 <- runif(n2)
  z2 <- runif(n2)

  y1 <- rnorm(n1, mean = 2 * x1 + 3)
  y2 <- rpois(n2, lambda = exp(2 * x2 + z2 + 3))

  df1 <- data.frame(y = y1, x = x1)
  df2 <- data.frame(y = y2, x = x2, z = z2)

  # Single likelihood models and inference using bru are done via

  cmp1 <- y ~ -1 + Intercept(1) + x
  fit1 <- bru(cmp1, family = "gaussian", data = df1)
  summary(fit1)

  cmp2 <- y ~ -1 + Intercept(1) + x + z
  fit2 <- bru(cmp2, family = "poisson", data = df2)
  summary(fit2)

  # A joint model has two likelihoods, which are set up using the bru_obs
  # function

  lik1 <- bru_obs(
    "gaussian",
    formula = y ~ x + Intercept,
    data = df1,
    tag = "norm"
  )
  lik2 <- bru_obs(
    "poisson",
    formula = y ~ x + z + Intercept,
    data = df2,
    tag = "pois"
  )

  # The union of effects of both models gives the components needed to run
  # bru

  jcmp <- ~ x + z + Intercept(1)
  jfit <- bru(jcmp, lik1, lik2)

  bru_index(jfit, "norm")
  bru_index(jfit, "pois")

  # Compare the estimates

  p1 <- ggplot() +
    gg(fit1$summary.fixed, bar = TRUE) +
    ylim(0, 4) +
    ggtitle("Model 1")
  p2 <- ggplot() +
    gg(fit2$summary.fixed, bar = TRUE) +
    ylim(0, 4) +
    ggtitle("Model 2")
  pj <- ggplot() +
    gg(jfit$summary.fixed, bar = TRUE) +
    ylim(0, 4) +
    ggtitle("Joint model")

  (p1 / p2 / pj)
}
}

}
\seealso{
\code{\link[=bru_response_size]{bru_response_size()}}, \code{\link[=bru_used]{bru_used()}}, \code{\link[=bru_comp]{bru_comp()}},
\code{\link[=bru_comp_eval]{bru_comp_eval()}}

\code{\link[=summary.bru_obs]{summary.bru_obs()}}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}

Finn Lindgren \email{finn.lindgren@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.robins_subset.R
\docType{data}
\name{robins_subset}
\alias{robins_subset}
\title{robins_subset}
\format{
The data are a data.frame with variables
\describe{
\item{\code{circle}:}{
Four-letter code of the CBC circle.
}
\item{\code{bcr}:}{
Numeric code for the bird conservation region encompassing the
count circle.
}
\item{\code{state}:}{
US state encompassing the count circle.
}
\item{\code{year}:}{
calendar year the count was conducted.
}
\item{\code{std_yr}:}{
transformed year, with 2016 = 0.
}
\item{\code{count}:}{
number of robins recorded.
}
\item{\code{log_hrs}:}{
the natural log of party hours.
}
\item{\code{lon}:}{
longitude of the count circle centroid.
}
\item{\code{lat}:}{
latitude of the count circle centroid.
}
\item{\code{obs}:}{
unique record identifier.
}
}
}
\source{
https://github.com/tmeeha/inlaSVCBC
}
\usage{
robins_subset
}
\description{
This is the \code{robins_subset} dataset, which is a subset of the
full robins data set used to demonstrate a spatially varying trend
coefficient model in Meehan et al. 2019. The dataset includes American
Robin counts, along with time, location, and effort information, from
Audubon Christimas Bird Counts (CBC) conducted in six US states between
1987 and 2016.
}
\examples{
if (require(ggplot2, quietly = TRUE)) {
  data(robins_subset, package = "inlabru") # get the data

  # plot the counts for one year of data
  ggplot(robins_subset[robins_subset$std_yr == 0, ]) +
    geom_point(aes(lon, lat, colour = count + 1)) +
    scale_colour_gradient(low = "blue", high = "red", trans = "log")
}
}
\references{
Meehan, T.D., Michel, N.L., and Rue, H. 2019. Spatial modeling of Audubon
Christmas Bird Counts reveals fine-scale patterns and drivers of relative
abundance trends. Ecosphere, 10(4), p.e02707.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inla.R
\name{bru_names}
\alias{bru_names}
\alias{bru_names.inla}
\alias{bru_names.bru}
\title{Extract standardised names from a bru or inla result object}
\usage{
bru_names(x)

\method{bru_names}{inla}(x)

\method{bru_names}{bru}(x)
}
\arguments{
\item{x}{an \code{inla} or \code{\link[=bru]{bru()}} result object}
}
\value{
A character vector with standardised names
}
\description{
Extracts the names of fixed effects, random effects, and hyperparameters,
converted with \code{\link[=bru_standardise_names]{bru_standardise_names()}}
}
\examples{
if (bru_safe_inla()) {
  fit <- bru(y ~ 1 + x + z(z, model = "iid"),
    data = data.frame(
      y = rnorm(10),
      x = rnorm(10),
      z = rep(seq_len(2), 5)
    )
  )
  bru_names(fit)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_call_options}
\alias{bru_call_options}
\title{Additional bru options}
\usage{
bru_call_options(...)
}
\arguments{
\item{\dots}{Options passed on to \code{\link[=as.bru_options]{as.bru_options()}}}
}
\value{
A \code{bru_options} object
}
\description{
Construct a \code{bru_options} object including the default and global options,
and converting deprecated option names.
}
\examples{
\donttest{

opts <- bru_call_options()

# Print them:
opts
}

}
\author{
Finn Lindgren \email{finn.lindgren@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_linear}
\alias{bm_linear}
\alias{bru_mapper_linear}
\title{Mapper for a linear effect}
\usage{
bm_linear()

bru_mapper_linear()
}
\description{
Create a mapper for linear effects
}
\examples{
m <- bm_linear()
ibm_eval(m, input = 1:4, state = 2)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inla.R
\name{bru_inla.stack.mjoin}
\alias{bru_inla.stack.mjoin}
\title{Join stacks intended to be run with different likelihoods}
\usage{
bru_inla.stack.mjoin(
  ...,
  compress = TRUE,
  remove.unused = TRUE,
  old.names = "BRU.response",
  new.name = "BRU.response"
)
}
\arguments{
\item{\dots}{List of stacks that contain vector observations (existing
multi-likelihood observation matrices are also permitted)}

\item{compress}{If \code{TRUE}, compress the model by removing duplicated rows of
effects, replacing the corresponding A-matrix columns with a single column
containing the sum.}

\item{remove.unused}{If \code{TRUE}, compress the model by removing rows of
effects corresponding to all-zero columns in the A matrix (and removing
those columns).}

\item{old.names}{A vector of strings with the names of the observation
vector/matrix for each stack. If a single string, this is assumed for all
the stacks. (default "BRU.response")}

\item{new.name}{The name to be used for the expanded observation matrix,
possibly the same as an old name. (default "BRU.response")}
}
\description{
Helper functions for multi-likelihood models
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_factor}
\alias{bm_factor}
\alias{bru_mapper_factor}
\title{Mapper for factor variables}
\usage{
bm_factor(values, factor_mapping, indexed = FALSE)

bru_mapper_factor(...)
}
\arguments{
\item{values}{Input values calculated by \code{\link[=bru_input.bru_input]{bru_input.bru_input()}}}

\item{factor_mapping}{character; selects the type of factor mapping.
\itemize{
\item \code{'contrast'} for leaving out the first factor level.
\item \code{'full'} for keeping all levels.
}}

\item{indexed}{logical; if \code{TRUE}, the \code{ibm_values()} method
will return an integer vector instead of the factor levels.
This is needed e.g. for \code{group} and \code{replicate} mappers, since
\code{INLA::f()} doesn't accept factor values. Default: \code{FALSE}, which
works for the main input mappers. The default mapper constructions
will set it the required setting.}

\item{\dots}{Arguments passed on to \code{\link[=bm_factor]{bm_factor()}}}
}
\description{
Create a factor mapper
}
\examples{
m <- bm_factor(factor(c("a", "b")), "full")
ibm_eval2(m, input = c("b", "a", "a", "b"), state = c(1, 3))

m <- bm_factor(factor(c("a", "b")), "contrast")
ibm_eval2(m, input = factor(c("b", "a", "a", "b")), state = 2)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/track_plotting.R
\name{bru_convergence_plot}
\alias{bru_convergence_plot}
\title{Plot inlabru convergence diagnostics}
\usage{
bru_convergence_plot(x, from = 1, to = NULL, type = NULL)
}
\arguments{
\item{x}{a \link{bru} object, typically a result from \code{\link[=bru]{bru()}} for a nonlinear
predictor model}

\item{from, to}{integer values for the range of iterations to plot.
Default \code{from = 1} (start from the first iteration) and \code{to = NULL} (end at
the last iteration).
Set \code{from = 0} to include the initial linearisation point in the track plot.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} character; "bru" (default)
for iterative nonlinear inlabru convergence diagnostics plots, or "inla"
for INLA optimiser trace plots.}
}
\value{
A ggplot object with four panels of convergence diagnostics:
\itemize{
\item \code{Tracks}: Mode and linearisation values for each effect
\item \code{Mode - Lin}: Difference between mode and linearisation values for each
effect
\item \verb{|Change| / sd}: Absolute change in mode and linearisation values
divided by the standard deviation for each effect
\item \code{Change & sd}: Absolute change in mode and linearisation values
and standard deviation for each effect
}

For multidimensional components, only the overall average, maximum, and
minimum values are shown.
}
\description{
Draws four panels of convergence diagnostics for an iterated INLA method
estimation
}
\details{
Requires the "dplyr", "ggplot2", and "patchwork"
packages to be installed.
}
\examples{
\dontrun{
fit <- bru(...)
bru_convergence_plot(fit)
}
}
\seealso{
\code{\link[=bru]{bru()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack.R
\name{bru_make_stack}
\alias{bru_make_stack}
\alias{bru_make_stack.bru_obs}
\alias{bru_make_stack.bru_obs_list}
\title{Build an inla data stack from linearisation information}
\usage{
bru_make_stack(...)

\method{bru_make_stack}{bru_obs}(lhood, lin, idx, ..., family_index = 1L)

\method{bru_make_stack}{bru_obs_list}(lhoods, lin, idx, ...)
}
\arguments{
\item{\dots}{Arguments passed on to other methods}

\item{lhood}{A \code{bru_obs} object}

\item{lin}{Linearisation information
\itemize{
\item For \code{.bru_obs}, a \code{bm_taylor} object
\item For \code{.bru_obs_list}, a list of \code{bm_taylor} objects
}}

\item{idx}{Output from \code{\link[=bru_index.bru_model]{bru_index.bru_model()}}}

\item{family_index}{integer specifying the family sequence index of the
observation model}

\item{lhoods}{A \code{bru_obs_list} object}
}
\description{
Combine linearisation for multiple likelihoods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg}
\alias{gg}
\title{ggplot2 geomes for inlabru related objects}
\usage{
gg(data, ...)
}
\arguments{
\item{data}{an object for which to generate a geom.}

\item{\dots}{Arguments passed on to the geom method.}
}
\value{
The form of the value returned by gg depends on the class of its
argument. See the documentation of the particular methods for details of
what is produced by that method.
}
\description{
gg is a generic function for generating geomes from various kinds of spatial
objects, e.g. Spatial* data, meshes, Raster objects and inla/inlabru
predictions. The function invokes particular methods which depend on the
\link{class} of the first argument.
}
\examples{
if (require("ggplot2", quietly = TRUE)) {
  # Load Gorilla data

  gorillas <- inlabru::gorillas_sf

  # Invoke ggplot and add geomes for the Gorilla nests and the survey
  # boundary

  ggplot() +
    gg(gorillas$boundary) +
    gg(gorillas$nests)
}
}
\seealso{
Other geomes: 
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_1d}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inla.R
\name{bru_inla.stack.mexpand}
\alias{bru_inla.stack.mexpand}
\title{Backwards compatibility to handle mexpand for INLA <= 24.06.02}
\usage{
bru_inla.stack.mexpand(
  ...,
  old.names = "BRU.response",
  new.name = "BRU.response"
)
}
\arguments{
\item{\dots}{List of stacks that contain vector observations (existing
multilikelihood observation matrices are also permitted)}

\item{old.names}{A vector of strings with the names of the observation
vector/matrix for each stack. If a single string, this is assumed for all
the stacks. (default "BRU.response")}

\item{new.name}{The name to be used for the expanded observation matrix,
possibly the same as an old name. (default "BRU.response")}
}
\value{
a list of modified stacks with multicolumn observations
}
\description{
Expand observation vectors/matrices in stacks into to a multicolumn matrix
for multiple likelihoods
}
\author{
Fabian E. Bachl \email{f.e.bachl@bath.ac.uk} and Finn Lindgren
\email{finn.lindgren@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.gorillas.R
\docType{data}
\name{gorillas_sf}
\alias{gorillas_sf}
\alias{gorillas_sf_gcov}
\alias{gorillas_sp}
\title{Gorilla nesting sites in sf format}
\format{
The data are a list that contains these elements:
\describe{
\item{\code{nests}:}{ An \code{sf} object containing the locations of
the gorilla nests.}
\item{\code{boundary}:}{ An \code{sf} object defining the boundary
of the region that was searched for the nests.}
\item{\code{mesh}:}{ An \code{fm_mesh_2d} object containing a mesh that can be used
with function \code{lgcp} to fit a LGCP to the nest data.}
\item{\code{gcov_file}:}{ The in-package filename of a \code{terra::SpatRaster}
object, with one layer for each of these spatial covariates:
\describe{
\item{\code{aspect}}{ Compass direction of the terrain slope. Categorical,
with levels
N, NE, E, SE, S, SW, W and NW, which are coded as integers 1 to 8.}
\item{\code{elevation}}{ Digital elevation of terrain, in metres.}
\item{\code{heat}}{ Heat Load Index at each point on the surface
(Beer's aspect),
discretised. Categorical with values Warmest (Beer's aspect between
0 and 0.999),
Moderate (Beer's aspect between 1 and 1.999), Coolest (Beer's aspect
equals 2). These are
coded as integers 1, 2 and 3, in that order.}
\item{\code{slopangle}}{ Terrain slope, in degrees.}
\item{\code{slopetype}}{ Type of slope. Categorical, with values Valley,
Toe (toe slope),
Flat, Midslope, Upper and Ridge. These are coded as integers 1 to 6.}
\item{\code{vegetation}}{ Vegetation type: a categorical variable with 6
levels coded as
integers 1 to 6 (in order of increasing expected habitat suitability)}
\item{\code{waterdist}}{ Euclidean distance from nearest water body, in
metres.}
}
Loading of the covariates can be done with \code{gorillas_sf_gcov()} or

\if{html}{\out{<div class="sourceCode">}}\preformatted{gorillas_sf$gcov <- terra::rast(
  system.file(gorillas_sf$gcov_file, package = "inlabru")
)
}\if{html}{\out{</div>}}

}
\item{\code{plotsample}}{Plot sample of gorilla nests, sampling 9x9 over the
region, with 60\\% coverage. Components:
\describe{
\item{\code{counts}}{An \code{sf} object with elements
\code{count}, \code{exposure}, and \code{geometry}, holding the point geometry for the
centre of each plot, the count in each
plot and the area of each plot.}
\item{\code{plots}}{An \code{sf} object with \code{MULTIPOLYGON} objects defining the
individual plot boundaries and an all-ones \code{weight} column.}
\item{\code{nests}}{An \code{sf} giving the locations of
each detected nests, \code{group} ("minor" or "major"),
\code{season} ("dry" or "rainy"), and \code{date} (in \code{Date} format).}
}
}
}
}
\source{
Library \code{spatstat.data}.
}
\usage{
gorillas_sf
data(gorillas_sf, package = "inlabru")

gorillas_sf_gcov()

gorillas_sp()
}
\description{
This is the \code{gorillas} dataset from the package \code{spatstat.data},
reformatted as point process data for use with \code{inlabru}.
}
\section{Functions}{
\itemize{
\item \code{gorillas_sf_gcov()}: Access the \code{gorillas_sf} covariates data as a
\code{terra::rast()} object.

\item \code{gorillas_sp()}: Access the \code{gorillas_sf} data in \code{sp} format.
The covariate data is added as \code{gcov}, a list of \code{sp::SpatialPixelsDataFrame}
objects. Requires the \code{sp}, \code{sf}, and \code{terra} packages to be installed.

}}
\examples{
if (interactive() &&
  require(ggplot2, quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  requireNamespace("tidyterra", quietly = TRUE)) {
  # plot all the nests, mesh and boundary
  ggplot() +
    gg(gorillas_sf$mesh) +
    geom_sf(
      data = gorillas_sf$boundary,
      alpha = 0.1, fill = "blue"
    ) +
    geom_sf(data = gorillas_sf$nests)

  # Plot the elevation covariate
  gorillas_sf$gcov <- gorillas_sf_gcov()
  ggplot() +
    tidyterra::geom_spatraster(data = gorillas_sf$gcov$elevation)

  # Plot the plot sample
  ggplot() +
    geom_sf(data = gorillas_sf$plotsample$plots) +
    geom_sf(data = gorillas_sf$plotsample$nests)
}
if (interactive() &&
  bru_safe_terra(quietly = TRUE)) {
  gorillas_sf$gcov <- gorillas_sf_gcov()
}
}
\references{
Funwi-Gabga, N. (2008) A pastoralist survey and fire impact assessment in the
Kagwene Gorilla Sanctuary, Cameroon. M.Sc. thesis, Geology and Environmental
Science, University of Buea, Cameroon.

Funwi-Gabga, N. and Mateu, J. (2012) Understanding the nesting spatial
behaviour of gorillas in the Kagwene Sanctuary, Cameroon. Stochastic
Environmental Research and Risk Assessment 26 (6), 793-811.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_info}
\alias{bru_info}
\alias{bru_info.character}
\alias{bru_info.bru}
\alias{as_bru_info}
\alias{as_bru_info.bru_info}
\alias{as_bru_info.bru}
\alias{summary.bru_info}
\alias{print.summary_bru_info}
\alias{print.bru_info}
\title{Methods for bru_info objects}
\usage{
bru_info(...)

\method{bru_info}{character}(method, ..., inlabru_version = NULL, INLA_version = NULL)

\method{bru_info}{bru}(object, ...)

as_bru_info(object, ...)

\method{as_bru_info}{bru_info}(object, ...)

\method{as_bru_info}{bru}(object, ...)

\method{summary}{bru_info}(object, verbose = TRUE, ...)

\method{print}{summary_bru_info}(x, ...)

\method{print}{bru_info}(x, ...)
}
\arguments{
\item{\dots}{Additional arguments to be stored in the \code{bru_info} object. For
\code{summary} and \code{print} methods, arguments passed on to submethods.}

\item{method}{character; The type of estimation method used}

\item{inlabru_version}{character; inlabru package version. Default: NULL, for
automatically detecting the version}

\item{INLA_version}{character; INLA package version. Default: NULL, for
automatically detecting the version}

\item{object}{A \code{bru_info} object}

\item{verbose}{logical; If \code{TRUE}, include more details of the
component definitions. If \code{FALSE}, only show basic component
definition information. Default: \code{FALSE}}

\item{x}{An  object to be printed}
}
\description{
The \code{bru_info} class is used to store metadata about \code{bru} models.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_info(character)}: Create a \code{bru_info} object

\item \code{bru_info(bru)}: Extract the \code{bru_info} object from an estimated \code{\link[=bru]{bru()}}
result object. The default print method shows information about model
components and observation models.

}}
\section{Methods (by generic)}{
\itemize{
\item \code{as_bru_info(bru_info)}: Extract a \code{bru_info} object.

}}
\section{Functions}{
\itemize{
\item \code{as_bru_info()}: Extract the \code{bru_info} object from an estimated \code{\link[=bru]{bru()}}
result object. The default print method shows information about model
components and observation models.

\item \code{as_bru_info(bru)}: Extract the \code{bru_info} object from an estimated \code{\link[=bru]{bru()}}
result object.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rgl.R
\name{glplot}
\alias{glplot}
\alias{glplot.SpatialPoints}
\alias{glplot.SpatialLines}
\alias{glplot.fm_mesh_2d}
\alias{globe}
\title{Render objects using RGL}
\usage{
glplot(object, ...)

\method{glplot}{SpatialPoints}(object, add = TRUE, color = "red", ...)

\method{glplot}{SpatialLines}(object, add = TRUE, ...)

\method{glplot}{fm_mesh_2d}(object, add = TRUE, col = NULL, ...)

globe(
  R = 1,
  R.grid = 1.05,
  specular = "black",
  axes = FALSE,
  box = FALSE,
  xlab = "",
  ylab = "",
  zlab = ""
)
}
\arguments{
\item{object}{an object used to select a method.}

\item{\dots}{Parameters passed on to \code{plot_rgl.fm_mesh_2d()}}

\item{add}{If TRUE, add the points to an existing plot. If FALSE, create new
plot.}

\item{color}{vector of R color characters. See material3d() for details.}

\item{col}{Color specification. A single named color, a vector of scalar
values, or a matrix of RGB values.}

\item{R}{Radius of the globe}

\item{R.grid}{Radius of the annotation sphere.}

\item{specular}{Light color of specular effect.}

\item{axes}{If TRUE, plot x, y and z axes.}

\item{box}{If TRUE, plot a box around the globe.}

\item{xlab, ylab, zlab}{Axes labels}
}
\description{
\code{glplot()} is a generic function for renders various kinds of spatial
objects, i.e. \verb{Spatial*} data and \code{fm_mesh_2d} objects. The function invokes
particular methods which depend on the class of the first argument.
}
\section{Methods (by class)}{
\itemize{
\item \code{glplot(SpatialPoints)}: This function will calculate the cartesian coordinates of
the points provided and use points3d() in order to render them.

\item \code{glplot(SpatialLines)}: This function will calculate a cartesian representation of
the lines provided and use \code{lines3d()} in order to render them.

\item \code{glplot(fm_mesh_2d)}: This function transforms the mesh to 3D cartesian
coordinates and uses \code{\link[fmesher:plot_rgl]{fmesher::plot_rgl()}} to plot the result.

}}
\section{Functions}{
\itemize{
\item \code{globe()}: Visualize a globe using RGL

Creates a textured sphere and lon/lat coordinate annotations.
This function requires the \code{rgl} and \code{sphereplot} packages.

}}
\examples{
\donttest{
if (interactive() &&
  require("rgl", quietly = TRUE) &&
  require("sphereplot", quietly = TRUE) &&
  bru_safe_sp() &&
  require("sp")) {
  # Show the globe:
  globe()

  # Load pantropoical dolphin data
  mexdolphin <- inlabru::mexdolphin_sp()

  # Add mesh, ship transects and dolphin sightings stored
  # as fm_mesh_2d, SpatialLines and SpatialPoints objects, respectively

  glplot(mexdolphin$mesh, alpha = 0.2)
  glplot(mexdolphin$samplers, lwd = 5)
  glplot(mexdolphin$points, size = 10)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.gof.R
\name{bincount}
\alias{bincount}
\title{1D LGCP bin count simulation and comparison with data}
\usage{
bincount(
  result,
  predictor,
  observations,
  breaks,
  nint = 20,
  probs = c(0.025, 0.5, 0.975),
  ...
)
}
\arguments{
\item{result}{A result object from a \code{\link[=bru]{bru()}} or \code{\link[=lgcp]{lgcp()}} call}

\item{predictor}{A formula describing the prediction of a 1D LGCP via
\code{\link[=predict]{predict()}}.}

\item{observations}{A vector of observed values}

\item{breaks}{A vector of bin boundaries}

\item{nint}{Number of integration intervals per bin. Increase this if the
bins are wide and the LGCP is not smooth.}

\item{probs}{numeric vector of probabilities with values in \verb{[0,1]}}

\item{\dots}{arguments passed on to \code{\link[=predict.bru]{predict.bru()}}}
}
\value{
An \code{data.frame} with a ggplot attribute \code{ggp}
}
\description{
A common procedure of analyzing the distribution of 1D points is to chose a
binning and plot the data's histogram with respect to this binning. This
function compares the counts that the histogram calculates to simulations
from a 1D log Gaussian Cox process conditioned on the number of data samples.
For each bin this results in a median number of counts as well as a
confidence interval. If the LGCP is a plausible model for the observed points
then most of the histogram counts (number of points within a bin) should be
within the confidence intervals. Note that a proper comparison  is a multiple
testing problem which the function does not solve for you.
}
\examples{
\dontrun{
if (require(ggplot2) && require(fmesher) && bru_safe_inla()) {
  # Load a point pattern
  data(Poisson2_1D)

  # Take a look at the point (and frequency) data

  ggplot(pts2) +
    geom_histogram(
      aes(x = x),
      binwidth = 55 / 20,
      boundary = 0,
      fill = NA,
      color = "black"
    ) +
    geom_point(aes(x), y = 0, pch = "|", cex = 4) +
    coord_fixed(ratio = 1)

  # Fit an LGCP model
  x <- seq(0, 55, length.out = 50)
  mesh1D <- fmesher::fm_mesh_1d(x, boundary = "free")
  matern <- INLA::inla.spde2.pcmatern(mesh1D,
    prior.range = c(1, 0.01),
    prior.sigma = c(1, 0.01),
    constr = TRUE
  )
  mdl <- x ~ spde1D(x, model = matern) + Intercept(1)
  fit.spde <- lgcp(mdl, pts2, domain = list(x = mesh1D))

  # Calculate bin statistics
  bc <- bincount(
    result = fit.spde,
    observations = pts2,
    breaks = seq(0, max(pts2), length.out = 12),
    predictor = x ~ exp(spde1D + Intercept)
  )

  # Plot them!
  attributes(bc)$ggp
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{iinla}
\alias{iinla}
\title{Iterated INLA}
\usage{
iinla(model, lhoods, inputs = NULL, initial = NULL, options)
}
\arguments{
\item{model}{A \link{bru_model} object}

\item{lhoods}{A list of likelihood objects from \code{\link[=bru_obs]{bru_obs()}}}

\item{inputs}{Optional pre-computed  list of per-likelihood component
evaluations, from \code{\link[=bru_input.bru_obs_list]{bru_input.bru_obs_list()}}.}

\item{initial}{A previous \code{bru} result or a list of named latent variable
initial states (missing elements are set to zero), to be used as starting
point, or \code{NULL}. If non-null, overrides \code{options$bru_initial}}

\item{options}{A \code{bru_options} object.}
}
\value{
An \code{iinla} object that inherits from \code{INLA::inla}, with an
added field \code{bru_iinla} with elements
\describe{
\item{log}{The diagnostic log messages produced by the run}
\item{states}{The list of linearisation points, one for each inla run}
\item{inla_stack}{The \code{inla.stack} object from the final inla run}
\item{track}{A list of convergence tracking vectors}
}
If an inla run is aborted by an error, the returned object also contains
an element \code{error} with the error object.
}
\description{
This is an internal wrapper for iterated runs of \code{INLA::inla}.
For nonlinear models, a linearisation is done with
\code{bru_compute_linearisation}, with a line search method between each
iteration. The \code{INLA::inla.stack} information is setup by \code{\link[=bru_make_stack]{bru_make_stack()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{ibm_is_rowwise}
\alias{ibm_is_rowwise}
\alias{ibm_is_rowwise.default}
\title{Check if a mapper is rowwise}
\usage{
ibm_is_rowwise(mapper, ...)

\method{ibm_is_rowwise}{default}(mapper, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{...}{Arguments passed on to other methods}
}
\description{
Implementations must return \code{TRUE} or \code{FALSE}.
If \code{TRUE} (returned by the default method unless the mapper
contains an \code{is_rowwise} variable), users of the mapper
may assume the mapper uses its inputs in "rowwise" manner, so that
blockwise evaluation is always possible.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_is_rowwise(default)}: Returns logical
\code{is_rowwise} from the mapper object if it exists, and otherwise \code{TRUE}.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotsample.R
\name{point2count}
\alias{point2count}
\title{Convert a plot sample of points into one of counts.}
\usage{
point2count(plots, dets)
}
\arguments{
\item{plots}{A \code{SpatialPolygonsDataFrame} object containing the plots that
were sampled.}

\item{dets}{A \code{SpatialPointsDataFrame} object containing the locations of the
points within the plots.}
}
\value{
A \code{SpatialPolygonsDataFrame} with counts in each plot contained in
slot \verb{@data$n}.
}
\description{
Converts a plot sample with locations of each point within each plot, into a
plot sample with only the count within each plot.
}
\examples{
\donttest{
# Some features require the raster package
if (bru_safe_sp() &&
  require("sp") &&
  require("raster", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE) &&
  bru_safe_terra(quietly = TRUE) &&
  require("sf", quietly = TRUE) &&
  require("patchwork", quietly = TRUE)) {
  gorillas <- gorillas_sp()
  plotpts <- plotsample(gorillas$nests, gorillas$boundary,
    x.ppn = 0.4, y.ppn = 0.4, nx = 5, ny = 5
  )
  p1 <- ggplot() +
    gg(plotpts$plots) +
    gg(plotpts$dets) +
    gg(gorillas$boundary)
  countdata <- point2count(plotpts$plots, plotpts$dets)
  x <- sp::coordinates(countdata)[, 1]
  y <- sp::coordinates(countdata)[, 2]
  count <- countdata@data$n
  p2 <- ggplot() +
    gg(gorillas$boundary) +
    gg(plotpts$plots) +
    geom_text(aes(label = count, x = x, y = y))
  (p1 | p2)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru_is.R
\name{bru_is_rowwise}
\alias{bru_is_rowwise}
\alias{bru_is_rowwise.bru_pred_expr}
\alias{bru_is_rowwise.bru_obs}
\alias{bru_is_rowwise.bru_obs_list}
\alias{bru_is_rowwise.bru_comp_list}
\alias{bru_is_rowwise.bru_comp}
\alias{bru_is_rowwise.bru_mapper}
\title{Check for predictor rowwise evaluability}
\usage{
bru_is_rowwise(x, ...)

\method{bru_is_rowwise}{bru_pred_expr}(x, ...)

\method{bru_is_rowwise}{bru_obs}(x, ...)

\method{bru_is_rowwise}{bru_obs_list}(x, ...)

\method{bru_is_rowwise}{bru_comp_list}(x, ...)

\method{bru_is_rowwise}{bru_comp}(x, ...)

\method{bru_is_rowwise}{bru_mapper}(x, ...)
}
\arguments{
\item{x}{An object containing a predictor definition}

\item{\dots}{Arguments passed on to submethods.}
}
\value{
\code{TRUE} if the expression is believed to be rowwise, \code{FALSE}
otherwise.
}
\description{
Checks if a predictor expression may be evaluated rowwise
}
\examples{
bru_is_rowwise(new_bru_pred_expr(~ x + y))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{lgcp}
\alias{lgcp}
\title{Log Gaussian Cox process (LGCP) inference using INLA}
\usage{
lgcp(
  components,
  data,
  domain = NULL,
  samplers = NULL,
  ips = NULL,
  formula = . ~ .,
  E = NULL,
  weights = NULL,
  ...,
  options = list(),
  .envir = parent.frame()
)
}
\arguments{
\item{components}{Latent component definitions, either as a \code{\link[=bru_comp_list]{bru_comp_list()}}
object, or a \code{formula}-like specification.
Also used to define a default linear additive predictor.  See
\code{\link[=bru_comp]{bru_comp()}} for details.}

\item{data}{Predictor expression-specific data, as a \code{data.frame}, \code{tibble},
or \code{sf}.  Since \verb{2.12.0.9023}, deprecated support for
\code{SpatialPoints[DataFrame]} objects.}

\item{domain, samplers, ips}{Arguments used for \code{family="cp"} and \verb{aggregate=}.
\describe{
\item{\code{domain}}{Named list of domain definitions, see \code{\link[fmesher:fm_int]{fmesher::fm_int()}}.}
\item{\code{samplers}}{Integration domain for \code{family="cp"} or subdomains for
\verb{aggregate=}, see \code{\link[fmesher:fm_int]{fmesher::fm_int()}}.}
\item{\code{ips}}{Integration points. Defaults
to \link[fmesher:fm_int]{fmesher::fm_int}\verb{(domain, samplers)}. If explicitly given,
overrides \code{domain} and \code{samplers}. \code{fmesher::new_fm_int()} (from fmesher
\verb{0.5.0.9013}) can be used for manually constructed integration schemes.}
}}

\item{formula}{a \code{formula} where the right hand side is a general R
expression defines the predictor used in the model.}

\item{E}{Exposure/effort parameter for family = 'poisson' passed on to
\code{INLA::inla}. Special case if family is 'cp': rescale all integration
weights by a scalar \code{E}. For sampler specific reweighting/effort, use a
\code{weight} column in the \code{samplers} object instead, see \code{\link[fmesher:fm_int]{fmesher::fm_int()}}.
Default taken from \code{options$E}, normally \code{1}.}

\item{weights}{Fixed (optional) weights parameters of the likelihood, so the
log-likelihood\verb{[i]} is changed into \code{weights[i] * log_likelihood[i]}.
Default value is \code{1}. WARNING: The normalizing constant for the likelihood
is NOT recomputed, so ALL marginals (and the marginal likelihood) must be
interpreted with great care.

For \code{family = "cp"}, the weights are applied as \code{sum(weights * eta)} in
the point location contribution part of the log-likelihood, where \code{eta} is
the linear predictor, and do not affect the integration part of the
likelihood. This can be used to implement approximative methods for point
location uncertainty.}

\item{\dots}{Further arguments passed on to \code{\link[=bru_obs]{bru_obs()}}.}

\item{options}{A \link{bru_options} options object or a list of options passed
on to \code{\link[=bru_options]{bru_options()}}}

\item{.envir}{The evaluation environment to use for special arguments (\code{E},
\code{Ntrials}, \code{weights}, and \code{scale}) if not found in \code{response_data} or
\code{data}. Defaults to the calling environment.}
}
\value{
An \code{\link[=bru]{bru()}} object
}
\description{
This function performs inference on a LGCP observed via points residing
possibly multiple dimensions. These dimensions are defined via the left hand
side of the formula provided via the model parameter. The left hand side
determines the intensity function that is assumed to drive the LGCP. This may
include effects that lead to a thinning (filtering) of the point process. By
default, the log intensity is assumed to be a linear combination of the
effects defined by the formula's RHS.

More sophisticated models, e.g.
non-linear thinning, can be achieved by using the predictor argument. The
latter requires multiple runs of INLA for improving the required
approximation of the predictor. In many applications the LGCP is only
observed through subsets of the dimensions the process is living in. For
example, spatial point realizations may only be known in sub-areas of the
modelled space. These observed subsets of the LGCP domain are called samplers
and can be provided via the respective parameter. If samplers is NULL it is
assumed that all of the LGCP's dimensions have been observed completely.
}
\details{
The \code{E} and \code{weights} arguments are evaluated in the data context, like for
\code{\link[=bru_obs]{bru_obs()}}.
}
\examples{
\donttest{
if (bru_safe_inla() &&
  require(ggplot2, quietly = TRUE) &&
  require(fmesher, quietly = TRUE) &&
  requireNamespace("sn", quietly = TRUE)) {
  # Load the Gorilla data
  data <- gorillas_sf

  # Plot the Gorilla nests, the mesh and the survey boundary
  ggplot() +
    fmesher::geom_fm(data = data$mesh) +
    gg(data$boundary, fill = "blue", alpha = 0.2) +
    gg(data$nests, col = "red", alpha = 0.2)

  # Define SPDE prior
  matern <- INLA::inla.spde2.pcmatern(
    data$mesh,
    prior.sigma = c(0.1, 0.01),
    prior.range = c(0.1, 0.01)
  )

  # Define domain of the LGCP as well as the model components (spatial SPDE
  # effect and Intercept)
  cmp <- geometry ~ field(geometry, model = matern) + Intercept(1)

  # Fit the model (with int.strategy="eb" to make the example take less time)
  fit <- lgcp(cmp, data$nests,
    samplers = data$boundary,
    domain = list(geometry = data$mesh),
    options = list(control.inla = list(int.strategy = "eb"))
  )

  # Predict the spatial intensity surface
  lambda <- predict(
    fit,
    fmesher::fm_pixels(data$mesh, mask = data$boundary),
    ~ exp(field + Intercept)
  )

  # Plot the intensity
  ggplot() +
    gg(lambda, geom = "tile") +
    gg(data$nests, col = "red", alpha = 0.2)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_pipe}
\alias{bm_pipe}
\alias{bru_mapper_pipe}
\title{Mapper for linking several mappers in sequence}
\usage{
bm_pipe(mappers)

bru_mapper_pipe(...)
}
\arguments{
\item{mappers}{A list of \code{bru_mapper} objects}

\item{\dots}{Arguments passed on to \code{\link[=bm_pipe]{bm_pipe()}}}
}
\description{
Create a pipe mapper, where \code{mappers} is a list of mappers,
and the evaluated output of each mapper is handed as the state to the next
mapper.
The \code{input} format for the \code{ibm_eval} and \code{ibm_jacobian} methods is
a list of inputs, one for each mapper.
}
\examples{
m <- bm_pipe(list(
  scale = bm_scale(),
  shift = bm_shift()
))
ibm_eval2(m, input = list(scale = 2, shift = 1:4), state = 1:4)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_env_get}
\alias{bru_env_get}
\title{Get access to the internal environment}
\usage{
bru_env_get()
}
\description{
Get access to the internal environment
}
\details{
The environment is defined in \verb{0_inlabru_envir.R} which is loaded
first.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{bru_model_mapper_methods}
\alias{bru_model_mapper_methods}
\alias{ibm_linear.bru_model}
\alias{ibm_linear.bru_comp_list}
\alias{ibm_simplify.bru_model}
\alias{ibm_simplify.bru_comp}
\alias{ibm_simplify.bru_comp_list}
\alias{ibm_linear.bm_list}
\alias{ibm_simplify.bm_list}
\title{Mapper methods for model objects}
\usage{
\method{ibm_linear}{bru_model}(mapper, input, state = NULL, ...)

\method{ibm_linear}{bru_comp_list}(mapper, input, state = NULL, ...)

\method{ibm_simplify}{bru_model}(mapper, input = NULL, state = NULL, ...)

\method{ibm_simplify}{bru_comp}(mapper, input = NULL, state = NULL, ...)

\method{ibm_simplify}{bru_comp_list}(mapper, input = NULL, state = NULL, ...)

\method{ibm_linear}{bm_list}(mapper, input, state = NULL, ...)

\method{ibm_simplify}{bm_list}(mapper, input = NULL, state = NULL, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{input}{Data input for the mapper.}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{...}{Arguments passed on to other methods}
}
\description{
Methods for the \code{ibm_linear()} and \code{ibm_simplify()} methods for
\link{bru} model objects and related classes.
}
\section{Functions}{
\itemize{
\item \code{ibm_linear(bru_model)}: Returns a list (one element per
observation model) of \link{bm_list} objects, each with one \link{bm_taylor}
entry for each included component.

\item \code{ibm_simplify(bru_model)}: Returns a list (one element per
observation model) of \link{bm_list} objects, each with one \link{bru_mapper}
entry for each included component.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_logsumexp}
\alias{bm_logsumexp}
\alias{bru_mapper_logsumexp}
\title{Mapper for log-sum-exp aggregation}
\usage{
bm_logsumexp(rescale = FALSE, n_block = NULL)

bru_mapper_logsumexp(...)
}
\arguments{
\item{rescale}{logical; For \code{bm_aggregate} and \code{bm_logsumexp},
specifies if the blockwise sums should be normalised by the blockwise weight
sums or not:
\itemize{
\item \code{FALSE}: (default) Straight weighted sum, no rescaling.
\item \code{TRUE}: Divide by the sum of the weight values within each block.
This is useful for integration averages, when the given weights are plain
integration weights. If the weights are \code{NULL} or all ones, this is
the same as dividing by the number of entries in each block.
}}

\item{n_block}{Predetermined number of output blocks. If \code{NULL}, overrides
the maximum block index in the inputs. The priority order is \code{input$n_block},
the mapper definition \code{n_block}, then \code{max(input$block)}.}

\item{\dots}{Arguments passed on to \code{\link[=bm_logsumexp]{bm_logsumexp()}}}
}
\description{
Constructs a mapper
that aggregates elements of \code{exp(state)}, with optional non-negative
weighting, and then takes the \code{log()}, so it can be used e.g.
for  \eqn{v_k=\log[\sum_{i\in I_k} w_i \exp(u_i)]
}{log(blocksum(w*exp(state)))}
and \eqn{v_k=\log[\sum_{i\in I_k} w_i \exp(u_i) / \sum_{i\in I_k} w_i]
}{log(blocksum(w*exp(state)/blocksum(w)))}
calculations.  Relies on the input handling methods for
\code{bm_aggregate}, but also allows the weights to be supplied on a
logarithmic scale as \code{log_weights}. To avoid numerical overflow, it uses the
common method of internally shifting the state blockwise;
\eqn{v_k=s_k+\log[\sum_{i\in I_k} \exp(u_i + \log(w_i)- s_k)]
}{log(blocksum(w*exp(state)))},
where \eqn{s_k=\max_{i\in I_k} u_i + \log(w_i)}{s=blockmax(u+log(w))} is the
shift for block \eqn{k}{k}.
}
\examples{
m <- bm_logsumexp()
ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4), 11:14)
ibm_eval2(m, list(block = c(1, 2, 1, 2), weights = 1:4, n_block = 3), 11:14)

}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{evaluate_effect_single_state}
\alias{evaluate_effect_single_state}
\alias{evaluate_effect_single_state.bru_mapper}
\alias{evaluate_effect_single_state.bm_list}
\alias{evaluate_effect_single_state.bru_comp_list}
\title{Evaluate a component effect}
\usage{
evaluate_effect_single_state(...)

\method{evaluate_effect_single_state}{bru_mapper}(component, input, state, ..., label = NULL)

\method{evaluate_effect_single_state}{bm_list}(components, input, state, ...)

\method{evaluate_effect_single_state}{bru_comp_list}(components, input, state, ...)
}
\arguments{
\item{\dots}{Optional additional parameters, e.g. \code{inla_f}. Normally unused.}

\item{component}{A \link{bru_mapper}, \link{bru_comp}, or
\link{bm_list}.}

\item{input}{Pre-evaluated component input}

\item{state}{Specification of one latent variable state:
\itemize{
\item \code{evaluate_effect_single_state.bru_mapper}:
A vector of the latent component state.
\item \verb{evaluate_effect_single_state.*_list}: list of named state vectors.
}}

\item{label}{Option label used for any warning messages, specifying the
affected component.}
}
\value{
\itemize{
\item \code{evaluate_effect_single_state.bm_list}: A list of
evaluated component effect values
}
}
\description{
Calculate latent component effects given some data and the state of the
component's internal random variables.
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com} and
Finn Lindgren \email{finn.lindgren@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bru.inference.R
\name{bru_timings}
\alias{bru_timings}
\alias{bru_timings.bru}
\title{Extract timing information from fitted \link{bru} object}
\usage{
bru_timings(object, ...)

\method{bru_timings}{bru}(object, ...)
}
\arguments{
\item{object}{A fitted \code{bru} object}

\item{\dots}{unused}
}
\description{
Extracts a data.frame or tibble with information about the \code{Time} (CPU),
\code{System}, and \code{Elapsed} time for each step of a \code{bru()} run.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R, R/mapper_sum.R
\name{ibm_is_linear}
\alias{ibm_is_linear}
\alias{ibm_is_linear.default}
\alias{ibm_is_linear.bm_multi}
\alias{ibm_is_linear.bm_collect}
\alias{ibm_is_linear.bm_sum}
\title{Check if a mapper is linear/affine}
\usage{
ibm_is_linear(mapper, ...)

\method{ibm_is_linear}{default}(mapper, ...)

\method{ibm_is_linear}{bm_multi}(mapper, multi = FALSE, ...)

\method{ibm_is_linear}{bm_collect}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_is_linear}{bm_sum}(mapper, multi = FALSE, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{...}{Arguments passed on to other methods}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}
}
\description{
Implementations must return \code{TRUE} or \code{FALSE}.
If \code{TRUE} (returned by the default method unless the mapper
contains an \code{is_linear} variable), users of the mapper
may assume the mapper is linear/affine.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_is_linear(default)}: Returns logical
\code{is_linear} from the mapper object if it exists, and otherwise \code{TRUE}.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{bru_safe_terra}
\alias{bru_safe_terra}
\title{Check for potential \code{terra} installation issues}
\usage{
bru_safe_terra(quietly = FALSE, minimum_version = "1.7-66")
}
\arguments{
\item{quietly}{logical; if \code{TRUE}, prints diagnostic messages. Default
\code{FALSE}}

\item{minimum_version}{character; the minimum required version.
Default 1.7-66 (should always match the requirement in the package
DESCRIPTION)}
}
\value{
Returns (invisibly) \code{FALSE} if a potential issue is detected, and
give a message if \code{quietly} is \code{FALSE}. Otherwise returns \code{TRUE}
}
\description{
Loads the terra package with \code{requireNamespace("terra", quietly = TRUE)},
and checks whether \code{eval_spatial()} successfully extracts non-NA values.
NA values may appear when there is a GDAL/PROJ installation issue that causes
it to not find the \code{proj.db} database file.

For use in package tests and examples that depend on \code{terra}.
}
\examples{
\dontrun{
if (bru_safe_terra()) {
  # Run terra dependent calculations
}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.mrsea.R
\docType{data}
\name{mrsea}
\alias{mrsea}
\title{Marine renewables strategic environmental assessment}
\format{
A list of objects:
\describe{
\item{\code{points}}{ A \code{sf} object containing the locations of
XXXXX.}
\item{\code{samplers}}{ A \code{sf} object containing the transect lines
that were surveyed.}
\item{\code{mesh}}{ An \code{fm_mesh_2d} object containing a Delaunay triangulation
mesh (a type of discretization of continuous space) covering the survey
region.}
\item{\code{boundary}}{ An \code{sf} object defining the boundary polygon of the
survey region.}
\item{\code{covar}}{ An \code{sf} containing sea depth estimates.}
}
}
\source{
Library \code{MRSea}.
}
\usage{
mrsea
}
\description{
Data imported from package MRSea, see
\url{https://www.creem.st-andrews.ac.uk/software/}
}
\examples{
if (require(ggplot2, quietly = TRUE)) {
  ggplot() +
    fmesher::geom_fm(data = mrsea$mesh) +
    gg(mrsea$samplers) +
    gg(mrsea$points) +
    gg(mrsea$boundary)
}
}
\references{
NONE YET
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmesher.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{fm_int}
\alias{fm_pixels}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{fmesher}{\code{\link[fmesher]{fm_int}}, \code{\link[fmesher]{fm_pixels}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/used.R
\name{new_bru_used}
\alias{new_bru_used}
\alias{new_bru_used.NULL}
\alias{new_bru_used.character}
\alias{new_bru_used.expression}
\alias{new_bru_used.quosure}
\alias{new_bru_used.formula}
\title{Store information about used components}
\usage{
new_bru_used(
  x = NULL,
  ...,
  effect = NULL,
  effect_exclude = NULL,
  latent = NULL,
  labels = NULL
)

\method{new_bru_used}{`NULL`}(
  x = NULL,
  ...,
  effect = NULL,
  effect_exclude = NULL,
  latent = NULL,
  labels = NULL
)

\method{new_bru_used}{character}(
  x,
  ...,
  effect = NULL,
  effect_exclude = NULL,
  latent = NULL,
  labels = NULL
)

\method{new_bru_used}{expression}(
  x,
  ...,
  effect = NULL,
  effect_exclude = NULL,
  latent = NULL,
  labels = NULL
)

\method{new_bru_used}{quosure}(
  x,
  ...,
  effect = NULL,
  effect_exclude = NULL,
  latent = NULL,
  labels = NULL
)

\method{new_bru_used}{formula}(
  x,
  ...,
  effect = NULL,
  effect_exclude = NULL,
  latent = NULL,
  labels = NULL
)
}
\arguments{
\item{x}{\code{NULL}, or an object representing an expression}

\item{\dots}{Parameters passed on to the other methods}

\item{effect}{character; components used as effects. When \code{NULL}, auto-detect
components to include all components in a predictor expression.}

\item{effect_exclude}{character; components to specifically exclude from
effect evaluation. When \code{NULL}, do not specifically exclude any components.}

\item{latent}{character; components used as \verb{_latent} or \verb{_eval()}. When
\code{NULL}, auto-detect components.}

\item{labels}{character; component labels passed on to
\code{\link[=bru_used_update]{bru_used_update()}}}
}
\value{
A \code{bru_used} object (a list with elements \code{effect}
and \code{latent})
}
\description{
Create a \code{bru_used} object from effect name character
vectors.
Create information about which components are used by a model, or
its individual observation models. If a non-NULL \code{labels} argument is
supplied, also calls \code{\link[=bru_used_update]{bru_used_update()}} on the \code{bru_used} object.
}
\details{
The arguments \code{effect}, \code{effect_exclude}, and \code{latent} control what
components and effects are available for use in predictor expressions.
\describe{
\item{\code{effect}}{
Character vector of component labels that are used as effects
by the predictor expression; If \code{NULL} (default), the names
are extracted from the formula.
}
\item{\code{exclude}}{
Character vector of component labels to be excluded from the effect list
even if they have been auto-detected as being necessary.
Default is \code{NULL}; do not remove any components from the inclusion list.
}
\item{\code{include_latent}}{Character vector.
Specifies which latent state variables need to be directly available to the
predictor expression, with a \verb{_latent} suffix. This also makes evaluator
functions with suffix \verb{_eval} available, taking parameters \code{main}, \code{group},
and \code{replicate}, taking values for where to evaluate the component effect
that are different than those defined in the component definition itself
(see \code{\link[=bru_comp_eval]{bru_comp_eval()}}). If \code{NULL}, the use of \verb{_latent} and \verb{_eval}
in the predictor expression is detected automatically.
}
}
}
\section{Methods (by class)}{
\itemize{
\item \code{new_bru_used(`NULL`)}: Create a \code{bru_used} object from effect name
character vectors.

\item \code{new_bru_used(character)}: Create a \code{bru_used} object from a \code{character}
representation of an expression.

\item \code{new_bru_used(expression)}: Create a \code{bru_used} object from an expression
object.

\item \code{new_bru_used(quosure)}: Create a \code{bru_used} object from an \code{rlang} \code{expr}
object.

\item \code{new_bru_used(formula)}: Create a \code{bru_used} object from a formula (only the
right-hand side is used).

}}
\examples{
(used <- new_bru_used(~.))
bru_used(used, labels = c("a", "c"))
(used <- new_bru_used(~ a + b + c_latent + d_latent))
bru_used(used, labels = c("a", "c"))
(used <- new_bru_used(expression(a + b + c_latent + d_latent)))
bru_used(used, labels = c("a", "c"))

}
\seealso{
Other bru_used: 
\code{\link{bru_used}()},
\code{\link{bru_used_update}()},
\code{\link{bru_used_vars}()}
}
\concept{bru_used}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_log_bookmark}
\alias{bru_log_bookmark}
\alias{bru_log_bookmarks}
\title{Methods for \code{bru_log} bookmarks}
\usage{
bru_log_bookmark(bookmark = "", offset = NULL, x = NULL)

bru_log_bookmarks(x = NULL)
}
\arguments{
\item{bookmark}{character; The label for a bookmark with a stored offset.}

\item{offset}{integer; a position offset in the log, with \code{0L} pointing at
the start of the log. If negative, denotes the point \code{abs(offset)} elements
from tail of the log. When \code{bookmark} is non-NULL, the \code{offset} applies a
shift (forwards or backwards) to the bookmark list.}

\item{x}{A \code{bru_log} object. If \code{NULL}, the global \code{inlabru} log is used.}
}
\value{
\code{bru_log_bookmark()}: Returns the modified \code{bru_log} object if \code{x}
is non-NULL.

\code{bru_log_bookmarks()}: Returns the bookmark vector associated with
\code{x}
}
\description{
Methods for \code{bru_log} bookmarks.
}
\section{Functions}{
\itemize{
\item \code{bru_log_bookmark()}: Set a log bookmark. If \code{offset} is \code{NULL} (the default),
the bookmark will point to the current end of the log.

\item \code{bru_log_bookmarks()}: Return a integer vector with named elements
being bookmarks into the global \code{inlabru} log with associated log position
offsets.

}}
\seealso{
Other inlabru log methods: 
\code{\link{bru_log}()},
\code{\link{bru_log_message}()},
\code{\link{bru_log_new}()},
\code{\link{bru_log_offset}()},
\code{\link{bru_log_reset}()}
}
\concept{inlabru log methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_index}
\alias{bm_index}
\alias{bru_mapper_index}
\title{Mapper for indexed variables}
\usage{
bm_index(n = 1L)

bru_mapper_index(...)
}
\arguments{
\item{n}{Size of a model for \code{bm_index}}

\item{\dots}{Arguments passed on to \code{\link[=bm_index]{bm_index()}}}
}
\description{
Create a an indexing mapper
}
\examples{
m <- bm_index(4)
ibm_eval(m, -2:6, 1:4)
}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bm_taylor}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{ibm_inla_subset}
\alias{ibm_inla_subset}
\alias{ibm_inla_subset.default}
\title{Find index subset of INLA visible states}
\usage{
ibm_inla_subset(mapper, ...)

\method{ibm_inla_subset}{default}(mapper, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{...}{Arguments passed on to other methods}
}
\description{
Implementations must return a logical vector of \code{TRUE/FALSE} for
the subset such that, given the full A matrix and values output,
\code{A[, subset, drop = FALSE]} and \code{values[subset]}
(or \code{values[subset, , drop = FALSE]} for data.frame values) are equal
to the \code{inla_f = TRUE} version of A and values. The default method uses
the \code{ibm_values} output to construct the subset indexing.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_inla_subset(default)}: Uses the [ibm_values\verb{()] output to construct the inla subset indexing as the difference between }inla_f=FALSE\code{and}inla_f=TRUE\verb{. Extra arguments such as }multi\verb{are passed on to [ibm_values()]. This means it supports both regular vector values and}multi=1` data.frame values.

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R,
%   R/mapper_repeat.R, R/mapper_sum.R
\name{ibm_values}
\alias{ibm_values}
\alias{ibm_values.default}
\alias{ibm_values.bm_fmesher}
\alias{ibm_values.bm_fm_mesh_1d}
\alias{ibm_values.bm_taylor}
\alias{ibm_values.bm_linear}
\alias{ibm_values.bm_matrix}
\alias{ibm_values.bm_factor}
\alias{ibm_values.bm_const}
\alias{ibm_values.bm_shift}
\alias{ibm_values.bm_scale}
\alias{ibm_values.bm_aggregate}
\alias{ibm_values.bm_marginal}
\alias{ibm_values.bm_pipe}
\alias{ibm_values.bm_multi}
\alias{ibm_values.bm_reparam}
\alias{ibm_values.bm_collect}
\alias{ibm_values.bm_repeat}
\alias{ibm_values.bm_sum}
\title{Value vector for a mapping}
\usage{
ibm_values(mapper, inla_f = FALSE, ...)

\method{ibm_values}{default}(mapper, inla_f = FALSE, ...)

\method{ibm_values}{bm_fmesher}(mapper, ...)

\method{ibm_values}{bm_fm_mesh_1d}(mapper, ...)

\method{ibm_values}{bm_taylor}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_values}{bm_linear}(mapper, ...)

\method{ibm_values}{bm_matrix}(mapper, ...)

\method{ibm_values}{bm_factor}(mapper, ...)

\method{ibm_values}{bm_const}(mapper, ...)

\method{ibm_values}{bm_shift}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_values}{bm_scale}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_values}{bm_aggregate}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_values}{bm_marginal}(mapper, ..., state = NULL, n_state = NULL)

\method{ibm_values}{bm_pipe}(mapper, ...)

\method{ibm_values}{bm_multi}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_values}{bm_reparam}(mapper, ...)

\method{ibm_values}{bm_collect}(mapper, inla_f = FALSE, multi = FALSE, ...)

\method{ibm_values}{bm_repeat}(mapper, ...)

\method{ibm_values}{bm_sum}(mapper, inla_f = FALSE, multi = FALSE, ...)
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{inla_f}{logical; when \code{TRUE} for \code{ibm_n()} and \code{ibm_values()}, the
result must be compatible with the \code{INLA::f(...)} and corresponding
\code{INLA::inla.stack(...)} constructions.  For \verb{ibm_\{eval,jacobian,linear\}},
the \code{input} interpretation may be different.
Implementations do not normally need to do anything different, except
for mappers of the type needed for hidden multicomponent models such
as "bym2", which can be handled by \code{bm_collect}.}

\item{...}{Arguments passed on to other methods}

\item{multi}{logical;
If \code{TRUE} (or positive), recurse one level into sub-mappers}

\item{state}{A vector of latent state values for the mapping,
of length \code{ibm_n(mapper, inla_f = FALSE)}}

\item{n_state}{integer giving the length of the state vector for mappers
that have state dependent output size.}
}
\description{
When \code{inla_f=TRUE}, implementations must return a vector that
would be interpretable by an \code{INLA::f(..., values = ...)} specification.
The exception is the method for \code{bm_multi}, that returns a
multi-column data frame if \code{multi=TRUE}.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_values(default)}: Returns a non-null element
'values' from the mapper object, and \code{seq_len(ibm_n(mapper))} if
it doesn't exist.

\item \code{ibm_values(bm_fmesher)}: Returns an index vector for the mesh basis functions.

\item \code{ibm_values(bm_fm_mesh_1d)}: Returns an index vector into the basis functions for
an indexed mapper. Otherwise, the \code{mid} values if present in the mesh being
mapped, and otherwise returns the \code{loc} values of the mesh.

\item \code{ibm_values(bm_linear)}: Returns \code{1.0}

\item \code{ibm_values(bm_matrix)}: For integer labels, the vector of labels.
For character labels, the labels as a factor variable.

\item \code{ibm_values(bm_factor)}: Returns the factor levels (minus the first level for
\code{factor_mapping} "contrast"), or an integer vector (if \code{indexed = TRUE} in
\code{\link[=bm_factor]{bm_factor()}}).

}}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_names}()},
\code{\link{ibm_simplify}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{format.bru_mapper}
\alias{format.bru_mapper}
\alias{format.bm_list}
\alias{summary.bru_mapper}
\alias{format.bm_multi}
\alias{format.bm_pipe}
\alias{format.bm_collect}
\alias{format.bm_sum}
\alias{format.bm_repeat}
\alias{format.bm_reparam}
\alias{print.summary_bru_mapper}
\alias{print.bru_mapper}
\alias{print.bm_list}
\title{mapper object summaries}
\usage{
\method{format}{bru_mapper}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_list}(
  x,
  ...,
  prefix = "",
  initial = prefix,
  depth = 1,
  collapse = ", ",
  labels = TRUE
)

\method{summary}{bru_mapper}(object, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_multi}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_pipe}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_collect}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_sum}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_repeat}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{format}{bm_reparam}(x, ..., prefix = "", initial = prefix, depth = 1)

\method{print}{summary_bru_mapper}(x, ..., sep = "\\n")

\method{print}{bru_mapper}(x, ..., sep = "\\n", prefix = "", initial = prefix, depth = 1)

\method{print}{bm_list}(
  x,
  ...,
  sep = "\\n",
  prefix = "",
  initial = prefix,
  depth = 1,
  labels = TRUE,
  collapse = ", "
)
}
\arguments{
\item{x}{Object to format/print}

\item{\dots}{Unused arguments}

\item{prefix}{character prefix for each line. Default \code{""}.}

\item{initial}{character prefix for the first line. Default \code{initial=prefix}.}

\item{depth}{The recursion depth for multi/collection/pipe mappers. Default
1, to only show the collection, and not the contents of the sub-mappers.}

\item{collapse}{character or NULL, as in \code{\link[base:paste]{base::paste()}}.}

\item{labels}{logical; if TRUE, include mapper names or numerical indices.
Default \code{TRUE}}

\item{object}{Object to summarise}

\item{sep}{character; separator for printing the summary.}
}
\description{
mapper object summaries
}
\examples{
mapper <-
  bm_pipe(
    list(
      bm_multi(list(
        A = bm_index(2),
        B = bm_index(3)
      )),
      bm_index(2)
    )
  )
summary(mapper, depth = 2)
mapper <-
  bm_repeat(
    bm_multi(
      list(
        A = bm_index(2),
        B = bm_index(3)
      )
    ),
    3
  )
summary(mapper)
summary(mapper, depth = 0)
mapper <-
  bm_reparam(
    bm_multi(
      list(
        A = bm_index(2),
        B = bm_index(3)
      )
    ),
    matrix(1:36, nrow = 6)
  )
summary(mapper)
summary(mapper, depth = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment.R
\name{bru_options}
\alias{bru_options}
\alias{as.bru_options}
\alias{bru_options_default}
\alias{bru_options_check}
\alias{bru_options_get}
\alias{bru_options_set}
\alias{bru_options_reset}
\alias{bru_options_set_local}
\title{Create or update an options objects}
\usage{
bru_options(...)

as.bru_options(x = NULL)

bru_options_default()

bru_options_check(options, ignore_null = TRUE)

bru_options_get(name = NULL, include_default = TRUE)

bru_options_set(..., .reset = FALSE)

bru_options_reset()

bru_options_set_local(..., .reset = FALSE, .envir = parent.frame())
}
\arguments{
\item{\dots}{A collection of named options, optionally including one or more
\code{\link{bru_options}} objects. Options specified later override the previous
options.}

\item{x}{An object to be converted to an \code{bru_options} object.}

\item{options}{An \code{bru_options} object to be checked}

\item{ignore_null}{Ignore missing or NULL options.}

\item{name}{Either \code{NULL}, or single option name string, or character vector
or list with option names,
Default: NULL}

\item{include_default}{logical; If \code{TRUE}, the default options are included
together with the global override options. Default: \code{TRUE}}

\item{.reset}{For \code{bru_options_set}, logical indicating if the global
override options list should be emptied before setting the new option(s).}

\item{.envir}{The environment in which to set the options.
Default: \code{parent.frame()}}
}
\value{
\code{bru_options()} returns a \code{bru_options} object.

For \code{as.bru_options()}, \code{NULL} or no input returns an empty
\code{bru_options} object, a \code{list} is converted via \code{bru_options(...)},
and \code{bru_options} input is passed through. Other types of input generates
an error.

\code{bru_options_default()} returns an \code{bru_options} object containing
default options.

\code{bru_options_check()} returns a \code{logical}; \code{TRUE} if the object
contains valid options for use by other functions

\code{bru_options_get} returns either an \code{\link{bru_options}} object, for
\code{name == NULL}, the contents of single option, if \code{name} is a options name
string, or a named list of option contents, if \code{name} is a list of option
name strings.

\code{bru_options_set()} returns a copy of the global override options,
invisibly (as \code{bru_options_get(include_default = FALSE)}).
}
\description{
Create a new options object, or merge information from several
objects.

The \verb{_get}, \verb{_set}, and \verb{_reset} functions operate on a global
package options override object. In many cases, setting options in
specific calls to \code{\link[=bru]{bru()}} is recommended instead.
}
\section{Functions}{
\itemize{
\item \code{as.bru_options()}: Coerces inputs to a \code{bru_options} object.

\item \code{bru_options_default()}: Returns the default options.

\item \code{bru_options_check()}: Checks for valid contents of a \code{bru_options}
object, and produces warnings for invalid options.

\item \code{bru_options_get()}: Used to access global package options.

\item \code{bru_options_set()}: Used to set global package options.

\item \code{bru_options_reset()}: Clears the global option overrides.

\item \code{bru_options_set_local()}: Sets local option overrides, that are
automatically reset using \code{withr::defer()}.

}}
\section{Valid options}{

For \code{bru_options} and \code{bru_options_set}, recognised options are:
\describe{
\item{bru_verbose}{
numeric or logical; if \code{TRUE}, log messages with \code{verbosity} \eqn{\le}
1 are printed by \code{\link[=bru_log_message]{bru_log_message()}}. If numeric, messages with
\code{verbosity} \eqn{\le} \code{bru_verbose} are printed.
For line search details, set \code{bru_verbose=2} or \code{3}.
Default: 0, to not print any messages.}
\item{bru_verbose_store}{
numeric or logical, as for \code{bru_verbose}, but controls what messages
are stored in the global log object.
Default: \code{Inf}, to store all messages.}
\item{bru_max_iter}{maximum number of inla iterations, default 10.
Also see \code{bru_method$rel_tol} and related options below.}
\item{bru_run}{
If \code{TRUE}, run inference. Otherwise only return configuration needed to
run inference.}
\item{bru_initial}{
An \code{inla} object returned from previous calls of \code{INLA::inla}, \code{\link[=bru]{bru()}}
or \code{\link[=lgcp]{lgcp()}}, or a list of named vectors of starting values for the
latent variables. This will be used as a starting point for further
improvement of the approximate posterior.}
\item{bru_int_args}{
List of arguments passed all the way to the integration method
\code{\link[fmesher:fm_int]{fmesher::fm_int()}} for 'cp' family models;
\describe{
\item{method}{
"stable" or "direct". For "stable" (default) integration points are
aggregated to mesh vertices.}
\item{nsub1}{Number of integration points per knot interval in 1D.
Default 30.}
\item{nsub2}{Number of integration points along a triangle edge for 2D.
Default 9.}
\item{nsub}{
Deprecated parameter that overrides \code{nsub1} and \code{nsub2} if set.
Default \verb{<not set>}.}
}
}
\item{bru_method}{
List of arguments controlling the iterative inlabru method:
\describe{
\item{taylor}{'pandemic'
(default, from version 2.1.15).}
\item{search}{Either 'all' (default), to use all available line search
methods, or one or more of
\describe{
\item{'finite'}{(reduce step size until predictor is finite)}
\item{'contract'}{(decrease step size until trust hypersphere reached)}
\item{'expand'}{(increase step size until no improvement)}
\item{'optimise'}{(fast approximate error norm minimisation)}
}
To disable line search, set to an empty vector. Line search is not
available for \code{taylor="legacy"}.}
\item{factor}{
Numeric, \eqn{> 1} determining the line search step scaling multiplier.
Default \eqn{(1 + \sqrt{5})/2}{(1+sqrt(5))/2}.}
\item{rel_tol}{
Stop the iterations when the largest change in linearisation point (the
conditional latent state mode) in relation to the estimated posterior
standard deviation is less than \code{rel_tol}. Default 0.1 (ten percent).}
\item{max_step}{
The largest allowed line search step factor. Factor 1 is the full INLA
step. Default is 2.}
\item{line_opt_method}{
Which method to use for the line search optimisation step. Default
"onestep", using a quadratic approximation based on the value and
gradient at zero, and the value at the current best step length guess.
The method "full" does line optimisation on the full nonlinear
predictor; this is slow and intended for debugging purposes only.}
}
}
\item{bru_compress_cp}{logical; when \code{TRUE}, compress the
\eqn{\sum_{i=1}^n \eta_i}{sum_i=1^n eta_i}
part of the Poisson process likelihood (\code{family = "cp"}) into
either a single term,
with \eqn{y=n}{y=n}, and predictor \code{mean(eta)}, or a blockwise version of
this. Default: \code{FALSE} (was \code{TRUE} prior to version \verb{2.13.0.9031}.)}
\item{bru_debug}{
logical; when \code{TRUE}, activate temporary debug features for package
development. Default: \code{FALSE}}
\item{bru_compat_pre_2_14_enable}{
logical; when \code{TRUE}, enable compatibility features for inlabru versions
prior to 2.14. Set to \code{FALSE} to test external package compatibility
updates. Default: \code{TRUE} before version 2.14, and will be set to \code{FALSE}
by default in a later version.}
\item{\code{inla()} options}{
All options not starting with \code{bru_} are passed on to \code{inla()}, sometimes
after altering according to the needs of the inlabru method.

Warning:
Due to how inlabru currently constructs the \code{inla()} call, the \code{mean},
\code{prec}, \code{mean.intercept}, and \code{prec.intercept} settings in
\code{control.fixed} will have no effect. Until a more elegant alternative
has been implemented, use explicit \code{mean.linear} and \code{prec.linear}
specifications in each \code{model="linear"} component instead.

The following \code{inla()} options have inlabru specific defaults:
\describe{
\item{\code{E}}{Default \code{1}.}
\item{\code{Ntrials}}{Default \code{1L}.}
\item{\code{control.compute}}{
Default \code{list(config = TRUE, control.gcpo = list())}.}
\item{\code{control.inla}}{Default \code{list(int.strategy = "auto")}.}
\item{\code{control.fixed}}{
Default \code{list(expand.factor.strategy = "inla")}.}
}
}
}
}

\examples{
\dontrun{
if (interactive()) {
  # Combine global and user options:
  options1 <- bru_options(bru_options_get(), bru_verbose = TRUE)
  # Create a proto-options object in two equivalent ways:
  options2 <- as.bru_options(bru_verbose = TRUE)
  options2 <- as.bru_options(list(bru_verbose = TRUE))
  # Combine options objects:
  options3 <- bru_options(options1, options2)
}
}
\dontrun{
if (interactive()) {
  bru_options_check(bru_options(bru_max_iter = "text"))
}
}
bru_options_get("bru_verbose")
\dontrun{
if (interactive()) {
  bru_options_set(
    bru_verbose = TRUE,
    verbose = TRUE
  )
}
}
my_fun <- function(val) {
  bru_options_set_local(bru_verbose = val)
  bru_options_get("bru_verbose")
}
# Inside the function, the bru_verbose option is changed.
# Outside the function, the bru_verbose option is unchanged.
print(my_fun(TRUE))
print(bru_options_get("bru_verbose"))
print(my_fun(FALSE))
print(bru_options_get("bru_verbose"))
}
\seealso{
\code{\link[=bru_options]{bru_options()}}, \code{\link[=bru_options_default]{bru_options_default()}}, \code{\link[=bru_options_get]{bru_options_get()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R, R/mapper_collect.R, R/mapper_sum.R
\name{ibm_names}
\alias{ibm_names}
\alias{ibm_names<-}
\alias{ibm_names.default}
\alias{ibm_names.bm_multi}
\alias{ibm_names<-.bm_multi}
\alias{ibm_names<-.bru_mapper_multi}
\alias{ibm_names.bm_collect}
\alias{ibm_names<-.bm_collect}
\alias{ibm_names<-.bru_mapper_collect}
\alias{ibm_names.bm_sum}
\alias{ibm_names<-.bm_sum}
\alias{ibm_names<-.bru_mapper_sum}
\title{Names of submapper}
\usage{
ibm_names(mapper)

ibm_names(mapper) <- value

\method{ibm_names}{default}(mapper, ...)

\method{ibm_names}{bm_multi}(mapper)

\method{ibm_names}{bm_multi}(mapper) <- value

\method{ibm_names}{bru_mapper_multi}(mapper) <- value

\method{ibm_names}{bm_collect}(mapper)

\method{ibm_names}{bm_collect}(mapper) <- value

\method{ibm_names}{bru_mapper_collect}(mapper) <- value

\method{ibm_names}{bm_sum}(mapper)

\method{ibm_names}{bm_sum}(mapper) <- value

\method{ibm_names}{bru_mapper_sum}(mapper) <- value
}
\arguments{
\item{mapper}{A mapper S3 object, inheriting from \code{bru_mapper}.}

\item{value}{a character vector of up to the same length as the number
of mappers in the multi-mapper x}

\item{...}{Arguments passed on to other methods}
}
\description{
Implementations must return a character vector of sub-mapper names, or
\code{NULL}. Intended for providing information about multi-mappers and mapper
collections.
}
\section{Methods (by class)}{
\itemize{
\item \code{ibm_names(default)}: Returns \code{NULL}

\item \code{ibm_names(bm_multi)}: Returns the names from the sub-mappers list

\item \code{ibm_names(bm_collect)}: Returns the names from the
sub-mappers list

\item \code{ibm_names(bm_sum)}: Returns the names from the sub-mappers list

}}
\section{Functions}{
\itemize{
\item \code{ibm_names(mapper) <- value}: Set mapper names.

}}
\examples{
# ibm_names
mapper <- bm_multi(list(A = bm_index(2), B = bm_index(2)))
ibm_names(mapper)
ibm_names(mapper) <- c("new", "names")
ibm_names(mapper)
}
\seealso{
Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}

Other mapper methods: 
\code{\link{bru_mapper_generics}},
\code{\link{ibm_eval}()},
\code{\link{ibm_eval2}()},
\code{\link{ibm_inla_subset}()},
\code{\link{ibm_invalid_output}()},
\code{\link{ibm_is_linear}()},
\code{\link{ibm_is_rowwise}()},
\code{\link{ibm_jacobian}()},
\code{\link{ibm_linear}()},
\code{\link{ibm_n}()},
\code{\link{ibm_n_output}()},
\code{\link{ibm_simplify}()},
\code{\link{ibm_values}()}
}
\concept{mapper methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inlabru-package.R
\docType{package}
\name{inlabru-package}
\alias{inlabru-package}
\alias{inlabru}
\title{inlabru}
\description{
Convenient model fitting using (iterated) INLA.
}
\details{
\code{inlabru} facilitates Bayesian spatial modelling using integrated nested
Laplace approximations. It is heavily based on R-inla
(\url{https://www.r-inla.org}) but adds additional modelling abilities and
simplified syntax for (in particular) spatial models.
Tutorials and more information can be found at
\url{https://inlabru-org.github.io/inlabru/} and \url{http://www.inlabru.org/}.
The iterative method used for non-linear predictors is documented in the
\code{method} vignette.

The main function for inference using inlabru is \code{\link[=bru]{bru()}}.
The general model specification details is documented in \code{\link[=bru_comp]{bru_comp()}}
and \code{\link[=bru_obs]{bru_obs()}}.
Posterior quantities beyond the basic summaries can be calculated with
a \code{predict()} method, documented in \code{\link[=predict.bru]{predict.bru()}}.
For point process inference \code{\link[=lgcp]{lgcp()}} can be used as a shortcut to
\code{bru(..., bru_obs(model="cp", ...))}.

The package comes with multiple real world data sets, namely \link{gorillas},
\link{gorillas_sf}, \link{mexdolphin_sf}. Plotting these data
sets is straight forward using inlabru's extensions
to \code{ggplot2}, e.g. the \code{\link[=gg]{gg()}} function. For educational purposes some
simulated data sets are available as well, e.g. \link{Poisson1_1D},
\link{Poisson2_1D}, \link{Poisson2_1D} and \link{toygroups}.
}
\seealso{
Useful links:
\itemize{
  \item \url{http://www.inlabru.org}
  \item \url{https://inlabru-org.github.io/inlabru/}
  \item \url{https://github.com/inlabru-org/inlabru}
  \item Report bugs at \url{https://github.com/inlabru-org/inlabru/issues}
}

}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com}
and Finn Lindgren \email{finn.lindgren@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/effect.R
\name{bru_comp_list}
\alias{bru_comp_list}
\alias{bru_component_list}
\alias{component_list}
\alias{bru_comp_list.formula}
\alias{bru_comp_list.list}
\alias{bru_comp_list.bru_comp}
\alias{bru_comp_list.bru_comp_list}
\alias{c.bru_comp_list}
\alias{c.bru_comp}
\alias{[.bru_comp_list}
\title{Methods for inlabru component lists}
\usage{
bru_comp_list(object, ..., .envir = parent.frame())

\method{bru_comp_list}{formula}(object, ..., lhoods = NULL, .envir = parent.frame())

\method{bru_comp_list}{list}(
  object,
  ...,
  lhoods = NULL,
  .envir = parent.frame(),
  inputs = NULL
)

\method{bru_comp_list}{bru_comp}(object, ..., .envir = parent.frame())

\method{bru_comp_list}{bru_comp_list}(object, ..., .envir = parent.frame())

\method{c}{bru_comp_list}(...)

\method{c}{bru_comp}(...)

\method{[}{bru_comp_list}(x, i)
}
\arguments{
\item{object}{The object to operate on}

\item{\dots}{Parameters passed on to other methods. Also see Details.}

\item{.envir}{An evaluation environment for non-formula input}

\item{lhoods}{A \link{bru_obs_list} object}

\item{inputs}{A tree-like list of component input evaluations,
from \code{\link[=bru_input.bru_obs_list]{bru_input.bru_obs_list()}}.}

\item{x}{\code{bru_comp_list} object from which to extract a sub-list}

\item{i}{indices specifying elements to extract}
}
\description{
Constructor methods for inlabru component lists. Syntax details are given in
\code{\link[=bru_comp]{bru_comp()}}.
}
\section{Methods (by class)}{
\itemize{
\item \code{bru_comp_list(formula)}: Convert a component formula
into a \code{bru_comp_list} object

\item \code{bru_comp_list(list)}: Combine a list of components, component lists,
and/or component formulas into a single \code{bru_comp_list} object

\item \code{bru_comp_list(bru_comp)}: Place a single \code{bru_comp} object into a
\code{bru_comp_list} object.

\item \code{bru_comp_list(bru_comp_list)}: Make sure a \code{bru_comp_list} object is fully
configured.

}}
\section{Methods (by generic)}{
\itemize{
\item \code{c(bru_comp_list)}: The \code{...} arguments should be \code{bru_comp_list}
objects. The environment from the first argument will be applied to the
resulting \code{bru_comp_list}.

}}
\section{Functions}{
\itemize{
\item \code{c(bru_comp)}: The \code{...} arguments should be \code{component}
objects from \code{\link[=bru_comp]{bru_comp()}}. The environment from the first argument
will be applied to the resulting \code{bru_comp_list}.

}}
\examples{
# As an example, let us create a linear component. Here, the component is
# called "myLinearEffectOfX" while the covariate the component acts on is
# called "x". Note that a list of components is returned because the
# formula may define multiple components

eff <- bru_comp_list(~ myLinearEffectOfX(main = x, model = "linear"))
summary(eff[[1]])
# Equivalent shortcuts:
eff <- bru_comp_list(~ myLinearEffectOfX(x, model = "linear"))
eff <- bru_comp_list(~ myLinearEffectOfX(x))
# Individual component
eff <- bru_comp("myLinearEffectOfX", main = x, model = "linear")
}
\seealso{
Other component constructors: 
\code{\link{bru_comp}()}

Other component constructors: 
\code{\link{bru_comp}()}
}
\author{
Fabian E. Bachl \email{bachlfab@gmail.com} and
Finn Lindgren \email{finn.lindgren@gmail.com}
}
\concept{component constructors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.R
\name{gg.fm_mesh_1d}
\alias{gg.fm_mesh_1d}
\title{Geom for fm_mesh_1d objects}
\usage{
\method{gg}{fm_mesh_1d}(
  data,
  mapping = ggplot2::aes(.data[["x"]], .data[["y"]]),
  y = 0,
  shape = 4,
  ...
)
}
\arguments{
\item{data}{An \link[fmesher:fm_mesh_1d]{fmesher::fm_mesh_1d} object.}

\item{mapping}{aesthetic mappings created by \code{aes}. These are passed on to
\code{geom_point}.}

\item{y}{Single or vector numeric defining the y-coordinates of the mesh
knots to plot.}

\item{shape}{Shape of the knot markers.}

\item{\dots}{parameters passed on to \code{geom_point}.}
}
\value{
An object generated by \code{geom_point}.
}
\description{
This function generates a \code{geom_point} object showing the knots (vertices)
of a 1D mesh.
Requires the \code{ggplot2} package.
}
\examples{
\donttest{
if (require("fmesher", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE)) {
  # Create a 1D mesh

  mesh <- fmesher::fm_mesh_1d(seq(0, 10, by = 0.5))

  # Plot it

  ggplot() +
    gg(mesh)

  # Plot it using a different shape and size for the mesh nodes

  ggplot() +
    gg(mesh, shape = "|", size = 5)
}
}

}
\seealso{
Other geomes: 
\code{\link{gg}()},
\code{\link{gg.RasterLayer}()},
\code{\link{gg.SpatRaster}()},
\code{\link{gg.Spatial}},
\code{\link{gg.data.frame}()},
\code{\link{gg.fm_mesh_2d}()},
\code{\link{gg.matrix}()},
\code{\link{gg.sf}()}
}
\concept{geomes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexagon_tiling.R
\name{cv_partition}
\alias{cv_partition}
\title{Grid partitioning}
\usage{
cv_partition(
  samplers,
  resolution = NULL,
  nrows = NULL,
  ncols = NULL,
  chess = TRUE,
  ...
)
}
\arguments{
\item{samplers}{A sf object containing region for which
partitions to be created}

\item{resolution}{resolution of the grids that are required}

\item{nrows}{number of rows of grids that are required}

\item{ncols}{number of columns of grids that are required}

\item{chess}{chessboard partitioning}

\item{\dots}{Unused.}
}
\value{
a partitioned sf object as required or a list of partitioned sf
objects if chess is TRUE
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Both the interface and function name
may change in future versions.

Partitions the region based on the given criteria for calculating residuals
in each partition. Parts of this function are taken from concepts in
https://rpubs.com/huanfaChen/grid_from_polygon
}
\author{
Man Ho Suen
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mappers.R
\name{bm_taylor}
\alias{bm_taylor}
\alias{bru_mapper_taylor}
\title{Mapper for linear Taylor approximations}
\usage{
bm_taylor(offset = NULL, jacobian = NULL, state0 = NULL, values_mapper = NULL)

bru_mapper_taylor(...)
}
\arguments{
\item{offset}{For \code{bm_taylor}, an offset vector giving the value of
the linearisation at \code{state0}.
May be \code{NULL}, interpreted as an all-zero vector of length determined by
a non-null Jacobian.}

\item{jacobian}{For \code{bm_taylor()}, the Jacobian matrix,
evaluated at \code{state0}, or, a named list of such matrices.
May be \code{NULL} or an empty list, for a constant mapping.}

\item{state0}{For \code{bm_taylor}, the reference \code{state} for the
linearisation, or a list of such states matching the \code{jacobian} list.
\code{NULL} is interpreted as 0.}

\item{values_mapper}{mapper object to be used for \code{ibm_n} and
\code{ibm_values} for \code{inla_f=TRUE} (experimental, currently unused)}

\item{\dots}{Arguments passed on to \code{\link[=bm_taylor]{bm_taylor()}}}
}
\description{
Provides a pre-computed affine mapping,
internally used to represent and evaluate linearisation information.
The \code{state0} information indicates for which state the \code{offset} was
evaluated;
The affine mapper output is defined as
\code{effect(state) = offset + jacobian \%*\% (state - state0)}
}
\examples{
m <- bm_taylor(
  offset = rep(2, 3),
  jacobian = matrix(1:6, 3, 2),
  state0 = c(1, 2)
)
ibm_eval2(m, state = 2:3)
}
\seealso{
\link{bru_mapper}, \link{bru_mapper_generics}

Other mappers: 
\code{\link{bm_aggregate}()},
\code{\link{bm_collect}()},
\code{\link{bm_const}()},
\code{\link{bm_factor}()},
\code{\link{bm_fm_mesh_1d}},
\code{\link{bm_fmesher}()},
\code{\link{bm_harmonics}()},
\code{\link{bm_index}()},
\code{\link{bm_linear}()},
\code{\link{bm_logitaverage}()},
\code{\link{bm_logsumexp}()},
\code{\link{bm_marginal}()},
\code{\link{bm_matrix}()},
\code{\link{bm_multi}()},
\code{\link{bm_pipe}()},
\code{\link{bm_reparam}()},
\code{\link{bm_repeat}()},
\code{\link{bm_scale}()},
\code{\link{bm_shift}()},
\code{\link{bm_sum}()},
\code{\link{bru_get_mapper}()},
\code{\link{bru_mapper}()}
}
\concept{mappers}
